/*
 * Copyright (C) 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "CallLinkInfo.h"

#include "DFGOperations.h"
#include "DFGThunks.h"
#include "Repatch.h"
#include <wtf/ListDump.h>
#include <wtf/NeverDestroyed.h>

#if ENABLE(JIT_VERBOSE_OSR)
static const bool verboseUnlinking = true;
#else
static const bool verboseUnlinking = false;
#endif

#if ENABLE(JIT)
namespace JSC {

CallLinkInfo::CallLinkInfo()
: m_registerPreservationMode(static_cast<unsigned>(RegisterPreservationNotRequired))
, m_hasSeenShouldRepatch(false)
, m_hasSeenClosure(false)
, m_clearedByGC(false)
, m_allowStubs(true)
, m_clearedByJettison(false)
, m_callType(None)
, m_calleeGPR(0)
, m_slowPathCount(0)
{
}

CallLinkInfo::~CallLinkInfo()
{
    clearStub();

    if (isOnList())
        remove();
}

void CallLinkInfo::clearStub()
{
    if (!stub())
        return;

    m_stub->clearCallNodesFor(this);
    m_stub.clear();
}

void CallLinkInfo::unlink(VM& vm)
{
    if (!isLinked()) {
        // We could be called even if we're not linked anymore because of how polymorphic calls
        // work. Each callsite within the polymorphic call stub may separately ask us to unlink().
        RELEASE_ASSERT(!isOnList());
        return;
    }

    unlinkFor(vm, *this);

    // It will be on a list if the callee has a code block.
    if (isOnList())
        remove();
}

void CallLinkInfo::visitWeak(VM& vm)
{
	class HandleSpecificCallee {
	public:
		HandleSpecificCallee(CallLinkInfo& info) : info(info) { }
		void operator()(JSFunction* callee) {
			if (Heap::isMarked(callee->executable()))
				info.setHasSeenClosure();
			else
				info.setClearedByGC();
		}
	private:
		CallLinkInfo& info;
    };
	HandleSpecificCallee handleSpecificCallee(*this);

    if (isLinked()) {
        if (stub()) {
            if (!stub()->visitWeak(vm)) {
#if 0
                if (verboseUnlinking) {
                    dataLog(
                        "Clearing closure call from ", *repatchBuffer.codeBlock(), " to ",
						listDump(stub->variants()), ", stub routine ", RawPointer(stub.get()),
						".\n");
                }
#endif
                unlink(vm);
                m_clearedByGC = true;
            }
        } else if (!Heap::isMarked(callee())) {
#if 0
            if (verboseUnlinking) {
                dataLog(
                    "Clearing call from ", *repatchBuffer.codeBlock(), " to ",
                    RawPointer(callee()), " (",
                    //callee.get()->executable()->hashFor(specializationKind()),
                    ").\n");
            }
#endif
            handleSpecificCallee(callee());
            unlink(vm);
        }
    }
    if (haveLastSeenCallee() && !Heap::isMarked(lastSeenCallee())) {
        handleSpecificCallee(lastSeenCallee());
        clearLastSeenCallee();
    }
}

} // namespace JSC
#endif // ENABLE(JIT)

