/*
 * Copyright (C) 2014 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "StructureSet.h"
#include "DFGAbstractValue.h"
#include "DFGStructureAbstractValue.h"

#include <wtf/CommaPrinter.h>

namespace JSC {
#if ENABLE(DFG_JIT)

void StructureSet::filter(const DFG::StructureAbstractValue& other)
{
	class Functor
	{
	public:
		Functor(const DFG::StructureAbstractValue& other) : other(other) { }
		bool operator()(Structure* structure) const { return other.contains(structure); }
	private:
		const DFG::StructureAbstractValue& other;
	};
	Functor functor(other);
    genericFilter(functor);
}

void StructureSet::filter(SpeculatedType type)
{
	class Functor
	{
	public:
		Functor(SpeculatedType type) : type(type) { }
		bool operator()(Structure* structure) const {
            return type & speculationFromStructure(structure);
		}
	private:
		const SpeculatedType type;
	};
	Functor functor(type);
    genericFilter(functor);
}

void StructureSet::filterArrayModes(ArrayModes arrayModes)
{
	class Functor
	{
	public:
		Functor(ArrayModes arrayModes) : arrayModes(arrayModes) { }
		bool operator()(Structure* structure) const {
            return arrayModes & arrayModeFromStructure(structure);
		}
	private:
		const ArrayModes arrayModes;
	};
	Functor functor(arrayModes);
    genericFilter(functor);
}

void StructureSet::filter(const DFG::AbstractValue& other)
{
    filter(other.m_structure);
    filter(other.m_type);
    filterArrayModes(other.m_arrayModes);
}

#endif // ENABLE(DFG_JIT)

SpeculatedType StructureSet::speculationFromStructures() const
{
	SpeculatedType result = SpecNone;
	class Functor
	{
	public:
		Functor(SpeculatedType& result) : result(result) { }
		void operator()(Structure* structure) {
			mergeSpeculation(result, speculationFromStructure(structure));
		}
	private:
		SpeculatedType& result;
	};
	Functor functor(result);
	forEach(functor);
	return result;
}

ArrayModes StructureSet::arrayModesFromStructures() const
{
	ArrayModes result = 0;
	class Functor
	{
	public:
		Functor(ArrayModes& result) : result(result) { }
		void operator()(Structure* structure) {
			mergeArrayModes(result, asArrayModes(structure->indexingType()));
		}
	private:
		ArrayModes& result;
	};
	Functor functor(result);
	forEach(functor);
	return result;
}

void StructureSet::dumpInContext(PrintStream& out, DumpContext* context) const
{
#if 0
    CommaPrinter comma;
    out.print("[");
    forEach([&] (Structure* structure) { out.print(comma, inContext(*structure, context)); });
    out.print("]");
#endif
}

void StructureSet::dump(PrintStream& out) const
{
    dumpInContext(out, 0);
}

} // namespace JSC

