/*
 * Copyright (C) 2011, 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "ValueRecovery.h"

#include "CodeBlock.h"
#include "Operations.h"

namespace JSC {

JSValue ValueRecovery::recover(ExecState* exec) const
{
    switch (technique()) {
    case DisplacedInJSStack:
        return exec->r(virtualRegister()).jsValue();
    case Int32DisplacedInJSStack:
        return jsNumber(exec->r(virtualRegister()).unboxedInt32());
    case Int52DisplacedInJSStack:
        return jsNumber(exec->r(virtualRegister()).unboxedInt52());
    case StrictInt52DisplacedInJSStack:
        return jsNumber(exec->r(virtualRegister()).unboxedStrictInt52());
    case DoubleDisplacedInJSStack:
        return jsNumber(purifyNaN(exec->r(virtualRegister()).unboxedDouble()));
    case CellDisplacedInJSStack:
        return exec->r(virtualRegister()).unboxedCell();
    case BooleanDisplacedInJSStack:
#if USE(JSVALUE64)
        return exec->r(virtualRegister()).jsValue();
#else
        return jsBoolean(exec->r(virtualRegister()).unboxedBoolean());
#endif
    case Constant:
        return constant();
    default:
        RELEASE_ASSERT_NOT_REACHED();
        return JSValue();
    }
}

void ValueRecovery::dumpInContext(PrintStream& out, DumpContext* context) const
{
    switch (technique()) {
    case InGPR:
        out.printf("%%r%d", gpr());
        return;
    case UnboxedInt32InGPR:
        out.printf("int32(%%r%d)", gpr());
        return;
    case UnboxedInt52InGPR:
        out.printf("int52(%%r%d)", gpr());
        return;
    case UnboxedStrictInt52InGPR:
        out.printf("strictInt52(%%r%d)", gpr());
        return;
    case UnboxedBooleanInGPR:
        out.printf("bool(%%r%d)", gpr());
        return;
    case UnboxedCellInGPR:
        out.printf("cell(%%r%d)", gpr());
        return;
    case InFPR:
        out.printf("%%fr%d", fpr());
        return;
    case UnboxedDoubleInFPR:
        out.print("double(", fpr(), ")");
        return;
#if USE(JSVALUE32_64)
    case InPair:
        out.printf("pair(%%r%d, %%r%d)", tagGPR(), payloadGPR());
        return;
#endif
    case DisplacedInJSStack:
        out.printf("*%d", virtualRegister());
        return;
    case Int32DisplacedInJSStack:
        out.printf("*int32(%d)", virtualRegister());
        return;
    case Int52DisplacedInJSStack:
        out.printf("*int52(%d)", virtualRegister());
        return;
    case StrictInt52DisplacedInJSStack:
        out.printf("*strictInt52(%d)", virtualRegister());
        return;
    case DoubleDisplacedInJSStack:
        out.printf("*double(%d)", virtualRegister());
        return;
    case CellDisplacedInJSStack:
        out.printf("*cell(%d)", virtualRegister());
        return;
    case BooleanDisplacedInJSStack:
        out.printf("*bool(%d)", virtualRegister());
        return;
#if ENABLE(DFG_JIT)
    case DirectArgumentsThatWereNotCreated:
        out.print("DirectArguments(", nodeID(), ")");
        return;
    case ClonedArgumentsThatWereNotCreated:
        out.print("ClonedArguments(", nodeID(), ")");
        return;
#endif
    case Constant:
        out.print("[", constant(), "]");
        return;
    case DontKnow:
        out.printf("!");
        return;
    default:
        out.printf("?%d", technique());
        return;
    }
}

} // namespace JSC

