/*
 * Copyright (C) 2011 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef VirtualRegister_h
#define VirtualRegister_h

#include "BytecodeConventions.h"
#include "CallFrame.h"
#include <wtf/Platform.h>
#include <wtf/PrintStream.h>

namespace JSC {

inline int localToOperand(int local) { return -1 - local; }
inline bool operandIsLocal(int operand) { return operand < 0; }
inline int operandToLocal(int operand) { return -1 - operand; }

// argument 0 is 'this'.
inline bool operandIsArgument(int operand) { return operand >= 0; }
inline int operandToArgument(int operand) { return operand - CallFrame::thisArgumentOffset(); }
inline int argumentToOperand(int argument) { return argument + CallFrame::thisArgumentOffset(); }

inline int virtualRegisterForLocal(int local)
{
    return localToOperand(local);
}

inline int virtualRegisterForArgument(int argument, int offset = 0)
{
    return argumentToOperand(argument) + offset;
}

// Type for a virtual register number (spill location).
// Using an enum to make this type-checked at compile time, to avert programmer errors.
enum VirtualRegister { InvalidVirtualRegister = 0x3fffffff };
COMPILE_ASSERT(sizeof(VirtualRegister) == sizeof(int), VirtualRegister_is_32bit);

class VirtualRegister_ {
public:
    friend int virtualRegisterForLocal(int);
    friend int virtualRegisterForArgument(int, int);

    VirtualRegister_()
        : m_virtualRegister(s_invalidVirtualRegister)
    { }

    explicit VirtualRegister_(int virtualRegister)
        : m_virtualRegister(virtualRegister)
    { }

    bool isValid() const { return (m_virtualRegister != s_invalidVirtualRegister); }
    bool isLocal() const { return operandIsLocal(m_virtualRegister); }
    bool isArgument() const { return operandIsArgument(m_virtualRegister); }
    bool isHeader() const { return m_virtualRegister >= 0 && m_virtualRegister < JSStack::ThisArgument; }
    bool isConstant() const { return m_virtualRegister >= s_firstConstantRegisterIndex; }
    int toLocal() const { ASSERT(isLocal()); return operandToLocal(m_virtualRegister); }
    int toArgument() const { ASSERT(isArgument()); return operandToArgument(m_virtualRegister); }
    int toConstantIndex() const { ASSERT(isConstant()); return m_virtualRegister - s_firstConstantRegisterIndex; }
    int offset() const { return m_virtualRegister; }
    int offsetInBytes() const { return m_virtualRegister * sizeof(Register); }

    bool operator==(VirtualRegister_ other) const { return m_virtualRegister == other.m_virtualRegister; }
    bool operator!=(VirtualRegister_ other) const { return m_virtualRegister != other.m_virtualRegister; }
    bool operator<(VirtualRegister_ other) const { return m_virtualRegister < other.m_virtualRegister; }
    bool operator>(VirtualRegister_ other) const { return m_virtualRegister > other.m_virtualRegister; }
    bool operator<=(VirtualRegister_ other) const { return m_virtualRegister <= other.m_virtualRegister; }
    bool operator>=(VirtualRegister_ other) const { return m_virtualRegister >= other.m_virtualRegister; }

    VirtualRegister_ operator+(int value) const
    {
        return VirtualRegister_(offset() + value);
    }
    VirtualRegister_ operator-(int value) const
    {
        return VirtualRegister_(offset() - value);
    }
    VirtualRegister_ operator+(VirtualRegister_ value) const
    {
        return VirtualRegister_(offset() + value.offset());
    }
    VirtualRegister_ operator-(VirtualRegister_ value) const
    {
        return VirtualRegister_(offset() - value.offset());
    }
    VirtualRegister_& operator+=(int value)
    {
        return *this = *this + value;
    }
    VirtualRegister_& operator-=(int value)
    {
        return *this = *this - value;
    }

    void dump(PrintStream& out) const;

private:
    static const int s_invalidVirtualRegister = InvalidVirtualRegister;
    static const int s_firstConstantRegisterIndex = FirstConstantRegisterIndex;

    static int localToOperand(int local) { return -1 - local; }
    static int operandToLocal(int operand) { return -1 - operand; }
    static int operandToArgument(int operand) { return operand - CallFrame::thisArgumentOffset(); }
    static int argumentToOperand(int argument) { return argument + CallFrame::thisArgumentOffset(); }

    int m_virtualRegister;
};

} // namespace JSC

namespace WTF {

inline void printInternal(PrintStream& out, JSC::VirtualRegister value)
{
    out.print(static_cast<int>(value));
}

} // namespace WTF

#endif // VirtualRegister_h
