/*
 * Copyright (C) 2008, 2009 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "DebuggerActivation.h"

#include "JSLexicalEnvironment.h"
#include "Operations.h"

namespace JSC {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(DebuggerActivation);

const ClassInfo DebuggerActivation::s_info = { "DebuggerActivation", &Base::s_info, 0, CREATE_METHOD_TABLE(DebuggerActivation) };

DebuggerActivation::DebuggerActivation(JSGlobalData& globalData)
    : JSNonFinalObject(globalData, globalData.debuggerActivationStructure.get())
{
}

void DebuggerActivation::finishCreation(JSGlobalData& globalData, JSObject* activation)
{
    Base::finishCreation(globalData);
    ASSERT(activation);
    ASSERT(activation->isActivationObject());
    m_lexicalEnvironment.set(globalData, this, jsCast<JSLexicalEnvironment*>(activation));
}

void DebuggerActivation::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    DebuggerActivation* thisObject = jsCast<DebuggerActivation*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);

    if (thisObject->m_lexicalEnvironment)
        visitor.append(&thisObject->m_lexicalEnvironment);
}

String DebuggerActivation::className(const JSObject* object)
{
    const DebuggerActivation* thisObject = jsCast<const DebuggerActivation*>(object);
    return thisObject->m_lexicalEnvironment->methodTable()->className(thisObject->m_lexicalEnvironment.get());
}

bool DebuggerActivation::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    DebuggerActivation* thisObject = jsCast<DebuggerActivation*>(object);
    return thisObject->m_lexicalEnvironment->methodTable()->getOwnPropertySlot(thisObject->m_lexicalEnvironment.get(), exec, propertyName, slot);
}

bool DebuggerActivation::put(JSCell* cell, ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    DebuggerActivation* thisObject = jsCast<DebuggerActivation*>(cell);
    return thisObject->m_lexicalEnvironment->methodTable()->put(thisObject->m_lexicalEnvironment.get(), exec, propertyName, value, slot);
}

bool DebuggerActivation::deleteProperty(JSCell* cell, ExecState* exec, PropertyName propertyName)
{
    DebuggerActivation* thisObject = jsCast<DebuggerActivation*>(cell);
    return thisObject->m_lexicalEnvironment->methodTable()->deleteProperty(thisObject->m_lexicalEnvironment.get(), exec, propertyName);
}

void DebuggerActivation::getOwnPropertyNames(JSObject* object, ExecState* exec, PropertyNameArray& propertyNames, EnumerationMode mode)
{
    DebuggerActivation* thisObject = jsCast<DebuggerActivation*>(object);
    thisObject->m_lexicalEnvironment->methodTable()->getPropertyNames(thisObject->m_lexicalEnvironment.get(), exec, propertyNames, mode);
}

bool DebuggerActivation::defineOwnProperty(JSObject* object, ExecState* exec, PropertyName propertyName, const PropertyDescriptor& descriptor, bool shouldThrow)
{
    DebuggerActivation* thisObject = jsCast<DebuggerActivation*>(object);
    return thisObject->m_lexicalEnvironment->methodTable()->defineOwnProperty(thisObject->m_lexicalEnvironment.get(), exec, propertyName, descriptor, shouldThrow);
}

} // namespace JSC
