/*
 * Copyright (C) 2015 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "DFGLiveCatchVariablePreservationPhase.h"

#if ENABLE(DFG_JIT)

#include "DFGBasicBlockInlines.h"
#include "DFGGraph.h"
#include "DFGInsertionSet.h"
#include "DFGPhase.h"
#include "FullBytecodeLiveness.h"
#include "JSCInlines.h"

namespace JSC { namespace DFG {

class FlushLiveCatchVariablesInsertionPhase : public Phase {
public:
    FlushLiveCatchVariablesInsertionPhase(Graph& graph)
        : Phase(graph, "live catch variable preservation phase")
    {
    }

    bool run()
    {
        if (!m_graph.m_hasExceptionHandlers)
            return true;

        //DFG_ASSERT(m_graph, 0, m_graph.m_form == LoadStore);

        m_currentBlockLiveness.resize(m_graph.block(0)->variablesAtTail.numberOfLocals());

        InsertionSet insertionSet(m_graph);
        auto naturalBlockIterable = m_graph.blocksInNaturalOrder();
        for (auto it = naturalBlockIterable.begin(); it != naturalBlockIterable.end(); ++it) {
        	BasicBlock* block = *it;
            handleBlock(block, insertionSet);
            insertionSet.execute(block);
        }

        return true;
    }

    bool willCatchException(CodeOrigin origin)
    {
        unsigned bytecodeIndexToCheck = origin.bytecodeIndex;
        m_currentBlockLiveness.clearAll();

        while (1) {
            InlineCallFrame* inlineCallFrame = origin.inlineCallFrame;
            CodeBlock* codeBlock = m_graph.baselineCodeBlockFor(inlineCallFrame);
            if (HandlerInfo* handler = codeBlock->handlerForBytecodeOffset(bytecodeIndexToCheck)) {
                unsigned catchBytecodeIndex = handler->target;
                class Functor
                {
                public:
                	Functor(FastBitVector& currentBlockLiveness) : m_currentBlockLiveness(currentBlockLiveness) { }
                	void operator()(int operand) {
                        m_currentBlockLiveness.set(VirtualRegister_(operand).toLocal(), true);
                    }
                private:
                    FastBitVector& m_currentBlockLiveness;
                };
                Functor functor(m_currentBlockLiveness);
                m_graph.forAllLocalsLiveInBytecode(CodeOrigin(catchBytecodeIndex, inlineCallFrame), functor);
                return true;
            }

            if (!inlineCallFrame)
                return false;

            bytecodeIndexToCheck = inlineCallFrame->directCaller.bytecodeIndex;
            origin = inlineCallFrame->directCaller;
        }
    }

    void handleBlock(BasicBlock* block, InsertionSet& insertionSet)
    {
        // Because precise jump targets ensures that the start of a "try" block is its
        // own basic block, we will never have two "try" statements in the same DFG
        // basic block. Therefore, checking the first node in the block is sufficient 
        // to checking if we're in a try block.
        if (!willCatchException(block->at(0)->origin.semantic))
            return;

        Operands<VariableAccessData*> currentBlockAccessData(block->variablesAtTail.numberOfArguments(), block->variablesAtTail.numberOfLocals(), 0);
        HashSet<InlineCallFrame*> seenInlineCallFrames;

        {
            for (unsigned i = 0; i < block->size(); i++) {
                Node* node = block->at(i);
                bool isPrimordialSetArgument = node->op() == SetArgument && VirtualRegister_(node->local()).isArgument() && node == m_graph.m_arguments[VirtualRegister_(node->local()).toArgument()];
                InlineCallFrame* inlineCallFrame = node->origin.semantic.inlineCallFrame;
                if (inlineCallFrame)
                    seenInlineCallFrames.add(inlineCallFrame);

                if (node->op() == SetLocal || (node->op() == SetArgument && !isPrimordialSetArgument)) {
                    VirtualRegister_ operand = VirtualRegister_(node->local());

                    int stackOffset = inlineCallFrame ? inlineCallFrame->stackOffset : 0;
                    if ((operand.isLocal() && m_currentBlockLiveness.get(operand.toLocal()))
                        || (operand.offset() == stackOffset + CallFrame::thisArgumentOffset())) {

                        VariableAccessData* flushAccessData = currentBlockAccessData.operand(operand.offset());
                        if (!flushAccessData)
                            flushAccessData = newVariableAccessData(static_cast<VirtualRegister>(operand.offset()));

                        insertionSet.insertNode(i, SpecNone, 
                            Flush, node->origin, OpInfo(flushAccessData));
                    }
                }

                if (node->hasVariableAccessData(m_graph))
                    currentBlockAccessData.operand(node->local()) = node->variableAccessData();
            }
        }

        // Flush everything at the end of the block.
        // FIXME: I think this will only be necessary if we have any successor
        // blocks who aren't inside this "try" statement. If all our successor's
        // are in this try statement, they will have Flushes for any live "catch"
        // variables.
        {
            NodeOrigin origin = block->at(block->size() - 1)->origin;
            class PreserveLivenessAtEndOfBlockFunctor
            {
            public:
            	PreserveLivenessAtEndOfBlockFunctor(FlushLiveCatchVariablesInsertionPhase& phase, BasicBlock* block, InsertionSet& insertionSet, Operands<VariableAccessData*>& currentBlockAccessData, NodeOrigin& origin)
            	: phase(phase)
            	, block(block)
            	, insertionSet(insertionSet)
            	, currentBlockAccessData(currentBlockAccessData)
            	, origin(origin)
            	{
            	}
				void operator()(int operand, bool alwaysFlush) {
					if ((VirtualRegister_(operand).isLocal() && phase.m_currentBlockLiveness.get(VirtualRegister_(operand).toLocal()))
						|| VirtualRegister_(operand).isArgument()
						|| alwaysFlush) {
						VariableAccessData* accessData = currentBlockAccessData.operand(operand);
						if (!accessData)
							accessData = phase.newVariableAccessData(static_cast<VirtualRegister>(operand));

						currentBlockAccessData.operand(operand) = accessData;

						insertionSet.insertNode(block->size(), SpecNone,
							Flush, origin, OpInfo(accessData));
					}
				}
            public:
				FlushLiveCatchVariablesInsertionPhase& phase;
				BasicBlock* block;
				InsertionSet& insertionSet;
				Operands<VariableAccessData*>& currentBlockAccessData;
				NodeOrigin& origin;
            };
            PreserveLivenessAtEndOfBlockFunctor preserveLivenessAtEndOfBlock(*this, block, insertionSet, currentBlockAccessData, origin);
            for (unsigned local = 0; local < block->variablesAtTail.numberOfLocals(); local++)
            	preserveLivenessAtEndOfBlock(virtualRegisterForLocal(local), false);
            for(auto it = seenInlineCallFrames.begin(); it != seenInlineCallFrames.end(); ++it)
            	preserveLivenessAtEndOfBlock(VirtualRegister((*it)->stackOffset + CallFrame::thisArgumentOffset()), true);
            preserveLivenessAtEndOfBlock(VirtualRegister(CallFrame::thisArgumentOffset()), true);
        }
    }

    VariableAccessData* newVariableAccessData(VirtualRegister operand)
    {
        ASSERT(!operand.isConstant());
        
        m_graph.m_variableAccessData.append(VariableAccessData(operand));
        return &m_graph.m_variableAccessData.last();
    }

    FastBitVector m_currentBlockLiveness;
};

bool performLiveCatchVariablePreservationPhase(Graph& graph)
{
    SamplingRegion samplingRegion("DFG Live Catch Variables Preservation Phase");
    return runPhase<FlushLiveCatchVariablesInsertionPhase>(graph);
}

} } // namespace JSC::DFG

#endif // ENABLE(DFG_JIT)
