/*
 *  Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2011 Apple Inc. All rights reserved.
 *  Copyright (C) 2007 Eric Seidel <eric@webkit.org>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "Heap.h"

#include "ArrayBuffer.h"
#include "CodeBlock.h"
#include "ConservativeRoots.h"
#include "CopiedSpace.h"
#include "CopiedSpaceInlineMethods.h"
#include "CopyVisitorInlineMethods.h"
#include "ExecutableAllocator.h"
#include "GCActivityCallback.h"
#include "GCIncomingRefCountedSetInlines.h"
#include "HeapIterationScope.h"
#include "HeapRootVisitor.h"
#include "IncrementalSweeper.h"
#include "Interpreter.h"
#include "JSGlobalData.h"
#include "JSGlobalObject.h"
#include "JSLock.h"
#include "JSONObject.h"
#include "JITStubRoutineSet.h"
#include "Operations.h"
#include "Options.h"
#include "RegExpCache.h"
#include "StructureRareData.h"
#include "Tracing.h"
#include "WeakSetInlines.h"
#include <algorithm>
#include <wtf/RAMSize.h>
#include <wtf/CurrentTime.h>
#include <wtf/MemoryMonitor.h>

using namespace std;
using namespace JSC;

namespace JSC {

extern int jitCodeCapBytes();
extern int jitCodeReservedBytes();

namespace { 

static const size_t smallHeapSize = 1 * MB;

#if ENABLE(GC_LOGGING)
#if COMPILER(CLANG)
#define DEFINE_GC_LOGGING_GLOBAL(type, name, arguments) \
_Pragma("clang diagnostic push") \
_Pragma("clang diagnostic ignored \"-Wglobal-constructors\"") \
_Pragma("clang diagnostic ignored \"-Wexit-time-destructors\"") \
static type name arguments; \
_Pragma("clang diagnostic pop")
#else
#define DEFINE_GC_LOGGING_GLOBAL(type, name, arguments) \
static type name arguments;
#endif // COMPILER(CLANG)

struct GCTimer {
    GCTimer(const char* name)
        : m_time(0)
        , m_min(100000000)
        , m_max(0)
        , m_count(0)
        , m_name(name)
    {
    }
    ~GCTimer()
    {
        dataLog("%s: %.2lfms (avg. %.2lf, min. %.2lf, max. %.2lf)\n", m_name, m_time * 1000, m_time * 1000 / m_count, m_min*1000, m_max*1000);
    }
    double m_time;
    double m_min;
    double m_max;
    size_t m_count;
    const char* m_name;
};

struct GCTimerScope {
    GCTimerScope(GCTimer* timer)
        : m_timer(timer)
        , m_start(WTF::monotonicallyIncreasingTime())
    {
    }
    ~GCTimerScope()
    {
        double delta = WTF::monotonicallyIncreasingTime() - m_start;
        if (delta < m_timer->m_min)
            m_timer->m_min = delta;
        if (delta > m_timer->m_max)
            m_timer->m_max = delta;
        m_timer->m_count++;
        m_timer->m_time += delta;
    }
    GCTimer* m_timer;
    double m_start;
};

struct GCCounter {
    GCCounter(const char* name)
        : m_name(name)
        , m_count(0)
        , m_total(0)
        , m_min(10000000)
        , m_max(0)
    {
    }
    
    void count(size_t amount)
    {
        m_count++;
        m_total += amount;
        if (amount < m_min)
            m_min = amount;
        if (amount > m_max)
            m_max = amount;
    }
    ~GCCounter()
    {
        dataLog("%s: %zu values (avg. %zu, min. %zu, max. %zu)\n", m_name, m_total, m_total / m_count, m_min, m_max);
    }
    const char* m_name;
    size_t m_count;
    size_t m_total;
    size_t m_min;
    size_t m_max;
};

#define GCPHASE(name) DEFINE_GC_LOGGING_GLOBAL(GCTimer, name##Timer, (#name)); GCTimerScope name##TimerScope(&name##Timer)
#define COND_GCPHASE(cond, name1, name2) DEFINE_GC_LOGGING_GLOBAL(GCTimer, name1##Timer, (#name1)); DEFINE_GC_LOGGING_GLOBAL(GCTimer, name2##Timer, (#name2)); GCTimerScope name1##CondTimerScope(cond ? &name1##Timer : &name2##Timer)
#define GCCOUNTER(name, value) do { DEFINE_GC_LOGGING_GLOBAL(GCCounter, name##Counter, (#name)); name##Counter.count(value); } while (false)
    
#else

#define GCPHASE(name) do { } while (false)
#define COND_GCPHASE(cond, name1, name2) do { } while (false)
#define GCCOUNTER(name, value) do { } while (false)
#endif

static inline size_t minHeapSize(HeapType heapType)
{
    return (heapType == LargeHeap) ? Options::largeHeapSize() : smallHeapSize;
}

static inline size_t proportionalHeapSize(size_t heapSize)
{
	WTF::MmapMonitor& monitor = WTF::MmapMonitor::instance();

	int availableSize = (monitor.capSize() > monitor.allocatedSize()) ? (monitor.capSize() - monitor.allocatedSize()) : 0;

	if(!monitor.capSize() || availableSize >= heapSize)
		return heapSize + heapSize;

	return heapSize + std::max(availableSize, Options::minHeapIncrement());
}

static inline bool isValidSharedInstanceThreadState(JSGlobalData* globalData)
{
    return globalData->apiLock().currentThreadIsHoldingLock();
}

static inline bool isValidThreadState(JSGlobalData* globalData)
{
    if (globalData->identifierTable != wtfThreadData().currentIdentifierTable())
        return false;

    if (globalData->isSharedInstance() && !isValidSharedInstanceThreadState(globalData))
        return false;

    return true;
}

struct MarkObject : public MarkedBlock::VoidFunctor {
    void operator()(JSCell* cell)
    {
        if (cell->isZapped())
            return;
        Heap::heap(cell)->setMarked(cell);
    }
};

struct Count : public MarkedBlock::CountFunctor {
    void operator()(JSCell*) { count(1); }
};

struct CountIfGlobalObject : MarkedBlock::CountFunctor {
    void operator()(JSCell* cell) {
        if (!cell->isObject())
            return;
        if (!asObject(cell)->isGlobalObject())
            return;
        count(1);
    }
};

class RecordType {
public:
    typedef PassOwnPtr<TypeCountSet> ReturnType;

    RecordType();
    void operator()(JSCell*);
    ReturnType returnValue();

private:
    const char* typeName(JSCell*);
    OwnPtr<TypeCountSet> m_typeCountSet;
};

inline RecordType::RecordType()
    : m_typeCountSet(adoptPtr(new TypeCountSet))
{
}

inline const char* RecordType::typeName(JSCell* cell)
{
    const ClassInfo* info = cell->classInfo();
    if (!info || !info->className)
        return "[unknown]";
    return info->className;
}

inline void RecordType::operator()(JSCell* cell)
{
    m_typeCountSet->add(typeName(cell));
}

inline PassOwnPtr<TypeCountSet> RecordType::returnValue()
{
    return m_typeCountSet.release();
}

class StorageStatistics : public MarkedBlock::VoidFunctor {
public:
    StorageStatistics();

    void operator()(JSCell*);

    size_t objectWithOutOfLineStorageCount();
    size_t objectCount();

    size_t storageSize();
    size_t storageCapacity();

private:
    size_t m_objectWithOutOfLineStorageCount;
    size_t m_objectCount;
    size_t m_storageSize;
    size_t m_storageCapacity;
};

inline StorageStatistics::StorageStatistics()
    : m_objectWithOutOfLineStorageCount(0)
    , m_objectCount(0)
    , m_storageSize(0)
    , m_storageCapacity(0)
{
}

inline void StorageStatistics::operator()(JSCell* cell)
{
    if (!cell->isObject())
        return;

    JSObject* object = jsCast<JSObject*>(cell);
    if (hasIndexedProperties(object->structure()->indexingType()))
        return;

    if (object->structure()->isUncacheableDictionary())
        return;

    ++m_objectCount;
    if (!object->hasInlineStorage())
        ++m_objectWithOutOfLineStorageCount;
    m_storageSize += object->structure()->totalStorageSize() * sizeof(WriteBarrierBase<Unknown>);
    m_storageCapacity += object->structure()->totalStorageCapacity() * sizeof(WriteBarrierBase<Unknown>);
}

inline size_t StorageStatistics::objectWithOutOfLineStorageCount()
{
    return m_objectWithOutOfLineStorageCount;
}

inline size_t StorageStatistics::objectCount()
{
    return m_objectCount;
}


inline size_t StorageStatistics::storageSize()
{
    return m_storageSize;
}


inline size_t StorageStatistics::storageCapacity()
{
    return m_storageCapacity;
}

} // anonymous namespace

Heap::Heap(JSGlobalData* globalData, HeapType heapType)
: m_heapType(heapType)
, m_ramSize(ramSize())
, m_minBytesPerCycle(minHeapSize(m_heapType))
, m_sizeAfterLastCollect(0)
, m_bytesAllocatedLimit(m_minBytesPerCycle)
, m_bytesAllocated(0)
, m_extraBytesAllocated(0)
, m_deprecatedExtraBytesAllocated(0)
, m_bytesAbandoned(0)
, m_shouldDoFullCollection(true)
, m_totalBytesVisited(0)
, m_totalBytesCopied(0)
, m_extraMemorySize(0)
, m_deprecatedExtraMemorySize(0)
, m_operationInProgress(NoOperation)
, m_objectSpace(this)
, m_storageSpace(this)
, m_machineThreads(this)
, m_sharedData(globalData)
, m_slotVisitor(m_sharedData)
, m_copyVisitor(m_sharedData)
, m_handleSet(globalData)
, m_jitStubRoutines(adoptPtr(new JITStubRoutineSet()))
, m_isSafeToCollect(false)
, m_vm(globalData)
, m_lastGCLength(0)
, m_lastOpaqueRootCount(0)
, m_lastCodeDiscardTime(0)
, m_indexOfNextLogicallyEmptyWeakBlockToSweep(WTF::notFound)
#if ENABLE(GC_ACTIVITY_CALLBACK)
, m_activityCallback(DefaultGCActivityCallback::create(this))
#endif
#if ENABLE(GC_INCREMENTALSWEEPER)
, m_sweeper(IncrementalSweeper::create(this))
#endif
, m_deferralDepth(0)
{
    m_storageSpace.init();
}

Heap::~Heap()
{
    for (int i = 0; i < m_logicallyEmptyWeakBlocks.size(); ++i)
        WeakBlock::destroy(m_logicallyEmptyWeakBlocks[i]);
}

bool Heap::isPagedOut(double deadline)
{
    return m_objectSpace.isPagedOut(deadline) || m_storageSpace.isPagedOut(deadline);
}

// The JSGlobalData is being destroyed and the collector will never run again.
// Run all pending finalizers now because we won't get another chance.
void Heap::lastChanceToFinalize()
{
    ASSERT(!m_vm->dynamicGlobalObject);
    ASSERT(m_operationInProgress == NoOperation);

    m_arrayBuffers.lastChanceToFinalize();
    m_objectSpace.lastChanceToFinalize();

    sweepAllLogicallyEmptyWeakBlocks();

#if ENABLE(SIMPLE_HEAP_PROFILING)
    m_slotVisitor.m_visitedTypeCounts.dump(WTF::dataFile(), "Visited Type Counts");
    m_destroyedTypeCounts.dump(WTF::dataFile(), "Destroyed Type Counts");
#endif
}

void Heap::reportExtraMemoryAllocatedSlowCase(size_t size)
{
    didAllocate(size);
    collectIfNecessaryOrDefer();
}

void Heap::deprecatedReportExtraMemorySlowCase(size_t size)
{ // FIXME: Change this to use SaturatedArithmetic when available.
    // https://bugs.webkit.org/show_bug.cgi?id=170411
    Checked<size_t, RecordOverflow> checkedNewSize = m_deprecatedExtraMemorySize;
    checkedNewSize += size;
    m_deprecatedExtraMemorySize = UNLIKELY(checkedNewSize.hasOverflowed()) ? std::numeric_limits<size_t>::max() : checkedNewSize.unsafeGet();
    reportExtraMemoryAllocatedSlowCase(size);
}

void Heap::reportAbandonedObjectGraph()
{
    // Our clients don't know exactly how much memory they
    // are abandoning so we just guess for them.
    double abandonedBytes = 0.10 * m_sizeAfterLastCollect;

    // We want to accelerate the next collection. Because memory has just
    // been abandoned, the next collection has the potential to
    // be more profitable. Since allocation is the trigger for collection,
    // we hasten the next collection by pretending that we've allocated more memory.
    didAbandon(abandonedBytes);
}

void Heap::protect(JSValue k)
{
    ASSERT(k);
    ASSERT(m_vm->apiLock().currentThreadIsHoldingLock());

    if (k.isCell())
    	m_protectedValues.add(k.asCell());
}

bool Heap::unprotect(JSValue k)
{
    ASSERT(k);
    ASSERT(m_vm->apiLock().currentThreadIsHoldingLock());

    if (k.isCell())
    	return m_protectedValues.remove(k.asCell());
    return false;
}

void Heap::addReference(JSCell* cell, ArrayBuffer* buffer)
{
    if (m_arrayBuffers.addReference(cell, buffer)) {
        collectIfNecessaryOrDefer();
        didAllocate(buffer->gcSizeEstimateInBytes());
    }
}

void Heap::markProtectedObjects(HeapRootVisitor& heapRootVisitor)
{
    ProtectCountSet::iterator end = m_protectedValues.end();
    for (ProtectCountSet::iterator it = m_protectedValues.begin(); it != end; ++it)
        heapRootVisitor.visit(&it->first);
}

void Heap::harvestWeakReferences()
{
    m_slotVisitor.harvestWeakReferences();
}

void Heap::finalizeUnconditionalFinalizers()
{
    m_slotVisitor.finalizeUnconditionalFinalizers();
}

inline JSStack& Heap::registerFile()
{
    return m_vm->interpreter->registerFile();
}

void Heap::willStartIterating()
{
    m_objectSpace.willStartIterating();
}

void Heap::didFinishIterating()
{
    m_objectSpace.didFinishIterating();
}

void Heap::getConservativeRegisterRoots(HashSet<JSCell*>& roots)
{
    ASSERT(isValidThreadState(m_vm));
    ConservativeRoots registerFileRoots(&m_objectSpace.blocks(), &m_storageSpace);
    registerFile().gatherConservativeRoots(registerFileRoots);
    size_t registerFileRootCount = registerFileRoots.size();
    JSCell** registerRoots = registerFileRoots.roots();
    for (size_t i = 0; i < registerFileRootCount; i++) {
        setMarked(registerRoots[i]);
        roots.add(registerRoots[i]);
    }
}

void Heap::markRoots(bool fullGC, void* stackOrigin, void* stackTop, MachineThreads::RegisterState& calleeSavedRegisters, bool noStackObjects)
{
    SamplingRegion samplingRegion("Garbage Collection: Tracing");

    COND_GCPHASE(fullGC, MarkFullRoots, MarkYoungRoots);
    UNUSED_PARAM(fullGC);
    ASSERT(isValidThreadState(m_vm));

#if ENABLE(OBJECT_MARK_LOGGING)
    double gcStartTime = WTF::monotonicallyIncreasingTime();
#endif

    m_jitStubRoutines->prepareForConservativeScan();
    
    // We gather conservative roots before clearing mark bits because conservative
    // gathering uses the mark bits to determine whether a reference is valid.
    ConservativeRoots machineThreadRoots(&m_objectSpace.blocks(), &m_storageSpace);
    m_jitStubRoutines->clearMarks();
    {
        GCPHASE(GatherConservativeRoots);
        m_machineThreads.gatherConservativeRoots(machineThreadRoots, stackOrigin, stackTop, calleeSavedRegisters, noStackObjects);
    }

    ConservativeRoots registerFileRoots(&m_objectSpace.blocks(), &m_storageSpace);
    m_codeBlocks.clearMarks();
    {
        GCPHASE(GatherRegisterFileRoots);
        registerFile().gatherConservativeRoots(registerFileRoots, *m_jitStubRoutines, m_codeBlocks);
    }

#if ENABLE(DFG_JIT)
    ConservativeRoots scratchBufferRoots(&m_objectSpace.blocks(), &m_storageSpace);
    {
        GCPHASE(GatherScratchBufferRoots);
        m_vm->gatherConservativeRoots(scratchBufferRoots, noStackObjects);
    }
#if 0
    {
        GCPHASE(VisitDFGWorklists);
        MARK_LOG_ROOT(visitor, "DFG Worklists");
        for (unsigned i = DFG::numberOfWorklists(); i--;) {
            if (DFG::Worklist* worklist = DFG::worklistForIndexOrNull(i))
                worklist->visitChildren(visitor, m_codeBlocks);
        }
    }
#endif
#endif

#if ENABLE(GGC)
    MarkedBlock::DirtyCellVector dirtyCells;
    if (!fullGC) {
        GCPHASE(GatheringDirtyCells);
        m_objectSpace.gatherDirtyCells(dirtyCells);
    } else
#endif
    {
        GCPHASE(ClearLivenessData);
        m_objectSpace.clearNewlyAllocated();
        m_objectSpace.clearMarks();
    }

    m_sharedData.didStartMarking();
    SlotVisitor& visitor = m_slotVisitor;
    HeapRootVisitor heapRootVisitor(visitor);

    {
        ParallelModeEnabler enabler(visitor);
#if ENABLE(GGC)
        {
            size_t dirtyCellCount = dirtyCells.size();
            GCPHASE(VisitDirtyCells);
            GCCOUNTER(DirtyCellCount, dirtyCellCount);
            for (size_t i = 0; i < dirtyCellCount; i++) {
                heapRootVisitor.visitChildren(dirtyCells[i]);
                visitor.donateAndDrain();
            }
        }
#endif
        {
            GCPHASE(VisitSmallStrings);
            MARK_LOG_ROOT(visitor, "Small strings");
            if (m_vm->smallStrings.needsToBeVisited(m_operationInProgress)) {
            	m_vm->smallStrings.visitStrongReferences(visitor);
            	m_slotVisitor.donateAndDrain();
            }
        }

        {
            GCPHASE(VisitMachineRoots);
            MARK_LOG_ROOT(visitor, "C++ Stack");
            visitor.append(machineThreadRoots);
            visitor.donateAndDrain();
        }
        {
            GCPHASE(VisitRegisterFileRoots);
            MARK_LOG_ROOT(visitor, "Register File");
            visitor.append(registerFileRoots);
            visitor.donateAndDrain();
        }
#if ENABLE(DFG_JIT)
        {
            GCPHASE(VisitScratchBufferRoots);
            MARK_LOG_ROOT(visitor, "Scratch Buffers");
            visitor.append(scratchBufferRoots);
            visitor.donateAndDrain();
        }
#endif
        {
            GCPHASE(VisitProtectedObjects);
            MARK_LOG_ROOT(visitor, "Protected Objects");
            markProtectedObjects(heapRootVisitor);
            visitor.donateAndDrain();
        }

        {
            GCPHASE(MarkingArgumentBuffers);
            if (m_markListSet && m_markListSet->size()) {
                MARK_LOG_ROOT(visitor, "Argument Buffers");
                MarkedArgumentBuffer::markLists(heapRootVisitor, *m_markListSet);
                visitor.donateAndDrain();
            }
        }

        visitException(heapRootVisitor);
    
        {
            GCPHASE(VisitStrongHandles);
            MARK_LOG_ROOT(visitor, "Strong Handles");
            m_handleSet.visitStrongHandles(heapRootVisitor);
            visitor.donateAndDrain();
        }

        {
            GCPHASE(TraceCodeBlocksAndJITStubRoutines);
            MARK_LOG_ROOT(visitor, "Trace Code Blocks and JIT Stub Routines");
            m_codeBlocks.traceMarked(visitor);
            m_jitStubRoutines->traceMarkedStubRoutines(visitor);
            visitor.donateAndDrain();
        }
    
#if ENABLE(PARALLEL_GC)
        {
            GCPHASE(Convergence);
            visitor.drainFromShared(SlotVisitor::MasterDrain);
        }
#endif
    }

    // Weak references must be marked last because their liveness depends on
    // the liveness of the rest of the object graph.
    {
        GCPHASE(VisitingLiveWeakHandles);
        MARK_LOG_ROOT(visitor, "Live Weak Handles");
        while (true) {
            m_objectSpace.visitWeakSets(heapRootVisitor);
            harvestWeakReferences();
            if (visitor.isEmpty())
                break;
            {
                ParallelModeEnabler enabler(visitor);
                visitor.donateAndDrain();
#if ENABLE(PARALLEL_GC)
                visitor.drainFromShared(SlotVisitor::MasterDrain);
#endif
            }
        }
    }

    GCCOUNTER(VisitedValueCount, visitor.visitCount());

    m_sharedData.didFinishMarking();
#if ENABLE(OBJECT_MARK_LOGGING)
    size_t visitCount = visitor.visitCount();
#if ENABLE(PARALLEL_GC)
    visitCount += m_sharedData.childVisitCount();
#endif
    MARK_LOG_MESSAGE2("\nNumber of live Objects after full GC %lu, took %.6f secs\n", visitCount, WTF::monotonicallyIncreasingTime() - gcStartTime);
#endif

    m_totalBytesVisited = visitor.bytesVisited();
    m_totalBytesCopied = visitor.bytesCopied();
    m_lastOpaqueRootCount = visitor.opaqueRootCount();
#if ENABLE(PARALLEL_GC)
    m_totalBytesVisited += m_sharedData.childBytesVisited();
    m_totalBytesCopied += m_sharedData.childBytesCopied();
#endif

    visitor.reset();
#if ENABLE(PARALLEL_GC)
    m_sharedData.resetChildren();
#endif
    m_sharedData.reset();
}

void Heap::visitException(HeapRootVisitor& visitor)
{
    GCPHASE(MarkingException);
    if (!m_vm->exception() && !m_vm->lastException())
        return;

    visitor.visit(m_vm->addressOfException());
    visitor.visit(m_vm->addressOfLastException());

    //if (Options::logGC() == GCLogging::Verbose)
    //    dataLog("Exceptions:\n", m_slotVisitor);

    m_slotVisitor.donateAndDrain();
}

void Heap::copyBackingStores()
{
    GCPHASE(CopyBackingStores);
    m_storageSpace.startedCopying();
    if (m_storageSpace.shouldDoCopyPhase()) {
        m_sharedData.didStartCopying();
        m_copyVisitor.startCopying();
        m_copyVisitor.copyFromShared();
        m_copyVisitor.doneCopying();
        // We need to wait for everybody to finish and return their CopiedBlocks
        // before signaling that the phase is complete.
        m_storageSpace.doneCopying();
        m_sharedData.didFinishCopying();
    } else
        m_storageSpace.doneCopying();
}

size_t Heap::objectCount()
{
    return m_objectSpace.objectCount();
}

size_t Heap::extraMemorySize()
{
    // FIXME: Change this to use SaturatedArithmetic when available.
    // https://bugs.webkit.org/show_bug.cgi?id=170411
    Checked<size_t, RecordOverflow> checkedTotal = m_extraMemorySize;
    checkedTotal += m_deprecatedExtraMemorySize;
    checkedTotal += m_arrayBuffers.size();
    size_t total = UNLIKELY(checkedTotal.hasOverflowed()) ? std::numeric_limits<size_t>::max() : checkedTotal.unsafeGet();

    ASSERT(m_objectSpace.capacity() >= m_objectSpace.size());
    return std::min(total, std::numeric_limits<size_t>::max() - m_objectSpace.capacity());
}

size_t Heap::size()
{
    return m_objectSpace.size() + m_storageSpace.size() + extraMemorySize();
}

size_t Heap::capacity()
{
    return m_objectSpace.capacity() + m_storageSpace.capacity() + extraMemorySize();
}

size_t Heap::sizeAfterCollect()
{
    // The result here may not agree with the normal Heap::size().
    // This is due to the fact that we only count live copied bytes
    // rather than all used (including dead) copied bytes, thus it's
    // always the case that m_totalBytesCopied <= m_storageSpace.size().
    ASSERT(m_totalBytesCopied <= m_storageSpace.size());
    return m_totalBytesVisited + m_totalBytesCopied + extraMemorySize();
}

size_t Heap::protectedGlobalObjectCount()
{
    return forEachProtectedCell<CountIfGlobalObject>();
}

size_t Heap::globalObjectCount()
{
    HeapIterationScope iterationScope(*this);
    return m_objectSpace.forEachLiveCell<CountIfGlobalObject>(iterationScope);
}

size_t Heap::protectedObjectCount()
{
    return forEachProtectedCell<Count>();
}

PassOwnPtr<TypeCountSet> Heap::protectedObjectTypeCounts()
{
    return forEachProtectedCell<RecordType>();
}

PassOwnPtr<TypeCountSet> Heap::objectTypeCounts()
{
    HeapIterationScope iterationScope(*this);
    return m_objectSpace.forEachLiveCell<RecordType>(iterationScope);
}

bool Heap::shouldDeleteCodeBlocks() const
{
	if(ExecutableAllocator::underMemoryPressure())
		return true;

	if(int interval = Options::jitCodesDeletionInterval())
		return (WTF::monotonicallyIncreasingTime() - m_lastCodeDiscardTime) >= interval;

	return false;
}

void Heap::deleteAllCodeBlocks(bool all)
{
	// If JavaScript is running, it's not safe to delete all JavaScript code, since
    // we'll end up returning to deleted code.
    if (m_vm->dynamicGlobalObject)
        return;

    int deletedBytes = 0;
    int targetDeleteBytes = 0;

    if(Options::jitCodesDeletionPercentage() >= 1)
    	all = true;

    if(all) {
        for(ExecutableBase* executable = m_executables.head(); executable; executable = executable->next()) {
            if (executable->isFunctionExecutable())
            	static_cast<FunctionExecutable*>(executable)->clearCode();
        }
    } else {
		int capBytes = jitCodeCapBytes();
		int reservedBytes = jitCodeReservedBytes();
		targetDeleteBytes = (reservedBytes > capBytes) ? (reservedBytes - capBytes) : 0;
		targetDeleteBytes += (int)(capBytes * Options::jitCodesDeletionPercentage());

        for(ExecutableBase* executable = m_executables.head(); executable; executable = executable->next()) {
            if (!executable->isFunctionExecutable())
            	continue;

            FunctionExecutable* functionExecutable = static_cast<FunctionExecutable*>(executable);

            CodeBlock* codeBlockForCall = functionExecutable->generatedBytecodeFor(CodeForCall);
            CodeBlock* codeBlockForConstruct = functionExecutable->generatedBytecodeFor(CodeForConstruct);

            if( deletedBytes < targetDeleteBytes ||
    			(codeBlockForCall && JITCode::isOptimizingJIT(codeBlockForCall->jitType())) ||
    			(codeBlockForConstruct && JITCode::isOptimizingJIT(codeBlockForConstruct->jitType()))) {

    			for(CodeBlock* codeBlock = codeBlockForCall; codeBlock; codeBlock = codeBlock->alternative()) {
    				codeBlock->unlinkIncomingCalls();
    				deletedBytes += codeBlock->jitCode()->size();
    			}

    			for(CodeBlock* codeBlock = codeBlockForConstruct; codeBlock; codeBlock = codeBlock->alternative()) {
    				codeBlock->unlinkIncomingCalls();
    				deletedBytes += codeBlock->jitCode()->size();
    			}

    			functionExecutable->clearCode();
            }
        }
    }

    m_codeBlocks.clearMarks();
    m_codeBlocks.deleteUnmarkedAndUnreferenced();

#if ENABLE(ASSEMBLER) && ENABLE(EXECUTABLE_ALLOCATOR_DEMAND)
    gcExecutableAllocators(all);
#endif

    m_lastCodeDiscardTime = WTF::monotonicallyIncreasingTime();

    WTF::MmapMonitor::didDeleteAllCodeBlocks(all, m_codeBlocks.size(), deletedBytes, targetDeleteBytes);
}

void Heap::deleteUnmarkedCompiledCode()
{
	ExecutableBase* current = m_executables.head();
	while(current) {
		ExecutableBase* next = current->next();
		if (!isMarked(current)) {
	        // We do this because executable memory is limited on some platforms and because
	        // CodeBlock requires eager finalization.
	        ExecutableBase::clearCodeVirtual(current);
	        m_executables.remove(current);
		}
		current = next;
	}

    m_codeBlocks.deleteUnmarkedAndUnreferenced();
    m_jitStubRoutines->deleteUnmarkedJettisonedStubRoutines();
}

bool Heap::collectAllGarbage(bool noStackObjects, bool force)
{
    if (!m_isSafeToCollect)
        return false;

    if(collect(noStackObjects, force)) {
		SamplingRegion samplingRegion("Garbage Collection: Sweeping");

		DeferGCForAWhile deferGC(*this);
		//DelayedReleaseScope delayedReleaseScope(m_objectSpace);
		m_objectSpace.sweep();
		m_objectSpace.shrink();

		sweepAllLogicallyEmptyWeakBlocks();
    	return true;
    }

    return false;
}

void Heap::pruneStaleEntriesFromWeakGCMaps()
{
    GCPHASE(PruningStaleEntriesFromWeakGCMaps);
    if (m_operationInProgress != FullCollection)
        return;
    for (auto it = m_weakGCMaps.begin(); it != m_weakGCMaps.end(); ++it)
        (*it)->pruneStaleEntries();
}

bool Heap::collect(bool noStackObjects, bool force)
{
    if(!force && WTF::MmapMonitor::shouldDeferGC(m_bytesAllocatedLimit))
    	return false;

    void* stackTop;
    ALLOCATE_AND_GET_REGISTER_STATE(registers);

    collectImpl(wtfThreadData().stack().origin(), &stackTop, registers, noStackObjects);

    sanitizeStackForVM(*m_vm);

    return true;
}

void Heap::collectImpl(void* stackOrigin, void* stackTop, MachineThreads::RegisterState& calleeSavedRegisters, bool noStackObjects)
{
    SamplingRegion samplingRegion("Garbage Collection");
    
    GCPHASE(Collect);
    ASSERT(globalData()->apiLock().currentThreadIsHoldingLock());
    ASSERT(globalData()->identifierTable == wtfThreadData().currentIdentifierTable());
    ASSERT(m_isSafeToCollect);
    JAVASCRIPTCORE_GC_BEGIN();
#ifndef NDEBUG
    if (m_operationInProgress != NoOperation)
        RELEASE_ASSERT_NOT_REACHED();
#endif
    m_operationInProgress = FullCollection;

    m_extraMemorySize = 0;
	m_deprecatedExtraMemorySize = 0;

    m_extraBytesAllocated = 0;
    m_deprecatedExtraBytesAllocated = 0;

#if ENABLE(GC_ACTIVITY_CALLBACK)
    m_activityCallback->willCollect();
#endif

    double gcStartTime = WTF::monotonicallyIncreasingTime();

#if ENABLE(GGC)
    bool fullGC = sweepToggle == DoSweep;
    if (!fullGC)
        fullGC = (capacity() > 4 * m_sizeAfterLastCollect);  
#else
    bool fullGC = true;
#endif

    m_structureRareDataWithObjectToStringAdaptiveInferredValueWatchpointMap.clear();

    {
        GCPHASE(StopAllocation);
        m_objectSpace.stopAllocating();
    }

    markRoots(fullGC, stackOrigin, stackTop, calleeSavedRegisters, noStackObjects);

    JAVASCRIPTCORE_GC_MARKED();
    
    {
        GCPHASE(ReapingWeakHandles);
        m_objectSpace.reapWeakSets();
    }

    pruneStaleEntriesFromWeakGCMaps();

    {
        GCPHASE(SweepingArrayBuffers);
        m_arrayBuffers.sweep();
    }

    {
        m_blockSnapshot.resize(m_objectSpace.blocks().set().size());
        MarkedBlockSnapshotFunctor functor(m_blockSnapshot);
        m_objectSpace.forEachBlock(functor);
    }

    copyBackingStores();

    {
        GCPHASE(FinalizeUnconditionalFinalizers);
        finalizeUnconditionalFinalizers();
    }

    {
        GCPHASE(DeleteCodeBlocks);
        deleteUnmarkedCompiledCode();
    }

    {
        GCPHASE(DeleteSourceProviderCaches);
        m_vm->clearSourceProviderCaches();
    }

    {
    	if (!m_logicallyEmptyWeakBlocks.isEmpty())
    		m_indexOfNextLogicallyEmptyWeakBlockToSweep = 0;
#if ENABLE(GC_INCREMENTALSWEEPER)
    	m_sweeper->startSweeping(m_blockSnapshot);
#endif
    }

    m_bytesAbandoned = 0;

    {
        GCPHASE(ResetAllocators);
        m_objectSpace.resetAllocators();
    }

    size_t currentHeapSize = sizeAfterCollect();
    if (fullGC) {
        m_sizeAfterLastCollect = currentHeapSize;

        // To avoid pathological GC churn in very small and very large heaps, we set
        // the new allocation limit based on the current size of the heap, with a
        // fixed minimum.
        size_t maxHeapSize = max(minHeapSize(m_heapType), proportionalHeapSize(currentHeapSize));
        m_bytesAllocatedLimit = maxHeapSize - currentHeapSize;
    }

#if ENABLE(ASSEMBLER) && ENABLE(EXECUTABLE_ALLOCATOR_DEMAND)
    // If JavaScript is running, it's not safe to delete all JavaScript code, since
	// we'll end up returning to deleted code.
	if(!m_vm->dynamicGlobalObject && shouldDeleteCodeBlocks()){
		GCPHASE(DeleteOldCode);
		m_vm->regExpCache()->deleteAllCode();
		deleteAllCodeBlocks(false);
	}
	else
	    gcExecutableAllocators(false);
#endif

    m_bytesAllocated = 0;

    double gcEndTime = WTF::monotonicallyIncreasingTime();
    m_lastGCLength = gcEndTime - gcStartTime;

#ifndef NDEBUG
    if (m_operationInProgress != FullCollection)
        RELEASE_ASSERT_NOT_REACHED();
#endif
    m_operationInProgress = NoOperation;
    JAVASCRIPTCORE_GC_END();

    WTF::MmapMonitor::didGC(m_lastGCLength, m_sizeAfterLastCollect, m_bytesAllocatedLimit, m_lastOpaqueRootCount);

    if (Options::useZombieMode())
        zombifyDeadObjects();

    if (Options::objectsAreImmortal())
        markDeadObjects();

    if (Options::showHeapStatistics())
        showStatistics();
}

void Heap::showStatistics()
{
    dataLog("\n=== Heap Statistics: ===\n");
    dataLog("size: %ldkB\n", static_cast<long>(m_sizeAfterLastCollect / KB));
    dataLog("capacity: %ldkB\n", static_cast<long>(capacity() / KB));
    dataLog("pause time: %lfms\n\n", m_lastGCLength);

    StorageStatistics storageStatistics;
    {
        HeapIterationScope iterationScope(*this);
        m_objectSpace.forEachLiveCell(iterationScope, storageStatistics);
    }
    dataLog("wasted .property storage: %ldkB (%ld%%)\n",
        static_cast<long>(
            (storageStatistics.storageCapacity() - storageStatistics.storageSize()) / KB),
        static_cast<long>(
            (storageStatistics.storageCapacity() - storageStatistics.storageSize()) * 100
                / storageStatistics.storageCapacity()));
    dataLog("objects with out-of-line .property storage: %ld (%ld%%)\n",
        static_cast<long>(
            storageStatistics.objectWithOutOfLineStorageCount()),
        static_cast<long>(
            storageStatistics.objectWithOutOfLineStorageCount() * 100
                / storageStatistics.objectCount()));
}

void Heap::markDeadObjects()
{
    HeapIterationScope iterationScope(*this);
    m_objectSpace.forEachDeadCell<MarkObject>(iterationScope);
}

#if ENABLE(GC_ACTIVITY_CALLBACK)
void Heap::setActivityCallback(GCActivityCallback* activityCallback)
{
    m_activityCallback = activityCallback;
}

GCActivityCallback* Heap::activityCallback()
{
    return m_activityCallback;
}
#endif

#if ENABLE(GC_INCREMENTALSWEEPER)
IncrementalSweeper* Heap::sweeper()
{
    return m_sweeper;
}
#endif

void Heap::setGarbageCollectionTimerEnabled(bool enable)
{
#if ENABLE(GC_ACTIVITY_CALLBACK)
    activityCallback()->setEnabled(enable);
#endif
}

bool Heap::isValidAllocation(size_t)
{
    if (!isValidThreadState(m_vm))
        return false;

    if (m_operationInProgress != NoOperation)
        return false;
    
    return true;
}

void Heap::addFinalizer(JSCell* cell, Finalizer finalizer)
{
    WeakSet::allocate(cell, &m_finalizerOwner, reinterpret_cast<void*>(finalizer)); // Balanced by FinalizerOwner::finalize().
}

void Heap::FinalizerOwner::finalize(Handle<Unknown> handle, void* context)
{
    HandleSlot slot = handle.slot();
    Finalizer finalizer = reinterpret_cast<Finalizer>(context);
    finalizer(slot->asCell());
    WeakSet::deallocate(WeakImpl::asWeakImpl(slot));
}

void Heap::addExecutable(ExecutableBase* executable)
{
    m_executables.append(executable);
}

void Heap::didInstallCode(ExecutableBase* executable)
{
	m_executables.remove(executable);
    m_executables.append(executable);
}

void Heap::didStartVMShutdown()
{
#if ENABLE(GC_ACTIVITY_CALLBACK)
    m_activityCallback->didStartVMShutdown();
    m_activityCallback = 0;
#endif
#if ENABLE(GC_INCREMENTALSWEEPER)
    m_sweeper->didStartVMShutdown();
    m_sweeper = 0;
#endif
    lastChanceToFinalize();
}

class Zombify : public MarkedBlock::VoidFunctor {
public:
    void operator()(JSCell* cell)
    {
        void** current = reinterpret_cast<void**>(cell);

        // We want to maintain zapped-ness because that's how we know if we've called
        // the destructor.
        if (cell->isZapped())
            current++;

        void* limit = static_cast<void*>(reinterpret_cast<char*>(cell) + MarkedBlock::blockFor(cell)->cellSize());
        for (; current < limit; current++)
            *current = zombifiedBits;
    }
};

void Heap::zombifyDeadObjects()
{
    // Sweep now because destructors will crash once we're zombified.
    m_objectSpace.sweep();
    HeapIterationScope iterationScope(*this);
    m_objectSpace.forEachDeadCell<Zombify>(iterationScope);
}

void Heap::addLogicallyEmptyWeakBlock(WeakBlock* block)
{
    m_logicallyEmptyWeakBlocks.append(block);
}

void Heap::sweepAllLogicallyEmptyWeakBlocks()
{
    if (m_logicallyEmptyWeakBlocks.isEmpty())
        return;

    m_indexOfNextLogicallyEmptyWeakBlockToSweep = 0;
    while (sweepNextLogicallyEmptyWeakBlock()) { }
}

bool Heap::sweepNextLogicallyEmptyWeakBlock()
{
    if (m_indexOfNextLogicallyEmptyWeakBlockToSweep == WTF::notFound)
        return false;

    WeakBlock* block = m_logicallyEmptyWeakBlocks[m_indexOfNextLogicallyEmptyWeakBlockToSweep];

    block->sweep();
    if (block->isEmpty()) {
        std::swap(m_logicallyEmptyWeakBlocks[m_indexOfNextLogicallyEmptyWeakBlockToSweep], m_logicallyEmptyWeakBlocks.last());
        m_logicallyEmptyWeakBlocks.removeLast();
        WeakBlock::destroy(block);
    } else
        m_indexOfNextLogicallyEmptyWeakBlockToSweep++;

    if (m_indexOfNextLogicallyEmptyWeakBlockToSweep >= m_logicallyEmptyWeakBlocks.size()) {
        m_indexOfNextLogicallyEmptyWeakBlockToSweep = WTF::notFound;
        return false;
    }

    return true;
}

} // namespace JSC
