/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003, 2007, 2008, 2011 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#ifndef CallFrame_h
#define CallFrame_h

#include "AbstractPC.h"
#include "MacroAssemblerCodeRef.h"
#include "JSStack.h"
#include "Register.h"
#include "StackVisitor.h"
#include "VM.h"

namespace JSC  {

class Arguments;
class ExecState;
class VirtualRegister_;
class Interpreter;
class JSScope;

typedef ExecState CallFrame;

struct CallSiteIndex {
    CallSiteIndex()
        : m_bits(UINT_MAX)
    {
    }

    explicit CallSiteIndex(uint32_t bits)
        : m_bits(bits)
    { }
#if USE(JSVALUE32_64)
    explicit CallSiteIndex(Instruction* instruction)
        : m_bits(bitwise_cast<uint32_t>(instruction))
    { }
#endif

    explicit operator bool() const { return m_bits != UINT_MAX; }

    bool operator==(const CallSiteIndex& other) const { return m_bits == other.m_bits; }

    inline uint32_t bits() const { return m_bits; }

private:
    uint32_t m_bits;
};

// Represents the current state of script execution.
// Passed as the first argument to most functions.
class ExecState : private Register {
public:
	ALWAYS_INLINE JSValue calleeAsValue() const { return this[JSStack::Callee].jsValue(); }
	ALWAYS_INLINE JSObject* jsCallee() const { return this[JSStack::Callee].object(); }
	ALWAYS_INLINE JSObject* callee() const { return jsCallee(); }
	ALWAYS_INLINE CodeBlock* codeBlock() const { return this[JSStack::CodeBlock].Register::codeBlock(); }
	CodeBlock** addressOfCodeBlock() const { return bitwise_cast<CodeBlock**>(this + JSStack::CodeBlock); }
	JSScope* scope(int scopeRegisterOffset) const
    {
        ASSERT(this[scopeRegisterOffset].Register::scope());
        return this[scopeRegisterOffset].Register::scope();
    }

    // Under the constructor implemented in C++, thisValue holds the newTarget instead of the automatically constructed value.
    // The result of this function is only effective under the "construct" context.
    JSValue newTarget() { return thisValue(); }

	// Global object in which execution began.
	JSGlobalObject* dynamicGlobalObject();
	JSGlobalObject* vmEntryGlobalObject() { return dynamicGlobalObject(); }

	// Global object in which the currently executing code was defined.
	// Differs from dynamicGlobalObject() during function calls across web browser frames.
	inline JSGlobalObject* lexicalGlobalObject() const;

	// Differs from lexicalGlobalObject because this will have DOM window shell rather than
	// the actual DOM window, which can't be "this" for security reasons.
	inline JSObject* globalThisValue() const;

	JSGlobalData& globalData() const;
	VM& vm() const;

	// Convenience functions for access to global data.
	// It takes a few memory references to get from a call frame to the global data
	// pointer, so these are inefficient, and should be used sparingly in new code.
	// But they're used in many places in legacy code, so they're not going away any time soon.

	ALWAYS_INLINE void clearException() { vm().clearException(); }

    Exception* exception() const { return vm().exception(); }
    bool hadException() const { return !!vm().exception(); }

    Exception* lastException() const { return vm().lastException(); }
    void clearLastException() { vm().clearLastException(); }

	const CommonIdentifiers& propertyNames() const { return *globalData().propertyNames; }

	static CallFrame* create(Register* callFrameBase) { return static_cast<CallFrame*>(callFrameBase); }
	Register* registers() { return this; }
    const Register* registers() const { return this; }

	CallFrame& operator=(const Register& r) { *static_cast<Register*>(this) = r; return *this; }

    CallFrame* callerFrame() const { return callerFrameAndPC().callerFrame; }
    static ptrdiff_t callerFrameOffset() { return OBJECT_OFFSETOF(CallerFrameAndPC, callerFrame); }

    ReturnAddressPtr returnPC() const { return ReturnAddressPtr(callerFrameAndPC().pc); }
    bool hasReturnPC() const { return !!callerFrameAndPC().pc; }
    void clearReturnPC() { callerFrameAndPC().pc = 0; }
    static ptrdiff_t returnPCOffset() { return OBJECT_OFFSETOF(CallerFrameAndPC, pc); }
    AbstractPC abstractReturnPC(JSGlobalData& globalData) { return AbstractPC(globalData, this); }

    bool callSiteBitsAreBytecodeOffset() const;
    bool callSiteBitsAreCodeOriginIndex() const;

    unsigned callSiteAsRawBits() const;
    CallSiteIndex callSiteIndex() const;
private:
    unsigned callSiteBitsAsBytecodeOffset() const;
public:

    // This will try to get you the bytecode offset, but you should be aware that
    // this bytecode offset may be bogus in the presence of inlining. This will
    // also return 0 if the call frame has no notion of bytecode offsets (for
    // example if it's native code).
    // https://bugs.webkit.org/show_bug.cgi?id=121754
    unsigned bytecodeOffset();

    // This will get you a CodeOrigin. It will always succeed. May return
    // CodeOrigin(0) if we're in native code.
    CodeOrigin codeOrigin();

	Register* frameExtent()
	{
        if (isVMEntrySentinel() || !codeBlock())
			return registers() - 1;
		return frameExtentInternal();
	}

	Register* frameExtentInternal();

    Instruction* currentVPC() const; // This only makes sense in the LLInt and baseline.
	void setCurrentVPC(Instruction* vpc);

	void setCallerFrame(CallFrame* frame) { callerFrameAndPC().callerFrame = frame; }
	void setScope(int scopeRegisterOffset, JSScope* scope) { static_cast<Register*>(this)[scopeRegisterOffset] = scope; }

	ALWAYS_INLINE void init(CodeBlock* codeBlock, Instruction* vPC, CallFrame* callerFrame, int argc, JSObject* callee)
	{
        ASSERT(callerFrame == noCaller() || callerFrame->isVMEntrySentinel() || callerFrame->stack()->containsAddress(this));

		setCodeBlock(codeBlock);
		setCallerFrame(callerFrame);
		setReturnPC(vPC); // This is either an Instruction* or a pointer into JIT generated code stored as an Instruction*.
		setArgumentCountIncludingThis(argc); // original argument count (for the sake of the "arguments" object)
		setCallee(callee);
	}

	// Read a register from the codeframe (or constant from the CodeBlock).
	inline Register& r(int);
    Register& r(VirtualRegister_);
	// Read a register for a non-constant
	inline Register& uncheckedR(int);
    Register& uncheckedR(VirtualRegister_);

	// Access to arguments as passed. (After capture, arguments may move to a different location.)
	ALWAYS_INLINE size_t argumentCount() const { return argumentCountIncludingThis() - 1; }
	ALWAYS_INLINE size_t argumentCountIncludingThis() const { return this[JSStack::ArgumentCount].payload(); }
	static ALWAYS_INLINE int argumentOffset(int argument) { return (JSStack::FirstArgument + argument); }
	static ALWAYS_INLINE int argumentOffsetIncludingThis(int argument) { return (JSStack::ThisArgument + argument); }

	// In the following (argument() and setArgument()), the 'argument'
	// parameter is the index of the arguments of the target function of
	// this frame. The index starts at 0 for the first arg, 1 for the
	// second, etc.
	//
	// The arguments (in this case) do not include the 'this' value.
	// arguments(0) will not fetch the 'this' value. To get/set 'this',
	// use thisValue() and setThisValue() below.

	ALWAYS_INLINE JSValue argument(size_t argument)
	{
		if (LIKELY(argument < argumentCount()))
			return getArgumentUnsafe(argument);
		else
			return jsUndefined();
	}
    JSValue uncheckedArgument(size_t argument)
    {
        ASSERT(argument < argumentCount());
        return getArgumentUnsafe(argument);
    }
	ALWAYS_INLINE void setArgument(size_t argument, JSValue value)
	{
		this[argumentOffset(argument)] = value;
	}

    JSValue getArgumentUnsafe(size_t argIndex)
    {
        // User beware! This method does not verify that there is a valid
        // argument at the specified argIndex. This is used for debugging
        // and verification code only. The caller is expected to know what
        // he/she is doing when calling this method.
        return this[argumentOffset(argIndex)].jsValue();
    }

	static ALWAYS_INLINE int thisArgumentOffset() { return argumentOffsetIncludingThis(0); }
	ALWAYS_INLINE JSValue thisValue() { return this[thisArgumentOffset()].jsValue(); }
	ALWAYS_INLINE void setThisValue(JSValue value) { this[thisArgumentOffset()] = value; }

	static int offsetFor(size_t argumentCountIncludingThis) { return argumentCountIncludingThis + JSStack::ThisArgument - 1; }

	// FIXME: Remove these.
	ALWAYS_INLINE int hostThisRegister() { return thisArgumentOffset(); }
	ALWAYS_INLINE JSValue hostThisValue() { return thisValue(); }

    static CallFrame* noCaller() { return 0; }

    bool isVMEntrySentinel() const
    {
        return !!this && codeBlock() == vmEntrySentinelCodeBlock();
    }

    CallFrame* vmEntrySentinelCallerFrame() const
    {
        ASSERT(isVMEntrySentinel());
        return this[JSStack::ScopeChain].callFrame();
    }

    void initializeVMEntrySentinelFrame(CallFrame* callFrame)
    {
        setCallerFrame(noCaller());
        setReturnPC(0);
        setCodeBlock(vmEntrySentinelCodeBlock());
        static_cast<Register*>(this)[JSStack::ScopeChain] = callFrame;
        setCallee(0);
        setArgumentCountIncludingThis(0);
    }

    CallFrame* callerFrameSkippingVMEntrySentinel()
    {
        CallFrame* caller = callerFrame();
        if (caller->isVMEntrySentinel())
            return caller->vmEntrySentinelCallerFrame();
        return caller;
    }

	void setArgumentCountIncludingThis(int count) { static_cast<Register*>(this)[JSStack::ArgumentCount].payload() = count; }
	void setCallee(JSObject* callee) { static_cast<Register*>(this)[JSStack::Callee] = callee; }
	void setCodeBlock(CodeBlock* codeBlock) { static_cast<Register*>(this)[JSStack::CodeBlock] = codeBlock; }
	void setReturnPC(void* value) { callerFrameAndPC().pc = reinterpret_cast<Instruction*>(value); }

    // CallFrame::iterate() expects a Functor that implements the following method:
    //     StackVisitor::Status operator()(StackVisitor&);

    template <typename Functor> void iterate(Functor& functor)
    {
        StackVisitor::visit<Functor>(this, functor);
    }

private:
	static const intptr_t s_VMEntrySentinel = 1;

#ifndef NDEBUG
	JSStack* registerFile();
#endif
	ExecState();
	~ExecState();

	// The following are for internal use in debugging and verification
	// code only and not meant as an API for general usage:

	size_t argIndexForRegister(Register* reg)
	{
		// The register at 'offset' number of slots from the frame pointer
		// i.e.
		//       reg = frame[offset];
		//   ==> reg = frame + offset;
		//   ==> offset = reg - frame;
		int offset = reg - this->registers();

        // The offset is defined (based on argumentOffset()) to be:
        //       offset = JSStack::FirstArgument - argIndex;
        // Hence:
        //       argIndex = JSStack::FirstArgument - offset;
        size_t argIndex = offset - JSStack::FirstArgument;
		return argIndex;
	}

    CallerFrameAndPC& callerFrameAndPC() { return *reinterpret_cast<CallerFrameAndPC*>(this); }
    const CallerFrameAndPC& callerFrameAndPC() const { return *reinterpret_cast<const CallerFrameAndPC*>(this); }

    static CodeBlock* vmEntrySentinelCodeBlock() { return reinterpret_cast<CodeBlock*>(s_VMEntrySentinel); }

	friend class JSStack;
};

} // namespace JSC

#endif // CallFrame_h
