/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef DateInstance_h
#define DateInstance_h

#include "JSDestructibleObject.h"

namespace JSC {

    class DateInstance : public JSDestructibleObject {
    protected:
        JS_EXPORT_PRIVATE DateInstance(VM&, Structure*);
        void finishCreation(JSGlobalData&);
        JS_EXPORT_PRIVATE void finishCreation(JSGlobalData&, double);

        static void destroy(JSCell*);
 
    public:
        typedef JSDestructibleObject Base;

        static DateInstance* create(VM& vm, Structure* structure, double date)
        {
            DateInstance* instance = new (NotNull, allocateCell<DateInstance>(vm.heap)) DateInstance(vm, structure);
            instance->finishCreation(vm, date);
            return instance;
        }

        static DateInstance* create(VM& vm, Structure* structure)
        {
            DateInstance* instance = new (NotNull, allocateCell<DateInstance>(vm.heap)) DateInstance(vm, structure);
            instance->finishCreation(vm);
            return instance;
        }

        double internalNumber() const { return m_internalNumber; }
        void setInternalNumber(double value) { m_internalNumber = value; }

        DECLARE_EXPORT_INFO;

        const GregorianDateTime* gregorianDateTime(JSGlobalData& vm) const
        {
            if (m_data && m_data->m_gregorianDateTimeCachedForMS == internalNumber())
                return &m_data->m_cachedGregorianDateTime;
            return calculateGregorianDateTime(vm);
        }
        
        const GregorianDateTime* gregorianDateTimeUTC(JSGlobalData& vm) const
        {
            if (m_data && m_data->m_gregorianDateTimeUTCCachedForMS == internalNumber())
                return &m_data->m_cachedGregorianDateTimeUTC;
            return calculateGregorianDateTimeUTC(vm);
        }

        static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype)
        {
            return Structure::create(globalData, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
        }

    private:
        const GregorianDateTime* calculateGregorianDateTime(JSGlobalData& vm) const;
        const GregorianDateTime* calculateGregorianDateTimeUTC(JSGlobalData& vm) const;

        double m_internalNumber;
        mutable RefPtr<DateInstanceData> m_data;
    };

    DateInstance* asDateInstance(JSValue);

    inline DateInstance* asDateInstance(JSValue value)
    {
        ASSERT(asObject(value)->inherits(DateInstance::info()));
        return static_cast<DateInstance*>(asObject(value));
    }

} // namespace JSC

#endif // DateInstance_h
