/*
 * Copyright (C) 2009, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "Executable.h"

#include "BytecodeGenerator.h"
#include "CodeBlock.h"
#include "DFGDriver.h"
#include "DeferGC.h"
#include "Exception.h"
#include "JIT.h"
#include "Operations.h"
#include "Parser.h"
#include <wtf/CurrentTime.h>
#include <wtf/Vector.h>
#include <wtf/text/StringBuilder.h>

namespace JSC {

void ScriptExecutable::installCode(CodeBlock* genericCodeBlock)
{
    RELEASE_ASSERT(genericCodeBlock->ownerExecutable() == this);

    JSGlobalData& vm = *genericCodeBlock->vm();

    ASSERT(vm.heap.isDeferred());

    CodeSpecializationKind kind = genericCodeBlock->specializationKind();

    RefPtr<CodeBlock> oldCodeBlock;

    switch (kind) {
    case CodeForCall:
        m_jitCodeForCall = genericCodeBlock->jitCode();
        m_jitCodeForCallWithArityCheck = MacroAssemblerCodePtr();
        m_jitCodeForCallWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
        m_numParametersForCall = genericCodeBlock->numParameters();
        break;
    case CodeForConstruct:
        m_jitCodeForConstruct = genericCodeBlock->jitCode();
        m_jitCodeForConstructWithArityCheck = MacroAssemblerCodePtr();
        m_jitCodeForConstructWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
        m_numParametersForConstruct = genericCodeBlock->numParameters();
        break;
    }

    switch (genericCodeBlock->codeType()) {
    case GlobalCode: {
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        ProgramCodeBlock* codeBlock = static_cast<ProgramCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_programCodeBlock;
        executable->m_programCodeBlock = codeBlock;
        break;
    }

    case ModuleCode: {
        ModuleProgramExecutable* executable = jsCast<ModuleProgramExecutable*>(this);
        ModuleProgramCodeBlock* codeBlock = static_cast<ModuleProgramCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_moduleProgramCodeBlock;
        executable->m_moduleProgramCodeBlock = codeBlock;
        break;
    }

    case EvalCode: {
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        EvalCodeBlock* codeBlock = static_cast<EvalCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_evalCodeBlock;
        executable->m_evalCodeBlock = codeBlock;
        break;
    }

    case FunctionCode: {
        FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
        FunctionCodeBlock* codeBlock = static_cast<FunctionCodeBlock*>(genericCodeBlock);

        switch (kind) {
        case CodeForCall:
            oldCodeBlock = executable->m_codeBlockForCall;
            executable->m_codeBlockForCall = codeBlock;
            break;
        case CodeForConstruct:
            oldCodeBlock = executable->m_codeBlockForConstruct;
            executable->m_codeBlockForConstruct = codeBlock;
            break;
        }
        break;
    }
    }

    if (oldCodeBlock)
        oldCodeBlock->unlinkIncomingCalls();

    genericCodeBlock->setInstallTime(WTF::monotonicallyIncreasingTime());

    vm.heap.didInstallCode(this);
}

PassRefPtr<CodeBlock> ScriptExecutable::newCodeBlockFor(
	ExecState* exec, CodeSpecializationKind kind, JSFunction* jsFunction, JSScope* scope, JSObject*& exception)
{
    VM* vm = scope->vm();
    JSGlobalObject* globalObject = scope->globalObject();

    ASSERT(vm->heap.isDeferred());

    if (classInfo() == EvalExecutable::info()) {
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_evalCodeBlock);
        RELEASE_ASSERT(!jsFunction);

        if (!globalObject->evalEnabled()) {
        	exception = vm->throwException(exec, createEvalError(exec, globalObject->evalDisabledErrorMessage()));
        	return nullptr;
        }
        ParserError error;
        JSParserStrictMode strictMode = executable->isStrictMode() ? JSParserStrictMode::Strict : JSParserStrictMode::NotStrict;
        DerivedContextType derivedContextType = executable->derivedContextType();
        EvalContextType evalContextType = executable->evalContextType();

        OwnPtr<EvalNode> evalNode = parse<EvalNode>(
        	vm, m_source, Identifier(), JSParserBuiltinMode::NotBuiltin, strictMode, SourceParseMode::ProgramMode, SuperBinding::NotNeeded, error, ConstructorKind::None, derivedContextType, evalContextType);

        if (globalObject->hasDebugger() && !EvalNode::scopeIsFunction)
        	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!evalNode) {
            exception = error.toErrorObject(globalObject, m_source);
            return nullptr;
        }

        unsigned arrowContextFeature = executable->isArrowFunctionContext() ? ArrowFunctionContextFeature : 0;
		recordParse(evalNode->features() | arrowContextFeature, evalNode->hasCapturedVariables(), evalNode->lineNo(), evalNode->lastLine(), startColumn());

        RefPtr<EvalCodeBlock> evalCodeBlock = adoptRef(new EvalCodeBlock(executable, executable->executableInfo(), scope->globalObject(), source().provider()));
        OwnPtr<BytecodeGenerator> generator(adoptPtr(new BytecodeGenerator(evalNode.get(), scope->globalObject(), evalCodeBlock.get(), &executable->variablesUnderTDZ())));

        error = generator->generate(exec, scope);
        if (error.isValid()) {
            exception = error.toErrorObject(globalObject, m_source);
        	return nullptr;
        }

        return evalCodeBlock.release();
    }

    if (classInfo() == ProgramExecutable::info()) {
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_programCodeBlock);
        RELEASE_ASSERT(!jsFunction);

        ParserError error;
        JSParserStrictMode strictMode = executable->isStrictMode() ? JSParserStrictMode::Strict : JSParserStrictMode::NotStrict;
        DerivedContextType derivedContextType = executable->derivedContextType();
        OwnPtr<ProgramNode> programNode = parse<ProgramNode>(
        	vm, m_source, Identifier(), JSParserBuiltinMode::NotBuiltin, strictMode, SourceParseMode::ProgramMode, SuperBinding::NotNeeded, error, ConstructorKind::None, derivedContextType);

        if (globalObject->hasDebugger() && !ProgramNode::scopeIsFunction)
        	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!programNode) {
            exception = error.toErrorObject(globalObject, m_source);
            return nullptr;
        }

        unsigned arrowContextFeature = executable->isArrowFunctionContext() ? ArrowFunctionContextFeature : 0;
        recordParse(programNode->features() | arrowContextFeature, programNode->hasCapturedVariables(), programNode->lineNo(), programNode->lastLine(), startColumn());

        executable->initializeGlobalProperties(*vm, programNode.get(), exec, scope);

        RefPtr<ProgramCodeBlock> programCodeBlock = adoptRef(new ProgramCodeBlock(executable, GlobalCode, executable->executableInfo(), scope->globalObject(), source().provider(), source().startColumn()));
        VariableEnvironment emptyParentTDZVariables;
        OwnPtr<BytecodeGenerator> generator(adoptPtr(new BytecodeGenerator(programNode.get(), scope->globalObject(), programCodeBlock.get(), &emptyParentTDZVariables)));

        error = generator->generate(exec, scope);
        if (error.isValid()) {
            exception = error.toErrorObject(globalObject, m_source);
        	return nullptr;
        }

        return programCodeBlock.release();
    }

    if (classInfo() == ModuleProgramExecutable::info()) {
        ModuleProgramExecutable* executable = jsCast<ModuleProgramExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_moduleProgramCodeBlock);
        RELEASE_ASSERT(!jsFunction);

        ParserError error;
        JSParserStrictMode strictMode = JSParserStrictMode::Strict;
        DerivedContextType derivedContextType = executable->derivedContextType();
        OwnPtr<ModuleProgramNode> moduleProgramNode = parse<ModuleProgramNode>(
        	vm, m_source, Identifier(), JSParserBuiltinMode::NotBuiltin, strictMode, SourceParseMode::ModuleEvaluateMode, SuperBinding::NotNeeded, error, ConstructorKind::None, derivedContextType);

        if (globalObject->hasDebugger() && !ModuleProgramNode::scopeIsFunction)
        	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!moduleProgramNode) {
            exception = error.toErrorObject(globalObject, m_source);
            return nullptr;
        }

        unsigned arrowContextFeature = executable->isArrowFunctionContext() ? ArrowFunctionContextFeature : 0;
		recordParse(moduleProgramNode->features() | arrowContextFeature, moduleProgramNode->hasCapturedVariables(), moduleProgramNode->lineNo(), moduleProgramNode->lastLine(), startColumn());

	    VariableEnvironment emptyParentTDZVariables;
        RefPtr<ModuleProgramCodeBlock> moduleProgramCodeBlock = adoptRef(new ModuleProgramCodeBlock(executable, executable->executableInfo(), scope->globalObject(), source().provider(), source().startColumn()));
        OwnPtr<BytecodeGenerator> generator(adoptPtr(new BytecodeGenerator(moduleProgramNode.get(), scope->globalObject(), moduleProgramCodeBlock.get(), &emptyParentTDZVariables)));

        error = generator->generate(exec, scope);
        if (error.isValid()) {
            exception = error.toErrorObject(globalObject, m_source);
        	return nullptr;
        }

        executable->m_moduleEnvironmentSymbolTable.set(vm, executable, jsCast<SymbolTable*>(moduleProgramCodeBlock->constantRegister(moduleProgramCodeBlock->moduleEnvironmentSymbolTableConstantRegisterOffset()).get())->cloneScopePart(*vm));

        return moduleProgramCodeBlock.release();
    }

    RELEASE_ASSERT(classInfo() == FunctionExecutable::info());
    RELEASE_ASSERT(jsFunction);
    FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
    RELEASE_ASSERT(!executable->codeBlockFor(kind));

    exception = 0;
    ParserError error;
    JSParserBuiltinMode builtinMode = executable->isBuiltinFunction() ? JSParserBuiltinMode::Builtin : JSParserBuiltinMode::NotBuiltin;
    JSParserStrictMode strictMode = executable->isStrictMode() ? JSParserStrictMode::Strict : JSParserStrictMode::NotStrict;
    ASSERT(isFunctionParseMode(executable->parseMode()));
    OwnPtr<FunctionNode> function = parse<FunctionNode>(
        vm, m_source, executable->name(), builtinMode, strictMode, executable->parseMode(), executable->superBinding(), error);

    if (globalObject->hasDebugger() && !FunctionNode::scopeIsFunction)
    	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

    if (!function) {
        exception = error.toErrorObject(globalObject, m_source);
        return nullptr;
    }
    function->finishParsing(executable->m_name, executable->functionMode());
    recordParse(function->features(), function->hasCapturedVariables(), function->lineNo(), function->lastLine(), startColumn());

    bool isClassContext = executable->superBinding() == SuperBinding::Needed;

    SourceProvider* provider = source().provider();
    unsigned sourceOffset = source().startOffset();
    unsigned startColumn = source().startColumn();

    RefPtr<FunctionCodeBlock> functionCodeBlock = adoptRef(new FunctionCodeBlock(executable, FunctionCode, ExecutableInfo(function->usesEval(), function->isStrictMode(), kind == CodeForConstruct, executable->isBuiltinFunction(), executable->constructorKind(), executable->superBinding(), executable->parseMode(), executable->derivedContextType(), false, isClassContext, EvalContextType::FunctionEvalContext), scope->globalObject(), provider, sourceOffset, startColumn));

    VariableEnvironment parentScopeTDZVariables = executable->parentScopeTDZVariables();
	OwnPtr<BytecodeGenerator> generator(adoptPtr(new BytecodeGenerator(function.get(), scope->globalObject(), functionCodeBlock.get(), &parentScopeTDZVariables)));
    error = generator->generate(exec, scope);
    if (error.isValid()) {
        exception = error.toErrorObject(globalObject, m_source);
		return nullptr;
    }

	return functionCodeBlock.release();
}

PassRefPtr<CodeBlock> ScriptExecutable::newReplacementCodeBlockFor(
    CodeSpecializationKind kind)
{
    if (classInfo() == EvalExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        RefPtr<EvalCodeBlock> result = adoptRef(new EvalCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_evalCodeBlock));
        result->setAlternative(executable->m_evalCodeBlock);
        return result;
    }

    if (classInfo() == ProgramExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        RefPtr<ProgramCodeBlock> result = adoptRef(new ProgramCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_programCodeBlock));
        result->setAlternative(executable->m_programCodeBlock);
        return result;
    }

    if (classInfo() == ModuleProgramExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        ModuleProgramExecutable* executable = jsCast<ModuleProgramExecutable*>(this);
        RefPtr<ModuleProgramCodeBlock> result = adoptRef(new ModuleProgramCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_moduleProgramCodeBlock));
        result->setAlternative(executable->m_moduleProgramCodeBlock);
        return result;
    }

    RELEASE_ASSERT(classInfo() == FunctionExecutable::info());
    FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
    RefPtr<FunctionCodeBlock> result = adoptRef(new FunctionCodeBlock(
        CodeBlock::CopyParsedBlock, *executable->codeBlockFor(kind)));
    result->setAlternative(executable->codeBlockFor(kind));
    return result;
}

JSObject* ScriptExecutable::prepareForExecutionImpl(ExecState* exec, JSFunction* jsFunction, JSScope* scope, CodeSpecializationKind kind, CodeBlock*& resultCodeBlock)
{
    VM& vm = exec->vm();
    DeferGCForAWhile deferGC(vm.heap);

    JSObject* exception = 0;
    RefPtr<CodeBlock> codeBlock = newCodeBlockFor(exec, kind, jsFunction, scope, exception);
    resultCodeBlock = codeBlock.get();
    if (!codeBlock) {
        RELEASE_ASSERT(exception);
        if(!exception)
        	exception = createOutOfMemoryError(exec);
        return exception;
    }

#if ENABLE(LLINT)
    LLInt::setEntrypoint(vm, codeBlock.get());
#else
    CompilationResult result = JIT::compile(&vm, codeBlock.get(), JITCompilationCanFail);
    RELEASE_ASSERT(result == CompilationSuccessful);
#endif

    if(result == CompilationSuccessful) {
    	installCode(codeBlock.get());
    	return 0;
    }

    return vm.throwException(exec, createOutOfMemoryError(exec));
}

const ClassInfo ExecutableBase::s_info = { "Executable", 0, 0, CREATE_METHOD_TABLE(ExecutableBase) };

void ExecutableBase::destroy(JSCell* cell)
{
    static_cast<ExecutableBase*>(cell)->ExecutableBase::~ExecutableBase();
}

void ExecutableBase::clearCode()
{
#if ENABLE(JIT)
    m_jitCodeForCall.clear();
    m_jitCodeForConstruct.clear();
    m_jitCodeForCallWithArityCheck = MacroAssemblerCodePtr();
    m_jitCodeForConstructWithArityCheck = MacroAssemblerCodePtr();
    m_jitCodeForCallWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
    m_jitCodeForConstructWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
#endif
    m_numParametersForCall = NUM_PARAMETERS_NOT_COMPILED;
    m_numParametersForConstruct = NUM_PARAMETERS_NOT_COMPILED;
}

#if ENABLE(DFG_JIT)
Intrinsic ExecutableBase::intrinsic() const
{
    if (const NativeExecutable* nativeExecutable = jsDynamicCast<const NativeExecutable*>(this))
        return nativeExecutable->intrinsic();
    return NoIntrinsic;
}
#else
Intrinsic ExecutableBase::intrinsic() const
{
    return NoIntrinsic;
}
#endif

const ClassInfo NativeExecutable::s_info = { "NativeExecutable", &ExecutableBase::s_info, 0, CREATE_METHOD_TABLE(NativeExecutable) };

void NativeExecutable::destroy(JSCell* cell)
{
    static_cast<NativeExecutable*>(cell)->NativeExecutable::~NativeExecutable();
}

#if ENABLE(DFG_JIT)
Intrinsic NativeExecutable::intrinsic() const
{
    return m_intrinsic;
}
#endif

const ClassInfo ScriptExecutable::s_info = { "ScriptExecutable", &ExecutableBase::s_info, 0, CREATE_METHOD_TABLE(ScriptExecutable) };

ScriptExecutable::ScriptExecutable(Structure* structure, JSGlobalData& globalData, const SourceCode& source, bool isInStrictContext, DerivedContextType derivedContextType, bool isInArrowFunctionContext, EvalContextType evalContextType)
: ExecutableBase(globalData, structure, NUM_PARAMETERS_NOT_COMPILED)
, m_features(isInStrictContext ? StrictModeFeature : 0)
, m_didTryToEnterInLoop(false)
, m_hasCapturedVariables(false)
, m_neverInline(false)
, m_isArrowFunctionContext(isInArrowFunctionContext)
, m_derivedContextType(static_cast<unsigned>(derivedContextType))
, m_evalContextType(static_cast<unsigned>(evalContextType))
, m_firstLine(-1)
, m_lastLine(-1)
, m_startColumn(UINT_MAX)
, m_source(source)
{
}

void ScriptExecutable::destroy(JSCell* cell)
{
    static_cast<ScriptExecutable*>(cell)->ScriptExecutable::~ScriptExecutable();
}

const ClassInfo EvalExecutable::s_info = { "EvalExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(EvalExecutable) };

EvalExecutable::EvalExecutable(JSGlobalData& vm, const SourceCode& source, bool inStrictContext, DerivedContextType derivedContextType, bool isArrowFunctionContext, EvalContextType evalContextType, const VariableEnvironment& variablesUnderTDZ)
    : ScriptExecutable(vm.evalExecutableStructure.get(), vm, source, inStrictContext, derivedContextType, isArrowFunctionContext, evalContextType)
	, m_variablesUnderTDZ(variablesUnderTDZ)
{
}

void EvalExecutable::destroy(JSCell* cell)
{
    static_cast<EvalExecutable*>(cell)->EvalExecutable::~EvalExecutable();
}

unsigned EvalExecutable::numVariables()
{
	return m_evalCodeBlock->numVariables();
}

unsigned EvalExecutable::numberOfFunctionDecls()
{
	return m_evalCodeBlock->numberOfFunctionDecls();
}

const ClassInfo ProgramExecutable::s_info = { "ProgramExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(ProgramExecutable) };

ProgramExecutable::ProgramExecutable(JSGlobalData& vm, const SourceCode& source)
    : ScriptExecutable(vm.programExecutableStructure.get(), vm, source, false, DerivedContextType::None, false, EvalContextType::None)
{
}

void ProgramExecutable::destroy(JSCell* cell)
{
    static_cast<ProgramExecutable*>(cell)->ProgramExecutable::~ProgramExecutable();
}

const ClassInfo ModuleProgramExecutable::s_info = { "ModuleProgramExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(ModuleProgramExecutable) };

ModuleProgramExecutable::ModuleProgramExecutable(JSGlobalData& vm, const SourceCode& source)
    : ScriptExecutable(vm.moduleProgramExecutableStructure.get(), vm, source, false, DerivedContextType::None, false, EvalContextType::None)
{
#if 0
    m_typeProfilingStartOffset = 0;
    m_typeProfilingEndOffset = source.length() - 1;
    if (exec->vm().typeProfiler() || exec->vm().controlFlowProfiler())
        exec->vm().functionHasExecutedCache()->insertUnexecutedRange(sourceID(), m_typeProfilingStartOffset, m_typeProfilingEndOffset);
#endif
}

ModuleProgramExecutable* ModuleProgramExecutable::create(JSGlobalData& vm, const SourceCode& source)
{
    ModuleProgramExecutable* executable = new (NotNull, allocateCell<ModuleProgramExecutable>(vm.heap)) ModuleProgramExecutable(vm, source);
    executable->finishCreation(vm);
    return executable;
}

void ModuleProgramExecutable::destroy(JSCell* cell)
{
    static_cast<ModuleProgramExecutable*>(cell)->ModuleProgramExecutable::~ModuleProgramExecutable();
}

const ClassInfo FunctionExecutable::s_info = { "FunctionExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(FunctionExecutable) };

FunctionExecutable* FunctionExecutable::create(
	JSGlobalData& globalData, FunctionMetadataNode* node, unsigned startColumn, bool isBuiltinFunction, ConstructAbility constructAbility, VariableEnvironment& parentScopeTDZVariables, DerivedContextType derivedContextType)
{
	FunctionExecutable* executable = new (NotNull, allocateCell<FunctionExecutable>(globalData.heap)) FunctionExecutable(globalData, node, startColumn, isBuiltinFunction, constructAbility, parentScopeTDZVariables, derivedContextType);
	executable->finishCreation(globalData);
	return executable;
}

FunctionExecutable::FunctionExecutable(JSGlobalData& vm, FunctionMetadataNode* node, unsigned startColumn, bool isBuiltinFunction, ConstructAbility constructAbility, VariableEnvironment& parentScopeTDZVariables, DerivedContextType derivedContextType)
    : ScriptExecutable(vm.functionExecutableStructure.get(), vm, node->source(), node->isInStrictContext(), derivedContextType, false, EvalContextType::None)
    , m_classSource(node->classSource())
	, m_parameterCount(node->parameterCount())
	, m_parseMode(node->parseMode())
	, m_functionMode(static_cast<unsigned>(node->functionMode()))
	, m_isBuiltinFunction(isBuiltinFunction)
	, m_constructorKind(static_cast<unsigned>(node->constructorKind()))
	, m_constructAbility(static_cast<unsigned>(constructAbility))
	, m_superBinding(static_cast<unsigned>(node->superBinding()))
    , m_name(node->ident())
	, m_ecmaName(node->ecmaName())
    , m_inferredName(node->inferredName().isNull() ? vm.propertyNames->emptyIdentifier : node->inferredName())
	, m_didParseForTheFirstTime(false)
	, m_parentScopeTDZVariables(vm.m_compactVariableMap->get(parentScopeTDZVariables))
{
	m_firstLine = node->lineNo();
	m_lastLine = node->lastLine();
	m_startColumn = startColumn;
    m_parametersStartOffset = node->parametersStart();

    // Make sure these bitfields are adequately wide.
    ASSERT(m_constructAbility == static_cast<unsigned>(constructAbility));
    ASSERT(m_constructorKind == static_cast<unsigned>(node->constructorKind()));
    ASSERT(m_functionMode == static_cast<unsigned>(node->functionMode()));
    ASSERT(m_superBinding == static_cast<unsigned>(node->superBinding()));
    ASSERT(m_derivedContextType == static_cast<unsigned>(derivedContextType));
    ASSERT(m_parseMode == static_cast<unsigned>(node->parseMode()));
}

FunctionCodeBlock* FunctionExecutable::codeBlockFor(CodeSpecializationKind kind)
{
	if (kind == CodeForCall)
		return m_codeBlockForCall.get();
	ASSERT(kind == CodeForConstruct);
	return m_codeBlockForConstruct.get();
}

void FunctionExecutable::finishCreation(JSGlobalData& vm)
{
	Base::finishCreation(vm);
    m_singletonFunction.set(vm, this, InferredValue::create(vm));
}

void FunctionExecutable::destroy(JSCell* cell)
{
    static_cast<FunctionExecutable*>(cell)->FunctionExecutable::~FunctionExecutable();
}

inline const char* samplingDescription(JITCode::JITType jitType)
{
    switch (jitType) {
    case JITCode::InterpreterThunk:
        return "Interpreter Compilation (TOTAL)";
    case JITCode::BaselineJIT:
        return "Baseline Compilation (TOTAL)";
    case JITCode::DFGJIT:
        return "DFG Compilation (TOTAL)";
    case JITCode::FTLJIT:
        return "FTL Compilation (TOTAL)";
    default:
        ASSERT_NOT_REACHED();
        return 0;
    }
}

EvalExecutable* EvalExecutable::create(JSGlobalData& vm, const SourceCode& source, bool isInStrictContext, DerivedContextType derivedContextType, bool isArrowFunctionContext, EvalContextType evalContextType, const VariableEnvironment& variablesUnderTDZ)
{
	EvalExecutable* executable = new (NotNull, allocateCell<EvalExecutable>(vm.heap)) EvalExecutable(vm, source, isInStrictContext, derivedContextType, isArrowFunctionContext, evalContextType, variablesUnderTDZ);
	executable->finishCreation(vm);
	return executable;
}

void EvalExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    EvalExecutable* thisObject = jsCast<EvalExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    if (thisObject->m_evalCodeBlock)
        thisObject->m_evalCodeBlock->visitAggregate(visitor);
}

void EvalExecutable::clearCode()
{
    m_evalCodeBlock.clear();
    Base::clearCode();
}

// http://www.ecma-international.org/ecma-262/6.0/index.html#sec-hasrestrictedglobalproperty
static bool hasRestrictedGlobalProperty(ExecState* exec, JSGlobalObject* globalObject, PropertyName propertyName)
{
    PropertyDescriptor descriptor;
    if (!globalObject->getOwnPropertyDescriptor(exec, propertyName, descriptor))
        return false;
    if (descriptor.configurable())
        return false;
    return true;
}

JSObject* ProgramExecutable::initializeGlobalProperties(VM& vm, ProgramNode* programNode, CallFrame* callFrame, JSScope* scope)
{
    RELEASE_ASSERT(scope);
    JSGlobalObject* globalObject = scope->globalObject();
    RELEASE_ASSERT(globalObject);
    ASSERT(&globalObject->vm() == &vm);

    JSValue nextPrototype = globalObject->getPrototypeDirect();
    while (nextPrototype && nextPrototype.isObject()) {
        if (UNLIKELY(asObject(nextPrototype)->structure()->typeInfo().type() == ProxyObjectType)) {
            ExecState* exec = globalObject->globalExec();
            return createTypeError(exec, ASCIILiteral("Proxy is not allowed in the global prototype chain."));
        }
        nextPrototype = asObject(nextPrototype)->getPrototypeDirect();
    }

    JSGlobalLexicalEnvironment* globalLexicalEnvironment = globalObject->globalLexicalEnvironment();
    const VariableEnvironment& variableDeclarations = programNode->varDeclarations();
    const VariableEnvironment& lexicalDeclarations = programNode->lexicalVariables();
    // The ES6 spec says that no vars/global properties/let/const can be duplicated in the global scope.
    // This carried out section 15.1.8 of the ES6 spec: http://www.ecma-international.org/ecma-262/6.0/index.html#sec-globaldeclarationinstantiation
    {
        ExecState* exec = globalObject->globalExec();
        // Check for intersection of "var" and "let"/"const"/"class"
        for (auto it = lexicalDeclarations.begin(); it != lexicalDeclarations.end(); ++it) {
            if (variableDeclarations.contains(it->first))
                return createSyntaxError(exec, makeString("Can't create duplicate variable: '", String(it->first.get()), "'"));
        }

        // Check if any new "let"/"const"/"class" will shadow any pre-existing global property names, or "var"/"let"/"const" variables.
        // It's an error to introduce a shadow.
        for (auto it = lexicalDeclarations.begin(); it != lexicalDeclarations.end(); ++it) {
            // The ES6 spec says that RestrictedGlobalProperty can't be shadowed.
            if (hasRestrictedGlobalProperty(exec, globalObject, reinterpret_cast<AtomicStringImpl*>(it->first.get())))
				return createSyntaxError(exec, makeString("Can't create duplicate variable that shadows a global property: '", String(it->first.get()), "'"));

            bool hasProperty = globalLexicalEnvironment->hasProperty(exec, reinterpret_cast<AtomicStringImpl*>(it->first.get()));
            RETURN_IF_EXCEPTION(vm, vm.exception());
            if (hasProperty) {
                if (UNLIKELY(it->second.isConst() && !vm.globalConstRedeclarationShouldThrow() && !isStrictMode())) {
                    // We only allow "const" duplicate declarations under this setting.
                    // For example, we don't "let" variables to be overridden by "const" variables.
                    if (globalLexicalEnvironment->isConstVariable(it->first.get()))
                        continue;
                }
                return createSyntaxError(exec, makeString("Can't create duplicate variable: '", String(it->first.get()), "'"));
            }
        }

        // Check if any new "var"s will shadow any previous "let"/"const"/"class" names.
        // It's an error to introduce a shadow.
        if (!globalLexicalEnvironment->isEmpty()) {
            for (auto it = variableDeclarations.begin(); it != variableDeclarations.end(); ++it) {
            	bool hasProperty = globalLexicalEnvironment->hasProperty(exec, reinterpret_cast<AtomicStringImpl*>(it->first.get()));
            	RETURN_IF_EXCEPTION(vm, vm.exception());
				if (hasProperty)
                    return createSyntaxError(exec, makeString("Can't create duplicate variable: '", String(it->first.get()), "'"));
            }
        }
    }

    BatchedTransitionOptimizer optimizer(vm, globalObject);

    const DeclarationStacks::FunctionStack& functionStack = programNode->functionStack();
    for (size_t i = 0; i < functionStack.size(); ++i) {
    	FunctionMetadataNode* function = functionStack[i];
        globalObject->addFunction(callFrame, function->ident());
    }

    for (auto it = variableDeclarations.begin(); it != variableDeclarations.end(); ++it) {
    	ASSERT(it->second.isVar());
		globalObject->addVar(callFrame, Identifier::fromUid(vm, it->first.get()));
        ASSERT(!vm.exception());
	}

    {
        JSGlobalLexicalEnvironment* globalLexicalEnvironment = jsCast<JSGlobalLexicalEnvironment*>(globalObject->globalScope());
        SymbolTable* symbolTable = globalLexicalEnvironment->symbolTable();
        //ConcurrentJITLocker locker(symbolTable->m_lock);
        for (auto it = lexicalDeclarations.begin(); it != lexicalDeclarations.end(); ++it) {
            if (UNLIKELY(it->second.isConst() && !vm.globalConstRedeclarationShouldThrow() && !isStrictMode())) {
                if (symbolTable->contains(it->first.get()))
                    continue;
            }
            ScopeOffset offset = symbolTable->takeNextScopeOffset();
            SymbolTableEntry newEntry(VarOffset(offset), it->second.isConst() ? ReadOnly : 0);
            newEntry.prepareToWatch();
            symbolTable->add(it->first.get(), newEntry);

            ScopeOffset offsetForAssert = globalLexicalEnvironment->addVariables(1, jsTDZValue());
            RELEASE_ASSERT(offsetForAssert == offset);
        }
    }
    return 0;
}

void ProgramExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    ProgramExecutable* thisObject = jsCast<ProgramExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    if (thisObject->m_programCodeBlock)
        thisObject->m_programCodeBlock->visitAggregate(visitor);
}

void ProgramExecutable::clearCode()
{
    m_programCodeBlock.clear();
    Base::clearCode();
}

void ModuleProgramExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    ModuleProgramExecutable* thisObject = jsCast<ModuleProgramExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    ScriptExecutable::visitChildren(thisObject, visitor);
    visitor.append(&thisObject->m_moduleEnvironmentSymbolTable);
    if (thisObject->m_moduleProgramCodeBlock)
        thisObject->m_moduleProgramCodeBlock->visitAggregate(visitor);
}

void ModuleProgramExecutable::clearCode()
{
    m_moduleProgramCodeBlock.clear();
    m_moduleEnvironmentSymbolTable.clear();
    Base::clearCode();
}

FunctionCodeBlock* FunctionExecutable::baselineCodeBlockFor(CodeSpecializationKind kind)
{
    FunctionCodeBlock* result;
    if (kind == CodeForCall)
        result = m_codeBlockForCall.get();
    else {
        ASSERT(kind == CodeForConstruct);
        result = m_codeBlockForConstruct.get();
    }
    if (!result)
        return 0;
    return static_cast<FunctionCodeBlock*>(result->baselineAlternative());
}

void FunctionExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    FunctionExecutable* thisObject = jsCast<FunctionExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    if (thisObject->m_codeBlockForCall)
        thisObject->m_codeBlockForCall->visitAggregate(visitor);
    if (thisObject->m_codeBlockForConstruct)
        thisObject->m_codeBlockForConstruct->visitAggregate(visitor);
    visitor.append(&thisObject->m_singletonFunction);
}

void FunctionExecutable::clearCode()
{
    m_codeBlockForCall.clear();
    m_codeBlockForConstruct.clear();
    Base::clearCode();
}

// FIXME: There's no need to add the function's name to the key here. It's already in the source code.
FunctionExecutable* FunctionExecutable::fromGlobalCode(const Identifier& name, ExecState* exec, Debugger* debugger, const SourceCode& source, JSObject** exception)
{
	JSGlobalData& vm = exec->vm();
    JSGlobalObject* lexicalGlobalObject = exec->lexicalGlobalObject();
    ParserError error;
    OwnPtr<ProgramNode> program = parse<ProgramNode>(
    	&vm, source, Identifier(), JSParserBuiltinMode::NotBuiltin,
		JSParserStrictMode::NotStrict, SourceParseMode::ProgramMode, SuperBinding::NotNeeded, error);

    if (lexicalGlobalObject->hasDebugger() && !ProgramNode::scopeIsFunction)
    	lexicalGlobalObject->debugger()->sourceParsed(exec, source.provider(), error.line(), error.message());

    if (!program) {
        *exception = error.toErrorObject(lexicalGlobalObject, source);
        return 0;
    }

    // This function assumes an input string that would result in a single function declaration.
    StatementNode* statement = program->singleStatement();
    if (UNLIKELY(!statement)) {
		JSToken token;
		error = ParserError(ParserError::SyntaxError, ParserError::SyntaxErrorIrrecoverable, token, "Parser error", -1);
		return 0;
	}
    ASSERT(statement->isBlock());

    StatementNode* funcDecl = static_cast<BlockNode*>(statement)->singleStatement();
    if (UNLIKELY(!funcDecl)) {
		JSToken token;
		error = ParserError(ParserError::SyntaxError, ParserError::SyntaxErrorIrrecoverable, token, "Parser error", -1);
		return 0;
	}
    ASSERT(funcDecl->isFuncDeclNode());

    FunctionMetadataNode* metadata = static_cast<FuncDeclNode*>(funcDecl)->metadata();
    ASSERT(metadata);
    if (!metadata)
        return 0;

    metadata->overrideName(name);
    unsigned startColumn = source.startColumn();
    // The Function constructor only has access to global variables, so no variables will be under TDZ.
	VariableEnvironment emptyTDZVariables;
    FunctionExecutable* functionExecutable = FunctionExecutable::create(vm, metadata, startColumn, false, ConstructAbility::CanConstruct, emptyTDZVariables, DerivedContextType::None);

    return functionExecutable;
}

}
