/*
 * Copyright (C) 2009, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef Executable_h
#define Executable_h

#include "ArityCheckMode.h"
#include "CallData.h"
#include "CodeSpecializationKind.h"
#include "CompilationResult.h"
#include "ConstructAbility.h"
//#include "DFGPlan.h"
#include "ExecutableInfo.h"
#include "InferredValue.h"
#include "JITCode.h"
#include "JSGlobalObject.h"
#include "RegisterPreservationMode.h"
#include "Nodes.h"
#include "SamplingTool.h"
#include "SourceCode.h"
#include <wtf/DoublyLinkedList.h>
#include <wtf/PassOwnPtr.h>

namespace JSC {

class BuiltinExecutables;
class CodeBlock;
class Debugger;
class EvalCodeBlock;
class FunctionCodeBlock;
class Heap;
class LLIntOffsetsExtractor;
class JSModuleRecord;
class ProgramCodeBlock;
class ModuleProgramCodeBlock;
class JSScope;

enum CompilationKind { FirstCompilation, OptimizingCompilation };

inline bool isCall(CodeSpecializationKind kind)
{
	if (kind == CodeForCall)
		return true;
	ASSERT(kind == CodeForConstruct);
	return false;
}

class ExecutableBase : public JSCell {
	friend class JIT;
	friend class DoublyLinkedList<ExecutableBase>;
	friend class Heap;

private:
    void setPrev(ExecutableBase* prev) { m_prev = prev; }
    void setNext(ExecutableBase* next) { m_next = next; }

    ExecutableBase* prev() const { return m_prev; }
    ExecutableBase* next() const { return m_next; }

	ExecutableBase* m_prev;
	ExecutableBase* m_next;

protected:
	static const int NUM_PARAMETERS_IS_HOST = 0;
	static const int NUM_PARAMETERS_NOT_COMPILED = -1;

	ExecutableBase(JSGlobalData& globalData, Structure* structure, int numParameters)
		: JSCell(globalData, structure)
		, m_prev(0)
		, m_next(0)
		, m_numParametersForCall(numParameters)
		, m_numParametersForConstruct(numParameters)
		, m_canDFGCompileCall(true)
		, m_canDFGCompileConstruct(true)
	{
	}

	void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
	}

public:
	typedef JSCell Base;

	static const bool needsDestruction = true;
	static const bool hasImmortalStructure = true;
	static void destroy(JSCell*);

    bool isEvalExecutable() const
    {
        return structure()->typeInfo().type() == EvalExecutableType;
    }
    bool isFunctionExecutable() const
    {
        return structure()->typeInfo().type() == FunctionExecutableType;
    }
    bool isProgramExecutable() const
    {
        return structure()->typeInfo().type() == ProgramExecutableType;
    }
    bool isModuleProgramExecutable()
    {
        return structure()->typeInfo().type() == ModuleProgramExecutableType;
    }

	bool isHostFunction() const
	{
		ASSERT((m_numParametersForCall == NUM_PARAMETERS_IS_HOST) == (m_numParametersForConstruct == NUM_PARAMETERS_IS_HOST));
		return m_numParametersForCall == NUM_PARAMETERS_IS_HOST;
	}

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto) { return Structure::create(globalData, globalObject, proto, TypeInfo(CellType, StructureFlags), info()); }

	void clearCode();

	DECLARE_EXPORT_INFO;

protected:
	static const unsigned StructureFlags = 0;
	int m_numParametersForCall;
	int m_numParametersForConstruct;

public:
	static void clearCodeVirtual(ExecutableBase*);

#if ENABLE(JIT)
	PassRefPtr<JITCode> generatedJITCodeForCall()
	{
		ASSERT(m_jitCodeForCall);
		return m_jitCodeForCall;
	}

	PassRefPtr<JITCode> generatedJITCodeForConstruct()
	{
		ASSERT(m_jitCodeForConstruct);
		return m_jitCodeForConstruct;
	}

	PassRefPtr<JITCode> generatedJITCodeFor(CodeSpecializationKind kind)
	{
		if (kind == CodeForCall)
			return generatedJITCodeForCall();
		ASSERT(kind == CodeForConstruct);
		return generatedJITCodeForConstruct();
	}

    MacroAssemblerCodePtr entrypointFor(
        VM& vm, CodeSpecializationKind kind, ArityCheckMode arity, RegisterPreservationMode registers)
    {
        // Check if we have a cached result. We only have it for arity check because we use the
        // no-arity entrypoint in non-virtual calls, which will "cache" this value directly in
        // machine code.
        if (arity == MustCheckArity) {
            switch (kind) {
            case CodeForCall:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    if (MacroAssemblerCodePtr result = m_jitCodeForCallWithArityCheck)
                        return result;
                    break;
                case MustPreserveRegisters:
                    if (MacroAssemblerCodePtr result = m_jitCodeForCallWithArityCheckAndPreserveRegs)
                        return result;
                    break;
                }
                break;
            case CodeForConstruct:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    if (MacroAssemblerCodePtr result = m_jitCodeForConstructWithArityCheck)
                        return result;
                    break;
                case MustPreserveRegisters:
                    if (MacroAssemblerCodePtr result = m_jitCodeForConstructWithArityCheckAndPreserveRegs)
                        return result;
                    break;
                }
                break;
            }
        }
        MacroAssemblerCodePtr result =
            generatedJITCodeFor(kind)->addressForCall(vm, this, arity, registers);
        if (arity == MustCheckArity) {
            // Cache the result; this is necessary for the JIT's virtual call optimizations.
            switch (kind) {
            case CodeForCall:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    m_jitCodeForCallWithArityCheck = result;
                    break;
                case MustPreserveRegisters:
                    m_jitCodeForCallWithArityCheckAndPreserveRegs = result;
                    break;
                }
                break;
            case CodeForConstruct:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    m_jitCodeForConstructWithArityCheck = result;
                    break;
                case MustPreserveRegisters:
                    m_jitCodeForConstructWithArityCheckAndPreserveRegs = result;
                    break;
                }
                break;
            }
        }
        return result;
    }

    static ptrdiff_t offsetOfJITCodeWithArityCheckFor(
        CodeSpecializationKind kind, RegisterPreservationMode registers)
    {
        switch (kind) {
        case CodeForCall:
            switch (registers) {
            case RegisterPreservationNotRequired:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForCallWithArityCheck);
            case MustPreserveRegisters:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForCallWithArityCheckAndPreserveRegs);
            }
        case CodeForConstruct:
            switch (registers) {
            case RegisterPreservationNotRequired:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForConstructWithArityCheck);
            case MustPreserveRegisters:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForConstructWithArityCheckAndPreserveRegs);
            }
        }
        RELEASE_ASSERT_NOT_REACHED();
        return 0;
    }

    static ptrdiff_t offsetOfNumParametersFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return OBJECT_OFFSETOF(ExecutableBase, m_numParametersForCall);
        ASSERT(kind == CodeForConstruct);
        return OBJECT_OFFSETOF(ExecutableBase, m_numParametersForConstruct);
    }
#endif // ENABLE(JIT)

	bool hasJITCodeForCall() const
	{
		return m_numParametersForCall >= 0;
	}

	bool hasJITCodeForConstruct() const
	{
		return m_numParametersForConstruct >= 0;
	}

	bool hasJITCodeFor(CodeSpecializationKind kind) const
	{
		if (kind == CodeForCall)
			return hasJITCodeForCall();
		ASSERT(kind == CodeForConstruct);
		return hasJITCodeForConstruct();
	}

	// Intrinsics are only for calls, currently.
	Intrinsic intrinsic() const;

	Intrinsic intrinsicFor(CodeSpecializationKind kind) const
	{
		if (isCall(kind))
			return intrinsic();
		return NoIntrinsic;
	}

	bool canDFGCompileCall() const { return m_canDFGCompileCall; }
	bool canDFGCompileConstruct() const { return m_canDFGCompileConstruct; }

	void disableDFGCompileCall()
	{
		m_canDFGCompileCall = false;
	}

	void disableDFGCompileConstruct()
	{
		m_canDFGCompileConstruct = false;
	}

protected:
#if ENABLE(JIT)
	RefPtr<JITCode> m_jitCodeForCall;
	RefPtr<JITCode> m_jitCodeForConstruct;
	MacroAssemblerCodePtr m_jitCodeForCallWithArityCheck;
	MacroAssemblerCodePtr m_jitCodeForConstructWithArityCheck;
    MacroAssemblerCodePtr m_jitCodeForCallWithArityCheckAndPreserveRegs;
    MacroAssemblerCodePtr m_jitCodeForConstructWithArityCheckAndPreserveRegs;
#endif

    bool m_canDFGCompileCall;
    bool m_canDFGCompileConstruct;
};

class NativeExecutable : public ExecutableBase {
	friend class JIT;
	friend class LLIntOffsetsExtractor;
public:
	typedef ExecutableBase Base;

#if ENABLE(JIT)
	static NativeExecutable* create(JSGlobalData& globalData, MacroAssemblerCodeRef callThunk, NativeFunction function, MacroAssemblerCodeRef constructThunk, NativeFunction constructor, Intrinsic intrinsic, const String& name)
	{
		NativeExecutable* executable;
		if (!callThunk) {
			executable = new (NotNull, allocateCell<NativeExecutable>(globalData.heap)) NativeExecutable(globalData, function, constructor);
			executable->finishCreation(globalData, nullptr, nullptr, intrinsic, name);
		} else {
			executable = new (NotNull, allocateCell<NativeExecutable>(globalData.heap)) NativeExecutable(globalData, function, constructor);
			executable->finishCreation(globalData, JITCode::hostFunction(callThunk), JITCode::hostFunction(constructThunk), intrinsic, name);
		}
		return executable;
	}
#endif

	static void destroy(JSCell*);

	NativeFunction function() { return m_function; }
	NativeFunction constructor() { return m_constructor; }

    NativeFunction nativeFunctionFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return function();
        ASSERT(kind == CodeForConstruct);
        return constructor();
    }

    static ptrdiff_t offsetOfNativeFunctionFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return OBJECT_OFFSETOF(NativeExecutable, m_function);
        ASSERT(kind == CodeForConstruct);
        return OBJECT_OFFSETOF(NativeExecutable, m_constructor);
    }

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto) { return Structure::create(globalData, globalObject, proto, TypeInfo(CellType, StructureFlags), info()); }

	DECLARE_INFO;

	Intrinsic intrinsic() const;

    const String& name() const { return m_name; }

protected:
#if ENABLE(JIT)
	void finishCreation(JSGlobalData& globalData, PassRefPtr<JITCode> callThunk, PassRefPtr<JITCode> constructThunk, Intrinsic intrinsic, const String& name)
	{
		Base::finishCreation(globalData);
		m_jitCodeForCall = callThunk;
		m_jitCodeForConstruct = constructThunk;
		m_intrinsic = intrinsic;
        m_name = name;
	}
#endif

private:
	NativeExecutable(JSGlobalData& globalData, NativeFunction function, NativeFunction constructor)
		: ExecutableBase(globalData, globalData.nativeExecutableStructure.get(), NUM_PARAMETERS_IS_HOST)
		, m_function(function)
		, m_constructor(constructor)
		, m_intrinsic(NoIntrinsic)
	{
	}

	NativeFunction m_function;
	NativeFunction m_constructor;

	Intrinsic m_intrinsic;

    String m_name;
};

class ScriptExecutable : public ExecutableBase {
public:
	typedef ExecutableBase Base;

	static void destroy(JSCell*);

	const SourceCode& source() { return m_source; }
	intptr_t sourceID() const { return m_source.providerID(); }
	const String& sourceURL() const { return m_source.provider()->url(); }
	int lineNo() const { return m_firstLine; }
	int lastLine() const { return m_lastLine; }
    unsigned startColumn() const { return m_startColumn; }

	bool usesEval() const { return m_features & EvalFeature; }
	bool usesArguments() const { return m_features & ArgumentsFeature; }
	bool isArrowFunctionContext() const { return m_isArrowFunctionContext; }
	bool isStrictMode() const { return m_features & StrictModeFeature; }
    DerivedContextType derivedContextType() const { return static_cast<DerivedContextType>(m_derivedContextType); }
    EvalContextType evalContextType() const { return static_cast<EvalContextType>(m_evalContextType); }

    void setNeverInline(bool value) { m_neverInline = value; }
    void setDidTryToEnterInLoop(bool value) { m_didTryToEnterInLoop = value; }
    bool neverInline() const { return m_neverInline; }
    bool didTryToEnterInLoop() const { return m_didTryToEnterInLoop; }
    bool isInliningCandidate() const { return !neverInline(); }

    bool* addressOfDidTryToEnterInLoop() { return &m_didTryToEnterInLoop; }

    CodeFeatures features() const { return m_features; }

    DECLARE_INFO;

    void recordParse(CodeFeatures features, bool hasCapturedVariables, int firstLine, int lastLine, unsigned startColumn)
    {
        m_features = features;
        m_hasCapturedVariables = hasCapturedVariables;
        m_firstLine = firstLine;
        m_lastLine = lastLine;
        m_startColumn = startColumn;
    }

    void installCode(CodeBlock*);
    PassRefPtr<CodeBlock> newCodeBlockFor(ExecState*, CodeSpecializationKind, JSFunction*, JSScope*, JSObject*& exception);
    PassRefPtr<CodeBlock> newReplacementCodeBlockFor(CodeSpecializationKind);

    // This function has an interesting GC story. Callers of this function are asking us to create a CodeBlock
	// that is not jettisoned before this function returns. Callers are essentially asking for a strong reference
	// to the CodeBlock. Because the Executable may be allocating the CodeBlock, we require callers to pass in
	// their CodeBlock*& reference because it's safe for CodeBlock to be jettisoned if Executable is the only thing
	// to point to it. This forces callers to have a CodeBlock* in a register or on the stack that will be marked
	// by conservative GC if a GC happens after we create the CodeBlock.
    template <typename ExecutableType>
    JSObject* prepareForExecution(ExecState* exec, JSFunction* function, JSScope* scope, CodeSpecializationKind kind, CodeBlock*& resultCodeBlock);

private:
    JSObject* prepareForExecutionImpl(ExecState*, JSFunction* function, JSScope* scope, CodeSpecializationKind, CodeBlock*& resultCodeBlock);

protected:
	ScriptExecutable(Structure*, JSGlobalData&, const SourceCode&, bool isInStrictContext, DerivedContextType, bool isInArrowFunctionContext, EvalContextType);

	void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
		globalData.heap.addExecutable(this); // Balanced by Heap::deleteUnmarkedCompiledCode().

#if ENABLE(CODEBLOCK_SAMPLING)
		if (SamplingTool* sampler = globalData.interpreter->sampler())
			sampler->notifyOfScope(globalData, this);
#endif
	}

	CodeFeatures m_features;
    bool m_didTryToEnterInLoop;
	bool m_hasCapturedVariables : 1;
    bool m_neverInline : 1;
    bool m_isArrowFunctionContext : 1;
    unsigned m_derivedContextType : 2;
    unsigned m_evalContextType : 2; // EvalContextType

	int m_firstLine;
	int m_lastLine;
    unsigned m_startColumn;
	SourceCode m_source;
};

class EvalExecutable : public ScriptExecutable {
	friend class LLIntOffsetsExtractor;
public:
	typedef ScriptExecutable Base;

	static void destroy(JSCell*);

    EvalCodeBlock* codeBlock()
    {
        return m_evalCodeBlock.get();
    }

	EvalCodeBlock* generatedBytecode()
	{
		ASSERT(m_evalCodeBlock);
		return m_evalCodeBlock.get();
	}

	static EvalExecutable* create(JSGlobalData&, const SourceCode&, bool isInStrictContext, DerivedContextType, bool isArrowFunctionContext, EvalContextType, const VariableEnvironment&);

#if ENABLE(JIT)
	PassRefPtr<JITCode> generatedJITCode()
	{
		return generatedJITCodeForCall();
	}
#endif
	static Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(vm, globalObject, proto, TypeInfo(EvalExecutableType, StructureFlags), info());
	}

    DECLARE_INFO;

    ExecutableInfo executableInfo() const { return ExecutableInfo(usesEval(), isStrictMode(), false, false, ConstructorKind::None, SuperBinding::NotNeeded, SourceParseMode::ProgramMode, derivedContextType(), isArrowFunctionContext(), false, evalContextType()); }

	void clearCode();

    unsigned numVariables();
    unsigned numberOfFunctionDecls();

    VariableEnvironment& variablesUnderTDZ() { return m_variablesUnderTDZ; }

private:
    friend class ScriptExecutable;

	EvalExecutable(JSGlobalData&, const SourceCode&, bool inStrictContext, DerivedContextType, bool isArrowFunctionContext, EvalContextType, const VariableEnvironment& variablesUnderTDZ);

	static void visitChildren(JSCell*, SlotVisitor&);

	RefPtr<EvalCodeBlock> m_evalCodeBlock;
	VariableEnvironment m_variablesUnderTDZ;
};

class ProgramExecutable : public ScriptExecutable {
	friend class LLIntOffsetsExtractor;
public:
	typedef ScriptExecutable Base;

	static ProgramExecutable* create(JSGlobalData& vm, const SourceCode& source)
	{
		ProgramExecutable* executable = new (NotNull, allocateCell<ProgramExecutable>(vm.heap)) ProgramExecutable(vm, source);
		executable->finishCreation(vm);
		return executable;
	}

    JSObject* initializeGlobalProperties(JSGlobalData&, ProgramNode*, CallFrame*, JSScope*);

	static void destroy(JSCell*);

	ProgramCodeBlock* codeBlock()
	{
		return m_programCodeBlock.get();
	}

	ProgramCodeBlock* generatedBytecode()
	{
		ASSERT(m_programCodeBlock);
		return m_programCodeBlock.get();
	}

#if ENABLE(JIT)
	PassRefPtr<JITCode> generatedJITCode()
	{
		return generatedJITCodeForCall();
	}
#endif

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(globalData, globalObject, proto, TypeInfo(ProgramExecutableType, StructureFlags), info());
	}

    DECLARE_INFO;

    ExecutableInfo executableInfo() const { return ExecutableInfo(usesEval(), isStrictMode(), false, false, ConstructorKind::None, SuperBinding::NotNeeded, SourceParseMode::ProgramMode, derivedContextType(), isArrowFunctionContext(), false, EvalContextType::None); }

	void clearCode();

private:
    friend class ScriptExecutable;

	ProgramExecutable(JSGlobalData&, const SourceCode&);

	static void visitChildren(JSCell*, SlotVisitor&);

	RefPtr<ProgramCodeBlock> m_programCodeBlock;
};

class ModuleProgramExecutable FINAL : public ScriptExecutable {
    friend class LLIntOffsetsExtractor;
public:
    typedef ScriptExecutable Base;
    static const unsigned StructureFlags = Base::StructureFlags | StructureIsImmortal;

    static ModuleProgramExecutable* create(JSGlobalData&, const SourceCode&);

    static const bool hasImmortalStructure = true;
    static void destroy(JSCell*);

    ModuleProgramCodeBlock* codeBlock()
	{
		return m_moduleProgramCodeBlock.get();
	}

    ModuleProgramCodeBlock* generatedBytecode()
    {
        return m_moduleProgramCodeBlock.get();
    }

    PassRefPtr<JITCode> generatedJITCode()
    {
        return generatedJITCodeForCall();
    }

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue proto)
    {
        return Structure::create(vm, globalObject, proto, TypeInfo(ModuleProgramExecutableType, StructureFlags), info());
    }

    DECLARE_INFO;

    ExecutableInfo executableInfo() const { return ExecutableInfo(usesEval(), isStrictMode(), false, false, ConstructorKind::None, SuperBinding::NotNeeded, SourceParseMode::ModuleEvaluateMode, derivedContextType(), isArrowFunctionContext(), false, EvalContextType::None); }

    void clearCode();

    SymbolTable* moduleEnvironmentSymbolTable() { return m_moduleEnvironmentSymbolTable.get(); }

private:
    friend class ScriptExecutable;

    ModuleProgramExecutable(JSGlobalData&, const SourceCode&);

    static void visitChildren(JSCell*, SlotVisitor&);

    WriteBarrier<SymbolTable> m_moduleEnvironmentSymbolTable;
    RefPtr<ModuleProgramCodeBlock> m_moduleProgramCodeBlock;
};

class FunctionExecutable : public ScriptExecutable {
	friend class JIT;
	friend class LLIntOffsetsExtractor;
public:
	typedef ScriptExecutable Base;

	static FunctionExecutable* create(JSGlobalData&, FunctionMetadataNode*, unsigned, bool, ConstructAbility, VariableEnvironment&, DerivedContextType derivedContextType);

	static FunctionExecutable* fromGlobalCode(const Identifier& name, ExecState*, Debugger*, const SourceCode&, JSObject** exception);

	static void destroy(JSCell*);

	// Returns either call or construct bytecode. This can be appropriate
	// for answering questions that that don't vary between call and construct --
	// for example, argumentsRegister().
	FunctionCodeBlock& generatedBytecode()
	{
		if (m_codeBlockForCall)
			return *m_codeBlockForCall;
		ASSERT(m_codeBlockForConstruct);
		return *m_codeBlockForConstruct;
	}

	bool isGeneratedForCall() const
	{
		return m_codeBlockForCall;
	}

	FunctionCodeBlock* generatedBytecodeForCall()
	{
		ASSERT(m_codeBlockForCall);
		return m_codeBlockForCall.get();
	}

	bool isGeneratedForConstruct() const
	{
		return m_codeBlockForConstruct;
	}

	FunctionCodeBlock* generatedBytecodeForConstruct()
	{
		ASSERT(m_codeBlockForConstruct);
		return m_codeBlockForConstruct.get();
	}

	bool isGeneratedFor(CodeSpecializationKind kind)
	{
		if (kind == CodeForCall)
			return isGeneratedForCall();
		ASSERT(kind == CodeForConstruct);
		return isGeneratedForConstruct();
	}

	FunctionCodeBlock* generatedBytecodeFor(CodeSpecializationKind kind)
	{
		if (kind == CodeForCall)
			return generatedBytecodeForCall();
		ASSERT(kind == CodeForConstruct);
		return generatedBytecodeForConstruct();
	}

	FunctionCodeBlock* baselineCodeBlockFor(CodeSpecializationKind);

	FunctionCodeBlock* profiledCodeBlockFor(CodeSpecializationKind kind)
	{
		return baselineCodeBlockFor(kind);
	}

    FunctionMode functionMode() const { return static_cast<FunctionMode>(m_functionMode); }
	bool isBuiltinFunction() const { return m_isBuiltinFunction; }
	bool isClassConstructorFunction() const { return constructorKind() != ConstructorKind::None; }
	ConstructorKind constructorKind() const { return static_cast<ConstructorKind>(m_constructorKind); }
	ConstructAbility constructAbility() const { return static_cast<ConstructAbility>(m_constructAbility); }
	bool isClass() const { return !classSource().isNull(); }
	bool isGetter() const { return parseMode() == SourceParseMode::GetterMode; }
	bool isSetter() const { return parseMode() == SourceParseMode::SetterMode; }
	const Identifier& name() { return m_name; }
    const Identifier& ecmaName() const { return m_ecmaName; }
    void setEcmaName(const Identifier& name) { m_ecmaName = name; }
	const Identifier& inferredName() { return m_inferredName; }
    unsigned parameterCount() const { return m_parameterCount; };
    unsigned parametersStartOffset() const { return m_parametersStartOffset; }
    SourceParseMode parseMode() const { return m_parseMode; };
	bool isArrowFunction() const { return parseMode() == SourceParseMode::ArrowFunctionMode; }
	SuperBinding superBinding() const { return static_cast<SuperBinding>(m_superBinding); }

    const SourceCode& classSource() const { return m_classSource; };
    void setClassSource(const SourceCode& source) { m_classSource = source; };

	static void visitChildren(JSCell*, SlotVisitor&);
	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(globalData, globalObject, proto, TypeInfo(FunctionExecutableType, StructureFlags), info());
	}

	DECLARE_INFO;

	void clearCode();

    InferredValue* singletonFunction() { return m_singletonFunction.get(); }

	VariableEnvironment parentScopeTDZVariables() { return m_parentScopeTDZVariables->environment().toVariableEnvironment(); }

protected:
	void finishCreation(JSGlobalData& globalData);

private:
	FunctionExecutable(JSGlobalData&, FunctionMetadataNode*, unsigned startColumn, bool isBuiltinFunction, ConstructAbility constructAbility, VariableEnvironment&, DerivedContextType derivedContextType);

	FunctionCodeBlock* codeBlockFor(CodeSpecializationKind kind);

    friend class ScriptExecutable;

	RefPtr<FunctionCodeBlock> m_codeBlockForCall;
	RefPtr<FunctionCodeBlock> m_codeBlockForConstruct;
    SourceCode m_classSource;
    OwnPtr<CompactVariableMap::Handle> m_parentScopeTDZVariables;
    unsigned m_parameterCount;
    unsigned m_parametersStartOffset;
    SourceParseMode m_parseMode;
    unsigned m_functionMode : 2; // FunctionMode
	unsigned m_isBuiltinFunction : 1;
    unsigned m_constructAbility: 1;
    unsigned m_constructorKind : 2;
    unsigned m_superBinding : 1;
	Identifier m_name;
    Identifier m_ecmaName;
    Identifier m_inferredName;
    bool m_didParseForTheFirstTime;
    WriteBarrier<InferredValue> m_singletonFunction;
};

inline void ExecutableBase::clearCodeVirtual(ExecutableBase* executable)
{
	switch (executable->structure()->typeInfo().type()) {
	case EvalExecutableType:
		return jsCast<EvalExecutable*>(executable)->clearCode();
	case ProgramExecutableType:
		return jsCast<ProgramExecutable*>(executable)->clearCode();
	case FunctionExecutableType:
		return jsCast<FunctionExecutable*>(executable)->clearCode();
    case ModuleProgramExecutableType:
        return jsCast<ModuleProgramExecutable*>(executable)->clearCode();
	default:
		return jsCast<NativeExecutable*>(executable)->clearCode();
	}
}

template <typename ExecutableType>
JSObject* ScriptExecutable::prepareForExecution(ExecState* exec, JSFunction* function, JSScope* scope, CodeSpecializationKind kind, CodeBlock*& resultCodeBlock)
{
    if (hasJITCodeFor(kind)) {
        if (classInfo() == EvalExecutable::info())
            resultCodeBlock = jsCast<CodeBlock*>(jsCast<EvalExecutable*>(this)->codeBlock());
        else if (classInfo() == ProgramExecutable::info())
            resultCodeBlock = jsCast<CodeBlock*>(jsCast<ProgramExecutable*>(this)->codeBlock());
        else if (classInfo() == ModuleProgramExecutable::info())
            resultCodeBlock = jsCast<CodeBlock*>(jsCast<ModuleProgramExecutable*>(this)->codeBlock());
        else if (classInfo() == FunctionExecutable::info())
            resultCodeBlock = jsCast<CodeBlock*>(jsCast<FunctionExecutable*>(this)->codeBlockFor(kind));
        else
            RELEASE_ASSERT_NOT_REACHED();
        return 0;
    }
    return prepareForExecutionImpl(exec, function, scope, kind, resultCodeBlock);
}

}

#include "JSFunctionInlines.h"

#endif
