/*
 *  Copyright (C) 2003, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#ifndef Identifier_h
#define Identifier_h

#include "JSGlobalData.h"
#include <wtf/Optional.h>
#include <wtf/ThreadSpecific.h>
#include <wtf/WTFThreadData.h>
#include <wtf/text/CString.h>
#include <wtf/text/SymbolImpl.h>
#include <wtf/text/UniquedStringImpl.h>
#include <wtf/text/WTFString.h>

namespace JSC {

class ExecState;

ALWAYS_INLINE bool isIndex(uint32_t index)
{
	return index != 0xFFFFFFFFU;
}

template <typename CharType>
ALWAYS_INLINE Optional<uint32_t> parseIndex(const CharType* characters, unsigned length)
{
	// An empty string is not a number.
	if (UNLIKELY(!length))
		return Nullopt;

	// Get the first character, turning it into a digit.
	uint32_t value = characters[0] - '0';
	if (value > 9)
		return Nullopt;

	// Check for leading zeros. If the first characher is 0, then the
	// length of the string must be one - e.g. "042" is not equal to "42".
	if (!value && length > 1)
		return Nullopt;

	while (--length) {
		// Multiply value by 10, checking for overflow out of 32 bits.
		if (value > 0xFFFFFFFFU / 10)
			return Nullopt;
		value *= 10;

		// Get the next character, turning it into a digit.
		uint32_t newValue = *(++characters) - '0';
		if (newValue > 9)
			return Nullopt;

		// Add in the old value, checking for overflow out of 32 bits.
		newValue += value;
		if (newValue < value)
			return Nullopt;
		value = newValue;
	}

	if (!isIndex(value))
		return Nullopt;
	return value;
}

ALWAYS_INLINE Optional<uint32_t> parseIndex(StringImpl& impl)
{
	if (impl.is8Bit())
		return parseIndex(impl.characters8(), impl.length());
	return parseIndex(impl.characters16(), impl.length());
}

class Identifier {
	friend class Structure;
public:
	ALWAYS_INLINE Identifier() { }
	enum EmptyIdentifierFlag { EmptyIdentifier };
	ALWAYS_INLINE Identifier(EmptyIdentifierFlag) : m_string(StringImpl::empty()) { }

	ALWAYS_INLINE ~Identifier() { }

	const String& string() const { return m_string; }
	const String& ustring() const { return m_string; }
    UniquedStringImpl* impl() const { return static_cast<UniquedStringImpl*>(m_string.impl()); }

	ALWAYS_INLINE int length() const { return m_string.length(); }

	CString ascii() const { return m_string.ascii(); }
	CString utf8() const { return m_string.utf8(); }

	// Only to be used with string literals.
	template<unsigned charactersCount>
	static Identifier fromString(JSGlobalData&, const char (&characters)[charactersCount]);
	static Identifier fromString(JSGlobalData&, const LChar*, int length);
	static Identifier fromString(JSGlobalData&, const UChar*, int length);
	static Identifier fromString(JSGlobalData&, const String&);
	static Identifier fromString(JSGlobalData&, const char*);
	static Identifier fromString(JSGlobalData&, AtomicStringImpl*);
	static Identifier fromString(JSGlobalData&, const AtomicString&);

	static Identifier fromUid(JSGlobalData&, UniquedStringImpl* uid);
	static Identifier fromUid(const PrivateName&);

	static Identifier createLCharFromUChar(JSGlobalData& vm, const UChar* s, int length) { return Identifier(vm, add8(vm, s, length)); }

	static Identifier from(JSGlobalData& vm, unsigned y);
	static Identifier from(JSGlobalData& vm, int y);
	static Identifier from(JSGlobalData&vm , double y);

	ALWAYS_INLINE bool isNull() const { return m_string.isNull(); }
	ALWAYS_INLINE bool isEmpty() const { return m_string.isEmpty(); }
	ALWAYS_INLINE bool isSymbol() const { return !isNull() && impl()->isSymbol(); }

	friend bool operator==(const Identifier&, const Identifier&);
	friend bool operator!=(const Identifier&, const Identifier&);

	friend bool operator==(const Identifier&, const LChar*);
	friend bool operator==(const Identifier&, const char*);
	friend bool operator!=(const Identifier&, const LChar*);
	friend bool operator!=(const Identifier&, const char*);

	static bool equal(const StringImpl*, const LChar*);
	static ALWAYS_INLINE bool equal(const StringImpl*a, const char*b) { return Identifier::equal(a, reinterpret_cast<const LChar*>(b)); };
	static bool equal(const StringImpl*, const LChar*, unsigned length);
	static bool equal(const StringImpl*, const UChar*, unsigned length);
	static ALWAYS_INLINE bool equal(const StringImpl* a, const StringImpl* b) { return ::equal(a, b); }

private:
	String m_string;

	// Only to be used with string literals.
	template<unsigned charactersCount>
	ALWAYS_INLINE Identifier(JSGlobalData& vm, const char (&characters)[charactersCount]) : m_string(add(vm, characters)) { }

	Identifier(JSGlobalData& vm, const LChar* s, int length) : m_string(add(vm, s, length)) { }
	Identifier(JSGlobalData& vm, const UChar* s, int length) : m_string(add(vm, s, length)) { }
	Identifier(JSGlobalData& vm, StringImpl* rep) : m_string(add(vm, rep)) { }
	Identifier(JSGlobalData& vm, const String& s) : m_string(add(vm, s.impl())) { }

	Identifier(JSGlobalData&, AtomicStringImpl*);
	Identifier(JSGlobalData&, const AtomicString&);

    Identifier(SymbolImpl& uid)
        : m_string(&uid)
    {
    }

	template <typename CharType>
	ALWAYS_INLINE static uint32_t toUInt32FromCharacters(const CharType* characters, unsigned length, bool& ok);

	static ALWAYS_INLINE bool equal(const Identifier& a, const Identifier& b) { return a.m_string.impl() == b.m_string.impl(); }
	static bool equal(const Identifier& a, const LChar* b) { return equal(a.m_string.impl(), b); }

	template <typename T> static PassRefPtr<StringImpl> add(JSGlobalData&, const T*, int length);
	static PassRefPtr<StringImpl> add8(JSGlobalData&, const UChar*, int length);
	template <typename T> ALWAYS_INLINE static bool canUseSingleCharacterString(T);

	static PassRefPtr<StringImpl> add(JSGlobalData& vm, StringImpl* r);

	// Only to be used with string literals.
	template<unsigned charactersCount>
	static PassRefPtr<StringImpl> add(JSGlobalData&, const char (&characters)[charactersCount]);

	JS_EXPORT_PRIVATE static void checkCurrentIdentifierTable(JSGlobalData&);
};

template <> ALWAYS_INLINE bool Identifier::canUseSingleCharacterString(LChar)
{
	ASSERT(maxSingleCharacterString == 0xff);
	return true;
}

template <> ALWAYS_INLINE bool Identifier::canUseSingleCharacterString(UChar c)
{
	return (c <= maxSingleCharacterString);
}

template <typename T>
struct CharBuffer {
	const T* s;
	unsigned int length;
};

template <typename T>
PassRefPtr<StringImpl> Identifier::add(JSGlobalData& vm, const T* s, int length)
{
	if (length == 1) {
		T c = s[0];
		if (canUseSingleCharacterString(c))
			return vm.smallStrings.singleCharacterStringRep(c);
	}

	if (!length)
		return StringImpl::empty();

	return AtomicString::add(s, length);
}

ALWAYS_INLINE bool operator==(const Identifier& a, const Identifier& b)
{
	return Identifier::equal(a, b);
}

ALWAYS_INLINE bool operator!=(const Identifier& a, const Identifier& b)
{
	return !Identifier::equal(a, b);
}

ALWAYS_INLINE bool operator==(const Identifier& a, const LChar* b)
{
	return Identifier::equal(a, b);
}

ALWAYS_INLINE bool operator==(const Identifier& a, const char* b)
{
	return Identifier::equal(a, reinterpret_cast<const LChar*>(b));
}

ALWAYS_INLINE bool operator!=(const Identifier& a, const LChar* b)
{
	return !Identifier::equal(a, b);
}

ALWAYS_INLINE bool operator!=(const Identifier& a, const char* b)
{
	return !Identifier::equal(a, reinterpret_cast<const LChar*>(b));
}

ALWAYS_INLINE bool Identifier::equal(const StringImpl* r, const LChar* s)
{
	return WTF::equal(r, s);
}

ALWAYS_INLINE bool Identifier::equal(const StringImpl* r, const LChar* s, unsigned length)
{
	return WTF::equal(r, s, length);
}

ALWAYS_INLINE bool Identifier::equal(const StringImpl* r, const UChar* s, unsigned length)
{
	return WTF::equal(r, s, length);
}

IdentifierTable* createIdentifierTable();
void deleteIdentifierTable(IdentifierTable*);

ALWAYS_INLINE Optional<uint32_t> parseIndex(const Identifier& identifier)
{
	auto* uid = identifier.impl();
	if (!uid)
		return Nullopt;
	if (uid->isSymbol())
		return Nullopt;
	return parseIndex(*uid);
}

JSValue identifierToJSValue(JSGlobalData&, const Identifier&);
// This will stringify private symbols. When leaking JSValues to
// non-internal code, make sure to use this function and not the above one.
JSValue identifierToSafePublicJSValue(JSGlobalData&, const Identifier&);

struct IdentifierRepHash : PtrHash<RefPtr<UniquedStringImpl>> {
	static unsigned hash(const RefPtr<UniquedStringImpl>& key) { return key->existingSymbolAwareHash(); }
	static unsigned hash(UniquedStringImpl* key) { return key->existingSymbolAwareHash(); }
};

struct IdentifierMapIndexHashTraits : HashTraits<int> {
	static ALWAYS_INLINE int emptyValue() { return std::numeric_limits<int>::max(); }
	static const bool emptyValueIsZero = false;
};

typedef HashSet<RefPtr<UniquedStringImpl>, IdentifierRepHash> IdentifierSet;
typedef HashMap<RefPtr<UniquedStringImpl>, int, IdentifierRepHash, HashTraits<RefPtr<UniquedStringImpl> >, IdentifierMapIndexHashTraits> IdentifierMap;
typedef HashMap<UniquedStringImpl*, int, IdentifierRepHash, HashTraits<UniquedStringImpl*>, IdentifierMapIndexHashTraits> BorrowedIdentifierMap;

} // namespace JSC

namespace WTF {

template <> struct VectorTraits<JSC::Identifier> : SimpleClassVectorTraits { };

} // namespace WTF

#endif // Identifier_h
