/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef JSArray_h
#define JSArray_h

#include "ArrayConventions.h"
#include "ButterflyInlineMethods.h"
#include "JSCellInlines.h"
#include "Exception.h"
#include "JSObject.h"

namespace JSC {

class JSArray;
class LLIntOffsetsExtractor;

class JSArray : public JSNonFinalObject {
	friend class LLIntOffsetsExtractor;
	friend class Walker;
	friend class JIT;

public:
	typedef JSNonFinalObject Base;

    static size_t allocationSize(size_t inlineCapacity)
    {
        ASSERT_UNUSED(inlineCapacity, !inlineCapacity);
        return sizeof(JSArray);
    }

protected:
	explicit JSArray(JSGlobalData& globalData, Structure* structure, Butterfly* butterfly)
		: JSNonFinalObject(globalData, structure, butterfly)
	{
	}

public:
    static JSArray* tryCreate(JSGlobalData&, Structure*, unsigned initialLength = 0);
	static JSArray* tryCreate(JSGlobalData&, Structure*, unsigned initialLength, unsigned vectorLengthHint);
	static JSArray* create(JSGlobalData&, Structure*, unsigned initialLength = 0);
    static JSArray* createWithButterfly(JSGlobalData&, Structure*, Butterfly*);

    // tryCreateUninitializedRestricted is used for fast construction of arrays whose size and
    // contents are known at time of creation. This is a restricted API for careful use only in
    // performance critical code paths. If you don't have a good reason to use it, you probably
    // shouldn't use it. Instead, you should go with
    //   - JSArray::tryCreate() or JSArray::create() instead of tryCreateUninitializedRestricted(), and
    //   - putDirectIndex() instead of initializeIndex().
    //
    // Clients of this interface must:
	//   - null-check the result (indicating out of memory, or otherwise unable to allocate vector).
	//   - call 'initializeIndex' for all properties in sequence, for 0 <= i < initialLength.
	static JSArray* tryCreateUninitializedRestricted(JSGlobalData&, Structure*, unsigned initialLength);

    static void eagerlyInitializeButterfly(JSArray*, unsigned initialLength);

	JS_EXPORT_PRIVATE static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool throwException);

	static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);

	DECLARE_EXPORT_INFO

	// OK if we know this is a JSArray, but not if it could be an object of a derived class; for RuntimeArray this always returns 0.
	unsigned length() const { return getArrayLength(); }

	// OK to use on new arrays, but not if it might be a RegExpMatchArray or RuntimeArray.
	bool setLength(ExecState*, unsigned, bool throwException = false);

	void push(ExecState*, JSValue);
	JSValue pop(ExecState*);

	JSArray* fastSlice(ExecState&, unsigned startIndex, unsigned count);

    bool canFastCopy(VM&, JSArray* otherArray);
    bool canDoFastIndexedAccess(VM&);
    // This function returns NonArray if the indexing types are not compatable for copying.
    IndexingType mergeIndexingTypeForCopying(IndexingType other);
    bool appendMemcpy(ExecState*, VM&, unsigned startIndex, JSArray* otherArray);

    static IndexingType fastConcatType(VM& vm, JSArray& firstArray, JSArray& secondArray)
    {
        IndexingType type = firstArray.structure()->indexingType();
        if (type != secondArray.structure()->indexingType())
            return NonArray;
        if (type != ArrayWithDouble && type != ArrayWithInt32 && type != ArrayWithContiguous)
            return NonArray;
        if (firstArray.structure(vm)->holesMustForwardToPrototype(vm)
            || secondArray.structure(vm)->holesMustForwardToPrototype(vm))
            return NonArray;
        return type;
    }
    EncodedJSValue fastConcatWith(ExecState&, JSArray&);

	enum ShiftCountMode {
		// This form of shift hints that we're doing queueing. With this assumption in hand,
		// we convert to ArrayStorage, which has queue optimizations.
		ShiftCountForShift,

		// This form of shift hints that we're just doing care and feeding on an array that
		// is probably typically used for ordinary accesses. With this assumption in hand,
		// we try to preserve whatever indexing type it has already.
		ShiftCountForSplice
	};

	bool shiftCountForShift(ExecState* exec, unsigned startIndex, unsigned count)
	{
		return shiftCountWithArrayStorage(exec->vm(), startIndex, count, ensureArrayStorage(exec->globalData()));
	}
	bool shiftCountForSplice(ExecState* exec, unsigned& startIndex, unsigned count)
	{
		return shiftCountWithAnyIndexingType(exec, startIndex, count);
	}
	template<ShiftCountMode shiftCountMode>
	bool shiftCount(ExecState* exec, unsigned& startIndex, unsigned count)
	{
		switch (shiftCountMode) {
		case ShiftCountForShift:
			return shiftCountForShift(exec, startIndex, count);
		case ShiftCountForSplice:
			return shiftCountForSplice(exec, startIndex, count);
		default:
			RELEASE_ASSERT_NOT_REACHED();
			return false;
		}
	}

	bool unshiftCountForShift(ExecState* exec, unsigned startIndex, unsigned count)
	{
		return unshiftCountWithArrayStorage(exec, startIndex, count, ensureArrayStorage(exec->globalData()));
	}
	bool unshiftCountForSplice(ExecState* exec, unsigned startIndex, unsigned count)
	{
		return unshiftCountWithAnyIndexingType(exec, startIndex, count);
	}
	template<ShiftCountMode shiftCountMode>
	bool unshiftCount(ExecState* exec, unsigned startIndex, unsigned count)
	{
		switch (shiftCountMode) {
		case ShiftCountForShift:
			return unshiftCountForShift(exec, startIndex, count);
		case ShiftCountForSplice:
			return unshiftCountForSplice(exec, startIndex, count);
		default:
			RELEASE_ASSERT_NOT_REACHED();
			return false;
		}
	}

	void fillArgList(ExecState*, MarkedArgumentBuffer&);
	void copyToArguments(ExecState*, CallFrame*, uint32_t copyLength, int32_t firstVarArgOffset);

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype, IndexingType indexingType)
	{
		return Structure::create(globalData, globalObject, prototype, TypeInfo(ArrayType, StructureFlags), info(), indexingType);
	}

protected:
	static const unsigned StructureFlags = OverridesGetOwnPropertySlot | OverridesGetPropertyNames | JSObject::StructureFlags;
	static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);

	static bool deleteProperty(JSCell*, ExecState*, PropertyName);
	JS_EXPORT_PRIVATE static void getOwnNonIndexPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);

private:
	bool isLengthWritable()
	{
		ArrayStorage* storage = arrayStorageOrNull();
		if (!storage)
			return true;
		SparseArrayValueMap* map = storage->m_sparseMap.get();
		return !map || !map->lengthIsReadOnly();
	}

	bool shiftCountWithAnyIndexingType(ExecState*, unsigned& startIndex, unsigned count);
	bool shiftCountWithArrayStorage(JSGlobalData&, unsigned startIndex, unsigned count, ArrayStorage*);

	bool unshiftCountWithAnyIndexingType(ExecState*, unsigned startIndex, unsigned count);
	bool unshiftCountWithArrayStorage(ExecState*, unsigned startIndex, unsigned count, ArrayStorage*);
	bool unshiftCountSlowCase(JSGlobalData&, bool, unsigned);

	bool setLengthWithArrayStorage(ExecState*, unsigned newLength, bool throwException, ArrayStorage*);
	void setLengthWritable(ExecState*, bool writable);
};

inline Butterfly* tryCreateArrayButterfly(VM& vm, JSCell* intendedOwner, unsigned initialLength)
{
    Butterfly* butterfly = Butterfly::tryCreate(
        vm, intendedOwner, 0, 0, true, baseIndexingHeaderForArrayStorage(initialLength),
        ArrayStorage::sizeFor(BASE_ARRAY_STORAGE_VECTOR_LEN));
    if (UNLIKELY(!butterfly))
        return 0;
    ArrayStorage* storage = butterfly->arrayStorage();
    storage->m_sparseMap.clear();
    storage->m_indexBias = 0;
    storage->m_numValuesInVector = 0;
    return butterfly;
}

Butterfly* createArrayButterflyInDictionaryIndexingMode(JSGlobalData&, JSCell* intendedOwner, unsigned initialLength);

inline JSArray* JSArray::tryCreate(JSGlobalData& globalData, Structure* structure, unsigned initialLength, unsigned vectorLengthHint)
{
	Butterfly* butterfly;
    IndexingType indexingType = structure->indexingType();
	if (LIKELY(!hasAnyArrayStorage(indexingType))) {
		ASSERT(
			hasUndecided(indexingType)
			|| hasInt32(indexingType)
			|| hasDouble(indexingType)
			|| hasContiguous(indexingType));

        if (UNLIKELY(vectorLengthHint > MAX_STORAGE_VECTOR_LENGTH))
            return 0;

		IndexingHeader header;
		unsigned vectorLength = std::max(vectorLengthHint, BASE_ARRAY_STORAGE_VECTOR_LEN);
		header.setVectorLength(vectorLength);
		header.setPublicLength(initialLength);
		butterfly = Butterfly::create(globalData, 0, 0, 0, true, header, vectorLength * sizeof(EncodedJSValue));
		ASSERT(initialLength < MIN_ARRAY_STORAGE_CONSTRUCTION_LENGTH);
		if (hasDouble(indexingType)) {
			for (unsigned i = 0; i < vectorLength; ++i)
				butterfly->contiguousDouble()[i] = PNaN;
		}
	} else {
		ASSERT(
			indexingType == ArrayWithSlowPutArrayStorage
			|| indexingType == ArrayWithArrayStorage);
		butterfly = tryCreateArrayButterfly(globalData, 0, initialLength);
        if (UNLIKELY(!butterfly))
            return 0;
	}
    return createWithButterfly(globalData, structure, butterfly);
}

inline JSArray* JSArray::tryCreate(JSGlobalData& vm, Structure* structure, unsigned initialLength)
{
    return tryCreate(vm, structure, initialLength, initialLength);
}

inline JSArray* JSArray::create(JSGlobalData& vm, Structure* structure, unsigned initialLength)
{
    JSArray* result = JSArray::tryCreate(vm, structure, initialLength);
    RELEASE_ASSERT(result);

    return result;
}

inline JSArray* JSArray::createWithButterfly(JSGlobalData& vm, Structure* structure, Butterfly* butterfly)
{
    JSArray* array = new (NotNull, allocateCell<JSArray>(vm.heap)) JSArray(vm, structure, butterfly);
    array->finishCreation(vm);
    return array;
}

JSArray* asArray(JSValue);

ALWAYS_INLINE JSArray* asArray(JSCell* cell)
{
	ASSERT(cell->inherits(JSArray::info()));
	return jsCast<JSArray*>(cell);
}

ALWAYS_INLINE JSArray* asArray(JSValue value)
{
	return asArray(value.asCell());
}

ALWAYS_INLINE bool isJSArray(const JSCell* cell)
{
    ASSERT((cell->classInfo() == JSArray::info()) == (cell->structure()->typeInfo().type() == ArrayType));
    return cell->structure()->typeInfo().type() == ArrayType;
}

ALWAYS_INLINE bool isJSArray(JSValue v) { return v.isCell() && isJSArray(v.asCell()); }

JSArray* constructArray(VM&, Structure*, const ArgList& values);
JSArray* constructArray(VM&, Structure*, const JSValue* values, unsigned length);
JSArray* constructArrayNegativeIndexed(VM&, Structure*, const JSValue* values, unsigned length);

ALWAYS_INLINE double toLength(ExecState* exec, JSObject* obj)
{
    if (LIKELY(isJSArray(obj)))
        return jsCast<JSArray*>(obj)->length();

    VM& vm = exec->vm();
    JSValue lengthValue = obj->get(exec, vm.propertyNames->length);
    RETURN_IF_EXCEPTION(vm, PNaN);
    return lengthValue.toLength(exec);
}

} // namespace JSC

#endif // JSArray_h
