/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#ifndef JSObject_h
#define JSObject_h

#include "ArgList.h"
#include "ArrayConventions.h"
#include "ArrayStorage.h"
#include "Butterfly.h"
#include "ClassInfo.h"
#include "CommonIdentifiers.h"
#include "CallFrame.h"
#include "CustomGetterSetter.h"
#include "DeferGC.h"
#include "ExceptionScope.h"
#include "JSCell.h"
#include "PropertySlot.h"
#include "PropertyStorage.h"
#include "PutDirectIndexMode.h"
#include "PutPropertySlot.h"
#include "RuntimeType.h"
#include "CustomGetterSetter.h"
#include "Structure.h"
#include "JSGlobalData.h"
#include "JSString.h"
#include "SlotVisitorInlines.h"
#include "SparseArrayValueMap.h"
#include <wtf/StdLibExtras.h>

namespace JSC {

ALWAYS_INLINE JSCell* getJSFunction(JSValue value)
{
	if (value.isCell() && (value.asCell()->structure()->typeInfo().type() == JSFunctionType))
		return value.asCell();
	return 0;
}

class GetterSetter;
class InternalFunction;
class JSFunction;
class LLIntOffsetsExtractor;
class MarkedBlock;
class PropertyDescriptor;
class PropertyNameArray;
class Structure;
struct HashTable;
struct HashTableValue;

JS_EXPORT_PRIVATE JSObject* throwTypeError(ExecState*, const String&);
extern JS_EXPORTDATA const char* ReadonlyPropertyWriteError;
extern JS_EXPORTDATA const char* UnconfigurablePropertyChangeAccessMechanismError;

COMPILE_ASSERT(None < FirstInternalAttribute, None_is_below_FirstInternalAttribute);
COMPILE_ASSERT(ReadOnly < FirstInternalAttribute, ReadOnly_is_below_FirstInternalAttribute);
COMPILE_ASSERT(DontEnum < FirstInternalAttribute, DontEnum_is_below_FirstInternalAttribute);
COMPILE_ASSERT(DontDelete < FirstInternalAttribute, DontDelete_is_below_FirstInternalAttribute);
COMPILE_ASSERT(Accessor < FirstInternalAttribute, Accessor_is_below_FirstInternalAttribute);

class JSFinalObject;

class JSObject : public JSCell {
	friend class BatchedTransitionOptimizer;
	friend class JIT;
	friend class JSCell;
	friend class JSFinalObject;
	friend class MarkedBlock;
	JS_EXPORT_PRIVATE friend bool setUpStaticFunctionSlot(VM&, const HashTableValue*, JSObject*, PropertyName, PropertySlot&);

	enum PutMode {
		PutModePut,
		PutModeDefineOwnProperty,
	};

public:
	typedef JSCell Base;

	JS_EXPORT_PRIVATE static void visitChildren(JSCell*, SlotVisitor&);
	JS_EXPORT_PRIVATE static void copyBackingStore(JSCell*, CopyVisitor&, CopyToken);

	JS_EXPORT_PRIVATE static String className(const JSObject*);
    JS_EXPORT_PRIVATE static String calculatedClassName(JSObject*);

    // This function is what Object.prototype.toString() will use to get the name of
    // an object when using Symbol.toStringTag fails. For the most part there is no
    // difference between this and className(). The main use case is for new JS language
    // objects to set the default tag to "Object".
    JS_EXPORT_PRIVATE static String toStringName(const JSObject*, ExecState*);

    // This is the fully virtual [[GetPrototypeOf]] internal function defined
    // in the ECMAScript 6 specification. Use this when doing a [[GetPrototypeOf]]
    // operation as dictated in the specification.
    JSValue getPrototype(VM&, ExecState*);
    JS_EXPORT_PRIVATE static JSValue getPrototype(JSObject*, ExecState*);
    // This gets the prototype directly off of the structure. This does not do
    // dynamic dispatch on the getPrototype method table method. It is not valid
    // to use this when performing a [[GetPrototypeOf]] operation in the specification.
    // It is valid to use though when you know that you want to directly get it
    // without consulting the method table. This is akin to getting the [[Prototype]]
    // internal field directly as described in the specification.
    JSValue getPrototypeDirect() const;

    // This sets the prototype without checking for cycles and without
    // doing dynamic dispatch on [[SetPrototypeOf]] operation in the specification.
    // It is not valid to use this when performing a [[SetPrototypeOf]] operation in
    // the specification. It is valid to use though when you know that you want to directly
    // set it without consulting the method table and when you definitely won't
    // introduce a cycle in the prototype chain. This is akin to setting the
    // [[Prototype]] internal field directly as described in the specification.
    JS_EXPORT_PRIVATE void setPrototypeDirect(VM&, JSValue prototype);
private:
    // This is OrdinarySetPrototypeOf in the specification. Section 9.1.2.1
    // https://tc39.github.io/ecma262/#sec-ordinarysetprototypeof
    JS_EXPORT_PRIVATE bool setPrototypeWithCycleCheck(VM&, ExecState*, JSValue prototype, bool shouldThrowIfCantSet);
public:
    // This is the fully virtual [[SetPrototypeOf]] internal function defined
    // in the ECMAScript 6 specification. Use this when doing a [[SetPrototypeOf]]
    // operation as dictated in the specification.
    bool setPrototype(VM&, ExecState*, JSValue prototype, bool shouldThrowIfCantSet = false);
	JS_EXPORT_PRIVATE static bool setPrototype(JSObject*, ExecState*, JSValue prototype, bool shouldThrowIfCantSet);

    inline bool mayInterceptIndexedAccesses() const;
    inline bool mayInterceptIndexedAccesses(VM&) const;

	JSValue get(ExecState*, PropertyName) const;
	JSValue get(ExecState*, unsigned propertyName) const;

	bool getPropertySlot(ExecState*, PropertyName, PropertySlot&);
	bool getPropertySlot(ExecState*, unsigned propertyName, PropertySlot&);
	//template<typename CallbackWhenNoException> typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type getPropertySlot(ExecState*, PropertyName, CallbackWhenNoException) const;
	//template<typename CallbackWhenNoException> typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type getPropertySlot(ExecState*, PropertyName, PropertySlot&, CallbackWhenNoException) const;

	static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);
	JS_EXPORT_PRIVATE static bool getOwnPropertySlotByIndex(JSObject*, ExecState*, unsigned propertyName, PropertySlot&);
	bool getOwnPropertyDescriptor(ExecState*, PropertyName, PropertyDescriptor&);

	unsigned getArrayLength() const
	{
		if (!hasIndexedProperties(structure()->indexingType()))
			return 0;
		return m_butterfly->publicLength();
	}

	unsigned getVectorLength()
	{
		if (!hasIndexedProperties(structure()->indexingType()))
			return 0;
		return m_butterfly->vectorLength();
	}

    static bool putInlineForJSObject(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);

	JS_EXPORT_PRIVATE static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);
	// putByIndex assumes that the receiver is this JSCell object.
	JS_EXPORT_PRIVATE static bool putByIndex(JSCell*, ExecState*, unsigned propertyName, JSValue, bool shouldThrow);

    // This performs the ECMAScript Set() operation.
	ALWAYS_INLINE bool putByIndexInline(ExecState* exec, unsigned propertyName, JSValue value, bool shouldThrow)
    {
        if (canSetIndexQuickly(propertyName)) {
            setIndexQuickly(exec->vm(), propertyName, value);
            return true;
        }
        return methodTable()->putByIndex(this, exec, propertyName, value, shouldThrow);
    }

    // This is similar to the putDirect* methods:
    //  - the prototype chain is not consulted
    //  - accessors are not called.
    //  - it will ignore extensibility and read-only properties if PutDirectIndexLikePutDirect is passed as the mode (the default).
    // This method creates a property with attributes writable, enumerable and configurable all set to true if attributes is zero,
    // otherwise, it creates a property with the provided attributes. Semantically, this is performing defineOwnProperty.
    bool putDirectIndex(ExecState* exec, unsigned propertyName, JSValue value, unsigned attributes, PutDirectIndexMode mode)
    {
        ASSERT(!value.isCustomGetterSetter());
        if (!attributes && canSetIndexQuicklyForPutDirect(propertyName)) {
            setIndexQuickly(exec->vm(), propertyName, value);
            return true;
        }
        return putDirectIndexSlowOrBeyondVectorLength(exec, propertyName, value, attributes, mode);
    }
    // This is semantically equivalent to performing defineOwnProperty(propertyName, {configurable:true, writable:true, enumerable:true, value:value}).
	bool putDirectIndex(ExecState* exec, unsigned propertyName, JSValue value)
    {
        return putDirectIndex(exec, propertyName, value, 0, PutDirectIndexLikePutDirect);
    }

	// A generally non-throwing version of putDirect and putDirectIndex.
	// However, it's only guaranteed to not throw based on what the receiver is.
	// For example, if the receiver is a ProxyObject, this is not guaranteed, since
	// it may call into arbitrary JS code. It's the responsibility of the user of
	// this API to ensure that the receiver object is a well known type if they
	// want to ensure that this won't throw an exception.
	bool putDirectMayBeIndex(ExecState*, VM&, PropertyName, JSValue);

    bool hasIndexingHeader() const
    {
        return structure()->hasIndexingHeader(this);
    }

    bool canGetIndexQuickly(unsigned i)
	{
		switch (structure()->indexingType()) {
		case ALL_BLANK_INDEXING_TYPES:
		case ALL_UNDECIDED_INDEXING_TYPES:
			return false;
		case ALL_INT32_INDEXING_TYPES:
		case ALL_CONTIGUOUS_INDEXING_TYPES:
			return i < m_butterfly->vectorLength() && m_butterfly->contiguous()[i];
		case ALL_DOUBLE_INDEXING_TYPES: {
			if (i >= m_butterfly->vectorLength())
				return false;
			double value = m_butterfly->contiguousDouble()[i];
			if (value != value)
				return false;
			return true;
		}
		case ALL_ARRAY_STORAGE_INDEXING_TYPES:
			return i < m_butterfly->arrayStorage()->vectorLength() && m_butterfly->arrayStorage()->m_vector[i];
		default:
			ASSERT_NOT_REACHED();
			return false;
		}
	}

    JSValue getIndexQuickly(unsigned i)
    {
        switch (structure()->indexingType()) {
        case ALL_INT32_INDEXING_TYPES:
            return jsNumber(m_butterfly->contiguous()[i].get().asInt32());
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            return m_butterfly->contiguous()[i].get();
        case ALL_DOUBLE_INDEXING_TYPES:
        	return JSValue(JSValue::EncodeAsDouble, m_butterfly->contiguousDouble()[i]);
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return m_butterfly->arrayStorage()->m_vector[i].get();
        default:
            ASSERT_NOT_REACHED();
            return JSValue();
        }
    }

    JSValue tryGetIndexQuickly(unsigned i) const
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
        case ALL_UNDECIDED_INDEXING_TYPES:
            break;
        case ALL_INT32_INDEXING_TYPES:
            if (i < m_butterfly->publicLength()) {
                JSValue result = m_butterfly->contiguous()[i].get();
                ASSERT(result.isInt32() || !result);
                return result;
            }
            break;
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            if (i < m_butterfly->publicLength())
                return m_butterfly->contiguous()[i].get();
            break;
        case ALL_DOUBLE_INDEXING_TYPES: {
            if (i >= m_butterfly->publicLength())
                break;
            double result = m_butterfly->contiguousDouble()[i];
            if (result != result)
                break;
            return JSValue(JSValue::EncodeAsDouble, result);
        }
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            if (i < m_butterfly->arrayStorage()->vectorLength())
                return m_butterfly->arrayStorage()->m_vector[i].get();
            break;
        default:
            ASSERT_NOT_REACHED();
            break;
        }
        return JSValue();
    }

    JSValue getDirectIndex(ExecState* exec, unsigned i)
    {
        if (JSValue result = tryGetIndexQuickly(i))
            return result;
        PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
        if (methodTable()->getOwnPropertySlotByIndex(this, exec, i, slot))
            return slot.getValue(exec, i);
        return JSValue();
    }

    JSValue getIndex(ExecState* exec, unsigned i) const
    {
        if (JSValue result = tryGetIndexQuickly(i))
            return result;
        return get(exec, i);
    }

    bool canSetIndexQuickly(unsigned i) const
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
        case ALL_UNDECIDED_INDEXING_TYPES:
            return false;
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
        case NonArrayWithArrayStorage:
        case ArrayWithArrayStorage:
            return i < m_butterfly->vectorLength();
        case NonArrayWithSlowPutArrayStorage:
        case ArrayWithSlowPutArrayStorage:
            return i < m_butterfly->arrayStorage()->vectorLength()
                && !!m_butterfly->arrayStorage()->m_vector[i];
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    }

    bool canSetIndexQuicklyForPutDirect(unsigned i) const
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
        case ALL_UNDECIDED_INDEXING_TYPES:
            return false;
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return i < m_butterfly->vectorLength();
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    }

    void setIndexQuickly(JSGlobalData& vm, unsigned i, JSValue v)
    {
        switch (structure()->indexingType()) {
        case ALL_INT32_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isInt32()) {
                convertInt32ToDoubleOrContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            // Fall through to contiguous case.
        }
        case ALL_CONTIGUOUS_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->vectorLength());
            m_butterfly->contiguous()[i].set(vm, this, v);
            if (i >= m_butterfly->publicLength())
                m_butterfly->setPublicLength(i + 1);
            break;
        }
        case ALL_DOUBLE_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isNumber()) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            double value = v.asNumber();
            if (value != value) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            m_butterfly->contiguousDouble()[i] = value;
            if (i >= m_butterfly->publicLength())
                m_butterfly->setPublicLength(i + 1);
            break;
        }
        case ALL_ARRAY_STORAGE_INDEXING_TYPES: {
            ArrayStorage* storage = m_butterfly->arrayStorage();
            WriteBarrier<Unknown>& x = storage->m_vector[i];
            JSValue old = x.get();
            x.set(vm, this, v);
            if (!old) {
                ++storage->m_numValuesInVector;
                if (i >= storage->length())
                    storage->setLength(i + 1);
            }
            break;
        }
        default:
            ASSERT_NOT_REACHED();
        }
    }

    void initializeIndex(JSGlobalData& vm, unsigned i, JSValue v)
    {
        initializeIndex(vm, i, v, structure()->indexingType());
    }

    void initializeIndex(JSGlobalData& vm, unsigned i, JSValue v, IndexingType indexingType)
    {
        switch (indexingType) {
        case ALL_UNDECIDED_INDEXING_TYPES: {
            setIndexQuicklyToUndecided(vm, i, v);
            break;
        }
        case ALL_INT32_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->publicLength());
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isInt32()) {
                convertInt32ToDoubleOrContiguousWhilePerformingSetIndex(vm, i, v);
                break;
            }
            // Fall through.
        }
        case ALL_CONTIGUOUS_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->publicLength());
            ASSERT(i < m_butterfly->vectorLength());
            m_butterfly->contiguous()[i].set(vm, this, v);
            break;
        }
        case ALL_DOUBLE_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->publicLength());
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isNumber()) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            double value = v.asNumber();
            if (value != value) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            m_butterfly->contiguousDouble()[i] = value;
            break;
        }
        case ALL_ARRAY_STORAGE_INDEXING_TYPES: {
            ArrayStorage* storage = m_butterfly->arrayStorage();
            ASSERT(i < storage->length());
            ASSERT(i < storage->m_numValuesInVector);
            storage->m_vector[i].set(vm, this, v);
            break;
        }
        default:
            ASSERT_NOT_REACHED();
        }
    }

    bool hasSparseMap()
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
        case ALL_UNDECIDED_INDEXING_TYPES:
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            return false;
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return !!m_butterfly->arrayStorage()->m_sparseMap;
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    }

    bool inSparseIndexingMode()
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
        case ALL_UNDECIDED_INDEXING_TYPES:
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            return false;
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return m_butterfly->arrayStorage()->inSparseMode();
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    }

    void enterDictionaryIndexingMode(JSGlobalData&);

	// putDirect is effectively an unchecked vesion of 'defineOwnProperty':
	//  - the prototype chain is not consulted
	//  - accessors are not called.
	//  - attributes will be respected (after the call the property will exist with the given attributes)
    //  - the property name is assumed to not be an index.
	bool putDirect(JSGlobalData&, PropertyName, JSValue, unsigned attributes = 0);
	bool putDirect(JSGlobalData&, PropertyName, JSValue, PutPropertySlot&);
	bool putDirectWithoutTransition(JSGlobalData&, PropertyName, JSValue, unsigned attributes = 0);
	bool putDirectNonIndexAccessor(VM&, PropertyName, JSValue, unsigned attributes);
	bool putDirectAccessor(ExecState*, PropertyName, JSValue, unsigned attributes);
	JS_EXPORT_PRIVATE bool putDirectCustomAccessor(VM&, PropertyName, JSValue, unsigned attributes);

	bool putGetter(ExecState*, PropertyName, JSValue, unsigned attributes);
	bool putSetter(ExecState*, PropertyName, JSValue, unsigned attributes);

	JS_EXPORT_PRIVATE bool hasProperty(ExecState*, PropertyName) const;
	JS_EXPORT_PRIVATE bool hasProperty(ExecState*, unsigned propertyName) const;
	bool hasPropertyGeneric(ExecState*, PropertyName, PropertySlot::InternalMethodType) const;
	bool hasPropertyGeneric(ExecState*, unsigned propertyName, PropertySlot::InternalMethodType) const;
	bool hasOwnProperty(ExecState*, PropertyName, PropertySlot&) const;
	bool hasOwnProperty(ExecState*, PropertyName) const;
    bool hasOwnProperty(ExecState*, unsigned) const;

	JS_EXPORT_PRIVATE static bool deleteProperty(JSCell*, ExecState*, PropertyName);
	JS_EXPORT_PRIVATE static bool deletePropertyByIndex(JSCell*, ExecState*, unsigned propertyName);

	JS_EXPORT_PRIVATE static JSValue defaultValue(const JSObject*, ExecState*, PreferredPrimitiveType);
	JSValue ordinaryToPrimitive(ExecState*, PreferredPrimitiveType) const;

	JS_EXPORT_PRIVATE bool hasInstance(ExecState*, JSValue value, JSValue hasInstanceValue);
	bool hasInstance(ExecState*, JSValue);
    static bool defaultHasInstance(ExecState*, JSValue, JSValue prototypeProperty);

	JS_EXPORT_PRIVATE static void getOwnPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);
	JS_EXPORT_PRIVATE static void getOwnNonIndexPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);
	JS_EXPORT_PRIVATE static void getPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);

    JS_EXPORT_PRIVATE static uint32_t getEnumerableLength(ExecState*, JSObject*);
    JS_EXPORT_PRIVATE static void getStructurePropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);
    JS_EXPORT_PRIVATE static void getGenericPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);

	JSValue toPrimitive(ExecState*, PreferredPrimitiveType = NoPreference) const;
	bool getPrimitiveNumber(ExecState*, double& number, JSValue&) const;
	JS_EXPORT_PRIVATE double toNumber(ExecState*) const;
	JS_EXPORT_PRIVATE JSString* toString(ExecState*) const;

    JS_EXPORT_PRIVATE static JSValue toThis(JSCell*, ExecState*, ECMAMode);

	// This get function only looks at the property map.
	ALWAYS_INLINE JSValue getDirect(JSGlobalData& vm, PropertyName propertyName) const
	{
		PropertyOffset offset = structure()->get(vm, propertyName);
		checkOffset(offset, structure()->inlineCapacity());
		return offset != invalidOffset ? getDirect(offset) : JSValue();
	}

	ALWAYS_INLINE JSValue getDirect(VM& vm, PropertyName propertyName, unsigned& attributes) const
    {
        PropertyOffset offset = structure()->get(vm, propertyName, attributes);
        checkOffset(offset, structure()->inlineCapacity());
        return offset != invalidOffset ? getDirect(offset) : JSValue();
    }

	ALWAYS_INLINE PropertyOffset getDirectOffset(JSGlobalData& vm, PropertyName propertyName)
	{
		PropertyOffset offset = structure()->get(vm, propertyName);
		checkOffset(offset, structure()->inlineCapacity());
		return offset;
	}

	ALWAYS_INLINE PropertyOffset getDirectOffset(VM& vm, PropertyName propertyName, unsigned& attributes)
    {
        PropertyOffset offset = structure()->get(vm, propertyName, attributes);
        checkOffset(offset, structure()->inlineCapacity());
        return offset;
    }

	ALWAYS_INLINE bool hasInlineStorage() const { return structure()->hasInlineStorage(); }
	ALWAYS_INLINE ConstPropertyStorage inlineStorageUnsafe() const
	{
		return bitwise_cast<ConstPropertyStorage>(this + 1);
	}
	ALWAYS_INLINE PropertyStorage inlineStorageUnsafe()
	{
		return bitwise_cast<PropertyStorage>(this + 1);
	}
	ALWAYS_INLINE ConstPropertyStorage inlineStorage() const
	{
		ASSERT(hasInlineStorage());
		return inlineStorageUnsafe();
	}
	ALWAYS_INLINE PropertyStorage inlineStorage()
	{
		ASSERT(hasInlineStorage());
		return inlineStorageUnsafe();
	}

    const Butterfly* butterfly() const { return m_butterfly; }
    Butterfly* butterfly() { return m_butterfly; }

    ALWAYS_INLINE ConstPropertyStorage outOfLineStorage() const { return m_butterfly->propertyStorage(); }
    ALWAYS_INLINE PropertyStorage outOfLineStorage() { return m_butterfly->propertyStorage(); }

    ALWAYS_INLINE const WriteBarrierBase<Unknown>* locationForOffset(PropertyOffset offset) const
	{
		if (isInlineOffset(offset))
			return &inlineStorage()[offsetInInlineStorage(offset)];
		return &outOfLineStorage()[offsetInOutOfLineStorage(offset)];
	}

    ALWAYS_INLINE WriteBarrierBase<Unknown>* locationForOffset(PropertyOffset offset)
	{
		if (isInlineOffset(offset))
			return &inlineStorage()[offsetInInlineStorage(offset)];
		return &outOfLineStorage()[offsetInOutOfLineStorage(offset)];
	}

    PropertyOffset offsetForLocation(WriteBarrierBase<Unknown>* location) const
    {
        PropertyOffset result;
        size_t offsetInInlineStorage = location - inlineStorageUnsafe();
        if (offsetInInlineStorage < static_cast<size_t>(firstOutOfLineOffset))
            result = offsetInInlineStorage;
        else
            result = outOfLineStorage() - location + (firstOutOfLineOffset - 1);
        validateOffset(result, structure()->inlineCapacity());
        return result;
    }

	void transitionTo(JSGlobalData&, Structure*);

	ALWAYS_INLINE bool hasCustomProperties() { return structure()->didTransition(); }
	ALWAYS_INLINE bool hasGetterSetterProperties() { return structure()->hasGetterSetterProperties(); }
	bool hasCustomGetterSetterProperties() { return structure()->hasCustomGetterSetterProperties(); }

	// putOwnDataProperty has 'put' like semantics, however this method:
	//  - assumes the object contains no own getter/setter properties.
	//  - provides no special handling for __proto__
	//  - does not walk the prototype chain (to check for accessors or non-writable properties).
	// This is used by JSLexicalEnvironment.
	bool putOwnDataProperty(JSGlobalData&, PropertyName, JSValue, PutPropertySlot&);
	bool putOwnDataPropertyMayBeIndex(ExecState*, PropertyName, JSValue, PutPropertySlot&);

	// Fast access to known property offsets.
	ALWAYS_INLINE JSValue getDirect(PropertyOffset offset) const { return locationForOffset(offset)->get(); }
	void putDirect(JSGlobalData& vm, PropertyOffset offset, JSValue value) { locationForOffset(offset)->set(vm, this, value); }
	void putDirectUndefined(PropertyOffset offset) { locationForOffset(offset)->setUndefined(); }

	JS_EXPORT_PRIVATE bool putDirectNativeIntrinsicGetter(VM&, JSGlobalObject*, Identifier, NativeFunction, Intrinsic, unsigned attributes);
	bool putDirectNativeFunction(VM&, JSGlobalObject*, PropertyName, unsigned functionLength, NativeFunction, Intrinsic, unsigned attributes);
	bool putDirectNativeFunction(VM&, JSGlobalObject*, Identifier, unsigned functionLength, NativeFunction, Intrinsic, unsigned attributes);
	bool putDirectNativeFunctionWithoutTransition(VM&, JSGlobalObject*, Identifier, unsigned functionLength, NativeFunction, Intrinsic, unsigned attributes);

    JSFunction* putDirectBuiltinFunction(VM&, JSGlobalObject*, PropertyName, FunctionExecutable*, unsigned attributes);
    JSFunction* putDirectBuiltinFunction(VM&, JSGlobalObject*, Identifier, FunctionExecutable*, unsigned attributes);
    JSFunction* putDirectBuiltinFunctionWithoutTransition(VM&, JSGlobalObject*, Identifier, FunctionExecutable*, unsigned attributes);

	JS_EXPORT_PRIVATE static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool shouldThrow);

	bool isGlobalObject() const;
    bool isJSLexicalEnvironment() const;
    bool isGlobalLexicalEnvironment() const;
	bool isErrorInstance() const;

	void seal(JSGlobalData&);
	void freeze(JSGlobalData&);
    JS_EXPORT_PRIVATE static bool preventExtensions(JSObject*, ExecState*);
    JS_EXPORT_PRIVATE static bool isExtensible(JSObject*, ExecState*);
	bool isSealed(JSGlobalData& vm) { return structure()->isSealed(vm); }
	bool isFrozen(JSGlobalData& vm) { return structure()->isFrozen(vm); }
private:
    ALWAYS_INLINE bool isExtensibleImpl() { return isStructureExtensible(); }
public:
    // You should only call isStructureExtensible() when:
    // - Performing this check in a way that isn't described in the specification
    //   as calling the virtual [[IsExtensible]] trap.
    // - When you're guaranteed that object->methodTable()->isExtensible isn't
    //   overridden.
    ALWAYS_INLINE bool isStructureExtensible() { return structure()->isStructureExtensible(); }
    // You should call this when performing [[IsExtensible]] trap in a place
    // that is described in the specification. This performs the fully virtual
    // [[IsExtensible]] trap.
    bool isExtensible(ExecState*);
	bool indexingShouldBeSparse()
	{
		return !isStructureExtensible()
			|| structure()->typeInfo().interceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero();
	}

	bool staticPropertiesReified() { return structure()->staticPropertiesReified(); }
    void reifyAllStaticProperties(JSGlobalData&);

    JS_EXPORT_PRIVATE Butterfly* growOutOfLineStorage(JSGlobalData&, size_t oldSize, size_t newSize);
    void setButterflyWithoutChangingStructure(Butterfly*); // You probably don't want to call this.

    void setStructure(JSGlobalData&, Structure*);
    void setStructureAndButterfly(JSGlobalData&, Structure*, Butterfly*);
	void setStructureAndReallocateStorageIfNecessary(JSGlobalData&, unsigned oldCapacity, Structure*);
	void setStructureAndReallocateStorageIfNecessary(JSGlobalData&, Structure*);

    JS_EXPORT_PRIVATE void convertToDictionary(JSGlobalData&);

	void flattenDictionaryObject(JSGlobalData& vm)
	{
		structure()->flattenDictionaryStructure(vm, this);
	}

    void shiftButterflyAfterFlattening(JSGlobalData&, size_t outOfLineCapacityBefore, size_t outOfLineCapacityAfter);

	ALWAYS_INLINE JSGlobalObject* globalObject() const
	{
		ASSERT(structure()->globalObject());
		ASSERT(!isGlobalObject() || ((JSObject*)structure()->globalObject()) == this);
		return structure()->globalObject();
	}

    // Does everything possible to return the global object. If it encounters an object
    // that does not have a global object, it returns 0 instead (for example
    // JSNotAnObject).

    void switchToSlowPutArrayStorage(JSGlobalData&);

    // The receiver is the prototype in this case. The following:
    //
    // asObject(foo->structure()->storedPrototype())->attemptToInterceptPutByIndexOnHoleForPrototype(...)
    //
    // is equivalent to:
    //
    // foo->attemptToInterceptPutByIndexOnHole(...);
    bool attemptToInterceptPutByIndexOnHoleForPrototype(ExecState*, JSGlobalData&, JSValue thisValue, unsigned propertyName, JSValue, bool shouldThrow, bool& putResult);

    // Returns 0 if int32 storage cannot be created - either because
    // indexing should be sparse, we're having a bad time, or because
    // we already have a more general form of storage (double,
    // contiguous, array storage).
    ContiguousJSValues ensureInt32(JSGlobalData& vm)
    {
        if (LIKELY(hasInt32(structure()->indexingType())))
            return m_butterfly->contiguousInt32();

        return ensureInt32Slow(vm);
    }

    // Returns 0 if double storage cannot be created - either because
    // indexing should be sparse, we're having a bad time, or because
    // we already have a more general form of storage (contiguous,
    // or array storage).
    ContiguousDoubles ensureDouble(JSGlobalData& vm)
    {
        if (LIKELY(hasDouble(structure()->indexingType())))
            return m_butterfly->contiguousDouble();

        return ensureDoubleSlow(vm);
    }

    // Returns 0 if contiguous storage cannot be created - either because
    // indexing should be sparse or because we're having a bad time.
    ContiguousJSValues ensureContiguous(JSGlobalData& vm)
    {
        if (LIKELY(hasContiguous(structure()->indexingType())))
            return m_butterfly->contiguous();

        return ensureContiguousSlow(vm);
    }

    // Ensure that the object is in a mode where it has array storage. Use
    // this if you're about to perform actions that would have required the
    // object to be converted to have array storage, if it didn't have it
    // already.
    ArrayStorage* ensureArrayStorage(JSGlobalData& vm)
    {
        if (LIKELY(hasAnyArrayStorage(structure()->indexingType())))
            return m_butterfly->arrayStorage();

        return ensureArrayStorageSlow(vm);
    }

	static size_t offsetOfInlineStorage();

    static ptrdiff_t butterflyOffset()
    {
        return OBJECT_OFFSETOF(JSObject, m_butterfly);
    }

    void* butterflyAddress()
    {
        return &m_butterfly;
    }

    JSValue getMethod(ExecState* exec, CallData& callData, CallType& callType, const Identifier& ident, const String& errorMessage);

    DECLARE_INFO;

protected:
	ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		ASSERT(inherits(info()));
		ASSERT(!structure()->outOfLineCapacity());
		ASSERT(structure()->isEmpty());
		ASSERT(getPrototypeDirect().isNull() || Heap::heap(this) == Heap::heap(getPrototypeDirect()));
		ASSERT(structure()->isObject());
		ASSERT(classInfo());
	}

	static ALWAYS_INLINE Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue prototype)
	{
		return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
	}

	// To instantiate objects you likely want JSFinalObject, below.
	// To create derived types you likely want JSNonFinalObject, below.
	JSObject(JSGlobalData&, Structure*, Butterfly* = 0);

    void visitButterfly(SlotVisitor&, Butterfly*, size_t storageSize);
    void copyButterfly(CopyVisitor&, Butterfly*, size_t storageSize);

    // Call this if you know that the object is in a mode where it has array
    // storage. This will assert otherwise.
    ALWAYS_INLINE ArrayStorage* arrayStorage()
    {
        ASSERT(hasAnyArrayStorage(structure()->indexingType()));
        return m_butterfly->arrayStorage();
    }

    // Call this if you want to predicate some actions on whether or not the
    // object is in a mode where it has array storage.
    ArrayStorage* arrayStorageOrNull()
    {
        switch (structure()->indexingType()) {
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return m_butterfly->arrayStorage();

        default:
            return 0;
        }
    }

    size_t butterflyTotalSize();
    size_t butterflyPreCapacity();

    Butterfly* createInitialUndecided(JSGlobalData&, unsigned length);
    ContiguousJSValues createInitialInt32(JSGlobalData&, unsigned length);
    ContiguousDoubles createInitialDouble(JSGlobalData&, unsigned length);
    ContiguousJSValues createInitialContiguous(JSGlobalData&, unsigned length);

    void convertUndecidedForValue(JSGlobalData&, JSValue);
    void createInitialForValueAndSet(VM&, unsigned index, JSValue);
    void convertInt32ForValue(JSGlobalData&, JSValue);

    static Butterfly* createArrayStorageButterfly(JSGlobalData&, JSCell* intendedOwner, Structure*, unsigned length, unsigned vectorLength, Butterfly* oldButterfly = 0);
    ArrayStorage* createArrayStorage(JSGlobalData&, unsigned length, unsigned vectorLength);
    ArrayStorage* createInitialArrayStorage(JSGlobalData&);

    ContiguousJSValues convertUndecidedToInt32(JSGlobalData&);
    ContiguousDoubles convertUndecidedToDouble(JSGlobalData&);
    ContiguousJSValues convertUndecidedToContiguous(JSGlobalData&);
    ArrayStorage* convertUndecidedToArrayStorage(JSGlobalData&, NonPropertyTransition);
    ArrayStorage* convertUndecidedToArrayStorage(JSGlobalData&);

    ContiguousDoubles convertInt32ToDouble(JSGlobalData&);
    ContiguousJSValues convertInt32ToContiguous(JSGlobalData&);
    ArrayStorage* convertInt32ToArrayStorage(JSGlobalData&, NonPropertyTransition);
    ArrayStorage* convertInt32ToArrayStorage(JSGlobalData&);

    ContiguousJSValues convertDoubleToContiguous(JSGlobalData&);
    ArrayStorage* convertDoubleToArrayStorage(JSGlobalData&, NonPropertyTransition);
    ArrayStorage* convertDoubleToArrayStorage(JSGlobalData&);

	ArrayStorage* convertContiguousToArrayStorage(JSGlobalData&, NonPropertyTransition);
	ArrayStorage* convertContiguousToArrayStorage(JSGlobalData&);

    ArrayStorage* ensureArrayStorageExistsAndEnterDictionaryIndexingMode(JSGlobalData&);

    bool defineOwnNonIndexProperty(ExecState*, PropertyName, const PropertyDescriptor&, bool throwException);

    template<IndexingType indexingShape>
    bool putByIndexBeyondVectorLengthWithoutAttributes(ExecState*, unsigned propertyName, JSValue);
    bool putByIndexBeyondVectorLengthWithArrayStorage(ExecState*, unsigned propertyName, JSValue, bool shouldThrow, ArrayStorage*);

    bool ensureVectorLength(JSGlobalData&, unsigned newLength);
    void deallocateSparseIndexMap();
    bool defineOwnIndexedProperty(ExecState*, unsigned, const PropertyDescriptor&, bool throwException);
    SparseArrayValueMap* allocateSparseIndexMap(JSGlobalData&);

    void notifyPresenceOfIndexedAccessors(JSGlobalData&);

    bool attemptToInterceptPutByIndexOnHole(ExecState*, VM&, unsigned index, JSValue, bool shouldThrow, bool& putResult);

    // Call this if you want setIndexQuickly to succeed and you're sure that
    // the array is contiguous.
    bool ensureLength(JSGlobalData& vm, unsigned length)
	{
    	RELEASE_ASSERT(length <= MAX_STORAGE_VECTOR_LENGTH);
		ASSERT(hasContiguous(structure()->indexingType()) || hasInt32(structure()->indexingType()) || hasDouble(structure()->indexingType()) || hasUndecided(structure()->indexingType()));

        bool result = true;
		if (m_butterfly->vectorLength() < length)
			result = ensureLengthSlow(vm, length);

        if (m_butterfly->publicLength() < length)
            m_butterfly->setPublicLength(length);

        return result;
    }

    // Call this if you want to shrink the butterfly backing store, and you're
    // sure that the array is contiguous.
    void reallocateAndShrinkButterfly(JSGlobalData&, unsigned length);

    template<IndexingType indexingShape>
    unsigned countElements(Butterfly*);

    // This is relevant to undecided, int32, double, and contiguous.
    unsigned countElements();

private:
	friend class LLIntOffsetsExtractor;

	// Nobody should ever ask any of these questions on something already known to be a JSObject.
	using JSCell::isAPIValueWrapper;
	using JSCell::isGetterSetter;
	void getObject();
	void getString(ExecState* exec);
	void isObject();
	void isString();

    Butterfly* createInitialIndexedStorage(JSGlobalData&, unsigned length, size_t elementSize);

    ArrayStorage* enterDictionaryIndexingModeWhenArrayStorageAlreadyExists(JSGlobalData&, ArrayStorage*);

	template<PutMode>
	bool putDirectInternal(JSGlobalData&, PropertyName, JSValue, unsigned attr, PutPropertySlot&);
	bool canPerformFastPutInline(ExecState* exec, JSGlobalData&, PropertyName);

    JS_EXPORT_PRIVATE NEVER_INLINE bool putInlineSlow(ExecState*, PropertyName, JSValue, PutPropertySlot&);

    bool getNonIndexPropertySlot(ExecState*, PropertyName, PropertySlot&);
	bool getOwnNonIndexPropertySlot(JSGlobalData&, Structure&, PropertyName, PropertySlot&);
	JS_EXPORT_PRIVATE void fillGetterPropertySlot(PropertySlot&, JSValue, unsigned, PropertyOffset);
	void fillCustomGetterPropertySlot(PropertySlot&, JSValue, unsigned, Structure&);

    JS_EXPORT_PRIVATE bool getOwnStaticPropertySlot(VM&, PropertyName, PropertySlot&);
	JS_EXPORT_PRIVATE const HashTableValue* findPropertyHashEntry(PropertyName) const;

	bool putIndexedDescriptor(ExecState*, SparseArrayEntry*, const PropertyDescriptor&, PropertyDescriptor& old);

	bool putByIndexBeyondVectorLength(ExecState*, unsigned propertyName, JSValue, bool shouldThrow);
    bool putDirectIndexBeyondVectorLengthWithArrayStorage(ExecState*, unsigned propertyName, JSValue, unsigned attributes, PutDirectIndexMode, ArrayStorage*);
    JS_EXPORT_PRIVATE bool putDirectIndexSlowOrBeyondVectorLength(ExecState*, unsigned propertyName, JSValue, unsigned attributes, PutDirectIndexMode);

    unsigned getNewVectorLength(unsigned currentVectorLength, unsigned currentLength, unsigned desiredLength);
    unsigned getNewVectorLength(unsigned desiredLength);

    ArrayStorage* constructConvertedArrayStorageWithoutCopyingElements(JSGlobalData&, unsigned neededLength);

    JS_EXPORT_PRIVATE void setIndexQuicklyToUndecided(JSGlobalData&, unsigned index, JSValue);
    JS_EXPORT_PRIVATE void convertInt32ToDoubleOrContiguousWhilePerformingSetIndex(JSGlobalData&, unsigned index, JSValue);
    JS_EXPORT_PRIVATE void convertDoubleToContiguousWhilePerformingSetIndex(JSGlobalData&, unsigned index, JSValue);

    bool ensureLengthSlow(JSGlobalData&, unsigned length);

    ContiguousJSValues ensureInt32Slow(JSGlobalData&);
    ContiguousDoubles ensureDoubleSlow(JSGlobalData&);
    ContiguousJSValues ensureContiguousSlow(JSGlobalData&);
    ArrayStorage* ensureArrayStorageSlow(JSGlobalData&);

protected:
    Butterfly* m_butterfly;
};

// JSNonFinalObject is a type of JSObject that has some internal storage,
// but also preserves some space in the collector cell for additional
// data members in derived types.
class JSNonFinalObject : public JSObject {
	friend class JSObject;

public:
	typedef JSObject Base;

	static ALWAYS_INLINE Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue prototype)
	{
		return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
	}

protected:
    explicit ALWAYS_INLINE JSNonFinalObject(JSGlobalData& vm, Structure* structure, Butterfly* butterfly = 0)
        : JSObject(vm, structure, butterfly)
    {
    }

    ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		ASSERT(!this->structure()->hasInlineStorage());
		ASSERT(classInfo());
	}
};

class JSFinalObject;

// JSFinalObject is a type of JSObject that contains sufficent internal
// storage to fully make use of the colloctor cell containing it.
class JSFinalObject : public JSObject {
	friend class JSObject;

public:
	typedef JSObject Base;

    static size_t allocationSize(size_t inlineCapacity)
    {
        return sizeof(JSObject) + inlineCapacity * sizeof(WriteBarrierBase<Unknown>);
    }

    static inline const TypeInfo typeInfo() { return TypeInfo(FinalObjectType, StructureFlags); }
    static const IndexingType defaultIndexingType = NonArray;

    static const unsigned defaultSize = 64;
    static inline unsigned defaultInlineCapacity()
    {
        return (defaultSize - allocationSize(0)) / sizeof(WriteBarrier<Unknown>);
    }

    static const unsigned maxSize = 512;
    static inline unsigned maxInlineCapacity()
    {
        return (maxSize - allocationSize(0)) / sizeof(WriteBarrier<Unknown>);
    }

	static JSFinalObject* create(ExecState*, Structure*);
    static JSFinalObject* create(VM&, Structure*);
	static ALWAYS_INLINE Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue prototype, unsigned inlineCapacity)
	{
		return Structure::create(vm, globalObject, prototype, typeInfo(), info(), defaultIndexingType, inlineCapacity);
	}

	JS_EXPORT_PRIVATE static void visitChildren(JSCell*, SlotVisitor&);

    DECLARE_INFO;

protected:
	void visitChildrenCommon(SlotVisitor&);

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		ASSERT(structure()->totalStorageCapacity() == structure()->inlineCapacity());
		ASSERT(classInfo());
	}

private:
	friend class LLIntOffsetsExtractor;

	explicit ALWAYS_INLINE JSFinalObject(JSGlobalData& vm, Structure* structure)
		: JSObject(vm, structure)
	{
	}

	static const unsigned StructureFlags = JSObject::StructureFlags;
};

JS_EXPORT_PRIVATE EncodedJSValue JSC_HOST_CALL objectPrivateFuncInstanceOf(ExecState*);

ALWAYS_INLINE JSFinalObject* JSFinalObject::create(VM& vm, Structure* structure)
{
    JSFinalObject* finalObject = new (
    	NotNull,
		allocateCell<JSFinalObject>(
			vm.heap,
			allocationSize(structure->inlineCapacity())
		)
	) JSFinalObject(vm, structure);
    finalObject->finishCreation(vm);
    return finalObject;
}

ALWAYS_INLINE JSFinalObject* JSFinalObject::create(ExecState* exec, Structure* structure)
{
	return create(exec->vm(), structure);
}

ALWAYS_INLINE bool isJSFinalObject(const JSCell* cell)
{
    return cell->classInfo() == JSFinalObject::info();
}

ALWAYS_INLINE bool isJSFinalObject(JSValue value)
{
    return value.isCell() && isJSFinalObject(value.asCell());
}

ALWAYS_INLINE size_t JSObject::offsetOfInlineStorage()
{
    return sizeof(JSObject);
}

ALWAYS_INLINE bool JSObject::isGlobalObject() const
{
    return structure()->typeInfo().type() == GlobalObjectType;
}

inline bool JSObject::isJSLexicalEnvironment() const
{
	JSType type = structure()->typeInfo().type();
    return type == LexicalEnvironmentType || type == ModuleEnvironmentType;
}

inline bool JSObject::isGlobalLexicalEnvironment() const
{
    return structure()->typeInfo().type() == GlobalLexicalEnvironmentType;
}

ALWAYS_INLINE bool JSObject::isErrorInstance() const
{
    return structure()->typeInfo().type() == ErrorInstanceType;
}

ALWAYS_INLINE void JSObject::setStructureAndButterfly(JSGlobalData& vm, Structure* structure, Butterfly* butterfly)
{
    m_butterfly = butterfly;
    setStructure(vm, structure);
}

ALWAYS_INLINE void JSObject::setStructure(VM& vm, Structure* structure)
{
    ASSERT(structure);
    ASSERT(!m_butterfly == !(structure->outOfLineCapacity() || structure->hasIndexingHeader(this)));
    JSCell::setStructure(vm, structure);
}

inline void JSObject::setButterflyWithoutChangingStructure(Butterfly* butterfly)
{
    m_butterfly = butterfly;
}

ALWAYS_INLINE CallType getCallData(JSValue value, CallData& callData)
{
    CallType result = CallType::None;

    if(LIKELY(value.isCell())){
    	JSCell* cell = value.asCell();
    	if(LIKELY(cell->classInfo() != NULL))
    		result = cell->methodTable()->getCallData(cell, callData);
    }
    ASSERT(result == CallType::None || value.isValidCallee());
    return result;
}

ALWAYS_INLINE ConstructType getConstructData(JSValue value, ConstructData& constructData)
{
    ConstructType result = value.isCell() ? value.asCell()->methodTable()->getConstructData(value.asCell(), constructData) : ConstructType::None;
    ASSERT(result == ConstructType::None || value.isValidCallee());
    return result;
}

ALWAYS_INLINE JSObject* asObject(JSCell* cell)
{
    ASSERT(cell->isObject());
    return jsCast<JSObject*>(cell);
}

ALWAYS_INLINE JSObject* asObject(JSValue value)
{
    return asObject(value.asCell());
}

ALWAYS_INLINE JSObject::JSObject(JSGlobalData& vm, Structure* structure, Butterfly* butterfly)
    : JSCell(vm, structure)
    , m_butterfly(butterfly)
{
}

ALWAYS_INLINE JSValue JSObject::getPrototypeDirect() const
{
    return structure()->storedPrototype();
}

inline JSValue JSObject::getPrototype(VM& vm, ExecState* exec)
{
    auto getPrototypeMethod = methodTable(vm)->getPrototype;
    MethodTable::GetPrototypeFunctionPtr defaultGetPrototype = JSObject::getPrototype;
    if (LIKELY(getPrototypeMethod == defaultGetPrototype))
        return getPrototypeDirect();
    return getPrototypeMethod(this, exec);
}

// It is safe to call this method with a PropertyName that is actually an index,
// but if so will always return false (doesn't search index storage).
ALWAYS_INLINE bool JSObject::getOwnNonIndexPropertySlot(JSGlobalData& vm, Structure& structure, PropertyName propertyName, PropertySlot& slot)
{
    unsigned attributes;
	PropertyOffset offset = structure.get(vm, propertyName, attributes);
    if (!isValidOffset(offset)) {
        if (!TypeInfo::hasStaticPropertyTable(this->structure(vm)->typeInfo().inlineTypeFlags()))
            return false;
        return getOwnStaticPropertySlot(vm, propertyName, slot);
    }

    // getPropertySlot relies on this method never returning index properties!
    ASSERT(!parseIndex(propertyName));

	JSValue value = getDirect(offset);
	if (value.isCell()) {
		ASSERT(value);
		if(UNLIKELY(!value))
			return false;
		JSCell* cell = value.asCell();
		JSType type = cell->structure()->typeInfo().type();
		switch (type) {
		case GetterSetterType:
			fillGetterPropertySlot(slot, value, attributes, offset);
			return true;
		case CustomGetterSetterType:
			fillCustomGetterPropertySlot(slot, value, attributes, structure);
			return true;
		default:
			break;
		}
	}

	slot.setValue(this, attributes, value, offset);
	return true;
}

ALWAYS_INLINE void JSObject::fillCustomGetterPropertySlot(PropertySlot& slot, JSValue customGetterSetter, unsigned attributes, Structure& structure)
{
    ASSERT(attributes & PropertyAttribute::CustomAccessorOrValue);
    if (structure.isUncacheableDictionary()) {
        slot.setCustom(this, attributes, jsCast<CustomGetterSetter*>(customGetterSetter)->getter());
        return;
    }

    // This access is cacheable because Structure requires an attributeChangeTransition
    // if this property stops being an accessor.
    slot.setCacheableCustom(this, attributes, jsCast<CustomGetterSetter*>(customGetterSetter)->getter());
}

// It may seem crazy to inline a function this large, especially a virtual function,
// but it makes a big difference to property lookup that derived classes can inline their
// base class call to this.
ALWAYS_INLINE bool JSObject::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    VM& vm = exec->vm();
    Structure& structure = *object->structure(vm);
    if(object->getOwnNonIndexPropertySlot(vm, structure, propertyName, slot))
    	return true;
    if (Optional<uint32_t> index = parseIndex(propertyName))
        return getOwnPropertySlotByIndex(object, exec, index.value(), slot);
    return false;
}

// It may seem crazy to inline a function this large but it makes a big difference
// since this is function very hot in variable lookup
ALWAYS_INLINE bool JSObject::getPropertySlot(ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    VM& vm = exec->vm();
    JSObject* object = this;
    while (true) {
        Structure& structure = *object->structure(vm);
        if (UNLIKELY(structure.typeInfo().overridesGetOwnPropertySlot())) {
            // If propertyName is an index then we may have missed it (as this loop is using
            // getOwnNonIndexPropertySlot), so we cannot safely call the overridden getOwnPropertySlot
            // (lest we return a property from a prototype that is shadowed). Check now for an index,
            // if so we need to start afresh from this object.
            if (Optional<uint32_t> index = parseIndex(propertyName))
                return getPropertySlot(exec, index.value(), slot);
            // Safe to continue searching from current position; call getNonIndexPropertySlot to avoid
            // parsing the int again.
            return object->getNonIndexPropertySlot(exec, propertyName, slot);
        }
        ASSERT(object->structure()->typeInfo().type() != ProxyObjectType);
        if (object->getOwnNonIndexPropertySlot(vm, structure, propertyName, slot))
            return true;
        JSValue prototype = structure.storedPrototype();
        if (!prototype.isObject())
        	break;
        object = asObject(prototype);
    }

    if (Optional<uint32_t> index = parseIndex(propertyName))
        return getPropertySlot(exec, index.value(), slot);
    return false;
}

ALWAYS_INLINE bool JSObject::getPropertySlot(ExecState* exec, unsigned propertyName, PropertySlot& slot)
{
    VM& vm = exec->vm();
    JSObject* object = this;
    MethodTable::GetPrototypeFunctionPtr defaultGetPrototype = JSObject::getPrototype;
    while (true) {
    	bool hasSlot = object->methodTable()->getOwnPropertySlotByIndex(object, exec, propertyName, slot);
    	RETURN_IF_EXCEPTION(vm, false);
		if (hasSlot)
			return true;
        JSValue prototype;
        if (LIKELY(object->methodTable()->getPrototype == defaultGetPrototype || slot.internalMethodType() == PropertySlot::InternalMethodType::VMInquiry))
            prototype = object->structure()->storedPrototype();
        else {
            prototype = object->getPrototype(vm, exec);
        	RETURN_IF_EXCEPTION(vm, false);
        }
        if (!prototype.isObject())
            return false;
        object = asObject(prototype);
    }
}

ALWAYS_INLINE bool JSObject::getNonIndexPropertySlot(ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    // This method only supports non-index PropertyNames.
    ASSERT(!parseIndex(propertyName));

    JSGlobalData& vm = exec->vm();
    JSObject* object = this;
    MethodTable::GetPrototypeFunctionPtr defaultGetPrototype = JSObject::getPrototype;
    while (true) {
        Structure& structure = *object->structure();
        if (LIKELY(!structure.typeInfo().overridesGetOwnPropertySlot())) {
            if (object->getOwnNonIndexPropertySlot(vm, structure, propertyName, slot))
                return true;
        } else {
        	bool hasSlot = object->methodTable()->getOwnPropertySlot(object, exec, propertyName, slot);
        	RETURN_IF_EXCEPTION(vm, false);
			if (hasSlot)
				return true;
        }
        JSValue prototype;
        if (LIKELY(object->methodTable()->getPrototype == defaultGetPrototype || slot.internalMethodType() == PropertySlot::InternalMethodType::VMInquiry))
            prototype = structure.storedPrototype();
        else {
            prototype = object->getPrototype(vm, exec);
        	RETURN_IF_EXCEPTION(vm, false);
        }
        if (!prototype.isObject())
            return false;
        object = asObject(prototype);
    }
}

ALWAYS_INLINE JSValue JSObject::get(ExecState* exec, PropertyName propertyName) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
    if (const_cast<JSObject*>(this)->getPropertySlot(exec, propertyName, slot))
        return slot.getValue(exec, propertyName);
    
    return jsUndefined();
}

ALWAYS_INLINE JSValue JSObject::get(ExecState* exec, unsigned propertyName) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
    if (const_cast<JSObject*>(this)->getPropertySlot(exec, propertyName, slot))
        return slot.getValue(exec, propertyName);

    return jsUndefined();
}

#if 0
template<typename CallbackWhenNoException>
ALWAYS_INLINE typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type JSObject::getPropertySlot(ExecState* exec, PropertyName propertyName, CallbackWhenNoException callback) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
    return getPropertySlot(exec, propertyName, slot, callback);
}

template<typename CallbackWhenNoException>
ALWAYS_INLINE typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type JSObject::getPropertySlot(ExecState* exec, PropertyName propertyName, PropertySlot& slot, CallbackWhenNoException callback) const
{
	VM& vm = exec->vm();
    bool found = const_cast<JSObject*>(this)->getPropertySlot(exec, propertyName, slot);
    if (UNLIKELY(vm.exception()))
        return { };
    return callback(found, slot);
}
#endif

template<JSObject::PutMode mode>
ALWAYS_INLINE bool JSObject::putDirectInternal(JSGlobalData& vm, PropertyName propertyName, JSValue value, unsigned attributes, PutPropertySlot& slot)
{
    ASSERT(value);
    ASSERT(value.isGetterSetter() == !!(attributes & Accessor));
    ASSERT(value.isCustomGetterSetter() == !!(attributes & PropertyAttribute::CustomAccessorOrValue));
	ASSERT(!Heap::heap(value) || Heap::heap(value) == Heap::heap(this));
    ASSERT(!parseIndex(propertyName));

    if (structure()->isDictionary()) {
        unsigned currentAttributes;
        PropertyOffset offset = structure()->get(vm, propertyName, currentAttributes);
        if (offset != invalidOffset) {
            if ((mode == PutModePut) && currentAttributes & ReadOnly)
                return false;

            putDirect(vm, offset, value);
            structure()->didReplaceProperty(vm, offset);

            if ((attributes & Accessor) != (currentAttributes & Accessor) || (attributes & CustomAccessorOrValue) != (currentAttributes & CustomAccessorOrValue)) {
                ASSERT(!(attributes & ReadOnly));
                setStructure(vm, Structure::attributeChangeTransition(vm, structure(), propertyName, attributes));
            }
            else
    			slot.setExistingProperty(this, offset);

            return true;
        }

        if ((mode == PutModePut) && !isStructureExtensible())
            return false;

        DeferGC deferGC(vm.heap);
        Butterfly* newButterfly = m_butterfly;
        if (UNLIKELY(structure()->putWillGrowOutOfLineStorage()))
            newButterfly = growOutOfLineStorage(vm, structure()->outOfLineCapacity(), structure()->suggestedNewOutOfLineStorageCapacity());

        offset = structure()->addPropertyWithoutTransition(vm, propertyName, attributes);
        setStructureAndButterfly(vm, structure(), newButterfly);

        validateOffset(offset);
        ASSERT(structure()->isValidOffset(offset));
        putDirect(vm, offset, value);
        slot.setNewProperty(this, offset);
        if (attributes & ReadOnly)
            structure()->setContainsReadOnlyProperties();
        return true;
    }

    PropertyOffset offset;
    size_t currentCapacity = structure()->outOfLineCapacity();
    if (Structure* structure = Structure::addPropertyTransitionToExistingStructure(this->structure(), propertyName.uid(), attributes, offset)) {
        DeferGC deferGC(vm.heap);
    	Butterfly* newButterfly = m_butterfly;
		if (currentCapacity != structure->outOfLineCapacity())
			newButterfly = growOutOfLineStorage(vm, currentCapacity, structure->outOfLineCapacity());

		validateOffset(offset);
		ASSERT(structure->isValidOffset(offset));

        // This assertion verifies that the concurrent GC won't read garbage if the concurrentGC
        // is running at the same time we put without transitioning.
        ASSERT(!getDirect(offset) || !JSValue::encode(getDirect(offset)));

        setStructureAndButterfly(vm, structure, newButterfly);

        putDirect(vm, offset, value);
        slot.setNewProperty(this, offset);
        return true;
    }

    unsigned currentAttributes;
    offset = structure()->get(vm, propertyName, currentAttributes);
    if (offset != invalidOffset) {
        if ((mode == PutModePut) && currentAttributes & ReadOnly)
            return false;

        structure()->didReplaceProperty(vm, offset);
        putDirect(vm, offset, value);

        if ((attributes & Accessor) != (currentAttributes & Accessor) || (attributes & CustomAccessorOrValue) != (currentAttributes & CustomAccessorOrValue)) {
            ASSERT(!(attributes & ReadOnly));
            setStructure(vm, Structure::attributeChangeTransition(vm, structure(), propertyName, attributes));
        }
        else
            slot.setExistingProperty(this, offset);

        return true;
    }

    if ((mode == PutModePut) && !isStructureExtensible())
        return false;

    // We want the structure transition watchpoint to fire after this object has switched
    // structure. This allows adaptive watchpoints to observe if the new structure is the one
    // we want.
    DeferredStructureTransitionWatchpointFire deferredWatchpointFire;

    Structure* structure = Structure::addNewPropertyTransition(vm, this->structure(), propertyName, attributes, offset, slot.context(), &deferredWatchpointFire);
    
    validateOffset(offset);
    ASSERT(structure->isValidOffset(offset));
    setStructureAndReallocateStorageIfNecessary(vm, structure);

    // This assertion verifies that the concurrent GC won't read garbage if the concurrentGC
    // is running at the same time we put without transitioning.
    ASSERT(!getDirect(offset) || !JSValue::encode(getDirect(offset)));
    putDirect(vm, offset, value);
    slot.setNewProperty(this, offset);
    if (attributes & ReadOnly)
        structure->setContainsReadOnlyProperties();
    return true;
}

inline void JSObject::setStructureAndReallocateStorageIfNecessary(JSGlobalData& vm, unsigned oldCapacity, Structure* newStructure)
{
    ASSERT(oldCapacity <= newStructure->outOfLineCapacity());
    
    if (oldCapacity == newStructure->outOfLineCapacity()) {
        setStructure(vm, newStructure);
        return;
    }

    DeferGC deferGC(vm.heap);
    Butterfly* newButterfly = growOutOfLineStorage(vm, oldCapacity, newStructure->outOfLineCapacity());
    setStructureAndButterfly(vm, newStructure, newButterfly);
}

ALWAYS_INLINE void JSObject::setStructureAndReallocateStorageIfNecessary(JSGlobalData& vm, Structure* newStructure)
{
    setStructureAndReallocateStorageIfNecessary(
        vm, structure()->outOfLineCapacity(), newStructure);
}

inline bool JSObject::putOwnDataProperty(JSGlobalData& vm, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    ASSERT(value);
    ASSERT(!Heap::heap(value) || Heap::heap(value) == Heap::heap(this));
    ASSERT(!structure()->hasGetterSetterProperties());
    ASSERT(!structure()->hasCustomGetterSetterProperties());

    return putDirectInternal<PutModePut>(vm, propertyName, value, 0, slot);
}

inline bool JSObject::putOwnDataPropertyMayBeIndex(ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    ASSERT(value);
    ASSERT(!Heap::heap(value) || Heap::heap(value) == Heap::heap(this));
    ASSERT(!structure()->hasGetterSetterProperties());
    ASSERT(!structure()->hasCustomGetterSetterProperties());

    if (Optional<uint32_t> index = parseIndex(propertyName))
        return putDirectIndex(exec, index.value(), value, 0, PutDirectIndexLikePutDirect);

    return putDirectInternal<PutModePut>(exec->vm(), propertyName, value, 0, slot);
}

inline bool JSObject::putDirect(JSGlobalData& vm, PropertyName propertyName, JSValue value, unsigned attributes)
{
    ASSERT(!value.isGetterSetter() && !(attributes & Accessor));
    ASSERT(!value.isCustomGetterSetter() && !(attributes & PropertyAttribute::CustomAccessorOrValue));
    PutPropertySlot slot(this);
    return putDirectInternal<PutModeDefineOwnProperty>(vm, propertyName, value, attributes, slot);
}

inline bool JSObject::putDirect(JSGlobalData& vm, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    ASSERT(!value.isGetterSetter());
    ASSERT(!value.isCustomGetterSetter());
    return putDirectInternal<PutModeDefineOwnProperty>(vm, propertyName, value, 0, slot);
}

inline bool JSObject::putDirectWithoutTransition(JSGlobalData& vm, PropertyName propertyName, JSValue value, unsigned attributes)
{
    DeferGC deferGC(vm.heap);
    ASSERT(!value.isGetterSetter() && !(attributes & Accessor));
    ASSERT(!value.isCustomGetterSetter());
    Butterfly* newButterfly = m_butterfly;
    if(UNLIKELY(structure()->putWillGrowOutOfLineStorage()))
        newButterfly = growOutOfLineStorage(vm, structure()->outOfLineCapacity(), structure()->suggestedNewOutOfLineStorageCapacity());
    Structure* structure = this->structure();
    PropertyOffset offset = structure->addPropertyWithoutTransition(vm, propertyName, attributes);
    if (attributes & ReadOnly)
		structure->setContainsReadOnlyProperties();
    setStructureAndButterfly(vm, structure, newButterfly);
    putDirect(vm, offset, value);
    return true;
}

ALWAYS_INLINE JSObject* Register::object() const
{
	return asObject(jsValue());
}

ALWAYS_INLINE Register& Register::operator=(JSObject* object)
{
    u.value = JSValue::encode(JSValue(object));
    return *this;
}

ALWAYS_INLINE size_t offsetInButterfly(PropertyOffset offset)
{
    return offsetInOutOfLineStorage(offset) + Butterfly::indexOfPropertyStorage();
}

ALWAYS_INLINE size_t JSObject::butterflyPreCapacity()
{
    if (UNLIKELY(hasIndexingHeader()))
        return butterfly()->indexingHeader()->preCapacity(structure());
    return 0;
}

ALWAYS_INLINE size_t JSObject::butterflyTotalSize()
{
    Structure* structure = this->structure();
    Butterfly* butterfly = this->butterfly();
    size_t preCapacity;
    size_t indexingPayloadSizeInBytes;
    bool hasIndexingHeader = this->hasIndexingHeader();

    if (UNLIKELY(hasIndexingHeader)) {
        preCapacity = butterfly->indexingHeader()->preCapacity(structure);
        indexingPayloadSizeInBytes = butterfly->indexingHeader()->indexingPayloadSizeInBytes(structure);
    } else {
        preCapacity = 0;
        indexingPayloadSizeInBytes = 0;
    }

    return Butterfly::totalSize(preCapacity, structure->outOfLineCapacity(), hasIndexingHeader, indexingPayloadSizeInBytes);
}

// Helpers for patching code where you want to emit a load or store and
// the base is:
// For inline offsets: a pointer to the out-of-line storage pointer.
// For out-of-line offsets: the base of the out-of-line storage.
inline size_t offsetRelativeToPatchedStorage(PropertyOffset offset)
{
    if (isOutOfLineOffset(offset))
        return sizeof(EncodedJSValue) * offsetInButterfly(offset);
    return JSObject::offsetOfInlineStorage() - JSObject::butterflyOffset() + sizeof(EncodedJSValue) * offsetInInlineStorage(offset);
}

// Returns the maximum offset (away from zero) a load instruction will encode.
inline size_t maxOffsetRelativeToPatchedStorage(PropertyOffset offset)
{
    ptrdiff_t addressOffset = static_cast<ptrdiff_t>(offsetRelativeToPatchedStorage(offset));
#if USE(JSVALUE32_64)
    if (addressOffset >= 0)
        return static_cast<size_t>(addressOffset) + OBJECT_OFFSETOF(EncodedValueDescriptor, asBits.tag);
#endif
    return static_cast<size_t>(addressOffset);
}

inline int indexRelativeToBase(PropertyOffset offset)
{
    if (isOutOfLineOffset(offset))
        return offsetInOutOfLineStorage(offset) + Butterfly::indexOfPropertyStorage();
    ASSERT(!(JSObject::offsetOfInlineStorage() % sizeof(EncodedJSValue)));
    return JSObject::offsetOfInlineStorage() / sizeof(EncodedJSValue) + offsetInInlineStorage(offset);
}

inline int offsetRelativeToBase(PropertyOffset offset)
{
    if (isOutOfLineOffset(offset))
        return offsetInOutOfLineStorage(offset) * sizeof(EncodedJSValue) + Butterfly::offsetOfPropertyStorage();
    return JSObject::offsetOfInlineStorage() + offsetInInlineStorage(offset) * sizeof(EncodedJSValue);
}

COMPILE_ASSERT(!(sizeof(JSObject) % sizeof(WriteBarrierBase<Unknown>)), JSObject_inline_storage_has_correct_alignment);

template<unsigned charactersCount>
ALWAYS_INLINE Identifier makeIdentifier(VM& vm, const char (&characters)[charactersCount])
{
    return Identifier::fromString(vm, characters);
}

ALWAYS_INLINE Identifier makeIdentifier(JSGlobalData& vm, const Identifier& name)
{
    return name;
}

// Section 7.3.17 of the spec.
template <typename AddFunction> // Add function should have a type like: (JSValue, RuntimeType) -> bool
void createListFromArrayLike(ExecState* exec, JSValue arrayLikeValue, RuntimeTypeMask legalTypesFilter, const String& errorMessage, AddFunction addFunction)
{
    VM& vm = exec->vm();
    Vector<JSValue> result;
    JSValue lengthProperty = arrayLikeValue.get(exec, exec->vm().propertyNames->length);
    if (vm.exception())
        return;
    double lengthAsDouble = lengthProperty.toLength(exec);
    if (vm.exception())
        return;
    RELEASE_ASSERT(lengthAsDouble >= 0.0 && lengthAsDouble == std::trunc(lengthAsDouble));
    uint64_t length = static_cast<uint64_t>(lengthAsDouble);
    for (uint64_t index = 0; index < length; index++) {
        JSValue  next = arrayLikeValue.get(exec, index);
        if (vm.exception())
            return;

        RuntimeType type = runtimeTypeForValue(vm, next);
        if (!(type & legalTypesFilter)) {
            throwTypeError(exec, errorMessage);
            return;
        }

        bool exitEarly = addFunction(next, type);
        if (exitEarly)
            return;
    }
}

bool validateAndApplyPropertyDescriptor(ExecState*, JSObject*, PropertyName, bool isExtensible,
    const PropertyDescriptor& descriptor, bool isCurrentDefined, const PropertyDescriptor& current, bool throwException);

JS_EXPORT_PRIVATE NEVER_INLINE bool ordinarySetSlow(ExecState*, JSObject*, PropertyName, JSValue, JSValue receiver, bool shouldThrow);

// Helper for defining native functions, if you're not using a static hash table.
// Use this macro from within finishCreation() methods in prototypes. This assumes
// you've defined variables called exec, globalObject, and vm, and they
// have the expected meanings.
#define JSC_NATIVE_INTRINSIC_FUNCTION(jsName, cppName, attributes, length, intrinsic) \
    putDirectNativeFunction(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (length), cppName, \
        (intrinsic), (attributes))

#define JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(jsName, cppName, attributes, length, intrinsic) \
    putDirectNativeFunctionWithoutTransition(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (length), cppName, \
        (intrinsic), (attributes))

// As above, but this assumes that the function you're defining doesn't have an
// intrinsic.
#define JSC_NATIVE_FUNCTION(jsName, cppName, attributes, length) \
    JSC_NATIVE_INTRINSIC_FUNCTION(jsName, cppName, (attributes), (length), NoIntrinsic)

#define JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(jsName, cppName, attributes, length) \
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(jsName, cppName, (attributes), (length), NoIntrinsic)

// Identical helpers but for builtins. Note that currently, we don't support builtins that are
// also intrinsics, but we probably will do that eventually.
#define JSC_BUILTIN_FUNCTION(jsName, generatorName, attributes) \
    putDirectBuiltinFunction(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (generatorName)(globalObject), (attributes))

#define JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(jsName, generatorName, attributes) \
    putDirectBuiltinFunctionWithoutTransition(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (generatorName)(globalObject), (attributes))

// Helper for defining native getters on properties.
#define JSC_NATIVE_INTRINSIC_GETTER(jsName, cppName, attributes, intrinsic)  \
    putDirectNativeIntrinsicGetter(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (cppName), \
        (intrinsic), ((attributes) | Accessor))

#define JSC_NATIVE_GETTER(jsName, cppName, attributes) \
    JSC_NATIVE_INTRINSIC_GETTER((jsName), (cppName), (attributes), NoIntrinsic)

// HasOwnProperty(O, P) from section 7.3.11 in the spec.
// http://www.ecma-international.org/ecma-262/6.0/index.html#sec-hasownproperty
ALWAYS_INLINE bool JSObject::hasOwnProperty(ExecState* exec, PropertyName propertyName, PropertySlot& slot) const
{
	ASSERT(slot.internalMethodType() == PropertySlot::InternalMethodType::GetOwnProperty);
    const MethodTable* methodTable = const_cast<JSObject*>(this)->methodTable();
    if (LIKELY(methodTable->getOwnPropertySlot == JSObject::getOwnPropertySlot))
        return JSObject::getOwnPropertySlot(const_cast<JSObject*>(this), exec, propertyName, slot);
    return methodTable->getOwnPropertySlot(const_cast<JSObject*>(this), exec, propertyName, slot);
}

ALWAYS_INLINE bool JSObject::hasOwnProperty(ExecState* exec, PropertyName propertyName) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::GetOwnProperty);
    return hasOwnProperty(exec, propertyName, slot);
}

ALWAYS_INLINE bool JSObject::hasOwnProperty(ExecState* exec, unsigned propertyName) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::GetOwnProperty);
    return const_cast<JSObject*>(this)->methodTable()->getOwnPropertySlotByIndex(const_cast<JSObject*>(this), exec, propertyName, slot);
}

ALWAYS_INLINE JSValue PropertySlot::getValue(ExecState* exec, unsigned propertyName) const
{
    if (m_propertyType == TypeValue)
        return JSValue::decode(m_data.value);
    if (m_propertyType == TypeGetter)
        return functionGetter(exec);
	VM& vm = exec->vm();
    return customGetter(exec, Identifier::from(vm, propertyName));
}

} // namespace JSC

#endif // JSObject_h
