/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef Lookup_h
#define Lookup_h

#include "BatchedTransitionOptimizer.h"
#include "CallFrame.h"
#include "CustomGetterSetter.h"
#include "IdentifierInlines.h"
#include "Intrinsic.h"
#include "JSGlobalObject.h"
#include "PropertySlot.h"
#include "PutPropertySlot.h"
#include "Reject.h"
#include <stdio.h>
#include <wtf/Assertions.h>

namespace JSC {

struct CompactHashIndex {
    const int16_t value;
    const int16_t next;
};

// FIXME: There is no reason this get function can't be simpler.
// ie. typedef JSValue (*GetFunction)(ExecState*, JSObject* baseObject)
typedef PropertySlot::GetValueFunc GetFunction;
typedef PutPropertySlot::PutValueFunc PutFunction;
typedef FunctionExecutable* (*BuiltinGenerator)(JSGlobalObject*);

// Hash table generated by the create_hash_table script.
struct HashTableValue {
    const char* m_key; // property name
    unsigned m_attributes; // JSObject attributes
    Intrinsic m_intrinsic;
    intptr_t m_value1;
    intptr_t m_value2;

	unsigned attributes() const { return m_attributes; }

	Intrinsic intrinsic() const { ASSERT(m_attributes & Function); return m_intrinsic; }

	BuiltinGenerator builtinGenerator() const { ASSERT(m_attributes & Builtin); return reinterpret_cast<BuiltinGenerator>(m_value1); }
	NativeFunction function() const { ASSERT(m_attributes & Function); return reinterpret_cast<NativeFunction>(m_value1); }
	unsigned char functionLength() const { ASSERT(m_attributes & Function); return static_cast<unsigned char>(m_value2); }

    GetFunction propertyGetter() const { ASSERT(!(m_attributes & BuiltinOrFunctionOrAccessorOrConstant)); return reinterpret_cast<GetFunction>(m_value1); }
    PutFunction propertyPutter() const { ASSERT(!(m_attributes & BuiltinOrFunctionOrAccessorOrConstant)); return reinterpret_cast<PutFunction>(m_value2); }

    NativeFunction accessorGetter() const { ASSERT(m_attributes & Accessor); return reinterpret_cast<NativeFunction>(m_value1); }
    NativeFunction accessorSetter() const { ASSERT(m_attributes & Accessor); return reinterpret_cast<NativeFunction>(m_value2); }
    BuiltinGenerator builtinAccessorGetterGenerator() const;
	BuiltinGenerator builtinAccessorSetterGenerator() const;

    intptr_t constantInteger() const { ASSERT(m_attributes & ConstantInteger); return m_value1; }

	intptr_t lexerValue() const { ASSERT(!m_attributes); return m_value1; }
};

struct HashTable {
    mutable int numberOfValues;
    int indexMask;
    bool hasSetterOrReadonlyProperties;

	const HashTableValue* values; // Fixed values generated by script.
    mutable const char** keys; // Table allocated at runtime.
    const CompactHashIndex* index;

	ALWAYS_INLINE HashTable copy() const
	{
		// Don't copy dynamic table since it's thread specific.
		HashTable result = { numberOfValues, indexMask, hasSetterOrReadonlyProperties, values, 0, index };
		return result;
	}

	ALWAYS_INLINE void initializeIfNeeded() const
	{
		if (UNLIKELY(!keys))
			createTable();
	}

	JS_EXPORT_PRIVATE void deleteTable() const;

	// Find an entry in the table, and return the entry.
	ALWAYS_INLINE const HashTableValue* entry(PropertyName propertyName) const
	{
		initializeIfNeeded();

		auto* impl = propertyName.uid();
		if (!impl)
			return 0;

		ASSERT(keys);

		int indexEntry = IdentifierRepHash::hash(impl) & indexMask;
		int valueIndex = index[indexEntry].value;
		if (valueIndex == -1)
			return 0;

		while (true) {
			if (WTF::equal(impl, keys[valueIndex]))
				return &values[valueIndex];

			indexEntry = index[indexEntry].next;
			if (indexEntry == -1)
				return 0;
			valueIndex = index[indexEntry].value;
			ASSERT(valueIndex != -1);
		};
	}

	class ConstIterator {
	public:
		ALWAYS_INLINE ConstIterator(const HashTable* table, int position)
			: m_table(table)
			, m_position(position)
		{
			skipInvalidKeys();
		}

		ALWAYS_INLINE const HashTableValue* value() const
        {
            return &m_table->values[m_position];
        }

		ALWAYS_INLINE const HashTableValue& operator*() const { return *value(); }

		ALWAYS_INLINE const char* key() const
        {
            return m_table->keys[m_position];
        }

		ALWAYS_INLINE const HashTableValue* operator->() const
        {
            return value();
        }

		ALWAYS_INLINE bool operator!=(const ConstIterator& other) const
		{
			ASSERT(m_table == other.m_table);
			return m_position != other.m_position;
		}

		ALWAYS_INLINE ConstIterator& operator++()
		{
			ASSERT(m_position < m_table->numberOfValues);
			++m_position;
			skipInvalidKeys();
			return *this;
		}

	private:
		ALWAYS_INLINE void skipInvalidKeys()
		{
            ASSERT(m_position <= m_table->numberOfValues);
            while (m_position < m_table->numberOfValues && !m_table->keys[m_position])
                ++m_position;
            ASSERT(m_position <= m_table->numberOfValues);
		}

		const HashTable* m_table;
		int m_position;
	};

	ALWAYS_INLINE ConstIterator begin() const
	{
		initializeIfNeeded();
		return ConstIterator(this, 0);
	}
	ALWAYS_INLINE ConstIterator end() const
	{
		initializeIfNeeded();
		return ConstIterator(this, numberOfValues);
	}

private:
	// Convert the hash table keys to identifiers.
	JS_EXPORT_PRIVATE void createTable() const;
};

JS_EXPORT_PRIVATE bool setUpStaticFunctionSlot(VM&, const HashTableValue*, JSObject* thisObject, PropertyName, PropertySlot&);
JS_EXPORT_PRIVATE void reifyStaticAccessor(VM&, const HashTableValue&, JSObject& thisObject, PropertyName);

inline BuiltinGenerator HashTableValue::builtinAccessorGetterGenerator() const
{
    ASSERT(m_attributes & Accessor);
    ASSERT(m_attributes & Builtin);
    return reinterpret_cast<BuiltinGenerator>(m_value1);
}

inline BuiltinGenerator HashTableValue::builtinAccessorSetterGenerator() const
{
    ASSERT(m_attributes & Accessor);
    ASSERT(m_attributes & Builtin);
    return reinterpret_cast<BuiltinGenerator>(m_value2);
}

inline bool getStaticPropertySlotFromTable(VM& vm, const HashTable& table, JSObject* thisObject, PropertyName propertyName, PropertySlot& slot)
{
    if (thisObject->staticPropertiesReified())
        return false;

    auto* entry = table.entry(propertyName);
    if (!entry)
        return false;

    if (entry->attributes() & BuiltinOrFunctionOrAccessor)
        return setUpStaticFunctionSlot(vm, entry, thisObject, propertyName, slot);

    if (entry->attributes() & ConstantInteger) {
        slot.setValue(thisObject, attributesForStructure(entry->attributes()), jsNumber(entry->constantInteger()));
        return true;
    }

    slot.setCacheableCustom(thisObject, attributesForStructure(entry->attributes()), entry->propertyGetter());
    return true;
}

inline bool replaceStaticPropertySlot(VM& vm, JSObject* thisObject, PropertyName propertyName, JSValue value)
{
    if (!thisObject->putDirect(vm, propertyName, value))
        return false;

    if (!thisObject->staticPropertiesReified())
        thisObject->JSObject::setStructure(vm, Structure::attributeChangeTransition(vm, thisObject->structure(), propertyName, 0));

    return true;
}

// 'base' means the object holding the property (possibly in the prototype chain of the object put was called on).
// 'thisValue' is the object that put is being applied to (in the case of a proxy, the proxy target).
// 'slot.thisValue()' is the object the put was originally performed on (in the case of a proxy, the proxy itself).
inline bool putEntry(ExecState* exec, VM& vm, const HashTableValue* entry, JSObject* base, JSObject* thisValue, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
	if (entry->attributes() & BuiltinOrFunction) {
        if (!(entry->attributes() & ReadOnly)) {
            // If this is a function put it as an override property.
			if (JSObject* thisObject = jsDynamicCast<JSObject*>(thisValue))
				thisObject->putDirect(exec->vm(), propertyName, value);
			return true;
        }
        return reject(exec, slot.isStrictMode(), ReadonlyPropertyWriteError);
    }

	if (entry->attributes() & Accessor)
        return reject(exec, slot.isStrictMode(), ReadonlyPropertyWriteError);

	if (!(entry->attributes() & ReadOnly)) {
        bool isAccessor = entry->attributes() & CustomAccessor;
        JSValue updateThisValue = entry->attributes() & CustomAccessor ? slot.thisValue() : JSValue(base);
        // We need to make sure that we decide to cache this property before we potentially execute aribitrary JS.
        if (isAccessor)
            slot.setCustomAccessor(base, entry->propertyPutter());
        else
            slot.setCustomValue(base, entry->propertyPutter());


        bool result = callCustomSetter(exec, entry->propertyPutter(), isAccessor, updateThisValue, value);
        RETURN_IF_EXCEPTION(vm, false);
        return result;
	}

    return reject(exec, slot.isStrictMode(), ReadonlyPropertyWriteError);
}

/**
 * This one is for "put".
 * It looks up a hash entry for the property to be set.  If an entry
 * is found it sets the value and returns true, else it returns false.
 */
inline bool lookupPut(ExecState* exec, VM& vm, PropertyName propertyName, JSObject* base, JSValue value, const HashTable& table, PutPropertySlot& slot, bool& putResult)
{
	const HashTableValue* entry = table.entry(propertyName);

	if (!entry)
		return false;

	putResult = putEntry(exec, vm, entry, base, base, propertyName, value, slot);
	return true;
}

inline void reifyStaticProperty(JSGlobalData& vm, PropertyName propertyName, const HashTableValue& value, JSObject& thisObject)
{
	if (value.attributes() & Builtin) {
        if (value.attributes() & Accessor)
            reifyStaticAccessor(vm, value, thisObject, propertyName);
        else
        	thisObject.putDirectBuiltinFunction(vm, thisObject.globalObject(), propertyName, value.builtinGenerator()(thisObject.globalObject()), attributesForStructure(value.attributes()));
		return;
	}

	if (value.attributes() & Function) {
		thisObject.putDirectNativeFunction(
			vm, thisObject.globalObject(), propertyName, value.functionLength(),
			value.function(), value.intrinsic(), attributesForStructure(value.attributes()));
		return;
	}

	if (value.attributes() & ConstantInteger) {
		thisObject.putDirect(vm, propertyName, jsNumber(value.constantInteger()), attributesForStructure(value.attributes()));
		return;
	}

	if (value.attributes() & Accessor) {
        reifyStaticAccessor(vm, value, thisObject, propertyName);
		return;
	}

	CustomGetterSetter* customGetterSetter = CustomGetterSetter::create(vm, value.propertyGetter(), value.propertyPutter());
	thisObject.putDirectCustomAccessor(vm, propertyName, customGetterSetter, attributesForStructure(value.attributes()));
}

template<unsigned numberOfValues>
inline void reifyStaticProperties(JSGlobalData& vm, const HashTableValue (&values)[numberOfValues], JSObject& thisObject)
{
    BatchedTransitionOptimizer transitionOptimizer(vm, &thisObject);
    for (int i = 0; i < numberOfValues; ++i) {
    	const HashTableValue& value = values[i];
        if (!value.m_key)
            continue;
        Identifier key = Identifier::fromString(vm, reinterpret_cast<const LChar*>(value.m_key), strlen(value.m_key));
        reifyStaticProperty(vm, key, value, thisObject);
    }
}

} // namespace JSC

#endif // Lookup_h
