/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "ObjectConstructor.h"

#include "ButterflyInlineMethods.h"
#include "CallFrameInlines.h"
#include "CopiedSpaceInlineMethods.h"
#include "Error.h"
#include "ExceptionHelpers.h"
#include "JSFunction.h"
#include "JSArray.h"
#include "JSGlobalObject.h"
#include "JSGlobalObjectFunctions.h"
#include "Lookup.h"
#include "ObjectPrototype.h"
#include "Operations.h"
#include "PropertyDescriptor.h"
#include "PropertyNameArray.h"
#include "StackVisitor.h"
#include "Symbol.h"
#include "ThrowScope.h"
#include <wtf/MemoryMonitor.h>

namespace JSC {

EncodedJSValue JSC_HOST_CALL objectConstructorAssign(ExecState*);
EncodedJSValue JSC_HOST_CALL objectConstructorValues(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorGetPrototypeOf(ExecState*);
EncodedJSValue JSC_HOST_CALL objectConstructorSetPrototypeOf(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperty(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperties(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorCreate(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorSeal(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorFreeze(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorPreventExtensions(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorIsSealed(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorIsFrozen(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorIsExtensible(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorIs(ExecState*);

}

#include "ObjectConstructor.lut.h"

namespace JSC {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(ObjectConstructor);

const ClassInfo ObjectConstructor::s_info = { "Function", &InternalFunction::s_info, &objectConstructorTable, CREATE_METHOD_TABLE(ObjectConstructor) };

/* Source for ObjectConstructor.lut.h
@begin objectConstructorTable
  getPrototypeOf            objectConstructorGetPrototypeOf             DontEnum|Function 1
  setPrototypeOf            objectConstructorSetPrototypeOf             DontEnum|Function 2
  getOwnPropertyDescriptor  objectConstructorGetOwnPropertyDescriptor   DontEnum|Function 2
  getOwnPropertyDescriptors objectConstructorGetOwnPropertyDescriptors  DontEnum|Function 1
  getOwnPropertyNames       objectConstructorGetOwnPropertyNames        DontEnum|Function 1
  getOwnPropertySymbols     objectConstructorGetOwnPropertySymbols      DontEnum|Function 1
  keys                      objectConstructorKeys                       DontEnum|Function 1
  defineProperty            objectConstructorDefineProperty             DontEnum|Function 3
  defineProperties          objectConstructorDefineProperties           DontEnum|Function 2
  create                    objectConstructorCreate                     DontEnum|Function 2
  seal                      objectConstructorSeal                       DontEnum|Function 1
  freeze                    objectConstructorFreeze                     DontEnum|Function 1
  preventExtensions         objectConstructorPreventExtensions          DontEnum|Function 1
  isSealed                  objectConstructorIsSealed                   DontEnum|Function 1
  isFrozen                  objectConstructorIsFrozen                   DontEnum|Function 1
  isExtensible              objectConstructorIsExtensible               DontEnum|Function 1
  is                        objectConstructorIs                         DontEnum|Function 2
  assign                    objectConstructorAssign                     DontEnum|Function 2
  values                    objectConstructorValues                     DontEnum|Function 1
  entries                   JSBuiltin                                   DontEnum|Function 1
@end
*/

ObjectConstructor::ObjectConstructor(VM& vm, Structure* structure)
    : InternalFunction(vm, structure)
{
}

void ObjectConstructor::finishCreation(VM& vm, JSGlobalObject* globalObject, ObjectPrototype* objectPrototype)
{
    Base::finishCreation(vm, objectPrototype->classInfo()->className);
    // ECMA 15.2.3.1
    putDirectWithoutTransition(vm, vm.propertyNames->prototype, objectPrototype, DontEnum | DontDelete | ReadOnly);
    // no. of arguments for constructor
    putDirectWithoutTransition(vm, vm.propertyNames->length, jsNumber(1), ReadOnly | DontEnum | DontDelete);

    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->createPrivateName, objectConstructorCreate, DontEnum, 2);
	JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->definePropertyPrivateName, objectConstructorDefineProperty, DontEnum, 3);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->getPrototypeOfPrivateName, objectConstructorGetPrototypeOf, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->getOwnPropertyNamesPrivateName, objectConstructorGetOwnPropertyNames, DontEnum, 1);
}

JSFunction* ObjectConstructor::addDefineProperty(ExecState* exec, JSGlobalObject* globalObject)
{
    VM& vm = exec->vm();
    JSFunction* definePropertyFunction = JSFunction::create(vm, globalObject, 3, vm.propertyNames->defineProperty.string(), objectConstructorDefineProperty);
    putDirectWithoutTransition(vm, vm.propertyNames->defineProperty, definePropertyFunction, DontEnum);
    return definePropertyFunction;
}

// ES 19.1.1.1 Object([value])
static ALWAYS_INLINE JSObject* constructObject(ExecState* exec, JSValue newTarget)
{
    ObjectConstructor* objectConstructor = jsCast<ObjectConstructor*>(exec->callee());
    JSGlobalObject* globalObject = objectConstructor->globalObject();

    VM& vm = globalObject->vm();

    // We need to check newTarget condition in this caller side instead of InternalFunction::createSubclassStructure side.
    // Since if we found this condition is met, we should not fall into the type conversion in the step 3.

    // 1. If NewTarget is neither undefined nor the active function, then
    if (newTarget && newTarget != objectConstructor) {
        // a. Return ? OrdinaryCreateFromConstructor(NewTarget, "%ObjectPrototype%").
        Structure* objectStructure = InternalFunction::createSubclassStructure(exec, newTarget, globalObject->objectStructureForObjectConstructor());
        RETURN_IF_EXCEPTION(vm, 0);
        return constructEmptyObject(exec, objectStructure);
    }

    // 2. If value is null, undefined or not supplied, return ObjectCreate(%ObjectPrototype%).
    ArgList args(exec);
    JSValue arg = args.at(0);
    if (arg.isUndefinedOrNull())
        return constructEmptyObject(exec, globalObject->objectStructureForObjectConstructor());

    // 3. Return ToObject(value).
    return arg.toObject(exec, globalObject);
}

static EncodedJSValue JSC_HOST_CALL constructWithObjectConstructor(ExecState* exec)
{
    return JSValue::encode(constructObject(exec, exec->newTarget()));
}

ConstructType ObjectConstructor::getConstructData(JSCell*, ConstructData& constructData)
{
    constructData.native.function = constructWithObjectConstructor;
    return ConstructType::Host;
}

static EncodedJSValue JSC_HOST_CALL callObjectConstructor(ExecState* exec)
{
    return JSValue::encode(constructObject(exec, JSValue()));
}

CallType ObjectConstructor::getCallData(JSCell*, CallData& callData)
{
    callData.native.function = callObjectConstructor;
    return CallType::Host;
}

EncodedJSValue JSC_HOST_CALL objectConstructorGetPrototypeOf(ExecState* exec)
{
    VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(object->getPrototype(vm, exec));
}

EncodedJSValue JSC_HOST_CALL objectConstructorSetPrototypeOf(ExecState* exec)
{
    VM& vm = exec->vm();

    JSValue objectValue = exec->argument(0);
    if (objectValue.isUndefinedOrNull())
        return throwVMTypeError(exec, ASCIILiteral("Cannot set prototype of undefined or null"));

    JSValue protoValue = exec->argument(1);
    if (!protoValue.isObject() && !protoValue.isNull())
        return throwVMTypeError(exec, ASCIILiteral("Prototype value can only be an object or null"));

    JSObject* object = objectValue.toObject(exec);
    if (vm.exception())
        return JSValue::encode(objectValue);

    bool shouldThrowIfCantSet = true;
    bool didSetPrototype = object->setPrototype(vm, exec, protoValue, shouldThrowIfCantSet);
    ASSERT_UNUSED(didSetPrototype, vm.exception() || didSetPrototype);
    return JSValue::encode(objectValue);
}

JSValue objectConstructorGetOwnPropertyDescriptor(ExecState* exec, JSObject* object, const Identifier& propertyName)
{
	VM& vm = exec->vm();
    PropertyDescriptor descriptor;
    if (!object->getOwnPropertyDescriptor(exec, propertyName, descriptor))
        return jsUndefined();
    RETURN_IF_EXCEPTION(vm, JSValue());

    JSObject* result = constructObjectFromPropertyDescriptor(exec, descriptor);
    if (!result)
        return jsUndefined();
    return result;
}

JSValue objectConstructorGetOwnPropertyDescriptors(ExecState* exec, JSObject* object)
{
	VM& vm = exec->vm();
    PropertyNameArray properties(vm, PropertyNameMode::Strings);
    object->methodTable()->getOwnPropertyNames(object, exec, properties, EnumerationMode(DontEnumPropertiesMode::Include));
    RETURN_IF_EXCEPTION(vm, JSValue());

    JSObject* descriptors = constructEmptyObject(exec);
    RETURN_IF_EXCEPTION(vm, JSValue());

    size_t numProperties = properties.size();
    for (size_t i = 0; i < numProperties; i++) {
    	PropertyDescriptor descriptor;
		bool didGetDescriptor = object->getOwnPropertyDescriptor(exec, properties[i], descriptor);
	    RETURN_IF_EXCEPTION(vm, JSValue());

        if (!didGetDescriptor)
            continue;

        JSObject* fromDescriptor = constructObjectFromPropertyDescriptor(exec, descriptor);
        if (!fromDescriptor)
            return jsUndefined();

        PutPropertySlot slot(descriptors);
        descriptors->putOwnDataPropertyMayBeIndex(exec, properties[i], fromDescriptor, slot);
        ASSERT(!vm.exception());
    }

    return descriptors;
}

EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertyDescriptor(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    Identifier propertyName = exec->argument(1).toPropertyKey(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(objectConstructorGetOwnPropertyDescriptor(exec, object, propertyName));
}

EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertyDescriptors(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(objectConstructorGetOwnPropertyDescriptors(exec, object));
}

// FIXME: Use the enumeration cache.
EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertyNames(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    JSArray* keys = ownPropertyKeys(exec, object, PropertyNameMode::Strings, EnumerationMode(DontEnumPropertiesMode::Include));
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(keys);
}

// FIXME: Use the enumeration cache.
EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertySymbols(ExecState* exec)
{
    VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    JSArray* keys = ownPropertyKeys(exec, object, PropertyNameMode::Symbols, EnumerationMode(DontEnumPropertiesMode::Include));
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(keys);
}

// FIXME: Use the enumeration cache.
EncodedJSValue JSC_HOST_CALL objectConstructorKeys(ExecState* exec)
{
    VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    JSArray* keys = ownPropertyKeys(exec, object, PropertyNameMode::Strings, EnumerationMode());
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(keys);
}

EncodedJSValue JSC_HOST_CALL objectConstructorAssign(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue targetValue = exec->argument(0);
    if (targetValue.isUndefinedOrNull())
        return throwVMTypeError(vm, exec, ASCIILiteral("Object.assign requires that input parameter not be null or undefined"));
    JSObject* target = targetValue.toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());

    unsigned argsCount = exec->argumentCount();
    for (unsigned i = 1; i < argsCount; ++i) {
        JSValue sourceValue = exec->uncheckedArgument(i);
        if (sourceValue.isUndefinedOrNull())
            continue;
        JSObject* source = sourceValue.toObject(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSValue());

        PropertyNameArray properties(vm, PropertyNameMode::StringsAndSymbols);
        source->methodTable(vm)->getOwnPropertyNames(source, exec, properties, EnumerationMode(DontEnumPropertiesMode::Include));
        RETURN_IF_EXCEPTION(vm, encodedJSValue());

        class Assign {
        public:
        	Assign(ExecState* exec, VM& vm, JSObject* target, JSObject* source) : exec(exec), vm(vm), target(target), source(source) { }
        	void operator()(PropertyName propertyName) {
				// FIXME: We can avoid this enumerable look up by checking Structure's status.
				// https://bugs.webkit.org/show_bug.cgi?id=173416
				PropertySlot slot(source, PropertySlot::InternalMethodType::GetOwnProperty);
				if (!source->methodTable(vm)->getOwnPropertySlot(source, exec, propertyName, slot))
					return;
				if (slot.attributes() & DontEnum)
					return;

				JSValue value = source->get(exec, propertyName);
				if(UNLIKELY(vm.exception()))
					return;

				PutPropertySlot putPropertySlot(target, true);
				target->putInline(exec, propertyName, value, putPropertySlot);
			}
        private:
        	ExecState* exec;
        	VM& vm;
        	JSObject* target;
        	JSObject* source;
        };
        Assign assign(exec, vm, target, source);

        // First loop is for strings. Second loop is for symbols to keep standardized order requirement in the spec.
        // https://tc39.github.io/ecma262/#sec-ordinaryownpropertykeys
        bool foundSymbol = false;
        unsigned numProperties = properties.size();
        for (unsigned j = 0; j < numProperties; j++) {
            const auto& propertyName = properties[j];
            if (propertyName.isSymbol()) {
                foundSymbol = true;
                continue;
            }

            assign(propertyName);
            RETURN_IF_EXCEPTION(vm, encodedJSValue());
        }

        if (foundSymbol) {
            for (unsigned j = 0; j < numProperties; j++) {
                const auto& propertyName = properties[j];
                if (propertyName.isSymbol() && !vm.propertyNames->isPrivateName(propertyName)) {
                    assign(propertyName);
                    RETURN_IF_EXCEPTION(vm, encodedJSValue());
                }
            }
        }
    }
    return JSValue::encode(target);
}

EncodedJSValue JSC_HOST_CALL objectConstructorValues(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue targetValue = exec->argument(0);
    if (targetValue.isUndefinedOrNull())
        return throwVMTypeError(vm, exec, ASCIILiteral("Object.values requires that input parameter not be null or undefined"));
    JSObject* target = targetValue.toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());

    JSArray* values = constructEmptyArray(exec, 0);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());

    PropertyNameArray properties(vm, PropertyNameMode::Strings);
    target->methodTable(vm)->getOwnPropertyNames(target, exec, properties, EnumerationMode(DontEnumPropertiesMode::Include));
    RETURN_IF_EXCEPTION(vm, encodedJSValue());

    unsigned index = 0;
    class AddValue {
    public:
    	AddValue(ExecState* exec, VM& vm, JSObject* target, JSArray* values, unsigned& index) : exec(exec), vm(vm), target(target), values(values), index(index) { }
    	void operator()(PropertyName propertyName) {
			PropertySlot slot(target, PropertySlot::InternalMethodType::GetOwnProperty);
			if (!target->methodTable(vm)->getOwnPropertySlot(target, exec, propertyName, slot))
				return;
			if (slot.attributes() & DontEnum)
				return;

			JSValue value;
			if (LIKELY(!slot.isTaintedByOpaqueObject()))
				value = slot.getValue(exec, propertyName);
			else
				value = target->get(exec, propertyName);
			if(UNLIKELY(vm.exception()))
				return;

	        values->putDirectIndex(exec, index++, value);
    	}
    private:
    	ExecState* exec;
    	VM& vm;
    	JSObject* target;
    	JSArray* values;
    	unsigned& index;
    };
    AddValue addValue(exec, vm, target, values, index);

    for (unsigned i = 0, numProperties = properties.size(); i < numProperties; i++) {
        const auto& propertyName = properties[i];
        if (propertyName.isSymbol())
            continue;

        addValue(propertyName);
        RETURN_IF_EXCEPTION(vm, encodedJSValue());
    }

    return JSValue::encode(values);
}

// ES6 6.2.4.5 ToPropertyDescriptor
// https://tc39.github.io/ecma262/#sec-topropertydescriptor
bool toPropertyDescriptor(ExecState* exec, JSValue in, PropertyDescriptor& desc)
{
	VM& vm = exec->vm();
    if (!in.isObject()) {
        vm.throwException(exec, createTypeError(exec, "Property description must be an object."));
        return false;
    }
    JSObject* description = asObject(in);

    bool hasProperty = description->hasProperty(exec, vm.propertyNames->enumerable);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(exec, vm.propertyNames->enumerable);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setEnumerable(value.toBoolean(exec));
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->configurable);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(exec, vm.propertyNames->configurable);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setConfigurable(value.toBoolean(exec));
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->value);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(exec, vm.propertyNames->value);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setValue(value);
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->writable);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(exec, vm.propertyNames->writable);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setWritable(value.toBoolean(exec));
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->get);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue get = description->get(exec, vm.propertyNames->get);
        RETURN_IF_EXCEPTION(vm, false);
        if (!get.isUndefined()) {
            CallData callData;
            if (getCallData(get, callData) == CallType::None) {
                vm.throwException(exec, createTypeError(exec, "Getter must be a function."));
                return false;
            }
        }
        desc.setGetter(get);
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->set);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue set = description->get(exec, vm.propertyNames->set);
        RETURN_IF_EXCEPTION(vm, false);
        if (!set.isUndefined()) {
            CallData callData;
            if (getCallData(set, callData) == CallType::None) {
                vm.throwException(exec, createTypeError(exec, "Setter must be a function."));
                return false;
            }
        }
        desc.setSetter(set);
    } else
        RETURN_IF_EXCEPTION(vm, false);

    if (!desc.isAccessorDescriptor())
        return true;

    if (desc.value()) {
        vm.throwException(exec, createTypeError(exec, "Invalid property.  'value' present on property with getter or setter."));
        return false;
    }

    if (desc.writablePresent()) {
        vm.throwException(exec, createTypeError(exec, "Invalid property.  'writable' present on property with getter or setter."));
        return false;
    }
    return true;
}

EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperty(ExecState* exec)
{
	VM& vm = exec->vm();
	if(UNLIKELY(WTF::gLowMemory))
		return JSValue::encode(vm.throwException(exec, createOutOfMemoryError(exec)));
    if (UNLIKELY(!exec->argument(0).isObject()))
        return JSValue::encode(vm.throwException(exec, createTypeError(exec, "Properties can only be defined on Objects.")));
    JSObject* O = asObject(exec->argument(0));
    Identifier propertyName = exec->argument(1).toPropertyKey(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    PropertyDescriptor descriptor;
    if (UNLIKELY(!toPropertyDescriptor(exec, exec->argument(2), descriptor)))
        return JSValue::encode(jsNull());
    ASSERT((descriptor.attributes() & Accessor) || (!descriptor.isAccessorDescriptor()));
    ASSERT(!vm.exception());
    O->methodTable()->defineOwnProperty(O, exec, propertyName, descriptor, true);
    return JSValue::encode(O);
}

static JSValue defineProperties(ExecState* exec, JSObject* object, JSObject* properties)
{
	VM& vm = exec->vm();
    PropertyNameArray propertyNames(vm, PropertyNameMode::StringsAndSymbols);
    asObject(properties)->methodTable()->getOwnPropertyNames(asObject(properties), exec, propertyNames, EnumerationMode(DontEnumPropertiesMode::Exclude));
    RETURN_IF_EXCEPTION(vm, JSValue());
    size_t numProperties = propertyNames.size();
    Vector<PropertyDescriptor> descriptors;
    MarkedArgumentBuffer markBuffer;
    for (size_t i = 0; i < numProperties; i++) {
        JSValue prop = properties->get(exec, propertyNames[i]);
        RETURN_IF_EXCEPTION(vm, JSValue());
        PropertyDescriptor descriptor;
        bool success = toPropertyDescriptor(exec, prop, descriptor);
        ASSERT(!vm.exception() || !success);
        if (UNLIKELY(!success))
            return jsNull();
        descriptors.append(descriptor);
        // Ensure we mark all the values that we're accumulating
        if (descriptor.isDataDescriptor() && descriptor.value())
            markBuffer.append(descriptor.value());
        if (descriptor.isAccessorDescriptor()) {
            if (descriptor.getter())
                markBuffer.append(descriptor.getter());
            if (descriptor.setter())
                markBuffer.append(descriptor.setter());
        }
    }
    for (size_t i = 0; i < numProperties; i++) {
    	Identifier propertyName = propertyNames[i];
		if (vm.propertyNames->isPrivateName(propertyName))
			continue;
		object->methodTable(vm)->defineOwnProperty(object, exec, propertyName, descriptors[i], true);
		RETURN_IF_EXCEPTION(vm, JSValue());
    }
    return object;
}

EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperties(ExecState* exec)
{
    if (!exec->argument(0).isObject())
        return throwVMError(exec, createTypeError(exec, "Properties can only be defined on Objects."));
    JSObject* targetObj = asObject(exec->argument(0));
	JSObject* props = exec->argument(1).toObject(exec);
    ASSERT(!!scope.exception() == !props);
    if (UNLIKELY(!props))
        return encodedJSValue();
	return JSValue::encode(defineProperties(exec, targetObj, props));
}

EncodedJSValue JSC_HOST_CALL objectConstructorCreate(ExecState* exec)
{
    JSValue proto = exec->argument(0);
    if (!proto.isObject() && !proto.isNull())
        return throwVMError(exec, createTypeError(exec, "Object prototype may only be an Object or null."));
    JSObject* newObject = proto.isObject()
		? constructEmptyObject(exec, asObject(proto))
		: constructEmptyObject(exec, exec->lexicalGlobalObject()->nullPrototypeObjectStructure());
   if (exec->argument(1).isUndefined())
        return JSValue::encode(newObject);
    if (!exec->argument(1).isObject())
        return throwVMError(exec, createTypeError(exec, "Property descriptor list must be an Object."));
    return JSValue::encode(defineProperties(exec, newObject, asObject(exec->argument(1))));
}

enum class IntegrityLevel {
    Sealed,
    Frozen
};

template<IntegrityLevel level>
bool setIntegrityLevel(ExecState* exec, VM& vm, JSObject* object)
{
    // See https://tc39.github.io/ecma262/#sec-setintegritylevel.
	ThrowScope scope(vm);

    bool success = object->methodTable(vm)->preventExtensions(object, exec);
    RETURN_IF_EXCEPTION(scope, false);
    if (UNLIKELY(!success))
        return false;

    PropertyNameArray properties(vm, PropertyNameMode::StringsAndSymbols);
    object->methodTable(vm)->getOwnPropertyNames(object, exec, properties, EnumerationMode(DontEnumPropertiesMode::Include));
    RETURN_IF_EXCEPTION(scope, false);

    PropertyNameArray::const_iterator end = properties.end();
    for (PropertyNameArray::const_iterator iter = properties.begin(); iter != end; ++iter) {
        Identifier propertyName = *iter;
        if (vm.propertyNames->isPrivateName(propertyName))
            continue;

        PropertyDescriptor desc;
        if (level == IntegrityLevel::Sealed)
            desc.setConfigurable(false);
        else {
            bool hasPropertyDescriptor = object->getOwnPropertyDescriptor(exec, propertyName, desc);
            RETURN_IF_EXCEPTION(scope, false);
            if (!hasPropertyDescriptor)
                continue;

            if (desc.isDataDescriptor())
                desc.setWritable(false);

            desc.setConfigurable(false);
        }

        object->methodTable(vm)->defineOwnProperty(object, exec, propertyName, desc, true);
        RETURN_IF_EXCEPTION(scope, false);
    }
    return true;
}

EncodedJSValue JSC_HOST_CALL objectConstructorSeal(ExecState* exec)
{
    VM& vm = exec->vm();
    ThrowScope scope(vm);

    // 1. If Type(O) is not Object, return O.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(obj);
    JSObject* object = asObject(obj);

    if (isJSFinalObject(object)) {
        object->seal(vm);
        return JSValue::encode(obj);
    }

    bool success = setIntegrityLevel<IntegrityLevel::Sealed>(exec, vm, object);
    RETURN_IF_EXCEPTION(scope, encodedJSValue());
    if (UNLIKELY(!success)) {
        throwTypeError(exec, ASCIILiteral("Unable to prevent extension in Object.seal"));
        return encodedJSValue();
    }

    return JSValue::encode(obj);
}

JSObject* objectConstructorFreeze(ExecState* exec, JSObject* object)
{
	VM& vm = exec->vm();
    ThrowScope scope(vm);

    if (isJSFinalObject(object) && !hasIndexedProperties(object->structure()->indexingType())) {
        object->freeze(vm);
        return object;
    }

    bool success = setIntegrityLevel<IntegrityLevel::Frozen>(exec, vm, object);
    RETURN_IF_EXCEPTION(scope, 0);
    if (UNLIKELY(!success))
        return throwTypeError(exec, ASCIILiteral("Unable to prevent extension in Object.freeze"));
    return object;
}

EncodedJSValue JSC_HOST_CALL objectConstructorFreeze(ExecState* exec)
{
	VM& vm = exec->vm();
    // 1. If Type(O) is not Object, return O.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(obj);
    JSObject* result = objectConstructorFreeze(exec, asObject(obj));
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(result);
}

EncodedJSValue JSC_HOST_CALL objectConstructorPreventExtensions(ExecState* exec)
{
	VM& vm = exec->vm();
    JSValue argument = exec->argument(0);
    if (!argument.isObject())
        return JSValue::encode(argument);
    JSObject* object = asObject(argument);
    object->methodTable(vm)->preventExtensions(object, exec);
    return JSValue::encode(object);
}

EncodedJSValue JSC_HOST_CALL objectConstructorIsSealed(ExecState* exec)
{
	VM& vm = exec->vm();

    // 1. If Type(O) is not Object, return true.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(jsBoolean(true));
    JSObject* object = asObject(obj);

    if (isJSFinalObject(object))
        return JSValue::encode(jsBoolean(object->isSealed(vm)));

    // 2. For each named own property name P of O,
    PropertyNameArray properties(vm, PropertyNameMode::StringsAndSymbols);
    object->methodTable()->getOwnPropertyNames(object, exec, properties, EnumerationMode(DontEnumPropertiesMode::Include));
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
	PropertyNameArray::const_iterator end = properties.end();
    for (PropertyNameArray::const_iterator iter = properties.begin(); iter != end; ++iter) {
    	Identifier propertyName = *iter;
		if (vm.propertyNames->isPrivateName(propertyName))
			continue;
        // a. Let desc be the result of calling the [[GetOwnProperty]] internal method of O with P.
        PropertyDescriptor desc;
        if (!object->getOwnPropertyDescriptor(exec, propertyName, desc))
            continue;
        // b. If desc.[[Configurable]] is true, then return false.
        if (desc.configurable())
            return JSValue::encode(jsBoolean(false));
    }

    // 3. If the [[Extensible]] internal property of O is false, then return true.
    // 4. Otherwise, return false.
    bool isExtensible = object->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(jsBoolean(!isExtensible));
}

EncodedJSValue JSC_HOST_CALL objectConstructorIsFrozen(ExecState* exec)
{
    // 1. If Type(O) is not Object, return true.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(jsBoolean(true));
    JSObject* object = asObject(obj);

	VM& vm = exec->vm();

    if (isJSFinalObject(object))
        return JSValue::encode(jsBoolean(object->isFrozen(vm)));

    // 2. For each named own property name P of O,
    PropertyNameArray properties(vm, PropertyNameMode::StringsAndSymbols);
    object->methodTable()->getOwnPropertyNames(object, exec, properties, EnumerationMode(DontEnumPropertiesMode::Include));
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    PropertyNameArray::const_iterator end = properties.end();
    for (PropertyNameArray::const_iterator iter = properties.begin(); iter != end; ++iter) {
    	 Identifier propertyName = *iter;
		if (vm.propertyNames->isPrivateName(propertyName))
			continue;
        // a. Let desc be the result of calling the [[GetOwnProperty]] internal method of O with P.
        PropertyDescriptor desc;
        if (!object->getOwnPropertyDescriptor(exec, propertyName, desc))
            continue;
        // b. If IsDataDescriptor(desc) is true then
        // i. If desc.[[Writable]] is true, return false. c. If desc.[[Configurable]] is true, then return false.
        if ((desc.isDataDescriptor() && desc.writable()) || desc.configurable())
            return JSValue::encode(jsBoolean(false));
    }

    // 3. If the [[Extensible]] internal property of O is false, then return true.
    // 4. Otherwise, return false.
    bool isExtensible = object->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(jsBoolean(!isExtensible));
}

EncodedJSValue JSC_HOST_CALL objectConstructorIsExtensible(ExecState* exec)
{
    VM& vm = exec->vm();
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(jsBoolean(false));
    JSObject* object = asObject(obj);
    bool isExtensible = object->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSValue());
    return JSValue::encode(jsBoolean(isExtensible));
}

EncodedJSValue JSC_HOST_CALL objectConstructorIs(ExecState* exec)
{
    return JSValue::encode(jsBoolean(sameValue(exec, exec->argument(0), exec->argument(1))));
}

// FIXME: Use the enumeration cache.
JSArray* ownPropertyKeys(ExecState* exec, JSObject* object, PropertyNameMode propertyNameMode, EnumerationMode mode)
{
    VM& vm = exec->vm();
    PropertyNameArray properties(vm, propertyNameMode);
    object->methodTable(vm)->getOwnPropertyNames(object, exec, properties, mode);
    RETURN_IF_EXCEPTION(vm, 0);

    if (propertyNameMode == PropertyNameMode::Strings && properties.size() < MIN_SPARSE_ARRAY_INDEX) {
    	auto* globalObject = exec->lexicalGlobalObject();
		if (LIKELY(!globalObject->isHavingABadTime())) {
			size_t numProperties = properties.size();
			JSArray* keys = JSArray::create(vm, globalObject->arrayStructureForIndexingTypeDuringAllocation(ArrayWithContiguous), numProperties);
			WriteBarrier<Unknown>* buffer = keys->butterfly()->contiguous().data();
			for (size_t i = 0; i < numProperties; i++) {
				const Identifier& identifier = properties[i];
				ASSERT(!identifier.isSymbol());
				buffer[i].set(vm, keys, jsOwnedString(&vm, identifier.string()));
			}
			return keys;
		}
    }

    JSArray* keys = constructEmptyArray(exec, 0);
    RETURN_IF_EXCEPTION(vm, 0);

    unsigned index = 0;
    /*
    auto pushDirect = [&] (ExecState* exec, JSArray* array, JSValue value) {
        array->putDirectIndex(exec, index++, value);
    };
    */

    switch (propertyNameMode) {
    case PropertyNameMode::Strings: {
        size_t numProperties = properties.size();
        for (size_t i = 0; i < numProperties; i++) {
            const Identifier& identifier = properties[i];
            ASSERT(!identifier.isSymbol());
			keys->putDirectIndex(exec, index++, jsOwnedString(vm, identifier.string()));
			RETURN_IF_EXCEPTION(vm, 0);
        }
        break;
    }

    case PropertyNameMode::Symbols: {
        size_t numProperties = properties.size();
        for (size_t i = 0; i < numProperties; i++) {
            const Identifier& identifier = properties[i];
            ASSERT(identifier.isSymbol());
            if (!vm.propertyNames->isPrivateName(identifier)) {
				keys->putDirectIndex(exec, index++, Symbol::create(vm, static_cast<SymbolImpl&>(*identifier.impl())));
                RETURN_IF_EXCEPTION(vm, 0);
            }
        }
        break;
    }

    case PropertyNameMode::StringsAndSymbols: {
        Vector<Identifier, 16> propertySymbols;
        size_t numProperties = properties.size();
        for (size_t i = 0; i < numProperties; i++) {
            const Identifier& identifier = properties[i];
            if (identifier.isSymbol() && !vm.propertyNames->isPrivateName(identifier))
				propertySymbols.append(identifier);
            else {
				keys->putDirectIndex(exec, index++, jsOwnedString(vm, identifier.string()));
                RETURN_IF_EXCEPTION(vm, 0);
            }
        }

        // To ensure the order defined in the spec (9.1.12), we append symbols at the last elements of keys.
        for(int i = 0; i < propertySymbols.size(); ++i) {
        	const Identifier& identifier = propertySymbols[i];
			keys->putDirectIndex(exec, index++, Symbol::create(vm, static_cast<SymbolImpl&>(*identifier.impl())));
			RETURN_IF_EXCEPTION(vm, 0);
        }

        break;
    }
    }

    return keys;
}

} // namespace JSC
