/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2002, 2005, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#ifndef Operations_h
#define Operations_h

#include "JSCInlines.h"

namespace JSC {

#define InvalidPrototypeChain (std::numeric_limits<size_t>::max())

NEVER_INLINE JSValue jsAddSlowCase(JSGlobalData&, CallFrame*, JSValue, JSValue);
JSValue jsTypeStringForValue(CallFrame*, JSValue);
JSValue jsTypeStringForValue(JSGlobalData&, JSGlobalObject*, JSValue);
bool jsIsObjectTypeOrNull(CallFrame*, JSValue);
bool jsIsFunctionType(JSValue);
size_t normalizePrototypeChain(CallFrame*, Structure*);

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm, ExecState* exec, const String& u1, JSString* s2)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    unsigned length1 = u1.length();
    if (!length1)
        return s2;
    unsigned length2 = s2->length();
    if (!length2)
        return jsString(vm, u1);
    //static_assert(JSString::MaxLength == std::numeric_limits<int32_t>::max(), "");
    if (sumOverflows<int32_t>(length1, length2)) {
        throwOutOfMemoryError(exec);
        return 0;
    }

    // (1) Cost of making JSString    : sizeof(JSString) (for new string) + sizeof(StringImpl header) + length1 + length2
    // (2) Cost of making JSRopeString: sizeof(JSString) (for u1) + sizeof(JSRopeString)
    // We do not account u1 cost in (2) since u1 may be shared StringImpl, and it may not introduce additional cost.
    // We conservatively consider the cost of u1. Currently, we are not considering about is8Bit() case because 16-bit
    // strings are relatively rare. But we can do that if we need to consider it.
    if (s2->isRope() || (StringImpl::headerSize<LChar>() + length1 + length2) >= sizeof(JSRopeString))
        return JSRopeString::create(vm, jsString(vm, u1), s2);

    ASSERT(!s2->isRope());
    const String& u2 = s2->value(exec);
    //scope.assertNoException();
    String newString = tryMakeString(u1, u2);
    if (!newString) {
        throwOutOfMemoryError(exec);
        return 0;
    }
    return JSString::create(vm, newString.releaseImpl());
}

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm, ExecState* exec, JSString* s1, const String& u2)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    unsigned length1 = s1->length();
    if (!length1)
        return jsString(&vm, u2);
    unsigned length2 = u2.length();
    if (!length2)
        return s1;
    //static_assert(JSString::MaxLength == std::numeric_limits<int32_t>::max(), "");
    if (sumOverflows<int32_t>(length1, length2)) {
        throwOutOfMemoryError(exec);
        return 0;
    }

    // (1) Cost of making JSString    : sizeof(JSString) (for new string) + sizeof(StringImpl header) + length1 + length2
    // (2) Cost of making JSRopeString: sizeof(JSString) (for u2) + sizeof(JSRopeString)
    if (s1->isRope() || (StringImpl::headerSize<LChar>() + length1 + length2) >= sizeof(JSRopeString))
        return JSRopeString::create(vm, s1, jsString(&vm, u2));

    ASSERT(!s1->isRope());
    const String& u1 = s1->value(exec);
    //scope.assertNoException();
    String newString = tryMakeString(u1, u2);
    if (!newString) {
        throwOutOfMemoryError(exec);
        return 0;
    }
    return JSString::create(vm, newString.releaseImpl());
}

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm, ExecState* exec, JSString* s1, JSString* s2)
{
	int32_t length1 = s1->length();
	if (!length1)
		return s2;
	int32_t length2 = s2->length();
	if (!length2)
		return s1;
    if (sumOverflows<int32_t>(length1, length2)) {
        throwOutOfMemoryError(exec);
        return 0;
    }

	return JSRopeString::create(vm, s1, s2);
}

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm, ExecState* exec, JSString* s1, JSString* s2, JSString* s3)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    int32_t length1 = s1->length();
    if (!length1) {
        //scope.release();
        return jsString(vm, exec, s2, s3);
    }
    int32_t length2 = s2->length();
    if (!length2) {
        //scope.release();
        return jsString(vm, exec, s1, s3);
    }
    int32_t length3 = s3->length();
    if (!length3) {
        //scope.release();
        return jsString(vm, exec, s1, s2);
    }

    if (sumOverflows<int32_t>(length1, length2, length3)) {
        throwOutOfMemoryError(exec);
        return 0;
    }

    return JSRopeString::create(vm, s1, s2, s3);
}

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm,ExecState* exec, const String& u1, const String& u2)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    unsigned length1 = u1.length();
    if (!length1)
        return jsString(&vm, u2);
    unsigned length2 = u2.length();
    if (!length2)
        return jsString(&vm, u1);
    //static_assert(JSString::MaxLength == std::numeric_limits<int32_t>::max(), "");
    if (sumOverflows<int32_t>(length1, length2)) {
        throwOutOfMemoryError(exec);
        return 0;
    }

    // (1) Cost of making JSString    : sizeof(JSString) (for new string) + sizeof(StringImpl header) + length1 + length2
    // (2) Cost of making JSRopeString: sizeof(JSString) (for u1) + sizeof(JSString) (for u2) + sizeof(JSRopeString)
    if ((StringImpl::headerSize<LChar>() + length1 + length2) >= (sizeof(JSRopeString) + sizeof(JSString)))
        return JSRopeString::create(vm, jsString(&vm, u1), jsString(&vm, u2));

    String newString = tryMakeString(u1, u2);
    if (!newString) {
        throwOutOfMemoryError(exec);
        return 0;
    }
    return JSString::create(vm, newString.releaseImpl());
}

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm, ExecState* exec, const String& u1, const String& u2, const String& u3)
{
	int32_t length1 = u1.length();
	int32_t length2 = u2.length();
	int32_t length3 = u3.length();

    if (length1 < 0 || length2 < 0 || length3 < 0) {
        throwOutOfMemoryError(exec);
        return 0;
    }

    if (!length1) {
        //scope.release();
        return jsString(vm, exec, u2, u3);
    }
    if (!length2) {
        //scope.release();
        return jsString(vm, exec, u1, u3);
    }
    if (!length3) {
        //scope.release();
        return jsString(vm, exec, u1, u2);
    }

    if (sumOverflows<int32_t>(length1, length2, length3)) {
        throwOutOfMemoryError(exec);
        return 0;
    }

    // (1) Cost of making JSString    : sizeof(JSString) (for new string) + sizeof(StringImpl header) + length1 + length2 + length3
    // (2) Cost of making JSRopeString: sizeof(JSString) (for u1) + sizeof(JSString) (for u2) + sizeof(JSString) (for u3) + sizeof(JSRopeString)
    if ((StringImpl::headerSize<LChar>() + length1 + length2 + length3) >= (sizeof(JSRopeString) + sizeof(JSString) * 2))
        return JSRopeString::create(vm, jsString(vm, u1), jsString(vm, u2), jsString(vm, u3));

    String newString = tryMakeString(u1, u2, u3);
    if (!newString) {
        throwOutOfMemoryError(exec);
        return 0;
    }
    return JSString::create(vm, newString.releaseImpl());
}

ALWAYS_INLINE JSValue jsStringFromRegisterArray(ExecState* exec, Register* strings, unsigned count)
{
	JSGlobalData* vm = &exec->globalData();
	JSRopeString::RopeBuilder<RecordOverflow> ropeBuilder(*vm);

	for (unsigned i = 0; i < count; ++i) {
		JSValue v = strings[-static_cast<int>(i)].jsValue();
        JSString* string = v.toString(exec);
        RETURN_IF_EXCEPTION(*vm, JSValue());
		if(UNLIKELY(!ropeBuilder.append(string)))
			return throwOutOfMemoryError(exec);
	}

	return ropeBuilder.release();
}

ALWAYS_INLINE JSValue jsStringFromArguments(ExecState* exec, JSValue thisValue)
{
	JSGlobalData* vm = &exec->globalData();
	JSRopeString::RopeBuilder<RecordOverflow> ropeBuilder(*vm);
    JSString* str = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(*vm, JSValue());
    ropeBuilder.append(str);

	for (unsigned i = 0; i < exec->argumentCount(); ++i) {
		JSValue v = exec->argument(i);
        JSString* str = v.toString(exec);
        RETURN_IF_EXCEPTION(*vm, JSValue());
        if (UNLIKELY(!ropeBuilder.append(str)))
			return throwOutOfMemoryError(exec);
	}

	return ropeBuilder.release();
}

// See ES5 11.8.1/11.8.2/11.8.5 for definition of leftFirst, this value ensures correct
// evaluation ordering for argument conversions for '<' and '>'. For '<' pass the value
// true, for leftFirst, for '>' pass the value false (and reverse operand order).
template<bool leftFirst>
ALWAYS_INLINE bool jsLess(CallFrame* callFrame, JSValue v1, JSValue v2)
{
	if (v1.isInt32() && v2.isInt32())
		return v1.asInt32() < v2.asInt32();

	if (v1.isNumber() && v2.isNumber())
		return v1.asNumber() < v2.asNumber();

	if (isJSString(v1) && isJSString(v2))
        return codePointCompareLessThan(asString(v1)->value(callFrame), asString(v2)->value(callFrame));

    VM& vm = callFrame->vm();

	double n1;
	double n2;
	JSValue p1;
	JSValue p2;
	bool wasNotString1;
	bool wasNotString2;
	if (leftFirst) {
		wasNotString1 = v1.getPrimitiveNumber(callFrame, n1, p1);
        RETURN_IF_EXCEPTION(vm, false);
		wasNotString2 = v2.getPrimitiveNumber(callFrame, n2, p2);
	} else {
		wasNotString2 = v2.getPrimitiveNumber(callFrame, n2, p2);
        RETURN_IF_EXCEPTION(vm, false);
		wasNotString1 = v1.getPrimitiveNumber(callFrame, n1, p1);
	}
    RETURN_IF_EXCEPTION(vm, false);

	if (wasNotString1 | wasNotString2)
		return n1 < n2;
	return codePointCompareLessThan(asString(p1)->value(callFrame), asString(p2)->value(callFrame));
}

// See ES5 11.8.3/11.8.4/11.8.5 for definition of leftFirst, this value ensures correct
// evaluation ordering for argument conversions for '<=' and '=>'. For '<=' pass the
// value true, for leftFirst, for '=>' pass the value false (and reverse operand order).
template<bool leftFirst>
ALWAYS_INLINE bool jsLessEq(CallFrame* callFrame, JSValue v1, JSValue v2)
{
	if (v1.isInt32() && v2.isInt32())
		return v1.asInt32() <= v2.asInt32();

	if (v1.isNumber() && v2.isNumber())
		return v1.asNumber() <= v2.asNumber();

	if (isJSString(v1) && isJSString(v2))
		return !codePointCompareLessThan(asString(v2)->value(callFrame), asString(v1)->value(callFrame));

    VM& vm = callFrame->vm();

	double n1;
	double n2;
	JSValue p1;
	JSValue p2;
	bool wasNotString1;
	bool wasNotString2;
	if (leftFirst) {
		wasNotString1 = v1.getPrimitiveNumber(callFrame, n1, p1);
	    RETURN_IF_EXCEPTION(vm, false);
		wasNotString2 = v2.getPrimitiveNumber(callFrame, n2, p2);
	} else {
		wasNotString2 = v2.getPrimitiveNumber(callFrame, n2, p2);
	    RETURN_IF_EXCEPTION(vm, false);
		wasNotString1 = v1.getPrimitiveNumber(callFrame, n1, p1);
	}
    RETURN_IF_EXCEPTION(vm, false);

	if (wasNotString1 | wasNotString2)
		return n1 <= n2;
	return !codePointCompareLessThan(asString(p2)->value(callFrame), asString(p1)->value(callFrame));
}

// Fast-path choices here are based on frequency data from SunSpider:
//    <times> Add case: <t1> <t2>
//    ---------------------------
//    5626160 Add case: 3 3 (of these, 3637690 are for immediate values)
//    247412  Add case: 5 5
//    20900   Add case: 5 6
//    13962   Add case: 5 3
//    4000    Add case: 3 5

ALWAYS_INLINE JSValue jsAddNonNumber(JSGlobalData& vm, CallFrame* callFrame, JSValue v1, JSValue v2)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);
    ASSERT(!v1.isNumber() || !v2.isNumber());

    if (LIKELY(v1.isString() && !v2.isObject())) {
        if (v2.isString())
            return jsString(vm, callFrame, asString(v1), asString(v2));
        String s2 = v2.toWTFString(callFrame);
        RETURN_IF_EXCEPTION(vm, JSValue());
        return jsString(vm, callFrame, asString(v1), s2);
    }

    // All other cases are pretty uncommon
    return jsAddSlowCase(vm, callFrame, v1, v2);
}

ALWAYS_INLINE JSValue jsAdd(JSGlobalData& vm, CallFrame* callFrame, JSValue v1, JSValue v2)
{
	if (v1.isNumber() && v2.isNumber())
		return jsNumber(v1.asNumber() + v2.asNumber());

    return jsAddNonNumber(vm, callFrame, v1, v2);
}

inline bool scribbleFreeCells()
{
    return !ASSERT_DISABLED || Options::scribbleFreeCells();
}

#define SCRIBBLE_WORD static_cast<intptr_t>(0xbadbeef0)

inline bool isScribbledValue(JSValue value)
{
    return JSValue::encode(value) == JSValue::encode(bitwise_cast<JSCell*>(SCRIBBLE_WORD));
}

inline void scribble(void* base, size_t size)
{
    for (size_t i = size / sizeof(EncodedJSValue); i--;) {
        // Use a 16-byte aligned value to ensure that it passes the cell check.
        static_cast<EncodedJSValue*>(base)[i] = JSValue::encode(bitwise_cast<JSCell*>(SCRIBBLE_WORD));
    }
}

} // namespace JSC

#endif // Operations_h
