/*
 *  Copyright (C) 2006, 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#ifndef PropertyNameArray_h
#define PropertyNameArray_h

#include "CallFrame.h"
#include "Identifier.h"
#include <wtf/HashSet.h>
#include <wtf/OwnArrayPtr.h>
#include <wtf/Vector.h>

namespace JSC {

class JSPropertyNameEnumerator;
class Structure;

// FIXME: Rename to PropertyNameArray.
class PropertyNameArrayData : public RefCounted<PropertyNameArrayData> {
public:
	typedef Vector<Identifier, 20> PropertyNameVector;

	static PassRefPtr<PropertyNameArrayData> create() { return adoptRef(new PropertyNameArrayData); }

	PropertyNameVector& propertyNameVector() { return m_propertyNameVector; }

private:
	PropertyNameArrayData()
	{
	}

	PropertyNameVector m_propertyNameVector;
};

// FIXME: Rename to PropertyNameArrayBuilder.
class PropertyNameArray {
public:
	ALWAYS_INLINE PropertyNameArray(JSGlobalData& globalData, PropertyNameMode mode)
		: m_data(PropertyNameArrayData::create())
		, m_vm(&globalData)
    	, m_mode(mode)
	{
	}

	ALWAYS_INLINE PropertyNameArray(JSGlobalData* globalData, PropertyNameMode mode)
		: m_data(PropertyNameArrayData::create())
		, m_vm(globalData)
    	, m_mode(mode)
	{
	}

	JSGlobalData* globalData() { return m_vm; }
	JSGlobalData* vm() { return m_vm; }

    void add(uint32_t index)
    {
        add(Identifier::from(*m_vm, index));
    }

	void add(const Identifier& identifier);
	void add(UniquedStringImpl*);
    void addUnchecked(UniquedStringImpl* identifier);

	ALWAYS_INLINE Identifier& operator[](unsigned i) { return m_data->propertyNameVector()[i]; }
	ALWAYS_INLINE const Identifier& operator[](unsigned i) const { return m_data->propertyNameVector()[i]; }

	void setData(PassRefPtr<PropertyNameArrayData> data) { m_data = data; }
	PropertyNameArrayData* data() { return m_data.get(); }
	PassRefPtr<PropertyNameArrayData> releaseData() { return m_data.release(); }

	// FIXME: Remove these functions.
	bool canAddKnownUniqueForStructure() const { return m_data->propertyNameVector().isEmpty(); }
	typedef PropertyNameArrayData::PropertyNameVector::const_iterator const_iterator;
	size_t size() const { return m_data->propertyNameVector().size(); }
	ALWAYS_INLINE const_iterator begin() const { return m_data->propertyNameVector().begin(); }
	ALWAYS_INLINE const_iterator end() const { return m_data->propertyNameVector().end(); }

    PropertyNameMode mode() const { return m_mode; }
    PropertyNameMode propertyNameMode() const { return m_mode; }
    bool includeSymbolProperties() const;
    bool includeStringProperties() const;

private:
    bool isUidMatchedToTypeMode(UniquedStringImpl* identifier);

	RefPtr<PropertyNameArrayData> m_data;
    HashSet<UniquedStringImpl*> m_set;
	JSGlobalData* m_vm;
    PropertyNameMode m_mode;
};

ALWAYS_INLINE void PropertyNameArray::add(const Identifier& identifier)
{
    add(identifier.impl());
}

ALWAYS_INLINE void PropertyNameArray::addUnchecked(UniquedStringImpl* identifier)
{
    if (!isUidMatchedToTypeMode(identifier))
        return;
    m_data->propertyNameVector().append(Identifier::fromUid(*m_vm, identifier));
}

ALWAYS_INLINE void PropertyNameArray::add(UniquedStringImpl* identifier)
{
    static const unsigned setThreshold = 20;

    ASSERT(identifier);

    if (!isUidMatchedToTypeMode(identifier))
        return;

    if (size() < setThreshold) {
        if (m_data->propertyNameVector().contains(identifier))
            return;
    } else {
        if (m_set.isEmpty()) {
        	for(int i = 0; i < m_data->propertyNameVector().size(); ++i)
                m_set.add(m_data->propertyNameVector()[i].impl());
        }
        if (!m_set.add(identifier).isNewEntry)
            return;
    }

    addUnchecked(identifier);
}

ALWAYS_INLINE bool PropertyNameArray::isUidMatchedToTypeMode(UniquedStringImpl* identifier)
{
    if (identifier->isSymbol())
        return includeSymbolProperties();
    return includeStringProperties();
}

ALWAYS_INLINE bool PropertyNameArray::includeSymbolProperties() const
{
    return static_cast<unsigned>(m_mode) & static_cast<unsigned>(PropertyNameMode::Symbols);
}

ALWAYS_INLINE bool PropertyNameArray::includeStringProperties() const
{
    return static_cast<unsigned>(m_mode) & static_cast<unsigned>(PropertyNameMode::Strings);
}

} // namespace JSC

#endif // PropertyNameArray_h
