/*
 *  Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef RegExpMatchesArray_h
#define RegExpMatchesArray_h

#include "ButterflyInlineMethods.h"
#include "JSArray.h"
#include "JSCInlines.h"
#include "JSGlobalObject.h"
#include "RegExpInlines.h"
#include "RegExpObject.h"

namespace JSC {

static const PropertyOffset RegExpMatchesArrayIndexPropertyOffset = 100;
static const PropertyOffset RegExpMatchesArrayInputPropertyOffset = 101;

inline JSArray* tryCreateUninitializedRegExpMatchesArray(VM& vm, Structure* structure, unsigned initialLength)
{
    unsigned vectorLength = std::max(BASE_ARRAY_STORAGE_VECTOR_LEN, initialLength);
    if (vectorLength > MAX_STORAGE_VECTOR_LENGTH)
        return 0;

    JSGlobalObject* globalObject = structure->globalObject();
    bool createUninitialized = globalObject->isOriginalArrayStructure(structure);
    void* temp;
    if (UNLIKELY(!vm.heap.tryAllocateStorage(Butterfly::totalSize(0, structure->outOfLineCapacity(), true, vectorLength * sizeof(EncodedJSValue)), &temp)))
        return 0;
    Butterfly* butterfly = Butterfly::fromBase(temp, 0, structure->outOfLineCapacity());
    butterfly->setVectorLength(vectorLength);
    butterfly->setPublicLength(initialLength);

    JSArray* result = JSArray::createWithButterfly(vm, structure, butterfly);
    return result;
}

inline JSArray* createRegExpMatchesArray(
    VM& vm, JSGlobalObject* globalObject, JSString* input, const String& inputValue,
    RegExp* regExp, unsigned startOffset, MatchResult& result)
{
    SamplingRegion samplingRegion("createRegExpMatchesArray");

    Vector<int, 32> subpatternResults;
    int position = regExp->matchInline(vm, inputValue, startOffset, subpatternResults);
    if (position == -1) {
        result = MatchResult::failed();
        return 0;
    }

    result.start = position;
    result.end = subpatternResults[1];

    JSArray* array;

    // FIXME: This should handle array allocation errors gracefully.
    // https://bugs.webkit.org/show_bug.cgi?id=155144

    unsigned numSubpatterns = regExp->numSubpatterns();

    DeferGC deferGC(vm.heap);

    if (UNLIKELY(globalObject->isHavingABadTime())) {
        array = JSArray::tryCreateUninitializedRestricted(vm, globalObject->regExpMatchesArrayStructure(), numSubpatterns + 1);
        // FIXME: we should probably throw an out of memory error here, but
		// when making this change we should check that all clients of this
		// function will correctly handle an exception being thrown from here.
		// https://bugs.webkit.org/show_bug.cgi?id=169786
		RELEASE_ASSERT(array);
        if(UNLIKELY(!array)) {
            result = MatchResult::failed();
            return 0;
        }

        array->initializeIndex(vm, 0, jsSubstringOfResolved(vm, input, result.start, result.end - result.start));

		for (unsigned i = 1; i <= numSubpatterns; ++i) {
			int start = subpatternResults[2 * i];
			JSValue value;
			if (start >= 0)
				value = JSRopeString::createSubstringOfResolved(vm, input, start, subpatternResults[2 * i + 1] - start);
			else
				value = jsUndefined();
			array->initializeIndex(vm, i, value);
		}
    } else {
        array = tryCreateUninitializedRegExpMatchesArray(vm, globalObject->regExpMatchesArrayStructure(), numSubpatterns + 1);
        if(UNLIKELY(!array)) {
			result = MatchResult::failed();
			return 0;
		}

        array->initializeIndex(vm, 0, jsSubstringOfResolved(vm, input, result.start, result.end - result.start), ArrayWithContiguous);

		for (unsigned i = 1; i <= numSubpatterns; ++i) {
			int start = subpatternResults[2 * i];
			JSValue value;
			if (start >= 0)
				value = JSRopeString::createSubstringOfResolved(vm, input, start, subpatternResults[2 * i + 1] - start);
			else
				value = jsUndefined();
			array->initializeIndex(vm, i, value, ArrayWithContiguous);
		}
    }

    array->putDirect(vm, RegExpMatchesArrayIndexPropertyOffset, jsNumber(result.start));
    array->putDirect(vm, RegExpMatchesArrayInputPropertyOffset, input);

    return array;
}

JSArray* createRegExpMatchesArray(ExecState*, JSGlobalObject*, JSString*, RegExp*, unsigned startOffset, MatchResult&);
inline JSArray* createRegExpMatchesArray(VM& vm, ExecState* exec, JSGlobalObject* globalObject, JSString* string, RegExp* regExp, unsigned startOffset)
{
    MatchResult ignoredResult;
    return createRegExpMatchesArray(vm, globalObject, string, string->value(exec), regExp, startOffset, ignoredResult);
}
inline JSArray* createRegExpMatchesArray(ExecState* exec, JSGlobalObject* globalObject, JSString* string, RegExp* regExp, unsigned startOffset)
{
    return createRegExpMatchesArray(globalObject->vm(), exec, globalObject, string, regExp, startOffset);
}
JSArray* createEmptyRegExpMatchesArray(JSGlobalObject*, JSString*, RegExp*);
Structure* createRegExpMatchesArrayStructure(JSGlobalData&, JSGlobalObject*);
Structure* createRegExpMatchesArraySlowPutStructure(JSGlobalData&, JSGlobalObject*);

}

#endif // RegExpMatchesArray_h
