/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2007, 2008, 2012 Apple Inc. All Rights Reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef RegExpObject_h
#define RegExpObject_h

#include "JSObject.h"
#include "RegExp.h"

namespace JSC {
    
class JSGlobalObject;

class RegExpObject : public JSNonFinalObject {
public:
	typedef JSNonFinalObject Base;

	static RegExpObject* create(JSGlobalData& vm, Structure* structure, RegExp* regExp)
	{
		RegExpObject* object = new (NotNull, allocateCell<RegExpObject>(vm.heap)) RegExpObject(vm, structure, regExp);
		object->finishCreation(vm);
		return object;
	}

	void setRegExp(JSGlobalData& globalData, RegExp* r) { m_regExp.set(globalData, this, r); }
	RegExp* regExp() const { return m_regExp.get(); }

	bool setLastIndex(ExecState* exec, size_t lastIndex)
	{
		if (LIKELY(m_lastIndexIsWritable)) {
			m_lastIndex.setWithoutWriteBarrier(jsNumber(lastIndex));
			return true;
		}
		throwTypeError(exec, ReadonlyPropertyWriteError);
		return false;
	}
	bool setLastIndex(ExecState* exec, JSValue lastIndex, bool shouldThrow)
	{
		if (LIKELY(m_lastIndexIsWritable)) {
			m_lastIndex.set(exec->vm(), this, lastIndex);
			return true;
		}

		if (shouldThrow)
			throwTypeError(exec, ReadonlyPropertyWriteError);
		return false;
	}
	JSValue getLastIndex() const
	{
		return m_lastIndex.get();
	}

	bool test(ExecState* exec, JSGlobalObject* globalObject, JSString* string) { return !!match(exec, globalObject, string); }
	bool testInline(ExecState* exec, JSGlobalObject* globalObject, JSString* string) { return !!matchInline(exec, globalObject, string); }
	JSValue exec(ExecState*, JSGlobalObject*, JSString*);
	JSValue execInline(ExecState*, JSGlobalObject*, JSString*);
    MatchResult match(ExecState*, JSGlobalObject*, JSString*);
    JSValue matchGlobal(ExecState*, JSGlobalObject*, JSString*);

	static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);
	static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);

	DECLARE_INFO;

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype)
	{
		return Structure::create(globalData, globalObject, prototype, TypeInfo(RegExpObjectType, StructureFlags), info());
	}

    static ptrdiff_t offsetOfRegExp()
    {
        return OBJECT_OFFSETOF(RegExpObject, m_regExp);
    }

	static ptrdiff_t offsetOfLastIndex()
	{
		return OBJECT_OFFSETOF(RegExpObject, m_lastIndex);
	}

	static ptrdiff_t offsetOfLastIndexIsWritable()
	{
		return OBJECT_OFFSETOF(RegExpObject, m_lastIndexIsWritable);
	}

    static size_t allocationSize(Checked<size_t> inlineCapacity)
    {
        ASSERT_UNUSED(inlineCapacity, !inlineCapacity);
        return sizeof(RegExpObject);
    }

    static unsigned advanceStringUnicode(const String&, unsigned length, unsigned currentIndex);

protected:
	JS_EXPORT_PRIVATE RegExpObject(JSGlobalData&, Structure*, RegExp*);
	JS_EXPORT_PRIVATE void finishCreation(JSGlobalData&);

	static const unsigned StructureFlags = OverridesGetOwnPropertySlot | OverridesGetPropertyNames | Base::StructureFlags;

	static void visitChildren(JSCell*, SlotVisitor&);

	JS_EXPORT_PRIVATE static bool deleteProperty(JSCell*, ExecState*, PropertyName);
	JS_EXPORT_PRIVATE static void getOwnNonIndexPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);
	JS_EXPORT_PRIVATE static void getPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);
	JS_EXPORT_PRIVATE static void getGenericPropertyNames(JSObject*, ExecState*, PropertyNameArray&, EnumerationMode);
	JS_EXPORT_PRIVATE static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool shouldThrow);

private:
	MatchResult matchInline(ExecState*, JSGlobalObject*, JSString*);

	WriteBarrier<RegExp> m_regExp;
	WriteBarrier<Unknown> m_lastIndex;
	uint8_t m_lastIndexIsWritable;
};

RegExpObject* asRegExpObject(JSValue);

inline RegExpObject* asRegExpObject(JSValue value)
{
	ASSERT(asObject(value)->inherits(RegExpObject::info()));
	return static_cast<RegExpObject*>(asObject(value));
}

} // namespace JSC

#endif // RegExpObject_h
