/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef StructureInlines_h
#define StructureInlines_h

#include "ArrayPrototype.h"
#include "JSArrayBufferView.h"
#include "JSCJSValueInlines.h"
#include "JSGlobalObject.h"
#include "ObjectPrototype.h"
#include "PropertyMapHashTable.h"
#include "Structure.h"

namespace JSC {

ALWAYS_INLINE Structure* Structure::create(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype, const TypeInfo& typeInfo, const ClassInfo* classInfo, IndexingType indexingType, unsigned inlineCapacity)
{
	ASSERT(globalData.structureStructure);
	ASSERT(classInfo);
	Structure* structure = new (NotNull, allocateCell<Structure>(globalData.heap)) Structure(globalData, globalObject, prototype, typeInfo, classInfo, indexingType, inlineCapacity);
	structure->finishCreation(globalData);
	return structure;
}

ALWAYS_INLINE Structure* Structure::createStructure(JSGlobalData& globalData)
{
	ASSERT(!globalData.structureStructure);
	Structure* structure = new (NotNull, allocateCell<Structure>(globalData.heap)) Structure(globalData);
	structure->finishCreation(globalData, CreatingEarlyCell);
	return structure;
}

ALWAYS_INLINE Structure* Structure::create(JSGlobalData& globalData, Structure* structure, DeferredStructureTransitionWatchpointFire* deferred)
{
	ASSERT(globalData.structureStructure);
	Structure* newStructure = new (NotNull, allocateCell<Structure>(globalData.heap)) Structure(globalData, structure, deferred);
	newStructure->finishCreation(globalData);
	return newStructure;
}

inline bool Structure::mayInterceptIndexedAccesses() const
{
    if (indexingTypeIncludingHistory() & MayHaveIndexedAccessors)
        return true;

    // Consider a scenario where object O (of global G1)'s prototype is set to A
    // (of global G2), and G2 is already having a bad time. If an object B with
    // indexed accessors is then set as the prototype of A:
    //      O -> A -> B
    // Then, O should be converted to SlowPutArrayStorage (because it now has an
    // object with indexed accessors in its prototype chain). But it won't be
    // converted because this conversion is done by JSGlobalObject::haveAbadTime(),
    // but G2 is already having a bad time. We solve this by conservatively
    // treating A as potentially having indexed accessors if its global is already
    // having a bad time. Hence, when A is set as O's prototype, O will be
    // converted to SlowPutArrayStorage.

    JSGlobalObject* globalObject = this->globalObject();
    if (!globalObject)
        return false;
    return globalObject->isHavingABadTime();
}

inline JSObject* Structure::storedPrototypeObject() const
{
    JSValue value = m_prototype.get();
    if (value.isNull())
        return 0;
    return asObject(value);
}

inline Structure* Structure::storedPrototypeStructure() const
{
    JSObject* object = storedPrototypeObject();
    if (!object)
        return 0;
    return object->structure();
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& globalData, UniquedStringImpl* uid)
{
    ASSERT(!isCompilationThread());
	ASSERT(structure()->classInfo() == info());
    PropertyTable* propertyTable;
    materializePropertyMapIfNecessary(globalData, propertyTable);
    if (!propertyTable)
		return invalidOffset;

	PropertyMapEntry* entry = propertyTable->get(uid);
	return entry ? entry->offset : invalidOffset;
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& globalData, PropertyName propertyName)
{
    return get(globalData, propertyName.uid());
}

inline PropertyOffset Structure::getConcurrently(VM& vm, UniquedStringImpl* uid)
{
    unsigned attributesIgnored;
    return getConcurrently(vm, uid, attributesIgnored);
}

ALWAYS_INLINE PropertyOffset Structure::getWithoutMaterializing(VM& vm, PropertyName propertyName, unsigned& attributes)
{
	return getWithoutMaterializing(vm, propertyName.uid(), attributes);
}

ALWAYS_INLINE PropertyOffset Structure::getWithoutMaterializing(VM& vm, UniquedStringImpl* uid)
{
    unsigned attributesIgnored;
    return getWithoutMaterializing(vm, uid, attributesIgnored);
}

ALWAYS_INLINE PropertyOffset Structure::getWithoutMaterializing(VM& vm, PropertyName propertyName)
{
    unsigned attributesIgnored;
    return getWithoutMaterializing(vm, propertyName, attributesIgnored);
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& globalData, UniquedStringImpl* uid, unsigned& attributes)
{
    ASSERT(!isCompilationThread());
    ASSERT(structure()->classInfo() == info());

    PropertyTable* propertyTable;
    materializePropertyMapIfNecessary(globalData, propertyTable);
    if (!propertyTable)
        return invalidOffset;

    PropertyMapEntry* entry = propertyTable->get(uid);
    if (!entry)
        return invalidOffset;

    attributes = entry->attributes;
    return entry->offset;
}

ALWAYS_INLINE PropertyOffset Structure::getConcurrently(VM& vm, UniquedStringImpl* uid, unsigned& attributes)
{
	return get(vm, uid, attributes);
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& globalData, PropertyName propertyName, unsigned& attributes)
{
    return get(globalData, propertyName.uid(), attributes);
}

inline bool Structure::hasIndexingHeader(const JSCell* cell) const
{
    if (hasIndexedProperties(indexingType()))
        return true;

    if (!isTypedView(m_classInfo->typedArrayStorageType))
        return false;

    return jsCast<const JSArrayBufferView*>(cell)->mode() == WastefulTypedArray;
}

ALWAYS_INLINE bool Structure::masqueradesAsUndefined(JSGlobalObject* lexicalGlobalObject)
{
	return typeInfo().masqueradesAsUndefined() && globalObject() == lexicalGlobalObject;
}

ALWAYS_INLINE bool Structure::transitivelyTransitionedFrom(Structure* structureToFind)
{
	for (Structure* current = this; current; current = current->previousID()) {
		if (current == structureToFind)
			return true;
	}
	return false;
}

ALWAYS_INLINE JSValue Structure::prototypeForLookup(JSGlobalObject* globalObject) const
{
	if (isObject())
		return m_prototype.get();
	if(typeInfo().type() == StringType)
		return globalObject->stringPrototype();
    if (typeInfo().type() == SymbolType)
        return globalObject->symbolPrototype();
	return jsNull();
}

ALWAYS_INLINE JSValue Structure::prototypeForLookup(ExecState* exec) const
{
	return prototypeForLookup(exec->lexicalGlobalObject());
}

ALWAYS_INLINE StructureChain* Structure::prototypeChain(JSGlobalData& vm, JSGlobalObject* globalObject) const
{
    // We cache our prototype chain so our clients can share it.
    if (!isValid(globalObject, m_cachedPrototypeChain.get())) {
        JSValue prototype = prototypeForLookup(globalObject);
        m_cachedPrototypeChain.set(vm, this, StructureChain::create(vm, prototype.isNull() ? 0 : asObject(prototype)->structure()));
    }
    return m_cachedPrototypeChain.get();
}

ALWAYS_INLINE StructureChain* Structure::prototypeChain(ExecState* exec) const
{
    return prototypeChain(exec->vm(), exec->lexicalGlobalObject());
}

inline bool Structure::isValid(JSGlobalObject* globalObject, StructureChain* cachedPrototypeChain) const
{
	if (!cachedPrototypeChain)
		return false;

	JSValue prototype = prototypeForLookup(globalObject);
	WriteBarrier<Structure>* cachedStructure = cachedPrototypeChain->head();
	while(*cachedStructure && !prototype.isNull()) {
		if (asObject(prototype)->structure() != cachedStructure->get())
			return false;
		++cachedStructure;
		prototype = asObject(prototype)->getPrototypeDirect();
	}
	return prototype.isNull() && !*cachedStructure;
}

ALWAYS_INLINE bool Structure::isValid(ExecState* exec, StructureChain* cachedPrototypeChain) const
{
    return isValid(exec->lexicalGlobalObject(), cachedPrototypeChain);
}

ALWAYS_INLINE bool Structure::putWillGrowOutOfLineStorage()
{
    checkOffsetConsistency();

    ASSERT(outOfLineCapacity() >= outOfLineSize());

    if (!propertyTable()) {
        unsigned currentSize = numberOfOutOfLineSlotsForLastOffset(m_offset);
        ASSERT(outOfLineCapacity() >= currentSize);
        return currentSize == outOfLineCapacity();
    }

    ASSERT(totalStorageCapacity() >= propertyTable()->propertyStorageSize());
    if (propertyTable()->hasDeletedOffset())
        return false;

    ASSERT(totalStorageCapacity() >= propertyTable()->size());
    return propertyTable()->size() == totalStorageCapacity();
}

ALWAYS_INLINE WriteBarrier<PropertyTable>& Structure::propertyTable()
{
    ASSERT(!globalObject() || !globalObject()->vm().heap.isBusy());
    return m_propertyTableUnsafe;
}

inline void Structure::didReplaceProperty(VM& vm, PropertyOffset offset)
{
    if (LIKELY(!hasRareData()))
        return;
    StructureRareData::PropertyWatchpointMap* map = rareData()->m_replacementWatchpointSets.get();
    if (LIKELY(!map))
        return;
    RefPtr<WatchpointSet> set = map->get(offset);
    if (LIKELY(!set))
        return;
    set->fireAll(vm, "Property did get replaced");
}

inline RefPtr<WatchpointSet> Structure::propertyReplacementWatchpointSet(PropertyOffset offset)
{
    //ConcurrentJITLocker locker(m_lock);
    if (!hasRareData())
        return nullptr;
    WTF::loadLoadFence();
    StructureRareData::PropertyWatchpointMap* map = rareData()->m_replacementWatchpointSets.get();
    if (!map)
        return nullptr;
    return map->get(offset);
}

ALWAYS_INLINE bool Structure::checkOffsetConsistency() const
{
#if !ASSERT_DISABLED
    PropertyTable* propertyTable = m_propertyTableUnsafe.get();

    if (!propertyTable) {
        ASSERT(!isPinnedPropertyTable());
        return true;
    }

    // We cannot reliably assert things about the property table in the concurrent
    // compilation thread. It is possible for the table to be stolen and then have
    // things added to it, which leads to the offsets being all messed up. We could
    // get around this by grabbing a lock here, but I think that would be overkill.
    if (isCompilationThread())
        return true;

    RELEASE_ASSERT(numberOfSlotsForLastOffset(m_offset, m_inlineCapacity) == propertyTable->propertyStorageSize());
    unsigned totalSize = propertyTable->propertyStorageSize();
    RELEASE_ASSERT((totalSize < inlineCapacity() ? 0 : totalSize - inlineCapacity()) == numberOfOutOfLineSlotsForLastOffset(m_offset));
#endif

    return true;
}

ALWAYS_INLINE size_t nextOutOfLineStorageCapacity(size_t currentCapacity)
{
    if (!currentCapacity)
        return initialOutOfLineCapacity;
    return currentCapacity * outOfLineGrowthFactor;
}

ALWAYS_INLINE size_t Structure::suggestedNewOutOfLineStorageCapacity()
{
    return nextOutOfLineStorageCapacity(outOfLineCapacity());
}

inline void Structure::setObjectToStringValue(ExecState* exec, VM& vm, JSString* value, PropertySlot toStringTagSymbolSlot)
{
    if (!hasRareData())
        allocateRareData(vm);
    rareData()->setObjectToStringValue(exec, vm, this, value, toStringTagSymbolSlot);
}

} // namespace JSC

#endif // StructureInlines_h

