/*
 * Copyright (C) 2009 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef WeakGCMap_h
#define WeakGCMap_h

#include "Weak.h"
#include <wtf/HashMap.h>

namespace JSC {

// A HashMap with Weak<JSCell> values, which automatically removes values once they're garbage collected.

class WeakGCMapBase {
public:
    virtual ~WeakGCMapBase() { }
    virtual void pruneStaleEntries() = 0;
};

template<typename KeyArg, typename RawMappedArg, typename HashArg = typename DefaultHash<KeyArg>::Hash,
    typename KeyTraitsArg = HashTraits<KeyArg> >
class WeakGCMap : public WeakGCMapBase {
    typedef Weak<RawMappedArg> MappedType;
    typedef HashMap<KeyArg, MappedType, HashArg, KeyTraitsArg> HashMapType;
    typedef HashTraits<MappedType> MappedTraits;
    typedef typename MappedTraits::PassInType MappedPassInType;

public:
    typedef typename HashMapType::KeyType KeyType;
    typedef typename HashMapType::AddResult AddResult;
    typedef typename HashMapType::iterator iterator;
    typedef typename HashMapType::const_iterator const_iterator;

    explicit WeakGCMap(JSGlobalData&);
    virtual ~WeakGCMap();

    ALWAYS_INLINE RawMappedArg* get(const KeyType& key) const
    {
        return m_map.get(key);
    }

    AddResult set(const KeyType& key, MappedPassInType value)
    {
        return m_map.set(key, value);
    }

    ALWAYS_INLINE void remove(const KeyType& key)
    {
        m_map.remove(key);
    }

    ALWAYS_INLINE void clear()
    {
        m_map.clear();
    }

    bool isEmpty() const
    {
        const_iterator it = m_map.begin();
        const_iterator end = m_map.end();
        while (it != end) {
            if (it->value)
                return true;
        }
        return false;
    }

    iterator find(const KeyType& key)
    {
        iterator it = m_map.find(key);
        iterator end = m_map.end();
        if (it != end && !it->second) // Found a zombie value.
            return end;
        return it;
    }

    template<typename Functor>
    void forEach(Functor functor)
    {
        for (auto iter = m_map.begin(); iter != m_map.end(); ++iter) {
            if (iter->second)
                functor(iter->first, iter->second.get());
        }
    }

    ALWAYS_INLINE const_iterator find(const KeyType& key) const
    {
        return const_cast<WeakGCMap*>(this)->find(key);
    }

    ALWAYS_INLINE bool contains(const KeyType& key) const
    {
        return find(key) != m_map.end();
    }

    virtual void pruneStaleEntries() OVERRIDE;

private:
    HashMapType m_map;
    JSGlobalData& m_vm;
};

} // namespace JSC

#endif // WeakGCMap_h
