/*
 * Copyright (C) 2005, 2006, 2007, 2008, 2011 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef WTF_HashSet_h
#define WTF_HashSet_h

#include <wtf/FastAllocBase.h>
#include <wtf/HashTable.h>

namespace WTF {

struct IdentityExtractor;

template<typename Value, typename HashFunctions, typename Traits, typename MappedTraitsArg> class HashSet;
template<typename Value, typename HashFunctions, typename Traits, typename MappedTraitsArg>
void deleteAllValues(const HashSet<Value, HashFunctions, Traits, MappedTraitsArg>&);
template<typename Value, typename HashFunctions, typename Traits, typename MappedTraitsArg>
void fastDeleteAllValues(const HashSet<Value, HashFunctions, Traits, MappedTraitsArg>&);

template<typename ValueArg, typename HashArg = typename DefaultHash<ValueArg>::Hash,
	typename TraitsArg = HashTraits<ValueArg>, typename MappedTraitsArg = HashTraits<ValueArg>> class HashSet {
	WTF_MAKE_FAST_ALLOCATED;
private:
	typedef HashArg HashFunctions;
	typedef TraitsArg ValueTraits;
    typedef MappedTraitsArg MappedTraits;

public:
	typedef typename ValueTraits::TraitType ValueType;

private:
    typedef typename MappedTraits::PassInType MappedPassInType;
    typedef typename MappedTraits::PassOutType MappedPassOutType;
    typedef typename MappedTraits::PeekType MappedPeekType;

	typedef HashTable<ValueType, ValueType, IdentityExtractor,
		HashFunctions, ValueTraits, ValueTraits> HashTableType;

public:
	typedef HashTableIteratorAdapter<HashTableType, ValueType> iterator;
	typedef HashTableConstIteratorAdapter<HashTableType, ValueType> const_iterator;
	typedef typename HashTableType::AddResult AddResult;

	void swap(HashSet&);

	int size() const;
	int capacity() const;
	bool isEmpty() const;

	iterator begin();
	iterator end();
	const_iterator begin() const;
	const_iterator end() const;

	iterator find(const ValueType&);
	const_iterator find(const ValueType&) const;
	bool contains(const ValueType&) const;

	// An alternate version of find() that finds the object by hashing and comparing
	// with some other type, to avoid the cost of type conversion. HashTranslator
	// must have the following function members:
	//   static unsigned hash(const T&);
	//   static bool equal(const ValueType&, const T&);
	// FIXME: We should reverse the order of the template arguments so that callers
	// can just pass the translator and let the compiler deduce T.
	template<typename T, typename HashTranslator> iterator find(const T&);
	template<typename T, typename HashTranslator> const_iterator find(const T&) const;
	template<typename T, typename HashTranslator> bool contains(const T&) const;

	// The return value is a pair of an interator to the new value's location,
	// and a bool that is true if an new entry was added.
	AddResult add(const ValueType&);
    AddResult add(ValueType&&);

	// An alternate version of add() that finds the object by hashing and comparing
	// with some other type, to avoid the cost of type conversion if the object is already
	// in the table. HashTranslator must have the following function members:
	//   static unsigned hash(const T&);
	//   static bool equal(const ValueType&, const T&);
	//   static translate(ValueType&, const T&, unsigned hashCode);
	// FIXME: We should reverse the order of the template arguments so that callers
	// can just pass the translator and let the compiler deduce T.
	template<typename T, typename HashTranslator> AddResult add(const T&);

	// Attempts to add a list of things to the set. Returns true if any of
	// them are new to the set. Returns false if the set is unchanged.
	template<typename IteratorType>
	bool add(IteratorType begin, IteratorType end);

	bool remove(const ValueType&);
	bool remove(iterator);
	void clear();

    static bool isValidValue(const ValueType&);

	MappedPassOutType take(const ValueType&);
	MappedPassOutType take(iterator);
	MappedPassOutType takeAny();

	bool operator==(const HashSet&) const;

private:
	friend void deleteAllValues<>(const HashSet&);
	friend void fastDeleteAllValues<>(const HashSet&);

	HashTableType m_impl;
};

struct IdentityExtractor {
	template<typename T> static ALWAYS_INLINE const T& extract(const T& t) { return t; }
};

template<typename Translator>
struct HashSetTranslatorAdapter {
	template<typename T> static ALWAYS_INLINE unsigned hash(const T& key) { return Translator::hash(key); }
	template<typename T, typename U> static ALWAYS_INLINE bool equal(const T& a, const U& b) { return Translator::equal(a, b); }
	template<typename T, typename U> static ALWAYS_INLINE void translate(T& location, const U& key, const U&, unsigned hashCode)
	{
		Translator::translate(location, key, hashCode);
	}
};

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE void HashSet<T, U, V, W>::swap(HashSet& other)
{
	m_impl.swap(other.m_impl);
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE int HashSet<T, U, V, W>::size() const
{
	return m_impl.size();
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE int HashSet<T, U, V, W>::capacity() const
{
	return m_impl.capacity();
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE bool HashSet<T, U, V, W>::isEmpty() const
{
	return m_impl.isEmpty();
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::iterator HashSet<T, U, V, W>::begin()
{
	return m_impl.begin();
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::iterator HashSet<T, U, V, W>::end()
{
	return m_impl.end();
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::const_iterator HashSet<T, U, V, W>::begin() const
{
	return m_impl.begin();
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::const_iterator HashSet<T, U, V, W>::end() const
{
	return m_impl.end();
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::iterator HashSet<T, U, V, W>::find(const ValueType& value)
{
	return m_impl.find(value);
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::const_iterator HashSet<T, U, V, W>::find(const ValueType& value) const
{
	return m_impl.find(value);
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE bool HashSet<T, U, V, W>::contains(const ValueType& value) const
{
	return m_impl.contains(value);
}

template<typename Value, typename HashFunctions, typename Traits, typename MappedTraitsArg>
template<typename T, typename HashTranslator>
typename HashSet<Value, HashFunctions, Traits, MappedTraitsArg>::iterator
ALWAYS_INLINE HashSet<Value, HashFunctions, Traits, MappedTraitsArg>::find(const T& value)
{
	return m_impl.template find<HashSetTranslatorAdapter<HashTranslator> >(value);
}

template<typename Value, typename HashFunctions, typename Traits, typename MappedTraitsArg>
template<typename T, typename HashTranslator>
typename HashSet<Value, HashFunctions, Traits, MappedTraitsArg>::const_iterator
ALWAYS_INLINE HashSet<Value, HashFunctions, Traits, MappedTraitsArg>::find(const T& value) const
{
	return m_impl.template find<HashSetTranslatorAdapter<HashTranslator> >(value);
}

template<typename Value, typename HashFunctions, typename Traits, typename MappedTraitsArg>
template<typename T, typename HashTranslator>
ALWAYS_INLINE bool HashSet<Value, HashFunctions, Traits, MappedTraitsArg>::contains(const T& value) const
{
	return m_impl.template contains<HashSetTranslatorAdapter<HashTranslator> >(value);
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::AddResult HashSet<T, U, V, W>::add(const ValueType& value)
{
	return m_impl.add(value);
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE typename HashSet<T, U, V, W>::AddResult HashSet<T, U, V, W>::add(ValueType&& value)
{
    return m_impl.add(WTFMove(value));
}

template<typename Value, typename HashFunctions, typename Traits, typename MappedTraitsArg>
template<typename T, typename HashTranslator>
ALWAYS_INLINE typename HashSet<Value, HashFunctions, Traits, MappedTraitsArg>::AddResult
HashSet<Value, HashFunctions, Traits, MappedTraitsArg>::add(const T& value)
{
	return m_impl.template addPassingHashCode<HashSetTranslatorAdapter<HashTranslator> >(value, value);
}

template<typename T, typename U, typename V, typename W>
template<typename IteratorType>
inline bool HashSet<T, U, V, W>::add(IteratorType begin, IteratorType end)
{
	bool changed = false;
	for (IteratorType iter = begin; iter != end; ++iter)
		changed |= add(*iter).isNewEntry;
	return changed;
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE bool HashSet<T, U, V, W>::remove(iterator it)
{
	if (it.m_impl == m_impl.end())
		return false;
	m_impl.internalCheckTableConsistency();
	m_impl.removeWithoutEntryConsistencyCheck(it.m_impl);
	return true;
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE bool HashSet<T, U, V, W>::remove(const ValueType& value)
{
	return remove(find(value));
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE void HashSet<T, U, V, W>::clear()
{
	m_impl.clear();
}

template<typename T, typename U, typename V, typename W>
inline bool HashSet<T, U, V, W>::isValidValue(const ValueType& value)
{
    if (ValueTraits::isDeletedValue(value))
        return false;

    if (HashFunctions::safeToCompareToEmptyOrDeleted) {
        if (value == ValueTraits::emptyValue())
            return false;
    } else {
        if (isHashTraitsEmptyValue<ValueTraits>(value))
            return false;
    }

    return true;
}

template<typename T, typename U, typename V, typename W>
inline auto HashSet<T, U, V, W>::take(iterator it) -> MappedPassOutType
{
	if (it == end())
		return MappedTraits::passOut(ValueTraits::emptyValue());

	MappedPassOutType result = MappedTraits::passOut(*it);
	remove(it);
	return result;
}

template<typename T, typename U, typename V, typename W>
inline auto HashSet<T, U, V, W>::take(const ValueType& value) -> MappedPassOutType
{
	return take(find(value));
}

template<typename T, typename U, typename V, typename W>
inline auto HashSet<T, U, V, W>::takeAny() -> MappedPassOutType
{
	return take(begin());
}

template<typename ValueType, typename HashTableType>
void deleteAllValues(HashTableType& collection)
{
	typedef typename HashTableType::const_iterator iterator;
	iterator end = collection.end();
	for (iterator it = collection.begin(); it != end; ++it)
		delete *it;
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE void deleteAllValues(const HashSet<T, U, V, W>& collection)
{
	deleteAllValues<typename HashSet<T, U, V, W>::ValueType>(collection.m_impl);
}

template<typename ValueType, typename HashTableType>
void fastDeleteAllValues(HashTableType& collection)
{
	typedef typename HashTableType::const_iterator iterator;
	iterator end = collection.end();
	for (iterator it = collection.begin(); it != end; ++it)
		fastDelete(*it);
}

template<typename T, typename U, typename V, typename W>
ALWAYS_INLINE void fastDeleteAllValues(const HashSet<T, U, V, W>& collection)
{
	fastDeleteAllValues<typename HashSet<T, U, V, W>::ValueType>(collection.m_impl);
}

template<typename T, typename U, typename V, typename W, typename VectorType>
inline void copyToVector(const HashSet<T, U, V, W>& collection, VectorType& vector)
{
	typedef typename HashSet<T, U, V, W>::const_iterator iterator;

	vector.resize(collection.size());

	iterator it = collection.begin();
	iterator end = collection.end();
	for (unsigned i = 0; it != end; ++it, ++i)
		vector[i] = *it;
}

template<typename T, typename U, typename V, typename W>
inline bool HashSet<T, U, V, W>::operator==(const HashSet& other) const
{
	if (size() != other.size())
		return false;
	for (const_iterator iter = begin(); iter != end(); ++iter) {
		if (!other.contains(*iter))
			return false;
	}
	return true;
}

} // namespace WTF

using WTF::HashSet;

#endif /* WTF_HashSet_h */
