/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CDMInstance_h
#define CDMInstance_h

#if ENABLE(ENCRYPTED_MEDIA)

#include "MediaKeyMessageType.h"
#include "MediaKeySessionType.h"
#include "MediaKeyStatus.h"
#include <utility>
#include <wtf/Forward.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/Optional.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>

namespace WebCore {

struct CDMResponseInfo;
class MediaPlayerPrivateInterface;
struct MediaKeySystemConfiguration;
class MediaKeySession;
class SharedBuffer;

class CDMInstance : public RefCounted<CDMInstance> {
public:
    virtual ~CDMInstance() { }

    enum SuccessValue {
    	Failed,
		Succeeded,
    };

    typedef MediaKeySessionType LicenseType;
    typedef MediaKeyStatus KeyStatus;
    typedef MediaKeyMessageType MessageType;

    virtual SuccessValue initializeWithConfiguration(const MediaKeySystemConfiguration&) = 0;
    virtual SuccessValue setDistinctiveIdentifiersAllowed(bool) = 0;
    virtual SuccessValue setPersistentStateAllowed(bool) = 0;
    virtual SuccessValue setServerCertificate(RefPtr<SharedBuffer>&) = 0;
    virtual bool attachToMediaPlayer(MediaPlayerPrivateInterface*) = 0;
    virtual bool isAesDecryptor() const = 0;

    class LicenseCallback
    {
    public:
    	virtual ~LicenseCallback() { }
        virtual void exec(RefPtr<SharedBuffer> message, const String& sessionId, bool needsIndividualization, SuccessValue succeeded) = 0;
    };

    virtual void requestLicense(LicenseType, const AtomicString& initDataType, RefPtr<SharedBuffer>& initData, PassOwnPtr<LicenseCallback>) = 0;

    typedef Vector<std::pair<RefPtr<SharedBuffer>, KeyStatus> > KeyStatusVector;
    typedef std::pair<MessageType, RefPtr<SharedBuffer> > Message;

    class LicenseUpdateCallback
    {
    public:
    	virtual ~LicenseUpdateCallback() { }
        virtual void exec(bool sessionWasClosed, Optional<KeyStatusVector> changedKeys, Optional<double> changedExpiration, Optional<Message> message, SuccessValue succeeded) = 0;
    };

    class LicenseRenewCallback : public RefCounted<LicenseRenewCallback>
    {
    public:
    	virtual ~LicenseRenewCallback() { }
        virtual void exec(const unsigned char* message, unsigned messageLength) = 0;
    };

    virtual void updateLicense(const String& sessionId, LicenseType, const SharedBuffer& response, const CDMResponseInfo&, PassOwnPtr<LicenseUpdateCallback>, PassRefPtr<LicenseRenewCallback>) = 0;

    virtual void keyMessage(const String& keySystem, const String& sessionId, const unsigned char* message, unsigned messageLength) = 0;

    enum SessionLoadFailure {
        None,
        NoSessionData,
        MismatchedSessionType,
        QuotaExceeded,
        Other,
    };

    class LoadSessionCallback
	{
	public:
		virtual ~LoadSessionCallback() { }
		virtual void exec(Optional<KeyStatusVector>, Optional<double>, Optional<Message>, SuccessValue, SessionLoadFailure) = 0;
	};
    virtual void loadSession(LicenseType, const String& sessionId, const String& origin, PassOwnPtr<LoadSessionCallback>) = 0;

	virtual void closedSession(const String& sessionId) = 0;

    class CloseSessionCallback
	{
	public:
		virtual ~CloseSessionCallback() { }
		virtual void exec() = 0;
	};
	virtual void closeSession(const String& sessionId, PassOwnPtr<CloseSessionCallback>) = 0;

    class RemoveSessionDataCallback
	{
	public:
		virtual ~RemoveSessionDataCallback() { }
		virtual void exec(const KeyStatusVector&, Optional<RefPtr<SharedBuffer> >, SuccessValue) = 0;
	};
	virtual void removeSessionData(const String& sessionId, LicenseType, PassOwnPtr<RemoveSessionDataCallback>) = 0;

    virtual void storeRecordOfKeyUsage(const String& sessionId) = 0;
};

} // WebCore

#endif // ENABLE(ENCRYPTED_MEDIA)

#endif // CDMInstance_h
