/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "InitDataRegistry.h"

#if ENABLE(ENCRYPTED_MEDIA)

#include "NotImplemented.h"
#include "SharedBuffer.h"
#include "inspector/InspectorValues.h"
#include <wtf/NeverDestroyed.h>
#include <wtf/text/Base64.h>

namespace WebCore {

static Vector<RefPtr<SharedBuffer> > extractKeyIDsKeyids(const SharedBuffer& buffer)
{
    // 1. Format
    // https://w3c.github.io/encrypted-media/format-registry/initdata/keyids.html#format
    String json(buffer.data(), buffer.size());

    RefPtr<InspectorValue> value = InspectorValue::parseJSON(json);
    if (!value)
        return Vector<RefPtr<SharedBuffer> >();

    RefPtr<InspectorObject> object;
    if (!value->asObject(&object))
        return Vector<RefPtr<SharedBuffer> >();

    RefPtr<InspectorArray> kidsArray = object->getArray("kids");
    if (!kidsArray)
        return Vector<RefPtr<SharedBuffer> >();

    Vector<RefPtr<SharedBuffer> > keyIDs;
    for (InspectorArray::iterator it = kidsArray->begin(); it != kidsArray->end(); ++it) {
    	RefPtr<InspectorValue> value = *it;
        String keyID;
        if (!value->asString(&keyID))
            continue;

        Vector<char> keyIDData;
        if (!WTF::base64URLDecode(keyID, keyIDData))
            continue;

        RefPtr<SharedBuffer> keyIDBuffer = SharedBuffer::adoptVector(keyIDData);
        keyIDs.append(keyIDBuffer);
    }

    return keyIDs;
}

static RefPtr<SharedBuffer> sanitizeKeyids(const SharedBuffer& buffer)
{
    // 1. Format
    // https://w3c.github.io/encrypted-media/format-registry/initdata/keyids.html#format
    Vector<RefPtr<SharedBuffer> > keyIDBuffer = extractKeyIDsKeyids(buffer);
    if (keyIDBuffer.isEmpty())
        return nullptr;

    RefPtr<InspectorObject> object = InspectorObject::create();
    RefPtr<InspectorArray> kidsArray = InspectorArray::create();
    for(int i = 0; i < keyIDBuffer.size(); ++i) {
    	RefPtr<SharedBuffer> buffer = keyIDBuffer[i];
        kidsArray->pushString(WTF::base64URLEncode(buffer->data(), buffer->size()));
    }
    object->setArray("kids", kidsArray);

    CString jsonData = object->toJSONString().utf8();
    return SharedBuffer::create(jsonData.data(), jsonData.length());
}

static RefPtr<SharedBuffer> sanitizeCenc(const SharedBuffer& buffer)
{
    // 4. Common SystemID and PSSH Box Format
    // https://w3c.github.io/encrypted-media/format-registry/initdata/cenc.html#common-system
    notImplemented();
    return buffer.copy();
}

static Vector<RefPtr<SharedBuffer> > extractKeyIDsCenc(const SharedBuffer&)
{
    // 4. Common SystemID and PSSH Box Format
    // https://w3c.github.io/encrypted-media/format-registry/initdata/cenc.html#common-system
    notImplemented();
    return Vector<RefPtr<SharedBuffer> >();
}

static RefPtr<SharedBuffer> sanitizeWebM(const SharedBuffer& buffer)
{
    // 1. Format
    // https://w3c.github.io/encrypted-media/format-registry/initdata/webm.html#format
    notImplemented();
    return buffer.copy();
}

static Vector<RefPtr<SharedBuffer> > extractKeyIDsWebM(const SharedBuffer&)
{
    // 1. Format
    // https://w3c.github.io/encrypted-media/format-registry/initdata/webm.html#format
    notImplemented();
    return Vector<RefPtr<SharedBuffer> >();
}

InitDataRegistry& InitDataRegistry::shared()
{
    static NeverDestroyed<InitDataRegistry> registry;
    return registry.get();
}

InitDataRegistry::InitDataRegistry()
{
    registerInitDataType("keyids", InitDataTypeCallbacks(sanitizeKeyids, extractKeyIDsKeyids));
    registerInitDataType("cenc", InitDataTypeCallbacks(sanitizeCenc, extractKeyIDsCenc));
    registerInitDataType("webm", InitDataTypeCallbacks(sanitizeWebM, extractKeyIDsWebM));
}

InitDataRegistry::~InitDataRegistry()
{
}

RefPtr<SharedBuffer> InitDataRegistry::sanitizeInitData(const AtomicString& initDataType, const SharedBuffer& buffer)
{
	HashMap<AtomicString, InitDataTypeCallbacks>::iterator iter = m_types.find(initDataType);
    if (iter == m_types.end() || !iter->second.sanitizeInitData)
        return nullptr;
    return iter->second.sanitizeInitData(buffer);
}

Vector<RefPtr<SharedBuffer> > InitDataRegistry::extractKeyIDs(const AtomicString& initDataType, const SharedBuffer& buffer)
{
	HashMap<AtomicString, InitDataTypeCallbacks>::iterator iter = m_types.find(initDataType);
    if (iter == m_types.end() || !iter->second.sanitizeInitData)
        return Vector<RefPtr<SharedBuffer> >();
    return iter->second.extractKeyIDs(buffer);
}

void InitDataRegistry::registerInitDataType(const AtomicString& initDataType, const InitDataTypeCallbacks& callbacks)
{
    ASSERT(!m_types.contains(initDataType));
    m_types.set(initDataType, callbacks);
}

}

#endif // ENABLE(ENCRYPTED_MEDIA)
