/*
 * Copyright (C) 2016 Metrological Group B.V.
 * Copyright (C) 2016 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MediaKeySession_h
#define MediaKeySession_h

#if ENABLE(ENCRYPTED_MEDIA)

#include "ActiveDOMObject.h"
#include "CDMInstance.h"
#include "EventTarget.h"
#include "GenericEventQueue.h"
#include "GenericTaskQueue.h"
#include "JSDOMPromise.h"
#include "MediaKeyMessageType.h"
#include "MediaKeySessionType.h"
#include "MediaKeyStatus.h"
#include "Timer.h"
#include <wtf/Deque.h>
#include <wtf/Optional.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>
#include <wtf/WeakPtr.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class BufferSource;
class CDM;
class MediaKeyStatusMap;
class MediaKeys;
class SharedBuffer;

class MediaKeySession FINAL : public RefCounted<MediaKeySession>, public EventTarget, public ActiveDOMObject {
public:
    static PassRefPtr<MediaKeySession> create(ScriptExecutionContext*, MediaKeySessionType, bool useDistinctiveIdentifier, RefPtr<CDM>, RefPtr<CDMInstance>);
    virtual ~MediaKeySession();

    using RefCounted<MediaKeySession>::ref;
    using RefCounted<MediaKeySession>::deref;

    DEFINE_ATTRIBUTE_EVENT_LISTENER(message);
    DEFINE_ATTRIBUTE_EVENT_LISTENER(keystatuseschange);

    const String& sessionId() const;
    double expiration() const;
    RefPtr<MediaKeyStatusMap> keyStatuses() const;

    typedef DOMPromise<std::nullptr_t, ExceptionCode> PromiseVoid;
    typedef DOMPromise<bool, ExceptionCode> PromiseBool;

    void generateRequest(const AtomicString&, const BufferSource&, PromiseVoid);
    void load(const String&, PromiseBool);
    void update(const BufferSource&, PromiseVoid);
    void close(PromiseVoid);
    void remove(PromiseVoid);

    typedef DOMPromise<std::nullptr_t, ExceptionCode> ClosedPromise;
    void registerClosedPromise(ClosedPromise);

    const Vector<std::pair<RefPtr<SharedBuffer>, MediaKeyStatus> >& statuses() const { return m_statuses; }

    // ActiveDOMObject
    virtual bool hasPendingActivity() const OVERRIDE;
    virtual const char* activeDOMObjectName() const OVERRIDE;
    virtual bool canSuspend() OVERRIDE const;
    virtual bool canSuspendForDocumentSuspension() const OVERRIDE;
    virtual void stop() OVERRIDE;

    // EventTarget
    virtual ScriptExecutionContext* scriptExecutionContext() const OVERRIDE { return ActiveDOMObject::scriptExecutionContext(); }

private:
    MediaKeySession(ScriptExecutionContext*, MediaKeySessionType, bool useDistinctiveIdentifier, RefPtr<CDM>, RefPtr<CDMInstance>);
    void enqueueMessage(MediaKeyMessageType, const unsigned char* message, unsigned messageLength);
    void enqueueMessage(MediaKeyMessageType, const SharedBuffer&);
    void updateKeyStatuses(const CDMInstance::KeyStatusVector&);
    void updateExpiration(double);
    void sessionClosed();

    // EventTarget
    virtual EventTargetInterface eventTargetInterface() const;
    void refEventTarget() OVERRIDE { ref(); }
    void derefEventTarget() OVERRIDE { deref(); }

    virtual EventTargetData* eventTargetData() OVERRIDE;
    virtual EventTargetData* ensureEventTargetData() OVERRIDE;

    WeakPtr<MediaKeySession> createWeakPtr() { return m_weakPtrFactory.createWeakPtr(this); }

    class GenerateRequestTask : public GenericTask<MediaKeySession>
	{
	public:
    	GenerateRequestTask(MediaKeySession* mediaKeySession, RefPtr<SharedBuffer> initData, const AtomicString& initDataType, PromiseVoid promise);
    	virtual ~GenerateRequestTask();
    	virtual void exec();

	private:
        RefPtr<SharedBuffer> initData;
    	AtomicString initDataType;
        PromiseVoid promise;
	};

    class LicenseCallbackImpl : public CDMInstance::LicenseCallback
    {
    public:
    	LicenseCallbackImpl(WeakPtr<MediaKeySession> weakThis, PromiseVoid promise);
    	virtual ~LicenseCallbackImpl();
        virtual void exec(RefPtr<SharedBuffer> message, const String& sessionId, bool needsIndividualization, CDMInstance::SuccessValue succeeded);

	private:
        WeakPtr<MediaKeySession> weakThis;
        PromiseVoid promise;
    };

    class LicenseFinishTask : public GenericTask<MediaKeySession>
    {
    public:
    	LicenseFinishTask(WeakPtr<MediaKeySession> weakThis, PromiseVoid promise, RefPtr<SharedBuffer> message, MediaKeyMessageType messageType, String sessionId, CDMInstance::SuccessValue succeeded);
    	virtual ~LicenseFinishTask();
        virtual void exec();

	private:
        PromiseVoid promise;
        RefPtr<SharedBuffer> message;
    	MediaKeyMessageType messageType;
        String sessionId;
        CDMInstance::SuccessValue succeeded;
    };

    class UpdateTask : public GenericTask<MediaKeySession>
	{
	public:
    	UpdateTask(MediaKeySession* mediaKeySession, RefPtr<SharedBuffer> response, PromiseVoid promise);
    	virtual ~UpdateTask();
    	virtual void exec();

	private:
        RefPtr<SharedBuffer> response;
        PromiseVoid promise;
	};

    class LicenseUpdateCallbackImpl : public CDMInstance::LicenseUpdateCallback
    {
    public:
    	LicenseUpdateCallbackImpl(WeakPtr<MediaKeySession> weakThis, PromiseVoid promise);
    	virtual ~LicenseUpdateCallbackImpl();
        virtual void exec(bool sessionWasClosed, Optional<CDMInstance::KeyStatusVector> changedKeys, Optional<double> changedExpiration, Optional<CDMInstance::Message> message, CDMInstance::SuccessValue succeeded);

    private:
        WeakPtr<MediaKeySession> weakThis;
        PromiseVoid promise;
    };

    class LicenseRenewCallbackImpl : public CDMInstance::LicenseRenewCallback
    {
    public:
    	LicenseRenewCallbackImpl(WeakPtr<MediaKeySession> weakThis);
    	virtual ~LicenseRenewCallbackImpl();
        virtual void exec(const unsigned char* message, unsigned messageLength);

    private:
        WeakPtr<MediaKeySession> weakThis;
    };

    class LicenseUpdateFinishTask : public GenericTask<MediaKeySession>
    {
    public:
    	LicenseUpdateFinishTask(WeakPtr<MediaKeySession> weakThis, bool sessionWasClosed, Optional<CDMInstance::KeyStatusVector> changedKeys, Optional<double> changedExpiration, Optional<CDMInstance::Message> message, PromiseVoid promise);
    	virtual ~LicenseUpdateFinishTask();
        virtual void exec();

	private:
        bool sessionWasClosed;
        Optional<CDMInstance::KeyStatusVector> changedKeys;
        Optional<double> changedExpiration;
        Optional<CDMInstance::Message> message;
        PromiseVoid promise;
    };

    class CloseTask : public GenericTask<MediaKeySession>
	{
	public:
    	CloseTask(MediaKeySession* mediaKeySession, PromiseVoid promise);
		virtual ~CloseTask();
		virtual void exec();

	private:
		PromiseVoid promise;
	};

	class CloseSessionCallbackImpl : public CDMInstance::CloseSessionCallback
	{
	public:
		CloseSessionCallbackImpl(WeakPtr<MediaKeySession> weakThis, PromiseVoid promise);
		virtual ~CloseSessionCallbackImpl();
		virtual void exec();

	private:
		WeakPtr<MediaKeySession> weakThis;
		PromiseVoid promise;
	};

	class CloseFinishTask : public GenericTask<MediaKeySession>
	{
	public:
		CloseFinishTask(WeakPtr<MediaKeySession> weakThis, PromiseVoid promise);
		virtual ~CloseFinishTask();
		virtual void exec();

	private:
		PromiseVoid promise;
	};

    class RemoveTask : public GenericTask<MediaKeySession>
	{
	public:
    	RemoveTask(MediaKeySession* mediaKeySession, PromiseVoid promise);
		virtual ~RemoveTask();
		virtual void exec();

	private:
		PromiseVoid promise;
	};

	class RemoveSessionDataCallbackImpl : public CDMInstance::RemoveSessionDataCallback
	{
	public:
		RemoveSessionDataCallbackImpl(WeakPtr<MediaKeySession> weakThis, PromiseVoid promise);
		virtual ~RemoveSessionDataCallbackImpl();
		virtual void exec(const CDMInstance::KeyStatusVector&, Optional<RefPtr<SharedBuffer> >, CDMInstance::SuccessValue);

	private:
		WeakPtr<MediaKeySession> weakThis;
		PromiseVoid promise;
	};

	class RemoveFinishTask : public GenericTask<MediaKeySession>
	{
	public:
		RemoveFinishTask(WeakPtr<MediaKeySession> weakThis, const CDMInstance::KeyStatusVector& keys, Optional<RefPtr<SharedBuffer> > message, CDMInstance::SuccessValue succeeded, PromiseVoid promise);
		virtual ~RemoveFinishTask();
		virtual void exec();

	private:
		CDMInstance::KeyStatusVector keys;
		Optional<RefPtr<SharedBuffer> > message;
		CDMInstance::SuccessValue succeeded;
		PromiseVoid promise;
	};

    class LoadTask : public GenericTask<MediaKeySession>
	{
	public:
    	LoadTask(MediaKeySession* mediaKeySession, const String& sessionId, PromiseBool promise);
		virtual ~LoadTask();
		virtual void exec();

	private:
		String sessionId;
		PromiseBool promise;
	};

	class LoadSessionCallbackImpl : public CDMInstance::LoadSessionCallback
	{
	public:
		LoadSessionCallbackImpl(WeakPtr<MediaKeySession> weakThis, PromiseBool promise, const String& sanitizedSessionId);
		virtual ~LoadSessionCallbackImpl();
		virtual void exec(Optional<CDMInstance::KeyStatusVector>, Optional<double>, Optional<CDMInstance::Message>, CDMInstance::SuccessValue, CDMInstance::SessionLoadFailure);

	private:
		WeakPtr<MediaKeySession> weakThis;
		String sanitizedSessionId;
		PromiseBool promise;
	};

	class LoadFinishTask : public GenericTask<MediaKeySession>
	{
	public:
		LoadFinishTask(WeakPtr<MediaKeySession> weakThis, Optional<CDMInstance::KeyStatusVector> knownKeys, Optional<double> expiration, Optional<CDMInstance::Message> message, const String& sanitizedSessionId, CDMInstance::SuccessValue succeeded, PromiseBool promise);
		virtual ~LoadFinishTask();
		virtual void exec();

	private:
		Optional<CDMInstance::KeyStatusVector> knownKeys;
		Optional<double> expiration;
		Optional<CDMInstance::Message> message;
		String sanitizedSessionId;
		CDMInstance::SuccessValue succeeded;
		PromiseBool promise;
	};

    EventTargetData m_eventTargetData;

    String m_sessionId;
    double m_expiration;
    Optional<ClosedPromise> m_closedPromise;
    RefPtr<MediaKeyStatusMap> m_keyStatuses;
    bool m_closed;
    bool m_uninitialized;
    bool m_callable;
    bool m_useDistinctiveIdentifier;
    MediaKeySessionType m_sessionType;
    RefPtr<CDM> m_implementation;
    RefPtr<CDMInstance> m_instance;
    GenericEventQueue m_eventQueue;
    GenericTaskQueue<MediaKeySession> m_taskQueue;
    Vector<RefPtr<SharedBuffer> > m_recordOfKeyUsage;
    double m_firstDecryptTime;
    double m_latestDecryptTime;
    Vector<std::pair<RefPtr<SharedBuffer>, MediaKeyStatus> > m_statuses;
    WeakPtrFactory<MediaKeySession> m_weakPtrFactory;
};

} // namespace WebCore

#endif // ENABLE(ENCRYPTED_MEDIA)
#endif // MediaKeySession_h
