/*
 * Copyright (C) 2016 Metrological Group B.V.
 * Copyright (C) 2016 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "MediaKeyStatusMap.h"

#if ENABLE(ENCRYPTED_MEDIA)

#include <string.h>
#include "JSMediaKeyStatusMap.h"
#include "MediaKeySession.h"
#include "SharedBuffer.h"

namespace WebCore {

MediaKeyStatusMap::MediaKeyStatusMap(const MediaKeySession& session)
    : m_session(&session)
{
}

MediaKeyStatusMap::~MediaKeyStatusMap()
{
}

void MediaKeyStatusMap::detachSession()
{
    m_session = 0;
}

unsigned long MediaKeyStatusMap::size()
{
    if (!m_session)
        return 0;
    return m_session->statuses().size();
}

static bool keyIdsMatch(RefPtr<SharedBuffer> a, const BufferSource& b)
{
    unsigned length = a->size();
    if (!length || length != b.length())
        return false;
    return !memcmp(a->data(), b.data(), length);
}

bool MediaKeyStatusMap::has(const BufferSource& keyId)
{
    if (!m_session)
        return false;

    const Vector<std::pair<RefPtr<SharedBuffer>, MediaKeyStatus> >& statuses = m_session->statuses();
    for(int i = 0; i < statuses.size(); ++i) {
    	if(keyIdsMatch(statuses[i].first, keyId))
    		return true;
    }

    return false;
}

ScriptValue MediaKeyStatusMap::get(JSC::ExecState* exec, const BufferSource& keyId)
{
    if (!m_session)
        return ScriptValue(exec->vm(), JSC::jsUndefined());

    const Vector<std::pair<RefPtr<SharedBuffer>, MediaKeyStatus> >& statuses = m_session->statuses();
    for(int i = 0; i < statuses.size(); ++i) {
		if(keyIdsMatch(statuses[i].first, keyId))
			return ScriptValue(exec->vm(), jsStringWithCache(exec, stringValue(statuses[i].second)));
	}

    return ScriptValue(exec->vm(), JSC::jsUndefined());
}

MediaKeyStatusMap::Iterator::Iterator(MediaKeyStatusMap& map)
: m_map(&map)
, m_index(0)
{
}

Optional<WTF::KeyValuePair<BufferSource, MediaKeyStatus>> MediaKeyStatusMap::Iterator::next()
{
    if (m_map && m_map->m_session && m_index < m_map->m_session->statuses().size()) {
		const std::pair<RefPtr<SharedBuffer>, MediaKeyStatus>& pair = m_map->m_session->statuses().at(m_index++);
		return WTF::KeyValuePair<BufferSource, MediaKeyStatus>(BufferSource(ArrayBuffer::create(reinterpret_cast<const uint8_t*>(pair.first->data()), pair.first->size())), pair.second);
	}

    m_map.clear();
    return Nullopt;
}

} // namespace WebCore

#endif // ENABLE(ENCRYPTED_MEDIA)
