/*
 * Copyright (C) 2016 Metrological Group B.V.
 * Copyright (C) 2016 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "MediaKeys.h"

#if ENABLE(ENCRYPTED_MEDIA)

#include "CDM.h"
#include "CDMInstance.h"
#include "JSMediaKeySessionType.h"
#include "MediaKeySession.h"
#include "SharedBuffer.h"

namespace WebCore {

PassRefPtr<MediaKeys> MediaKeys::create(bool useDistinctiveIdentifier, bool persistentStateAllowed, const Vector<MediaKeySessionType>& supportedSessionTypes, RefPtr<CDM> implementation, RefPtr<CDMInstance> instance)
{
    return adoptRef(new MediaKeys(useDistinctiveIdentifier, persistentStateAllowed, supportedSessionTypes, implementation, instance));
}

MediaKeys::MediaKeys(bool useDistinctiveIdentifier, bool persistentStateAllowed, const Vector<MediaKeySessionType>& supportedSessionTypes, RefPtr<CDM> implementation, RefPtr<CDMInstance> instance)
: m_useDistinctiveIdentifier(useDistinctiveIdentifier)
, m_persistentStateAllowed(persistentStateAllowed)
, m_supportedSessionTypes(supportedSessionTypes)
, m_implementation(implementation)
, m_instance(instance)
{
}

MediaKeys::~MediaKeys()
{
}

RefPtr<MediaKeySession> MediaKeys::createSession(ScriptExecutionContext* context, const String& sessionType, ExceptionCode& ec)
{
	MediaKeySessionType mediaKeySessionType = enumerationValueMediaKeySessionType(sessionType).valueOr(MediaKeySessionType::Temporary);

    // https://w3c.github.io/encrypted-media/#dom-mediakeys-setservercertificate
    // W3C Editor's Draft 09 November 2016

    // When this method is invoked, the user agent must run the following steps:
    // 1. If this object's supported session types value does not contain sessionType, throw [WebIDL] a NotSupportedError.
    if (!m_supportedSessionTypes.contains(mediaKeySessionType)) {
        ec = NOT_SUPPORTED_ERR;
        return nullptr;
    }

    // 2. If the implementation does not support MediaKeySession operations in the current state, throw [WebIDL] an InvalidStateError.
    if (!m_implementation->supportsSessions()) {
        ec = INVALID_STATE_ERR;
        return nullptr;
    }

    // 3. Let session be a new MediaKeySession object, and initialize it as follows:
    // NOTE: Continued in MediaKeySession.
    // 4. Return session.
    return MediaKeySession::create(context, mediaKeySessionType, m_useDistinctiveIdentifier, m_implementation, m_instance);
}

MediaKeys::Task::Task(MediaKeys* mediaKeys, RefPtr<SharedBuffer> certificate, PromiseBool promise)
: GenericTask<MediaKeys>(mediaKeys->createWeakPtr())
, certificate(certificate)
, promise(promise)
{
}

MediaKeys::Task::~Task()
{
}

void MediaKeys::Task::exec()
{
	if(!weakThis) {
        promise.reject(INVALID_STATE_ERR);
		return;
	}

    // 5.1. Use this object's cdm instance to process certificate.
    if (weakThis->m_instance->setServerCertificate(certificate) == CDMInstance::Failed) {
        // 5.2. If the preceding step failed, resolve promise with a new DOMException whose name is the appropriate error name.
        promise.reject(INVALID_STATE_ERR);
        return;
    }

    // 5.1. Resolve promise with true.
    promise.resolve(true);
}

void MediaKeys::setServerCertificate(const BufferSource& serverCertificate, PromiseBool promise)
{
    // https://w3c.github.io/encrypted-media/#dom-mediakeys-setservercertificate
    // W3C Editor's Draft 09 November 2016

    // When this method is invoked, the user agent must run the following steps:
    // 1. If the Key System implementation represented by this object's cdm implementation value does not support
    //    server certificates, return a promise resolved with false.
    if (!m_implementation->supportsServerCertificates()) {
        promise.resolve(false);
        return;
    }

    // 2. If serverCertificate is an empty array, return a promise rejected with a new a newly created TypeError.
    if (!serverCertificate.length()) {
        promise.reject(TypeError);
        return;
    }

    // 3. Let certificate be a copy of the contents of the serverCertificate parameter.
    RefPtr<SharedBuffer> certificate = SharedBuffer::create(serverCertificate.data(), serverCertificate.length());

    // 4. Let promise be a new promise.
    // 5. Run the following steps in parallel:
	m_taskQueue.enqueueTask(adoptPtr(new Task(this, certificate, promise)));

    // 6. Return promise.
}

} // namespace WebCore

#endif // ENABLE(ENCRYPTED_MEDIA)
