/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef LegacyCDM_h
#define LegacyCDM_h

#if ENABLE(LEGACY_ENCRYPTED_MEDIA_V2)

#include <runtime/Uint8Array.h>
#include <wtf/Forward.h>
#include <wtf/PassRefPtr.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class LegacyCDM;
class LegacyCDMPrivateInterface;
class LegacyCDMSession;
class MediaPlayer;

typedef PassOwnPtr<LegacyCDMPrivateInterface> (*CreateCDM)(LegacyCDM*);
typedef bool (*CDMSupportsKeySystem)(const String&);
typedef bool (*CDMSupportsKeySystemAndMimeType)(const String&, const String&);

class LegacyCDMClient {
public:
    virtual ~LegacyCDMClient() { }

    virtual MediaPlayer* cdmMediaPlayer(const LegacyCDM*) const = 0;
};

class LegacyCDMSession {
public:
	LegacyCDMSession() { }
    virtual ~LegacyCDMSession() { }

    virtual const String& sessionId() const = 0;
    virtual Vector<unsigned char> generateKeyRequest(const String& mimeType, Uint8Array* initData, Uint8Array* cdmData, String& destinationURL, unsigned short& errorCode, unsigned long& systemCode) = 0;
    virtual void releaseKeys() = 0;
    virtual bool update(Uint8Array*, RefPtr<Uint8Array>& nextMessage, unsigned short& errorCode, unsigned long& systemCode) = 0;
};

class LegacyCDM {
public:

	enum CDMErrorCode { UnknownError = 1, ClientError, ServiceError, OutputError, HardwareChangeError, DomainError };
    static bool supportsKeySystem(const String&);
    static bool keySystemSupportsMimeType(const String& keySystem, const String& mimeType);
    static PassOwnPtr<LegacyCDM> create(const String& keySystem);

    static void registerCDMFactory(CreateCDM, CDMSupportsKeySystem);
    static void registerCDMFactory(CreateCDM, CDMSupportsKeySystem, CDMSupportsKeySystemAndMimeType);

    ~LegacyCDM();

    bool supportsMIMEType(const String&) const;
    PassOwnPtr<LegacyCDMSession> createSession();

    const String& keySystem() const { return m_keySystem; }

    LegacyCDMClient* client() const { return m_client; }
    void setClient(LegacyCDMClient* client) { m_client = client; }

    MediaPlayer* mediaPlayer() const;

private:
    LegacyCDM(const String& keySystem);

    String m_keySystem;
    OwnPtr<LegacyCDMPrivateInterface> m_private;
    LegacyCDMClient* m_client;
};

inline MediaPlayer* LegacyCDM::mediaPlayer() const
{
    if (!m_client)
        return 0;
    return m_client->cdmMediaPlayer(this);
}

}

#endif // ENABLE(LEGACY_ENCRYPTED_MEDIA_V2)

#endif // LegacyCDM_h
