/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef WebKitMediaKeySession_h
#define WebKitMediaKeySession_h

#if ENABLE(LEGACY_ENCRYPTED_MEDIA_V2)

#include "ActiveDOMObject.h"
#include "EventTarget.h"
#include "ExceptionCode.h"
#include "GenericEventQueue.h"
#include "Timer.h"
#include <runtime/Uint8Array.h>
#include <wtf/Deque.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class GenericEventQueue;
class WebKitMediaKeyError;
class WebKitMediaKeys;
class LegacyCDMSession;

class WebKitMediaKeySession : public RefCounted<WebKitMediaKeySession>, public EventTarget, public ActiveDOMObject {
public:
    static PassRefPtr<WebKitMediaKeySession> create(ScriptExecutionContext*, WebKitMediaKeys*, const String& keySystem);
    ~WebKitMediaKeySession();

    const String& keySystem() const { return m_keySystem; }
    const String& sessionId() const;

    void setError(WebKitMediaKeyError*);
    WebKitMediaKeyError* error() { return m_error.get(); }

    void setKeys(WebKitMediaKeys* keys) { m_keys = keys; }
    WebKitMediaKeys* keys() const { return m_keys; }

    void generateKeyRequest(const String& mimeType, Uint8Array* initData, Uint8Array* cdmData);
    void update(Uint8Array* key, ExceptionCode&);

    bool isClosed() const { return !m_session; }
    void close();

    using RefCounted<WebKitMediaKeySession>::ref;
    using RefCounted<WebKitMediaKeySession>::deref;

    // ActiveDOMObject
    virtual bool hasPendingActivity() const OVERRIDE { return (m_keys && !isClosed()) || m_asyncEventQueue->hasPendingEvents(); }
    virtual void stop() OVERRIDE { close(); }

    DEFINE_ATTRIBUTE_EVENT_LISTENER(webkitkeyadded);
    DEFINE_ATTRIBUTE_EVENT_LISTENER(webkitkeyerror);
    DEFINE_ATTRIBUTE_EVENT_LISTENER(webkitkeymessage);

    virtual EventTargetInterface eventTargetInterface() const OVERRIDE;
    virtual ScriptExecutionContext* scriptExecutionContext() const OVERRIDE { return ActiveDOMObject::scriptExecutionContext(); }

protected:
    WebKitMediaKeySession(ScriptExecutionContext*, WebKitMediaKeys*, const String& keySystem);
    void keyRequestTimerFired(Timer<WebKitMediaKeySession>*);
    void addKeyTimerFired(Timer<WebKitMediaKeySession>*);

    WebKitMediaKeys* m_keys;
    String m_keySystem;
    String m_sessionId;
    RefPtr<WebKitMediaKeyError> m_error;
    OwnPtr<GenericEventQueue> m_asyncEventQueue;
    OwnPtr<LegacyCDMSession> m_session;

    struct PendingKeyRequest {
        PendingKeyRequest(const String& mimeType, Uint8Array* initData, Uint8Array* cdmData)
        : mimeType(mimeType)
        , initData(initData)
        , cdmData(cdmData)
        { }
        String mimeType;
        RefPtr<Uint8Array> initData;
        RefPtr<Uint8Array> cdmData;
    };
    Deque<PendingKeyRequest> m_pendingKeyRequests;
    Timer<WebKitMediaKeySession> m_keyRequestTimer;

    Deque<RefPtr<Uint8Array> > m_pendingKeys;
    Timer<WebKitMediaKeySession> m_addKeyTimer;

private:
    virtual void refEventTarget() OVERRIDE { ref(); }
    virtual void derefEventTarget() OVERRIDE { deref(); }

    virtual const char* activeDOMObjectName() const OVERRIDE { return "WebKitMediaKeySession"; }

    virtual EventTargetData* eventTargetData() OVERRIDE { return &m_eventTargetData; }
    virtual EventTargetData* ensureEventTargetData() OVERRIDE { return &m_eventTargetData; }

    EventTargetData m_eventTargetData;
};

}

#endif // ENABLE(LEGACY_ENCRYPTED_MEDIA_V2)

#endif // WebKitMediaKeySession_h
