/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FetchBody.h"

#if ENABLE(FETCH_API)

#include "DOMRequestState.h"
#include "Dictionary.h"
#include "FetchBodyOwner.h"
#include "FetchHeaders.h"
#include "FetchResponseSource.h"
#include "FormData.h"
#include "HTTPParsers.h"
#include "JSBlob.h"
#include "JSDOMFormData.h"
#include "JSDOMPromise.h"
#include "JSReadableStream.h"
#include "JSURLSearchParams.h"
#include "ReadableStreamSource.h"
#include "TextResourceDecoder.h"
#include "WebKitBlobBuilder.h"
#include <runtime/ArrayBuffer.h>
#include <runtime/ArrayBufferView.h>

namespace WebCore {

// FIXME: This implementation is not efficient and we should probably use UTF8Encoding().
static Vector<uint8_t> extractBytesFromText(const String& text)
{
    CString data = text.utf8();
    Vector<uint8_t> value(data.length());
    memcpy(value.data(), data.data(), data.length());
    return value;
}

FetchBody::FetchBody(PassRefPtr<Blob> blob)
    : m_blob(blob)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(PassRefPtr<FormData> formData)
    : m_formData(formData)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(const String& text)
    : m_text(text)
	, m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::~FetchBody()
{
}

FetchBody::FetchBody(PassRefPtr<ArrayBuffer> data)
    : m_data(data)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(PassRefPtr<ArrayBufferView> dataView)
    : m_dataView(dataView)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(PassRefPtr<URLSearchParams> url)
    : m_URLSearchParams(url)
	, m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(const FetchBodyConsumer& consumer)
    : m_consumer(consumer)
{
}

PassOwnPtr<FetchBody> FetchBody::extract(ScriptExecutionContext& context, JSC::ExecState& state, JSC::JSValue value, String& contentType)
{
    if (value.inherits(JSBlob::info())) {
    	RefPtr<Blob> blob = JSBlob::toWrapped(value);
    	contentType = blob->type();
        return adoptPtr(new FetchBody(blob.release()));
    }
    if (value.inherits(JSDOMFormData::info())) {
        ASSERT(!context.isWorkerGlobalScope());
        RefPtr<DOMFormData> domFormData = JSDOMFormData::toWrapped(value);
        RefPtr<FormData> formData = FormData::createMultiPart(*domFormData, domFormData->encoding(), &static_cast<Document&>(context));
        contentType = makeString("multipart/form-data;boundary=", formData->boundary().data());
        return adoptPtr(new FetchBody(formData.release()));
    }
    if (value.isString()) {
        contentType = ASCIILiteral("text/plain;charset=UTF-8");
        return adoptPtr(new FetchBody(value.toWTFString(&state)));
    }
    if (value.inherits(JSURLSearchParams::info())){
        contentType = ASCIILiteral("application/x-www-form-urlencoded;charset=UTF-8");
        return adoptPtr(new FetchBody(JSURLSearchParams::toWrapped(value)));
    }
    if (value.inherits(JSReadableStream::info()))
    	return adoptPtr(new FetchBody());
    if (value.inherits(JSC::JSArrayBuffer::info())) {
    	RefPtr<ArrayBuffer> data = toArrayBuffer(value);
        ASSERT(data);
        return adoptPtr(new FetchBody(data.releaseNonNull()));
    }
    if (value.inherits(JSC::JSArrayBufferView::info())) {
        RefPtr<JSC::ArrayBufferView> data = toArrayBufferView(value);
        ASSERT(data);
        return adoptPtr(new FetchBody(data.releaseNonNull()));
    }
    return PassOwnPtr<FetchBody>();
}

void FetchBody::arrayBuffer(FetchBodyOwner& owner, RefPtr<DeferredPromise> promise)
{
    m_consumer.setType(FetchBodyConsumer::Type::ArrayBuffer);
    consume(owner, promise);
}

void FetchBody::blob(FetchBodyOwner& owner, RefPtr<DeferredPromise> promise, const String& contentType)
{
    m_consumer.setType(FetchBodyConsumer::Type::Blob);
    m_consumer.setContentType(Blob::normalizedContentType(extractMIMETypeFromMediaType(contentType)));
    consume(owner, promise);
}

void FetchBody::json(FetchBodyOwner& owner, RefPtr<DeferredPromise> promise)
{
    if (isText()) {
        fulfillPromiseWithJSON(promise, m_text);
        return;
    }
    m_consumer.setType(FetchBodyConsumer::Type::JSON);
    consume(owner, promise);
}

void FetchBody::text(FetchBodyOwner& owner, RefPtr<DeferredPromise> promise)
{
    if (isText()) {
        promise->resolve(m_text);
        return;
    }
    m_consumer.setType(FetchBodyConsumer::Type::Text);
    consume(owner, promise);
}

void FetchBody::formData(FetchBodyOwner&, RefPtr<DeferredPromise> promise)
{
	promise->reject<ExceptionCode>(0);
}

void FetchBody::consumeOnceLoadingFinished(FetchBodyConsumer::Type type, RefPtr<DeferredPromise> promise, const String& contentType)
{
    m_consumer.setType(type);
    m_consumePromise = promise;

    if (type == FetchBodyConsumer::Type::Blob)
        m_consumer.setContentType(Blob::normalizedContentType(extractMIMETypeFromMediaType(contentType)));
}

void FetchBody::consume(FetchBodyOwner& owner, RefPtr<DeferredPromise> promise)
{
    if (isArrayBuffer()) {
        consumeArrayBuffer(promise);
        return;
    }
    if (isArrayBufferView()) {
        consumeArrayBufferView(promise);
        return;
    }
    if (isText()) {
        consumeText(promise, textBody());
        return;
    }
    if(isURLSearchParams()) {
        consumeText(promise, urlSearchParamsBody().toString());
        return;
    }
    if (isBlob()) {
        consumeBlob(owner, promise);
        return;
    }
    if (isFormData()) {
        // FIXME: Support consuming FormData.
        promise->reject<ExceptionCode>(0);
        return;
    }
    m_consumer.resolve(promise);
}

#if ENABLE(STREAMS_API)
void FetchBody::consumeAsStream(FetchBodyOwner& owner, FetchResponseSource& source)
{
    bool closeStream = false;
    if (isArrayBuffer()) {
        ASSERT(m_data);
        closeStream = source.enqueue(ArrayBuffer::tryCreate(m_data->data(), m_data->byteLength()));
        m_data = 0;
    } else if (isArrayBufferView()) {
        ASSERT(m_dataView);
        closeStream = source.enqueue(ArrayBuffer::tryCreate(m_dataView->baseAddress(), m_dataView->byteLength()));
        m_dataView = 0;
    } else if (isText()) {
        Vector<uint8_t> data = extractBytesFromText(textBody());
        closeStream = source.enqueue(ArrayBuffer::tryCreate(data.data(), data.size()));
        m_text = String();
    } else if(isURLSearchParams()) {
        Vector<uint8_t> data = extractBytesFromText(urlSearchParamsBody().toString());
        closeStream = source.enqueue(ArrayBuffer::tryCreate(data.data(), data.size()));
        m_URLSearchParams = 0;
    } else if (isBlob()) {
        ASSERT(m_blob);
        owner.loadBlob(*m_blob, 0);
        m_blob = 0;
    } else if (isFormData())
        source.error(ASCIILiteral("not implemented"));
    else if (m_consumer.hasData())
        closeStream = source.enqueue(m_consumer.takeAsArrayBuffer());
    else
        closeStream = true;

    if (closeStream)
        source.close();
}
#endif

void FetchBody::consumeArrayBuffer(RefPtr<DeferredPromise> promise)
{
    ASSERT(m_data);
	m_consumer.resolveWithData(promise, static_cast<const uint8_t*>(m_data->data()), m_data->byteLength());
	m_data = 0;
}

void FetchBody::consumeArrayBufferView(RefPtr<DeferredPromise> promise)
{
    ASSERT(m_dataView);
    m_consumer.resolveWithData(promise, static_cast<const uint8_t*>(m_dataView->baseAddress()), m_dataView->byteLength());
    m_dataView = 0;
}

void FetchBody::consumeText(RefPtr<DeferredPromise> promise, const String& text)
{
    Vector<uint8_t> data = extractBytesFromText(text);
	m_consumer.resolveWithData(promise, reinterpret_cast<const uint8_t*>(data.data()), data.size());
	m_text = String();
}

void FetchBody::consumeBlob(FetchBodyOwner& owner, RefPtr<DeferredPromise> promise)
{
    ASSERT(m_blob);

    m_consumePromise = promise;
    owner.loadBlob(*m_blob, &m_consumer);
    m_blob = 0;
}

void FetchBody::loadingFailed()
{
    if (m_consumePromise) {
        m_consumePromise->reject<ExceptionCode>(0);
        m_consumePromise = 0;
    }
}

void FetchBody::loadingSucceeded()
{
    if (m_consumePromise) {
        m_consumer.resolve(m_consumePromise);
        m_consumePromise = 0;
    }
}

RefPtr<FormData> FetchBody::bodyForInternalRequest(ScriptExecutionContext& context) const
{
    if (isText())
        return FormData::create(UTF8Encoding().encode(m_text, EntitiesForUnencodables));
    if(isURLSearchParams())
        return FormData::create(UTF8Encoding().encode(urlSearchParamsBody().toString(), EntitiesForUnencodables));
    if (isBlob()) {
		ASSERT(m_blob);
        RefPtr<FormData> body = FormData::create();
        body->appendBlob(m_blob->url());
        return body;
    }
    if (isArrayBuffer())
        return FormData::create(m_data->data(), m_data->byteLength());
    if (isArrayBufferView())
        return FormData::create(m_dataView->baseAddress(), m_dataView->byteLength());
    if (isFormData()) {
        ASSERT(m_formData);
        ASSERT(!context.isWorkerGlobalScope());
        auto body = m_formData;
        body->generateFiles(static_cast<Document*>(&context));
        return body;
    }
	ASSERT_NOT_REACHED();
	return nullptr;
}

PassOwnPtr<FetchBody> FetchBody::clone() const
{
    ASSERT(!m_consumePromise);
    OwnPtr<FetchBody> clone = adoptPtr(new FetchBody(m_consumer));

    if (isArrayBuffer())
        clone->m_data = m_data;
    else if (isArrayBufferView())
        clone->m_dataView = m_dataView;
    else if (isBlob())
        clone->m_blob = m_blob;
    else if (isFormData())
        clone->m_formData = m_formData;
    else if (isText())
        clone->m_text = m_text;
    else if(isURLSearchParams())
        clone->m_URLSearchParams = m_URLSearchParams;
    return clone.release();
}

}

#endif // ENABLE(FETCH_API)
