/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef FetchBody_h
#define FetchBody_h

#if ENABLE(FETCH_API)

#include "Blob.h"
#include "FetchBodyConsumer.h"
#include "FetchLoader.h"
#include "FormData.h"
#include "URLSearchParams.h"
#include <wtf/FastAllocBase.h>
#include <wtf/Optional.h>
#include <wtf/PassOwnPtr.h>

namespace JSC {
class ArrayBuffer;
class ArrayBufferView;
class ExecState;
class JSValue;
};

namespace WebCore {

class DOMFormData;
class DeferredPromise;
class FetchBodyOwner;
class FetchHeaders;
class FetchResponseSource;
class ScriptExecutionContext;

class FetchBody {
	WTF_MAKE_FAST_ALLOCATED;
public:
    void arrayBuffer(FetchBodyOwner&, RefPtr<DeferredPromise>);
    void blob(FetchBodyOwner&, RefPtr<DeferredPromise>, const String&);
    void json(FetchBodyOwner&, RefPtr<DeferredPromise>);
    void text(FetchBodyOwner&, RefPtr<DeferredPromise>);
    void formData(FetchBodyOwner&, RefPtr<DeferredPromise> promise);

#if ENABLE(STREAMS_API)
    void consumeAsStream(FetchBodyOwner&, FetchResponseSource&);
#endif

    bool isBlob() const { return m_blob; }
    bool isFormData() const { return m_formData; }
    bool isArrayBuffer() const { return m_data; }
    bool isArrayBufferView() const { return m_dataView; }
    bool isURLSearchParams() const { return m_URLSearchParams; }
    bool isText() const { return !m_text.isNull(); }

    static PassOwnPtr<FetchBody> extract(ScriptExecutionContext&, JSC::ExecState&, JSC::JSValue, String&);
    static PassOwnPtr<FetchBody> loadingBody() { return adoptPtr(new FetchBody()); }

    ~FetchBody();

    void loadingFailed();
    void loadingSucceeded();

    RefPtr<FormData> bodyForInternalRequest(ScriptExecutionContext&) const;

    FetchBodyConsumer& consumer() { return m_consumer; }

    void consumeOnceLoadingFinished(FetchBodyConsumer::Type, RefPtr<DeferredPromise>, const String&);
	void cleanConsumePromise() { m_consumePromise.clear(); }

	PassOwnPtr<FetchBody> clone() const;

private:
	explicit FetchBody(PassRefPtr<Blob>);
	explicit FetchBody(PassRefPtr<JSC::ArrayBuffer>);
	explicit FetchBody(PassRefPtr<JSC::ArrayBufferView>);
	explicit FetchBody(PassRefPtr<FormData>);
	explicit FetchBody(const String&);
	explicit FetchBody(PassRefPtr<URLSearchParams>);
	explicit FetchBody(const FetchBodyConsumer& consumer);
    FetchBody() : m_consumer(FetchBodyConsumer::Type::None) { }

    void consume(FetchBodyOwner&, RefPtr<DeferredPromise>);

    void consumeArrayBuffer(RefPtr<DeferredPromise>);
    void consumeArrayBufferView(RefPtr<DeferredPromise>);
    void consumeText(RefPtr<DeferredPromise>, const String&);
    void consumeBlob(FetchBodyOwner&, RefPtr<DeferredPromise>);

    const Blob& blobBody() const { return *m_blob; }
    FormData& formDataBody() { return *m_formData; }
    const FormData& formDataBody() const { return *m_formData; }
    const JSC::ArrayBuffer& arrayBufferBody() const { return *m_data; }
    const JSC::ArrayBufferView& arrayBufferViewBody() const { return *m_dataView; }
    String& textBody() { return m_text; }
    const String& textBody() const { return m_text; }
    const URLSearchParams& urlSearchParamsBody() const { return *m_URLSearchParams; }

    // std::experimental::variant<std::nullptr_t, Ref<const Blob>, Ref<FormData>, Ref<const ArrayBuffer>, Ref<const ArrayBufferView>, Ref<const URLSearchParams>, String> m_data { nullptr };

    // FIXME: Add support for URLSearchParams.
    RefPtr<Blob> m_blob;
    RefPtr<FormData> m_formData;
    RefPtr<JSC::ArrayBuffer> m_data;
    RefPtr<JSC::ArrayBufferView> m_dataView;
    RefPtr<URLSearchParams> m_URLSearchParams;
    String m_text;

    FetchBodyConsumer m_consumer;
    RefPtr<DeferredPromise> m_consumePromise;
};

} // namespace WebCore

#endif // ENABLE(FETCH_API)

#endif // FetchBody_h
