/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FetchRequest.h"

#if ENABLE(FETCH_API)

#include "Dictionary.h"
#include "ExceptionCode.h"
#include "HTTPParsers.h"
#include "ScriptExecutionContext.h"
#include "SecurityOrigin.h"

namespace WebCore {

static bool setReferrerPolicy(FetchOptions& options, const String& referrerPolicy)
{
    if (referrerPolicy.isEmpty())
        options.setReferrerPolicy(ReferrerHeaderPolicy::EmptyString);
    else if (referrerPolicy == "no-referrer")
        options.setReferrerPolicy(ReferrerHeaderPolicy::NoReferrer);
    else if (referrerPolicy == "no-referrer-when-downgrade")
        options.setReferrerPolicy(ReferrerHeaderPolicy::NoReferrerWhenDowngrade);
    else if (referrerPolicy == "same-origin")
        options.setReferrerPolicy(ReferrerHeaderPolicy::SameOrigin);
    else if (referrerPolicy == "origin")
        options.setReferrerPolicy(ReferrerHeaderPolicy::Origin);
    else if (referrerPolicy == "strict-origin")
        options.setReferrerPolicy(ReferrerHeaderPolicy::StrictOrigin);
    else if (referrerPolicy == "origin-when-cross-origin")
        options.setReferrerPolicy(ReferrerHeaderPolicy::OriginWhenCrossOrigin);
    else if (referrerPolicy == "strict-origin-when-cross-origin")
        options.setReferrerPolicy(ReferrerHeaderPolicy::StrictOriginWhenCrossOrigin);
    else if (referrerPolicy == "unsafe-url")
        options.setReferrerPolicy(ReferrerHeaderPolicy::UnsafeUrl);
    else
        return false;
    return true;
}

static bool setMode(FetchOptions& options, const String& mode)
{
    if (mode == "navigate")
        options.setMode(FetchOptions::Mode::Navigate);
    else if (mode == "same-origin")
        options.setMode(FetchOptions::Mode::SameOrigin);
    else if (mode == "no-cors")
        options.setMode(FetchOptions::Mode::NoCors);
    else if (mode == "cors")
        options.setMode(FetchOptions::Mode::Cors);
    else
        return false;
    return true;
}

static bool setCredentials(FetchOptions& options, const String& credentials)
{
    if (credentials == "omit")
        options.setCredentials(FetchOptions::Credentials::Omit);
    else if (credentials == "same-origin")
        options.setCredentials(FetchOptions::Credentials::SameOrigin);
    else if (credentials == "include")
        options.setCredentials(FetchOptions::Credentials::Include);
    else
        return false;
    return true;
}

static bool setCache(FetchOptions& options, const String& cache)
{
    if (cache == "default")
        options.setCache(FetchOptions::Cache::Default);
    else if (cache == "no-store")
        options.setCache(FetchOptions::Cache::NoStore);
    else if (cache == "reload")
        options.setCache(FetchOptions::Cache::Reload);
    else if (cache == "no-cache")
        options.setCache(FetchOptions::Cache::NoCache);
    else if (cache == "force-cache")
        options.setCache(FetchOptions::Cache::ForceCache);
    else if (cache == "only-if-cached")
        options.setCache(FetchOptions::Cache::OnlyIfCached);
    else
        return false;
    return true;
}

static bool setRedirect(FetchOptions& options, const String& redirect)
{
    if (redirect == "follow")
        options.setRedirect(FetchOptions::Redirect::Follow);
    else if (redirect == "error")
        options.setRedirect(FetchOptions::Redirect::Error);
    else if (redirect == "manual")
        options.setRedirect(FetchOptions::Redirect::Manual);
    else
        return false;
    return true;
}

static bool setMethod(ResourceRequest& request, const String& initMethod)
{
    if (!isValidHTTPToken(initMethod))
        return false;

    String method = initMethod.convertToASCIIUppercase();
    if (method == "CONNECT" || method == "TRACE" || method == "TRACK")
        return false;

    request.setHTTPMethod((method == "DELETE" || method == "GET" || method == "HEAD" || method == "OPTIONS" || method == "POST" || method == "PUT") ? method : initMethod);

    return true;
}

static bool setReferrer(FetchRequest::InternalRequest& request, ScriptExecutionContext& context, const Dictionary& init)
{
    String referrer;
    if (!init.get("referrer", referrer))
        return true;
    if (referrer.isEmpty()) {
        request.referrer = ASCIILiteral("no-referrer");
        return true;
    }
    // FIXME: Tighten the URL parsing algorithm according https://url.spec.whatwg.org/#concept-url-parser.
    KURL referrerURL = context.completeURL(referrer);
    if (!referrerURL.isValid())
        return false;

    if (referrerURL.protocolIs("about") && referrerURL.path() == "client") {
        request.referrer = ASCIILiteral("client");
        return true;
    }

    if (!(context.securityOrigin() && context.securityOrigin()->canRequest(referrerURL)))
        return false;

    request.referrer = referrerURL.string();
    return true;
}

static bool buildOptions(FetchRequest::InternalRequest& request, ScriptExecutionContext& context, const Dictionary& init)
{
    JSC::JSValue window;
    if (init.get("window", window)) {
        if (!window.isNull())
            return false;
    }

    if (!setReferrer(request, context, init))
        return false;

    String value;
    if (init.get("referrerPolicy", value) && !setReferrerPolicy(request.options, value))
        return false;

    if (init.get("mode", value) && !setMode(request.options, value))
        return false;
    if (request.options.mode() == FetchOptions::Mode::Navigate)
        return false;

    if (init.get("credentials", value) && !setCredentials(request.options, value))
        return false;

    if (init.get("cache", value) && !setCache(request.options, value))
        return false;

    if (request.options.cache() == FetchOptions::Cache::OnlyIfCached && request.options.mode() != FetchOptions::Mode::SameOrigin)
        return false; // Exception { TypeError, ASCIILiteral("only-if-cached cache option requires fetch mode to be same-origin.")  };

    if (init.get("redirect", value) && !setRedirect(request.options, value))
        return false;

    init.get("integrity", request.integrity);

    if (init.get("method", value) && !setMethod(request.request, value))
        return false;

    return true;
}

static bool methodCanHaveBody(const FetchRequest::InternalRequest& internalRequest)
{
    return internalRequest.request.httpMethod() != "GET" && internalRequest.request.httpMethod() != "HEAD";
}

void FetchRequest::initializeOptions(const Dictionary& init, ExceptionCode& ec)
{
    ASSERT(scriptExecutionContext());
    if (!buildOptions(m_internalRequest, *scriptExecutionContext(), init)) {
        ec = TypeError;
        return;
    }

    if (m_internalRequest.options.mode() == FetchOptions::Mode::NoCors) {
        const String& method = m_internalRequest.request.httpMethod();
        if (method != "GET" && method != "POST" && method != "HEAD") {
            ec = TypeError;
            return;
        }
        if (!m_internalRequest.integrity.isEmpty()) {
            ec = TypeError;
            return;
        }
        m_headers->setGuard(FetchHeaders::Guard::RequestNoCors);
    }
}

FetchHeaders* FetchRequest::initializeWith(const String& url, const Dictionary& init, ExceptionCode& ec)
{
    ASSERT(scriptExecutionContext());
    // FIXME: Tighten the URL parsing algorithm according https://url.spec.whatwg.org/#concept-url-parser.
    KURL requestURL = scriptExecutionContext()->completeURL(url);
    if (!requestURL.isValid() || !requestURL.user().isEmpty() || !requestURL.pass().isEmpty()) {
        ec = TypeError;
        return 0;
    }

    m_internalRequest.options.setMode(FetchOptions::Mode::Cors);
    m_internalRequest.options.setCredentials(FetchOptions::Credentials::Omit);
    m_internalRequest.referrer = ASCIILiteral("client");
    m_internalRequest.request.setURL(requestURL);

    initializeOptions(init, ec);
    return m_headers.get();
}

FetchHeaders* FetchRequest::initializeWith(FetchRequest* input, const Dictionary& init, ExceptionCode& ec)
{
    if (input->isDisturbedOrLocked()) {
        ec = TypeError;
        return 0;
    }

    m_internalRequest = input->m_internalRequest;

    initializeOptions(init, ec);
    return m_headers.get();
}

void FetchRequest::setBody(JSC::ExecState* execState, JSC::JSValue body, FetchRequest* request, ExceptionCode& ec)
{
    if (!body.isNull()) {
        if (!methodCanHaveBody(m_internalRequest)) {
            ec = TypeError;
            return;
        }

        ASSERT(scriptExecutionContext());
        extractBody(*scriptExecutionContext(), *execState, body);
        if (isBodyNull()) {
            ec = TypeError;
            return;
        }
    } else if (request && !request->isBodyNull()) {
        if (!methodCanHaveBody(m_internalRequest)) {
            ec = TypeError;
            return;
        }

        m_body = request->m_body.release();
        request->setDisturbed();
    }
    updateContentType();
}

String FetchRequest::referrer() const
{
    if (m_internalRequest.referrer == "no-referrer")
        return String();
    if (m_internalRequest.referrer == "client")
        return ASCIILiteral("about:client");
    return m_internalRequest.referrer;
}

const String& FetchRequest::url() const
{
    if (m_requestURL.isNull())
        m_requestURL = m_internalRequest.request.url().serialize();
    return m_requestURL;
}

ResourceRequest FetchRequest::internalRequest() const
{
    ASSERT(scriptExecutionContext());

    ResourceRequest request = m_internalRequest.request;
    request.setHTTPHeaderFields(m_headers->internalHeaders());

    if (!isBodyNull())
    	request.setHTTPBody(body().bodyForInternalRequest(*scriptExecutionContext()));

    return request;
}

RefPtr<FetchRequest> FetchRequest::clone(ScriptExecutionContext* context, ExceptionCode& ec)
{
    if (isDisturbedOrLocked()) {
        ec = TypeError;
        return nullptr;
    }

    RefPtr<FetchRequest> clone = adoptRef(new FetchRequest(context, PassOwnPtr<FetchBody>(), FetchHeaders::create(*m_headers), FetchRequest::InternalRequest(m_internalRequest)));
    clone->cloneBody(*this);
    return clone.release();
}

const char* FetchRequest::activeDOMObjectName() const
{
    return "Request";
}

bool FetchRequest::canSuspendForDocumentSuspension() const
{
    // FIXME: We can probably do the same strategy as XHR.
    return !isActive();
}

} // namespace WebCore

#endif // ENABLE(FETCH_API)
