/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef FetchResponse_h
#define FetchResponse_h

#if ENABLE(FETCH_API)

#include "FetchBodyOwner.h"
#include "ResourceResponse.h"
#include <runtime/TypedArrays.h>

namespace JSC {
class ArrayBuffer;
class ExecState;
class JSValue;
};

namespace WebCore {

class DeferredPromise;
class Dictionary;
class FetchRequest;
class ReadableStreamSource;

typedef int ExceptionCode;

typedef ResourceResponse::Type FetchResponseType;

class FetchResponse : public FetchBodyOwner {
public:
	typedef ResourceResponse::Type Type;

    static PassRefPtr<FetchResponse> create(ScriptExecutionContext* context) { return adoptRef(new FetchResponse(context, PassOwnPtr<FetchBody>(), FetchHeaders::create(FetchHeaders::Guard::Response), ResourceResponse())); }
    static PassRefPtr<FetchResponse> error(ScriptExecutionContext*);
    static RefPtr<FetchResponse> redirect(ScriptExecutionContext*, const String&, int, ExceptionCode&);
    // FIXME: Binding generator should not require below method to handle optional status parameter.
    static RefPtr<FetchResponse> redirect(ScriptExecutionContext* context, const String& url, ExceptionCode& ec) { return redirect(context, url, 302, ec); }

    static void fetch(ScriptExecutionContext&, FetchRequest&, RefPtr<DeferredPromise>);

    void consume(unsigned, RefPtr<DeferredPromise>);
#if ENABLE(STREAMS_API)
    void startConsumingStream(unsigned);
    void consumeChunk(RefPtr<JSC::Uint8Array>);
    void finishConsumingStream(RefPtr<DeferredPromise>);
#endif

    void setStatus(int, const String&, ExceptionCode&);
    void initializeWith(JSC::ExecState*, JSC::JSValue);

    Type type() const { return m_response.type(); }
    const String& url() const;
    bool redirected() const { return m_response.isRedirected(); }
    int status() const { return m_response.httpStatusCode(); }
    bool ok() const { return m_response.isSuccessful(); }
    const String& statusText() const { return m_response.httpStatusText(); }

    FetchHeaders* headers() { return m_headers.get(); }
    RefPtr<FetchResponse> cloneForJS();

#if ENABLE(STREAMS_API)
    ReadableStreamSource* createReadableStreamSource();
    void consumeBodyAsStream();
    void feedStream();
    void cancel();
#endif
    bool isLoading() const { return !!m_bodyLoader; }

private:
    FetchResponse(ScriptExecutionContext*, PassOwnPtr<FetchBody>, RefPtr<FetchHeaders>, const ResourceResponse&);

    // ActiveDOMObject API
    virtual void stop() FINAL;
    virtual const char* activeDOMObjectName() const FINAL;
    virtual bool canSuspend() const { return canSuspendForDocumentSuspension(); }

    virtual bool canSuspendForDocumentSuspension() const FINAL;

#if ENABLE(STREAMS_API)
    void closeStream();
#endif

    class BodyLoader FINAL : public FetchLoaderClient {
    public:
        BodyLoader(FetchResponse&, RefPtr<DeferredPromise>);
        ~BodyLoader();

        bool start(ScriptExecutionContext&, const FetchRequest&);
        void stop();

#if ENABLE(STREAMS_API)
        RefPtr<SharedBuffer> startStreaming();
#endif

    private:
        // FetchLoaderClient API
        virtual void didSucceed() FINAL;
        virtual void didFail() FINAL;
        virtual void didReceiveResponse(const ResourceResponse&) FINAL;
        void didReceiveData(const char*, size_t) FINAL;

        FetchResponse& m_response;
        RefPtr<DeferredPromise> m_promise;
        OwnPtr<FetchLoader> m_loader;
    };

    ResourceResponse m_response;
    OwnPtr<BodyLoader> m_bodyLoader;
    mutable String m_responseURL;

    FetchBodyConsumer m_consumer;
};

} // namespace WebCore

#endif // ENABLE(FETCH_API)

#endif // FetchResponse_h
