/*
 *  Copyright (C) 2010 Espial Group Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "ArgList.h"
#include "Error.h"
#include "ErrorInstance.h"
#include "DateInstance.h"
#include "JSDOMWindow.h"
#include "JSObjectBridge.h"
#include "JSGlobalObject.h"
#include "JSDocument.h"
#include "JSFunction.h"
#include "JSGlobalObject.h"
#include "JSMainThreadExecState.h"
#include "JSNode.h"
#include "HTMLImageElement.h"
#include "File.h"
#include "Frame.h"
#include "JSFile.h"
#include "JSNode.h"
#include "JSHTMLObjectElement.h"
#include "JSWebGLTexture.h"
#include "DOMWrapperWorld.h"
#include "SlotVisitorInlineMethods.h"
#include "MarkStack.h"
#include "Operations.h"
#include "StructureInlines.h"
#include <wtf/MathExtras.h>
#include <wtf/NakedPtr.h>

using namespace JSC;

namespace
{

class JSCustomFunction : public InternalFunction {
private:
	struct Data
	{
		WTF_MAKE_FAST_ALLOCATED;
	public:
		Data(WebCoreAPIs::JSCustomCall customCall, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
		: m_customCall(customCall)
		, m_customDestroy(customDestroy)
		, m_customContext(customContext)
		{
		}

		EncodedJSValue call(ExecState* exec){
			if(m_customCall)
				return m_customCall(exec, m_customContext);
			else
				return JSValue::encode(jsUndefined());
		}

		~Data()
		{
	    	if(m_customDestroy)
	    		m_customDestroy(m_customContext);
		}

	    WebCoreAPIs::JSCustomCall m_customCall;
	    WebCoreAPIs::JSCustomDestroy m_customDestroy;
	    void* m_customContext;
	};
public:
    typedef InternalFunction Base;

    DECLARE_INFO;

    static FunctionPrototype* createPrototype(ExecState* exec, JSGlobalObject* globalObject)
    {
        return exec->lexicalGlobalObject()->functionPrototype();
    }

    static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype)
    {
        return Structure::create(globalData, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

	JSCustomFunction(ExecState* exec, JSGlobalObject* globalObject, WebCoreAPIs::JSCustomCall customCall, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
    : InternalFunction(globalObject->vm(), createStructure(globalObject->globalData(), globalObject, createPrototype(exec, globalObject)))
	, m_private(new Data(customCall, customDestroy, customContext))
    {
    }

    static void destroy(JSCell* cell)
    {
    	static_cast<JSCustomFunction*>(cell)->JSCustomFunction::~JSCustomFunction();
    }

    void finishCreation(JSGlobalData& globalData, const String& name)
    {
    	Base::finishCreation(globalData, name);
    }

    static const unsigned StructureFlags = OverridesGetCallData | Base::StructureFlags;

private:
    static CallType getCallData(JSCell*, CallData&);
    static EncodedJSValue JSC_HOST_CALL call(ExecState* exec);

    ~JSCustomFunction()
    {
    	delete m_private;
    }

private:
    Data* m_private;
};

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSCustomFunction);

const ClassInfo JSCustomFunction::s_info = { "CustomFunction", &Base::s_info, 0, CREATE_METHOD_TABLE(JSCustomFunction) };

CallType JSCustomFunction::getCallData(JSCell*, CallData& callData)
{
	callData.native.function = call;
	return CallType::Host;
}

EncodedJSValue JSC_HOST_CALL JSCustomFunction::call(ExecState* exec)
{
	if(asObject(exec->callee())->inherits(JSCustomFunction::info())){
		return jsCast<JSCustomFunction*>(asObject(exec->callee()))->m_private->call(exec);
	}

	return JSValue::encode(jsUndefined());
}

} // anonymous namespace

namespace
{

class JSCustomConstructor : public JSDestructibleObject
{
private:
	struct Data
	{
		WTF_MAKE_FAST_ALLOCATED;
	public:
		Data(WebCoreAPIs::JSCustomConstruct customConstruct, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
		: m_customConstruct(customConstruct)
		, m_customDestroy(customDestroy)
		, m_customContext(customContext)
		{
	    }

		~Data()
		{
	    	if(m_customDestroy)
	    		m_customDestroy(m_customContext);
		}

		EncodedJSValue construct(ExecState* exec)
		{
			if(m_customConstruct)
				return m_customConstruct(exec, m_customContext);

			return JSValue::encode(jsUndefined());
		}

		WebCoreAPIs::JSCustomConstruct m_customConstruct;
		WebCoreAPIs::JSCustomDestroy m_customDestroy;
		void* m_customContext;
	};

public:
    typedef JSDestructibleObject Base;

    static JSCustomConstructor* create(ExecState* exec, WebCoreAPIs::JSCustomConstruct customConstruct, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
    {
    	VM& vm = exec->vm();
    	JSCustomConstructor* constructor = new (NotNull, allocateCell<JSCustomConstructor>(vm.heap)) JSCustomConstructor(exec, exec->lexicalGlobalObject(), customConstruct, customDestroy, customContext);
    	constructor->Base::finishCreation(vm);
        return constructor;
    }

    ~JSCustomConstructor()
    {
    	delete m_private;
    }

    DECLARE_INFO;

    static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
    {
        return Structure::create(globalData, globalObject, proto, TypeInfo(ObjectType, StructureFlags), info());
    }

    static void destroy(JSCell* cell)
    {
    	static_cast<JSCustomConstructor*>(cell)->JSCustomConstructor::~JSCustomConstructor();
    }

private:
    static const unsigned StructureFlags = ImplementsHasInstance | JSObject::StructureFlags;

    static EncodedJSValue JSC_HOST_CALL construct(ExecState* exec);

    static ConstructType getConstructData(JSCell*, ConstructData& constructData)
    {
        constructData.native.function = construct;
        return ConstructType::Host;
    }

private:
    JSCustomConstructor(ExecState* exec, JSGlobalObject* globalObject, WebCoreAPIs::JSCustomConstruct customConstruct, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
	: JSDestructibleObject(globalObject->globalData(), createStructure(globalObject->globalData(), globalObject, globalObject->objectPrototype()))
	, m_private(new Data(customConstruct, customDestroy, customContext))
	{
	}

    Data* m_private;
};

const ClassInfo JSCustomConstructor::s_info = { "CustomConstructor", &Base::s_info, 0, CREATE_METHOD_TABLE(JSCustomConstructor) };

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSCustomConstructor);

EncodedJSValue JSC_HOST_CALL JSCustomConstructor::construct(ExecState* exec)
{
	if(asObject(exec->callee())->inherits(JSCustomConstructor::info())){
		return jsCast<JSCustomConstructor*>(asObject(exec->callee()))->m_private->construct(exec);
	}

	return JSValue::encode(jsUndefined());
}

} // // anonymous namespace

namespace WebCoreAPIs
{

JSCustomInstance::~JSCustomInstance()
{
}

Structure* JSCustomInstance::createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
{
	return JSC::Structure::create(globalData, globalObject, proto, JSC::TypeInfo(JSC::ObjectType, StructureFlags), info());
}

bool JSCustomInstance::put(JSC::JSCell* cell, JSC::ExecState* exec, JSC::PropertyName propertyName, JSC::JSValue value, JSC::PutPropertySlot& slot)
{
	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(cell);

	if(customInstance->m_private->put(customInstance, exec, propertyName, value, slot))
		return true;

	return customInstance->Base::putDirect(exec->globalData(), propertyName, value, slot);
}

bool JSCustomInstance::putByIndex(JSC::JSCell* cell, JSC::ExecState* exec, unsigned propertyName, JSC::JSValue value, bool shouldThrow)
{
	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(cell);

	if(customInstance->m_private->putByIndex(customInstance, exec, propertyName, value, shouldThrow))
		return true;

	VM& vm = exec->vm();
    PutPropertySlot slot(cell, shouldThrow);
    return JSCustomInstance::put(cell, exec, Identifier::from(vm, propertyName), value, slot);
}

bool JSCustomInstance::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(object);

	if(customInstance->m_private->get(customInstance, exec, propertyName, slot))
		return true;

	return JSObject::getOwnPropertySlot(customInstance, exec, propertyName, slot);
}

bool JSCustomInstance::getOwnPropertySlotByIndex(JSObject* object, ExecState* exec, unsigned propertyName, PropertySlot& slot)
{
	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(object);

	if(customInstance->m_private->getByIndex(customInstance, exec, propertyName, slot))
		return true;

	VM& vm = exec->vm();
	return JSCustomInstance::getOwnPropertySlot(object, exec, Identifier::from(vm, propertyName), slot);
}

void JSCustomInstance::destroy(JSCell* cell)
{
	static_cast<JSCustomInstance*>(cell)->JSCustomInstance::~JSCustomInstance();
}

void JSCustomInstance::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
	Base::visitChildren(cell, visitor);
	jsCast<JSCustomInstance*>(cell)->m_private->visitChildren(jsCast<JSCustomInstance*>(cell), visitor);
}

DLL_PUBLIC void append(JSC::SlotVisitor& visitor, JSC::JSValue v)
{
	if(v.isUndefinedOrNull())
		return;

	visitor.appendUnbarrieredValue(&v);
}

ALWAYS_INLINE JSCustomInstance::JSCustomInstance(ExecState* exec)
: JSDestructibleObject(exec->globalData(), createStructure(exec->globalData(), exec->lexicalGlobalObject(), exec->lexicalGlobalObject()->objectPrototype()))
, m_private(adoptPtr(new Data()))
{
}

ALWAYS_INLINE JSCustomInstance::JSCustomInstance(ExecState* exec, const JSCustomObjectConstructorData& constructorData)
: JSDestructibleObject(exec->globalData(), createStructure(exec->globalData(), exec->lexicalGlobalObject(), exec->lexicalGlobalObject()->objectPrototype()))
, m_private(adoptPtr(new Data(constructorData)))
{
}

void JSCustomInstance::finishCreation(JSGlobalData& globalData)
{
	Base::finishCreation(globalData);
}

const ClassInfo JSCustomInstance::s_info = { "CustomObject", &Base::s_info, 0, CREATE_METHOD_TABLE(JSCustomInstance) };

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSCustomInstance);

DLL_PUBLIC JSCustomInstance* asJSCustomInstance(JSC::JSValue value)
{
	if(value.isCell()){
		JSCell* cell = value.asCell();

		if(cell->inherits(JSCustomInstance::info()))
			return jsCast<JSCustomInstance*>(cell);
	}

	return 0;
}

DLL_PUBLIC JSC::PrivateName createPrivateName()
{
	return JSC::PrivateName();
}

DLL_PUBLIC JSC::JSValue get(JSC::JSValue value, JSC::ExecState* exec, JSC::PropertyName propertyName)
{
	return value.get(exec, propertyName);
}

DLL_PUBLIC JSC::JSValue get(JSC::JSValue value, JSC::ExecState* exec, const String& propertyName)
{
	return value.get(exec, Identifier::fromString(exec->vm(), propertyName));
}

DLL_PUBLIC JSC::JSValue get(JSC::JSValue value, JSC::ExecState* exec, unsigned propertyName)
{
	return value.get(exec, propertyName);
}

DLL_PUBLIC JSC::JSValue get(JSC::ExecState* exec, JSC::PropertySlot slot, JSC::PropertyName propertyName)
{
	return slot.getValue(exec, propertyName);
}

DLL_PUBLIC bool getDirect(JSC::JSObject* thisObject, JSC::ExecState* exec, const String& propertyName, JSC::PropertySlot& slot)
{
	return JSObject::getOwnPropertySlot(thisObject, exec, Identifier::fromString(exec->vm(), propertyName), slot);
}

DLL_PUBLIC bool getDirect(JSC::JSObject* thisObject, JSC::ExecState* exec, JSC::PropertyName propertyName, JSC::PropertySlot& slot)
{
	return JSObject::getOwnPropertySlot(thisObject, exec, propertyName, slot);
}

DLL_PUBLIC void putDirect(JSC::JSObject* thisObject, JSC::JSGlobalData& globalData, JSC::PropertyName propertyName, JSC::JSValue value, unsigned attributes)
{
	thisObject->JSObject::putDirect(globalData, propertyName, value, attributes);
}

DLL_PUBLIC void putDirect(JSC::JSObject* thisObject, JSC::ExecState* exec, const String& propertyName, JSC::JSValue value, unsigned attributes)
{
	JSC::VM& vm = exec->vm();
	thisObject->JSObject::putDirect(vm, Identifier::fromString(vm, propertyName), value, attributes);
}

DLL_PUBLIC WebCore::Node* toNode(JSC::JSValue value)
{
	return WebCore::toNode(value);
}

DLL_PUBLIC WebCore::JSHTMLObjectElement* toJSHTMLObjectElement(JSC::JSValue thisValue)
{
	return JSC::jsDynamicCast<WebCore::JSHTMLObjectElement*>(thisValue);
}

#if ENABLE(WEBGL)
DLL_PUBLIC unsigned int toWebGLTextureObject(JSC::JSValue value)
{
    if(WebCore::WebGLTexture* texture = WebCore::toWebGLTexture(value))
    	return texture->object();
    return 0;
}
#endif

DLL_PUBLIC WebCore::HTMLObjectElement* toHTMLObjectElement(JSC::JSValue thisValue)
{
	if(WebCore::JSHTMLObjectElement* jsElement = JSC::jsDynamicCast<WebCore::JSHTMLObjectElement*>(thisValue))
		return static_cast<WebCore::HTMLObjectElement*>(&jsElement->wrapped());

	return 0;
}

DLL_PUBLIC JSC::JSObject* createJSFunction(JSC::ExecState* exec, JSCustomCall customCall, JSCustomDestroy customDestroy, void* customContext)
{
	VM& vm = exec->vm();
	JSCustomFunction* function = new (NotNull, allocateCell<JSCustomFunction>(vm.heap)) JSCustomFunction(exec, exec->lexicalGlobalObject(), customCall, customDestroy, customContext);
    function->finishCreation(vm, String(""));
    return function;
}

DLL_PUBLIC JSC::JSObject* createJSError(JSC::ExecState* exec, const String& message)
{
	return JSC::createError(exec, message, 0);
}

DLL_PUBLIC JSC::JSObject* createJSObject(JSC::ExecState* exec, const JSCustomObjectConstructorData& constructorData)
{
	VM& vm = exec->vm();
	JSCustomInstance* customInstance = new (NotNull, JSC::allocateCell<JSCustomInstance>(vm.heap)) JSCustomInstance(exec, constructorData);
	customInstance->finishCreation(vm);
	return customInstance;
}

DLL_PUBLIC JSC::JSObject* createJSConstructor(JSC::ExecState* exec, JSCustomConstruct customConstruct, JSCustomDestroy customDestroy, void* customContext)
{
	return JSCustomConstructor::create(exec, customConstruct, customDestroy, customContext);
}

DLL_PUBLIC bool isJSArray(JSValue v)
{
	return JSC::isJSArray(v);
}

DLL_PUBLIC JSArray* JSArrayBridge::create(ExecState* exec)
{
	return JSArray::create(exec->globalData(), exec->lexicalGlobalObject()->arrayStructureForIndexingTypeDuringAllocation(ArrayWithUndecided));
}

DLL_PUBLIC void JSArrayBridge::push(JSArray* jsArray, ExecState* exec, JSValue value)
{
	jsArray->push(exec, value);
}

DLL_PUBLIC void JSArrayBridge::setLength(JSArray* jsArray, ExecState* exec, unsigned newLength)
{
	jsArray->setLength(exec, newLength);
}

DLL_PUBLIC JSValue newDateInstance(ExecState* exec, double d)
{
    return DateInstance::create(exec->globalData(), exec->lexicalGlobalObject()->dateStructure(), d);
}

DLL_PUBLIC JSValue newDateInstance(
		ExecState* exec,
		int year,	/* specified as an number, i.e. 2010 */
		int month, 	/* specified as a number (0-11), i.e. 5 */
		int day,  	/* day of the month specified as a number (1-31), i.e. 18 */
		int hour, 	/* Hour specified as a number (0-23), i.e. 13 */
		int minute, /* specified as a number (0-59), i.e. 34 */
		int second, /* specified as a number (0-59), i.e. 23 */
		int ms		/* milli-second specified as a number (0-999), i.e. 240 */
		)
{
    GregorianDateTime gregorianDateTime;
    
    gregorianDateTime.setYear(year);
	gregorianDateTime.setMonth(month);
	gregorianDateTime.setMonthDay(day);
	gregorianDateTime.setHour(hour);
	gregorianDateTime.setMinute(minute);
	gregorianDateTime.setSecond(second);

	return newDateInstance(exec, gregorianDateTimeToMS(exec->vm(), gregorianDateTime, 0, 0));
}

DLL_PUBLIC bool parseDateInstance(
		ExecState* exec,
		JSValue value,
		int& year,	/* specified as an number, i.e. 2010 */
		int& month,	/* specified as a number (0-11), i.e. 5 */
		int& day,  	/* day of the month specified as a number (1-31), i.e. 18 */
		int& hour, 	/* Hour specified as a number (0-23), i.e. 13 */
		int& minute,/* specified as a number (0-59), i.e. 34 */
		int& second,/* specified as a number (0-59), i.e. 23 */
		int& ms		/* milli-second specified as a number (0-999), i.e. 240 */
		)
{
	DateInstance* dateInstance = jsDynamicCast<DateInstance*>(value);

	if(!dateInstance)
		return false;

	VM& vm = exec->vm();
    const GregorianDateTime* gregorianDateTime = dateInstance->gregorianDateTime(vm);

    year    = gregorianDateTime->year();
	month   = gregorianDateTime->month();
	day     = gregorianDateTime->monthDay();
	hour    = gregorianDateTime->hour();
	minute  = gregorianDateTime->minute();
	second  = gregorianDateTime->second();
	ms      = 0;
	
	return true;
}

DLL_PUBLIC JSValue newJSNaN()
{
	return jsNaN();
}

DLL_PUBLIC double Inf()
{
	return std::numeric_limits<double>::infinity();
}

DLL_PUBLIC JSValue toPrimitive(ExecState* exec, JSValue v)
{
	return v.toPrimitive(exec);
}

DLL_PUBLIC JSValue toJS(ExecState* exec, WebCore::Frame* frame, RefPtr<WebCore::HTMLImageElement> imageElement)
{
	if(!imageElement || !frame)
        return jsUndefined();

	WebCore::JSDOMWindow* globalObject = frame->script()->globalObject(WebCore::mainThreadNormalWorld());
    if (!globalObject)
        return jsUndefined();

	return WebCore::toJS(exec, globalObject, imageElement.get());
}

DLL_PUBLIC JSValue toJS(ExecState* exec, WebCore::Frame* frame, PassRefPtr<WebCore::File> file)
{
	if(!file || !frame)
        return jsUndefined();

	WebCore::JSDOMWindow* globalObject = frame->script()->globalObject(WebCore::mainThreadNormalWorld());
    if (!globalObject)
        return jsUndefined();

	return WebCore::toJS(exec, globalObject, file.get());
}

DLL_PUBLIC JSC::JSValue toJS(JSC::ExecState* exec , WebCore::DOMWindow* domWindow)
{
	return WebCore::toJS(exec, domWindow);
}

DLL_PUBLIC bool toBoolean(ExecState* exec, JSValue v)
{
	return v.toBoolean(exec);
}

DLL_PUBLIC int32_t toInt32(ExecState* exec, JSValue v)
{
	return v.toInt32(exec);
}

DLL_PUBLIC uint32_t toUInt32(ExecState* exec, JSValue v)
{
	return v.toUInt32(exec);
}

DLL_PUBLIC double toInteger(ExecState* exec, JSValue v)
{
	return v.toInteger(exec);
}

DLL_PUBLIC double toNumber(ExecState* exec, JSValue v)
{
	return v.toNumber(exec);
}

DLL_PUBLIC JSString* newJSString(ExecState* exec, const char* s)
{
	return jsString(exec->vm(), String(s));
}

DLL_PUBLIC JSValue newJSString(ExecState* exec, const WTF::String& s)
{
	return jsString(exec->vm(), s);
}

DLL_PUBLIC JSValue newJSSingleCharacterString(ExecState* exec, UChar c)
{
	return jsSingleCharacterString(exec->vm(), c);
}

DLL_PUBLIC void append(JSC::MarkedArgumentBuffer& args, JSC::JSValue v)
{
	args.append(v);
}

DLL_PUBLIC WebCore::DOMWrapperWorld* mainThreadNormalWorld()
{
	return WebCore::mainThreadNormalWorld();
}

DLL_PUBLIC WebCore::JSDOMWindow* globalObject(WebCore::Frame* frame)
{
	return frame->script()->globalObject(WebCore::mainThreadNormalWorld());
}

DLL_PUBLIC JSC::Bindings::RootObject* bindingRootObject(WebCore::Frame* frame)
{
	return frame->script()->bindingRootObject();
}

#if ENABLE(MOTTO_TV)
DLL_PUBLIC Vector<String> getPropertyNames(JSC::ExecState* exec, JSC::JSValue v)
{
	Vector<String> names;
	JSObject* object = v.toObject(exec);

	if(object){
		JSPropertyNameIterator* it = JSPropertyNameIterator::create(exec, object);
		for(size_t i = 0; i < it->size(); ++i){
			names.append(it->get(exec, object, i).toWTFString(exec));
		}
	}

	return names;
}
#endif

DLL_PUBLIC bool dispatchEvent(JSC::ExecState* exec, JSC::JSGlobalObject* globalObject, JSC::JSValue listener, const JSC::MarkedArgumentBuffer& args, JSC::JSValue& jsReturnValue)
{
	bool dispatched = false;

	CallData callData;
	CallType callType;

    JSGlobalData& vm = exec->vm();
    JSLockHolder lock(vm);

	callType = getCallData(listener, callData);

	if (callType != CallType::None){
		NakedPtr<JSC::Exception> exception;
        DynamicGlobalObjectScope globalObjectScope(vm, vm.dynamicGlobalObject ? vm.dynamicGlobalObject : globalObject);
        jsReturnValue = WebCore::JSMainThreadExecState::call(exec, listener, callType, callData, exec->globalThisValue(), args, exception);
		dispatched = true;
	}

	return dispatched;
}

/**
 * dispatch event to "obj.a.b" with optional arguments
 */
DLL_PUBLIC bool dispatchEvent(WebCore::Frame* frame, const Vector<String, 2>& objects, const MarkedArgumentBuffer& args, JSC::JSValue& jsReturnValue)
{
	if(!frame)
		return false;

	if(objects.size() == 0)
		return false;

	WebCore::JSDOMWindow* globalObject = frame->script()->globalObject(WebCore::mainThreadNormalWorld());

	if(!globalObject)
		return false;

	bool dispatched = false;
	ExecState* exec = globalObject->globalExec();
	VM& vm = globalObject->vm();

	JSValue value = globalObject->get(exec, Identifier::fromString(vm, objects[0]));

	if(value.isUndefinedOrNull())
		goto OUT;

	for(int i = 1; i < objects.size(); ++i){
		JSValue property = value.get(exec, Identifier::fromString(vm, objects[i]));

		if(property.isUndefinedOrNull())
			goto OUT;

		value = property;
	}

	dispatched = dispatchEvent(exec, globalObject, value, args, jsReturnValue);

OUT:
	exec->clearException();

	return dispatched;
}

DLL_PUBLIC JSObject* throwException(ExecState* exec, JSObject* error)
{
	return exec->globalData().throwException(exec, error);
}

DLL_PUBLIC JSValue throwException(ExecState* exec, JSValue error)
{
	return exec->globalData().throwException(exec, error);
}

} //namespace WebCoreAPIs


namespace WebCore
{

DLL_PUBLIC Document* getDocument(JSC::ExecState* exec, JSC::JSGlobalObject* globalObject)
{
	return toDocument(globalObject->get(exec, globalObject->vm().propertyNames->document));
}

DLL_PUBLIC String toStringWithNullCheck(JSC::ExecState* exec, JSC::JSValue value)
{
    if (value.isNull())
        return String();
    return value.toWTFString(exec);
}

DLL_PUBLIC JSC::JSValue getProperty(JSC::JSObject* object, JSC::ExecState* exec, const String& propertyName)
{
	VM& vm = exec->vm();
	return object->get(exec, JSC::Identifier::fromString(vm, propertyName));
}

} //namespace WebCore

