/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 * Copyright (C) 2013 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef DocumentRuleSets_h
#define DocumentRuleSets_h

#include "RuleFeature.h"
#include "RuleSet.h"

#include <wtf/HashMap.h>
#include <wtf/OwnPtr.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>

namespace WebCore {

class CSSStyleRule;
class CSSStyleSheet;
class ExtensionStyleSheets;
class InspectorCSSOMWrappers;
class MediaQueryEvaluator;
class MatchRequest;
class RuleSet;
class StyleResolver;

#if ENABLE(SHADOW_DOM)
class ShadowDistributedRules {
	WTF_MAKE_FAST_ALLOCATED;
public:
    void addRule(StyleRule*, size_t selectorIndex, ContainerNode* scope, AddRuleFlags);
    void collectMatchRequests(bool includeEmptyRules, Vector<MatchRequest>&);
    void clear() { m_shadowDistributedRuleSetMap.clear(); }
    bool isEmpty() const { return m_shadowDistributedRuleSetMap.isEmpty(); }
private:
    typedef HashMap<const ContainerNode*, OwnPtr<RuleSet> > ShadowDistributedRuleSetMap;
    ShadowDistributedRuleSetMap m_shadowDistributedRuleSetMap;
};
#endif

class DocumentRuleSets {
	WTF_MAKE_FAST_ALLOCATED;
public:
    DocumentRuleSets();
    ~DocumentRuleSets();
    RuleSet* authorStyle() const { return m_authorStyle.get(); }
    RuleSet* userStyle() const { return m_userStyle.get(); }
    RuleFeatureSet& features() { return m_features; }
    const RuleFeatureSet& features() const { return m_features; }
    RuleSet* sibling() const { return m_siblingRuleSet.get(); }
    RuleSet* uncommonAttribute() const { return m_uncommonAttributeRuleSet.get(); }
    RuleSet* ancestorClassRules(AtomicStringImpl* className) const;

    struct AttributeRules {
    	WTF_MAKE_FAST_ALLOCATED;
    public:
        Vector<const CSSSelector*> attributeSelectors;
        OwnPtr<RuleSet> ruleSet;
    };
    const AttributeRules* ancestorAttributeRulesForHTML(AtomicStringImpl*) const;

    void initUserStyle(ExtensionStyleSheets&, const MediaQueryEvaluator&, StyleResolver&);
    void resetAuthorStyle();
    void appendAuthorStyleSheets(const Vector<RefPtr<CSSStyleSheet> >&, MediaQueryEvaluator*, InspectorCSSOMWrappers&, bool isViewSource, StyleResolver*);

    void collectFeatures(bool isViewSource);

private:
    void collectRulesFromUserStyleSheets(const Vector<RefPtr<CSSStyleSheet> >&, RuleSet& userStyle, const MediaQueryEvaluator&, StyleResolver&);
    OwnPtr<RuleSet> m_authorStyle;
    OwnPtr<RuleSet> m_userStyle;
    RuleFeatureSet m_features;
    OwnPtr<RuleSet> m_siblingRuleSet;
    OwnPtr<RuleSet> m_uncommonAttributeRuleSet;
    mutable HashMap<AtomicStringImpl*, OwnPtr<RuleSet>> m_ancestorClassRuleSets;
    mutable HashMap<AtomicStringImpl*, OwnPtr<AttributeRules>> m_ancestorAttributeRuleSetsForHTML;
};

} // namespace WebCore

#endif // DocumentRuleSets_h
