/*
 * Copyright (C) 2001 Peter Kelly (pmk@post.com)
 * Copyright (C) 2001 Tobias Anton (anton@stud.fbi.fh-darmstadt.de)
 * Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
 * Copyright (C) 2003, 2005, 2006, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "Event.h"

#include "EventNames.h"
#include "EventPath.h"
#include "EventTarget.h"
#include "UserGestureIndicator.h"
#include <wtf/CurrentTime.h>
#include <wtf/HashSet.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/CString.h>

namespace WebCore {

#if ENABLE(DUMP_EVENTS)
static Mutex& mutex()
{
	static Mutex mutex;
	return mutex;
}

static HashSet<Event*>& events()
{
	static HashSet<Event*> evts;
	return evts;
}

DLL_PUBLIC void dumpEvents(Vector<pair<String, int> >& dumpEvts)
{
	MutexLocker locker(mutex());

	HashSet<Event*>& evts = events();

	for(HashSet<Event*>::iterator it = evts.begin(); it != evts.end(); ++it) {
		String name = (*it)->type();
		bool found = false;

		for(int j = 0; j < dumpEvts.size(); ++j) {
			if(name == dumpEvts[j].first) {
				++dumpEvts[j].second;
				found = true;
				break;
			}
		}

		if(!found)
			dumpEvts.append(pair<String, int>(name, 1));
	}
}

#endif

Event::Event()
    : m_isInitialized(false)
	, m_canBubble(false)
    , m_cancelable(false)
	, m_composed(false)
    , m_propagationStopped(false)
    , m_immediatePropagationStopped(false)
    , m_defaultPrevented(false)
    , m_defaultHandled(false)
    , m_cancelBubble(false)
	, m_isTrusted(false)
    , m_eventPhase(0)
    , m_currentTarget(0)
	, m_eventPath(0)
    , m_createTime(convertSecondsToDOMTimeStamp(currentTime()))
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().add(this);
#endif
}

Event::Event(const AtomicString& eventType, bool canBubbleArg, bool cancelableArg)
    : m_type(eventType)
	, m_isInitialized(true)
    , m_canBubble(canBubbleArg)
    , m_cancelable(cancelableArg)
	, m_composed(false)
    , m_propagationStopped(false)
    , m_immediatePropagationStopped(false)
    , m_defaultPrevented(false)
    , m_defaultHandled(false)
    , m_cancelBubble(false)
	, m_isTrusted(true)
    , m_eventPhase(0)
    , m_currentTarget(0)
	, m_eventPath(0)
    , m_createTime(convertSecondsToDOMTimeStamp(currentTime()))
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().add(this);
#endif
}

Event::Event(const AtomicString& eventType, const EventInit& initializer)
    : m_type(eventType)
	, m_isInitialized(true)
    , m_canBubble(initializer.bubbles)
    , m_cancelable(initializer.cancelable)
	, m_composed(initializer.composed)
    , m_propagationStopped(false)
    , m_immediatePropagationStopped(false)
    , m_defaultPrevented(false)
    , m_defaultHandled(false)
    , m_cancelBubble(false)
	, m_isTrusted(true)
    , m_eventPhase(0)
    , m_currentTarget(0)
	, m_eventPath(0)
    , m_createTime(convertSecondsToDOMTimeStamp(currentTime()))
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().add(this);
#endif
}

Event::~Event()
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().remove(this);
#endif
}

void Event::initEvent(const AtomicString& eventTypeArg, bool canBubbleArg, bool cancelableArg)
{
    if (dispatched())
        return;

    m_isInitialized = true;
    m_propagationStopped = false;
    m_immediatePropagationStopped = false;
    m_defaultPrevented = false;
    m_isTrusted = false;

    m_type = eventTypeArg;
    m_canBubble = canBubbleArg;
    m_cancelable = cancelableArg;
}

bool Event::composed() const
{
    if (m_composed)
        return true;

    // http://w3c.github.io/webcomponents/spec/shadow/#scoped-flag
    if (!isTrusted())
        return false;

    return m_type == eventNames().inputEvent
        || m_type == eventNames().textInputEvent
        || isCompositionEvent()
        || isClipboardEvent()
        || isFocusEvent()
        || isKeyboardEvent()
        || isMouseEvent()
        || isTouchEvent();
}

EventInterface Event::eventInterface() const
{
    return EventInterfaceType;
}

bool Event::isUIEvent() const
{
    return false;
}

bool Event::isMouseEvent() const
{
    return false;
}

bool Event::isFocusEvent() const
{
    return false;
}

bool Event::isKeyboardEvent() const
{
    return false;
}

bool Event::isCompositionEvent() const
{
    return false;
}

#if ENABLE(CEHTML)
bool Event::isKeyEvent() const
{
    return false;
}
#endif

bool Event::isTouchEvent() const
{
    return false;
}

bool Event::isDragEvent() const
{
    return false;
}

bool Event::isClipboardEvent() const
{
    return false;
}

bool Event::storesResultAsString() const
{
    return false;
}

bool Event::isBeforeTextInsertedEvent() const
{
    return false;
}

void Event::storeResult(const String&)
{
}

void Event::setTarget(PassRefPtr<EventTarget> target)
{
    if (m_target == target)
        return;

    m_target = target;
    if (m_target)
        receivedTarget();
}

Vector<EventTarget*> Event::composedPath() const
{
    if (!m_eventPath)
        return Vector<EventTarget*>();
    return m_eventPath->computePathDisclosedToTarget(*m_target);
}

void Event::receivedTarget()
{
}

void Event::setUnderlyingEvent(PassRefPtr<Event> ue)
{
    // Prohibit creation of a cycle -- just do nothing in that case.
    for (Event* e = ue.get(); e; e = e->underlyingEvent())
        if (e == this)
            return;
    m_underlyingEvent = ue;
}

#if ENABLE(RMAI)
bool Event::isMOMEvent() const
{
    return false;
}

bool Event::isGalioConfigurationEvent() const
{
    return false;
}
#endif

} // namespace WebCore
