/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004-2017 Apple Inc. All rights reserved.
 * Copyright (C) 2006 Alexey Proskuryakov (ap@webkit.org)
 *           (C) 2007, 2008 Nikolas Zimmermann <zimmermann@kde.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *
 */

#ifndef EventTarget_h
#define EventTarget_h

#include "EventListenerMap.h"
#include "EventNames.h"
#include "EventTargetInterfaces.h"
#include "ScriptWrappable.h"
#include <wtf/Forward.h>
#include <wtf/HashMap.h>
#include <wtf/OwnPtr.h>
#include <wtf/text/AtomicStringHash.h>

namespace WebCore {

class AudioNode;
class AudioContext;
class AudioTrackList;
class DedicatedWorkerGlobalScope;
class DOMApplicationCache;
class DOMWindow;
class DOMWrapperWorld;
class Event;
class EventListener;
class EventSource;
class FileReader;
class FileWriter;
class IDBDatabase;
class IDBRequest;
class IDBTransaction;
class ScriptProcessorNode;
class LocalMediaStream;
class MediaController;
class MediaStream;
class MessagePort;
class Node;
class Notification;
class ScriptExecutionContext;
class SharedWorker;
class SharedWorkerGlobalScope;
class TextTrack;
class TextTrackCue;
class VideoTrackList;
class WebSocket;
class WebKitNamedFlow;
class Worker;
class XMLHttpRequest;
class XMLHttpRequestUpload;
#if ENABLE(RMAI)
class MOM;
class MOMContentItem;
class MOMDRMAgent;
class MOMDVR;
class MOMDevice;
class MOMGuideSearch;
class MOMNetworkInterface;
class MOMPackageManager;
class MOMRenderer;
class MOMStream;
class MOMVODManager;
class GalioConfigurationComponent;
#endif

typedef int ExceptionCode;

struct FiringEventIterator {
	WTF_MAKE_FAST_ALLOCATED;
public:
	FiringEventIterator(const AtomicString& eventType, size_t& iterator, size_t& end)
		: eventType(eventType)
		, iterator(iterator)
		, end(end)
	{
	}

	const AtomicString& eventType;
	size_t& iterator;
	size_t& end;
};
typedef Vector<FiringEventIterator, 1> FiringEventIteratorVector;

struct EventTargetData {
	WTF_MAKE_NONCOPYABLE(EventTargetData); WTF_MAKE_FAST_ALLOCATED;
public:
	EventTargetData();
	~EventTargetData();

	EventListenerMap eventListenerMap;
	bool isFiringEventListeners;
};

struct EventListenerOptions {
	EventListenerOptions(bool capture = false)
		: capture(capture)
	{ }

	bool capture;
};

struct AddEventListenerOptions : public EventListenerOptions {
	AddEventListenerOptions(bool capture = false, bool passive = false, bool once = false)
		: EventListenerOptions(capture)
		, passive(passive)
		, once(once)
	{ }

	bool passive;
	bool once;
};

enum EventTargetInterface {

#define DOM_EVENT_INTERFACE_DECLARE(name) name##EventTargetInterfaceType,
DOM_EVENT_TARGET_INTERFACES_FOR_EACH(DOM_EVENT_INTERFACE_DECLARE)
#undef DOM_EVENT_INTERFACE_DECLARE

};

class EventTarget : public ScriptWrappable {
public:
	ALWAYS_INLINE void ref() { refEventTarget(); }
	ALWAYS_INLINE void deref() { derefEventTarget(); }

#if ENABLE(RMAI)
	virtual MOM* toMOM() { return 0; }
	virtual MOMContentItem* toMOMContentItem() { return 0; }
	virtual MOMDRMAgent* toMOMDRMAgent() { return 0; }
	virtual MOMDVR* toMOMDVR() { return 0; }
	virtual MOMDevice* toMOMDevice() { return 0; }
	virtual MOMGuideSearch* toMOMGuideSearch() { return 0; }
	virtual MOMNetworkInterface* toMOMNetworkInterface() { return 0; }
	virtual MOMPackageManager* toMOMPackageManager() { return 0; }
	virtual MOMRenderer* toMOMRenderer() { return 0; }
	virtual MOMStream* toMOMStream() { return 0; }
	virtual MOMVODManager* toMOMVODManager() { return 0; }
	virtual GalioConfigurationComponent* toGalioConfigurationComponent() { return 0; }
#endif

	virtual EventTargetInterface eventTargetInterface() const = 0;
	virtual ScriptExecutionContext* scriptExecutionContext() const = 0;

	virtual Node* toNode();
	virtual DOMWindow* toDOMWindow();
	virtual bool isMessagePort() const;

	void addEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener>, bool useCapture = false);
	void removeEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener>, bool useCapture = false);
	void addEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener>, const AddEventListenerOptions&);
	void removeEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener>, const EventListenerOptions&);
	virtual bool addEventListener(const AtomicString& eventType, PassRefPtr<EventListener>, bool useCapture);
	virtual bool removeEventListener(const AtomicString& eventType, EventListener*, bool useCapture);
	virtual void removeAllEventListeners();
	virtual bool dispatchEvent(PassRefPtr<Event>);
	bool dispatchEventForBindings(PassRefPtr<Event>, ExceptionCode&); // DOM API
	virtual void uncaughtExceptionInEventHandler();

	// Used for legacy "onEvent" attribute APIs.
	bool setAttributeEventListener(const AtomicString& eventType, PassRefPtr<EventListener>, DOMWrapperWorld*);
	EventListener* attributeEventListener(const AtomicString& eventType, DOMWrapperWorld*);

	bool hasEventListeners();
	bool hasEventListeners(const AtomicString& eventType);
	bool hasCapturingEventListeners(const AtomicString& eventType);
	const EventListenerVector& getEventListeners(const AtomicString& eventType);

	bool fireEventListeners(Event*);
	bool isFiringEventListeners();

#if USE(JSC)
	void visitJSEventListeners(JSC::SlotVisitor&);
	void invalidateJSEventListeners(JSC::JSObject*);
#endif

protected:
	virtual ~EventTarget();

	virtual EventTargetData* eventTargetData() = 0;
	virtual EventTargetData* ensureEventTargetData() = 0;

private:
	virtual void refEventTarget() = 0;
	virtual void derefEventTarget() = 0;

	void fireEventListeners(Event&, EventListenerVector);

	friend class EventListenerIterator;
};

// FIXME: These macros should be split into separate DEFINE and DECLARE
// macros to avoid causing so many header includes.
#define DEFINE_ATTRIBUTE_EVENT_LISTENER(attribute) \
	ALWAYS_INLINE EventListener* on##attribute(DOMWrapperWorld* isolatedWorld) { return attributeEventListener(eventNames().attribute##Event, isolatedWorld); } \
	ALWAYS_INLINE void setOn##attribute(PassRefPtr<EventListener> listener, DOMWrapperWorld* isolatedWorld) { setAttributeEventListener(eventNames().attribute##Event, listener, isolatedWorld); } \

#define DECLARE_VIRTUAL_ATTRIBUTE_EVENT_LISTENER(attribute) \
	virtual EventListener* on##attribute(DOMWrapperWorld*); \
	virtual void setOn##attribute(PassRefPtr<EventListener> listener, DOMWrapperWorld*); \

#define DEFINE_VIRTUAL_ATTRIBUTE_EVENT_LISTENER(type, attribute) \
	EventListener* type::on##attribute(DOMWrapperWorld* isolatedWorld) { return attributeEventListener(eventNames().attribute##Event, isolatedWorld); } \
	void type::setOn##attribute(PassRefPtr<EventListener> listener, DOMWrapperWorld* isolatedWorld) { setAttributeEventListener(eventNames().attribute##Event, listener, isolatedWorld); } \

#define DEFINE_WINDOW_ATTRIBUTE_EVENT_LISTENER(attribute) \
	EventListener* on##attribute(DOMWrapperWorld* isolatedWorld) { return document()->getWindowAttributeEventListener(eventNames().attribute##Event, isolatedWorld); } \
	void setOn##attribute(PassRefPtr<EventListener> listener, DOMWrapperWorld* isolatedWorld) { document()->setWindowAttributeEventListener(eventNames().attribute##Event, listener, isolatedWorld); } \

#define DEFINE_MAPPED_ATTRIBUTE_EVENT_LISTENER(attribute, eventName) \
	EventListener* on##attribute(DOMWrapperWorld* isolatedWorld) { return attributeEventListener(eventNames().eventName##Event, isolatedWorld); } \
	void setOn##attribute(PassRefPtr<EventListener> listener, DOMWrapperWorld* isolatedWorld) { setAttributeEventListener(eventNames().eventName##Event, listener, isolatedWorld); } \

#define DECLARE_FORWARDING_ATTRIBUTE_EVENT_LISTENER(recipient, attribute) \
	EventListener* on##attribute(DOMWrapperWorld*); \
	void setOn##attribute(PassRefPtr<EventListener> listener, isolatedWorld);

#define DEFINE_FORWARDING_ATTRIBUTE_EVENT_LISTENER(type, recipient, attribute) \
	EventListener* type::on##attribute(DOMWrapperWorld* isolatedWorld) { return recipient ? recipient->attributeEventListener(eventNames().attribute##Event, isolatedWorld) : 0; } \
	void type::setOn##attribute(PassRefPtr<EventListener> listener, DOMWrapperWorld* isolatedWorld) { if (recipient) recipient->setAttributeEventListener(eventNames().attribute##Event, listener, isolatedWorld); }

#ifndef NDEBUG
void forbidEventDispatch();
void allowEventDispatch();
bool eventDispatchForbidden();
#else
inline void forbidEventDispatch() { }
inline void allowEventDispatch() { }
#endif

#if USE(JSC)
ALWAYS_INLINE void EventTarget::visitJSEventListeners(JSC::SlotVisitor& visitor)
{
	EventListenerIterator iterator(this);
	while (EventListener* listener = iterator.nextListener())
		listener->visitJSFunction(visitor);
}
#endif

ALWAYS_INLINE bool EventTarget::isFiringEventListeners()
{
	EventTargetData* d = eventTargetData();
	if (!d)
		return false;
    return d->isFiringEventListeners;
}

ALWAYS_INLINE bool EventTarget::hasEventListeners()
{
	EventTargetData* d = eventTargetData();
	if (!d)
		return false;
	return !d->eventListenerMap.isEmpty();
}

ALWAYS_INLINE bool EventTarget::hasEventListeners(const AtomicString& eventType)
{
	EventTargetData* d = eventTargetData();
	if (!d)
		return false;
	return d->eventListenerMap.contains(eventType);
}

inline bool EventTarget::hasCapturingEventListeners(const AtomicString& eventType)
{
	EventTargetData* d = eventTargetData();
	if (!d)
		return false;
	return d->eventListenerMap.containsCapturing(eventType);
}

} // namespace WebCore

#endif // EventTarget_h
