/*
 * Copyright (C) 2012 Victor Carbune (victor@rosedu.org)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "GenericEventQueue.h"

#include "Event.h"
#include <wtf/CurrentTime.h>
#include <wtf/MainThread.h>
#include <wtf/MemoryMonitor.h>


namespace WebCore {

// Fire timers for this length of time, and then quit to let the run loop process user input events.
// 100ms is about a perceptable delay in UI, so use a half of that as a threshold.
// This is to prevent UI freeze when there are too many timers or machine performance is low.
static const double maxDurationOfFiringTimers = 0.050;

PassOwnPtr<GenericEventQueue> GenericEventQueue::create(EventTarget* owner)
{
    return adoptPtr(new GenericEventQueue(owner));
}

GenericEventQueue::GenericEventQueue(EventTarget* owner)
: m_owner(owner)
, m_isClosed(false)
{
}

GenericEventQueue::~GenericEventQueue()
{
}

bool GenericEventQueue::enqueueEvent(PassRefPtr<Event> event)
{
    if (m_isClosed)
        return false;

    if (event->target() == m_owner)
        event->setTarget(nullptr);

    m_pendingEvents.append(event);
    pendingQueues().append(m_weakPtrFactory.createWeakPtr(this));
    if (!sharedTimer().isActive())
        sharedTimer().startOneShot(0);

    return true;
}

StaticTimer& GenericEventQueue::sharedTimer()
{
    ASSERT(isMainThread());
    static StaticTimer timer(GenericEventQueue::sharedTimerFired);
    return timer;
}

void GenericEventQueue::sharedTimerFired()
{
	Deque<WeakPtr<GenericEventQueue> >& queues = pendingQueues();
    ASSERT(!sharedTimer().isActive());
    ASSERT(!queues.isEmpty());

    double fireTime = WTF::monotonicallyIncreasingTime();
    double timeToQuit = fireTime + maxDurationOfFiringTimers;

    while (!queues.isEmpty()) {
        WeakPtr<GenericEventQueue> queue = queues.takeFirst();
        if (!queue)
            continue;
        queue->dispatchOneEvent();

        if(WTF::gLowMemory || timeToQuit <= WTF::monotonicallyIncreasingTime())
            break;
    }

    if(queues.isEmpty() || sharedTimer().isActive())
    	return;

    sharedTimer().startOneShot(0);
}

Deque<WeakPtr<GenericEventQueue> >& GenericEventQueue::pendingQueues()
{
    ASSERT(isMainThread());
    static Deque<WeakPtr<GenericEventQueue> > queues;
    return queues;
}

void GenericEventQueue::dispatchOneEvent()
{
    ASSERT(!m_pendingEvents.isEmpty());

    RefPtr<EventTarget> protect(m_owner);
    RefPtr<Event> event = m_pendingEvents.takeFirst();
    EventTarget* target = event->target() ? event->target() : m_owner;
    target->dispatchEvent(event.release());
}

void GenericEventQueue::close()
{
    m_isClosed = true;

    m_weakPtrFactory.revokeAll();
    m_pendingEvents.clear();
}

void GenericEventQueue::cancelAllEvents()
{
    m_weakPtrFactory.revokeAll();
    m_pendingEvents.clear();
}

bool GenericEventQueue::hasPendingEvents() const
{
    return !m_pendingEvents.isEmpty();
}

}
