/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2006, 2007 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef LiveNodeList_h
#define LiveNodeList_h

#include "CollectionIndexCache.h"
#include "CollectionType.h"
#include "Document.h"
#include "ElementTraversal.h"
#include "HTMLNames.h"
#include "NodeList.h"
#include <wtf/Forward.h>
#include <wtf/RefPtr.h>

namespace WebCore {

class Element;
class Node;

static bool shouldInvalidateTypeOnAttributeChange(NodeListInvalidationType, const QualifiedName&);

class LiveNodeList : public NodeList {
public:
    LiveNodeList(PassRefPtr<ContainerNode> ownerNode, NodeListInvalidationType);
    virtual ~LiveNodeList();

    virtual Node* namedItem(const AtomicString&) const OVERRIDE FINAL;

    virtual bool nodeMatches(Element*) const = 0;
    virtual bool isRootedAtDocument() const = 0;

	ALWAYS_INLINE NodeListInvalidationType invalidationType() const { return static_cast<NodeListInvalidationType>(m_invalidationType); }
	ContainerNode& ownerNode() const { return *m_ownerNode; }
    ALWAYS_INLINE void invalidateCacheForAttribute(const QualifiedName* attrName) const
    {
        if (!attrName || shouldInvalidateTypeOnAttributeChange(invalidationType(), *attrName))
            invalidateCache(document());
    }
    virtual void invalidateCache(Document*) const = 0;

    bool isRegisteredForInvalidationAtDocument() const { return m_isRegisteredForInvalidationAtDocument; }
    void setRegisteredForInvalidationAtDocument(bool f) { m_isRegisteredForInvalidationAtDocument = f; }

protected:
    Document* document() const { return m_ownerNode->document(); }

private:
    virtual bool isLiveNodeList() const OVERRIDE FINAL { return true; }

    ContainerNode& rootNode() const;

    Element* iterateForPreviousElement(Element* current) const;

    RefPtr<ContainerNode> m_ownerNode;

    const unsigned m_invalidationType;
    bool m_isRegisteredForInvalidationAtDocument;
};

template <class NodeListType>
class CachedLiveNodeList : public LiveNodeList {
public:
    virtual ~CachedLiveNodeList();

    virtual unsigned length() const OVERRIDE FINAL;
    virtual Node* item(unsigned offset) const OVERRIDE FINAL;

    // For CollectionIndexCache
    Element* collectionFirst() const;
    Element* collectionLast() const;
    Element* collectionTraverseForward(Element&, unsigned count, unsigned& traversedCount) const;
    Element* collectionTraverseBackward(Element&, unsigned count) const;
    bool collectionCanTraverseBackward() const { return true; }
    void willValidateIndexCache() const;

    virtual void invalidateCache(Document*) const;
    virtual size_t memoryCost() const OVERRIDE;

protected:
    CachedLiveNodeList(PassRefPtr<ContainerNode> rootNode, NodeListInvalidationType);

private:
    ContainerNode& rootNode() const;

    mutable CollectionIndexCache<NodeListType, Element> m_indexCache;
};

ALWAYS_INLINE bool shouldInvalidateTypeOnAttributeChange(NodeListInvalidationType type, const QualifiedName& attrName)
{
    switch (type) {
    case InvalidateOnClassAttrChange:
        return attrName == HTMLNames::classAttr;
    case InvalidateOnNameAttrChange:
        return attrName == HTMLNames::nameAttr;
    case InvalidateOnIdNameAttrChange:
        return attrName == HTMLNames::idAttr || attrName == HTMLNames::nameAttr;
    case InvalidateOnForTypeAttrChange:
        return attrName == HTMLNames::forAttr || attrName == HTMLNames::typeAttr;
    case InvalidateForFormControls:
        return attrName == HTMLNames::nameAttr || attrName == HTMLNames::idAttr || attrName == HTMLNames::forAttr
            || attrName == HTMLNames::formAttr || attrName == HTMLNames::typeAttr;
    case InvalidateOnHRefAttrChange:
        return attrName == HTMLNames::hrefAttr;
    case InvalidateOnItemAttrChange:
#if ENABLE(MICRODATA)
        return attrName == HTMLNames::itemscopeAttr || attrName == HTMLNames::itempropAttr
			|| attrName == HTMLNames::itemtypeAttr || attrName == HTMLNames::itemrefAttr || attrName == HTMLNames::idAttr;
#endif // Intentionally fall through
    case DoNotInvalidateOnAttributeChanges:
        return false;
    case InvalidateOnAnyAttrChange:
        return true;
    }
    return false;
}

template <class NodeListType>
CachedLiveNodeList<NodeListType>::CachedLiveNodeList(PassRefPtr<ContainerNode> ownerNode, NodeListInvalidationType invalidationType)
    : LiveNodeList(ownerNode, invalidationType)
{
}

template <class NodeListType>
CachedLiveNodeList<NodeListType>::~CachedLiveNodeList()
{
    if (m_indexCache.hasValidCache())
        document()->unregisterNodeListForInvalidation(*this);
}

template <class NodeListType>
inline ContainerNode& CachedLiveNodeList<NodeListType>::rootNode() const
{
    if (static_cast<const NodeListType&>(*this).isRootedAtDocument() && ownerNode().inDocument())
        return *ownerNode().document();

    return ownerNode();
}

template <class NodeListType>
Element* CachedLiveNodeList<NodeListType>::collectionFirst() const
{
    auto& root = rootNode();
    Element* element = ElementTraversal::firstWithin(&root);
    while (element && !static_cast<const NodeListType*>(this)->nodeMatches(element))
        element = ElementTraversal::next(element, &root);
    return element;
}

template <class NodeListType>
Element* CachedLiveNodeList<NodeListType>::collectionLast() const
{
    auto& root = rootNode();
    Element* element = ElementTraversal::lastWithin(&root);
    while (element && !static_cast<const NodeListType*>(this)->nodeMatches(element))
        element = ElementTraversal::previous(element, &root);
    return element;
}

template <class NodeListType>
inline Element* nextMatchingElement(const NodeListType* nodeList, Element* current, ContainerNode& root)
{
    do {
        current = ElementTraversal::next(current, &root);
    } while (current && !static_cast<const NodeListType*>(nodeList)->nodeMatches(current));
    return current;
}

template <class NodeListType>
Element* CachedLiveNodeList<NodeListType>::collectionTraverseForward(Element& current, unsigned count, unsigned& traversedCount) const
{
    auto& root = rootNode();
    Element* element = &current;
    for (traversedCount = 0; traversedCount < count; ++traversedCount) {
        element = nextMatchingElement(static_cast<const NodeListType*>(this), element, root);
        if (!element)
            return 0;
    }
    return element;
}

template <class NodeListType>
inline Element* previousMatchingElement(const NodeListType* nodeList, Element* current, ContainerNode& root)
{
    do {
        current = ElementTraversal::previous(current, &root);
    } while (current && !nodeList->nodeMatches(current));
    return current;
}

template <class NodeListType>
Element* CachedLiveNodeList<NodeListType>::collectionTraverseBackward(Element& current, unsigned count) const
{
    auto& root = rootNode();
    Element* element = &current;
    for (; count; --count) {
        element = previousMatchingElement(static_cast<const NodeListType*>(this), element, root);
        if (!element)
            return 0;
    }
    return element;
}
template <class NodeListType>
void CachedLiveNodeList<NodeListType>::willValidateIndexCache() const
{
    document()->registerNodeListForInvalidation(const_cast<NodeListType&>(static_cast<const NodeListType&>(*this)));
}

template <class NodeListType>
void CachedLiveNodeList<NodeListType>::invalidateCache(Document* document) const
{
    if (!m_indexCache.hasValidCache())
        return;
    document->unregisterNodeListForInvalidation(const_cast<NodeListType&>(static_cast<const NodeListType&>(*this)));
    m_indexCache.invalidate();
}

template <class NodeListType>
unsigned CachedLiveNodeList<NodeListType>::length() const
{
    return m_indexCache.nodeCount(static_cast<const NodeListType&>(*this));
}

template <class NodeListType>
Node* CachedLiveNodeList<NodeListType>::item(unsigned offset) const
{
    return m_indexCache.nodeAt(static_cast<const NodeListType&>(*this), offset);
}

template <class NodeListType>
size_t CachedLiveNodeList<NodeListType>::memoryCost() const
{
    return m_indexCache.memoryCost();
}

} // namespace WebCore

#endif // LiveNodeList_h
