/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ShadowRoot_h
#define ShadowRoot_h

#include "ContainerNode.h"
#include "Document.h"
#include "DocumentFragment.h"
#include "Element.h"
#include "ExceptionCode.h"
#include "SlotAssignment.h"
#include "TreeScope.h"

namespace WebCore {

class AuthorStyleSheets;
class HTMLSlotElement;
class SlotAssignment;
class StyleResolver;

class ShadowRoot FINAL : public DocumentFragment, public TreeScope {
	WTF_MAKE_FAST_ALLOCATED;
public:
    enum class Type : uint8_t {
        UserAgent = 0,
        Closed,
        Open,
    };

    static PassRefPtr<ShadowRoot> create(Document* document, Type type)
    {
        return adoptRef(new ShadowRoot(document, type));
    }

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    static PassRefPtr<ShadowRoot> create(Document* document, PassOwnPtr<SlotAssignment> assignment)
    {
        return adoptRef(new ShadowRoot(document, assignment));
    }
#endif

    virtual ~ShadowRoot();

    StyleResolver& styleResolver();
    AuthorStyleSheets& authorStyleSheets();

    void updateStyle();
    void resetStyleResolver();

    bool resetStyleInheritance() const;
    void setResetStyleInheritance(bool);

    Element* host() const { return m_host; }
    void setHost(Element* host) { m_host = host; }

    String innerHTML() const;
    void setInnerHTML(const String&, ExceptionCode&);

    Element* activeElement() const;

    Type type() const { return m_type; }

    void removeAllEventListeners();

    using TreeScope::getElementById;

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    HTMLSlotElement* findAssignedSlot(const Node&);

    void addSlotElementByName(const AtomicString&, HTMLSlotElement&);
    void removeSlotElementByName(const AtomicString&, HTMLSlotElement&);

    void invalidateSlotAssignments();
    void invalidateDefaultSlotAssignments();

    const Vector<Node*>* assignedNodesForSlot(const HTMLSlotElement&);
#endif

protected:
    explicit ShadowRoot(Document*, Type);

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    ShadowRoot(Document*, PassOwnPtr<SlotAssignment>);
#endif

    // FIXME: This shouldn't happen. https://bugs.webkit.org/show_bug.cgi?id=88834
    bool isOrphan() const { return !m_host; }

private:
    virtual bool childTypeAllowed(NodeType) const OVERRIDE;

    virtual RefPtr<Node> cloneNodeInternal(Document&, CloningOperation) OVERRIDE;

    bool m_resetStyleInheritance : 1;
    Type m_type;

    Element* m_host;

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    OwnPtr<SlotAssignment> m_slotAssignment;
#endif

    OwnPtr<StyleResolver> m_styleResolver;
    OwnPtr<AuthorStyleSheets> m_authorStyleSheets;
};

inline Element* ShadowRoot::activeElement() const
{
    return treeScope()->focusedElement();
}

inline bool isShadowRoot(const Node& node)
{
	return node.isShadowRoot();
}

inline ShadowRoot* toShadowRoot(Node* node)
{
    ASSERT(!node || node->isShadowRoot());
    return static_cast<ShadowRoot*>(node);
}

inline const ShadowRoot* toShadowRoot(const Node* node)
{
    ASSERT(!node || node->isShadowRoot());
    return static_cast<const ShadowRoot*>(node);
}

inline ShadowRoot& toShadowRoot(Node& node)
{
    ASSERT(node.isShadowRoot());
    return static_cast<ShadowRoot&>(node);
}

inline const ShadowRoot& toShadowRoot(const Node& node)
{
    ASSERT(node.isShadowRoot());
    return static_cast<const ShadowRoot&>(node);
}

inline ShadowRoot* Node::shadowRoot() const
{
    if (!isElementNode())
        return 0;
    return toElement(this)->shadowRoot();
}

inline ContainerNode* Node::parentOrShadowHostNode() const
{
    ASSERT(isMainThreadOrGCThread());
    if (isShadowRoot())
        return toShadowRoot(this)->host();
    return parentNode();
}

inline bool hasShadowRootParent(const Node* node)
{
    return node->parentNode() && node->parentNode()->isShadowRoot();
}

} // namespace

#endif
