/*
 * Copyright (C) 2010, 2011 Nokia Corporation and/or its subsidiary(-ies)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "HTMLDetailsElement.h"

#if ENABLE(DETAILS_ELEMENT)
#include "ElementIterator.h"
#include "HTMLSlotElement.h"
#include "HTMLNames.h"
#include "HTMLSummaryElement.h"
#include "LocalizedStrings.h"
#include "MouseEvent.h"
#include "RenderBlockFlow.h"
#include "ShadowRoot.h"
#include "Text.h"
#include <wtf/NeverDestroyed.h>

namespace WebCore {

using namespace HTMLNames;

static const AtomicString& summarySlotName()
{
    static NeverDestroyed<AtomicString> summarySlot("summarySlot");
    return summarySlot;
}

static AtomicString slotNameFunction(const Node& child)
{
    auto& parent = *child.parentNode();
    ASSERT(is<HTMLDetailsElement>(parent));
    auto& details = toHTMLDetailsElement(parent);

    // The first summary child gets assigned to the summary slot.
    if (isHTMLSummaryElement(child)) {
        if (&child == childrenOfType<HTMLSummaryElement>(&details).first())
            return summarySlotName();
    }
    // Everything else is assigned to the default slot if details is open.
    if (details.isOpen())
        return SlotAssignment::defaultSlotName();

    // Otherwise don't render the content.
    return nullAtom;
};

PassRefPtr<HTMLDetailsElement> HTMLDetailsElement::create(const QualifiedName& tagName, Document* document)
{
    RefPtr<HTMLDetailsElement> details = adoptRef(new HTMLDetailsElement(tagName, document));
    details->addShadowRoot(ShadowRoot::create(document, adoptPtr(new SlotAssignment(slotNameFunction))));
    return details.release();
}

inline HTMLDetailsElement::HTMLDetailsElement(const QualifiedName& tagName, Document* document)
    : HTMLElement(tagName, document)
    , m_isOpen(false)
	, m_summarySlot(0)
	, m_defaultSummary(0)
{
    ASSERT(hasTagName(detailsTag));
}

RenderElement* HTMLDetailsElement::createElementRenderer(PassRefPtr<RenderStyle> style, const RenderTreePosition&)
{
    return new RenderBlockFlow(*this, style);
}

void HTMLDetailsElement::didAddUserAgentShadowRoot(ShadowRoot* root)
{
    auto summarySlot = HTMLSlotElement::create(slotTag, document());
    summarySlot->setAttribute(nameAttr, summarySlotName());
    m_summarySlot = summarySlot.get();

    auto defaultSummary = HTMLSummaryElement::create(summaryTag, document());
    defaultSummary->appendChild(Text::create(document(), defaultDetailsSummaryText()), ASSERT_NO_EXCEPTION);
    m_defaultSummary = defaultSummary.get();

    summarySlot->appendChild(defaultSummary);
    root->appendChild(summarySlot);

    auto defaultSlot = HTMLSlotElement::create(slotTag, document());
    root->appendChild(defaultSlot);
}

bool HTMLDetailsElement::isActiveSummary(const HTMLSummaryElement& summary) const
{
    if (!m_summarySlot->assignedNodes())
        return &summary == m_defaultSummary;

    if (summary.parentNode() != this)
        return false;

    auto* slot = shadowRoot()->findAssignedSlot(summary);
    if (!slot)
        return false;
    return slot == m_summarySlot;
}

void HTMLDetailsElement::parseAttribute(const Attribute& attribute)
{
    if (attribute.name() == openAttr) {
        bool oldValue = m_isOpen;
        m_isOpen =  !attribute.isNull();
        if (oldValue != m_isOpen)
            shadowRoot()->invalidateSlotAssignments();
    } else
        HTMLElement::parseAttribute(attribute);
}

void HTMLDetailsElement::toggleOpen()
{
    setAttribute(openAttr, m_isOpen ? nullAtom : emptyAtom);
}

}

#endif
