/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2009 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef HTMLElement_h
#define HTMLElement_h

#include "StyledElement.h"

namespace WebCore {

class DocumentFragment;
class FormNamedItem;
class HTMLCollection;
class HTMLFormElement;

#if ENABLE(MICRODATA)
class HTMLPropertiesCollection;
class MicroDataItemValue;
#endif

enum TranslateAttributeMode {
    TranslateAttributeYes,
    TranslateAttributeNo,
    TranslateAttributeInherit
};

class HTMLElement : public StyledElement {
public:
    static PassRefPtr<HTMLElement> create(const QualifiedName& tagName, Document*);

    virtual String title() const;

    virtual short tabIndex() const;
    void setTabIndex(int);

    void setInnerText(const String&, ExceptionCode&);
    void setOuterText(const String&, ExceptionCode&);

    Element* insertAdjacentElement(const String& where, Element* newChild, ExceptionCode&);
    void insertAdjacentHTML(const String& where, const String& html, ExceptionCode&);
    void insertAdjacentText(const String& where, const String& text, ExceptionCode&);

    virtual bool supportsFocus() const;

    String contentEditable() const;
    void setContentEditable(const String&, ExceptionCode&);

    virtual bool draggable() const;
    void setDraggable(bool);

    bool spellcheck() const;
    void setSpellcheck(bool);

    bool translate() const;
    void setTranslate(bool);

    void click();

    virtual void accessKeyAction(bool sendMouseEvents);

    bool ieForbidsInsertHTML() const;

    virtual bool rendererIsNeeded(const RenderStyle&);
    virtual RenderElement* createElementRenderer(PassRefPtr<RenderStyle>, const RenderTreePosition&);

    ALWAYS_INLINE HTMLFormElement* form() const { return virtualForm(); }

    HTMLFormElement* findFormAncestor() const;

    bool hasDirectionAuto() const;
    TextDirection directionalityIfhasDirAutoAttribute(bool& isAuto) const;

#if ENABLE(MICRODATA)
    void setItemValue(const String&, ExceptionCode&);
    PassRefPtr<MicroDataItemValue> itemValue() const;
    PassRefPtr<HTMLPropertiesCollection> properties();
    void getItemRefElements(Vector<HTMLElement*>&);
#endif

    virtual bool isHTMLUnknownElement() const { return false; }
    virtual bool isTextControlInnerTextElement() const { return false; }
    virtual bool canHaveUserAgentShadowRoot() const { return false; }

    virtual bool isLabelable() const { return false; }
    virtual bool isHTMLTableElement() const { return false; }
    virtual FormNamedItem* asFormNamedItem() { return 0; }

    bool hasTagName(const HTMLQualifiedName& name) const { return hasLocalName(name.localName()); }

protected:
    HTMLElement(const QualifiedName& tagName, Document*, ConstructionType);

    void addHTMLLengthToStyle(MutableStylePropertySet*, CSSPropertyID, const String& value);
    void addHTMLColorToStyle(MutableStylePropertySet*, CSSPropertyID, const String& color);

    void applyAlignmentAttributeToStyle(const Attribute&, MutableStylePropertySet*);
    void applyBorderAttributeToStyle(const Attribute&, MutableStylePropertySet*);

    virtual bool matchesReadWritePseudoClass() const;
    virtual void parseAttribute(const Attribute&) OVERRIDE;
    virtual bool isPresentationAttribute(const QualifiedName&) const OVERRIDE;
    virtual void collectStyleForPresentationAttribute(const Attribute&, MutableStylePropertySet*) OVERRIDE;

    virtual void childrenChanged(const ChildChange&) OVERRIDE;
	void calculateAndAdjustDirectionality();

    virtual bool isURLAttribute(const Attribute&) const OVERRIDE;

private:
    virtual String nodeName() const;

    void mapLanguageAttributeToLocale(const Attribute&, MutableStylePropertySet*);

    virtual HTMLFormElement* virtualForm() const;

    Node* insertAdjacent(const String& where, Node* newChild, ExceptionCode&);

    void dirAttributeChanged(const Attribute&);
    void adjustDirectionalityIfNeededAfterChildAttributeChanged(Element* child);
    void adjustDirectionalityIfNeededAfterChildrenChanged(Element* beforeChange, ChildChangeType);
    TextDirection directionality(Node** strongDirectionalityTextNode= 0) const;

    TranslateAttributeMode translateAttributeMode() const;

    AtomicString eventNameForAttributeName(const QualifiedName& attrName) const;

#if ENABLE(MICRODATA)
    virtual String itemValueText() const;
    virtual void setItemValueText(const String&, ExceptionCode&);
#endif
};

ALWAYS_INLINE HTMLElement* toHTMLElement(Node* node)
{
    ASSERT(!node || node->isHTMLElement());
    return static_cast<HTMLElement*>(node);
}

ALWAYS_INLINE const HTMLElement* toHTMLElement(const Node* node)
{
    ASSERT(!node || node->isHTMLElement());
    return static_cast<const HTMLElement*>(node);
}

ALWAYS_INLINE HTMLElement& toHTMLElement(Node& node)
{
    ASSERT(node.isHTMLElement());
    return static_cast<HTMLElement&>(node);
}

ALWAYS_INLINE const HTMLElement& toHTMLElement(const Node& node)
{
    ASSERT(node.isHTMLElement());
    return static_cast<const HTMLElement&>(node);
}

// This will catch anyone doing an unnecessary cast.
void toHTMLElement(const HTMLElement*);
void toHTMLElement(const HTMLElement&);

ALWAYS_INLINE HTMLElement::HTMLElement(const QualifiedName& tagName, Document* document, ConstructionType type = CreateHTMLElement)
    : StyledElement(tagName, document, type)
{
    ASSERT(tagName.localName().impl());
}

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLElement, ArgType> {
    static bool is(ArgType& node) { return node.isHTMLElement(); }
};

#if !ENABLE(MICRODATA)
inline bool HTMLElement::isURLAttribute(const Attribute& attribute) const
{
    UNUSED_PARAM(attribute);
    return false;
}
#endif

inline bool Node::hasTagName(const HTMLQualifiedName& name) const
{
    return isHTMLElement() && toHTMLElement(*this).hasTagName(name);
}

} // namespace WebCore

#endif // HTMLElement_h
