/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2008, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef HTMLImageElement_h
#define HTMLImageElement_h

#include "FormNamedItem.h"
#include "GraphicsTypes.h"
#include "HTMLElement.h"
#include "HTMLImageLoader.h"

#if ENABLE(WEBGL)
#include "WebGLTexture.h"
#endif

namespace WebCore {

class HTMLFormElement;
class HTMLPictureElement;
struct ImageCandidate;

class HTMLImageElement : public HTMLElement, public FormNamedItem {
    friend class HTMLFormElement;
public:
    static PassRefPtr<HTMLImageElement> create(Document*);
    static PassRefPtr<HTMLImageElement> create(const QualifiedName&, Document*, HTMLFormElement*);
    static PassRefPtr<HTMLImageElement> createForJSConstructor(Document*, Optional<unsigned> width, Optional<unsigned> height);

    virtual ~HTMLImageElement();

    unsigned width(bool ignorePendingStylesheets = false);
    unsigned height(bool ignorePendingStylesheets = false);

    int naturalWidth() const;
    int naturalHeight() const;
    const AtomicString& currentSrc() const { return m_currentSrc; }

    bool isServerMap() const;

    bool isAccelerated();

#if ENABLE(WEBGL)
    PassRefPtr<WebGLTexture> webglTexture();
    ALWAYS_INLINE PassRefPtr<WebGLTexture> texture() { return webglTexture(); }
    void clearTexture();
#endif

    String altText() const;

    CompositeOperator compositeOperator() const { return m_compositeOperator; }

    ALWAYS_INLINE CachedImage* cachedImage() const { return m_imageLoader.image(); }
    ALWAYS_INLINE void setCachedImage(CachedImage* i) { m_imageLoader.setImage(i); };

    ALWAYS_INLINE void setLoadManually(bool loadManually) { m_imageLoader.setLoadManually(loadManually); }

    bool matchesLowercasedUsemap(const AtomicStringImpl&) const;

    const AtomicString& alt() const;

    void setHeight(unsigned);

    KURL src() const;
    void setSrc(const String&);

    void setCrossOrigin(const AtomicString&);
    String crossOrigin() const;

    void setWidth(unsigned);

    int x() const;
    int y() const;

    bool complete() const;

    ALWAYS_INLINE bool hasPendingActivity() const { return m_imageLoader.hasPendingActivity(); }

    virtual bool canContainRangeEndPoint() const { return false; }

    virtual const AtomicString& imageSourceURL() const OVERRIDE;

    bool hasShadowControls() const { return m_experimentalImageMenuEnabled; }

    HTMLPictureElement* pictureElement() const;
    void setPictureElement(HTMLPictureElement*);

protected:
    HTMLImageElement(const QualifiedName&, Document*, HTMLFormElement* = 0);

    virtual void didMoveToNewDocument(Document* oldDocument) OVERRIDE;

    virtual void parseAttribute(const Attribute&) OVERRIDE;

private:
    virtual bool isPresentationAttribute(const QualifiedName&) const OVERRIDE;
    virtual void collectStyleForPresentationAttribute(const Attribute&, MutableStylePropertySet*) OVERRIDE;

    virtual void didAttachRenderers() OVERRIDE;
    virtual RenderElement* createElementRenderer(PassRefPtr<RenderStyle>, const RenderTreePosition&);
    void setBestFitURLAndDPRFromImageCandidate(const ImageCandidate&);

    virtual bool canStartSelection() const;

    virtual bool isURLAttribute(const Attribute&) const OVERRIDE;

    virtual bool draggable() const;

    virtual void addSubresourceAttributeURLs(ListHashSet<KURL>&) const;

    virtual InsertionNotificationRequest insertedInto(ContainerNode*) OVERRIDE;
    virtual void removedFrom(ContainerNode*) OVERRIDE;
    virtual bool shouldRegisterAsExtraNamedItem() const OVERRIDE { return true; }

    virtual bool isFormAssociatedElement() OVERRIDE FINAL { return false; }
    virtual FormNamedItem* asFormNamedItem() OVERRIDE FINAL { return this; }
    virtual HTMLElement* asHTMLElement() OVERRIDE FINAL { return this; }

#if ENABLE(MICRODATA)
    virtual String itemValueText() const OVERRIDE;
    virtual void setItemValueText(const String&, ExceptionCode&) OVERRIDE;
#endif

    void selectImageSource();

    ImageCandidate bestFitSourceFromPictureElement();

    HTMLImageLoader m_imageLoader;
    HTMLFormElement* m_form;

    CompositeOperator m_compositeOperator;
    AtomicString m_bestFitImageURL;
    AtomicString m_currentSrc;
    float m_imageDevicePixelRatio;
    AtomicString m_lowercasedUsemap;
    bool m_experimentalImageMenuEnabled;

#if ENABLE(IMAGE_CONTROLS)
    void updateImageControls();
    void createImageControls();
    void destroyImageControls();
    bool hasImageControls() const;
    virtual bool childShouldCreateRenderer(const Node&) const override;
#endif

    friend class HTMLPictureElement;
};

inline bool isHTMLImageElement(const Node* node)
{
    return node->hasTagName(HTMLNames::imgTag);
}

inline bool isHTMLImageElement(const Element* element)
{
    return element->hasTagName(HTMLNames::imgTag);
}

inline bool isHTMLImageElement(const Node& node)
{
    return node.hasTagName(HTMLNames::imgTag);
}

inline bool isHTMLImageElement(const Element& element)
{
    return element.hasTagName(HTMLNames::imgTag);
}

inline HTMLImageElement* toHTMLImageElement(Node* node)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!node || isHTMLImageElement(node));
    return static_cast<HTMLImageElement*>(node);
}

inline HTMLImageElement& toHTMLImageElement(Node& node)
{
    ASSERT_WITH_SECURITY_IMPLICATION(isHTMLImageElement(node));
    return static_cast<HTMLImageElement&>(node);
}

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLImageElement, ArgType> {
    static bool is(ArgType& node) { return isHTMLImageElement(node); }
};

} //namespace

#endif
