/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef HTMLObjectElement_h
#define HTMLObjectElement_h

#include "FormAssociatedElement.h"
#include "HTMLPlugInImageElement.h"

namespace JSC
{
class ExecState;
}

namespace WebCore {

class HTMLFormElement;
class PluginView;
class JSHTMLObjectElement;

class HTMLObjectElement : public HTMLPlugInImageElement, public FormAssociatedElement {

public:
	class JSPropertyDelegate
	{
	public:
		virtual ~JSPropertyDelegate() { }

		virtual bool jsGet(JSC::ExecState*, JSC::PropertyName, JSC::PropertySlot&, JSHTMLObjectElement*) = 0;
		virtual bool jsPut(JSC::ExecState*, JSC::PropertyName, JSC::JSValue, JSC::PutPropertySlot&, JSHTMLObjectElement*) = 0;

		virtual void updateNativePlayerID() { }
		virtual void resetEndedEventDispatched() { }

	protected:
		JSPropertyDelegate() { }
	};

public:
    static PassRefPtr<HTMLObjectElement> create(const QualifiedName&, Document*, HTMLFormElement*, bool createdByParser);
    virtual ~HTMLObjectElement();

    bool isDocNamedItem() const { return m_docNamedItem; }

    const String& classId() const { return m_classId; }

    bool containsJavaApplet() const;

    bool jsGet(JSC::ExecState*, JSC::PropertyName, JSC::PropertySlot&, JSHTMLObjectElement*);
    bool jsPut(JSC::ExecState*, JSC::PropertyName, JSC::JSValue, JSC::PutPropertySlot&, JSHTMLObjectElement*);

    virtual bool useFallbackContent() const { return m_useFallbackContent; }
    virtual void renderFallbackContent();

    virtual bool willValidate() const OVERRIDE { return false; }

    // Implementations of FormAssociatedElement
    HTMLFormElement* form() const { return FormAssociatedElement::form(); }

    virtual bool isFormControlElement() const { return false; }

    virtual bool isEnumeratable() const { return true; }
    virtual bool appendFormData(FormDataList&, bool);

    // Implementations of constraint validation API.
    // Note that the object elements are always barred from constraint validation.
    virtual String validationMessage() const OVERRIDE { return String(); }
    bool checkValidity() { return true; }
    virtual void setCustomValidity(const String&) OVERRIDE { }

    using Node::ref;
    using Node::deref;

    virtual bool canContainRangeEndPoint() const { return useFallbackContent(); }

    bool hasFallbackContent() const;

    // FIXME: This function should not deal with url or serviceType
    // so that we can better share code between <object> and <embed>.
    void parametersForPlugin(Vector<String>& paramNames, Vector<String>& paramValues, String& url, String& serviceType);

protected:
    HTMLObjectElement(const QualifiedName&, Document*, HTMLFormElement*, bool createdByParser);

    virtual void updateWidget(PluginCreationOption);

    virtual void removedFrom(ContainerNode*) OVERRIDE;

    virtual void attributeChanged(const QualifiedName&, const AtomicString& oldValue, const AtomicString& newValue);

    virtual void createJSPropertyDelegate() {}

    virtual void didMoveToNewDocument(Document* oldDocument) OVERRIDE;

private:
    virtual void parseAttribute(const Attribute&) OVERRIDE;
    virtual bool isPresentationAttribute(const QualifiedName&) const OVERRIDE;
    virtual void collectStyleForPresentationAttribute(const Attribute&, MutableStylePropertySet*) OVERRIDE;

    virtual InsertionNotificationRequest insertedInto(ContainerNode*) OVERRIDE;
    virtual void finishedInsertingSubtree() OVERRIDE FINAL;

    virtual bool rendererIsNeeded(const RenderStyle&);

    virtual void childrenChanged(const ChildChange&) OVERRIDE;

    virtual bool isURLAttribute(const Attribute&) const OVERRIDE;
    virtual const AtomicString& imageSourceURL() const OVERRIDE;

    virtual RenderWidget* renderWidgetLoadingPlugin() const;

    virtual void addSubresourceAttributeURLs(ListHashSet<KURL>&) const;

    void updateDocNamedItem();
    
    bool shouldAllowQuickTimeClassIdQuirk();
    bool hasValidClassId();

    virtual void refFormAssociatedElement() { ref(); }
    virtual void derefFormAssociatedElement() { deref(); }
    virtual HTMLFormElement* virtualForm() const;

    virtual FormNamedItem* asFormNamedItem() OVERRIDE FINAL { return this; }
    virtual HTMLElement* asHTMLElement() OVERRIDE FINAL { return this; }

#if ENABLE(MICRODATA)
    virtual String itemValueText() const OVERRIDE;
    virtual void setItemValueText(const String&, ExceptionCode&) OVERRIDE;
#endif

    String m_classId;
    bool m_docNamedItem : 1;
    bool m_useFallbackContent : 1;

protected:
    OwnPtr<JSPropertyDelegate> m_jsPropertyDelegate;
};

inline bool isHTMLObjectElement(const Node* node)
{
    ASSERT(node);
    return node->hasTagName(HTMLNames::objectTag);
}

inline bool isHTMLObjectElement(const Element* element)
{
    ASSERT(element);
    return element->hasTagName(HTMLNames::objectTag);
}

inline HTMLObjectElement* toHTMLObjectElement(Node* node)
{
    ASSERT(!node || isHTMLObjectElement(node));
    return static_cast<HTMLObjectElement*>(node);
}

}

#endif
