/*
 * Copyright (C) 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2009 Torch Mobile, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "CachedFont.h"

#if !PLATFORM(WIN_CAIRO) && !PLATFORM(WX)
#define STORE_FONT_CUSTOM_PLATFORM_DATA
#endif

#include "CachedResourceClient.h"
#include "CachedResourceClientWalker.h"
#include "CachedResourceLoader.h"
#include "FontDescription.h"
#include "FontPlatformData.h"
#include "MemoryCache.h"
#include "ResourceBuffer.h"
#include "TextResourceDecoder.h"
#include "WOFFFileFormat.h"
#include "WOFF2FileFormat.h"
#include <wtf/Vector.h>

#ifdef STORE_FONT_CUSTOM_PLATFORM_DATA
#include "FontCustomPlatformData.h"
#endif

#if ENABLE(SVG_FONTS)
#include "NodeList.h"
#include "SVGDocument.h"
#include "SVGElement.h"
#include "SVGFontElement.h"
#include "SVGGElement.h"
#include "SVGNames.h"
#endif

namespace WebCore {

CachedFont::CachedFont(CachedResourceRequest& request, Type type)
    : CachedResource(request, type)
    , m_loadInitiated(false)
	, m_hasCreatedFontDataWrappingResource(false)
	, m_fontPlatformDataDeletionTimer(this, &CachedFont::fontPlatformDataDeletionTimerCallback)
{
}

CachedFont::~CachedFont()
{
}

void CachedFont::load(CachedResourceLoader*)
{
    // Don't load the file yet.  Wait for an access before triggering the load.
    setLoading(true);
}

void CachedFont::didAddClient(CachedResourceClient* client)
{
	if(m_fontPlatformDataDeletionTimer.isActive())
		m_fontPlatformDataDeletionTimer.stop();

    ASSERT(client->resourceClientType() == CachedFontClient::expectedType());

	CachedResource::didAddClient(client);

    if (!isLoading())
        static_cast<CachedFontClient*>(client)->fontLoaded(this);
}

void CachedFont::finishLoading(ResourceBuffer* data)
{
    m_data = data;
    setEncodedSize(m_data.get() ? m_data->size() : 0);
    setLoading(false);
    checkNotify();
}

void CachedFont::beginLoadIfNeeded(CachedResourceLoader* dl)
{
    if (!m_loadInitiated) {
        m_loadInitiated = true;
        CachedResource::load(dl);
    }
}

bool CachedFont::ensureCustomFontData(const AtomicString&)
{
    return ensureCustomFontData(m_data->sharedBuffer());
}

bool CachedFont::ensureCustomFontData(SharedBuffer* data)
{
#ifdef STORE_FONT_CUSTOM_PLATFORM_DATA
    if (!m_fontData && !errorOccurred() && !isLoading() && data) {
        RefPtr<SharedBuffer> buffer(data);

#if (!PLATFORM(MAC) || __MAC_OS_X_VERSION_MIN_REQUIRED <= 1090) && (!PLATFORM(IOS) || __IPHONE_OS_VERSION_MIN_REQUIRED < 80000)
        if (isWOFF(buffer.get())) {
            Vector<char> convertedFont;
            if (!convertWOFFToSfnt(buffer.get(), convertedFont))
                buffer = nullptr;
            else
                buffer = SharedBuffer::adoptVector(convertedFont);
        }
        else if (isWOFF2(buffer.get())) {
    		Vector<char> convertedFont;
    		if (!convertWOFF2ToSfnt(buffer.get(), convertedFont))
                buffer = nullptr;
    		else
                buffer = SharedBuffer::adoptVector(convertedFont);
    	}
#endif

        m_fontData = buffer ? adoptPtr(createFontCustomPlatformData(*buffer)) : nullptr;
        m_hasCreatedFontDataWrappingResource = m_fontData && (buffer == data);
        if (!m_fontData)
            setStatus(DecodeError);
    }
#endif
    return m_fontData;
}

PassRefPtr<SimpleFontData> CachedFont::getFontData(const FontDescription& fontDescription, const AtomicString&, bool syntheticBold, bool syntheticItalic)
{
    return platformDataFromCustomData(fontDescription, syntheticBold, syntheticItalic);
}

PassRefPtr<SimpleFontData> CachedFont::platformDataFromCustomData(const FontDescription& fontDescription, bool syntheticBold, bool syntheticItalic)
{
#ifdef STORE_FONT_CUSTOM_PLATFORM_DATA
    ASSERT(m_fontData);
    return m_fontData->fontPlatformData(fontDescription, syntheticBold, syntheticItalic, fontDescription.orientation());
#else
    return FontPlatformData();
#endif
}

void CachedFont::fontPlatformDataDeletionTimerCallback(Timer<CachedFont>*)
{
#ifdef STORE_FONT_CUSTOM_PLATFORM_DATA
	m_fontData.clear();
#endif
}

void CachedFont::purge()
{
#ifdef STORE_FONT_CUSTOM_PLATFORM_DATA
	if(m_fontData)
		m_fontData->purge();
#endif
}

void CachedFont::allClientsRemoved()
{
    if (double interval = memoryCache().deadFontPlatformDataDeletionInterval())
		m_fontPlatformDataDeletionTimer.startOneShot(interval);

    CachedResource::allClientsRemoved();
}

void CachedFont::checkNotify()
{
    if (isLoading())
        return;
    
    CachedResourceClientWalker<CachedFontClient> walker(m_clients);
    while (CachedFontClient* client = walker.next())
    	client->fontLoaded(this);
}

void CachedFont::error(CachedResource::Status status)
{
    setStatus(status);
    ASSERT(errorOccurred());
    setLoading(false);
    checkNotify();
}

bool CachedFont::mayTryReplaceEncodedData() const
{
    // If a FontCustomPlatformData has ever been constructed to wrap the internal resource buffer then it still might be in use somewhere.
    // That platform font object might directly reference the encoded data buffer behind this CachedFont,
    // so replacing it is unsafe.

    return !m_hasCreatedFontDataWrappingResource;
}

}
