/*
    Copyright (C) 1998 Lars Knoll (knoll@mpi-hd.mpg.de)
    Copyright (C) 2001 Dirk Mueller <mueller@kde.org>
    Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
    Copyright (C) 2004, 2005, 2006, 2007 Apple Inc. All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#ifndef CachedImage_h
#define CachedImage_h

#include "CachedResource.h"
#include "CachedResourceClient.h"
#include "ImageObserver.h"
#include "IntRect.h"
#include "IntSizeHash.h"
#include "LayoutTypesInlineMethods.h"
#include "SVGImageCache.h"
#include <wtf/HashMap.h>
#include <wtf/Vector.h>

namespace WebCore {

class CachedImageClient;
class CachedResourceLoader;
class FloatSize;
class MemoryCache;
class RenderElement;
class RenderObject;
class SecurityOrigin;

struct Length;

class CachedImage : public CachedResource {
    friend class MemoryCache;

public:
    CachedImage(CachedResourceRequest&, const String& mimeType);

    static CachedImage* create(Image*);

    virtual ~CachedImage();
    
    virtual void load(CachedResourceLoader*);

    Image* image(); // Returns the nullImage() if the image is not available yet.
    Image* imageForRenderer(const RenderObject*); // Returns the nullImage() if the image is not available yet.
    bool hasImage() const { return m_image.get(); }
    bool currentFrameKnownToBeOpaque(const RenderElement*); // Side effect: ensures decoded image is in cache, therefore should only be called when about to draw the image.

    std::pair<Image*, float> brokenImage(float deviceScaleFactor) const; // Returns an image and the image's resolution scale factor.
    bool willPaintBrokenImage() const; 

    bool canRender(const RenderObject* renderer, float multiplier) { return !errorOccurred() && !imageSizeForRenderer(renderer, multiplier).isEmpty(); }

    void setContainerSizeForRenderer(const CachedImageClient*, const LayoutSize&, float);
    bool usesImageContainerSize() const;
    bool imageHasRelativeWidth() const;
    bool imageHasRelativeHeight() const;

    enum SizeType {
        UsedSize,
        IntrinsicSize
    };
    // This method takes a zoom multiplier that can be used to increase the natural size of the image by the zoom.
    LayoutSize imageSizeForRenderer(const RenderObject*, float multiplier, SizeType = UsedSize); // returns the size of the complete image.
    void computeIntrinsicDimensions(Length& intrinsicWidth, Length& intrinsicHeight, FloatSize& intrinsicRatio);

    bool isManuallyCached() const { return m_isManuallyCached; }
    virtual RevalidationDecision makeRevalidationDecision(CachePolicy) const OVERRIDE;

    virtual void didAddClient(CachedResourceClient*);
    virtual void didRemoveClient(CachedResourceClient*);

    virtual void allClientsRemoved();
    virtual void handleOutOfMemory(bool clearRawData);
    virtual void destroyDecodedData();

    virtual void addDataBuffer(ResourceBuffer*) OVERRIDE;
    virtual void finishLoading(ResourceBuffer*) OVERRIDE;

    virtual void addData(const char* data, unsigned length) OVERRIDE;
    virtual void error(CachedResource::Status);
    virtual void responseReceived(const ResourceResponse&);
    
    // For compatibility, images keep loading even if there are HTTP errors.
    virtual bool shouldIgnoreHTTPStatusCodeErrors() const { return true; }

    virtual bool isImage() const { return true; }
    virtual bool stillNeedsLoad() const OVERRIDE { return !errorOccurred() && status() == Unknown && !isLoading(); }

    class CachedImageObserver FINAL : public RefCounted<CachedImageObserver>, public ImageObserver {
    public:
        static PassRefPtr<CachedImageObserver> create(CachedImage& image) { return adoptRef(new CachedImageObserver(image)); }
        void add(CachedImage& image) { m_cachedImages.append(&image); }
        void remove(CachedImage& image) { m_cachedImages.removeFirst(&image); }

    private:
        explicit CachedImageObserver(CachedImage&);

        // ImageObserver API
		virtual void decodedSizeChanged(const Image* image, int delta);
		virtual void didDraw(const Image*);

		virtual void animationAdvanced(const Image*);
		virtual void changedInRect(const Image*, const IntRect&);

        Vector<CachedImage*> m_cachedImages;
    };

    void decodedSizeChanged(const Image* image, int delta);
    void didDraw(const Image*);

    void animationAdvanced(const Image*);
    void changedInRect(const Image*, const IntRect&);

    void acceleratedSurfaceDeletionTimerCallback(Timer<CachedImage>*);

    bool isOriginClean(SecurityOrigin*);

private:
    void clear();

    CachedImage(CachedImage&, const ResourceRequest&);
    CachedImage(Image*);

    virtual void setBodyDataFrom(const CachedResource&) FINAL;

    void createImage();
    void clearImage();
    bool canBeDrawn() const;
    // If not null, changeRect is the changed part of the image.
    void notifyObservers(const IntRect* changeRect, bool allDataReceived);
    virtual PurgePriority purgePriority() const { return PurgeFirst; }
    void checkShouldPaintBrokenImage();

    virtual void switchClientsToRevalidatedResource() OVERRIDE;

    void addIncrementalDataBuffer(ResourceBuffer*);

    typedef pair<LayoutSize, float> SizeAndZoom;
    typedef HashMap<const CachedImageClient*, SizeAndZoom> ContainerSizeRequests;
    ContainerSizeRequests m_pendingContainerSizeRequests;

    RefPtr<CachedImageObserver> m_imageObserver;
    RefPtr<Image> m_image;
#if ENABLE(SVG)
    OwnPtr<SVGImageCache> m_svgImageCache;
#endif
    unsigned m_isManuallyCached : 1;
    unsigned m_shouldPaintBrokenImage : 1;

	Timer<CachedImage> m_acceleratedSurfaceDeletionTimer;
};

class CachedImageClient : public CachedResourceClient {
	WTF_MAKE_FAST_ALLOCATED;
public:
    virtual ~CachedImageClient() { }
    static CachedResourceClientType expectedType() { return ImageType; }
    virtual CachedResourceClientType resourceClientType() const { return expectedType(); }

    // Called whenever a frame of an image changes because we got more data from the network.
    // If not null, the IntRect is the changed rect of the image.
    virtual void imageChanged(CachedImage*, const IntRect* = 0) { }
    virtual void imageChanged(CachedImage* image, const IntRect* rect, bool allDataReceived)
    {
    	imageChanged(image, rect);
    }

    // Called when GIF animation progresses.
    virtual void newImageAnimationFrameAvailable(CachedImage& image) { imageChanged(&image); }
};

}

#endif
