/*
 * Copyright (C) 2012 Google, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GOOGLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CachedResourceRequest_h
#define CachedResourceRequest_h

#include "CachedResource.h"
#include "ReferrerPolicy.h"
#include "ResourceLoadPriority.h"
#include "ResourceLoaderOptions.h"
#include "ResourceRequest.h"
#include "SecurityOrigin.h"
#include <wtf/RefPtr.h>
#include <wtf/text/AtomicString.h>

namespace WebCore {
namespace ContentExtensions {
struct BlockedStatus;
}
class Document;
class Element;
class FrameLoader;

bool isRequestCrossOrigin(SecurityOrigin*, const URL& requestURL, const ResourceLoaderOptions&);

class CachedResourceRequest {
public:
    enum DeferOption { NoDefer, DeferredByClient };

    explicit CachedResourceRequest(const ResourceRequest&, const String& charset = String(), ResourceLoadPriority = ResourceLoadPriorityUnresolved);
    CachedResourceRequest(const ResourceRequest&, const ResourceLoaderOptions&, ResourceLoadPriority = ResourceLoadPriorityUnresolved);
    CachedResourceRequest(const ResourceRequest&, ResourceLoadPriority);
    ~CachedResourceRequest();

    ResourceRequest& releaseResourceRequest() { return m_resourceRequest; }
    const ResourceRequest& resourceRequest() const { return m_resourceRequest; }
    const String& charset() const { return m_charset; }
    void setCharset(const String& charset) { m_charset = charset; }
    const String& mimeType() const { return m_mimeType; }
    void setMimeType(const String& mimeType) { m_mimeType = mimeType; }
    const ResourceLoaderOptions& options() const { return m_options; }
    void setOptions(const ResourceLoaderOptions& options) { m_options = options; }
    ResourceLoadPriority priority() const { return m_priority; }
    bool forPreload() const { return m_forPreload; }
    void setForPreload(bool forPreload) { m_forPreload = forPreload; }
    DeferOption defer() const { return m_defer; }
    void setDefer(DeferOption defer) { m_defer = defer; }
    void setInitiator(PassRefPtr<Element>);
    void setInitiator(const AtomicString& name);
    bool allowsCaching() const { return m_options.cachingPolicy() == CachingPolicy::AllowCaching; }
    void setCachingPolicy(CachingPolicy policy) { m_options.setCachingPolicy(policy); }

    void setAsPotentiallyCrossOrigin(const String&, Document&);
    void updateForAccessControl(Document&);

    void updateReferrerOriginAndUserAgentHeaders(FrameLoader&, ReferrerHeaderPolicy);
    void upgradeInsecureRequestIfNeeded(Document&);
    void setAcceptHeaderIfNone(CachedResource::Type);
    void updateAccordingCacheMode();
    void updateAcceptEncodingHeader();

    void removeFragmentIdentifierIfNeeded();
#if ENABLE(CONTENT_EXTENSIONS)
    void applyBlockedStatus(const ContentExtensions::BlockedStatus&);
#endif
#if ENABLE(CACHE_PARTITIONING)
    void setDomainForCachePartition(Document&);
#endif

    void setOrigin(PassRefPtr<SecurityOrigin> origin) { m_origin = origin; }
    PassRefPtr<SecurityOrigin> releaseOrigin() { return m_origin.release(); }
    SecurityOrigin* origin() const { return m_origin.get(); }

    String releaseFragmentIdentifier()
    {
    	String fragmentIdentifier = m_fragmentIdentifier;
    	m_fragmentIdentifier = String();
    	return fragmentIdentifier;
    }

    void clearFragmentIdentifier() { m_fragmentIdentifier = String(); }

    static String splitFragmentIdentifierFromRequestURL(ResourceRequest&);

private:
    ResourceRequest m_resourceRequest;
    String m_charset;
    String m_mimeType;
    ResourceLoaderOptions m_options;
    ResourceLoadPriority m_priority;
    bool m_forPreload;
    DeferOption m_defer;
    RefPtr<SecurityOrigin> m_origin;
    String m_fragmentIdentifier;
};

} // namespace WebCore

#endif
