/*
    Copyright (C) 1998 Lars Knoll (knoll@mpi-hd.mpg.de)
    Copyright (C) 2001 Dirk Mueller (mueller@kde.org)
    Copyright (C) 2002 Waldo Bastian (bastian@kde.org)
    Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include "config.h"
#include "MemoryCache.h"

#if ENABLE(BML)
#include "CachedClut.h"
#endif
#include "CachedCSSStyleSheet.h"
#include "CachedFont.h"
#include "CachedImage.h"
#include "CachedScript.h"
#include "CachedXSLStyleSheet.h"
#include "CachedResourceLoader.h"
#include "CrossThreadTask.h"
#include "Document.h"
#include "FrameLoader.h"
#include "FrameLoaderTypes.h"
#include "FrameView.h"
#include "Image.h"
#include "Logging.h"
#include "ResourceHandle.h"
#include "SecurityOrigin.h"
#include "SecurityOriginHash.h"
#include "WorkerGlobalScope.h"
#include "WorkerLoaderProxy.h"
#include "WorkerThread.h"
#include <stdio.h>
#include <wtf/CurrentTime.h>
#include <wtf/TemporaryChange.h>
#include <wtf/text/CString.h>

using namespace std;

namespace WebCore {

static const int cDefaultCacheCapacity = 16 * 1024 * 1024;
static const double cMinDelayBeforeLiveDecodedPrune = 1; // Seconds.
static const float cTargetPrunePercentage = .75f; // Percentage of capacity toward which we prune, to avoid immediately pruning again.
static const double cDefaultDecodedDataDeletionInterval = 0;

DLL_PUBLIC MemoryCache& memoryCache()
{
    static MemoryCache* staticCache = new MemoryCache;
    ASSERT(WTF::isMainThread());

    return *staticCache;
}

MemoryCache::MemoryCache()
: m_disabled(false)
, m_pruneEnabled(true)
, m_inPruneResources(false)
, m_shouldPurgeImageFrameBufferWhenPossible(false)
, m_inHandleOutOfMemory(false)
, m_capacity(cDefaultCacheCapacity)
, m_minDeadCapacity(0)
, m_maxDeadCapacity(cDefaultCacheCapacity)
, m_decodedDataDeletionInterval(0)
, m_deadDecodedDataDeletionInterval(cDefaultDecodedDataDeletionInterval)
, m_deadFontPlatformDataDeletionInterval(8)
, m_acceleratedSurfaceDeletionInterval(0)
, m_simpleFontDataDeletionInterval(8)
, m_liveSize(0)
, m_deadSize(0)
, m_largeAnimationCutoff(1920 * 1080 * 4)
, m_resourcesInUpdating(false)
, m_pruneTimer(this, &MemoryCache::pruneTimerFired)
{
}

bool MemoryCache::shouldRemoveFragmentIdentifier(const KURL& originalURL)
{
    if (!originalURL.hasFragmentIdentifier())
        return false;
    // Strip away fragment identifier from HTTP URLs.
    // Data URLs must be unmodified. For file and custom URLs clients may expect resources
    // to be unique even when they differ by the fragment identifier only.
    return originalURL.protocolIsInHTTPFamily();
}

KURL MemoryCache::removeFragmentIdentifierIfNeeded(const KURL& originalURL)
{
    if (!shouldRemoveFragmentIdentifier(originalURL))
        return originalURL;
    KURL url = originalURL;
    url.removeFragmentIdentifier();
    return url;
}

bool MemoryCache::add(CachedResource* resource)
{
    if (disabled())
        return false;

    ASSERT(WTF::isMainThread());

	m_resourcesInUpdating = true;
    m_resources.set(resource->url(), resource);
	m_resourcesInUpdating = false;
	
    resource->setInCache(true);
    
    resourceAccessed(resource);
    
    LOG(ResourceLoading, "MemoryCache::add Added '%s', resource %p\n", resource->url().string().latin1().data(), resource);
    return true;
}

void MemoryCache::revalidationSucceeded(CachedResource* revalidatingResource, const ResourceResponse& response)
{
    CachedResource* resource = revalidatingResource->resourceToRevalidate();
    ASSERT(resource);
    ASSERT(!resource->inCache());
    ASSERT(resource->isLoaded());

    // Calling evict() can potentially delete revalidatingResource, which we use
    // below. This mustn't be the case since revalidation means it is loaded
    // and so canDelete() is false.
    ASSERT(!revalidatingResource->canDelete());

    evict(revalidatingResource);

    ASSERT(!m_resources.get(resource->url()));

	m_resourcesInUpdating = true;
    m_resources.set(resource->url(), resource);
	m_resourcesInUpdating = false;
	
    resource->setInCache(true);
    resource->updateResponseAfterRevalidation(response);

	removeFromAccessOrderedResourcesList(resource);
    insertInAccessOrderedResourcesList(resource);

    int delta = resource->size();
    if (delta)
        adjustSize(resource->hasClients(), delta);
    
    revalidatingResource->switchClientsToRevalidatedResource();
    ASSERT(!revalidatingResource->m_deleted);
    // this deletes the revalidating resource
    if(revalidatingResource->resourceToRevalidate())
    	revalidatingResource->clearResourceToRevalidate();
}

void MemoryCache::revalidationFailed(CachedResource* revalidatingResource)
{
    ASSERT(WTF::isMainThread());
    LOG(ResourceLoading, "Revalidation failed for %p", revalidatingResource);
    // this deletes the revalidating resource
    if(revalidatingResource->resourceToRevalidate())
    	revalidatingResource->clearResourceToRevalidate();
}

CachedResource* MemoryCache::resourceForURL(const KURL& originalURL)
{
    ASSERT(WTF::isMainThread());

    KURL url = removeFragmentIdentifierIfNeeded(originalURL);

    CachedResource* resource = m_resources.get(url);
    bool wasPurgeable = MemoryCache::shouldMakeResourcePurgeableOnEviction() && resource && resource->isPurgeable();
    if (resource && !resource->makePurgeable(false)) {
        ASSERT(!resource->hasClients());    	        
        revalidationFailed(resource); // this deletes the revalidating resource
        evict(resource);
        return 0;
    }

    if(resource && resource->errorOccurred()) {    	  
        revalidationFailed(resource); // this deletes the revalidating resource 
        evict(resource);
        return 0;
    }
    
    // Add the size back since we had subtracted it when we marked the memory as purgeable.
    if (wasPurgeable)
        adjustSize(resource->hasClients(), resource->size());
    return resource;
}

CachedResource* MemoryCache::resourceForRequest(const CachedResourceRequest& cachedResourceRequest)
{
    // FIXME: Change all clients to make sure HTTP(s) URLs have no fragment identifiers before calling here.
    // CachedResourceLoader is now doing this. Add an assertion once all other clients are doing it too.
	return resourceForURL(cachedResourceRequest.resourceRequest().url());
}

DLL_PUBLIC int MemoryCache::prune(int requiredSize, const Image* forImage, bool canReuse)
{
	int oldSize = (m_liveSize + m_deadSize);
	
	pruneDeadResources(true);
	
	if((oldSize - m_liveSize - m_deadSize) < requiredSize){
		CachedResource* current = m_accessOrderedResources.m_tail;

		while (current) {
			CachedResource* prev = current->m_prevInAccessOrderedResourcesList;

			if(current->isImage() && (static_cast<CachedImage*>(current)->image() != forImage))
				current->destroyDecodedData();

			if((oldSize - m_liveSize - m_deadSize) >= requiredSize)
				break;

			if(!canReuse)
				current->disableReuse(); // do not reuse it, while delete when possible

			current = prev;
		}
	}

    return (oldSize - (m_liveSize + m_deadSize));
}

DLL_PUBLIC int MemoryCache::handleOutOfMemory(int requiredSize)
{
	int oldSize = (m_liveSize + m_deadSize);
	CachedResource* current = m_accessOrderedResources.m_tail;

	m_inHandleOutOfMemory = true;

	while (current) {
		CachedResource* prev = current->m_prevInAccessOrderedResourcesList;

    	current->handleOutOfMemory(false);

        if((oldSize - m_liveSize - m_deadSize) >= requiredSize)
        	break;
        
		current = prev;
	}

	m_inHandleOutOfMemory = false;

	return (oldSize - (m_liveSize + m_deadSize));
}

DLL_PUBLIC void MemoryCache::purge()
{
	CachedResource* current = m_accessOrderedResources.m_tail;
	while (current) {
		CachedResource* prev = current->m_prevInAccessOrderedResourcesList;
    	current->purge();
		current = prev;
	}
}

void MemoryCache::pruneLiveResources()
{
    if (!m_pruneEnabled)
        return;

    unsigned capacity = liveCapacity();
    if (capacity && m_liveSize <= capacity)
        return;

    unsigned targetSize = static_cast<unsigned>(capacity * cTargetPrunePercentage); // Cut by a percentage to avoid immediately pruning again.

    pruneLiveResourcesToSize(targetSize);
}

void MemoryCache::pruneLiveResourcesToPercentage(float prunePercentage)
{
    if (!m_pruneEnabled)
        return;

    if (prunePercentage < 0.0f  || prunePercentage > 0.95f)
        return;

    unsigned currentSize = m_liveSize + m_deadSize;
    unsigned targetSize = static_cast<unsigned>(currentSize * prunePercentage);

    pruneLiveResourcesToSize(targetSize);
}

void MemoryCache::pruneLiveResourcesToSize(unsigned targetSize)
{
    if (m_inPruneResources)
        return;

    TemporaryChange<bool> reentrancyProtector(m_inPruneResources, true);
    
    // Destroy any decoded data in live objects that we can.
    // Start from the tail, since this is the least recently accessed of the objects.

    // The list might not be sorted by the m_lastDecodedAccessTime. The impact
    // of this weaker invariant is minor as the below if statement to check the
    // elapsedTime will evaluate to false as the currentTime will be a lot
    // greater than the current->m_lastDecodedAccessTime.
    // For more details see: https://bugs.webkit.org/show_bug.cgi?id=30209
    CachedResource* current = m_accessOrderedResources.m_tail;
    while (current) {
        CachedResource* prev = current->m_prevInAccessOrderedResourcesList;
        ASSERT(current->hasClients());
        if (current->isLoaded() && current->decodedSize()) {
            current->destroyDecodedData();

            if (targetSize && m_liveSize <= targetSize)
                return;
        }
        current = prev;
    }
}

DLL_PUBLIC void MemoryCache::pruneDeadResources(bool all)
{
    if (!m_pruneEnabled)
        return;

    if(all){
    	pruneDeadResourcesToSize(0);
    	return;
    }
    
    unsigned capacity = deadCapacity();
    if (capacity && m_deadSize <= capacity)
        return;

    unsigned targetSize = static_cast<unsigned>(capacity * cTargetPrunePercentage); // Cut by a percentage to avoid immediately pruning again.
    pruneDeadResourcesToSize(targetSize);
}

void MemoryCache::pruneDeadResourcesToPercentage(float prunePercentage)
{
    if (!m_pruneEnabled)
        return;

    if (prunePercentage < 0.0f  || prunePercentage > 0.95f)
        return;

    unsigned currentSize = m_liveSize + m_deadSize;
    unsigned targetSize = static_cast<unsigned>(currentSize * prunePercentage);

    pruneDeadResourcesToSize(targetSize);
}

void MemoryCache::pruneDeadResourcesToSize(unsigned targetSize)
{
    if (m_inPruneResources)
        return;

    TemporaryChange<bool> reentrancyProtector(m_inPruneResources, true);
 
    // Now evict objects from this queue.
    CachedResource* current = m_accessOrderedResources.m_tail;
	while (current) {
        // Protect 'previous' so it can't get deleted during destroyDecodedData().
        CachedResourceHandle<CachedResource> previous = current->m_prevInAccessOrderedResourcesList;
        ASSERT(!previous || previous->inCache());
		if (!current->hasClients() && !current->isPreloaded() && !current->isCacheValidator()) {
			if (!makeResourcePurgeable(current))
				evict(current);

			if (targetSize && m_deadSize <= targetSize)
				return;
		}
        // Decoded data may reference other resources. Stop iterating if 'previous' somehow got
        // kicked out of cache during destroyDecodedData().
        if (previous && !previous->inCache())
            break;
        current = previous.get();
	}
}

DLL_PUBLIC void MemoryCache::pruneDeadResourcesSinceLastAccess(double interval)
{
    if (m_inPruneResources)
        return;

    if(m_deadSize == 0)
    	return;

    TemporaryChange<bool> reentrancyProtector(m_inPruneResources, true);

    double lastAccessTime = WTF::monotonicallyIncreasingTime() - interval;

    // Now evict objects from this queue.
    CachedResource* current = m_accessOrderedResources.m_tail;
	while (current) {
        // Protect 'previous' so it can't get deleted during destroyDecodedData().
        CachedResourceHandle<CachedResource> previous = current->m_prevInAccessOrderedResourcesList;
        ASSERT(!previous || previous->inCache());
		if (!current->hasClients() && !current->isPreloaded() && !current->isCacheValidator()) {
			if(current->m_lastAccessTime < lastAccessTime)
				evict(current);
			else
				break;
		}

        // Decoded data may reference other resources. Stop iterating if 'previous' somehow got
        // kicked out of cache during destroyDecodedData().
        if (previous && !previous->inCache())
            break;
        current = previous.get();
	}
}

DLL_PUBLIC void MemoryCache::setCapacities(unsigned minDeadBytes, unsigned maxDeadBytes, unsigned totalBytes)
{
    ASSERT(minDeadBytes <= maxDeadBytes);
    ASSERT(maxDeadBytes <= totalBytes);
    m_minDeadCapacity = minDeadBytes;
    m_maxDeadCapacity = maxDeadBytes;
    m_capacity = totalBytes;
    prune();
}

bool MemoryCache::makeResourcePurgeable(CachedResource* resource)
{
    if (!MemoryCache::shouldMakeResourcePurgeableOnEviction())
        return false;

    if (!resource->inCache())
        return false;

    if (resource->isPurgeable())
        return true;

    if (!resource->isSafeToMakePurgeable())
        return false;

    if (!resource->makePurgeable(true))
        return false;

    adjustSize(resource->hasClients(), -static_cast<int>(resource->size()));

    return true;
}

DLL_PUBLIC int MemoryCache::evict(const String& urlPrefix)
{
	int total = 0;
evict_again:
    for(CachedResourceMap::iterator i = m_resources.begin(); i != m_resources.end(); ++i){
    	if(i->second->url().string().startsWith(urlPrefix)){
    		//fprintf(stderr, ">>>>Cache::evict(%s)\n", i->second->url().latin1().data());
    		
    		++total;
    		
    		// this deletes the revalidating resource        
    		revalidationFailed(i->second);
        
    		evict(i->second);
    		
    		goto evict_again;
    	}
    }
    
    return total;
}

void MemoryCache::evict(CachedResource* resource)
{
    ASSERT(WTF::isMainThread());
    LOG(ResourceLoading, "Evicting resource %p for '%s' from cache", resource, resource->url().string().latin1().data());
    // The resource may have already been removed by someone other than our caller,
    // who needed a fresh copy for a reload. See <http://bugs.webkit.org/show_bug.cgi?id=12479#c6>.
    if (resource->inCache()) {
        // Remove from the resource map.

    	m_resourcesInUpdating = true;
        m_resources.remove(resource->url());
    	m_resourcesInUpdating = false;
    	
        resource->setInCache(false);

        // Remove from the appropriate LRU list.
        removeFromAccessOrderedResourcesList(resource);

        // If the resource was purged, it means we had already decremented the size when we made the
        // resource purgeable in makeResourcePurgeable(). So adjust the size if we are evicting a
        // resource that was not marked as purgeable.
        if (!MemoryCache::shouldMakeResourcePurgeableOnEviction() || !resource->isPurgeable())
            adjustSize(resource->hasClients(), -static_cast<int>(resource->size()));
    } else
        ASSERT(m_resources.get(resource->url()) != resource);

    if (resource->canDelete())
        delete resource;
    else if(!resource->hasClients())
    	resource->destroyDecodedDataIfNeeded();
}

void MemoryCache::insertInAccessOrderedResourcesList(CachedResource* resource)
{
	if(!m_accessOrderedResources.m_head) {
		m_accessOrderedResources.m_head = resource;
		m_accessOrderedResources.m_tail = resource;
		resource->m_nextInAccessOrderedResourcesList = 0;
		resource->m_prevInAccessOrderedResourcesList = 0;
		return;
	}

	m_accessOrderedResources.m_head->m_prevInAccessOrderedResourcesList = resource;
    resource->m_nextInAccessOrderedResourcesList = m_accessOrderedResources.m_head;
    resource->m_prevInAccessOrderedResourcesList = 0;
    m_accessOrderedResources.m_head = resource;
}

void MemoryCache::removeFromAccessOrderedResourcesList(CachedResource* resource)
{
	// check resource is on list
	if(resource == m_accessOrderedResources.m_head || resource->m_prevInAccessOrderedResourcesList) {
		if(resource->m_prevInAccessOrderedResourcesList)
			resource->m_prevInAccessOrderedResourcesList->m_nextInAccessOrderedResourcesList = resource->m_nextInAccessOrderedResourcesList;
		else
			m_accessOrderedResources.m_head = resource->m_nextInAccessOrderedResourcesList;

		if(resource->m_nextInAccessOrderedResourcesList)
			resource->m_nextInAccessOrderedResourcesList->m_prevInAccessOrderedResourcesList = resource->m_prevInAccessOrderedResourcesList;
		else
			m_accessOrderedResources.m_tail = resource->m_prevInAccessOrderedResourcesList;

		resource->m_nextInAccessOrderedResourcesList = 0;
		resource->m_prevInAccessOrderedResourcesList = 0;
	}
}

void MemoryCache::didAccessDecodedData(CachedResource* resource)
{
	removeFromAccessOrderedResourcesList(resource);
	insertInAccessOrderedResourcesList(resource);
}

void MemoryCache::resourceAccessed(CachedResource* resource)
{
    ASSERT(resource->inCache());
    
    // Need to make sure to remove before we increase the access count, since
    // the queue will possibly change.
    removeFromAccessOrderedResourcesList(resource);
    
    // If this is the first time the resource has been accessed, adjust the size of the cache to account for its initial size.
    if (!resource->accessCount())
        adjustSize(resource->hasClients(), resource->size());
    
    // Add to our access count.
    resource->increaseAccessCount();
    
    // Now insert into the new queue.
    insertInAccessOrderedResourcesList(resource);
}

void MemoryCache::removeResourcesWithOrigin(SecurityOrigin* origin)
{
    Vector<CachedResource*> resourcesWithOrigin;

    CachedResourceMap::iterator e = m_resources.end();
    for (CachedResourceMap::iterator it = m_resources.begin(); it != e; ++it) {
        CachedResource* resource = it->second;
        RefPtr<SecurityOrigin> resourceOrigin = SecurityOrigin::createFromString(resource->url());
        if (!resourceOrigin)
            continue;
        if (resourceOrigin->equal(origin))
            resourcesWithOrigin.append(resource);
    }

    for (size_t i = 0; i < resourcesWithOrigin.size(); ++i)
        remove(resourcesWithOrigin[i]);
}

void MemoryCache::getOriginsWithCache(SecurityOriginSet& origins)
{
    CachedResourceMap::iterator e = m_resources.end();
    for (CachedResourceMap::iterator it = m_resources.begin(); it != e; ++it)
        origins.add(SecurityOrigin::createFromString(it->second->url()));
}

void MemoryCache::addToLiveResourcesSize(CachedResource* resource)
{
    m_liveSize += resource->size();
    m_deadSize -= resource->size();
}

void MemoryCache::removeFromLiveResourcesSize(CachedResource* resource)
{
    m_liveSize -= resource->size();
    m_deadSize += resource->size();
}

void MemoryCache::adjustSize(bool live, int delta)
{
    if (live) {
        ASSERT(delta >= 0 || ((int)m_liveSize + delta >= 0));
        m_liveSize += delta;
    } else {
        ASSERT(delta >= 0 || ((int)m_deadSize + delta >= 0));
        m_deadSize += delta;
    }
}


void MemoryCache::removeUrlFromCache(ScriptExecutionContext* context, const String& urlString) 
{
#if ENABLE(WORKERS)
    if (context->isWorkerGlobalScope()) {
      WorkerGlobalScope* workerGlobalScope = static_cast<WorkerGlobalScope*>(context);
      workerGlobalScope->thread()->workerLoaderProxy().postTaskToLoader(
          createCallbackTask(&removeUrlFromCacheImpl, urlString));
      return;
    }
#endif
    removeUrlFromCacheImpl(context, urlString);
}

void MemoryCache::removeUrlFromCacheImpl(ScriptExecutionContext*, const String& urlString)
{
    KURL url(KURL(), urlString);

    MemoryCache& cache = memoryCache();

    if (CachedResource* resource = cache.resourceForURL(url))
        cache.remove(resource);
}

void MemoryCache::TypeStatistic::addResource(CachedResource* o)
{
    bool purged = o->wasPurged();
    bool purgeable = o->isPurgeable() && !purged; 
    int pageSize = (o->encodedSize() + o->overheadSize() + 4095) & ~4095;
    count++;
    size += purged ? 0 : o->size(); 
    liveSize += o->hasClients() ? o->size() : 0;
    decodedSize += o->decodedSize();
    purgeableSize += purgeable ? pageSize : 0;
    purgedSize += purged ? pageSize : 0;
}

MemoryCache::Statistics MemoryCache::getStatistics()
{
    Statistics stats;
    CachedResourceMap::iterator end = m_resources.end();
    for (CachedResourceMap::iterator i = m_resources.begin(); i != end; ++i) {
        CachedResource* resource = i->second;
        switch (resource->type()) {
        case CachedResource::ImageResource:
            stats.images.addResource(resource);
            break;
        case CachedResource::CSSStyleSheet:
            stats.cssStyleSheets.addResource(resource);
            break;
        case CachedResource::Script:
            stats.scripts.addResource(resource);
            break;
#if ENABLE(XSLT)
        case CachedResource::XSLStyleSheet:
            stats.xslStyleSheets.addResource(resource);
            break;
#endif
        case CachedResource::FontResource:
#if ENABLE(SVG_FONTS)
		case CachedResource::SVGFontResource:
#endif
            stats.fonts.addResource(resource);
            break;
        case CachedResource::RawResource:
            stats.rawResources.addResource(resource);
            break;
        default:
        	stats.others.addResource(resource);
            break;
        }
    }
    return stats;
}

void MemoryCache::setDisabled(bool disabled)
{
    m_disabled = disabled;
    if (!m_disabled)
        return;

    for (;;) {
        CachedResourceMap::iterator i = m_resources.begin();
        if (i == m_resources.end())
            break;
    	        
        revalidationFailed(i->second); // this deletes the revalidating resource
        evict(i->second);
    }
}

void MemoryCache::evictResources()
{
    if (disabled())
        return;

    setDisabled(true);
    setDisabled(false);
}

void MemoryCache::pruneTimerFired(Timer<MemoryCache>*)
{
    prune();
}

void MemoryCache::pruneSoon()
{
     if (m_pruneTimer.isActive())
        return;
     if (!needsPruning())
         return;
     m_pruneTimer.startOneShot(0);
}

void MemoryCache::pruneToPercentage(float targetPercentLive)
{
    pruneDeadResourcesToPercentage(targetPercentLive); // Prune dead first, in case it was "borrowing" capacity from live.
    pruneLiveResourcesToPercentage(targetPercentLive);
}

static void dumpStat(const char* name, MemoryCache::TypeStatistic& statistics, int (*logFunction)(const char * format, ... ))
{
	if(statistics.count == 0)
		return;
	logFunction("%-13s %13d %13d %13d %13d\n", name, statistics.count, statistics.size, statistics.liveSize, statistics.decodedSize);
}

DLL_PUBLIC void MemoryCache::dumpStats(int (*logFunction)(const char * format, ... ))
{
	if(m_resourcesInUpdating){
		logFunction("    >Updating resources, forbidden statistics\n");		    
		return;
	}
	
    logFunction("%-13s %-13s %-13s %-13s %-13s\n", "", "        Count", "         Size", "     LiveSize", "  DecodedSize");
    logFunction("%-13s %-13s %-13s %-13s %-13s\n", "-------------", "-------------", "-------------", "-------------", "-------------");

	if(m_resources.size() == 0)
		return;

    Statistics s = getStatistics();
    dumpStat("Images", s.images, logFunction);
    dumpStat("CSS", s.cssStyleSheets, logFunction);
#if ENABLE(XSLT)
    dumpStat("XSL", s.xslStyleSheets, logFunction);
#endif
    dumpStat("JavaScript", s.scripts, logFunction);
    dumpStat("Fonts", s.fonts, logFunction);
    dumpStat("Raw", s.rawResources, logFunction);
    dumpStat("Others", s.others, logFunction);
    logFunction("%-13s %-13s %-13s %-13s %-13s\n", "-------------", "-------------", "-------------", "-------------", "-------------");
}

#ifndef NDEBUG
void MemoryCache::dumpLRULists(bool includeLive) const
{
    printf("LRU-SP lists in eviction order (Kilobytes decoded, Kilobytes encoded, Access count, Referenced, isPurgeable, wasPurged):\n");

    int size = m_allResources.size();
    for (int i = size - 1; i >= 0; i--) {
        printf("\n\nList %d: ", i);
        CachedResource* current = m_allResources[i].m_tail;
        while (current) {
            CachedResource* prev = current->m_prevInAllResourcesList;
            if (includeLive || !current->hasClients())
                printf("(%.1fK, %.1fK, %uA, %dR, %d, %d); ", current->decodedSize() / 1024.0f, (current->encodedSize() + current->overheadSize()) / 1024.0f, current->accessCount(), current->hasClients(), current->isPurgeable(), current->wasPurged());

            current = prev;
        }
    }
}
#endif

} // namespace WebCore
