/*
 * Copyright (C) 2011 Google, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GOOGLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ContentSecurityPolicy.h"

#include "ContentSecurityPolicyDirective.h"
#include "ContentSecurityPolicyDirectiveList.h"
#include "ContentSecurityPolicyDirectiveNames.h"
#include "ContentSecurityPolicySource.h"
#include "ContentSecurityPolicySourceList.h"
#include "Console.h"
#include "CryptoDigest.h"
#include "DOMStringList.h"
#include "Document.h"
#include "DocumentLoader.h"
#include "FormData.h"
#include "FormDataList.h"
#include "Frame.h"
#include "HTMLParserIdioms.h"
#include "InspectorInstrumentation.h"
#include "InspectorValues.h"
#include "JSDOMWindowShell.h"
#include "JSMainThreadExecState.h"
#include "ParsingUtilities.h"
#include "PingLoader.h"
#include "ResourceRequest.h"
#include "SchemeRegistry.h"
#include "ScriptCallStack.h"
#include "ScriptCallStackFactory.h"
#include "SecurityOrigin.h"
#include "SecurityPolicyViolationEvent.h"
#include "TextEncoding.h"
#include "TextResourceDecoder.h"
#include <wtf/TemporaryChange.h>
#include <wtf/text/StringBuilder.h>
#include <wtf/text/TextPosition.h>

namespace WebCore {

static String consoleMessageForViolation(const char* effectiveViolatedDirective, const ContentSecurityPolicyDirective& violatedDirective, const KURL& blockedURL, const char* prefix, const char* subject = "it")
{
    StringBuilder result;
    if (violatedDirective.directiveList().isReportOnly())
        result.appendLiteral("[Report Only] ");
    result.append(prefix);
    if (!blockedURL.isEmpty()) {
        result.append(' ');
        result.append(blockedURL.stringCenterEllipsizedToLength());
    }
    result.appendLiteral(" because ");
    result.append(subject);
    if (violatedDirective.isDefaultSrc()) {
        result.appendLiteral(" appears in neither the ");
        result.append(effectiveViolatedDirective);
        result.appendLiteral(" directive nor the default-src directive of the Content Security Policy.");
    } else {
        result.appendLiteral(" does not appear in the ");
        result.append(effectiveViolatedDirective);
        result.appendLiteral(" directive of the Content Security Policy.");
    }
    return result.toString();
}

ContentSecurityPolicy::ContentSecurityPolicy(ScriptExecutionContext& scriptExecutionContext)
    : m_scriptExecutionContext(&scriptExecutionContext)
	, m_frame(0)
    , m_sandboxFlags(SandboxNone)
	, m_overrideInlineStyleAllowed(false)
	, m_isReportingEnabled(true)
	, m_upgradeInsecureRequests(false)
{
    ASSERT(scriptExecutionContext.securityOrigin());
    SecurityOrigin& securityOrigin = *scriptExecutionContext.securityOrigin();
    m_selfSourceProtocol = securityOrigin.protocol();
    m_selfSource = adoptPtr(new ContentSecurityPolicySource(*this, m_selfSourceProtocol, securityOrigin.host(), securityOrigin.port(), emptyString(), false, false));
}

ContentSecurityPolicy::ContentSecurityPolicy(const SecurityOrigin& securityOrigin, const Frame* frame)
	: m_scriptExecutionContext(0)
	, m_frame(frame)
	, m_sandboxFlags(SandboxNone)
	, m_overrideInlineStyleAllowed(false)
	, m_isReportingEnabled(true)
	, m_upgradeInsecureRequests(false)
{
    m_selfSourceProtocol = securityOrigin.protocol();
    m_selfSource = adoptPtr(new ContentSecurityPolicySource(*this, m_selfSourceProtocol, securityOrigin.host(), securityOrigin.port(), emptyString(), false, false));
}

ContentSecurityPolicy::~ContentSecurityPolicy()
{
}

void ContentSecurityPolicy::copyStateFrom(const ContentSecurityPolicy* other) 
{
    ASSERT(m_policies.isEmpty());
    for (auto iter = other->m_policies.begin(); iter != other->m_policies.end(); ++iter)
        didReceiveHeader((*iter)->header(), (*iter)->headerType(), ContentSecurityPolicy::PolicyFrom::Inherited);
}

void ContentSecurityPolicy::copyUpgradeInsecureRequestStateFrom(const ContentSecurityPolicy* other)
{
    m_upgradeInsecureRequests = other->m_upgradeInsecureRequests;
    for(auto iter = other->m_insecureNavigationRequestsToUpgrade.begin(); iter != other->m_insecureNavigationRequestsToUpgrade.end(); ++iter)
    	m_insecureNavigationRequestsToUpgrade.add(iter->isolatedCopy()); // used by worker
}

void ContentSecurityPolicy::didCreateWindowShell(JSDOMWindowShell& windowShell) const
{
    JSDOMWindow* window = windowShell.window();
    ASSERT(window);
    ASSERT(window->scriptExecutionContext());
    ASSERT(window->scriptExecutionContext()->contentSecurityPolicy() == this);
    if (!windowShell.world()->isNormal()) {
        window->setEvalEnabled(true);
        return;
    }
    window->setEvalEnabled(m_lastPolicyEvalDisabledErrorMessage.isNull(), m_lastPolicyEvalDisabledErrorMessage);
}

ContentSecurityPolicyResponseHeaders ContentSecurityPolicy::responseHeaders() const
{
    ContentSecurityPolicyResponseHeaders result;
    result.m_headers.reserveInitialCapacity(m_policies.size());
    for (CSPDirectiveListVector::const_iterator iter = m_policies.begin(); iter != m_policies.end(); ++iter)
        result.m_headers.uncheckedAppend(std::pair<String, ContentSecurityPolicyHeaderType>((*iter)->header(), (*iter)->headerType()));
    return result;
}

void ContentSecurityPolicy::didReceiveHeaders(const ContentSecurityPolicyResponseHeaders& headers, ReportParsingErrors reportParsingErrors)
{
	TemporaryChange<bool> isReportingEnabled(m_isReportingEnabled, reportParsingErrors == ReportParsingErrors::Yes);
	for(Vector<std::pair<String, ContentSecurityPolicyHeaderType>>::const_iterator it = headers.m_headers.begin(); it != headers.m_headers.end(); ++it)
        didReceiveHeader(it->first, it->second, ContentSecurityPolicy::PolicyFrom::HTTPHeader);
}

void ContentSecurityPolicy::didReceiveHeader(const String& header, ContentSecurityPolicyHeaderType type, ContentSecurityPolicy::PolicyFrom policyFrom)
{
    // RFC2616, section 4.2 specifies that headers appearing multiple times can
    // be combined with a comma. Walk the header string, and parse each comma
    // separated chunk as a separate header.
    auto characters = StringView(header).upconvertedCharacters();
    const UChar* begin = characters;
    const UChar* position = begin;
    const UChar* end = begin + header.length();
    while (position < end) {
        skipUntil<UChar>(position, end, ',');

        // header1,header2 OR header1
        //        ^                  ^
        OwnPtr<ContentSecurityPolicyDirectiveList> policy = ContentSecurityPolicyDirectiveList::create(*this, String(begin, position - begin), type, policyFrom);
        const ContentSecurityPolicyDirective* violatedDirective = policy->violatedDirectiveForUnsafeEval();
		if (violatedDirective && !violatedDirective->directiveList().isReportOnly())
			m_lastPolicyEvalDisabledErrorMessage = policy->evalDisabledErrorMessage();

		m_policies.append(policy.release());

        // Skip the comma, and begin the next header from the current position.
        ASSERT(position == end || *position == ',');
        skipExactly<UChar>(position, end, ',');
        begin = position;
    }

    if (m_scriptExecutionContext)
        applyPolicyToScriptExecutionContext();
}

void ContentSecurityPolicy::applyPolicyToScriptExecutionContext()
{
    ASSERT(m_scriptExecutionContext);
    if (!m_lastPolicyEvalDisabledErrorMessage.isNull())
        m_scriptExecutionContext->disableEval(m_lastPolicyEvalDisabledErrorMessage);
    if (m_sandboxFlags != SandboxNone && m_scriptExecutionContext->isDocument())
        m_scriptExecutionContext->enforceSandboxFlags(m_sandboxFlags);
}

void ContentSecurityPolicy::setOverrideAllowInlineStyle(bool value)
{
    m_overrideInlineStyleAllowed = value;
}

bool ContentSecurityPolicy::urlMatchesSelf(const KURL& url) const
{
    return m_selfSource->matches(url);
}

bool ContentSecurityPolicy::protocolMatchesSelf(const KURL& url) const
{
#if ENABLE(CSP_NEXT)
    if (equalLettersIgnoringASCIICase(m_selfSourceProtocol, "http"))
        return url.protocolIsInHTTPFamily();
#endif
    return equalIgnoringASCIICase(url.protocol(), m_selfSourceProtocol);
}

namespace {
class ViolatedDirectiveCallback
{
public:
	ViolatedDirectiveCallback(const ContentSecurityPolicy& contentSecurityPolicy)
	: m_contentSecurityPolicy(contentSecurityPolicy)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective&)
	{
	}

	virtual ~ViolatedDirectiveCallback() { }

protected:
	void reportViolation(const String& effectiveViolatedDirective, const ContentSecurityPolicyDirective& violatedDirective, const KURL& blockedURL, const String& consoleMessage, const String& sourceURL, const TextPosition& sourcePosition) const
	{
		m_contentSecurityPolicy.reportViolation(effectiveViolatedDirective, violatedDirective, blockedURL, consoleMessage, sourceURL, sourcePosition);
	}

	const ContentSecurityPolicy& m_contentSecurityPolicy;
};
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)(const Frame&) const>
bool allPoliciesAllowObjectWithFrame(ViolatedDirectiveCallback& callback, const CSPDirectiveListVector& policies, const Frame& frame)
{
    bool isAllowed = true;
    for (size_t i = 0; i < policies.size(); ++i) {
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[i].get()->*predicate)(frame)) {
        	if (!violatedDirective->directiveList().isReportOnly())
				isAllowed = false;
			callback(*violatedDirective);
        }
    }
    return isAllowed;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)() const>
bool allPoliciesAllowObjectWithContext(ViolatedDirectiveCallback& callback, const CSPDirectiveListVector& policies)
{
    bool isAllowed = true;
    for (size_t i = 0; i < policies.size(); ++i) {
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[i].get()->*predicate)()) {
        	if (!violatedDirective->directiveList().isReportOnly())
				isAllowed = false;
			callback(*violatedDirective);
        }
    }
    return isAllowed;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)(const String& nonce) const>
bool allPoliciesWithDispositionAllow(ContentSecurityPolicy::Disposition disposition, ViolatedDirectiveCallback& callback, const CSPDirectiveListVector& policies, const String& nonce)
{
    bool isReportOnly = disposition == ContentSecurityPolicy::Disposition::ReportOnly;
    bool isAllowed = true;
    for (size_t i = 0; i < policies.size(); ++i) {
        if (policies[i]->isReportOnly() != isReportOnly)
            continue;
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[i].get()->*predicate)(nonce)) {
        	isAllowed = false;
        	callback(*violatedDirective);
        }
    }
    return isAllowed;
}

static CryptoDigest::Algorithm toCryptoDigestAlgorithm(ContentSecurityPolicyHashAlgorithm algorithm)
{
    switch (algorithm) {
    case ContentSecurityPolicyHashAlgorithm::SHA_256:
        return CryptoDigest::Algorithm::SHA_256;
    case ContentSecurityPolicyHashAlgorithm::SHA_384:
        return CryptoDigest::Algorithm::SHA_384;
    case ContentSecurityPolicyHashAlgorithm::SHA_512:
        return CryptoDigest::Algorithm::SHA_512;
    }
    ASSERT_NOT_REACHED();
    return CryptoDigest::Algorithm::SHA_512;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)(const KURL&, bool) const>
bool allPoliciesAllowWithURI(ViolatedDirectiveCallback& callback, const CSPDirectiveListVector& policies, const KURL& url, bool didReceiveRedirectResponse)
{
    bool isAllowed = true;
    for (size_t i = 0; i < policies.size(); ++i) {
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[i].get()->*predicate)(url, didReceiveRedirectResponse)) {
        	if (!violatedDirective->directiveList().isReportOnly())
				isAllowed = false;
			callback(*violatedDirective);
        }
    }
    return isAllowed;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)(const KURL&) const>
bool allPoliciesAllowWithBaseURI(ViolatedDirectiveCallback& callback, const CSPDirectiveListVector& policies, const KURL& url)
{
    bool isAllowed = true;
    for (size_t i = 0; i < policies.size(); ++i) {
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[i].get()->*predicate)(url)) {
        	if (!violatedDirective->directiveList().isReportOnly())
				isAllowed = false;
			callback(*violatedDirective);
        }
    }
    return isAllowed;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)(const KURL&, bool, ContentSecurityPolicySourceListDirective::ShouldAllowEmptyURLIfSourceListIsNotNone shouldAllowEmptyURLIfSourceListEmpty) const>
bool allPoliciesAllowObjectWithURL(ViolatedDirectiveCallback& callback, const CSPDirectiveListVector& policies, const KURL& url, bool didReceiveRedirectResponse, ContentSecurityPolicySourceListDirective::ShouldAllowEmptyURLIfSourceListIsNotNone shouldAllowEmptyURLIfSourceListEmpty)
{
    bool isAllowed = true;
    for (size_t i = 0; i < policies.size(); ++i) {
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[i].get()->*predicate)(url, didReceiveRedirectResponse, shouldAllowEmptyURLIfSourceListEmpty)) {
        	if (!violatedDirective->directiveList().isReportOnly())
				isAllowed = false;
			callback(*violatedDirective);
        }
    }
    return isAllowed;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)(const String&, const String&) const>
bool allPoliciesAllowObjectWithPlugin(ViolatedDirectiveCallback& callback, const CSPDirectiveListVector& policies, const String& type, const String& typeAttribute)
{
    bool isAllowed = true;
    for (size_t i = 0; i < policies.size(); ++i) {
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[i].get()->*predicate)(type, typeAttribute)) {
        	if (!violatedDirective->directiveList().isReportOnly())
				isAllowed = false;
			callback(*violatedDirective);
        }
    }
    return isAllowed;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*predicate)(const ContentSecurityPolicyHash&) const>
const ContentSecurityPolicyDirective* violatedDirectiveInAnyPolicyWithHash(const CSPDirectiveListVector& policies, const ContentSecurityPolicyHash& hash)
{
    for (int j = 0; j < policies.size(); ++j) {
        if (const ContentSecurityPolicyDirective* violatedDirective = (policies[j].get()->*predicate)(hash))
            return violatedDirective;
    }
    return 0;
}

template<const ContentSecurityPolicyDirective* (ContentSecurityPolicyDirectiveList::*allowed)(const ContentSecurityPolicyHash&) const>
bool foundHashOfContentInAllPolicies(const CSPDirectiveListVector& policies, const String& content, const TextEncoding& encoding, const HashSet<ContentSecurityPolicyHashAlgorithm>& algorithms)
{
    if (algorithms.isEmpty() || content.isEmpty())
        return false;

    // FIXME: Compute the digest with respect to the raw bytes received from the page.
    // See <https://bugs.webkit.org/show_bug.cgi?id=155184>.
    CString contentCString = encoding.encode(content, EntitiesForUnencodables);
    for (HashSet<ContentSecurityPolicyHashAlgorithm>::const_iterator it = algorithms.begin(); it != algorithms.end(); ++it) {
    	ContentSecurityPolicyHashAlgorithm algorithm = *it;
    	OwnPtr<CryptoDigest> cryptoDigest = CryptoDigest::create(toCryptoDigestAlgorithm(algorithm));
        cryptoDigest->addBytes(contentCString.data(), contentCString.length());
        Vector<uint8_t> digest = cryptoDigest->computeHash();
        if (!violatedDirectiveInAnyPolicyWithHash<allowed>(policies, std::make_pair(algorithm, digest)))
			return true;
    }
    return false;
}

namespace {
class HandleViolatedDirectiveForJavaScriptURLs : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForJavaScriptURLs(const ContentSecurityPolicy& contentSecurityPolicy, const String& contextURL, const WTF::OrdinalNumber& contextLine)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_contextURL(contextURL)
	, m_contextLine(contextLine)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), "Refused to execute a script", "its hash, its nonce, or 'unsafe-inline'");
		m_contentSecurityPolicy.reportViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), consoleMessage, m_contextURL, TextPosition(m_contextLine, WTF::OrdinalNumber()));
		bool isReportOnly = violatedDirective.directiveList().isReportOnly();
		if (!isReportOnly)
			m_contentSecurityPolicy.reportBlockedScriptExecutionToInspector(violatedDirective.text());
	}

private:
	const String& m_contextURL;
	const WTF::OrdinalNumber& m_contextLine;
};
}

bool ContentSecurityPolicy::allowJavaScriptURLs(const String& contextURL, const WTF::OrdinalNumber& contextLine, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    HandleViolatedDirectiveForJavaScriptURLs callback(*this, contextURL, contextLine);
    return allPoliciesAllowObjectWithContext<&ContentSecurityPolicyDirectiveList::violatedDirectiveForUnsafeInlineScript>(callback, m_policies);
}

namespace {
class HandleViolatedDirectiveForInlineEventHandlers : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForInlineEventHandlers(const ContentSecurityPolicy& contentSecurityPolicy, const String& contextURL, const WTF::OrdinalNumber& contextLine)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_contextURL(contextURL)
	, m_contextLine(contextLine)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), "Refused to execute a script for an inline event handler", "'unsafe-inline'");
		m_contentSecurityPolicy.reportViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), consoleMessage, m_contextURL, TextPosition(m_contextLine, WTF::OrdinalNumber()));
		bool isReportOnly = violatedDirective.directiveList().isReportOnly();
		if (!isReportOnly)
			m_contentSecurityPolicy.reportBlockedScriptExecutionToInspector(violatedDirective.text());
	}

private:
	const String& m_contextURL;
	const WTF::OrdinalNumber& m_contextLine;
};
}

bool ContentSecurityPolicy::allowInlineEventHandlers(const String& contextURL, const WTF::OrdinalNumber& contextLine, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    HandleViolatedDirectiveForInlineEventHandlers callback(*this, contextURL, contextLine);
    return allPoliciesAllowObjectWithContext<&ContentSecurityPolicyDirectiveList::violatedDirectiveForUnsafeInlineScript>(callback, m_policies);
}

bool ContentSecurityPolicy::allowScriptWithNonce(const String& nonce, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    String strippedNonce = stripLeadingAndTrailingHTMLSpaces(nonce);
    if (strippedNonce.isEmpty())
        return false;
    ViolatedDirectiveCallback callback(*this);
    // FIXME: We need to report violations in report-only policies. See <https://bugs.webkit.org/show_bug.cgi?id=159830>.
    return allPoliciesWithDispositionAllow<&ContentSecurityPolicyDirectiveList::violatedDirectiveForScriptNonce>(ContentSecurityPolicy::Disposition::Enforce, callback, m_policies, strippedNonce);
}

bool ContentSecurityPolicy::allowStyleWithNonce(const String& nonce, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    String strippedNonce = stripLeadingAndTrailingHTMLSpaces(nonce);
    if (strippedNonce.isEmpty())
        return false;
    ViolatedDirectiveCallback callback(*this);
    // FIXME: We need to report violations in report-only policies. See <https://bugs.webkit.org/show_bug.cgi?id=159830>.
    return allPoliciesWithDispositionAllow<&ContentSecurityPolicyDirectiveList::violatedDirectiveForStyleNonce>(ContentSecurityPolicy::Disposition::Enforce, callback, m_policies, strippedNonce);
}

// FIXME: We should compute the document encoding once and cache it instead of computing it on each invocation.
const TextEncoding& ContentSecurityPolicy::documentEncoding() const
{
    if (!m_scriptExecutionContext->isDocument())
        return UTF8Encoding();
    Document& document = static_cast<Document&>(*m_scriptExecutionContext);
    if (TextResourceDecoder* decoder = document.decoder())
        return decoder->encoding();
    return UTF8Encoding();
}

namespace {
class HandleViolatedDirectiveForInlineScript : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForInlineScript(const ContentSecurityPolicy& contentSecurityPolicy, const String& contextURL, const WTF::OrdinalNumber& contextLine)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_contextURL(contextURL)
	, m_contextLine(contextLine)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), "Refused to execute a script", "its hash, its nonce, or 'unsafe-inline'");
		m_contentSecurityPolicy.reportViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), consoleMessage, m_contextURL, TextPosition(m_contextLine, WTF::OrdinalNumber()));
		bool isReportOnly = violatedDirective.directiveList().isReportOnly();
		if (!isReportOnly)
			m_contentSecurityPolicy.reportBlockedScriptExecutionToInspector(violatedDirective.text());
	}

private:
	const String& m_contextURL;
	const WTF::OrdinalNumber& m_contextLine;
};
}

bool ContentSecurityPolicy::allowInlineScript(const String& contextURL, const WTF::OrdinalNumber& contextLine, const String& scriptContent, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (foundHashOfContentInAllPolicies<&ContentSecurityPolicyDirectiveList::violatedDirectiveForScriptHash>(m_policies, scriptContent, documentEncoding(), m_hashAlgorithmsForInlineScripts))
        return true;
    HandleViolatedDirectiveForInlineScript callback(*this, contextURL, contextLine);
    return allPoliciesAllowObjectWithContext<&ContentSecurityPolicyDirectiveList::violatedDirectiveForUnsafeInlineScript>(callback, m_policies);
}

namespace {
class HandleViolatedDirectiveForInlineStyle : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForInlineStyle(const ContentSecurityPolicy& contentSecurityPolicy, const String& contextURL, const WTF::OrdinalNumber& contextLine)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_contextURL(contextURL)
	, m_contextLine(contextLine)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::styleSrc, violatedDirective, KURL(), "Refused to apply a stylesheet", "its hash, its nonce, or 'unsafe-inline'");
		reportViolation(ContentSecurityPolicyDirectiveNames::styleSrc, violatedDirective, KURL(), consoleMessage, m_contextURL, TextPosition(m_contextLine, WTF::OrdinalNumber()));
	}

private:
	const String& m_contextURL;
	const WTF::OrdinalNumber& m_contextLine;
};
}

bool ContentSecurityPolicy::allowInlineStyle(const String& contextURL, const WTF::OrdinalNumber& contextLine, const String& styleContent, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (m_overrideInlineStyleAllowed)
        return true;
    if (foundHashOfContentInAllPolicies<&ContentSecurityPolicyDirectiveList::violatedDirectiveForStyleHash>(m_policies, styleContent, documentEncoding(), m_hashAlgorithmsForInlineStylesheets))
        return true;
    HandleViolatedDirectiveForInlineStyle callback(*this, contextURL, contextLine);
    return allPoliciesAllowObjectWithContext<&ContentSecurityPolicyDirectiveList::violatedDirectiveForUnsafeInlineStyle>(callback, m_policies);
}

namespace {
class HandleViolatedDirectiveForEval : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForEval(const ContentSecurityPolicy& contentSecurityPolicy, PassRefPtr<ScriptCallStack> callStack)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_callStack(callStack)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), "Refused to execute a script", "'unsafe-eval'");
		m_contentSecurityPolicy.reportViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, KURL(), consoleMessage, m_callStack);
		bool isReportOnly = violatedDirective.directiveList().isReportOnly();
		if (!isReportOnly)
			m_contentSecurityPolicy.reportBlockedScriptExecutionToInspector(violatedDirective.text());
	}

private:
	RefPtr<ScriptCallStack> m_callStack;
};
}

bool ContentSecurityPolicy::allowEval(PassRefPtr<ScriptCallStack> callStack, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    HandleViolatedDirectiveForEval callback(*this, callStack);
    return allPoliciesAllowObjectWithContext<&ContentSecurityPolicyDirectiveList::violatedDirectiveForUnsafeEval>(callback, m_policies);
}

namespace {
class HandleViolatedDirectiveForFrameAncestors : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForFrameAncestors(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::frameAncestors, violatedDirective, m_url, "Refused to load");
		reportViolation(ContentSecurityPolicyDirectiveNames::frameAncestors, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowFrameAncestors(const Frame& frame, const KURL& url, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    const Frame& topFrame = *frame.tree()->top();
    if (&frame == &topFrame)
        return true;
    HandleViolatedDirectiveForFrameAncestors callback(*this, url);
    return allPoliciesAllowObjectWithFrame<&ContentSecurityPolicyDirectiveList::violatedDirectiveForFrameAncestor>(callback, m_policies, frame);
}

namespace {
class HandleViolatedDirectiveForPluginType : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForPluginType(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::pluginTypes, violatedDirective, m_url, "Refused to load", "its MIME type");
		reportViolation(ContentSecurityPolicyDirectiveNames::pluginTypes, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowPluginType(const String& type, const String& typeAttribute, const KURL& url, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    HandleViolatedDirectiveForPluginType callback(*this, url);
    return allPoliciesAllowObjectWithPlugin<&ContentSecurityPolicyDirectiveList::violatedDirectiveForPluginType>(callback, m_policies, type, typeAttribute);
}

namespace {
class HandleViolatedDirectiveForScriptFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForScriptFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, m_url, "Refused to load");
		reportViolation(ContentSecurityPolicyDirectiveNames::scriptSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowScriptFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForScriptFromSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForScript>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForObjectFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForObjectFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::objectSrc, violatedDirective, m_url, "Refused to load");
		reportViolation(ContentSecurityPolicyDirectiveNames::objectSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowObjectFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForObjectFromSource callback(*this, url);
	// As per section object-src of the Content Security Policy Level 3 spec., <http://w3c.github.io/webappsec-csp> (Editor's Draft, 29 February 2016),
	// "If plugin content is loaded without an associated KURL (perhaps an object element lacks a data attribute, but loads some default plugin based
	// on the specified type), it MUST be blocked if object-src's value is 'none', but will otherwise be allowed".
    return allPoliciesAllowObjectWithURL<&ContentSecurityPolicyDirectiveList::violatedDirectiveForObjectSource>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes, ContentSecurityPolicySourceListDirective::ShouldAllowEmptyURLIfSourceListIsNotNone::Yes);
}

namespace {
class HandleViolatedDirectiveForChildFrameFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForChildFrameFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::frameSrc, violatedDirective, m_url, "Refused to load");
		reportViolation(ContentSecurityPolicyDirectiveNames::frameSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowChildFrameFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForChildFrameFromSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForFrame>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForChildContextFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForChildContextFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::childSrc, violatedDirective, m_url, "Refused to load");
		reportViolation(ContentSecurityPolicyDirectiveNames::childSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowChildContextFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForChildContextFromSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForChildContext>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForImageFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForImageFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	 , m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
	    String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::imgSrc, violatedDirective, m_url, "Refused to load");
	    reportViolation(ContentSecurityPolicyDirectiveNames::imgSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowImageFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForImageFromSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForImage>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForStyleFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForStyleFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
	    String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::styleSrc, violatedDirective, m_url, "Refused to load");
	    reportViolation(ContentSecurityPolicyDirectiveNames::styleSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowStyleFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForStyleFromSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForStyle>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForFontFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForFontFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
	    String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::fontSrc, violatedDirective, m_url, "Refused to load");
	    reportViolation(ContentSecurityPolicyDirectiveNames::fontSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowFontFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForFontFromSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForFont>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForMediaFromSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForMediaFromSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
	    String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::mediaSrc, violatedDirective, m_url, "Refused to load");
	    reportViolation(ContentSecurityPolicyDirectiveNames::mediaSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowMediaFromSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForMediaFromSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForMedia>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForConnectToSource : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForConnectToSource(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
	    String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::connectSrc, violatedDirective, m_url, "Refused to connect to");
	    reportViolation(ContentSecurityPolicyDirectiveNames::connectSrc, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowConnectToSource(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForConnectToSource callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForConnectSource>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForFormAction : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForFormAction(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
	    String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::formAction, violatedDirective, m_url, "Refused to load");
	    reportViolation(ContentSecurityPolicyDirectiveNames::formAction, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowFormAction(const KURL& url, bool overrideContentSecurityPolicy, RedirectResponseReceived redirectResponseReceived) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForFormAction callback(*this, url);
    return allPoliciesAllowWithURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForFormAction>(callback, m_policies, url, redirectResponseReceived == RedirectResponseReceived::Yes);
}

namespace {
class HandleViolatedDirectiveForBaseURI : public ViolatedDirectiveCallback
{
public:
	HandleViolatedDirectiveForBaseURI(const ContentSecurityPolicy& contentSecurityPolicy, const KURL& url)
	: ViolatedDirectiveCallback(contentSecurityPolicy)
	, m_url(url)
	{
	}

	virtual void operator()(const ContentSecurityPolicyDirective& violatedDirective) {
		String consoleMessage = consoleMessageForViolation(ContentSecurityPolicyDirectiveNames::baseURI, violatedDirective, m_url, "Refused to change the document base URL to");
		reportViolation(ContentSecurityPolicyDirectiveNames::baseURI, violatedDirective, m_url, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber()));
	}

private:
	const KURL& m_url;
};
}

bool ContentSecurityPolicy::allowBaseURI(const KURL& url, bool overrideContentSecurityPolicy) const
{
    if (overrideContentSecurityPolicy)
        return true;
    if (SchemeRegistry::schemeShouldBypassContentSecurityPolicy(url.protocol().toStringWithoutCopying()))
        return true;
    HandleViolatedDirectiveForBaseURI callback(*this, url);
    return allPoliciesAllowWithBaseURI<&ContentSecurityPolicyDirectiveList::violatedDirectiveForBaseURI>(callback, m_policies, url);
}

bool ContentSecurityPolicy::isActive() const
{
    return !m_policies.isEmpty();
}

#if ENABLE(CSP_NEXT)
static String stripURLForUseInReport(Document* document, const KURL& url)
{
    if (!url.isValid())
        return String();
    if (!url.isHierarchical() || url.protocolIs("file"))
        return url.protocol().toString();
    return document->securityOrigin()->canRequest(url) ? url.strippedForUseAsReferrer() : SecurityOrigin::create(url)->toString();
}

#endif

static bool isDocument(const ScriptExecutionContext* scriptExecutionContext)
{
	return scriptExecutionContext && scriptExecutionContext->isDocument();
}

void ContentSecurityPolicy::reportViolation(const String& violatedDirective, const ContentSecurityPolicyDirective& effectiveViolatedDirective, const KURL& blockedURL, const String& consoleMessage, PassRefPtr<ScriptCallStack> callStack) const
{
	// FIXME: Extract source file and source position from JSC::ExecState.
	return reportViolation(violatedDirective, effectiveViolatedDirective, blockedURL, consoleMessage, String(), TextPosition(WTF::OrdinalNumber::beforeFirst(), WTF::OrdinalNumber::beforeFirst()), callStack);
}

void ContentSecurityPolicy::reportViolation(const String& effectiveViolatedDirective, const ContentSecurityPolicyDirective& violatedDirective, const KURL& blockedURL, const String& consoleMessage, const String& sourceURL, const TextPosition& sourcePosition, PassRefPtr<ScriptCallStack> callStack) const
{
	logToConsole(consoleMessage, sourceURL, sourcePosition.m_line, callStack);

    if (!m_isReportingEnabled)
        return;

    // FIXME: Support sending reports from worker.
    if (!isDocument(m_scriptExecutionContext) && !m_frame)
        return;

    ASSERT(!m_frame || effectiveViolatedDirective == ContentSecurityPolicyDirectiveNames::frameAncestors);

    Document& document = isDocument(m_scriptExecutionContext) ? static_cast<Document&>(*m_scriptExecutionContext) : *m_frame->document();
    Frame* frame = document.frame();
    ASSERT(!m_frame || m_frame == frame);
    if (!frame)
        return;

    String documentURI;
    String blockedURI;
    if (isDocument(m_scriptExecutionContext)) {
        documentURI = document.url().strippedForUseAsReferrer();
        blockedURI = stripURLForUseInReport(&document, blockedURL);
    } else {
        // The KURL of |document| may not have been initialized (say, when reporting a frame-ancestors violation).
        // So, we use the KURL of the blocked document for the protected document KURL.
        documentURI = blockedURL;
        blockedURI = blockedURL;
    }
    String violatedDirectiveText = violatedDirective.text();
    String originalPolicy = violatedDirective.directiveList().header();
    String referrer = document.referrer();

    ASSERT(document.loader());
    unsigned short statusCode = document.url().protocolIs("http") && document.loader() ? document.loader()->response().httpStatusCode() : 0;

    String sourceFile;
    int lineNumber = 0;
    int columnNumber = 0;
    RefPtr<ScriptCallStack> stack = createScriptCallStack(JSMainThreadExecState::currentState(), 2);
    const ScriptCallFrame* callFrame = stack->firstNonNativeCallFrame();
    if (callFrame && callFrame->lineNumber()) {
		sourceFile = stripURLForUseInReport(&document, KURL(KURL(), callFrame->sourceURL()));
		lineNumber = callFrame->lineNumber();
        columnNumber = callFrame->columnNumber();
	}

    // 1. Dispatch violation event.
    bool canBubble = false;
    bool cancelable = false;
    document.enqueueDocumentEvent(SecurityPolicyViolationEvent::create(eventNames().securitypolicyviolationEvent, canBubble, cancelable, documentURI, referrer, blockedURI, violatedDirectiveText, effectiveViolatedDirective, originalPolicy, sourceFile, statusCode, lineNumber, columnNumber));

    // 2. Send violation report (if applicable).
    const Vector<String>& reportURIs = violatedDirective.directiveList().reportURIs();
	if (reportURIs.isEmpty())
        return;

    // We need to be careful here when deciding what information to send to the
    // report-uri. Currently, we send only the current document's KURL and the
    // directive that was violated. The document's KURL is safe to send because
    // it's the document itself that's requesting that it be sent. You could
    // make an argument that we shouldn't send HTTPS document URLs to HTTP
    // report-uris (for the same reasons that we supress the Referer in that
    // case), but the Referer is sent implicitly whereas this request is only
    // sent explicitly. As for which directive was violated, that's pretty
    // harmless information.

    RefPtr<InspectorObject> cspReport = InspectorObject::create();
    cspReport->setString(ASCIILiteral("document-uri"), documentURI);
    cspReport->setString(ASCIILiteral("referrer"), referrer);
    cspReport->setString(ASCIILiteral("violated-directive"), violatedDirectiveText);
    cspReport->setString(ASCIILiteral("effective-directive"), effectiveViolatedDirective);
    cspReport->setString(ASCIILiteral("original-policy"), originalPolicy);
    cspReport->setString(ASCIILiteral("blocked-uri"), blockedURI);
    cspReport->setInteger(ASCIILiteral("status-code"), statusCode);
    if (!sourceFile.isNull()) {
        cspReport->setString(ASCIILiteral("source-file"), sourceFile);
        cspReport->setInteger(ASCIILiteral("line-number"), lineNumber);
        cspReport->setInteger(ASCIILiteral("column-number"), columnNumber);
    }

    RefPtr<InspectorObject> reportObject = InspectorObject::create();
    reportObject->setObject("csp-report", cspReport.release());

    RefPtr<FormData> report = FormData::create(reportObject->toJSONString().utf8());
    for (size_t i = 0; i < reportURIs.size(); ++i)
        PingLoader::sendViolationReport(frame, isDocument(m_scriptExecutionContext) ? document.completeURL(reportURIs[i]) : document.completeURL(reportURIs[i], blockedURL), report);
}

void ContentSecurityPolicy::reportUnsupportedDirective(const String& name) const
{
    DEFINE_STATIC_LOCAL(String, allow, (ASCIILiteral("allow")));
    DEFINE_STATIC_LOCAL(String, options, (ASCIILiteral("options")));
    DEFINE_STATIC_LOCAL(String, policyURI, (ASCIILiteral("policy-uri")));
    DEFINE_STATIC_LOCAL(String, allowMessage, (ASCIILiteral("The 'allow' directive has been replaced with 'default-src'. Please use that directive instead, as 'allow' has no effect.")));
    DEFINE_STATIC_LOCAL(String, optionsMessage, (ASCIILiteral("The 'options' directive has been replaced with 'unsafe-inline' and 'unsafe-eval' source expressions for the 'script-src' and 'style-src' directives. Please use those directives instead, as 'options' has no effect.")));
    DEFINE_STATIC_LOCAL(String, policyURIMessage, (ASCIILiteral("The 'policy-uri' directive has been removed from the specification. Please specify a complete policy via the Content-Security-Policy header.")));

    String message = makeString("Unrecognized Content-Security-Policy directive '", name, "'.\n");
    if (equalIgnoringCase(name, allow))
        message = allowMessage;
    else if (equalIgnoringCase(name, options))
        message = optionsMessage;
    else if (equalIgnoringCase(name, policyURI))
        message = policyURIMessage;

    logToConsole(message);
}

void ContentSecurityPolicy::reportDirectiveAsSourceExpression(const String& directiveName, const String& sourceExpression) const
{
    String message = "The Content Security Policy directive '" + directiveName + "' contains '" + sourceExpression + "' as a source expression. Did you mean '" + directiveName + " ...; " + sourceExpression + "...' (note the semicolon)?";
    logToConsole(message);
}

void ContentSecurityPolicy::reportDuplicateDirective(const String& name) const
{
    String message = makeString("Ignoring duplicate Content-Security-Policy directive '", name, "'.\n");
    logToConsole(message);
}

void ContentSecurityPolicy::reportInvalidPluginTypes(const String& pluginType) const
{
    String message;
    if (pluginType.isNull())
        message = "'plugin-types' Content Security Policy directive is empty; all plugins will be blocked.\n";
    else
        message = makeString("Invalid plugin type in 'plugin-types' Content Security Policy directive: '", pluginType, "'.\n");
    logToConsole(message);
}

void ContentSecurityPolicy::reportInvalidSandboxFlags(const String& invalidFlags) const
{
    logToConsole("Error while parsing the 'sandbox' Content Security Policy directive: " + invalidFlags);
}

void ContentSecurityPolicy::reportInvalidDirectiveInReportOnlyMode(const String& directiveName) const
{
    logToConsole("The Content Security Policy directive '" + directiveName + "' is ignored when delivered in a report-only policy.");
}

void ContentSecurityPolicy::reportInvalidDirectiveInHTTPEquivMeta(const String& directiveName) const
{
    logToConsole("The Content Security Policy directive '" + directiveName + "' is ignored when delivered via an HTML meta element.");
}

void ContentSecurityPolicy::reportInvalidDirectiveValueCharacter(const String& directiveName, const String& value) const
{
    String message = makeString("The value for Content Security Policy directive '", directiveName, "' contains an invalid character: '", value, "'. Non-whitespace characters outside ASCII 0x21-0x7E must be percent-encoded, as described in RFC 3986, section 2.1: http://tools.ietf.org/html/rfc3986#section-2.1.");
    logToConsole(message);
}

void ContentSecurityPolicy::reportInvalidPathCharacter(const String& directiveName, const String& value, const char invalidChar) const
{
    ASSERT(invalidChar == '#' || invalidChar == '?');

    String ignoring = "The fragment identifier, including the '#', will be ignored.";
    if (invalidChar == '?')
        ignoring = "The query component, including the '?', will be ignored.";
    String message = makeString("The source list for Content Security Policy directive '", directiveName, "' contains a source with an invalid path: '", value, "'. ", ignoring);
    logToConsole(message);
}

void ContentSecurityPolicy::reportInvalidSourceExpression(const String& directiveName, const String& source) const
{
    String message = makeString("The source list for Content Security Policy directive '", directiveName, "' contains an invalid source: '", source, "'. It will be ignored.");
    if (equalIgnoringCase(source, "'none'"))
		message = makeString(message, " Note that 'none' has no effect unless it is the only expression in the source list.");
    logToConsole(message);
}

void ContentSecurityPolicy::reportMissingReportURI(const String& policy) const
{
    logToConsole("The Content Security Policy '" + policy + "' was delivered in report-only mode, but does not specify a 'report-uri'; the policy will have no effect. Please either add a 'report-uri' directive, or deliver the policy via the 'Content-Security-Policy' header.");
}

void ContentSecurityPolicy::logToConsole(const String& message, const String& contextURL, const WTF::OrdinalNumber& contextLine, PassRefPtr<ScriptCallStack> callStack) const
{
    if (!m_isReportingEnabled)
        return;

	if (m_scriptExecutionContext)
		m_scriptExecutionContext->addConsoleMessage(JSMessageSource, LogMessageType, ErrorMessageLevel, message, contextURL, contextLine.oneBasedInt(), 0, callStack);
}

void ContentSecurityPolicy::reportBlockedScriptExecutionToInspector(const String& directiveText) const
{
	if (m_scriptExecutionContext)
		InspectorInstrumentation::scriptExecutionBlockedByCSP(m_scriptExecutionContext, directiveText);
}

void ContentSecurityPolicy::upgradeInsecureRequestIfNeeded(ResourceRequest& request, InsecureRequestType requestType) const
{
    KURL url = request.url();
    upgradeInsecureRequestIfNeeded(url, requestType);
    request.setURL(url);
}

void ContentSecurityPolicy::upgradeInsecureRequestIfNeeded(KURL& url, InsecureRequestType requestType) const
{
    if (!url.protocolIs("http") && !url.protocolIs("ws"))
        return;

    bool upgradeRequest = m_insecureNavigationRequestsToUpgrade.contains(SecurityOriginData::fromURL(url));
    if (requestType == InsecureRequestType::Load || requestType == InsecureRequestType::FormSubmission)
        upgradeRequest |= m_upgradeInsecureRequests;

    if (!upgradeRequest)
        return;

    if (url.protocolIs("http"))
        url.setProtocol("https");
    else if (url.protocolIs("ws"))
        url.setProtocol("wss");
    else
        return;

    if (url.port() && url.port().value() == 80)
        url.setPort(443);
}

void ContentSecurityPolicy::setUpgradeInsecureRequests(bool upgradeInsecureRequests)
{
    m_upgradeInsecureRequests = upgradeInsecureRequests;
    if (!m_upgradeInsecureRequests)
        return;

    if (!m_scriptExecutionContext)
        return;

    // Store the upgrade domain as an 'insecure' protocol so we can quickly identify
    // origins we should upgrade.
    KURL upgradeURL = m_scriptExecutionContext->url();
    if (upgradeURL.protocolIs("https"))
        upgradeURL.setProtocol("http");
    else if (upgradeURL.protocolIs("wss"))
        upgradeURL.setProtocol("ws");

    m_insecureNavigationRequestsToUpgrade.add(SecurityOriginData::fromURL(upgradeURL));
}

void ContentSecurityPolicy::inheritInsecureNavigationRequestsToUpgradeFromOpener(const ContentSecurityPolicy& other)
{
    m_insecureNavigationRequestsToUpgrade.add(other.m_insecureNavigationRequestsToUpgrade.begin(), other.m_insecureNavigationRequestsToUpgrade.end());
}

void ContentSecurityPolicy::takeNavigationRequestsToUpgrade(HashSet<SecurityOriginData>& insecureNavigationRequests)
{
    m_insecureNavigationRequestsToUpgrade.swap(insecureNavigationRequests);
}

void ContentSecurityPolicy::setInsecureNavigationRequestsToUpgrade(const HashSet<SecurityOriginData>& insecureNavigationRequests)
{
    m_insecureNavigationRequestsToUpgrade = insecureNavigationRequests;
}

}
