/*
 * Copyright (C) 2011 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ScrollingCoordinator_h
#define ScrollingCoordinator_h

#include "GraphicsLayer.h"
#include "IntRect.h"
#include "LayoutTypes.h"
#include "PlatformWheelEvent.h"
#include "ScrollTypes.h"
#include "Timer.h"
#include <wtf/Forward.h>
#include <wtf/text/WTFString.h>

#if ENABLE(THREADED_SCROLLING)
#include <wtf/HashMap.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Threading.h>
#endif

#if PLATFORM(MAC)
#include <wtf/RetainPtr.h>
#endif

namespace WebCore {

typedef unsigned MainThreadScrollingReasons;
typedef uint64_t ScrollingNodeID;

enum ScrollingNodeType { ScrollingNode, FixedNode, StickyNode };

class Document;
class FrameView;
class GraphicsLayer;
class Page;
class Region;
class ScrollableArea;
class ScrollingCoordinatorPrivate;
class ScrollingStateNode;
class ScrollingStateScrollingNode;
class ScrollingTreeState;
class ViewportConstraints;

#if ENABLE(THREADED_SCROLLING)
class ScrollingTree;
#endif

class ScrollingCoordinator : public ThreadSafeRefCounted<ScrollingCoordinator> {
public:
    static PassRefPtr<ScrollingCoordinator> create(Page*);
    virtual ~ScrollingCoordinator();

    void pageDestroyed();

#if ENABLE(THREADED_SCROLLING)
    ScrollingTree* scrollingTree() const;
    void commitTreeStateIfNeeded();
#endif

    // Return whether this scrolling coordinator handles scrolling for the given frame view.
    bool coordinatesScrollingForFrameView(FrameView*) const;

    // Should be called whenever the given frame view has been laid out.
    void frameViewLayoutUpdated(FrameView*);

    // Should be called whenever a wheel event handler is added or removed in the 
    // frame view's underlying document.
    void frameViewWheelEventHandlerCountChanged(FrameView*);

    // Should be called whenever the slow repaint objects counter changes between zero and one.
    void frameViewHasSlowRepaintObjectsDidChange(FrameView*);

    // Should be called whenever the set of fixed objects changes.
    void frameViewFixedObjectsDidChange(FrameView*);

    // Should be called whenever the root layer for the given frame view changes.
    void frameViewRootLayerDidChange(FrameView*);

    // Should be called whenever the horizontal scrollbar layer for the given frame view changes.
    void frameViewHorizontalScrollbarLayerDidChange(FrameView*, GraphicsLayer* horizontalScrollbarLayer);

    // Should be called whenever the vertical scrollbar layer for the given frame view changes.
    void frameViewVerticalScrollbarLayerDidChange(FrameView*, GraphicsLayer* verticalScrollbarLayer);

    // Dispatched by the scrolling tree whenever the main frame scroll position changes.
    void updateMainFrameScrollPosition(const IntPoint&);

    // Dispatched by the scrolling tree whenever the main frame scroll position changes and the scroll layer position needs to be updated as well.
    void updateMainFrameScrollPositionAndScrollLayerPosition();

#if PLATFORM(MAC) || (PLATFORM(CHROMIUM) && OS(DARWIN))
    // Dispatched by the scrolling tree during handleWheelEvent. This is required as long as scrollbars are painted on the main thread.
    void handleWheelEventPhase(PlatformWheelEventPhase);
#endif

    // Force all scroll layer position updates to happen on the main thread.
    void setForceMainThreadScrollLayerPositionUpdates(bool);

    // These virtual functions are currently unique to the threaded scrolling architecture.
    // Their meaningful implementations are in ScrollingCoordinatorMac.
    virtual void commitTreeStateIfNeeded() { }
    virtual bool requestScrollPositionUpdate(FrameView*, const IntPoint&) { return false; }
    virtual bool handleWheelEvent(FrameView*, const PlatformWheelEvent&) { return true; }
    virtual ScrollingNodeID attachToStateTree(ScrollingNodeType, ScrollingNodeID newNodeID, ScrollingNodeID /*parentID*/) { return newNodeID; }
    virtual void detachFromStateTree(ScrollingNodeID) { }
    virtual void clearStateTree() { }
    virtual void updateViewportConstrainedNode(ScrollingNodeID, const ViewportConstraints&, GraphicsLayer*) { }
    virtual void updateScrollingNode(ScrollingNodeID, GraphicsLayer* /*scrollLayer*/, GraphicsLayer* /*counterScrollingLayer*/) { }
    virtual void syncChildPositions(const LayoutRect&) { }
    virtual String scrollingStateTreeAsText() const { return String(); }

    // Generates a unique id to associate with a layer that will be represented by a node in the scrolling tree.
    ScrollingNodeID uniqueScrollLayerID();

    // Return whether this scrolling coordinator can keep fixed position layers fixed to their
    // containers while scrolling.
    bool supportsFixedPositionLayers() const;

    enum MainThreadScrollingReasonFlags {
        ForcedOnMainThread = 1 << 0,
        HasSlowRepaintObjects = 1 << 1,
        HasViewportConstrainedObjectsWithoutSupportingFixedLayers = 1 << 2,
        HasNonLayerFixedObjects = 1 << 3,
        IsImageDocument = 1 << 4
    };

    virtual void setScrollPinningBehavior(ScrollPinningBehavior) { }

    // These virtual functions are currently unique to Chromium's WebLayer approach. Their meaningful
    // implementations are in ScrollingCoordinatorChromium.
    virtual void willDestroyScrollableArea(ScrollableArea*) { }
    virtual void scrollableAreaScrollLayerDidChange(ScrollableArea*) { }
    virtual void scrollableAreaScrollbarLayerDidChange(ScrollableArea*, ScrollbarOrientation) { }
    virtual void setLayerIsContainerForFixedPositionLayers(GraphicsLayer*, bool) { }
    virtual void setLayerIsFixedToContainerLayer(GraphicsLayer*, bool) { }
    virtual void touchEventTargetRectsDidChange(const Document*) { }

private:
    explicit ScrollingCoordinator(Page*);

#if USE(ACCELERATED_COMPOSITING)
    static GraphicsLayer* scrollLayerForScrollableArea(ScrollableArea*);
    static GraphicsLayer* horizontalScrollbarLayerForScrollableArea(ScrollableArea*);
    static GraphicsLayer* verticalScrollbarLayerForScrollableArea(ScrollableArea*);
#endif

    void ensureRootStateNodeForFrameView(FrameView*);
    ScrollingStateScrollingNode* stateNodeForID(ScrollingNodeID);

    void recomputeWheelEventHandlerCountForFrameView(FrameView*);
    virtual bool hasVisibleSlowRepaintFixedObjects(FrameView*) const;
    void updateShouldUpdateScrollLayerPositionOnMainThread();

    struct ScrollParameters {
        ScrollElasticity horizontalScrollElasticity;
        ScrollElasticity verticalScrollElasticity;

        bool hasEnabledHorizontalScrollbar;
        bool hasEnabledVerticalScrollbar;

        ScrollbarMode horizontalScrollbarMode;
        ScrollbarMode verticalScrollbarMode;

        IntPoint scrollOrigin;

        IntRect viewportRect;
        IntSize contentsSize;
    };

    // FIXME: These functions are used only by ScrollingCoordinatorChromium. ScrollingCoordinator
    // now uses new versions: setScrollParametersForNode(), setWheelEventHandlerCountForNode(), etc.
    // To resolve this and all of the if-defs in this class, we should make ScrollingCoordinator
    // just be a typedef that defines ScrollingCoordinatorMac to be ScrollingCoordinator, and same
    // for Chromium and other platforms. Any shared functionality can be in something called
    // ScrollingCoordinatorBase which all of the platform implementations will inherit from.
    // https://bugs.webkit.org/show_bug.cgi?id=98700
    void setScrollParameters(const ScrollParameters&);
    void setWheelEventHandlerCount(unsigned);
    void setNonFastScrollableRegion(const Region&);
    void setScrollLayer(GraphicsLayer*);

#if ENABLE(THREADED_SCROLLING)
    void setScrollLayerForNode(GraphicsLayer*, ScrollingStateNode*);
    void setNonFastScrollableRegionForNode(const Region&, ScrollingStateScrollingNode*);
    void setScrollParametersForNode(const ScrollParameters&, ScrollingStateScrollingNode*);
    void setWheelEventHandlerCountForNode(unsigned, ScrollingStateScrollingNode*);
#endif

    void setShouldUpdateScrollLayerPositionOnMainThread(bool);

    Page* m_page;

    bool m_forceMainThreadScrollLayerPositionUpdates;

#if ENABLE(THREADED_SCROLLING)
    void scheduleTreeStateCommit();

    void scrollingTreeStateCommitterTimerFired(Timer<ScrollingCoordinator>*);
    void commitTreeState();

    OwnPtr<ScrollingTreeState> m_scrollingTreeState;
    RefPtr<ScrollingTree> m_scrollingTree;
    Timer<ScrollingCoordinator> m_scrollingTreeStateCommitterTimer;

    HashMap<ScrollingNodeID, ScrollingStateNode*> m_stateNodeMap;
#endif

    ScrollingCoordinatorPrivate* m_private;
};

} // namespace WebCore

#endif // ScrollingCoordinator_h
