/*
 * Copyright (C) 2015, 2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef GenericTaskQueue_h
#define GenericTaskQueue_h

#include "Timer.h"
#include <wtf/Deque.h>
#include <wtf/FastAllocBase.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/WeakPtr.h>

namespace WebCore {

template <typename T>
class GenericTask
{
    WTF_MAKE_FAST_ALLOCATED;
public:
	GenericTask(WeakPtr<T> t)
	: weakThis(t)
	{
	}

	virtual ~GenericTask() { };
	virtual void exec() = 0;

protected:
    WeakPtr<T> weakThis;
};

template <typename T>
class GenericTaskQueue {
public:
    GenericTaskQueue()
    : m_dispatchTaskTimer(this, &GenericTaskQueue::dispatchTasks)
	, m_isClosed(false)
    {
    }

    void enqueueTask(PassOwnPtr<GenericTask<T> > task)
    {
        if (m_isClosed)
            return;

    	m_pendingTasks.append(task);

    	if(!m_dispatchTaskTimer.isActive())
    		m_dispatchTaskTimer.startOneShot(0);
    }

    void close()
    {
        m_isClosed = true;

        m_pendingTasks.clear();
        m_dispatchTaskTimer.stop();
    }

    bool hasPendingTasks() const { return m_pendingTasks.size() > 0; }

private:
    void dispatchTasks(Timer<GenericTaskQueue>*)
    {
    	if(m_pendingTasks.isEmpty())
    		return;

    	OwnPtr<GenericTask<T> > task = m_pendingTasks.takeFirst();
    	task->exec();

    	if(m_pendingTasks.isEmpty() || m_dispatchTaskTimer.isActive())
    		return;

    	m_dispatchTaskTimer.startOneShot(0);
    }

    Deque<OwnPtr<GenericTask<T> > > m_pendingTasks;
    Timer<GenericTaskQueue> m_dispatchTaskTimer;
    bool m_isClosed;
};

}

#endif
