/*
 * Copyright (c) 2012, Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef LayoutUnit_h
#define LayoutUnit_h

#include <limits.h>
#include <limits>
#include <math.h>
#include <stdlib.h>
#include <wtf/MathExtras.h>

#if PLATFORM(QT)
#include <QDataStream>
#endif

namespace WebCore {

#ifdef NDEBUG

#define REPORT_OVERFLOW(doesOverflow) ((void)0)

#else

#define REPORT_OVERFLOW(doesOverflow) do \
    if (!(doesOverflow)) { \
        WTFReportError(__FILE__, __LINE__, WTF_PRETTY_FUNCTION, "!(%s)", #doesOverflow); \
    } \
while (0)

#endif

static const int kFixedPointDenominator = 64;
const int intMaxForLayoutUnit = INT_MAX / kFixedPointDenominator;
const int intMinForLayoutUnit = INT_MIN / kFixedPointDenominator;

class LayoutUnit {
public:
    LayoutUnit() : m_value(0) { }
    LayoutUnit(int value) { setValue(value); }
    LayoutUnit(unsigned short value) { setValue(value); }
    LayoutUnit(unsigned value) { setValue(value); }
    LayoutUnit(unsigned long value)
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        m_value = clampTo<int>(value * kFixedPointDenominator);
#else
        REPORT_OVERFLOW(isInBounds(static_cast<unsigned>(value)));
        m_value = value * kFixedPointDenominator;
#endif
    }
    LayoutUnit(unsigned long long value)
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        m_value = clampTo<int>(value * kFixedPointDenominator);
#else
        REPORT_OVERFLOW(isInBounds(static_cast<unsigned>(value)));
        m_value = static_cast<int>(value * kFixedPointDenominator);
#endif
    }
    LayoutUnit(float value)
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        m_value = clampTo<float>(value * kFixedPointDenominator, static_cast<float>(INT_MIN), static_cast<float>(INT_MAX));
#else
        REPORT_OVERFLOW(isInBounds(value));
        m_value = value * kFixedPointDenominator;
#endif
    }
    LayoutUnit(double value)
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        m_value = clampTo<double>(value * kFixedPointDenominator, static_cast<double>(INT_MIN), static_cast<double>(INT_MAX));
#else
        REPORT_OVERFLOW(isInBounds(value));
        m_value = value * kFixedPointDenominator;
#endif
    }

    static LayoutUnit fromPixel(int value)
    {
        return LayoutUnit(value);
    }

    static LayoutUnit fromFloatCeil(float value)
    {
        LayoutUnit v;
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        v.m_value = clampToInteger(ceilf(value * kFixedPointDenominator));
#else
        REPORT_OVERFLOW(isInBounds(value));
        v.m_value = ceilf(value * kFixedPointDenominator);
#endif
        return v;
    }

    static LayoutUnit fromFloatFloor(float value)
    {
        LayoutUnit v;
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        v.m_value = clampToInteger(floorf(value * kFixedPointDenominator));
#else
        REPORT_OVERFLOW(isInBounds(value));
        v.m_value = floorf(value * kFixedPointDenominator);
#endif
        return v;
    }

    static LayoutUnit fromFloatRound(float value)
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        if (value >= 0)
            return clamp(value + epsilon() / 2.0f);
        return clamp(value - epsilon() / 2.0f);
#else
        if (value >= 0) {
            REPORT_OVERFLOW(isInBounds(value + epsilon() / 2.0f));
            return LayoutUnit(value + epsilon() / 2.0f);
        }
        REPORT_OVERFLOW(isInBounds(value - epsilon() / 2.0f));
        return LayoutUnit(value - epsilon() / 2.0f);
#endif
    }

    ALWAYS_INLINE int toInt() const { return m_value / kFixedPointDenominator; }
    ALWAYS_INLINE float toFloat() const { return static_cast<float>(m_value) / kFixedPointDenominator; }
    ALWAYS_INLINE double toDouble() const { return static_cast<double>(m_value) / kFixedPointDenominator; }
    ALWAYS_INLINE unsigned toUnsigned() const { REPORT_OVERFLOW(m_value >= 0); return toInt(); }

    ALWAYS_INLINE operator int() const { return toInt(); }
    ALWAYS_INLINE operator float() const { return toFloat(); }
    ALWAYS_INLINE operator double() const { return toDouble(); }
    ALWAYS_INLINE operator bool() const { return m_value; }

    LayoutUnit& operator++()
    {
        m_value += kFixedPointDenominator;
        return *this;
    }

    ALWAYS_INLINE int rawValue() const { return m_value; }
    ALWAYS_INLINE void setRawValue(int value) { m_value = value; }
    ALWAYS_INLINE void setRawValue(long long value)
    {
        REPORT_OVERFLOW(value > std::numeric_limits<int>::min() && value < std::numeric_limits<int>::max());
        m_value = static_cast<int>(value);
    }

    LayoutUnit abs() const
    {
        LayoutUnit returnValue;
        returnValue.setRawValue(::abs(m_value));
        return returnValue;
    }
#if OS(DARWIN)
    int wtf_ceil() const
#else
    int ceil() const
#endif
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        if (UNLIKELY(m_value >= INT_MAX - kFixedPointDenominator + 1))
            return intMaxForLayoutUnit;
#endif
        if (m_value >= 0)
            return (m_value + kFixedPointDenominator - 1) / kFixedPointDenominator;
        return toInt();
    }

    int round() const
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        if (m_value > 0)
            return saturatedAddition(rawValue(), kFixedPointDenominator / 2) / kFixedPointDenominator;
        return saturatedSubtraction(rawValue(), (kFixedPointDenominator / 2) - 1) / kFixedPointDenominator;
#else
        if (m_value > 0)
            return (m_value + (kFixedPointDenominator / 2)) / kFixedPointDenominator;
        return (m_value - ((kFixedPointDenominator / 2) - 1)) / kFixedPointDenominator;
#endif
    }

    int floor() const
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        if (UNLIKELY(m_value <= INT_MIN + kFixedPointDenominator - 1))
            return intMinForLayoutUnit;
#endif
        if (m_value >= 0)
            return toInt();
        return (m_value - kFixedPointDenominator + 1) / kFixedPointDenominator;
    }

    float ceilToFloat() const
    {
        float floatValue = toFloat();
        if (static_cast<int>(floatValue * kFixedPointDenominator) == m_value)
            return floatValue;
        if (floatValue > 0)
            return nextafterf(floatValue, std::numeric_limits<float>::max());
        return nextafterf(floatValue, std::numeric_limits<float>::min());
    }

    LayoutUnit fraction() const
    {
        // Add the fraction to the size (as opposed to the full location) to avoid overflows.
        // Compute fraction using the mod operator to preserve the sign of the value as it may affect rounding.
    	LayoutUnit fraction;
        fraction.setRawValue(rawValue() % kFixedPointDenominator);
        return fraction;
    }

    static float epsilon() { return 1.0f / kFixedPointDenominator; }

    static const LayoutUnit max()
    {
        LayoutUnit m;
        m.m_value = std::numeric_limits<int>::max();
        return m;
    }
    static const LayoutUnit min()
    {
        LayoutUnit m;
        m.m_value = std::numeric_limits<int>::min();
        return m;
    }

    // Versions of max/min that are slightly smaller/larger than max/min() to allow for roinding without overflowing.
    static const LayoutUnit nearlyMax()
    {
    	LayoutUnit m;
        m.m_value = std::numeric_limits<int>::max() - kFixedPointDenominator / 2;
        return m;
    }
    static const LayoutUnit nearlyMin()
    {
    	LayoutUnit m;
        m.m_value = std::numeric_limits<int>::min() + kFixedPointDenominator / 2;
        return m;
    }

    static LayoutUnit clamp(double value)
    {
        return clampTo<LayoutUnit>(value, LayoutUnit::min(), LayoutUnit::max());
    }
    
private:

    static bool isInBounds(int value)
    {
        return ::abs(value) <= std::numeric_limits<int>::max() / kFixedPointDenominator;
    }
    static bool isInBounds(unsigned value)
    {
        return value <= static_cast<unsigned>(std::numeric_limits<int>::max()) / kFixedPointDenominator;
    }
    static bool isInBounds(double value)
    {
        return ::fabs(value) <= std::numeric_limits<int>::max() / kFixedPointDenominator;
    }

    inline void setValue(int value)
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        if (value > intMaxForLayoutUnit)
            m_value = std::numeric_limits<int>::max();
        else if (value < intMinForLayoutUnit)
            m_value = std::numeric_limits<int>::min();
        else
            m_value = value * kFixedPointDenominator;
#else
        REPORT_OVERFLOW(isInBounds(value));
        m_value = value * kFixedPointDenominator;
#endif
    }
    inline void setValue(unsigned value)
    {
#if ENABLE(SATURATED_LAYOUT_ARITHMETIC)
        if (value >= static_cast<unsigned>(intMaxForLayoutUnit))
            m_value = std::numeric_limits<int>::max();
        else
            m_value = value * kFixedPointDenominator;
#else
        REPORT_OVERFLOW(isInBounds(value));
        m_value = value * kFixedPointDenominator;
#endif
    }

    int m_value;
};

ALWAYS_INLINE bool operator<=(const LayoutUnit& a, const LayoutUnit& b)
{
    return a.rawValue() <= b.rawValue();
}

ALWAYS_INLINE bool operator<=(const LayoutUnit& a, float b)
{
    return a.toFloat() <= b;
}

ALWAYS_INLINE bool operator<=(const LayoutUnit& a, int b)
{
    return a <= LayoutUnit(b);
}

ALWAYS_INLINE bool operator<=(const float a, const LayoutUnit& b)
{
    return a <= b.toFloat();
}

ALWAYS_INLINE bool operator<=(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) <= b;
}

ALWAYS_INLINE bool operator>=(const LayoutUnit& a, const LayoutUnit& b)
{
    return a.rawValue() >= b.rawValue();
}

ALWAYS_INLINE bool operator>=(const LayoutUnit& a, int b)
{
    return a >= LayoutUnit(b);
}

ALWAYS_INLINE bool operator>=(const float a, const LayoutUnit& b)
{
    return a >= b.toFloat();
}

ALWAYS_INLINE bool operator>=(const LayoutUnit& a, float b)
{
    return a.toFloat() >= b;
}

ALWAYS_INLINE bool operator>=(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) >= b;
}

ALWAYS_INLINE bool operator<(const LayoutUnit& a, const LayoutUnit& b)
{
    return a.rawValue() < b.rawValue();
}

ALWAYS_INLINE bool operator<(const LayoutUnit& a, int b)
{
    return a < LayoutUnit(b);
}

ALWAYS_INLINE bool operator<(const LayoutUnit& a, float b)
{
    return a.toFloat() < b;
}

ALWAYS_INLINE bool operator<(const LayoutUnit& a, double b)
{
    return a.toDouble() < b;
}

ALWAYS_INLINE bool operator<(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) < b;
}

ALWAYS_INLINE bool operator<(const float a, const LayoutUnit& b)
{
    return a < b.toFloat();
}

ALWAYS_INLINE bool operator>(const LayoutUnit& a, const LayoutUnit& b)
{
    return a.rawValue() > b.rawValue();
}

ALWAYS_INLINE bool operator>(const LayoutUnit& a, double b)
{
    return a.toDouble() > b;
}

ALWAYS_INLINE bool operator>(const LayoutUnit& a, float b)
{
    return a.toFloat() > b;
}

ALWAYS_INLINE bool operator>(const LayoutUnit& a, int b)
{
    return a > LayoutUnit(b);
}

ALWAYS_INLINE bool operator>(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) > b;
}

ALWAYS_INLINE bool operator>(const float a, const LayoutUnit& b)
{
    return a > b.toFloat();
}

ALWAYS_INLINE bool operator>(const double a, const LayoutUnit& b)
{
    return a > b.toDouble();
}

ALWAYS_INLINE bool operator!=(const LayoutUnit& a, const LayoutUnit& b)
{
    return a.rawValue() != b.rawValue();
}

ALWAYS_INLINE bool operator!=(const LayoutUnit& a, float b)
{
    return a != LayoutUnit(b);
}

ALWAYS_INLINE bool operator!=(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) != b;
}

ALWAYS_INLINE bool operator!=(const LayoutUnit& a, int b)
{
    return a != LayoutUnit(b);
}

ALWAYS_INLINE bool operator==(const LayoutUnit& a, const LayoutUnit& b)
{
    return a.rawValue() == b.rawValue();
}

ALWAYS_INLINE bool operator==(const LayoutUnit& a, int b)
{
    return a == LayoutUnit(b);
}

ALWAYS_INLINE bool operator==(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) == b;
}

ALWAYS_INLINE bool operator==(const LayoutUnit& a, float b)
{
    return a.toFloat() == b;
}

ALWAYS_INLINE bool operator==(const float a, const LayoutUnit& b)
{
    return a == b.toFloat();
}

// For multiplication that's prone to overflow, this bounds it to LayoutUnit::max() and ::min()
ALWAYS_INLINE LayoutUnit boundedMultiply(const LayoutUnit& a, const LayoutUnit& b)
{
    LayoutUnit returnVal;
    long long rawVal = static_cast<long long>(a.rawValue()) * b.rawValue() / kFixedPointDenominator;
    if (rawVal > std::numeric_limits<int>::max())
        return LayoutUnit::max();
    if (rawVal < std::numeric_limits<int>::min())
        return LayoutUnit::min();
    returnVal.setRawValue(rawVal);
    return returnVal;
}

ALWAYS_INLINE LayoutUnit operator*(const LayoutUnit& a, const LayoutUnit& b)
{
    LayoutUnit returnVal;
    long long rawVal = static_cast<long long>(a.rawValue()) * b.rawValue() / kFixedPointDenominator;
    returnVal.setRawValue(rawVal);
    return returnVal;
}    

ALWAYS_INLINE double operator*(const LayoutUnit& a, double b)
{
    return a.toDouble() * b;
}

ALWAYS_INLINE float operator*(const LayoutUnit& a, float b)
{
    return a.toFloat() * b;
}

ALWAYS_INLINE LayoutUnit operator*(const LayoutUnit& a, int b)
{
    return a * LayoutUnit(b);
}

inline LayoutUnit operator*(const LayoutUnit& a, unsigned short b)
{
    return a * LayoutUnit(b);
}

ALWAYS_INLINE LayoutUnit operator*(const LayoutUnit& a, unsigned b)
{
    return a * LayoutUnit(b);
}

inline LayoutUnit operator*(const LayoutUnit& a, unsigned long b)
{
    return a * LayoutUnit(b);
}

inline LayoutUnit operator*(const LayoutUnit& a, unsigned long long b)
{
    return a * LayoutUnit(b);
}

inline LayoutUnit operator*(unsigned short a, const LayoutUnit& b)
{
    return LayoutUnit(a) * b;
}

ALWAYS_INLINE LayoutUnit operator*(unsigned a, const LayoutUnit& b)
{
    return LayoutUnit(a) * b;
}

inline LayoutUnit operator*(unsigned long a, const LayoutUnit& b)
{
    return LayoutUnit(a) * b;
}

inline LayoutUnit operator*(unsigned long long a, const LayoutUnit& b)
{
    return LayoutUnit(a) * b;
}

ALWAYS_INLINE LayoutUnit operator*(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) * b;
}

ALWAYS_INLINE float operator*(const float a, const LayoutUnit& b)
{
    return a * b.toFloat();
}

ALWAYS_INLINE double operator*(const double a, const LayoutUnit& b)
{
    return a * b.toDouble();
}

ALWAYS_INLINE LayoutUnit operator/(const LayoutUnit& a, const LayoutUnit& b)
{
    LayoutUnit returnVal;
    long long rawVal = static_cast<long long>(kFixedPointDenominator) * a.rawValue() / b.rawValue();
    returnVal.setRawValue(rawVal);
    return returnVal;
}    

ALWAYS_INLINE float operator/(const LayoutUnit& a, float b)
{
    return a.toFloat() / b;
}

ALWAYS_INLINE double operator/(const LayoutUnit& a, double b)
{
    return a.toDouble() / b;
}

ALWAYS_INLINE LayoutUnit operator/(const LayoutUnit& a, int b)
{
    return a / LayoutUnit(b);
}

inline LayoutUnit operator/(const LayoutUnit& a, unsigned short b)
{
    return a / LayoutUnit(b);
}

ALWAYS_INLINE LayoutUnit operator/(const LayoutUnit& a, unsigned int b)
{
    return a / LayoutUnit(b);
}

inline LayoutUnit operator/(const LayoutUnit& a, unsigned long b)
{
    return a / LayoutUnit(b);
}

inline LayoutUnit operator/(const LayoutUnit& a, unsigned long long b)
{
    return a / LayoutUnit(b);
}

ALWAYS_INLINE float operator/(const float a, const LayoutUnit& b)
{
    return a / b.toFloat();
}

ALWAYS_INLINE double operator/(const double a, const LayoutUnit& b)
{
    return a / b.toDouble();
}

ALWAYS_INLINE LayoutUnit operator/(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) / b;
}

inline LayoutUnit operator/(unsigned short a, const LayoutUnit& b)
{
    return LayoutUnit(a) / b;
}

ALWAYS_INLINE LayoutUnit operator/(unsigned int a, const LayoutUnit& b)
{
    return LayoutUnit(a) / b;
}

inline LayoutUnit operator/(unsigned long a, const LayoutUnit& b)
{
    return LayoutUnit(a) / b;
}

inline LayoutUnit operator/(unsigned long long a, const LayoutUnit& b)
{
    return LayoutUnit(a) / b;
}

ALWAYS_INLINE LayoutUnit operator+(const LayoutUnit& a, const LayoutUnit& b)
{
    LayoutUnit returnVal;
    returnVal.setRawValue(a.rawValue() + b.rawValue());
    return returnVal;
}

ALWAYS_INLINE LayoutUnit operator+(const LayoutUnit& a, int b)
{
    return a + LayoutUnit(b);
}

ALWAYS_INLINE float operator+(const LayoutUnit& a, float b)
{
    return a.toFloat() + b;
}

ALWAYS_INLINE double operator+(const LayoutUnit& a, double b)
{
    return a.toDouble() + b;
}

ALWAYS_INLINE LayoutUnit operator+(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) + b;
}

ALWAYS_INLINE float operator+(const float a, const LayoutUnit& b)
{
    return a + b.toFloat();
}

ALWAYS_INLINE double operator+(const double a, const LayoutUnit& b)
{
    return a + b.toDouble();
}

ALWAYS_INLINE LayoutUnit operator-(const LayoutUnit& a, const LayoutUnit& b)
{
    LayoutUnit returnVal;
    returnVal.setRawValue(a.rawValue() - b.rawValue());
    return returnVal;
}

ALWAYS_INLINE LayoutUnit operator-(const LayoutUnit& a, int b)
{
    return a - LayoutUnit(b);
}

ALWAYS_INLINE LayoutUnit operator-(const LayoutUnit& a, unsigned b)
{
    return a - LayoutUnit(b);
}

ALWAYS_INLINE float operator-(const LayoutUnit& a, float b)
{
    return a.toFloat() - b;
}

ALWAYS_INLINE LayoutUnit operator-(const int a, const LayoutUnit& b)
{
    return LayoutUnit(a) - b;
}

ALWAYS_INLINE float operator-(const float a, const LayoutUnit& b)
{
    return a - b.toFloat();
}

ALWAYS_INLINE LayoutUnit operator-(const LayoutUnit& a)
{
    LayoutUnit returnVal;
    returnVal.setRawValue(-a.rawValue());
    return returnVal;
}

// For returning the remainder after a division with integer results.
ALWAYS_INLINE LayoutUnit intMod(const LayoutUnit& a, const LayoutUnit& b)
{
    // This calculates the modulo so that: a = static_cast<int>(a / b) * b + intMod(a, b).
    LayoutUnit returnVal;
    returnVal.setRawValue(a.rawValue() % b.rawValue());
    return returnVal;
}

ALWAYS_INLINE LayoutUnit operator%(const LayoutUnit& a, const LayoutUnit& b)
{
    // This calculates the modulo so that: a = (a / b) * b + a % b.
    LayoutUnit returnVal;
    long long rawVal = (static_cast<long long>(kFixedPointDenominator) * a.rawValue()) % b.rawValue();
    returnVal.setRawValue(rawVal / kFixedPointDenominator);
    return returnVal;
}

ALWAYS_INLINE LayoutUnit operator%(const LayoutUnit& a, int b)
{
    return a % LayoutUnit(b);
}

ALWAYS_INLINE LayoutUnit operator%(int a, const LayoutUnit& b)
{
    return LayoutUnit(a) % b;
}

ALWAYS_INLINE LayoutUnit& operator+=(LayoutUnit& a, const LayoutUnit& b)
{
    a = a + b;
    return a;
}

ALWAYS_INLINE LayoutUnit& operator+=(LayoutUnit& a, int b)
{
    a = a + b;
    return a;
}

ALWAYS_INLINE LayoutUnit& operator+=(LayoutUnit& a, float b)
{
    a = a + b;
    return a;
}

ALWAYS_INLINE float& operator+=(float& a, const LayoutUnit& b)
{
    a = a + b;
    return a;
}

ALWAYS_INLINE LayoutUnit& operator-=(LayoutUnit& a, int b)
{
    a = a - b;
    return a;
}

ALWAYS_INLINE LayoutUnit& operator-=(LayoutUnit& a, const LayoutUnit& b)
{
    a = a - b;
    return a;
}

ALWAYS_INLINE LayoutUnit& operator-=(LayoutUnit& a, float b)
{
    a = a - b;
    return a;
}

ALWAYS_INLINE float& operator-=(float& a, const LayoutUnit& b)
{
    a = a - b;
    return a;
}

ALWAYS_INLINE LayoutUnit& operator*=(LayoutUnit& a, const LayoutUnit& b)
{
    a = a * b;
    return a;
}
// operator*=(LayoutUnit& a, int b) is supported by the operator above plus LayoutUnit(int).

ALWAYS_INLINE LayoutUnit& operator*=(LayoutUnit& a, float b)
{
    a = a * b;
    return a;
}

ALWAYS_INLINE float& operator*=(float& a, const LayoutUnit& b)
{
    a = a * b;
    return a;
}

ALWAYS_INLINE LayoutUnit& operator/=(LayoutUnit& a, const LayoutUnit& b)
{
    a = a / b;
    return a;
}
// operator/=(LayoutUnit& a, int b) is supported by the operator above plus LayoutUnit(int).

ALWAYS_INLINE LayoutUnit& operator/=(LayoutUnit& a, float b)
{
    a = a / b;
    return a;
}

ALWAYS_INLINE float& operator/=(float& a, const LayoutUnit& b)
{
    a = a / b;
    return a;
}

inline int roundToInt(LayoutUnit value)
{
    return value.round();
}

inline int floorToInt(LayoutUnit value)
{
    return value.floor();
}

inline float roundToDevicePixel(LayoutUnit value, const float pixelSnappingFactor, bool needsDirectionalRounding = false)
{
    double valueToRound = value.toDouble();
    if (needsDirectionalRounding)
        valueToRound -= LayoutUnit::epsilon() / (2 * kFixedPointDenominator);

    if (valueToRound >= 0)
        return round(valueToRound * pixelSnappingFactor) / pixelSnappingFactor;

    // This adjusts directional rounding on negative halfway values. It produces the same direction for both negative and positive values.
    // Instead of rounding negative halfway cases away from zero, we translate them to positive values before rounding.
    // It helps snapping relative negative coordinates to the same position as if they were positive absolute coordinates.
    unsigned translateOrigin = -value.rawValue();
    return (round((valueToRound + translateOrigin) * pixelSnappingFactor) / pixelSnappingFactor) - translateOrigin;
}

inline float floorToDevicePixel(LayoutUnit value, float pixelSnappingFactor)
{
    return floor((value.rawValue() * pixelSnappingFactor) / kFixedPointDenominator) / pixelSnappingFactor;
}

inline float ceilToDevicePixel(LayoutUnit value, float pixelSnappingFactor)
{
    return ceilf((value.rawValue() * pixelSnappingFactor) / kFixedPointDenominator) / pixelSnappingFactor;
}

inline LayoutUnit roundedLayoutUnit(float value)
{
    return LayoutUnit::fromFloatRound(value);
}

inline LayoutUnit ceiledLayoutUnit(float value)
{
    return LayoutUnit::fromFloatCeil(value);
}

inline LayoutUnit absoluteValue(const LayoutUnit& value)
{
    return value.abs();
}

inline bool isIntegerValue(const LayoutUnit value)
{
    return value.toInt() == value;
}

} // namespace WebCore

#endif // LayoutUnit_h
