/*
 *  Copyright (C) 2011 Igalia S.L
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"

#if ENABLE(WEB_AUDIO)

#include "AudioDestinationSDL.h"

#include "AudioChannel.h"
#include "AudioSourceProvider.h"

#include "AudioOutput.h"
#include <wtf/CurrentTime.h>

namespace WebCore {

typedef int (*LogFunction)(const char * format, ... );

LogFunction gWebAudioLog = 0;

DLL_PUBLIC void setWebAudioLogFunction(LogFunction func)
{
	gWebAudioLog = func;
}

class AudioInputPrivate : public AudioInput {
public:
    AudioInputPrivate (AudioSourceProvider& provider)
	: m_provider (provider)
	, m_logCount(0)
	, m_lastLogTimeStamp(0)
	{
	}

    virtual int readInterleaved (float* buf, int nb_frames, bool& isSilent) {
        //gWebAudioLog ("AudioInputPrivate::readInterleaved (%p, %d)\n", buf, nb_frames);
        RefPtr<AudioBus> renderBus = AudioBus::create(2, nb_frames/2);
        m_provider.provideInput(renderBus.get (), nb_frames/2);
        isSilent = renderBus->isSilent();
        if(isSilent)
        	return 0;

        float maxAbsValue = renderBus->maxAbsValue();

        if(maxAbsValue > 1.0) {
        	if(gWebAudioLog) {
        		double now = WTF::monotonicallyIncreasingTime();
        		if(m_logCount == 0) {
        			gWebAudioLog("(%s, %s, %i) numberOfRenderingConnections %i, maxAbsValue %f, scale output.\n", __FILE__, __FUNCTION__, __LINE__, renderBus->numberOfRenderingConnections(), maxAbsValue);
        			m_lastLogTimeStamp = now;
        		}
        		else {
        			if((now - m_lastLogTimeStamp) >= 5)
        				m_logCount = 0;
        		}
                ++m_logCount;
        	}
        	renderBus->scale(1.0f / maxAbsValue);
        }

        for (int i = 0 ; i < renderBus->length (); i++)
            for (int j = 0; j < 2; j++)
                buf[i * 2 + j] = (renderBus->channel (j)->data ())[i];
        return nb_frames;
    }

private:
    AudioSourceProvider& m_provider;
    bool m_logCount;
    double m_lastLogTimeStamp;
};

PassOwnPtr<AudioDestination> AudioDestination::create(AudioSourceProvider& provider, float sampleRate)
{
    return adoptPtr(new AudioDestinationSDL(provider, sampleRate));
}

AudioDestinationSDL::AudioDestinationSDL(AudioSourceProvider& provider, float sampleRate)
: m_provider(provider)
, m_sampleRate(sampleRate)
, m_isPlaying(false)
, m_output(adoptPtr(AudioOutput::create ()))
, m_input(adoptPtr(new AudioInputPrivate (m_provider)))
{
    AudioFormat fmt;
    fmt.m_channels = 2;
    fmt.m_sample_rate = sampleRate;
    m_output->Open(fmt);
}

AudioDestinationSDL::~AudioDestinationSDL()
{
    //gWebAudioLog ("AudioDestinationSDL::~AudioDestinationSDL()\n");
    m_output->Close();
}

void AudioDestinationSDL::start()
{
    //gWebAudioLog ("AudioDestinationSDL::start()\n");
    if (m_isPlaying)
    	return;
    m_output->Start(m_input.get());
}

void AudioDestinationSDL::stop()
{
    //gWebAudioLog ("AudioDestinationSDL::stop()\n");
    if (!m_isPlaying)
    	return;
    m_output->Stop();
    m_isPlaying = false;
}

} // namespace WebCore

#endif // ENABLE(WEB_AUDIO)
