/*
 *  Copyright (C) 2011, 2012 Igalia S.L
 *  Copyright (C) 2011 Zan Dobersek  <zandobersek@gmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"

#if ENABLE(WEB_AUDIO)

#include "AudioFileReader.h"
#include "AudioBus.h"
#include <wtf/CurrentTime.h>
#include <wtf/Vector.h>

#if USE(WEBAUDIO_FFMPEG)
#define __STDC_CONSTANT_MACROS
#include <stdint.h>

#ifndef INT64_C
#define INT64_C(c) (c ## LL)
#define UINT64_C(c) (c ## ULL)
#endif

extern "C" {
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libavformat/avio.h>
}

#ifndef DUMP_AUDIO_TO_FILE
#define DUMP_AUDIO_TO_FILE 0
#endif

#if DUMP_AUDIO_TO_FILE
#include <stdio.h>
#endif

namespace WebCore {

typedef int (*LogFunction)(const char * format, ... );
extern LogFunction gWebAudioLog;

static const char* sample_format_string (int fmt) {
    switch (fmt) {
	case AV_SAMPLE_FMT_U8:
        return "AV_SAMPLE_FMT_U8";
	case AV_SAMPLE_FMT_U8P:
        return "AV_SAMPLE_FMT_U8P";
	case AV_SAMPLE_FMT_S16:
        return "AV_SAMPLE_FMT_S16";
	case AV_SAMPLE_FMT_S16P:
        return "AV_SAMPLE_FMT_S16P";
	case AV_SAMPLE_FMT_S32:
        return "AV_SAMPLE_FMT_S32";
	case AV_SAMPLE_FMT_S32P:
        return "AV_SAMPLE_FMT_S32P";
	case AV_SAMPLE_FMT_FLT:
        return "AV_SAMPLE_FMT_FLT";
	case AV_SAMPLE_FMT_FLTP:
        return "AV_SAMPLE_FMT_FLTP";
	case AV_SAMPLE_FMT_DBL:
        return "AV_SAMPLE_FMT_DBL";
	case AV_SAMPLE_FMT_DBLP:
        return "AV_SAMPLE_FMT_DBLP";
	default:
		return "UNKOWN";
    }
}

static ALWAYS_INLINE float U82FLT (uint8_t src) {
    return (double)src/128.0;
}

static ALWAYS_INLINE float S162FLT (int16_t src) {
    return (double)src/32768.0;
}

static ALWAYS_INLINE float S322FLT (int32_t src) {
    return (double)src/2147483648.0;
}

static ALWAYS_INLINE float DBL2FLT (double src) {
    return (float)src;
}

static void samplecpy_from_U8 (float* dstData, uint8_t* srcData, size_t nb_samples) {
    for (uint32_t i = 0; i < nb_samples; i++)
        dstData[i] = U82FLT (srcData[i]);
}

static void samplecpy_from_S16 (float* dstData, int16_t* srcData, size_t nb_samples) {
    for (uint32_t i = 0; i < nb_samples; i++)
        dstData[i] = S162FLT (srcData[i]);
}

static void samplecpy_from_S32 (float* dstData, int32_t* srcData, size_t nb_samples) {
    for (uint32_t i = 0; i < nb_samples; i++)
        dstData[i] = S322FLT (srcData[i]);
}

static void samplecpy_from_FLT (float* dstData, float* srcData, size_t nb_samples) {
    memcpy (dstData, srcData, sizeof (float)*nb_samples);
}

static void samplecpy_from_DBL (float* dstData, double* srcData, size_t nb_samples) {
    for (uint32_t i = 0; i < nb_samples; i++)
        dstData[i] = DBL2FLT (srcData[i]);
}

class AVFrameList;
static int read_packet (void* userdata, uint8_t* buf, int buf_size);

class AudioStreamReaderFFMPEG {
public:
    AudioStreamReaderFFMPEG (const char* filePath);
    AudioStreamReaderFFMPEG (const void* data, size_t dataSize);
    ~AudioStreamReaderFFMPEG ();
    PassRefPtr<AudioBus> createBus(float sampleRate, bool mixToMono);

private:
    void initMemStreamContext ();

    AVFormatContext* m_avfmtctx;
    AVIOContext* m_probe_ctx;
    bool m_error;

    int readBuffer (uint8_t* buf, int size);

    int m_buf_index;
    uint8_t* m_buf;
    int m_buf_size;

    friend int read_packet (void* userdata, uint8_t* buf, int buf_size);
};

static int read_packet (void* userdata, uint8_t* buf, int buf_size) {
    AudioStreamReaderFFMPEG* reader = static_cast<AudioStreamReaderFFMPEG*>(userdata);
    return reader->readBuffer(buf, buf_size);
}

int AudioStreamReaderFFMPEG::readBuffer (uint8_t* buf, int size)
{
    int sz = m_buf_index + size <= m_buf_size ? size : m_buf_size - m_buf_index;
    if (!sz)
    	return 0;
    memcpy (buf, (const char*)&(m_buf[m_buf_index]), sz);
    m_buf_index += sz;
    return sz;
}

AudioStreamReaderFFMPEG::~AudioStreamReaderFFMPEG ()
{
    if (m_avfmtctx){
    	avformat_free_context (m_avfmtctx);
    	m_avfmtctx = NULL;
    }

    if (m_probe_ctx){
    	//  It may be freed and replaced with a new buffer by libavformat.
    	// AVIOContext.buffer holds the buffer currently in use, which must be later freed with av_free().
    	if(m_probe_ctx->buffer)
    		av_free(m_probe_ctx->buffer);

    	av_free (m_probe_ctx);
    	m_probe_ctx = NULL;
    }
}

static const int bufferSize = 256;

void AudioStreamReaderFFMPEG::initMemStreamContext() {
	av_register_all();

    m_avfmtctx = avformat_alloc_context ();

	//  It may be freed and replaced with a new buffer by libavformat.
	// AVIOContext.buffer holds the buffer currently in use, which must be later freed with av_free().
    unsigned char * av_buffer = (unsigned char*)av_malloc(bufferSize);
    m_probe_ctx = avio_alloc_context (av_buffer, bufferSize, 0, this, read_packet, NULL, NULL);

    if(!m_probe_ctx)
    	av_free(av_buffer);

    m_avfmtctx->pb = m_probe_ctx;
}

AudioStreamReaderFFMPEG::AudioStreamReaderFFMPEG (const char* filePath)
: m_avfmtctx (NULL)
, m_probe_ctx (NULL)
, m_error (false)
, m_buf_index (0)
, m_buf (NULL)
, m_buf_size (0)
{
	if(gWebAudioLog)
        gWebAudioLog("(%s, %s, %i) %s\n", __FILE__, __FUNCTION__, __LINE__, filePath);
}

AudioStreamReaderFFMPEG::AudioStreamReaderFFMPEG (const void* data, size_t dataSize)
: m_avfmtctx (NULL)
, m_probe_ctx (NULL)
, m_error (false)
, m_buf_index (0)
, m_buf((uint8_t*)data)
, m_buf_size(dataSize)
{
	if(gWebAudioLog)
        gWebAudioLog("(%s, %s, %i) with data length %i\n", __FILE__, __FUNCTION__, __LINE__, dataSize);

    initMemStreamContext ();

    AVProbeData probeData;
    memset(&probeData, 0x00, sizeof(probeData));

    probeData.buf = (unsigned char*)data;
    probeData.buf_size = dataSize;

    m_avfmtctx->iformat = av_probe_input_format (&probeData, 1);
    m_avfmtctx->flags = AVFMT_FLAG_CUSTOM_IO;

    int av_error;
    if (av_error = avformat_open_input(&m_avfmtctx, "", NULL, NULL)) {
        char av_err_buf [4096];
        memset (av_err_buf, 0, 4096);
        av_strerror (av_error, av_err_buf, 4096);
        m_error = true;
    	if(gWebAudioLog)
            gWebAudioLog("(%s, %s, %i) %s\n", __FILE__, __FUNCTION__, __LINE__, av_err_buf);
        return;
    }

	if(gWebAudioLog)
        gWebAudioLog("(%s, %s, %i) AVInputFormat %s (%s)\n", __FILE__, __FUNCTION__, __LINE__, m_avfmtctx->iformat->name, m_avfmtctx->iformat->long_name);

    if (avformat_find_stream_info (m_avfmtctx, NULL) < 0) {
        m_error = true;
    	if(gWebAudioLog)
            gWebAudioLog("(%s, %s, %i)\n", __FILE__, __FUNCTION__, __LINE__);
    }
}

static bool isPlanarSample(AVSampleFormat sample_fmt)
{
	switch (sample_fmt) {
	case AV_SAMPLE_FMT_U8P:
	case AV_SAMPLE_FMT_S16P:
	case AV_SAMPLE_FMT_S32P:
	case AV_SAMPLE_FMT_FLTP:
	case AV_SAMPLE_FMT_DBLP:
		return true;
	default:
		return false;
	}
}

#if DUMP_AUDIO_TO_FILE
static FILE* audio_file0 = NULL;
static FILE* audio_file1 = NULL;
#endif

static void fill(AVCodecContext* codecctx, Vector<Vector<float> >& audioFloatArrays, AVFrame* decoded_frame)
{
	if(isPlanarSample(codecctx->sample_fmt)){
		for (int i = 0; i < codecctx->channels; i++) {
			int oldSize = audioFloatArrays[i].size();
			audioFloatArrays[i].grow(oldSize + decoded_frame->nb_samples);
			float* data = audioFloatArrays[i].data() + oldSize;

			switch (codecctx->sample_fmt) {
			case AV_SAMPLE_FMT_U8P:
				samplecpy_from_U8(data, (uint8_t*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_S16P:
#if DUMP_AUDIO_TO_FILE
				if(i == 0){
					if(audio_file0){
						fwrite(decoded_frame->data[i], 1, decoded_frame->nb_samples * 2, audio_file0);
						fflush(audio_file0);
					}
				}
				else if(i == 1){
					if(audio_file1){
						fwrite(decoded_frame->data[i], 1, decoded_frame->nb_samples * 2, audio_file1);
						fflush(audio_file1);
					}
				}
#endif
				samplecpy_from_S16(data, (int16_t*)(decoded_frame->extended_data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_S32P:
				samplecpy_from_S32(data, (int32_t*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_FLTP:
				samplecpy_from_FLT(data, (float*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_DBLP:
				samplecpy_from_DBL(data, (double*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			default:
				break;
			}
		}
	}
	else{
		for (int i = 0; i < codecctx->channels; i++) {
			int oldSize = audioFloatArrays[i].size();
			audioFloatArrays[i].grow(oldSize + decoded_frame->nb_samples);
			float* data = audioFloatArrays[i].data() + oldSize;

			switch (codecctx->sample_fmt) {
			case AV_SAMPLE_FMT_U8:
				samplecpy_from_U8(data, (uint8_t*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_S16:
				samplecpy_from_S16(data, (int16_t*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_S32:
				samplecpy_from_S32(data, (int32_t*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_FLT:
				samplecpy_from_FLT(data, (float*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			case AV_SAMPLE_FMT_DBL:
				samplecpy_from_DBL(data, (double*)(decoded_frame->data[i]), decoded_frame->nb_samples);
				break;
			default:
				break;
			}
		}
	}
}

static PassRefPtr<AudioBus> createEmptyAudioBus(float sampleRate)
{
	RefPtr<AudioBus> audioBus = AudioBus::create(0, 0, true);
	audioBus->setSampleRate(sampleRate);
	return audioBus.release();
}

PassRefPtr<AudioBus> AudioStreamReaderFFMPEG::createBus(float sampleRate, bool mixToMono)
{
	if(gWebAudioLog)
        gWebAudioLog("(%s, %s, %i) sampleRate %f mixToMono %i\n", __FILE__, __FUNCTION__, __LINE__, sampleRate, mixToMono);

    if (m_error)
    	return createEmptyAudioBus(sampleRate);

    int i = 0;
    for (; i < (int)m_avfmtctx->nb_streams; i++) {
        if (m_avfmtctx->streams[i]->codec->codec_type == AVMEDIA_TYPE_AUDIO)
            break;
    }

    if (i == (int)m_avfmtctx->nb_streams) {
        m_error = true;
    	if(gWebAudioLog)
            gWebAudioLog("(%s, %s, %i)\n", __FILE__, __FUNCTION__, __LINE__);
    	return createEmptyAudioBus(sampleRate);
    }

    AVCodecContext* codecctx = m_avfmtctx->streams[i]->codec;
    AVCodec* codec = avcodec_find_decoder(codecctx->codec_id);

    if (!codec) {
        m_error = true;
    	if(gWebAudioLog)
            gWebAudioLog("(%s, %s, %i)\n", __FILE__, __FUNCTION__, __LINE__);
    	return createEmptyAudioBus(sampleRate);
    }

	if(gWebAudioLog)
        gWebAudioLog("(%s, %s, %i) decoder %s (%s)\n", __FILE__, __FUNCTION__, __LINE__, codec->name, codec->long_name);

    if (avcodec_open2(codecctx, codec, NULL)) {
        m_error = true;
    	if(gWebAudioLog)
            gWebAudioLog("(%s, %s, %i)\n", __FILE__, __FUNCTION__, __LINE__);
    	return createEmptyAudioBus(sampleRate);
    }

	if(gWebAudioLog)
        gWebAudioLog("(%s, %s, %i) sample format %s channels %i sample rate %i\n", __FILE__, __FUNCTION__, __LINE__,
        		sample_format_string(codecctx->sample_fmt), codecctx->channels, codecctx->sample_rate);

    if (m_error || !codecctx->channels || codecctx->channels > 2){
    	if(gWebAudioLog)
            gWebAudioLog("(%s, %s, %i) error\n", __FILE__, __FUNCTION__, __LINE__);
    	return createEmptyAudioBus(sampleRate);
    }

    int actual_channels = codecctx->channels;

    AVPacket input_packet;
    av_init_packet(&input_packet);

    input_packet.data = (uint8_t*)m_buf;
    input_packet.size = m_buf_size;

	AVFrame* decoded_frame = av_frame_alloc();

	Vector<Vector<float> > audioFloatArrays;
	audioFloatArrays.resize(actual_channels);

	for(int i = 0; i < actual_channels; ++i)
		audioFloatArrays[i].reserveCapacity(24000); // 0.5s data

#if DUMP_AUDIO_TO_FILE
	if (!audio_file0)
		audio_file0 = fopen("/3rd/espial/tmp/webaudio0.pcm", "wb");
	if (!audio_file1)
		audio_file1 = fopen("/3rd/espial/tmp/webaudio1.pcm", "wb");
#endif

    while(input_packet.size > 0){
		int gotFrame = 0;
		int len = avcodec_decode_audio4(codecctx, decoded_frame, &gotFrame, &input_packet);

		if(len < 0) {
			if(gWebAudioLog)
				gWebAudioLog("(%s, %s, %i) avcodec_decode_audio4 error %i\n", __FILE__, __FUNCTION__, __LINE__, len);
			break;
		}

		if (!gotFrame)
			break;

		fill(codecctx, audioFloatArrays, decoded_frame);

		input_packet.size -= len;
		input_packet.data += len;
		input_packet.dts = input_packet.pts = AV_NOPTS_VALUE;
    }

	av_frame_free(&decoded_frame);
	av_free_packet(&input_packet);

    RefPtr<AudioBus> audioBus = AudioBus::create(actual_channels, 0, false);
    audioBus->setSampleRate(codecctx->sample_rate);

	for(int i = 0; i < actual_channels; ++i){
		OwnPtr<AudioFloatArray> buffer = adoptPtr(new AudioFloatArray(audioFloatArrays[i].size()));
		memcpy(buffer->data(), audioFloatArrays[i].data(), sizeof(float) * audioFloatArrays[i].size());
		audioBus->setChannelMemory(i, buffer.release());
		audioFloatArrays[i].resize(0);
	}

	if(gWebAudioLog)
		gWebAudioLog("(%s, %s, %i) sample-frames %i\n", __FILE__, __FUNCTION__, __LINE__, audioBus->length());

#if DUMP_AUDIO_TO_FILE
	if(audio_file0){
		fclose(audio_file0);
		audio_file0 = 0;
	}
	if(audio_file1){
		fclose(audio_file1);
		audio_file1 = 0;
	}
#endif

    if(codecctx->sample_rate == sampleRate)
        return audioBus;

	if(gWebAudioLog)
		gWebAudioLog("(%s, %s, %i) sample rate does not match.\n", __FILE__, __FUNCTION__, __LINE__);

	return AudioBus::createBySampleRateConverting(audioBus.get(), mixToMono, sampleRate);
}

PassRefPtr<AudioBus> createBusFromAudioFile(const char* filePath, bool mixToMono, float sampleRate)
{
    return AudioStreamReaderFFMPEG (filePath).createBus(sampleRate, mixToMono);
}

PassRefPtr<AudioBus> createBusFromInMemoryAudioFile(const void* data, size_t dataSize, bool mixToMono, float sampleRate)
{
	double startTime = 0;

	if(gWebAudioLog)
		startTime = monotonicallyIncreasingTime();

	RefPtr<AudioBus> bus = AudioStreamReaderFFMPEG (data, dataSize).createBus(sampleRate, mixToMono);

	if(gWebAudioLog){
		double endTime = monotonicallyIncreasingTime();
		gWebAudioLog("(%s, %s, %i) createBusFromInMemoryAudioFile takes %ims\n", __FILE__, __FUNCTION__, __LINE__, (int)((endTime - startTime) * 1000));
	}

	return bus.release();
}

} // WebCore

#else
namespace WebCore
{

PassRefPtr<AudioBus> createBusFromAudioFile(const char*, bool, float sampleRate)
{
	RefPtr<AudioBus> audioBus = AudioBus::create(0, 0, true);
	audioBus->setSampleRate(sampleRate);
	return audioBus.release();
}

PassRefPtr<AudioBus> createBusFromInMemoryAudioFile(const void*, size_t, bool, float sampleRate)
{
	RefPtr<AudioBus> audioBus = AudioBus::create(0, 0, true);
	audioBus->setSampleRate(sampleRate);
	return audioBus.release();
}

} // WebCore
#endif

#endif // ENABLE(WEB_AUDIO)
