/*
 * Copyright (C) 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2009, 2010, 2011 Espial Group Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FileSystem.h"
#include "NotImplemented.h"
#include "PlatformString.h"
#include "TextEncoding.h"
#include <wtf/text/CString.h>
#include <wtf/text/StringBuilder.h>
#include "KURL.h"
#include "PageConsole.h"
#include "Settings.h"
#include "SharedBuffer.h"
#include "BitmapImage.h"
#include <stdlib.h>
#include <fnmatch.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <dlfcn.h>
#include <errno.h>
#include <signal.h>


namespace WebCore {

/* On linux file names are just raw bytes, so also strings that cannot be encoded in any way
 * are valid file names. This mean that we cannot just store a file name as-is in a String
 * but we have to escape it.
 * On Windows the GLib file name encoding is always UTF-8 so we can optimize this case. */
String filenameToString(const char* filename)
{
    if (!filename)
        return String();

    return String::fromUTF8(filename);
}

char* filenameFromString(const String& string)
{
	notImplemented();
	
	return NULL;
}

// Converts a string to something suitable to be displayed to the user.
String filenameForDisplay(const String& string)
{
    return string;
}

static bool fileExists(const char* path)
{	
	struct stat fileInfo;
	return (stat(path, &fileInfo) == 0);
}

DLL_PUBLIC bool fileExists(const String& path)
{
	return fileExists(path.utf8().data());
}

DLL_PUBLIC bool deleteFile(const String& path)
{
	bool deleted = TRUE;
	const CString file(path.utf8());

    if(::unlink(file.data()) != 0) {
        deleted = FALSE;
    }

    return deleted;
}

static PlatformFileHandle openFile(const char* filename, FileOpenMode mode)
{
    int platformFlag;

    if (mode == OpenForRead)
        platformFlag = O_RDONLY;
    else if(fileExists(filename))
		platformFlag = (O_WRONLY | O_TRUNC);
	else
		platformFlag = (O_WRONLY | O_CREAT | O_EXCL);

    return open(filename, platformFlag, 0666);
}

PlatformFileHandle openFile(const String& path, FileOpenMode mode)
{
    return openFile(path.utf8().data(), mode);
}

bool deleteEmptyDirectory(const String& path)
{
	notImplemented();

    return FALSE;
}

long getFileSize(const char* filename)
{			
	struct stat fileInfo;

	if(stat(filename, &fileInfo) == 0){
		if(S_ISREG(fileInfo.st_mode)){
			return fileInfo.st_size;
		}
		
		if(S_ISLNK(fileInfo.st_mode)){
			if(lstat(filename, &fileInfo) == 0){
				return fileInfo.st_size;
			}
		}
	}
		
    return -1;
}


DLL_PUBLIC bool getFileSize(const String& path, long long& resultSize)
{
	const CString file(path.utf8());
				
	resultSize = getFileSize(file.data());
		
    return (resultSize >= 0);
}


bool getFileModificationTime(const String& path, time_t& modifiedTime)
{
	const CString file(path.utf8());
	
	struct stat fileInfo;

	if(stat(file.data(), &fileInfo) != 0) {	
		return FALSE;
	}
	
	modifiedTime = fileInfo.st_mtime;

	return TRUE;
}

bool getFileMetadata(const String& path, FileMetadata& metadata)
{
	const CString file(path.utf8());

    if (path.isNull())
        return false;

    //metadata.modificationTime = statResult.st_mtime;
    //metadata.length = statResult.st_size;
    //metadata.type = S_ISDIR(statResult.st_mode) ? FileMetadata::TypeDirectory : FileMetadata::TypeFile;
    return false;
}

DLL_PUBLIC String pathByAppendingComponent(const String& path, const String& component)
{
	if(path.length() == 0 || component.startsWith("/"))
		return component;

	StringBuilder builder;

	if(path.endsWith("/")) {
		builder.append(path);
		builder.append(component);
	}
	else {
		builder.append(path);
		builder.append('/');
		builder.append(component);
	}

	return builder.toString();
}

DLL_PUBLIC bool makeAllDirectories(const String& path)
{
    CString fullPath(path.utf8());
    if (!access(fullPath.data(), F_OK))
        return true;

    char* p = fullPath.mutableData() + 1;
    int length = fullPath.length();

    if (length > 1) {
        if(p[length - 1] == '/')
            p[length - 1] = '\0';
        for (; *p; ++p)
            if (*p == '/') {
                *p = '\0';
                if (access(fullPath.data(), F_OK))
                    if (mkdir(fullPath.data(), S_IRWXU))
                        return false;
                *p = '/';
            }
    }

    if (access(fullPath.data(), F_OK))
        if (mkdir(fullPath.data(), S_IRWXU))
            return false;

    return true;
}

String homeDirectoryPath()
{
	const char *homedir = getenv("HOME");
	
	//std::cerr << "Home dir [" << homedir << "]\n";
	
	return String(homedir ? homedir: "");
}


String pathGetFileName(const String& pathName)
{
	String fileName = pathName.substring(pathName.reverseFind('/') + 1);
	
	//std::cerr << "pathGetFileName from [" << std_string(pathName) << "] : result [" << std_string(fileName) << "]\n";
	
	return fileName;
}

DLL_PUBLIC String directoryName(const String& pathName)
{
	String dir;
	
	int lastSlash = pathName.reverseFind('/');
	
	if(lastSlash > 0){
		dir = pathName.substring(0, lastSlash);
	}
	else if(lastSlash == 0){
		dir = String("/");
	}
	else{
		dir = String(".");
	}
	
	//std::cerr << "directoryName from [" << std_string(pathName) << "] : result [" << std_string(dir) << "]\n";
		
	return dir;
}

DLL_PUBLIC Vector<String> listDirectory(const String& path, const String& filter)
{
    Vector<String> entries;
    
    struct dirent **namelist;    
    
    const CString dir(path.utf8());
    const CString pattern(filter.utf8());
    
    int n = scandir(dir.data(), &namelist, 0, alphasort);
    
    if(n <= 0)
    	return entries;
    
	while(n--) {
		if(fnmatch(pattern.data(), namelist[n]->d_name, 0) == 0)
			entries.prepend(pathByAppendingComponent(path, String(namelist[n]->d_name)));

		free(namelist[n]);
	}
	
	free(namelist);

    return entries;
}

DLL_PUBLIC String openTemporaryFile(const String& prefix, PlatformFileHandle& handle)
{
    // 1. If TMPDIR exists and it's a valid directory, use it.
    // 2. Otherwise, use P_tmpdir which is "/tmp".
	// 3. This should be pretty standard for unix systems... for filesystems that cant be written to, a TMPDIR
	//	  will have to be set.
    //
    char *tmpdir = getenv("TMPDIR");
    if (tmpdir) {
        struct stat buf;
        if (!stat(tmpdir, &buf) == 0 || !S_ISDIR(buf.st_mode)) {
            tmpdir = NULL;
        }
    }
	
	char path[256];
	snprintf(path, 256, "%s/%s_XXXXXX", (tmpdir ? tmpdir : P_tmpdir), prefix.utf8().data());

    int fd = mkstemp(path);
    
    if (fd >= 0) {		
		handle = fd;
		
        return String(path);
    }

	return String();
}


String openTemporaryFile(const char* prefix, const char* ext, PlatformFileHandle& handle, int extraFlags)
{
    // 1. If TMPDIR exists and it's a valid directory, use it.
    // 2. Otherwise, use P_tmpdir which is "/tmp".
	// 3. This should be pretty standard for unix systems... for filesystems that cant be written to, a TMPDIR
	//	  will have to be set.
    //
    char *tmpdir = getenv("TMPDIR");
    if (tmpdir) {
        struct stat buf;
        if (!stat(tmpdir, &buf) == 0 || !S_ISDIR(buf.st_mode)) {
            tmpdir = NULL;
        }
    }
	
	char tmpStr[256];
	snprintf(tmpStr, 256, "%s/%s_XXXXXX", (tmpdir ? tmpdir : P_tmpdir), prefix);

    int fd = mkstemp(tmpStr);
    
    if(fd >= 0){
    	deleteFile(String(tmpStr));
    	
    	close(fd);
    }
    	
    const CString cStr(tmpStr);//create a temporay file with extension
    	
    snprintf(tmpStr, 256, "%s%s", cStr.data(), ext);
    		
    fd = open(tmpStr, O_CREAT | O_EXCL | O_RDWR | extraFlags);

#if (defined( DEBUG_EVO_PLUGIN ) && DEBUG_EVO_PLUGIN)
    fprintf(stderr, "====>openTemporaryFile(prefix=%s, ext=%s) : temp=%s, fd=%i\n", prefix, ext, tmpStr, fd);
#endif
    					
    if(fd >= 0){
    	close(fd);
    		
    	fd = open(tmpStr, O_RDWR | extraFlags);
    		
    	if(fd >= 0){
    		handle = fd;		
    		return String(tmpStr);
    	}
    }

	return String();
}

DLL_PUBLIC void closeFile(PlatformFileHandle& handle)
{
	if(handle)
	{
		close(handle);
	}
}

long long seekFile(PlatformFileHandle handle, long long offset, FileSeekOrigin origin)
{
    int whence = SEEK_SET;
    switch (origin) {
    case SeekFromBeginning:
        whence = SEEK_SET;
        break;
    case SeekFromCurrent:
        whence = SEEK_CUR;
        break;
    case SeekFromEnd:
        whence = SEEK_END;
        break;
    default:
        ASSERT_NOT_REACHED();
    }
    return static_cast<long long>(lseek(handle, offset, whence));
}

int readFromFile(PlatformFileHandle handle, char* data, int length)
{
    do {
        int bytesRead = read(handle, data, static_cast<size_t>(length));
        if (bytesRead >= 0)
            return bytesRead;
    } while (errno == EINTR);
    return -1;
}

DLL_PUBLIC int writeToFile(const String& path, const char* data, int length)
{
	int written = 0;
	
	PlatformFileHandle handle = openFile(path, OpenForWrite);
	
	if(isHandleValid(handle)){
		written = writeToFile(handle, data, length);
		
		closeFile(handle);
	}
	
	return written;
}

DLL_PUBLIC int writeToFile(PlatformFileHandle handle, const char* data, int length)
{
	unsigned long written = 0;

	if(isHandleValid(handle))
	{
		written = write(handle, data, length);
	}
	
	return written;
}

bool unloadModule(PlatformModule module)
{
	return !dlclose(module);
}


String KURL::fileSystemPath() const
{
	const String p(path());
	
	//std::cerr << "====>KURL::fileSystemPath()\n";
	//std::cerr << "\turl    : " << prettyURL() << std::endl;
	//std::cerr << "\tResult : " << p << std::endl;
	
	return p;
}

static PassRefPtr<SharedBuffer> loadResource(const char* resourceName)
{
	const long fileSize = getFileSize(resourceName);
	
	if(fileSize <= 0)
		return nullptr;

	PlatformFileHandle handle = openFile(resourceName, OpenForRead);
	
	if(!isHandleValid(handle))
		return nullptr;
	
	//std::cerr << "Load resource : " << resourceName << "\n";
	
	RefPtr<SharedBuffer> sbuf = SharedBuffer::create();
	
	if(sbuf){
		Vector<char> buffer;

		buffer.resize(4096);

		size_t remain = fileSize;

		while(remain > 0){
			size_t result = read(handle, buffer.data(), (remain > buffer.size() ? buffer.size() : remain));

			if(result > 0){
				sbuf->append(buffer.data(), result);
				remain -= result;
			}
			else
				break;
		}
	}

	closeFile(handle);

    return sbuf.release();
}

DLL_PUBLIC PassRefPtr<SharedBuffer> SharedBuffer::createWithContentsOfFile(const String& filePath)
{
	if (filePath.isEmpty())
	    return nullptr;
	
	return loadResource(filePath.utf8().data());
}

DLL_PUBLIC PassRefPtr<Image> Image::loadPlatformResource(const char *name)
{
	String fileName = pathByAppendingComponent(Settings::platformResourcePath(), String(name));

	RefPtr<SharedBuffer> sbuf(loadResource(fileName.utf8().data()));
	
	if(!sbuf){
		String message = String::fromLiteral("Image::loadPlatformResource(") + fileName + ") failed";
		addMessageToConsole(OtherMessageSource, LogMessageType, WarningMessageLevel, message);
		return nullptr;
	}
	
    RefPtr<BitmapImage> img = BitmapImage::create();
        
    img->setData(sbuf, true);
        
	//IntSize size = img->size();
	//std::cerr << "loadPlatformResource : " << name << " [width " << size.width() << ", height " << size.height() << "]\n";
	
    return img.release();
}

}
