/*
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "ScrollbarThemeEspial.h"

#include "PlatformMouseEvent.h"
#include "RenderThemeEspial.h"
#include "ScrollView.h"
#include "Scrollbar.h"

namespace WebCore {

DLL_PUBLIC RefPtr<Image> ScrollbarThemeEspial::s_scrollArrowUp;
DLL_PUBLIC RefPtr<Image> ScrollbarThemeEspial::s_scrollArrowDown;
DLL_PUBLIC RefPtr<Image> ScrollbarThemeEspial::s_scrollArrowLeft;
DLL_PUBLIC RefPtr<Image> ScrollbarThemeEspial::s_scrollArrowRight;
DLL_PUBLIC RefPtr<Image> ScrollbarThemeEspial::s_scrollBackground;
DLL_PUBLIC RefPtr<Image> ScrollbarThemeEspial::s_scrollRail;
DLL_PUBLIC RefPtr<Image> ScrollbarThemeEspial::s_scrollThumb;

ScrollbarTheme* ScrollbarTheme::nativeTheme()
{
    static ScrollbarThemeEspial theme;
    return &theme;
}

ScrollbarThemeEspial::ScrollbarThemeEspial()
: m_arrowBgColor(200, 200, 255)
, m_arrowColor(0, 0, 0)
{
	platformInit();
}

ScrollbarThemeEspial::~ScrollbarThemeEspial()
{
}

bool ScrollbarThemeEspial::hasThumb(ScrollbarThemeClient* scrollbar)
{
    // This method is just called as a paint-time optimization to see if
    // painting the thumb can be skipped.  We don't have to be exact here.
    return thumbLength(scrollbar) > 0;
}

void ScrollbarThemeEspial::paintArrow(GraphicsContext* ctx, const IntRect& rect, Image* image, float rotation)
{
    float w = rect.width();
    float h = rect.height();

    // paint image
    if (image) {
        int iw = image->width();
        int ih = image->height();
        ctx->drawImage(image, IntRect(rect.x() + (w - iw) / 2, rect.y() + (h - ih) / 2, iw, ih));
    }
    // paint triangle
    else {
        ctx->save();
        ctx->translate(rect.x(), rect.y());
        ctx->fillRect(FloatRect(0, 0, w, h), m_arrowBgColor);

        ctx->setFillColor(m_arrowColor);
        ctx->setStrokeColor(NoStroke);
    
        ctx->translate(w / 2.0f, h / 2.0f);
        ctx->rotate(rotation);

        FloatPoint topArrow[3];
        topArrow[0] = FloatPoint(0.0f, 1.0f - h / 2.0f);
        topArrow[1] = FloatPoint( w / 2.0f - 2.0f, h / 2.0f - 3.0f);
        topArrow[2] = FloatPoint(-w / 2.0f + 2.0f, h / 2.0f - 3.0f);
        ctx->drawConvexPolygon(3, topArrow, true);
        ctx->restore();
    }
}

static inline void drawImage(GraphicsContext* ctx, Image* image, const FloatRect& rect)
{
    ctx->drawImage(image, rect);
}

bool ScrollbarThemeEspial::paint(ScrollbarThemeClient* scrollbar, GraphicsContext* ctx, const IntRect& damageRect)
{
    if (ctx->paintingDisabled())
        return false;

    int x = scrollbar->x();
    int y = scrollbar->y();
    
    const IntRect frameRect(scrollbar->frameRect());
    
    // clip and translate
    ctx->save();
    ctx->clip(intersection(damageRect, frameRect));
    ctx->translate(x, y);

    // paint border
    ctx->fillRect(IntRect(0, 0, frameRect.width(), frameRect.height()), m_borderColor);
    ctx->translate(m_borderWidth, m_borderWidth);
    
    int w = frameRect.width()  - 2.0f * m_borderWidth;
    int h = frameRect.height() - 2.0f * m_borderWidth;
    
    // paint background
    IntRect bgRect(0, 0, w, h);
    if (s_scrollBackground) {
        drawImage(ctx, s_scrollBackground.get(), bgRect);
    } else {
        ctx->fillRect(bgRect, m_bgColor);
    }

    // layout arrows
    IntRect upRect, downRect;
    if (scrollbar->orientation() == HorizontalScrollbar) {
        upRect   = IntRect(    0, 0, h, h);
        downRect = IntRect(w - h, 0, h, h);
    } else {
        upRect   = IntRect(0,     0, w, w);
        downRect = IntRect(0, h - w, w, w);
    }

    // layout rail and thumb
    float sep = separatorWidth();
    float size = (float)scrollbar->visibleSize() / scrollbar->totalSize();
    float offset = (float)scrollbar->currentPos() / scrollbar->totalSize();
    IntRect railRect;
    IntRect thumbRect;
    if (scrollbar->orientation() == HorizontalScrollbar) {
        int railWidth = (m_railWidth < 0) ? h : m_railWidth;
        railRect = IntRect(h + sep, (h - railWidth) / 2, w - 2 * (h + sep), railWidth);
        thumbRect = railRect;
        thumbRect.setWidth(railRect.width() * size);
        thumbRect.setX(railRect.x() + railRect.width() * offset);
    } else {
        int railWidth = (m_railWidth < 0) ? w : m_railWidth;
        railRect = IntRect((w - railWidth) / 2, w + sep, railWidth, h - 2 * (w + sep));
        thumbRect = railRect;
        thumbRect.setHeight(railRect.height() * size);
        thumbRect.setY(railRect.y() + railRect.height() * offset);
    }

    // paint rail
    if (s_scrollRail) {
        drawImage(ctx, s_scrollRail.get(), railRect);
    } else {
        ctx->fillRect(railRect, m_railColor);
    }

    // paint separator
    // Note: The separator is only painted if there are no arrow images. This is
    // for backwards compatibility with the only scrollbar painting code. If you
    // want to use no arrow images with no separator, set a SeparatorColor of
    // transparent (e.g. 00000000). If you want separators with your arrow images,
    // put them in the image itself.
    if (scrollbar->orientation() == HorizontalScrollbar) {
        if (!s_scrollArrowLeft.get())  ctx->fillRect(IntRect(          h, 0, sep, h), m_separatorColor);
        if (!s_scrollArrowRight.get()) ctx->fillRect(IntRect(w - h - sep, 0, sep, h), m_separatorColor);
    } else {
        if (!s_scrollArrowUp.get())    ctx->fillRect(IntRect(0,           w, w, sep), m_separatorColor);
        if (!s_scrollArrowDown.get())  ctx->fillRect(IntRect(0, h - w - sep, w, sep), m_separatorColor);
    }
    
    // paint thumb
    if (s_scrollThumb) {
        drawImage(ctx, s_scrollThumb.get(), thumbRect);
    } else {
        ctx->fillRect(thumbRect, m_thumbColor);
    }

    // paint arrows
    if (scrollbar->orientation() == HorizontalScrollbar) {
        paintArrow(ctx, upRect,   s_scrollArrowLeft.get(),  -M_PI/2.0f);
        paintArrow(ctx, downRect, s_scrollArrowRight.get(),  M_PI/2.0f);
    } else {
        paintArrow(ctx, upRect,   s_scrollArrowUp.get(),    0.0f);
        paintArrow(ctx, downRect, s_scrollArrowDown.get(),  M_PI);
    }

    ctx->restore();
    
    return true;
}

}

