/*
 * Copyright (C) 2006, 2008 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2007-2008 Torch Mobile, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef FontCache_h
#define FontCache_h

#include "FontDescription.h"
#include <limits.h>
#include <wtf/Forward.h>
#include <wtf/HashSet.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>
#include <wtf/unicode/Unicode.h>

#if OS(WINDOWS)
#include <windows.h>
#include <objidl.h>
#include <mlang.h>
#endif

#include "FontPlatformData.h"
#include "RefPtrCairo.h"

namespace WebCore
{

class Font;
class FontPlatformData;
class FontSelector;
class OpenTypeVerticalData;
class SimpleFontData;

// This key contains the FontDescription fields other than family that matter when fetching FontDatas (platform fonts).
struct FontDescriptionKey {
    explicit FontDescriptionKey(unsigned size = 0)
        : m_size(size)
    	, m_horizontalSize(0)
        , m_weight(0)
    	, m_stretch(0)
        , m_flags(0)
    { }
    FontDescriptionKey(const FontDescription& description)
        : m_size(description.computedSize())
    	, m_horizontalSize(description.computedHorizontalSize())
        , m_weight(description.weight())
    	, m_stretch(description.stretch())
        , m_flags(makeFlagKey(description))
    { }

    explicit FontDescriptionKey(WTF::HashTableDeletedValueType)
        : m_size(cHashTableDeletedSize)
		, m_horizontalSize(0)
		, m_weight(0)
		, m_stretch(0)
		, m_flags(0)
    { }

    bool operator==(const FontDescriptionKey& other) const
    {
        return m_size == other.m_size
        	&& m_horizontalSize == other.m_horizontalSize
			&& m_weight == other.m_weight
            && m_stretch == other.m_stretch
			&& m_flags == other.m_flags;
    }
    bool operator!=(const FontDescriptionKey& other) const
    {
        return !(*this == other);
    }

    float size() const { return m_size; }

    bool isHashTableDeletedValue() const { return m_size == cHashTableDeletedSize; }

    inline unsigned computeHash() const
    {
        IntegerHasher hasher;
        hasher.add(m_size);
        hasher.add(m_horizontalSize);
        hasher.add(m_weight);
        hasher.add(m_stretch);
		hasher.add(m_flags);
        return hasher.hash();
    }

private:
    static unsigned makeFlagKey(const FontDescription& description)
    {
    	//static_assert(USCRIPT_CODE_LIMIT < 0x1000, "Script code must fit in an unsigned along with the other flags");
        return static_cast<unsigned>(description.script()) << 11
			| static_cast<unsigned>(description.fontSizeClamp()) << 10
			| static_cast<unsigned>(description.smallCaps()) << 9
			| static_cast<unsigned>(description.fontSynthesis()) << 7
			| static_cast<unsigned>(description.widthVariant()) << 5
            | static_cast<unsigned>(description.nonCJKGlyphOrientation()) << 4
            | static_cast<unsigned>(description.orientation()) << 3
            | static_cast<unsigned>(description.italic()) << 2
            | static_cast<unsigned>(description.usePrinterFont()) << 1
            | static_cast<unsigned>(description.renderingMode());
    }

    static const unsigned cHashTableDeletedSize = 0xFFFFFFFFU;

    float m_size;
    float m_horizontalSize;
    unsigned m_weight;
    float m_stretch;
    unsigned m_flags;
};

struct FontDescriptionKeyHash {
    static unsigned hash(const FontDescriptionKey& key)
    {
        return key.computeHash();
    }

    static bool equal(const FontDescriptionKey& a, const FontDescriptionKey& b)
    {
        return a == b;
    }

    static const bool safeToCompareToEmptyOrDeleted = true;
};

class FontCache {
    friend class FontCachePurgePreventer;

    WTF_MAKE_NONCOPYABLE(FontCache); WTF_MAKE_FAST_ALLOCATED;
public:
    friend FontCache& fontCache();

    // This method is implemented by the platform.
    PassRefPtr<SimpleFontData> systemFallbackForCharacters(const FontDescription&, const SimpleFontData* originalFontData, bool isPlatformFont, const UChar* characters, int length);

    // Also implemented by the platform.
    void platformInit();

#if OS(WINCE) && !PLATFORM(QT)
#if defined(IMLANG_FONT_LINK) && (IMLANG_FONT_LINK == 2)
    IMLangFontLink2* getFontLinkInterface();
#else
    IMLangFontLink* getFontLinkInterface();
#endif
    static void comInitialize();
    static void comUninitialize();
    static IMultiLanguage* getMultiLanguageInterface();
#elif PLATFORM(WIN)
    IMLangFontLink2* getFontLinkInterface();
#endif

    void getTraitsInFamily(const AtomicString&, Vector<unsigned>&);

    PassRefPtr<SimpleFontData> fontForFamily(const FontDescription&, const AtomicString&, bool checkingAlternateName = false);
    PassRefPtr<SimpleFontData> lastResortFallbackFont(const FontDescription&);
    SimpleFontData* getNonRetainedLastResortFallbackFont(const FontDescription&);
    PassRefPtr<SimpleFontData> fontForPlatformData(const FontPlatformData&);
    PassRefPtr<SimpleFontData> similarFontPlatformData(const FontDescription&);

    UChar32 missingCharacter(UChar32 c) const { return c == m_fontMissingCharacter ? 0 : m_fontMissingCharacter; }
    void setMissingCharacter(UChar32 c) { m_fontMissingCharacter = c; }
    
    void addClient(FontSelector*);
    void removeClient(FontSelector*);

    unsigned short generation();
    void invalidate();

    const FontCharReplaceMap& similarCharacters() const { return m_similarCharacters; }

    void addSimilarCharacter(UChar32 c, UChar32 replacement)
    {
    	m_similarCharacters.add(c, replacement);
    }

    void clearSimilarCharacters()
    {
    	m_similarCharacters.clear();
    }

    const Vector<String>& systemFontFamilyNames() const { return m_systemFontFamilyNames; }
    void scanSystemFontFamilyNames();

    size_t fontDataCount();
    size_t inactiveFontDataCount();
    void purgeInactiveFontData(int count = INT_MAX);
    void purgeInactiveFontDataIfNeeded();

    void schedulePurgeInactiveFontData();

#if PLATFORM(WIN)
    PassRefPtr<SimpleFontData> fontDataFromDescriptionAndLogFont(const FontDescription&, ShouldRetain, const LOGFONT& font, AtomicString& outFontFamilyName);
#elif PLATFORM(CHROMIUM) && OS(WINDOWS)
    PassRefPtr<SimpleFontData> fontDataFromDescriptionAndLogFont(const FontDescription&, ShouldRetain, const LOGFONT& font, wchar_t* outFontFamilyName);
#endif

#if ENABLE(OPENTYPE_VERTICAL)
#if USE(SKIA)
    typedef uint32_t FontFileKey;
#else
    typedef AtomicString FontFileKey;
#endif
    PassRefPtr<OpenTypeVerticalData> getVerticalData(const FontFileKey&, const FontPlatformData&);
#endif

private:
    FontCache();
    ~FontCache();

    void disablePurging() { m_purgePreventCount++; }
    void enablePurging()
    {
        ASSERT(m_purgePreventCount);
        if (!--m_purgePreventCount)
            purgeInactiveFontDataIfNeeded();
    }

    // FIXME: This method should eventually be removed.
    FontPlatformData* getCachedFontPlatformData(const FontDescription&, const AtomicString& family, bool checkingAlternateName = false);

    // These methods are implemented by each platform.
    FontPlatformData* createFontPlatformData(const FontDescription&, const AtomicString& family);

    // Don't purge if this count is > 0;
    int m_purgePreventCount;

    UChar32 m_fontMissingCharacter;

    FontCharReplaceMap m_similarCharacters;

    Vector<String> m_systemFontFamilyNames;
    RefPtr<FcPattern> m_perCharactersPattern;

#if PLATFORM(MAC) || (PLATFORM(CHROMIUM) && OS(DARWIN)) || OS(ANDROID)
    friend class ComplexTextController;
#endif
    friend class SimpleFontData; // For getCachedFontData(const FontPlatformData*)
};

// Get the global fontCache.
FontCache& fontCache();

class FontCachePurgePreventer {
public:
    FontCachePurgePreventer() { fontCache().disablePurging(); }
    ~FontCachePurgePreventer() { fontCache().enablePurging(); }
};

#if !PLATFORM(COCOA)
inline PassRefPtr<SimpleFontData> FontCache::similarFontPlatformData(const FontDescription&)
{
    return nullptr;
}
#endif

}

#endif
