/*
 * Copyright (C) 2000 Lars Knoll (knoll@kde.org)
 *           (C) 2000 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Nicholas Shanks <webkit@nickshanks.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIother.m_  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USm_
 *
 */

#ifndef FontDescription_h
#define FontDescription_h

#include "FontFamily.h"
#include "FontFeatureSettings.h"
#include "FontOrientation.h"
#include "FontRenderingMode.h"
#include "FontSmoothingMode.h"
#include "FontTraitsMask.h"
#include "FontWidthVariant.h"
#include "TextFlags.h"
#include "TextRenderingMode.h"
#include "WebKitFontFamilyNames.h"
#include <wtf/MathExtras.h>
#include <wtf/RefCountedArray.h>
#include <wtf/RefPtr.h>

namespace WebCore {

using namespace WebKitFontFamilyNames;

enum FontWeight {
    FontWeight100,
    FontWeight200,
    FontWeight300,
    FontWeight400,
    FontWeight500,
    FontWeight600,
    FontWeight700,
    FontWeight800,
    FontWeight900,
    FontWeightNormal = FontWeight400,
    FontWeightBold = FontWeight700
};

enum FontItalic {
    FontItalicOff = 0,
    FontItalicOn = 1
};

enum FontSmallCaps {
    FontSmallCapsOff = 0,
    FontSmallCapsOn = 1
};

class FontDescription {
public:
    enum Kerning { AutoKerning, NormalKerning, NoneKerning };

    enum LigaturesState { NormalLigaturesState, DisabledLigaturesState, EnabledLigaturesState };

    FontDescription()
		: m_families(1)
		, m_specifiedSize(0)
        , m_computedSize(0)
    	, m_specifiedHorizontalSize(0)
    	, m_computedHorizontalSize(0)
		, m_stretch(100)
        , m_orientation(Horizontal)
    	, m_nonCJKGlyphOrientation(static_cast<unsigned>(NonCJKGlyphOrientation::Mixed))
        , m_widthVariant(RegularWidth)
        , m_italic(FontItalicOff)
        , m_smallCaps(FontSmallCapsOff)
        , m_isAbsoluteSize(false)
    	, m_isAbsoluteHorizontalSize(false)
        , m_weight(FontWeightNormal)
    	, m_renderingMode(static_cast<unsigned>(FontRenderingMode::Normal))
        , m_kerning(AutoKerning)
        , m_commonLigaturesState(NormalLigaturesState)
        , m_discretionaryLigaturesState(NormalLigaturesState)
        , m_historicalLigaturesState(NormalLigaturesState)
        , m_keywordSize(0)
        , m_fontSmoothing(AutoSmoothing)
        , m_textRendering(AutoTextRendering)
        , m_isSpecifiedFont(false)
        , m_script(USCRIPT_COMMON)
    	, m_fontSynthesis(initialFontSynthesis())
    	, m_fontSizeClamp(false)
    {
    }

    bool operator==(const FontDescription&) const;
    ALWAYS_INLINE bool operator!=(const FontDescription& other) const { return !(*this == other); }

    unsigned familyCount() const { return m_families.size(); }
    const AtomicString& firstFamily() const { return familyAt(0); }
    const AtomicString& familyAt(unsigned i) const { return m_families[i]; }
    const RefCountedArray<AtomicString>& families() const { return m_families; }

    ALWAYS_INLINE float specifiedSize() const { return m_specifiedSize; }
    ALWAYS_INLINE float computedSize() const { return m_computedSize; }
    ALWAYS_INLINE unsigned computedPixelSize() const { return unsigned(m_computedSize + 0.5f); }
    ALWAYS_INLINE float specifiedHorizontalSize() const { return m_specifiedHorizontalSize; }
    ALWAYS_INLINE float computedHorizontalSize() const { return m_computedHorizontalSize; }
    ALWAYS_INLINE FontItalic italic() const { return static_cast<FontItalic>(m_italic); }
    float stretch() const { return m_stretch; }
    bool usePrinterFont() const { return false; }
    ALWAYS_INLINE int computedHorizontalPixelSize() const { return int(m_computedHorizontalSize + 0.5f); }
    ALWAYS_INLINE FontSmallCaps smallCaps() const { return static_cast<FontSmallCaps>(m_smallCaps); }
    ALWAYS_INLINE bool isAbsoluteSize() const { return m_isAbsoluteSize; }
    ALWAYS_INLINE bool isAbsoluteHorizontalSize() const { return m_isAbsoluteHorizontalSize; }
    ALWAYS_INLINE FontWeight weight() const { return static_cast<FontWeight>(m_weight); }
    FontWeight lighterWeight() const;
    FontWeight bolderWeight() const;
    // only use fixed default size when there is only one font family, and that family is "monospace"
    ALWAYS_INLINE bool useFixedDefaultSize() const { return familyCount() == 1 && firstFamily() == monospaceFamily; }
    ALWAYS_INLINE FontRenderingMode renderingMode() const { return static_cast<FontRenderingMode>(m_renderingMode); }
    ALWAYS_INLINE Kerning kerning() const { return static_cast<Kerning>(m_kerning); }
    ALWAYS_INLINE LigaturesState commonLigaturesState() const { return static_cast<LigaturesState>(m_commonLigaturesState); }
    ALWAYS_INLINE LigaturesState discretionaryLigaturesState() const { return static_cast<LigaturesState>(m_discretionaryLigaturesState); }
    ALWAYS_INLINE LigaturesState historicalLigaturesState() const { return static_cast<LigaturesState>(m_historicalLigaturesState); }
    ALWAYS_INLINE unsigned keywordSize() const { return m_keywordSize; }
    ALWAYS_INLINE FontSmoothingMode fontSmoothing() const { return static_cast<FontSmoothingMode>(m_fontSmoothing); }
    ALWAYS_INLINE TextRenderingMode textRenderingMode() const { return static_cast<TextRenderingMode>(m_textRendering); }
    ALWAYS_INLINE UScriptCode script() const { return static_cast<UScriptCode>(m_script); }

    FontTraitsMask traitsMask() const;
    ALWAYS_INLINE bool isSpecifiedFont() const { return m_isSpecifiedFont; }
    ALWAYS_INLINE FontOrientation orientation() const { return static_cast<FontOrientation>(m_orientation); }
    ALWAYS_INLINE NonCJKGlyphOrientation nonCJKGlyphOrientation() const { return static_cast<NonCJKGlyphOrientation>(m_nonCJKGlyphOrientation); }
    ALWAYS_INLINE FontWidthVariant widthVariant() const { return static_cast<FontWidthVariant>(m_widthVariant); }
    ALWAYS_INLINE FontFeatureSettings* featureSettings() const { return m_featureSettings.get(); }
    ALWAYS_INLINE FontSynthesis fontSynthesis() const { return static_cast<FontSynthesis>(m_fontSynthesis); }
    ALWAYS_INLINE bool fontSizeClamp() const { return m_fontSizeClamp; }
    FontDescription makeNormalFeatureSettings() const;

    ALWAYS_INLINE void setOneFamily(const AtomicString& family) { ASSERT(m_families.size() == 1); m_families[0] = family; }
    void setFamilies(const Vector<AtomicString>& families) { m_families = RefCountedArray<AtomicString>(families); }
	void setFamilies(const RefCountedArray<AtomicString>& families) { m_families = families; }
	ALWAYS_INLINE void setComputedSize(float s) { m_computedSize = clampToFloat(s); }
    ALWAYS_INLINE void setSpecifiedSize(float s) { m_specifiedSize = clampToFloat(s); }
	ALWAYS_INLINE void setComputedHorizontalSize(float s) { m_computedHorizontalSize = clampToFloat(s); }
    ALWAYS_INLINE void setSpecifiedHorizontalSize(float s) { m_specifiedHorizontalSize = clampToFloat(s); }
    ALWAYS_INLINE void setItalic(FontItalic i) { m_italic = i; }
    void setStretch(float stretch) { m_stretch = stretch; }
    ALWAYS_INLINE void setItalic(bool i) { setItalic(i ? FontItalicOn : FontItalicOff); }
    ALWAYS_INLINE void setSmallCaps(FontSmallCaps c) { m_smallCaps = c; }
    ALWAYS_INLINE void setSmallCaps(bool c) { setSmallCaps(c ? FontSmallCapsOn : FontSmallCapsOff); }
    ALWAYS_INLINE void setIsAbsoluteSize(bool s) { m_isAbsoluteSize = s; }
    ALWAYS_INLINE void setIsAbsoluteHorizontalSize(bool s) { m_isAbsoluteHorizontalSize = s; }
    ALWAYS_INLINE void setWeight(FontWeight w) { m_weight = w; }
    ALWAYS_INLINE void setRenderingMode(FontRenderingMode mode) { m_renderingMode = static_cast<unsigned>(mode); }
    ALWAYS_INLINE void setKerning(Kerning kerning) { m_kerning = kerning; }
    ALWAYS_INLINE void setCommonLigaturesState(LigaturesState commonLigaturesState) { m_commonLigaturesState = commonLigaturesState; }
    ALWAYS_INLINE void setDiscretionaryLigaturesState(LigaturesState discretionaryLigaturesState) { m_discretionaryLigaturesState = discretionaryLigaturesState; }
    ALWAYS_INLINE void setHistoricalLigaturesState(LigaturesState historicalLigaturesState) { m_historicalLigaturesState = historicalLigaturesState; }
    ALWAYS_INLINE void setKeywordSize(unsigned s) { m_keywordSize = s; }
    ALWAYS_INLINE void setFontSmoothing(FontSmoothingMode smoothing) { m_fontSmoothing = smoothing; }
    ALWAYS_INLINE void setTextRenderingMode(TextRenderingMode rendering) { m_textRendering = rendering; }
    ALWAYS_INLINE void setIsSpecifiedFont(bool isSpecifiedFont) { m_isSpecifiedFont = isSpecifiedFont; }
    ALWAYS_INLINE void setOrientation(FontOrientation orientation) { m_orientation = orientation; }
    ALWAYS_INLINE void setNonCJKGlyphOrientation(NonCJKGlyphOrientation orientation) { m_nonCJKGlyphOrientation = static_cast<unsigned>(orientation); }
    ALWAYS_INLINE void setWidthVariant(FontWidthVariant widthVariant) { m_widthVariant = widthVariant; }
    ALWAYS_INLINE void setScript(UScriptCode s) { m_script = s; }
    ALWAYS_INLINE void setFeatureSettings(PassRefPtr<FontFeatureSettings> settings) { m_featureSettings = settings; }
    ALWAYS_INLINE void setFontSynthesis(FontSynthesis fontSynthesis) { m_fontSynthesis = fontSynthesis; }
    ALWAYS_INLINE void setFontSizeClamp(bool clamp) { m_fontSizeClamp = clamp; }

    static FontSynthesis initialFontSynthesis() { return FontSynthesisWeight | FontSynthesisStyle; }

private:
    RefCountedArray<AtomicString> m_families;
    RefPtr<FontFeatureSettings> m_featureSettings;

    float m_specifiedSize;   // Specified CSS value. Independent of rendering issues such as integer
                             // rounding, minimum font sizes, and zooming.
    float m_computedSize;    // Computed size adjusted for the minimum font size and the zoom factor.  

    float m_specifiedHorizontalSize;   // Specified CSS value. Independent of rendering issues such as integer
                             // rounding, minimum font sizes, and zooming.
    float m_computedHorizontalSize;    // Computed size adjusted for the minimum font size and the zoom factor.

    float m_stretch; // Stretch, or "width," of the font
    unsigned m_orientation : 1; // FontOrientation - Whether the font is rendering on a horizontal line or a vertical line.
    unsigned m_nonCJKGlyphOrientation : 1; // NonCJKGlyphOrientation - Only used by vertical text. Determines the default orientation for non-ideograph glyphs.

    unsigned m_widthVariant : 2; // FontWidthVariant

    unsigned m_italic : 1; // FontItalic
    unsigned m_smallCaps : 1; // FontSmallCaps
    unsigned m_isAbsoluteSize : 1; // Whether or not CSS specified an explicit size
                                  // (logical sizes like "medium" don't count).
    unsigned m_isAbsoluteHorizontalSize : 1; // Whether or not CSS specified an explicit size
                                  // (logical sizes like "medium" don't count).
    unsigned m_weight : 8; // FontWeight

    unsigned m_renderingMode : 1;  // Used to switch between CG and GDI text on Windows.
    unsigned m_kerning : 2; // Kerning

    unsigned m_commonLigaturesState : 2;
    unsigned m_discretionaryLigaturesState : 2;
    unsigned m_historicalLigaturesState : 2;

    unsigned m_keywordSize : 4; // We cache whether or not a font is currently represented by a CSS keyword (e.g., medium).  If so,
                           // then we can accurately translate across different generic families to adjust for different preference settings
                           // (e.g., 13px monospace vs. 16px everything else).  Sizes are 1-8 (like the HTML size values for <font>).

    unsigned m_fontSmoothing : 2; // FontSmoothingMode
    unsigned m_textRendering : 2; // TextRenderingMode
    unsigned m_isSpecifiedFont : 1; // True if a web page specifies a non-generic font family as the first font family.
    unsigned m_script : 7; // Used to help choose an appropriate font for generic font families.
    unsigned m_fontSynthesis : 2; // FontSynthesis type

    unsigned m_fontSizeClamp : 1;
};

inline bool FontDescription::operator==(const FontDescription& other) const
{
    return m_families == other.m_families
        && m_specifiedSize == other.m_specifiedSize
        && m_computedSize == other.m_computedSize
		&& m_specifiedHorizontalSize == other.m_specifiedHorizontalSize
		&& m_computedHorizontalSize == other.m_computedHorizontalSize
        && m_stretch == other.m_stretch
        && m_italic == other.m_italic
        && m_smallCaps == other.m_smallCaps
        && m_isAbsoluteSize == other.m_isAbsoluteSize
        && m_isAbsoluteHorizontalSize == other.m_isAbsoluteHorizontalSize
        && m_weight == other.m_weight
        && m_renderingMode == other.m_renderingMode
        && m_kerning == other.m_kerning
        && m_commonLigaturesState == other.m_commonLigaturesState
        && m_discretionaryLigaturesState == other.m_discretionaryLigaturesState
        && m_historicalLigaturesState == other.m_historicalLigaturesState
        && m_keywordSize == other.m_keywordSize
        && m_fontSmoothing == other.m_fontSmoothing
        && m_textRendering == other.m_textRendering
        && m_isSpecifiedFont == other.m_isSpecifiedFont
        && m_orientation == other.m_orientation
        && m_nonCJKGlyphOrientation == other.m_nonCJKGlyphOrientation
        && m_widthVariant == other.m_widthVariant
        && m_script == other.m_script
        && m_featureSettings == other.m_featureSettings
        && m_fontSynthesis == other.m_fontSynthesis
		&& m_fontSizeClamp == other.m_fontSizeClamp;
}

String getFamilyNameStringFromFontDescriptionAndFamily(const FontDescription& fontDescription, const AtomicString& family);

#if 0
// FIXME: Move to a file of its own.
class FontCascadeDescription : public FontDescription {
public:
    FontCascadeDescription();

    bool operator==(const FontCascadeDescription&) const;
    bool operator!=(const FontCascadeDescription& other) const { return !(*this == other); }

    unsigned familyCount() const { return m_families.size(); }
    const AtomicString& firstFamily() const { return familyAt(0); }
    const AtomicString& familyAt(unsigned i) const { return m_families[i]; }
    const RefCountedArray<AtomicString>& families() const { return m_families; }

    float specifiedSize() const { return m_specifiedSize; }
    bool isAbsoluteSize() const { return m_isAbsoluteSize; }
    FontWeight lighterWeight() const;
    FontWeight bolderWeight() const;

    // only use fixed default size when there is only one font family, and that family is "monospace"
    bool useFixedDefaultSize() const { return familyCount() == 1 && firstFamily() == monospaceFamily; }

    Kerning kerning() const { return static_cast<Kerning>(m_kerning); }
    unsigned keywordSize() const { return m_keywordSize; }
    CSSValueID keywordSizeAsIdentifier() const
    {
        CSSValueID identifier = m_keywordSize ? static_cast<CSSValueID>(CSSValueXxSmall + m_keywordSize - 1) : CSSValueInvalid;
        ASSERT(identifier == CSSValueInvalid || (identifier >= CSSValueXxSmall && identifier <= CSSValueWebkitXxxLarge));
        return identifier;
    }
    FontSmoothingMode fontSmoothing() const { return static_cast<FontSmoothingMode>(m_fontSmoothing); }
    bool isSpecifiedFont() const { return m_isSpecifiedFont; }

    void setOneFamily(const AtomicString& family) { ASSERT(m_families.size() == 1); m_families[0] = family; }
    void setFamilies(const Vector<AtomicString>& families) { m_families = RefCountedArray<AtomicString>(families); }
    void setFamilies(const RefCountedArray<AtomicString>& families) { m_families = families; }
    void setSpecifiedSize(float s) { m_specifiedSize = clampToFloat(s); }
    void setIsAbsoluteSize(bool s) { m_isAbsoluteSize = s; }
    void setKerning(Kerning kerning) { m_kerning = static_cast<unsigned>(kerning); }
    void setKeywordSize(unsigned size)
    {
        ASSERT(size <= 8);
        m_keywordSize = size;
        ASSERT(m_keywordSize == size); // Make sure it fits in the bitfield.
    }
    void setKeywordSizeFromIdentifier(CSSValueID identifier)
    {
        ASSERT(!identifier || (identifier >= CSSValueXxSmall && identifier <= CSSValueWebkitXxxLarge));
        static_assert(CSSValueWebkitXxxLarge - CSSValueXxSmall + 1 == 8, "Maximum keyword size should be 8.");
        setKeywordSize(identifier ? identifier - CSSValueXxSmall + 1 : 0);
    }
    void setFontSmoothing(FontSmoothingMode smoothing) { m_fontSmoothing = smoothing; }
    void setIsSpecifiedFont(bool isSpecifiedFont) { m_isSpecifiedFont = isSpecifiedFont; }

#if ENABLE(TEXT_AUTOSIZING)
    bool familiesEqualForTextAutoSizing(const FontCascadeDescription& other) const;

    bool equalForTextAutoSizing(const FontCascadeDescription& other) const
    {
        return familiesEqualForTextAutoSizing(other)
            && m_specifiedSize == other.m_specifiedSize
            && variantSettings() == other.variantSettings()
            && m_isAbsoluteSize == other.m_isAbsoluteSize;
    }
#endif

    // Initial values for font properties.
    static FontItalic initialItalic() { return FontItalicOff; }
    static float initialStretch() { return 100; }
    static FontSmallCaps initialSmallCaps() { return FontSmallCapsOff; }
    static Kerning initialKerning() { return Kerning::Auto; }
    static FontSmoothingMode initialFontSmoothing() { return AutoSmoothing; }
    static TextRenderingMode initialTextRenderingMode() { return AutoTextRendering; }
    static FontSynthesis initialFontSynthesis() { return FontSynthesisWeight | FontSynthesisStyle | FontSynthesisSmallCaps; }
    static FontVariantPosition initialVariantPosition() { return FontVariantPosition::Normal; }
    static FontVariantCaps initialVariantCaps() { return FontVariantCaps::Normal; }
    static FontVariantAlternates initialVariantAlternates() { return FontVariantAlternates::Normal; }
    static const AtomicString& initialLocale() { return nullAtom; }

private:
    RefCountedArray<AtomicString> m_families { 1 };

    float m_specifiedSize { 0 };   // Specified CSS value. Independent of rendering issues such as integer
                             // rounding, minimum font sizes, and zooming.
    unsigned m_isAbsoluteSize : 1; // Whether or not CSS specified an explicit size
                                  // (logical sizes like "medium" don't count).
    unsigned m_kerning : 2; // Kerning

    unsigned m_keywordSize : 4; // We cache whether or not a font is currently represented by a CSS keyword (e.g., medium).  If so,
                           // then we can accurately translate across different generic families to adjust for different preference settings
                           // (e.g., 13px monospace vs. 16px everything else).  Sizes are 1-8 (like the HTML size values for <font>).

    unsigned m_fontSmoothing : 2; // FontSmoothingMode
    unsigned m_isSpecifiedFont : 1; // True if a web page specifies a non-generic font family as the first font family.
};

inline bool FontCascadeDescription::operator==(const FontCascadeDescription& other) const
{
    return FontDescription::operator==(other)
        && m_families == other.m_families
        && m_specifiedSize == other.m_specifiedSize
        && m_isAbsoluteSize == other.m_isAbsoluteSize
        && m_kerning == other.m_kerning
        && m_keywordSize == other.m_keywordSize
        && m_fontSmoothing == other.m_fontSmoothing
        && m_isSpecifiedFont == other.m_isSpecifiedFont;
}
#endif

}

#endif
