/*
 * Copyright (C) 2006, 2009 Apple Inc. All rights reserved.
 * Copyright (C) 2009, 2010, 2011 Espial Group Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef GlyphBuffer_h
#define GlyphBuffer_h

#include "FloatSize.h"
#include "Glyph.h"
#include <wtf/PassOwnPtr.h>
#include <wtf/UnusedParam.h>
#include <wtf/Vector.h>

#if USE(CG)
#include <CoreGraphics/CGGeometry.h>
#endif

#if OS(DARWIN) && (PLATFORM(WX) || PLATFORM(CHROMIUM))
#include <ApplicationServices/ApplicationServices.h>
#endif

#if USE(CAIRO) || (PLATFORM(WX) && defined(wxUSE_CAIRO) && wxUSE_CAIRO)
#include <cairo.h>
#endif

namespace WebCore {

class SimpleFontData;

#if USE(CAIRO)
typedef Glyph GlyphBufferGlyph;
#elif (PLATFORM(WX) && defined(wxUSE_CAIRO) && wxUSE_CAIRO)
// FIXME: Why does Cairo use such a huge struct instead of just an offset into an array?
typedef cairo_glyph_t GlyphBufferGlyph;
#elif OS(WINCE)
typedef wchar_t GlyphBufferGlyph;
#else
typedef Glyph GlyphBufferGlyph;
#endif

// CG uses CGSize instead of FloatSize so that the result of advances()
// can be passed directly to CGContextShowGlyphsWithAdvances in FontMac.mm
#if USE(CG) || (OS(DARWIN) && (PLATFORM(WX) || PLATFORM(CHROMIUM)))
struct GlyphBufferAdvance : CGSize {
public:
    GlyphBufferAdvance() : CGSize(CGSizeZero) { }
    GlyphBufferAdvance(CGSize size) : CGSize(size)
    {
    }

    void setWidth(CGFloat width) { this->CGSize::width = width; }
    CGFloat width() const { return this->CGSize::width; }
    CGFloat height() const { return this->CGSize::height; }
};
#else
typedef float GlyphBufferAdvance;
#endif


class GlyphBuffer {	
public:
	ALWAYS_INLINE bool isEmpty() const { return m_fontData.isEmpty(); }
	ALWAYS_INLINE int size() const { return m_fontData.size(); }
     
    void clear()
    {    	
        m_fontData.clear();
        m_glyphs.clear();
        m_advances.clear();
        if (m_offsetsInString)
            m_offsetsInString->clear();
#if PLATFORM(WIN)
        m_offsets.clear();
#endif
    }


    ALWAYS_INLINE GlyphBufferGlyph* glyphs(int from) { return m_glyphs.data() + from; }
    ALWAYS_INLINE GlyphBufferAdvance* advances(int from) { return m_advances.data() + from; }
    ALWAYS_INLINE const GlyphBufferGlyph* glyphs(int from) const { return m_glyphs.data() + from; }
    ALWAYS_INLINE const GlyphBufferAdvance* advances(int from) const { return m_advances.data() + from; }

    ALWAYS_INLINE const SimpleFontData* fontDataAt(int index) const { return m_fontData[index]; }

    //void setInitialAdvance(GlyphBufferAdvance initialAdvance) { m_initialAdvance = initialAdvance; }
    //const GlyphBufferAdvance& initialAdvance() const { return m_initialAdvance; }
    
    void swap(int index1, int index2)
    {
        const SimpleFontData* f = m_fontData[index1];
        m_fontData[index1] = m_fontData[index2];
        m_fontData[index2] = f;

        GlyphBufferGlyph g = m_glyphs[index1];
        m_glyphs[index1] = m_glyphs[index2];
        m_glyphs[index2] = g;

        GlyphBufferAdvance s = m_advances[index1];
        m_advances[index1] = m_advances[index2];
        m_advances[index2] = s;

#if PLATFORM(WIN)
        FloatSize offset = m_offsets[index1];
        m_offsets[index1] = m_offsets[index2];
        m_offsets[index2] = offset;
#endif
    }

    ALWAYS_INLINE Glyph glyphAt(int index) const
    {
#if (PLATFORM(WX) && defined(wxUSE_CAIRO) && wxUSE_CAIRO)
        return m_glyphs[index].index;
#else
        return m_glyphs[index];
#endif
    }

    ALWAYS_INLINE GlyphBufferAdvance advanceAt(int index) const
    {
        return m_advances[index];
    }

    ALWAYS_INLINE FloatSize offsetAt(int index) const
    {
#if PLATFORM(WIN)
        return m_offsets[index];
#else
        UNUSED_PARAM(index);
        return FloatSize();
#endif
    }

    static const int kNoOffset = -1;
    void add(Glyph glyph, const SimpleFontData* font, float width, int offsetInString = kNoOffset, const FloatSize* offset = 0)
    {
        m_fontData.append(font);

#if (PLATFORM(WX) && defined(wxUSE_CAIRO) && wxUSE_CAIRO)
        cairo_glyph_t cairoGlyph;
        cairoGlyph.index = glyph;
        m_glyphs.append(cairoGlyph);
#else
        m_glyphs.append(glyph);
#endif

#if USE(CG) || (OS(DARWIN) && (PLATFORM(WX) || PLATFORM(CHROMIUM)))
        CGSize advance = { width, 0 };
        m_advances.append(advance);
#else
        m_advances.append(width);
#endif

#if PLATFORM(WIN)
        if (offset)
            m_offsets.append(*offset);
        else
            m_offsets.append(FloatSize());
#else
        UNUSED_PARAM(offset);
#endif

        if (offsetInString != kNoOffset && m_offsetsInString)
            m_offsetsInString->append(offsetInString);
    }

#if USE(CG) || (OS(DARWIN) && (PLATFORM(WX) || PLATFORM(CHROMIUM)))
    void add(Glyph glyph, const SimpleFontData* font, GlyphBufferAdvance advance)
    {
        m_fontData.append(font);
#if (PLATFORM(WX) && defined(wxUSE_CAIRO) && wxUSE_CAIRO)
        cairo_glyph_t cairoGlyph;
        cairoGlyph.index = glyph;
        m_glyphs.append(cairoGlyph);
#else
        m_glyphs.append(glyph);
#endif

        m_advances.append(advance);
    }
#endif

    ALWAYS_INLINE void expandLastAdvance(float width)
    {
        ASSERT(!isEmpty());
        GlyphBufferAdvance& lastAdvance = m_advances.last();
        lastAdvance += width;
    }

    void shrink(int truncationPoint)
    {
    	m_fontData.shrink(truncationPoint);
        m_glyphs.shrink(truncationPoint);
        m_advances.shrink(truncationPoint);
        //if (m_offsetsInString)
        //    m_offsetsInString->shrink(truncationPoint);
#if PLATFORM(WIN)
        m_offsets.shrink(truncationPoint);
#endif
    }

    void saveOffsetsInString()
    {
        m_offsetsInString = adoptPtr(new Vector<int, 2048>());
    }

    int offsetInString(int index) const
    {
        ASSERT(m_offsetsInString);
        return (*m_offsetsInString)[index];
    }

private:	
	Vector<const SimpleFontData*, 128> m_fontData;
	Vector<GlyphBufferGlyph, 128> m_glyphs;
	Vector<GlyphBufferAdvance, 128> m_advances;
    //GlyphBufferAdvance m_initialAdvance;
    OwnPtr<Vector<int, 2048>> m_offsetsInString;
#if PLATFORM(WIN)
    Vector<FloatSize, 2048> m_offsets;
#endif
};

}
#endif
