/*
 * Copyright (C) 2003, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 * Copyright (C) 2008-2009 Torch Mobile, Inc.
 * Copyright (C) 2009, 2010, 2011 Espial Group Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef GraphicsContext_h
#define GraphicsContext_h

#include "DashArray.h"
#include "FloatRect.h"
#include "Gradient.h"
#include "Image.h"
#include "ImageOrientation.h"
#include "IntRect.h"
#include "Path.h"
#include "Pattern.h"
#include "TextDirection.h"
#include <wtf/Noncopyable.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/AlwaysInline.h>
#include <wtf/text/AtomicString.h>

#if USE(CG)
typedef struct CGContext PlatformGraphicsContext;
#elif USE(CAIRO)
namespace WebCore {
class PlatformContextCairo;
}
typedef WebCore::PlatformContextCairo PlatformGraphicsContext;
#elif USE(OPENVG)
namespace WebCore {
class SurfaceOpenVG;
}
typedef class WebCore::SurfaceOpenVG PlatformGraphicsContext;
#elif PLATFORM(QT)
#include <QPainter>
namespace WebCore {
class ShadowBlur;
}
typedef QPainter PlatformGraphicsContext;
#elif PLATFORM(WX)
class wxGCDC;
class wxWindowDC;

// wxGraphicsContext allows us to support Path, etc.
// but on some platforms, e.g. Linux, it requires fairly
// new software.
#if USE(WXGC)
// On OS X, wxGCDC is just a typedef for wxDC, so use wxDC explicitly to make
// the linker happy.
#ifdef __APPLE__
    class wxDC;
    typedef wxDC PlatformGraphicsContext;
#else
    typedef wxGCDC PlatformGraphicsContext;
#endif
#else
    typedef wxWindowDC PlatformGraphicsContext;
#endif
#elif USE(SKIA)
namespace WebCore {
class PlatformContextSkia;
typedef PlatformContextSkia GraphicsContextPlatformPrivate;
}
typedef WebCore::PlatformContextSkia PlatformGraphicsContext;
#elif OS(WINCE)
typedef struct HDC__ PlatformGraphicsContext;
#else
typedef void PlatformGraphicsContext;
#endif

#if PLATFORM(WIN)
#include "DIBPixelData.h"
typedef struct HDC__* HDC;
#if !USE(CG)
// UInt8 is defined in CoreFoundation/CFBase.h
typedef unsigned char UInt8;
#endif
#endif

#if PLATFORM(QT) && OS(WINDOWS)
#include <windows.h>
#endif

namespace WebCore {

#if OS(WINCE) && !PLATFORM(QT)
class SharedBitmap;
class SimpleFontData;
class GlyphBuffer;
#endif

const int cMisspellingLineThickness = 3;
const int cMisspellingLinePatternWidth = 4;
const int cMisspellingLinePatternGapWidth = 1;

class AffineTransform;
class DrawingBuffer;
class Font;
class FloatRoundedRect;
#if !USE(SKIA)
class GraphicsContextPlatformPrivate;
#endif
class GlyphBuffer;
class Gradient;
class ImageBuffer;
class IntRect;
class RoundedRect;
class KURL;
class GraphicsContext3D;
#if USE(ACCELERATED_COMPOSITING)
class GraphicsLayer;
#endif
class SimpleFontData;
class TextRun;
class TransformationMatrix;

enum TextDrawingMode {
	TextModeInvisible = 0,
	TextModeFill      = 1 << 0,
	TextModeStroke    = 1 << 1,
	TextModeClip      = 1 << 2
};
typedef unsigned TextDrawingModeFlags;

enum StrokeStyle {
	NoStroke,
	SolidStroke,
	DottedStroke,
	DashedStroke,
#if ENABLE(CSS3_TEXT)
	DoubleStroke,
	WavyStroke,
#endif // CSS3_TEXT
};

enum InterpolationQuality {
	InterpolationDefault,
	InterpolationNone,
	InterpolationLow,
	InterpolationMedium,
	InterpolationHigh
};

struct GraphicsContextState {
	WTF_MAKE_FAST_ALLOCATED;
public:
	GraphicsContextState()
		:strokeThickness(0)
		, shadowBlur(0)
#if USE(ACCELERATED_COMPOSITING)
		, graphicsLayer(0)
#endif
		, textDrawingMode(TextModeFill)
		, strokeColor(Color::black)
		, fillColor(Color::black)
		, strokeStyle(SolidStroke)
		, fillRule(RULE_NONZERO)
		, compositeOperator(CompositeSourceOver)
		, blendMode(BlendModeNormal)
		, shouldAntialias(true)
		, shouldSmoothFonts(true)
		, shouldSubpixelQuantizeFonts(true)
		, paintingDisabled(false)
		, shadowsIgnoreTransforms(false)
#if USE(CG)
		// Core Graphics incorrectly renders shadows with radius > 8px (<rdar://problem/8103442>),
		// but we need to preserve this buggy behavior for canvas and -webkit-box-shadow.
		, shadowsUseLegacyRadius(false)
#endif
	{
	}

	RefPtr<Gradient> strokeGradient;
	RefPtr<Pattern> strokePattern;

	RefPtr<Gradient> fillGradient;
	RefPtr<Pattern> fillPattern;

	FloatSize shadowOffset;

	float strokeThickness;
	float shadowBlur;

	FloatPoint position; // relative to top Frame

#if USE(ACCELERATED_COMPOSITING)
	GraphicsLayer* graphicsLayer;
#endif

	TextDrawingModeFlags textDrawingMode;

	Color strokeColor;
	Color fillColor;
	Color shadowColor;

	StrokeStyle strokeStyle;
	WindRule fillRule;

	CompositeOperator compositeOperator;
    BlendMode blendMode;

	bool shouldAntialias : 1;
	bool shouldSmoothFonts : 1;
    bool shouldSubpixelQuantizeFonts : 1;
	bool paintingDisabled : 1;
	bool shadowsIgnoreTransforms : 1;
#if USE(CG)
	bool shadowsUseLegacyRadius : 1;
#endif
};

class GraphicsContext {
	WTF_MAKE_NONCOPYABLE(GraphicsContext); WTF_MAKE_FAST_ALLOCATED;
public:
	GraphicsContext(PlatformGraphicsContext*);
	~GraphicsContext();

#if !OS(WINCE) || PLATFORM(QT)
	PlatformGraphicsContext* platformContext() const;
#endif

	ALWAYS_INLINE float strokeThickness() const { return m_state.strokeThickness; }
	void setStrokeThickness(float);
	StrokeStyle strokeStyle() const;
	void setStrokeStyle(StrokeStyle);
	Color strokeColor() const;
	void setStrokeColor(const Color&);

	void setStrokePattern(PassRefPtr<Pattern>);
	Pattern* strokePattern() const;

	void setStrokeGradient(PassRefPtr<Gradient>);
	Gradient* strokeGradient() const;

	WindRule fillRule() const;
	void setFillRule(WindRule);
	ALWAYS_INLINE Color fillColor() const { return m_state.fillColor; }
	void setFillColor(const Color&);

	void setFillPattern(PassRefPtr<Pattern>);
	Pattern* fillPattern() const;

	void setFillGradient(PassRefPtr<Gradient>);
	Gradient* fillGradient() const;

	void setShadowsIgnoreTransforms(bool);
	bool shadowsIgnoreTransforms() const;

	void setShouldAntialias(bool);
	bool shouldAntialias() const;

	void setShouldSmoothFonts(bool);
	bool shouldSmoothFonts() const;

    // Normally CG enables subpixel-quantization because it improves the performance of aligning glyphs.
    // In some cases we have to disable to to ensure a high-quality output of the glyphs.
    void setShouldSubpixelQuantizeFonts(bool);
    bool shouldSubpixelQuantizeFonts() const;

	ALWAYS_INLINE const GraphicsContextState& state() const { return m_state; }

	// relative to top Frame
	ALWAYS_INLINE const FloatPoint& position() const { return m_state.position; }
	ALWAYS_INLINE void setPosition(const FloatPoint& p) { m_state.position = p; }

#if USE(ACCELERATED_COMPOSITING)
	GraphicsLayer* graphicsLayer() const { return m_state.graphicsLayer; }
	void setGraphicsLayer(GraphicsLayer* graphicsLayer) { m_state.graphicsLayer = graphicsLayer; }
#endif

#if USE(CG)
	void applyStrokePattern();
	void applyFillPattern();
	void drawPath(const Path&);

	void drawNativeImage(NativeImagePtr, const FloatSize& selfSize, const FloatRect& destRect, const FloatRect& srcRect, CompositeOperator = CompositeSourceOver, BlendMode = BlendModeNormal, ImageOrientation = DefaultImageOrientation);

	// Allow font smoothing (LCD antialiasing). Not part of the graphics state.
	void setAllowsFontSmoothing(bool);

	void setIsCALayerContext(bool);
	bool isCALayerContext() const;

	void setIsAcceleratedContext(bool);
#endif

#if USE(CAIRO)
	void setIsAcceleratedContext(bool);
#endif

	bool isAcceleratedContext() const;

	void save();
	void restore();

	// These draw methods will do both stroking and filling.
	// FIXME: ...except drawRect(), which fills properly but always strokes
	// using a 1-pixel stroke inset from the rect borders (of the correct
	// stroke color).
    void drawRect(const FloatRect&, float borderThickness = 1);
	void drawLine(const FloatPoint&, const FloatPoint&);
	void drawLine(const IntPoint& point1, const IntPoint& point2) { drawLine(FloatPoint(point1), FloatPoint(point2)); }
	void drawEllipse(const IntRect&);
	void drawConvexPolygon(size_t numPoints, const FloatPoint*, bool shouldAntialias = false);

	void fillPath(const Path&);
	void strokePath(const Path&);

	void fillEllipse(const FloatRect&);
	void strokeEllipse(const FloatRect&);

	// Arc drawing (used by border-radius in CSS) just supports stroking at the moment.
	void strokeArc(const IntRect&, int startAngle, int angleSpan);

	void fillRect(const FloatRect&);
	void fillRect(const FloatRect&, const Color&);
	void fillRect(const FloatRect&, Gradient&);
	void fillRect(const FloatRect&, const Color&, CompositeOperator, BlendMode = BlendModeNormal);
	void fillRoundedRect(const FloatRoundedRect&, const Color&, BlendMode = BlendModeNormal);
	void fillRectWithRoundedHole(const FloatRect&, const FloatRoundedRect& roundedHoleRect, const Color&);

	void clearRect(const FloatRect&);

	void strokeRect(const FloatRect&, float lineWidth);

	void drawImage(Image*, const FloatPoint&, CompositeOperator = CompositeSourceOver, RespectImageOrientationEnum = DoNotRespectImageOrientation);
	void drawImage(Image*, const FloatRect&, CompositeOperator = CompositeSourceOver, RespectImageOrientationEnum = DoNotRespectImageOrientation, bool useLowQualityScale = false);
	void drawImage(Image*, const FloatPoint& destPoint, const IntRect& srcRect, CompositeOperator = CompositeSourceOver, RespectImageOrientationEnum = DoNotRespectImageOrientation);
	void drawImage(Image*, const FloatRect& destRect, const FloatRect& srcRect, CompositeOperator = CompositeSourceOver, BlendMode = BlendModeNormal, RespectImageOrientationEnum = DoNotRespectImageOrientation, bool useLowQualityScale = false);

	void drawTiledImage(Image*, const FloatRect& destRect, const FloatPoint& srcPoint, const FloatSize& tileSize, const FloatSize& spacing, CompositeOperator = CompositeSourceOver, bool useLowQualityScale = false, BlendMode = BlendModeNormal);
	void drawTiledImage(Image*, const FloatRect& destRect, const FloatRect& srcRect,
						const FloatSize& tileScaleFactor, Image::TileRule hRule = Image::StretchTile, Image::TileRule vRule = Image::StretchTile,
						CompositeOperator = CompositeSourceOver, bool useLowQualityScale = false);

	void drawImageBuffer(ImageBuffer*, const FloatPoint&, CompositeOperator = CompositeSourceOver, BlendMode = BlendModeNormal);
	void drawImageBuffer(ImageBuffer*, const FloatRect&, CompositeOperator = CompositeSourceOver, BlendMode = BlendModeNormal, bool useLowQualityScale = false);
	void drawImageBuffer(ImageBuffer*, const FloatPoint& destPoint, const FloatRect& srcRect, CompositeOperator = CompositeSourceOver, BlendMode = BlendModeNormal);
	void drawImageBuffer(ImageBuffer*, const FloatRect& destRect, const FloatRect& srcRect, CompositeOperator = CompositeSourceOver, BlendMode = BlendModeNormal, bool useLowQualityScale = false);

	void setImageInterpolationQuality(InterpolationQuality);
	InterpolationQuality imageInterpolationQuality() const;

	void clip(const IntRect&);
	void clip(const FloatRect&);
    void clipRoundedRect(const FloatRoundedRect&);

	void clipOut(const FloatRect&);
	void clipOutRoundedRect(const FloatRoundedRect&);
	void clipPath(const Path&, WindRule);
	void clipConvexPolygon(size_t numPoints, const FloatPoint*, bool antialias = true);
	void clipToImageBuffer(ImageBuffer*, const FloatRect&);

	IntRect clipBounds() const;

	ALWAYS_INLINE TextDrawingModeFlags textDrawingMode() const { return m_state.textDrawingMode; }
	void setTextDrawingMode(TextDrawingModeFlags);

	void drawText(const Font&, const TextRun&, const FloatPoint&, int from = 0, int to = -1);
	void drawGlyphs(const Font&, const SimpleFontData&, const GlyphBuffer&, int from, int numGlyphs, const FloatPoint&);
	void drawEmphasisMarks(const Font&, const TextRun& , const AtomicString& mark, const FloatPoint&, int from = 0, int to = -1);
	void drawBidiText(const Font&, const TextRun&, const FloatPoint&);

	enum RoundingMode {
		RoundAllSides,
		RoundOriginAndDimensions
	};
	FloatRect roundToDevicePixels(const FloatRect&, RoundingMode = RoundAllSides);

	FloatRect computeLineBoundsForText(const FloatPoint&, float width, bool printing);
	void drawLineForText(const FloatPoint&, float width, bool printing, bool doubleLines = false, StrokeStyle = SolidStroke);
    void drawLinesForText(const FloatPoint&, const DashArray& widths, bool printing, bool doubleLines = false, StrokeStyle = SolidStroke);
	enum DocumentMarkerLineStyle {
		DocumentMarkerSpellingLineStyle,
		DocumentMarkerGrammarLineStyle,
		DocumentMarkerAutocorrectionReplacementLineStyle,
		DocumentMarkerDictationAlternativesLineStyle
	};
	void drawLineForDocumentMarker(const FloatPoint&, float width, DocumentMarkerLineStyle);

	bool paintingDisabled() const;
	void setPaintingDisabled(bool);

	bool updatingControlTints() const;
	void setUpdatingControlTints(bool);

	void beginTransparencyLayer(float opacity);
	void endTransparencyLayer();
	bool isInTransparencyLayer() const;

	bool hasShadow() const;
	void setShadow(const FloatSize&, float blur, const Color&);
	// Legacy shadow blur radius is used for canvas, and -webkit-box-shadow.
	// It has different treatment of radii > 8px.
	void setLegacyShadow(const FloatSize&, float blur, const Color&);

	bool getShadow(FloatSize&, float&, Color&) const;
	void clearShadow();

    bool hasBlurredShadow() const;

	void drawFocusRing(const Vector<IntRect>&, int width, int offset, const Color&);
	void drawFocusRing(const Path&, int width, int offset, const Color&);

	void setLineCap(LineCap);
	void setLineDash(const DashArray&, float dashOffset);
	void setLineJoin(LineJoin);
	void setMiterLimit(float);

	void setAlpha(float);

	void setCompositeOperation(CompositeOperator, BlendMode = BlendModeNormal);
	ALWAYS_INLINE CompositeOperator compositeOperation() const { return m_state.compositeOperator; }
    BlendMode blendModeOperation() const { return m_state.blendMode; }

	void clip(const Path&, WindRule = RULE_EVENODD);

	// This clip function is used only by <canvas> code. It allows
	// implementations to handle clipping on the canvas differently since
	// the discipline is different.
	void canvasClip(const Path&, WindRule = RULE_EVENODD);
	void clipOut(const Path&);

	void scale(const FloatSize&);
	void rotate(float angleInRadians);
	ALWAYS_INLINE void translate(const FloatSize& size) { translate(size.width(), size.height()); }
	void translate(float x, float y);

	void setURLForRect(const KURL&, const IntRect&);

	void concatCTM(const AffineTransform&);
	void setCTM(const AffineTransform&);

	enum IncludeDeviceScale { DefinitelyIncludeDeviceScale, PossiblyIncludeDeviceScale };
	AffineTransform getCTM(IncludeDeviceScale includeScale = PossiblyIncludeDeviceScale) const;

    float pixelSnappingFactor() const { return m_pixelSnappingFactor; }

#if ENABLE(3D_RENDERING) && USE(TEXTURE_MAPPER)
	// This is needed when using accelerated-compositing in software mode, like in TextureMapper.
	void concat3DTransform(const TransformationMatrix&);
	void set3DTransform(const TransformationMatrix&);
	TransformationMatrix get3DTransform() const;
#endif

	// Create an image buffer compatible with this context, with suitable resolution
	// for drawing into the buffer and then into this context.
	PassOwnPtr<ImageBuffer> createCompatibleBuffer(const FloatSize&) const;
	bool isCompatibleWithBuffer(ImageBuffer*) const;

	// This function applies the device scale factor to the context, making the context capable of
	// acting as a base-level context for a HiDPI environment.
	void applyDeviceScaleFactor(float);
	void platformApplyDeviceScaleFactor(float);

#if OS(WINCE) && !PLATFORM(QT)
	void setBitmap(PassRefPtr<SharedBitmap>);
	const AffineTransform& affineTransform() const;
	AffineTransform& affineTransform();
	void resetAffineTransform();
	void fillRect(const FloatRect&, const Gradient*);
	void drawText(const SimpleFontData* fontData, const GlyphBuffer& glyphBuffer, int from, int numGlyphs, const FloatPoint& point);
	void drawFrameControl(const IntRect& rect, unsigned type, unsigned state);
	void drawFocusRect(const IntRect& rect);
	void paintTextField(const IntRect& rect, unsigned state);
	void drawBitmap(SharedBitmap*, const IntRect& dstRect, const IntRect& srcRect, CompositeOperator compositeOp);
	void drawBitmapPattern(SharedBitmap*, const FloatRect& tileRectIn, const AffineTransform& patternTransform, const FloatPoint& phase, CompositeOperator op, const FloatRect& destRect, const IntSize& origSourceSize);
	void drawIcon(HICON icon, const IntRect& dstRect, UINT flags);
	HDC getWindowsContext(const IntRect&, bool supportAlphaBlend = false, bool mayCreateBitmap = true); // The passed in rect is used to create a bitmap for compositing inside transparency layers.
	void releaseWindowsContext(HDC, const IntRect&, bool supportAlphaBlend = false, bool mayCreateBitmap = true);    // The passed in HDC should be the one handed back by getWindowsContext.
	void drawRoundCorner(bool newClip, RECT clipRect, RECT rectWin, HDC dc, int width, int height);
#elif PLATFORM(WIN)
	GraphicsContext(HDC, bool hasAlpha = false); // FIXME: To be removed.
	HDC getWindowsContext(const IntRect&, bool supportAlphaBlend = true, bool mayCreateBitmap = true); // The passed in rect is used to create a bitmap for compositing inside transparency layers.
	void releaseWindowsContext(HDC, const IntRect&, bool supportAlphaBlend = true, bool mayCreateBitmap = true);    // The passed in HDC should be the one handed back by getWindowsContext.

	// When set to true, child windows should be rendered into this context
	// rather than allowing them just to render to the screen. Defaults to
	// false.
	// FIXME: This is a layering violation. GraphicsContext shouldn't know
	// what a "window" is. It would be much more appropriate for this flag
	// to be passed as a parameter alongside the GraphicsContext, but doing
	// that would require lots of changes in cross-platform code that we
	// aren't sure we want to make.
	void setShouldIncludeChildWindows(bool);
	bool shouldIncludeChildWindows() const;

	class WindowsBitmap {
		WTF_MAKE_NONCOPYABLE(WindowsBitmap);
	public:
		WindowsBitmap(HDC, const IntSize&);
		~WindowsBitmap();

		HDC hdc() const { return m_hdc; }
		UInt8* buffer() const { return m_pixelData.buffer(); }
		unsigned bufferLength() const { return m_pixelData.bufferLength(); }
		const IntSize& size() const { return m_pixelData.size(); }
		unsigned bytesPerRow() const { return m_pixelData.bytesPerRow(); }
		unsigned short bitsPerPixel() const { return m_pixelData.bitsPerPixel(); }
		const DIBPixelData& windowsDIB() const { return m_pixelData; }

	private:
		HDC m_hdc;
		HBITMAP m_bitmap;
		DIBPixelData m_pixelData;
	};

	PassOwnPtr<WindowsBitmap> createWindowsBitmap(const IntSize&);
	// The bitmap should be non-premultiplied.
	void drawWindowsBitmap(WindowsBitmap*, const IntPoint&);
#endif

#if (PLATFORM(GTK) && OS(WINDOWS)) || (PLATFORM(QT) && OS(WINDOWS)) || (PLATFORM(WX) && OS(WINDOWS))
	HDC getWindowsContext(const IntRect&, bool supportAlphaBlend = true, bool mayCreateBitmap = true);
	void releaseWindowsContext(HDC, const IntRect&, bool supportAlphaBlend = true, bool mayCreateBitmap = true);
	bool shouldIncludeChildWindows() const { return false; }
#endif

#if PLATFORM(WX)
	// This is needed because of a bug whereby getting an HDC from a GDI+ context
	// loses the scale operations applied to the context.
	FloatSize currentScale();
#endif

#if PLATFORM(QT)
	void pushTransparencyLayerInternal(const QRect&, qreal, QImage&);
	void takeOwnershipOfPlatformContext();
#endif

#if USE(CAIRO)
	GraphicsContext(cairo_surface_t*);
	GraphicsContext(cairo_t*);
#endif

#if 0 && PLATFORM(GTK)
	void setGdkExposeEvent(GdkEventExpose*);
	GdkWindow* gdkWindow() const;
	GdkEventExpose* gdkExposeEvent() const;
#endif

	static void adjustLineToPixelBoundaries(FloatPoint& p1, FloatPoint& p2, float strokeWidth, StrokeStyle);

private:
	void platformInit(PlatformGraphicsContext*);
	void platformDestroy();

#if PLATFORM(WIN) && !OS(WINCE)
	void platformInit(HDC, bool hasAlpha = false);
#endif

	void savePlatformState();
	void restorePlatformState();

	void setPlatformTextDrawingMode(TextDrawingModeFlags);

	void setPlatformStrokeColor(const Color&);
	void setPlatformStrokeGradient(Gradient* gradient);
	void setPlatformStrokePattern(Pattern* p);
	void setPlatformStrokeStyle(StrokeStyle);
	void setPlatformStrokeThickness(float);

	void setPlatformFillColor(const Color&);
	void setPlatformFillGradient(Gradient* gradient);
	void setPlatformFillPattern(Pattern* p);

	void setPlatformShouldAntialias(bool);
	void setPlatformShouldSmoothFonts(bool);

	void setPlatformShadow(const FloatSize&, float blur, const Color&);
	void clearPlatformShadow();

	void setPlatformCompositeOperation(CompositeOperator, BlendMode = BlendModeNormal);

	void beginPlatformTransparencyLayer(float opacity);
	void endPlatformTransparencyLayer();
	static bool supportsTransparencyLayers();

	void fillEllipseAsPath(const FloatRect&);
	void strokeEllipseAsPath(const FloatRect&);

	void platformFillEllipse(const FloatRect&);
	void platformStrokeEllipse(const FloatRect&);

    void platformFillRoundedRect(const FloatRoundedRect&, const Color&);

    FloatRect computeLineBoundsAndAntialiasingModeForText(const FloatPoint&, float width, bool printing, Color&);

	GraphicsContextPlatformPrivate* m_data;

	GraphicsContextState m_state;
	Vector<GraphicsContextState, 1> m_stack;
	bool m_updatingControlTints;
	unsigned m_transparencyCount;
    float m_pixelSnappingFactor;
};

class GraphicsContextStateSaver {
    WTF_MAKE_FAST_ALLOCATED;
public:
	ALWAYS_INLINE GraphicsContextStateSaver(GraphicsContext& context, bool saveAndRestore = true)
	: m_context(context)
	, m_saveAndRestore(saveAndRestore)
	{
		if (m_saveAndRestore)
			m_context.save();
	}

	ALWAYS_INLINE ~GraphicsContextStateSaver()
	{
		if (m_saveAndRestore)
			m_context.restore();
	}

	ALWAYS_INLINE void save()
	{
		ASSERT(!m_saveAndRestore);
		m_context.save();
		m_saveAndRestore = true;
	}

	ALWAYS_INLINE void restore()
	{
		ASSERT(m_saveAndRestore);
		m_context.restore();
		m_saveAndRestore = false;
	}

	GraphicsContext* context() const { return &m_context; }

private:
	GraphicsContext& m_context;
	bool m_saveAndRestore;
};

ALWAYS_INLINE Gradient* GraphicsContext::fillGradient() const
{
    return m_state.fillGradient.get();
}

ALWAYS_INLINE Gradient* GraphicsContext::strokeGradient() const
{
    return m_state.strokeGradient.get();
}

ALWAYS_INLINE Pattern* GraphicsContext::fillPattern() const
{
    return m_state.fillPattern.get();
}

ALWAYS_INLINE Pattern* GraphicsContext::strokePattern() const
{
    return m_state.strokePattern.get();
}

ALWAYS_INLINE void GraphicsContext::setShadowsIgnoreTransforms(bool ignoreTransforms)
{
    m_state.shadowsIgnoreTransforms = ignoreTransforms;
}

ALWAYS_INLINE bool GraphicsContext::shadowsIgnoreTransforms() const
{
    return m_state.shadowsIgnoreTransforms;
}

ALWAYS_INLINE bool GraphicsContext::paintingDisabled() const
{
	return m_state.paintingDisabled;
}

ALWAYS_INLINE void GraphicsContext::setPaintingDisabled(bool f)
{
	m_state.paintingDisabled = f;
}

ALWAYS_INLINE bool GraphicsContext::shouldAntialias() const
{
    return m_state.shouldAntialias;
}

ALWAYS_INLINE bool GraphicsContext::shouldSmoothFonts() const
{
    return m_state.shouldSmoothFonts;
}

ALWAYS_INLINE StrokeStyle GraphicsContext::strokeStyle() const
{
    return m_state.strokeStyle;
}

ALWAYS_INLINE Color GraphicsContext::strokeColor() const
{
    return m_state.strokeColor;
}

ALWAYS_INLINE WindRule GraphicsContext::fillRule() const
{
    return m_state.fillRule;
}

ALWAYS_INLINE void GraphicsContext::setFillRule(WindRule fillRule)
{
    m_state.fillRule = fillRule;
}

#if !PLATFORM(QT)
ALWAYS_INLINE bool GraphicsContext::isInTransparencyLayer() const
{
    return (m_transparencyCount > 0) && supportsTransparencyLayers();
}
#endif

ALWAYS_INLINE bool GraphicsContext::updatingControlTints() const
{
    return m_updatingControlTints;
}

ALWAYS_INLINE void GraphicsContext::setUpdatingControlTints(bool b)
{
    setPaintingDisabled(b);
    m_updatingControlTints = b;
}

#if !PLATFORM(QT)
ALWAYS_INLINE void GraphicsContext::clip(const IntRect& rect)
{
    clip(FloatRect(rect));
}
#endif

} // namespace WebCore

#if USE(CAIRO)
#include "GraphicsContextPlatformPrivateCairo.h"

namespace WebCore
{

ALWAYS_INLINE PlatformGraphicsContext* GraphicsContext::platformContext() const
{
	return m_data->platformContext;
}

// Cairo contexts can't hold separate fill and stroke colors
// so we set them just before we actually fill or stroke
ALWAYS_INLINE void GraphicsContext::setPlatformStrokeGradient(Gradient* gradient) {}
ALWAYS_INLINE void GraphicsContext::setPlatformStrokePattern(Pattern* p) {}
ALWAYS_INLINE void GraphicsContext::setPlatformFillGradient(Gradient* gradient) {}
ALWAYS_INLINE void GraphicsContext::setPlatformFillPattern(Pattern* p) {}
}

#endif

#endif // GraphicsContext_h

