/*
 * Copyright (c) 2012, Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef LayoutRect_h
#define LayoutRect_h

#include "LayoutBoxExtent.h"
#include "LayoutPoint.h"
#include "IntRect.h"
#include <wtf/Vector.h>

namespace WebCore {

class FloatRect;

class LayoutRect {
public:
	ALWAYS_INLINE LayoutRect() { }
    ALWAYS_INLINE LayoutRect(const LayoutPoint& location, const LayoutSize& size)
        : m_location(location), m_size(size) { }
    ALWAYS_INLINE LayoutRect(const LayoutUnit& x, const LayoutUnit& y, const LayoutUnit& width, const LayoutUnit& height)
        : m_location(LayoutPoint(x, y)), m_size(LayoutSize(width, height)) { }
    ALWAYS_INLINE LayoutRect(const FloatPoint& location, const FloatSize& size)
        : m_location(location), m_size(size) { }
    ALWAYS_INLINE LayoutRect(const IntRect& rect) : m_location(rect.location()), m_size(rect.size()) { }

    explicit LayoutRect(const FloatRect&); // don't do this implicitly since it's lossy

    ALWAYS_INLINE LayoutRect(const LayoutRect& other) : m_location(other.m_location), m_size(other.m_size) { }
    ALWAYS_INLINE LayoutRect& operator=(const LayoutRect& other) { m_location = other.m_location; m_size = other.m_size; return *this; }

    ALWAYS_INLINE ~LayoutRect() {}

    const LayoutPoint& location() const { return m_location; }
    const LayoutSize& size() const { return m_size; }

    IntPoint pixelSnappedLocation() const { return roundedIntPoint(m_location); }
    IntSize pixelSnappedSize() const { return snappedIntSize(m_size, m_location); }

    ALWAYS_INLINE void setLocation(const LayoutPoint& location) { m_location = location; }
    ALWAYS_INLINE void setSize(const LayoutSize& size) { m_size = size; }

    ALWAYS_INLINE const LayoutUnit& x() const { return m_location.x(); }
    ALWAYS_INLINE const LayoutUnit& y() const { return m_location.y(); }
    ALWAYS_INLINE LayoutUnit maxX() const { return x() + width(); }
    ALWAYS_INLINE LayoutUnit maxY() const { return y() + height(); }
    ALWAYS_INLINE const LayoutUnit& width() const { return m_size.width(); }
    ALWAYS_INLINE const LayoutUnit& height() const { return m_size.height(); }

    ALWAYS_INLINE int pixelSnappedX() const { return x().round(); }
    ALWAYS_INLINE int pixelSnappedY() const { return y().round(); }
    ALWAYS_INLINE int pixelSnappedWidth() const { return snappedIntSize(m_size, m_location).width(); }
    ALWAYS_INLINE int pixelSnappedHeight() const { return snappedIntSize(m_size, m_location).height(); }
    ALWAYS_INLINE int pixelSnappedMaxX() const { return (m_location.x() + m_size.width()).round(); }
    ALWAYS_INLINE int pixelSnappedMaxY() const { return (m_location.y() + m_size.height()).round(); }

    ALWAYS_INLINE void setX(const LayoutUnit& x) { m_location.setX(x); }
    ALWAYS_INLINE void setY(const LayoutUnit& y) { m_location.setY(y); }
    ALWAYS_INLINE void setWidth(const LayoutUnit& width) { m_size.setWidth(width); }
    ALWAYS_INLINE void setHeight(const LayoutUnit& height) { m_size.setHeight(height); }

    ALWAYS_INLINE bool isEmpty() const { return m_size.isEmpty(); }

    // NOTE: The result is rounded to integer values, and thus may be not the exact
    // center point.
    LayoutPoint center() const { return LayoutPoint(x() + width() / 2, y() + height() / 2); }

    ALWAYS_INLINE void move(const LayoutSize& size) { m_location += size; }
    ALWAYS_INLINE void moveBy(const LayoutPoint& offset) { m_location.move(offset.x(), offset.y()); }
    ALWAYS_INLINE void move(const LayoutUnit& dx, const LayoutUnit& dy) { m_location.move(dx, dy); }

    ALWAYS_INLINE void expand(const LayoutSize& size) { m_size += size; }
    void expand(const LayoutBoxExtent& box)
    {
        m_location.move(-box.left(), -box.top());
        m_size.expand(box.left() + box.right(), box.top() + box.bottom());
    }
    ALWAYS_INLINE void expand(const LayoutUnit& dw, const LayoutUnit& dh) { m_size.expand(dw, dh); }
    ALWAYS_INLINE void contract(const LayoutSize& size) { m_size -= size; }
    ALWAYS_INLINE void contract(const LayoutUnit& dw, const LayoutUnit& dh) { m_size.expand(-dw, -dh); }

    void shiftXEdgeTo(const LayoutUnit& edge)
    {
        LayoutUnit delta = edge - x();
        setX(edge);
        setWidth(std::max<LayoutUnit>(0, width() - delta));
    }
    void shiftMaxXEdgeTo(const LayoutUnit& edge)
    {
        LayoutUnit delta = edge - maxX();
        setWidth(std::max<LayoutUnit>(0, width() + delta));
    }
    void shiftYEdgeTo(const LayoutUnit& edge)
    {
        LayoutUnit delta = edge - y();
        setY(edge);
        setHeight(std::max<LayoutUnit>(0, height() - delta));
    }
    void shiftMaxYEdgeTo(const LayoutUnit& edge)
    {
        LayoutUnit delta = edge - maxY();
        setHeight(std::max<LayoutUnit>(0, height() + delta));
    }

    LayoutPoint minXMinYCorner() const { return m_location; } // typically topLeft
    LayoutPoint maxXMinYCorner() const { return LayoutPoint(m_location.x() + m_size.width(), m_location.y()); } // typically topRight
    LayoutPoint minXMaxYCorner() const { return LayoutPoint(m_location.x(), m_location.y() + m_size.height()); } // typically bottomLeft
    LayoutPoint maxXMaxYCorner() const { return LayoutPoint(m_location.x() + m_size.width(), m_location.y() + m_size.height()); } // typically bottomRight
    
    bool intersects(const LayoutRect&) const;
    bool contains(const LayoutRect&) const;

    // This checks to see if the rect contains x,y in the traditional sense.
    // Equivalent to checking if the rect contains a 1x1 rect below and to the right of (px,py).
    bool contains(const LayoutUnit& px, const LayoutUnit& py) const
        { return px >= x() && px < maxX() && py >= y() && py < maxY(); }
    ALWAYS_INLINE bool contains(const LayoutPoint& point) const { return contains(point.x(), point.y()); }

    void intersect(const LayoutRect&);
    void unite(const LayoutRect&);
    void uniteIfNonZero(const LayoutRect&);

    ALWAYS_INLINE void inflateX(const LayoutUnit& dx)
    {
        m_location.setX(m_location.x() - dx);
        m_size.setWidth(m_size.width() + dx + dx);
    }
    ALWAYS_INLINE void inflateY(const LayoutUnit& dy)
    {
        m_location.setY(m_location.y() - dy);
        m_size.setHeight(m_size.height() + dy + dy);
    }
    ALWAYS_INLINE void inflate(const LayoutUnit& d) { inflateX(d); inflateY(d); }
    void scale(float s);
    void scale(float xScale, float yScale);

    LayoutRect transposedRect() const { return LayoutRect(m_location.transposedPoint(), m_size.transposedSize()); }
    bool isInfinite() const { return *this == LayoutRect::infiniteRect(); }

    static LayoutRect infiniteRect()
    {
        // Return a rect that is slightly smaller than the true max rect to allow pixelSnapping to round up to the nearest IntRect without overflowing.
        return LayoutRect(LayoutUnit::nearlyMin() / 2, LayoutUnit::nearlyMin() / 2, LayoutUnit::nearlyMax(), LayoutUnit::nearlyMax());
    }

    operator FloatRect() const { return FloatRect(m_location, m_size); }

private:
    LayoutPoint m_location;
    LayoutSize m_size;
};

ALWAYS_INLINE LayoutRect intersection(const LayoutRect& a, const LayoutRect& b)
{
    LayoutRect c = a;
    c.intersect(b);
    return c;
}

ALWAYS_INLINE LayoutRect unionRect(const LayoutRect& a, const LayoutRect& b)
{
    LayoutRect c = a;
    c.unite(b);
    return c;
}

LayoutRect unionRect(const Vector<LayoutRect>&);

ALWAYS_INLINE bool operator==(const LayoutRect& a, const LayoutRect& b)
{
    return a.location() == b.location() && a.size() == b.size();
}

ALWAYS_INLINE bool operator!=(const LayoutRect& a, const LayoutRect& b)
{
    return a.location() != b.location() || a.size() != b.size();
}

IntRect enclosingIntRect(const LayoutRect&);
LayoutRect enclosingLayoutRect(const FloatRect&);

inline IntRect snappedIntRect(const LayoutRect& rect)
{
    return IntRect(roundedIntPoint(rect.location()), snappedIntSize(rect.size(), rect.location()));
}

inline IntRect snappedIntRect(LayoutUnit left, LayoutUnit top, LayoutUnit width, LayoutUnit height)
{
    return IntRect(IntPoint(left.round(), top.round()), snappedIntSize(LayoutSize(width, height), LayoutPoint(left, top)));
}

inline IntRect snappedIntRectFromEdges(LayoutUnit left, LayoutUnit top, LayoutUnit right, LayoutUnit bottom)
{
    return IntRect(IntPoint(left.round(), top.round()), snappedIntSize(LayoutSize(right - left, bottom - top), LayoutPoint(left, top)));
}

inline IntRect snappedIntRect(LayoutPoint location, LayoutSize size)
{
    return IntRect(roundedIntPoint(location), snappedIntSize(size, location));
}

FloatRect encloseRectToDevicePixels(const LayoutRect&, float pixelSnappingFactor);

inline FloatRect snapRectToDevicePixels(const LayoutRect& rect, float pixelSnappingFactor)
{
    return FloatRect(FloatPoint(roundToDevicePixel(rect.x(), pixelSnappingFactor), roundToDevicePixel(rect.y(), pixelSnappingFactor)), snapSizeToDevicePixel(rect.size(), rect.location(), pixelSnappingFactor));
}

inline FloatRect snapRectToDevicePixels(LayoutUnit x, LayoutUnit y, LayoutUnit width, LayoutUnit height, float pixelSnappingFactor)
{
    return snapRectToDevicePixels(LayoutRect(x, y, width, height), pixelSnappingFactor);
}

// FIXME: This needs to take vertical centering into account too.
inline FloatRect snapRectToDevicePixelsWithWritingDirection(const LayoutRect& rect, float deviceScaleFactor, bool ltr)
{
    if (!ltr) {
        FloatPoint snappedTopRight = roundPointToDevicePixels(rect.maxXMinYCorner(), deviceScaleFactor, ltr);
        FloatSize snappedSize = snapSizeToDevicePixel(rect.size(), rect.maxXMinYCorner(), deviceScaleFactor);
        return FloatRect(snappedTopRight.x() - snappedSize.width(), snappedTopRight.y(), snappedSize.width(), snappedSize.height());
    }
    return snapRectToDevicePixels(rect, deviceScaleFactor);
}

} // namespace WebCore

#endif // LayoutRect_h
