/*
 * Copyright (C) 2005, 2008, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2006 Alexey Proskuryakov
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "SimpleFontData.h"

#include "Font.h"
#include "FontCache.h"
#include "OpenTypeMathData.h"
#include <wtf/MathExtras.h>
#include <wtf/NeverDestroyed.h>
#include <wtf/UnusedParam.h>

using namespace std;

namespace WebCore {

unsigned GlyphPage::s_count = 0;

DLL_PUBLIC int gSimpleFontDataCount;

DLL_PUBLIC SimpleFontData::SimpleFontData(const FontPlatformData& platformData, bool isCustomFont, bool isLoading, bool isTextOrientationFallback)
    : m_maxCharWidth(-1)
    , m_avgCharWidth(-1)
    , m_platformData(platformData)
    , m_treatAsFixedPitch(false)
    , m_isCustomFont(isCustomFont)
    , m_isLoading(isLoading)
    , m_isTextOrientationFallback(isTextOrientationFallback)
    , m_isBrokenIdeographFallback(false)
	, m_isUsedInSystemFallbackCache(false)
    , m_hasVerticalGlyphs(false)
	, m_spaceGlyph(0)
	, m_spaceWidth(0)
	, m_zeroGlyph(0)
	, m_adjustedSpaceWidth(0)
	, m_zeroWidthSpaceGlyph(0)
	, m_syntheticBoldOffset(0)
{
    platformInit();
    platformGlyphInit();
    platformCharWidthInit();
#if ENABLE(OPENTYPE_VERTICAL)
    if (platformData.orientation() == Vertical && !isTextOrientationFallback) {
        m_verticalData = platformData.verticalData();
        m_hasVerticalGlyphs = m_verticalData.get() && m_verticalData->hasVerticalMetrics();
    }
#endif

    ++gSimpleFontDataCount;
}

SimpleFontData::SimpleFontData(PassOwnPtr<AdditionalFontData> fontData, const FontDescription& fontDescription, bool syntheticBold, bool syntheticItalic)
    : m_maxCharWidth(-1)
	, m_avgCharWidth(-1)
	, m_platformData(FontPlatformData(fontDescription, syntheticBold, syntheticItalic))
    , m_fontData(fontData)
    , m_treatAsFixedPitch(false)
    , m_isCustomFont(true)
    , m_isLoading(false)
    , m_isTextOrientationFallback(false)
    , m_isBrokenIdeographFallback(false)
	, m_isUsedInSystemFallbackCache(false)
    , m_hasVerticalGlyphs(false)
	, m_spaceGlyph(0)
	, m_spaceWidth(0)
	, m_zeroGlyph(0)
	, m_adjustedSpaceWidth(0)
	, m_zeroWidthSpaceGlyph(0)
	, m_syntheticBoldOffset(0)
{
    m_fontData->initializeFontData(this, fontDescription.computedSize());

    ++gSimpleFontDataCount;
}

#if !(PLATFORM(QT) && !HAVE(QRAWFONT))
// Estimates of avgCharWidth and maxCharWidth for platforms that don't support accessing these values from the font.
DLL_PUBLIC void SimpleFontData::initCharWidths()
{
    auto* glyphPageZero = glyphPage(GlyphPage::pageNumberForCodePoint('0'));

    // Treat the width of a '0' as the avgCharWidth.
    if (m_avgCharWidth <= 0.f && glyphPageZero) {
        Glyph digitZeroGlyph = glyphPageZero->glyphDataForCharacter('0').glyph;
        if (digitZeroGlyph)
            m_avgCharWidth = widthForGlyph(digitZeroGlyph);
    }

    // If we can't retrieve the width of a '0', fall back to the x height.
    if (m_avgCharWidth <= 0.f)
        m_avgCharWidth = m_fontMetrics.xHeight();

    if (m_maxCharWidth <= 0.f)
        m_maxCharWidth = max(m_avgCharWidth, m_fontMetrics.floatAscent());
}

void SimpleFontData::platformGlyphInit()
{
    auto* glyphPageZero = glyphPage(0);
    auto* glyphPageCharacterZero = glyphPage(GlyphPage::pageNumberForCodePoint('0'));
    auto* glyphPageSpace = glyphPage(GlyphPage::pageNumberForCodePoint(space));

    // Ask for the glyph for 0 to avoid paging in ZERO WIDTH SPACE. Control characters, including 0,
    // are mapped to the ZERO WIDTH SPACE glyph.
    if (glyphPageZero)
        m_zeroWidthSpaceGlyph = glyphPageZero->glyphDataForCharacter(0).glyph;

    // Nasty hack to determine if we should round or ceil space widths.
    // If the font is monospace or fake monospace we ceil to ensure that 
    // every character and the space are the same width. Otherwise we round.
    if (glyphPageSpace)
        m_spaceGlyph = glyphPageSpace->glyphDataForCharacter(space).glyph;
    float width = widthForGlyph(m_spaceGlyph);
    m_spaceWidth = width;
    if (glyphPageCharacterZero)
        m_zeroGlyph = glyphPageCharacterZero->glyphDataForCharacter('0').glyph;
    m_fontMetrics.setZeroWidth(widthForGlyph(m_zeroGlyph));
    determinePitch();
    m_adjustedSpaceWidth = m_treatAsFixedPitch ? ceilf(width) : roundf(width);

    // Force the glyph for ZERO WIDTH SPACE to have zero width, unless it is shared with SPACE.
    // Helvetica is an example of a non-zero width ZERO WIDTH SPACE glyph.
    // See <http://bugs.webkit.org/show_bug.cgi?id=13178>
    // Ask for the glyph for 0 to avoid paging in ZERO WIDTH SPACE. Control characters, including 0,
    // are mapped to the ZERO WIDTH SPACE glyph.
    if (m_zeroWidthSpaceGlyph == m_spaceGlyph) {
        m_zeroWidthSpaceGlyph = 0;
        LOG_ERROR("Font maps SPACE and ZERO WIDTH SPACE to the same glyph. Glyph width will not be overridden.");
    }
}
#endif

DLL_PUBLIC SimpleFontData::~SimpleFontData()
{
    if (!m_fontData)
        platformDestroy();

    removeFromSystemFallbackCache();

    --gSimpleFontDataCount;
}

static bool fillGlyphPage(GlyphPage& pageToFill, UChar* buffer, unsigned bufferLength, const SimpleFontData& fontData)
{
#if ENABLE(SVG_FONTS)
    if (SimpleFontData::AdditionalFontData* additionalFontData = fontData.fontData())
        return additionalFontData->fillSVGGlyphPage(&pageToFill, buffer, bufferLength);
#endif
    bool hasGlyphs = pageToFill.fill(buffer, bufferLength, &fontData);
#if ENABLE(OPENTYPE_VERTICAL)
    if (hasGlyphs && fontData.verticalData())
        fontData.verticalData()->substituteWithVerticalGlyphs(&fontData, &pageToFill);
#endif
    return hasGlyphs;
}

static void overwriteCodePoints(UChar* buffer, unsigned start, unsigned end, unsigned minimum, unsigned maximum, UChar newCodePoint)
{
	unsigned begin = std::max(start, minimum);
	unsigned complete = std::min(end, maximum);
	for (unsigned i = begin; i < complete; ++i)
		buffer[i - start] = newCodePoint;
}

static void overwriteCodePoint(UChar* buffer, unsigned start, unsigned end, UChar codePoint, UChar newCodePoint)
{
	if (codePoint >= start && codePoint < end)
		buffer[codePoint - start] = newCodePoint;
}

static RefPtr<GlyphPage> createAndFillGlyphPage(unsigned pageNumber, const SimpleFontData& fontData)
{
#if PLATFORM(IOS)
    // FIXME: Times New Roman contains Arabic glyphs, but Core Text doesn't know how to shape them. See <rdar://problem/9823975>.
    // Once we have the fix for <rdar://problem/9823975> then remove this code together with SimpleFontData::shouldNotBeUsedForArabic()
    // in <rdar://problem/12096835>.
    if (GlyphPage::pageNumberIsUsedForArabic(pageNumber) && font.shouldNotBeUsedForArabic())
        return nullptr;
#endif

    unsigned glyphPageSize = GlyphPage::sizeForPageNumber(pageNumber);

    unsigned start = GlyphPage::startingCodePointInPageNumber(pageNumber);
    unsigned end = start + glyphPageSize;
    UChar buffer[glyphPageSize * 2 + 2];
    unsigned bufferLength;
    // Fill in a buffer with the entire "page" of characters that we want to look up glyphs for.
    if (U_IS_BMP(start)) {
        bufferLength = glyphPageSize;
        for (unsigned i = 0; i < bufferLength; i++)
            buffer[i] = start + i;

        // Code points 0x0 - 0x20 and 0x7F - 0xA0 are control character and shouldn't render. Map them to ZERO WIDTH SPACE.
        overwriteCodePoints(buffer, start, end, 0x0, 0x20, zeroWidthSpace);
        overwriteCodePoints(buffer, start, end, 0x7F, 0xA0, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, softHyphen, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, '\n', space);
        overwriteCodePoint(buffer, start, end, '\t', space);
        overwriteCodePoint(buffer, start, end, noBreakSpace, space);
        overwriteCodePoint(buffer, start, end, leftToRightMark, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, rightToLeftMark, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, leftToRightEmbed, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, rightToLeftEmbed, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, leftToRightOverride, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, rightToLeftOverride, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, leftToRightIsolate, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, rightToLeftIsolate, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, zeroWidthNonJoiner, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, zeroWidthJoiner, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, popDirectionalFormatting, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, popDirectionalIsolate, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, firstStrongIsolate, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, objectReplacementCharacter, zeroWidthSpace);
        overwriteCodePoint(buffer, start, end, zeroWidthNoBreakSpace, zeroWidthSpace);
    } else {
        bufferLength = glyphPageSize * 2;
        for (unsigned i = 0; i < glyphPageSize; i++) {
            int c = i + start;
            buffer[i * 2] = U16_LEAD(c);
            buffer[i * 2 + 1] = U16_TRAIL(c);
        }
    }

    // Now that we have a buffer full of characters, we want to get back an array
    // of glyph indices. This part involves calling into the platform-specific
    // routine of our glyph map for actually filling in the page with the glyphs.
    // Success is not guaranteed. For example, Times fails to fill page 260, giving glyph data
    // for only 128 out of 256 characters.
    RefPtr<GlyphPage> glyphPage = GlyphPage::create(fontData);

    bool haveGlyphs = fillGlyphPage(*glyphPage, buffer, bufferLength, fontData);
    if (!haveGlyphs)
        return nullptr;

    return glyphPage.releaseNonNull();
}

const GlyphPage* SimpleFontData::glyphPage(unsigned pageNumber) const
{
    if (pageNumber == 0) {
        if (!m_glyphPageZero)
            m_glyphPageZero = createAndFillGlyphPage(0, *this);
        return m_glyphPageZero.get();
    }
    auto addResult = m_glyphPages.add(pageNumber, nullptr);
    if (addResult.isNewEntry)
        addResult.iterator->second = createAndFillGlyphPage(pageNumber, *this);

    return addResult.iterator->second.get();
}

Glyph SimpleFontData::glyphForCharacter(UChar32 character) const
{
    auto* page = glyphPage(GlyphPage::pageNumberForCodePoint(character));
    if (!page)
        return 0;
    return page->glyphForCharacter(character);
}

GlyphData SimpleFontData::glyphDataForCharacter(UChar32 character) const
{
    auto* page = glyphPage(GlyphPage::pageNumberForCodePoint(character));
    if (!page)
        return GlyphData();
    return page->glyphDataForCharacter(character);
}

PassRefPtr<SimpleFontData> SimpleFontData::verticalRightOrientationFontData() const
{
    if (!m_derivedFontData)
        m_derivedFontData = DerivedFontData::create(isCustomFont());
    if (!m_derivedFontData->verticalRightOrientation) {
        FontPlatformData verticalRightPlatformData(m_platformData);
        verticalRightPlatformData.setOrientation(Horizontal);
        m_derivedFontData->verticalRightOrientation = create(verticalRightPlatformData, isCustomFont(), false, true);
    }
    return m_derivedFontData->verticalRightOrientation;
}

PassRefPtr<SimpleFontData> SimpleFontData::uprightOrientationFontData() const
{
    if (!m_derivedFontData)
        m_derivedFontData = DerivedFontData::create(isCustomFont());
    if (!m_derivedFontData->uprightOrientation)
        m_derivedFontData->uprightOrientation = create(m_platformData, isCustomFont(), false, true);
    return m_derivedFontData->uprightOrientation;
}

PassRefPtr<SimpleFontData> SimpleFontData::brokenIdeographFontData() const
{
    if (!m_derivedFontData)
        m_derivedFontData = DerivedFontData::create(isCustomFont());
    if (!m_derivedFontData->brokenIdeograph) {
        m_derivedFontData->brokenIdeograph = create(m_platformData, isCustomFont(), false);
        m_derivedFontData->brokenIdeograph->m_isBrokenIdeographFallback = true;
    }
    return m_derivedFontData->brokenIdeograph;
}

#ifndef NDEBUG
String SimpleFontData::description() const
{
    if (isSVGFont())
        return "[SVG font]";
    if (isCustomFont())
        return "[custom font]";

    return platformData().description();
}
#endif

const OpenTypeMathData* SimpleFontData::mathData() const
{
    if (m_isLoading)
        return 0;
    if (!m_mathData) {
        m_mathData = OpenTypeMathData::create(m_platformData);
        if (!m_mathData->hasMathData())
            m_mathData.clear();
    }
    return m_mathData.get();
}

DLL_PUBLIC PassOwnPtr<SimpleFontData::DerivedFontData> SimpleFontData::DerivedFontData::create(bool forCustomFont)
{
    return adoptPtr(new DerivedFontData(forCustomFont));
}

DLL_PUBLIC SimpleFontData::DerivedFontData::~DerivedFontData()
{
#if PLATFORM(COCOA)
    if (compositeFontReferences) {
        // FIXME: Why don't we use WebKit types here?
        CFDictionaryRef dictionary = CFDictionaryRef(compositeFontReferences.get());
        CFIndex count = CFDictionaryGetCount(dictionary);
        if (count > 0) {
            Vector<SimpleFontData*, 2> stash(count);
            SimpleFontData** fonts = stash.data();
            CFDictionaryGetKeysAndValues(dictionary, 0, (const void **)fonts);
            // This deletes the fonts.
            while (count-- > 0 && *fonts)
                adoptRef(*fonts++);
        }
    }
#endif
}

PassRefPtr<SimpleFontData> SimpleFontData::createScaledFontData(const FontDescription& fontDescription, float scaleFactor) const
{
    // FIXME: Support scaled fonts that used AdditionalFontData.
    if (m_fontData)
        return nullptr;

    return platformCreateScaledFontData(fontDescription, scaleFactor);
}

// FontDatas are not refcounted to avoid cycles.
typedef HashMap<std::pair<UChar32, unsigned>, SimpleFontData*> CharacterFallbackMap;
typedef HashMap<const SimpleFontData*, CharacterFallbackMap> SystemFallbackCache;

static SystemFallbackCache& systemFallbackCache()
{
    static NeverDestroyed<SystemFallbackCache> map;
    return map.get();
}

RefPtr<SimpleFontData> SimpleFontData::systemFallbackFontDataForCharacter(UChar32 character, const FontDescription& description, bool isForPlatformFont) const
{
	if(!character)
		return nullptr;

    auto fontAddResult = systemFallbackCache().add(this, CharacterFallbackMap());

    auto key = std::make_pair(character, isForPlatformFont);
    auto characterAddResult = fontAddResult.iterator->second.add(key, reinterpret_cast<SimpleFontData*>(0));

    SimpleFontData*& fallbackFontData = characterAddResult.iterator->second;

    if (!fallbackFontData) {
        UChar codeUnits[2];
        int codeUnitsLength;
        if (U_IS_BMP(character)) {
            codeUnits[0] = Font::normalizeSpaces(character);
            codeUnitsLength = 1;
        } else {
            codeUnits[0] = U16_LEAD(character);
            codeUnits[1] = U16_TRAIL(character);
            codeUnitsLength = 2;
        }

        fallbackFontData = fontCache().systemFallbackForCharacters(description, this, isForPlatformFont, codeUnits, codeUnitsLength).get();
        if (fallbackFontData)
            fallbackFontData->m_isUsedInSystemFallbackCache = true;
    }

    return fallbackFontData;
}

void SimpleFontData::removeFromSystemFallbackCache()
{
    systemFallbackCache().remove(this);

    if (!m_isUsedInSystemFallbackCache)
        return;

    SystemFallbackCache& cache = systemFallbackCache();
    for(auto it1 = cache.begin(); it1 != cache.end(); ++it1) {
    	CharacterFallbackMap& characterMap = it1->second;
        Vector<std::pair<UChar32, unsigned>, 512> toRemove;
        for(auto it2 = characterMap.begin(); it2 != characterMap.end(); ++it2) {
            if (it2->second == this)
                toRemove.append(it2->first);
        }
        for(int i = 0; i < toRemove.size(); ++i)
            characterMap.remove(toRemove[i]);
    }
}

} // namespace WebCore
