/*
 * Copyright (C) 2004, 2005, 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2009 Dirk Schulze <krit@webkit.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "BitmapImage.h"

#include "AffineTransform.h"
#include "Color.h"
#include "FloatRect.h"
#include "GraphicsContext.h"
#include "PlatformContextCairo.h"
#include "ImageBuffer.h"
#include "ImageObserver.h"
#include <math.h>
#include <runtime/JSValueInlines.h>
#include <wtf/OwnPtr.h>
#include "Timer.h"
#include <wtf/PlatformRunState.h>
#include "CairoUtilities.h"

#ifndef CAIRO_HAS_GLESV2_SURFACE
#define CAIRO_HAS_GLESV2_SURFACE 0
#endif

#if ENABLE(WEBGL)
#if CAIRO_HAS_GLESV2_SURFACE
#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include "cairo/cairo-gl-ext.h"
#endif
#endif

namespace WebCore {

DLL_PUBLIC PassRefPtr<BitmapImage> BitmapImage::create(cairo_surface_t* surface)
{
    return BitmapImage::create(new NativeImageCairo(surface));
}

BitmapImage::BitmapImage(NativeImageCairo* nativeImage, ImageObserver* observer, bool hasAlpha)
: Image(observer)
, m_source(this)
, m_currentFrame(0)
, m_frames(0)
, m_repetitionCount(cAnimationNone)
, m_repetitionCountStatus(Unknown)
, m_repetitionsComplete(0)
, m_decodedSize(0)
, m_frameCount(1)
, m_isSolidColor(false)
, m_checkedForSolidColor(false)
, m_animationFinished(true)
, m_allDataReceived(true)
, m_haveSize(true)
, m_sizeAvailable(true)
, m_haveFrameCount(true)
, m_scaledDataDeletionTimer(this, &BitmapImage::scaledDataDeletionTimerCallback)
{
    int width  = 0;
    int height = 0;

    getSurfaceSize(nativeImage->surface(), width, height);

	m_decodedSize = width * height * 4;
	m_size = IntSize(width, height);

	m_frames.grow(1);
	m_frames[0].m_frame = nativeImage;
	m_frames[0].m_hasAlpha = hasAlpha ? (cairo_surface_get_content(nativeImage->surface()) != CAIRO_CONTENT_COLOR) : false;
	m_frames[0].m_haveMetadata = true;

	checkForSolidColor();
}

void BitmapImage::draw(GraphicsContext* context, const FloatRect& dst, const FloatRect& src, CompositeOperator op, BlendMode blendMode)
{
	Locker locker(this);

    FloatRect srcRect(src);
    FloatRect dstRect(dst);

    if (dstRect.width() == 0.0f || dstRect.height() == 0.0f ||
        srcRect.width() == 0.0f || srcRect.height() == 0.0f)
        return;

    if(gGfxDisableRequested || (gPlatformRunState != WTF::Platform_State_Running))
    	return;

    startAnimation();

    FrameData* frameData = frameAtIndex(m_currentFrame);
    if(!frameData || !frameData->m_frame) // If it's too early we won't have an image yet.
        return;

    if (mayFillWithSolidColor()) {
        fillWithSolidColor(context, dstRect, solidColor(), op);
        return;
    }

    cairo_surface_t* surface = frameData->m_frame->surface();

#if !(CAIRO_HAS_GLESV2_SURFACE || CAIRO_HAS_GL_SURFACE)
    if(dstRect.size() != srcRect.size()){ // scaled image
    	if(frameData->m_scaledFrame && srcRect == frameData->m_srcRect && dstRect.size() == frameData->m_dstSize && !frameData->m_isPattern){
    		surface = frameData->m_scaledFrame->surface();
    		srcRect = FloatRect(0, 0, dstRect.width(), dstRect.height());
    	}
    	else if(context->imageInterpolationQuality() == InterpolationHigh || ((dstRect.width() * dstRect.height()) <= (128 * 128 /* 64K */))){
    		cairo_surface_t* scaledSurface;

    		if((dstRect.width() * dstRect.height()) <= (32 * 32))
    			scaledSurface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, dstRect.width(), dstRect.height());
    		else
    			scaledSurface = createAcceleratedSurface(CAIRO_FORMAT_ARGB32, dstRect.width(), dstRect.height());

		    frameData->clearScaledFrame();

    		if(scaledSurface){
    			RefPtr<cairo_t> scaledContext = adoptRef(cairo_create(scaledSurface));
    		    RefPtr<cairo_pattern_t> scaledPattern = adoptRef(cairo_pattern_create_for_surface(surface));

    		    switch(context->imageInterpolationQuality()) {
					case InterpolationNone:
					case InterpolationLow:
						cairo_pattern_set_filter(scaledPattern.get(), CAIRO_FILTER_FAST);
						break;
					case InterpolationMedium:
						cairo_pattern_set_filter(scaledPattern.get(), CAIRO_FILTER_GOOD);
						break;
					case InterpolationHigh:
						cairo_pattern_set_filter(scaledPattern.get(), CAIRO_FILTER_BILINEAR);
						break;
					case InterpolationDefault:
						cairo_pattern_set_filter(scaledPattern.get(), CAIRO_FILTER_GOOD);
						break;
    		    }

    		    cairo_pattern_set_extend(scaledPattern.get(), CAIRO_EXTEND_PAD);

    		    float scaleX = srcRect.width() / dstRect.width();
    		    float scaleY = srcRect.height() / dstRect.height();
    		    cairo_matrix_t matrix = { scaleX, 0, 0, scaleY, srcRect.x(), srcRect.y() };
    		    cairo_pattern_set_matrix(scaledPattern.get(), &matrix);

    		    cairo_set_source(scaledContext.get(), scaledPattern.get());
    		    cairo_set_operator(scaledContext.get(), CAIRO_OPERATOR_SOURCE);
    		    cairo_rectangle(scaledContext.get(), 0, 0, dstRect.width(), dstRect.height());
    		    cairo_fill(scaledContext.get());
    		    cairo_surface_flush(scaledSurface);

    			frameData->m_scaledFrame = new NativeImageCairo(scaledSurface);
    			frameData->m_srcRect = srcRect;
    			frameData->m_dstSize = dstRect.size();
    			frameData->m_isPattern = false;

    			surface = scaledSurface;
        		srcRect = FloatRect(0, 0, dstRect.width(), dstRect.height());

        		gBitmapImages.add(this);
    	    }
    	}
    }
#endif

    context->save();

    // Set the compositing operation.
    if (op == CompositeSourceOver && blendMode == BlendModeNormal && !frameHasAlphaAtIndex(m_currentFrame))
        context->setCompositeOperation(CompositeCopy);
    else
        context->setCompositeOperation(op, blendMode);

    context->platformContext()->drawSurfaceToContext(surface, dstRect, srcRect, context);

    context->restore();

    if (imageObserver())
        imageObserver()->didDraw(this);

	if(m_scaledDataDeletionTimer.isActive())
		m_scaledDataDeletionTimer.stop();

    if(frameData->m_scaledFrame){
		double delay;

		if((dstRect.width() * dstRect.height()) > (1920 * 1080))
			delay = 1.0;
		else
			delay = 5.0;

		m_scaledDataDeletionTimer.startOneShot(delay);
    }
}

void BitmapImage::drawPattern(GraphicsContext* context, const FloatRect& src, const AffineTransform& patternTransform,
                        const FloatPoint& phase, const FloatSize& spacing, CompositeOperator op, const FloatRect& destRect, BlendMode)
{
	Locker locker(this);

    if(gGfxDisableRequested || (gPlatformRunState != WTF::Platform_State_Running))
    	return;

    // Set the compositing operation.
    if (op == CompositeSourceOver && currentFrameKnownToBeOpaque())
    	op = CompositeCopy;

    FrameData* frameData = frameAtIndex(m_currentFrame);
    if(!frameData || !frameData->m_frame) // If it's too early we won't have an image yet.
        return;

    if (mayFillWithSolidColor()) {
        fillWithSolidColor(context, destRect, solidColor(), op);
        return;
    }

    IntSize imageSize(size());

    cairo_surface_t* surface = frameData->m_frame->surface();

    if( cairo_surface_get_type(surface) == CAIRO_SURFACE_TYPE_IMAGE &&
    	(imageSize.width() <= 32 || imageSize.height() <= 32) &&
    	((imageSize.width() * imageSize.height()) <= (256 * 256)))
    {
    	/*
    	 * Cairo should cache the replicated surface
    	 */
    	FloatRect imageRect = FloatRect(0, 0, imageSize.width(), imageSize.height());

    	if(frameData->m_scaledFrame && imageRect == frameData->m_srcRect && frameData->m_isPattern) {
    		surface = frameData->m_scaledFrame->surface();
    	} else {
			frameData->clearScaledFrame();

			if(cairo_surface_t* copiedSurface = createAcceleratedSurface(CAIRO_FORMAT_ARGB32, imageSize.width(), imageSize.height())){
				RefPtr<cairo_t> scaledContext = adoptRef(cairo_create(copiedSurface));
				RefPtr<cairo_pattern_t> scaledPattern = adoptRef(cairo_pattern_create_for_surface(surface));

				cairo_pattern_set_extend(scaledPattern.get(), CAIRO_EXTEND_PAD);

				cairo_set_source(scaledContext.get(), scaledPattern.get());
				cairo_set_operator(scaledContext.get(), CAIRO_OPERATOR_SOURCE);
				cairo_paint(scaledContext.get());
				cairo_surface_flush(copiedSurface);

				frameData->m_scaledFrame = new NativeImageCairo(copiedSurface);
				frameData->m_srcRect = imageRect;
				frameData->m_dstSize = imageSize;
				frameData->m_isPattern = true;

				surface = copiedSurface;

				gBitmapImages.add(this);
			}
    	}
    }

    drawPatternToCairoContext(context, surface, imageSize, src, patternTransform, phase, toCairoOperator(op), destRect);

    if (imageObserver())
        imageObserver()->didDraw(this);

	if(m_scaledDataDeletionTimer.isActive())
		m_scaledDataDeletionTimer.stop();

    if(frameData->m_scaledFrame){
		double delay;

		if((src.width() * src.height()) > (1920 * 1080))
			delay = 1.0;
		else
			delay = 5.0;

		m_scaledDataDeletionTimer.startOneShot(delay);
    }
}

void BitmapImage::drawTiled(GraphicsContext* ctxt, const FloatRect& destRect, const FloatPoint& srcPoint, const FloatSize& scaledTileSize, const FloatSize& spacing, CompositeOperator op)
{
	Locker locker(this);

    if(gGfxDisableRequested || (gPlatformRunState != WTF::Platform_State_Running))
    	return;

    if (mayFillWithSolidColor()) {
        fillWithSolidColor(ctxt, destRect, solidColor(), op);
        return;
    }

    // See <https://webkit.org/b/59043>.
#if !PLATFORM(WX)
    ASSERT(!isBitmapImage() || notSolidColor());
#endif

    FloatSize intrinsicTileSize = size();
    if (hasRelativeWidth())
        intrinsicTileSize.setWidth(scaledTileSize.width());
    if (hasRelativeHeight())
        intrinsicTileSize.setHeight(scaledTileSize.height());

    FloatSize scale(scaledTileSize.width() / intrinsicTileSize.width(), scaledTileSize.height() / intrinsicTileSize.height());

    FloatRect oneTileRect;
    oneTileRect.setX(destRect.x() + fmodf(fmodf(-srcPoint.x(), scaledTileSize.width()) - scaledTileSize.width(), scaledTileSize.width()));
    oneTileRect.setY(destRect.y() + fmodf(fmodf(-srcPoint.y(), scaledTileSize.height()) - scaledTileSize.height(), scaledTileSize.height()));
    oneTileRect.setSize(scaledTileSize);

    if((scaledTileSize == intrinsicTileSize) || (ctxt->imageInterpolationQuality() != InterpolationHigh))
    {
		// Check and see if a single draw of the image can cover the entire area we are supposed to tile.
		if (oneTileRect.contains(destRect)) {
			FloatRect visibleSrcRect;
			visibleSrcRect.setX((destRect.x() - oneTileRect.x()) / scale.width());
			visibleSrcRect.setY((destRect.y() - oneTileRect.y()) / scale.height());
			visibleSrcRect.setWidth(destRect.width() / scale.width());
			visibleSrcRect.setHeight(destRect.height() / scale.height());
			draw(ctxt, destRect, visibleSrcRect, op, BlendModeNormal);
			return;
		}
    }

    AffineTransform patternTransform = AffineTransform().scaleNonUniform(scale.width(), scale.height());
	FloatRect tileRect(FloatPoint(), intrinsicTileSize);

	drawPattern(ctxt, tileRect, patternTransform, oneTileRect.location(), spacing, op, destRect);

    startAnimation();
}

void Image::drawPattern(GraphicsContext* context, const FloatRect& srcRect, const AffineTransform& patternTransform,
                        const FloatPoint& phase, const FloatSize& spacing, CompositeOperator op, const FloatRect& destRect, BlendMode)
{
	Locker locker(this);

    if(gGfxDisableRequested || (gPlatformRunState != WTF::Platform_State_Running))
    	return;

    NativeImageCairo* image = nativeImageForCurrentFrame();
    if (!image) // If it's too early we won't have an image yet.
        return;

    // Set the compositing operation.
    if (op == CompositeSourceOver && currentFrameKnownToBeOpaque())
    	op = CompositeCopy;

    drawPatternToCairoContext(context, image->surface(), IntSize(size()), srcRect, patternTransform, phase, toCairoOperator(op), destRect);

    if (imageObserver())
        imageObserver()->didDraw(this);
}

#if ENABLE(WEBGL)
uint32_t Image::webglTexture()
{
#if (CAIRO_HAS_GLESV2_SURFACE || CARIO_HAS_GL_SURFACE)
    if(NativeImageCairo* image = nativeImageForCurrentFrame()){
    	GLuint tex;
    	GLuint fb;

    	if(cairo_gl_surface_get_texture(image->surface(), &tex, &fb) == CAIRO_STATUS_SUCCESS)
    		return tex;
    }
#endif

	return 0;
}

void Image::clearTexture()
{
#if (CAIRO_HAS_GLESV2_SURFACE || CAIRO_HAS_GL_SURFACE)
	destroyAllAcceleratedSurfaces();
#endif
}

#endif

void BitmapImage::checkForSolidColor()
{
	Locker locker(this);

    if (frameCount() == 0) // do not force decoding
        return;

    m_isSolidColor = false;
    m_checkedForSolidColor = true;

    if (frameCount() > 1)
        return;

    FrameData* frameData = frameAtIndex(m_currentFrame);
    if(!frameData || !frameData->m_frame) // If it's too early we won't have an image yet.
        return;

    cairo_surface_t* surface = frameData->m_frame->surface();

    if(cairo_surface_get_type(surface) == CAIRO_SURFACE_TYPE_IMAGE){
	    int width = cairo_image_surface_get_width(surface);
	    int height = cairo_image_surface_get_height(surface);

	    if (width != 1 || height != 1)
        	return;

	    unsigned* pixelColor = reinterpret_cast<unsigned*>(cairo_image_surface_get_data(surface));
	    m_solidColor = colorFromPremultipliedARGB(*pixelColor);
	    m_isSolidColor = true;
    }
}

}

