/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2009 Dirk Schulze <krit@webkit.org>
 * Copyright (C) 2010 Holger Hans Peter Freyther
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "Font.h"
#include "AffineTransform.h"
#include "GlyphBuffer.h"
#include "Gradient.h"
#include "GraphicsContext.h"
#include "PlatformContextCairo.h"
#include "ImageBuffer.h"
#include "Pattern.h"
#include "FontPlatformData.h"
#include "SimpleFontData.h"
#include "CairoUtilities.h"
#include <cairo-ft.h>
#include <cairo.h>
    
#define SYNTHETIC_OBLIQUE_ANGLE 14

namespace WebCore {

static void drawGlyphsToContext(cairo_t* cr, FontPlatformData* font, cairo_glyph_t* glyphs, int numGlyphs)
{
    cairo_matrix_t originalTransform;
    float syntheticBoldOffset = font->syntheticBoldOffset();
    if (syntheticBoldOffset)
        cairo_get_matrix(cr, &originalTransform);

    cairo_set_scaled_font(cr, font->scaledFont());

    cairo_scaled_font_t* used_scaled_font = cairo_get_scaled_font(cr);

    if(used_scaled_font != font->scaledFont())
    	font->cacheScaledFont(used_scaled_font);

    cairo_show_glyphs(cr, glyphs, numGlyphs);

    if (syntheticBoldOffset) {
        cairo_translate(cr, syntheticBoldOffset, 0);
        cairo_show_glyphs(cr, glyphs, numGlyphs);
    }

    if (syntheticBoldOffset)
        cairo_set_matrix(cr, &originalTransform);
}

static void drawGlyphsShadow(GraphicsContext* graphicsContext, const FloatPoint& point, FontPlatformData* font, cairo_glyph_t* glyphs, int numGlyphs)
{
	ShadowState shadowState(graphicsContext->state());
	ShadowBlur shadow(FloatSize(shadowState.blur, shadowState.blur), shadowState.offset, shadowState.color, shadowState.ignoreTransforms);

    if (!(graphicsContext->textDrawingMode() & TextModeFill) || shadow.type() == ShadowBlur::NoShadow)
        return;

    if (!shadowState.isRequired(*graphicsContext->platformContext())) {
        // Optimize non-blurry shadows, by just drawing text without the ShadowBlur.
        cairo_t* context = graphicsContext->platformContext()->cr();
        cairo_save(context);

        FloatSize shadowOffset(graphicsContext->state().shadowOffset);
        cairo_translate(context, shadowOffset.width(), shadowOffset.height());

		float red, green, blue, alpha;
        graphicsContext->state().shadowColor.getRGBA(red, green, blue, alpha);
        cairo_set_source_rgba(context, red, green, blue, alpha);

        drawGlyphsToContext(context, font, glyphs, numGlyphs);

        cairo_restore(context);
        return;
    }

    cairo_text_extents_t extents;
    cairo_scaled_font_glyph_extents(font->scaledFont(), glyphs, numGlyphs, &extents);
    FloatRect fontExtentsRect(point.x() + extents.x_bearing, point.y() + extents.y_bearing, extents.width, extents.height);

    bool forceImageSurface = false;

#if USE(ACCELERATED_COMPOSITING)
    forceImageSurface = true;
#endif

    if (GraphicsContext* shadowContext = shadow.beginShadowLayer(graphicsContext, fontExtentsRect, forceImageSurface)) {
        drawGlyphsToContext(shadowContext->platformContext()->cr(), font, glyphs, numGlyphs);
        shadow.endShadowLayer(graphicsContext);
    }
}

void FontPlatformData::drawGlyphs(GraphicsContext* context, const GlyphBuffer& glyphBuffer, int from, int numGlyphs, const FloatPoint& point)
{
    cairo_t* cr = context->platformContext()->cr();

	if(UNLIKELY(!m_scaledFont))
		return;

    Vector<cairo_glyph_t, 128> cairo_glyphs(numGlyphs);

    auto xOffset = point.x();
    {
        auto* glyphsData = glyphBuffer.glyphs(from);
        auto* advances = glyphBuffer.advances(from);

        auto yOffset = point.y();
		for (int i = 0; i < numGlyphs; i++) {
			cairo_glyph_t& cairo_glyph = cairo_glyphs[i];
			cairo_glyph.index = glyphsData[i];
			cairo_glyph.x = xOffset;
			cairo_glyph.y = yOffset;

			xOffset += advances[i];
		}
    }
    
    drawGlyphsShadow(context, point, this, cairo_glyphs.data(), numGlyphs);

    cairo_save(cr);

    if (context->textDrawingMode() & TextModeFill) {
    	context->platformContext()->prepareForFilling(context->state(), PlatformContextCairo::AdjustPatternForGlobalAlpha);
        drawGlyphsToContext(cr, this, cairo_glyphs.data(), numGlyphs);
    }

    // Prevent running into a long computation within cairo. If the stroke width is
    // twice the size of the width of the text we will not ask cairo to stroke
    // the text as even one single stroke would cover the full wdth of the text.
    //  See https://bugs.webkit.org/show_bug.cgi?id=33759.
    if ((context->textDrawingMode() & TextModeStroke) && context->strokeThickness() < 2 * xOffset) {
    	context->platformContext()->prepareForStroking(context->state());
        cairo_set_line_width(cr, context->strokeThickness());

        // This may disturb the CTM, but we are going to call cairo_restore soon after.
        cairo_set_scaled_font(cr, m_scaledFont.get());
        cairo_glyph_path(cr, cairo_glyphs.data(), numGlyphs);
        cairo_stroke(cr);
    }

    cairo_restore(cr);
}

void Font::drawGlyphs(GraphicsContext* context, const SimpleFontData* font, const GlyphBuffer& glyphBuffer, int from, int numGlyphs, const FloatPoint& point) const
{
	const_cast<FontPlatformData&>(font->platformData()).drawGlyphs(context, glyphBuffer, from, numGlyphs, point);
}

}
