/*
 * Copyright (C) 2007, 2009 Apple Inc.  All rights reserved.
 * Copyright (C) 2007 Collabora Ltd. All rights reserved.
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2009, 2010, 2011 Espial Group Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * aint with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef MediaPlayerPrivateEvo_h
#define MediaPlayerPrivateEvo_h

#if ENABLE(VIDEO)

#include "MediaPlayerPrivate.h"
#include <wtf/OwnPtr.h>
#include <wtf/RefPtr.h>
#include <wtf/text/WTFString.h>
#include <wtf/HashSet.h>
#include "Timer.h"

namespace WebCore {
#if USE(VIDEO_AS_TEXTURE)
class VideoLayerEvo;
#endif

#if USE(ACCELERATED_COMPOSITING)
class GraphicsLayer;
#endif

class GraphicsContext;
class HTMLMediaElement;
class IntSize;
class IntRect;

class SubresourceLoader;
class HTML5MediaPlayerCallbackClient;
class MediaPlayerPrivatePlatformData;
class MediaSourcePrivateEvo;
class InbandTextTrackPrivateEvo;
class SourceBufferPrivateEvo;

class MediaPlayerPrivate : public MediaPlayerPrivateInterface
{
	friend class MediaSourcePrivateEvo;
	friend class SourceBufferPrivateEvo;

public:
	static void registerMediaEngine(MediaEngineRegistrar);

#if ENABLE(LEGACY_ENCRYPTED_MEDIA_V2)
	static void initCDM();
#endif

	static bool supportsMIMEType(const String& mimeType);

	virtual ~MediaPlayerPrivate();

    virtual void attributeChanged(const String& name, const String& value);

	virtual IntSize naturalSize() const;
	virtual bool hasVideo() const;
	virtual bool hasAudio() const;

    virtual unsigned videoWidth() const;
    virtual unsigned videoHeight() const;

	virtual void load(const KURL& url, double fragmentStartTime);
#if ENABLE(MEDIA_SOURCE)
    virtual void load(const String& url, PassRefPtr<HTMLMediaSource>);
#endif

	virtual void cancelLoad() { }

	virtual void prepareToPlay();
	virtual void play();
	virtual void pause();
    virtual void stop();

	virtual bool paused() const;
	virtual bool seeking() const;

	virtual double durationDouble() const;
	virtual double currentTimeDouble() const;
	virtual double startTimeDouble() const;
	virtual void seekWithTolerance(double time, double negativeTolerance, double positiveTolerance);

	virtual void setRateDouble(double);
	virtual double rateDouble();
	virtual void setVolumeDouble(double);
	virtual double volume();
	virtual void setMuted(bool);
	virtual bool muted();
	virtual void setPoster(const String&);
	virtual bool canLoadPoster(void) const;
	virtual void setClosedCaptionsVisible(bool);

	virtual MediaPlayer::NetworkState networkState() const { return m_networkState; }
	virtual MediaPlayer::ReadyState readyState() const { return m_readyState; }

	void mediaStateChanged();
	void naturalSizeChanged();
	void timeChanged(bool playEnded);
    void needsRedraw();

	void rateChanged()
	{
		m_player->rateChanged();
	}

	void muteChanged()
	{
		m_player->muteChanged(muted());
	}

	void volumeChanged()
	{
		m_player->volumeChanged(volume());
	}

	void durationChanged()
	{
		m_player->durationChanged();
	}

	virtual PassRefPtr<TimeRanges> buffered() const;
	virtual PassRefPtr<TimeRanges> seekable() const;
	virtual double maxTimeSeekableDouble() const;

	virtual void setVisible(bool);

	virtual void setSize(const IntSize& size);
	void setBounds(GraphicsContext* ctx, IntRect bounds);

	void* webView();
	void webViewDestroyed();

	virtual bool didLoadingProgress() const { return true; }

	void repaint();
	virtual void paint(GraphicsContext*, const IntRect&);
    virtual void paintCurrentFrameInContext(GraphicsContext*, const IntRect&);

    virtual bool copyVideoTextureToPlatformTexture(GraphicsContext3D*, Platform3DObject, GC3Dint, GC3Denum, GC3Denum, bool, bool);
	virtual void setPreload(MediaPlayer::Preload);

	virtual bool hasSingleSecurityOrigin() const;

	unsigned int getMediaCapabilities() const;
	void loadMediaCapabilities();

	virtual bool supportsFullscreen() const;
	virtual bool supportsMuting() const;
	virtual bool supportsSave() const;
	virtual bool supportsLoadingNewURL() const;
    virtual bool supportsFragmentStartTime() const;

	virtual void prepareForLoad();

	virtual bool hasClosedCaptions() const;
	virtual bool handleRemote() const;

	MediaPlayer* player() const { return m_player; }
	HTMLMediaElement* element() const;

	const IntRect& playerBounds() const { return m_playerBounds; }

	virtual int mediaPlayerID() const { return m_playerID; }
	void setupPlayer();

	bool isSeekable() const;

	void handlePlayKey();
	void handlePauseKey();
	void handleStopKey();
	void handleFastForwardKey(float factor);
	void handleFastBackKey(float factor);

#if ENABLE(MEDIA_STATISTICS)
    virtual unsigned decodedFrameCount() const;
    virtual unsigned droppedFrameCount() const;
    virtual unsigned audioDecodedByteCount() const;
    virtual unsigned videoDecodedByteCount() const;
#endif

#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    virtual MediaPlayer::MediaKeyException addKey(const String& keySystem, const unsigned char* key, unsigned keyLength, const unsigned char* initData, unsigned initDataLength, const String& sessionId);
    virtual MediaPlayer::MediaKeyException generateKeyRequest(const String& keySystem, const unsigned char* initData, unsigned initDataLength);
    virtual MediaPlayer::MediaKeyException cancelKeyRequest(const String& keySystem, const String& sessionId);
#endif

#if ENABLE(ENCRYPTED_MEDIA)
    virtual void attachMediaKeys(MediaKeys* mediaKeys);
#endif

#if ENABLE(VIDEO_TRACK)
    virtual void audioTrackEnabledChanged(AudioTrack*, AudioTrackList*);

    void clearInbandTextTracks();

    void addTextTrack(RefPtr<InbandTextTrackPrivateEvo>);

    void textTrackModeChanged(InbandTextTrackPrivateEvo*);
#endif

#if ENABLE(MEDIA_SOURCE)
    virtual unsigned long totalVideoFrames();
    virtual unsigned long droppedVideoFrames();
    virtual unsigned long corruptedVideoFrames();
    virtual double totalFrameDelay();
#endif

#if USE(VIDEO_AS_TEXTURE)
    void* eglImage(size_t pixmap_idx);
    size_t videoFrameIndex();
#endif

#if USE(ACCELERATED_COMPOSITING)
    // only for comparison
	GraphicsLayer* graphicsLayer() { return m_graphicsLayer; }
#endif

private:
	MediaPlayerPrivate(MediaPlayer*,int);

	static PassOwnPtr<MediaPlayerPrivateInterface> create(MediaPlayer* player);

	bool updateMediaState(bool& ended);

	void updateNaturalSizeIfNeeded() const;

	void doLoad();

    void doLoadTimerCallback(Timer<MediaPlayerPrivate>*);

#if ENABLE(VIDEO_TRACK) 
    void configureInbandTracks(); 
#endif 

#if USE(VIDEO_AS_TEXTURE)
    bool supportsAcceleratedRendering() const;
    void acceleratedRenderingStateChanged();
    PlatformLayer* platformLayer() const;
    VideoLayerEvo* m_videoLayer;
#endif

private:
	MediaPlayer* m_player;
	bool m_visible;

	bool m_seeking;
	bool m_paused;
	bool m_stopped;
	bool m_delayingLoad;
	bool m_firstTimeUpdate;
	MediaPlayer::Preload m_preload;
	KURL m_url;
	String m_mimeType;
	double m_fragmentStartTime;
	unsigned int m_playerCapabilities;
	unsigned int m_mediaCapabilities;

	MediaPlayer::ReadyState m_readyState;
	MediaPlayer::NetworkState m_networkState;

	double m_duration;
	double m_rate;

	int m_playerID;
	int m_zIndex;
	IntRect m_bounds;
	IntRect m_playerBounds;

	OwnPtr<HTML5MediaPlayerCallbackClient> m_callbackClient;
	OwnPtr<MediaPlayerPrivatePlatformData> m_platformData;

	void* m_webView;

#if USE(ACCELERATED_COMPOSITING)
	GraphicsLayer* m_graphicsLayer;
#endif

#if ENABLE(VIDEO_TRACK)
	Vector<RefPtr<InbandTextTrackPrivateEvo> > m_inbandTextTracks;
#endif 

	bool m_sawNaturalSizeUpdate;
	mutable IntSize m_cachedNaturalSize;

	Timer<MediaPlayerPrivate> m_loadTimer;
};

}

#endif
#endif
