/*
 * Copyright (C) 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2010 Igalia S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "FontCustomPlatformData.h"

#include "FontPlatformData.h"
#include "MemoryCache.h"
#include "SharedBuffer.h"
#include "WOFFFileFormat.h"
#include "WOFF2FileFormat.h"
#include <cairo-ft.h>
#include <cairo.h>
#include <ft2build.h>
#include FT_MODULE_H

namespace WebCore {

DLL_PUBLIC int gFontCustomPlatformDataCount;

static void releaseCustomFontData(void* data)
{
    static_cast<SharedBuffer*>(data)->deref();
}

FontCustomPlatformData::FontCustomPlatformData(FT_Face freeTypeFace, SharedBuffer& buffer)
    : m_freeTypeFace(freeTypeFace)
	, m_fontFace(cairo_ft_font_face_create_for_ft_face(freeTypeFace, FT_LOAD_DEFAULT))
	, m_simpleFontDataPruneTimer(this, &FontCustomPlatformData::simpleFontDataPruneTimerCallback)
{
    buffer.ref(); // This is balanced by the buffer->deref() in releaseCustomFontData.

    // Cairo doesn't do FreeType reference counting, so we need to ensure that when
    // this cairo_font_face_t is destroyed, it cleans up the FreeType face as well.
    static cairo_user_data_key_t freeTypeFaceKey;
    cairo_font_face_set_user_data(m_fontFace, &freeTypeFaceKey, freeTypeFace,
         reinterpret_cast<cairo_destroy_func_t>(FT_Done_Face));

    static cairo_user_data_key_t bufferKey;
    cairo_font_face_set_user_data(m_fontFace, &bufferKey, &buffer,
         static_cast<cairo_destroy_func_t>(releaseCustomFontData));

    ++gFontCustomPlatformDataCount;
}

FontCustomPlatformData::~FontCustomPlatformData()
{
	m_cache.clear();

    // m_freeTypeFace will be destroyed along with m_fontFace. See the constructor.
    cairo_font_face_destroy(m_fontFace);

    --gFontCustomPlatformDataCount;
}

PassRefPtr<SimpleFontData> FontCustomPlatformData::fontPlatformData(const FontDescription& fontDescription, bool syntheticBold, bool syntheticItalic, FontOrientation orientation)
{
	double now = WTF::monotonicallyIncreasingTime();

	for(int i = 0; i < m_cache.size(); ++i) {
		const FontPlatformData& platformData = m_cache[i].first->platformData();

		if( platformData.fontDescription().computedSize() == fontDescription.computedSize()
			&& platformData.fontDescription().computedHorizontalSize() == fontDescription.computedHorizontalSize()
			&& platformData.fontDescription().stretch() == fontDescription.stretch()
			&& platformData.fontDescription().fontSizeClamp() == fontDescription.fontSizeClamp()
			&& platformData.syntheticBold() == syntheticBold
			&& platformData.syntheticOblique() == syntheticItalic
			&& platformData.orientation() == orientation) {
			//WTF::logMessage(String("FontCustomPlatformData hit cache : size ") + String::number(fontDescription.computedSize()) + String(", horizontal size ") + String::number(fontDescription.computedHorizontalSize()));
			m_cache[i].second = now;
			return m_cache[i].first;
		}
	}

	//WTF::logMessage(String("FontCustomPlatformData miss cache : size ") + String::number(fontDescription.computedSize()) + String(", horizontal size ") + String::number(fontDescription.computedHorizontalSize()));

	RefPtr<SimpleFontData> simpleFontData = SimpleFontData::create(FontPlatformData(fontDescription, m_fontFace, syntheticBold, syntheticItalic), true, false);

	m_cache.append(std::pair<RefPtr<SimpleFontData>, double>(simpleFontData, now));

	if(double interval = memoryCache().simpleFontDataDeletionInterval())
		if(!m_simpleFontDataPruneTimer.isActive())
			m_simpleFontDataPruneTimer.startOneShot(0.25); // every 250ms

	return simpleFontData.release();
}

void FontCustomPlatformData::simpleFontDataPruneTimerCallback(Timer<FontCustomPlatformData>*)
{
	double interval = memoryCache().simpleFontDataDeletionInterval();

	if(!interval)
		return;

	double now = WTF::monotonicallyIncreasingTime();

	for(int i = m_cache.size() - 1; i >= 0; --i) {
		if(m_cache[i].first->refCount() > 1)
			continue;

		if((now - m_cache[i].second) <= interval)
			continue;

		m_cache.remove(i);
	}

	if(m_cache.size() > 0)
		if(!m_simpleFontDataPruneTimer.isActive())
			m_simpleFontDataPruneTimer.startOneShot(0.25); // every 250ms
}

void FontCustomPlatformData::purge()
{
	for(int i = m_cache.size() - 1; i >= 0; --i) {
		if(m_cache[i].first->refCount() > 1)
			continue;

		m_cache.remove(i);
	}
}

static void* alloc_(FT_Memory, long size)
{
    return fastMalloc(size);
}

static void free_(FT_Memory, void* block)
{
	fastFree(block);
}

static void* realloc_(FT_Memory, long, long newSize, void* block)
{
	return fastRealloc(block, newSize);
}

static bool initializeFreeTypeLibrary(FT_Library& library)
{
    // https://www.freetype.org/freetype2/docs/design/design-4.html
    // https://lists.nongnu.org/archive/html/freetype-devel/2004-10/msg00022.html

    FT_Memory memory = bitwise_cast<FT_Memory>(ft_smalloc(sizeof(*memory)));
    if (!memory)
        return false;

    memory->user = 0;
    memory->alloc = alloc_;
    memory->free = free_;
    memory->realloc = realloc_;

    if (FT_New_Library(memory, &library)) {
        ft_sfree(memory);
        return false;
    }

    FT_Add_Default_Modules(library);
    return true;
}

FontCustomPlatformData* createFontCustomPlatformData(SharedBuffer& buffer)
{
    static FT_Library library;
    if (!library && !initializeFreeTypeLibrary(library)) {
        library = 0;
        return 0;
    }

    FT_Face freeTypeFace;
    if (FT_New_Memory_Face(library, reinterpret_cast<const FT_Byte*>(buffer.data()), buffer.size(), 0, &freeTypeFace))
        return 0;
    return new FontCustomPlatformData(freeTypeFace, buffer);
}

bool FontCustomPlatformData::supportsFormat(const String& format)
{
    return
    	equalIgnoringCase(format, ASCIILiteral("truetype")) ||
		equalIgnoringCase(format, ASCIILiteral("opentype")) ||
		equalIgnoringCase(format, ASCIILiteral("woff")) ||
		equalIgnoringCase(format, ASCIILiteral("woff2"));
}

}
