
#include "config.h"

#include "CairoUtilities.h"
#include "FontBaseline.h"
#include "FontPlatformData.h"
#include "PlatformString.h"
#include "Font.h"
#include "FontCache.h"
#include "FontDescription.h"
#include "SimpleFontData.h"
#include "TextEncoding.h"
#include "UTF16UChar32Iterator.h"
#include <wtf/HashSet.h>
#include <wtf/HexNumber.h>
#include <wtf/MathExtras.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/StringBuilder.h>
#include <wtf/unicode/UTF8.h>
#include <freetype/tttables.h>

#ifndef CAIRO_HAS_GLESV2_SURFACE
#define CAIRO_HAS_GLESV2_SURFACE 0
#endif

namespace WebCore {

static bool sFontAntialiased = false;

DLL_PUBLIC void setFontAntialiased(bool antialias)
{
	sFontAntialiased = antialias;
}

static HashSet<FontPlatformData*>& fontPlatformDataCache()
{
	static HashSet<FontPlatformData*> cache;
	return cache;
}

static cairo_subpixel_order_t convertFontConfigSubpixelOrder(int fontConfigOrder)
{
    switch (fontConfigOrder) {
    case FC_RGBA_RGB:
        return CAIRO_SUBPIXEL_ORDER_RGB;
    case FC_RGBA_BGR:
        return CAIRO_SUBPIXEL_ORDER_BGR;
    case FC_RGBA_VRGB:
        return CAIRO_SUBPIXEL_ORDER_VRGB;
    case FC_RGBA_VBGR:
        return CAIRO_SUBPIXEL_ORDER_VBGR;
    case FC_RGBA_NONE:
    case FC_RGBA_UNKNOWN:
        return CAIRO_SUBPIXEL_ORDER_DEFAULT;
    }
    return CAIRO_SUBPIXEL_ORDER_DEFAULT;
}

static cairo_hint_style_t convertFontConfigHintStyle(int fontConfigStyle)
{
    switch (fontConfigStyle) {
    case FC_HINT_NONE:
        return CAIRO_HINT_STYLE_NONE;
    case FC_HINT_SLIGHT:
        return CAIRO_HINT_STYLE_SLIGHT;
    case FC_HINT_MEDIUM:
        return CAIRO_HINT_STYLE_MEDIUM;
    case FC_HINT_FULL:
        return CAIRO_HINT_STYLE_FULL;
    }
    return CAIRO_HINT_STYLE_NONE;
}

static void setCairoFontOptionsFromFontConfigPattern(cairo_font_options_t* options, FcPattern* pattern)
{
    FcBool booleanResult;
    int integerResult;

    if (FcPatternGetInteger(pattern, FC_RGBA, 0, &integerResult) == FcResultMatch) {
        cairo_font_options_set_subpixel_order(options, convertFontConfigSubpixelOrder(integerResult));

        // Based on the logic in cairo-ft-font.c in the cairo source, a font with
        // a subpixel order implies that is uses subpixel antialiasing.
        if (integerResult != FC_RGBA_NONE)
            cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_SUBPIXEL);
    }

    if (FcPatternGetBool(pattern, FC_ANTIALIAS, 0, &booleanResult) == FcResultMatch) {
        // Only override the anti-aliasing setting if was previously turned off. Otherwise
        // we'll override the preference which decides between gray anti-aliasing and
        // subpixel anti-aliasing.
        if (!booleanResult)
            cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_NONE);
        else if (cairo_font_options_get_antialias(options) == CAIRO_ANTIALIAS_NONE)
            cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_GRAY);
    }

    if (FcPatternGetInteger(pattern, FC_HINT_STYLE, 0, &integerResult) == FcResultMatch)
        cairo_font_options_set_hint_style(options, convertFontConfigHintStyle(integerResult));
    if (FcPatternGetBool(pattern, FC_HINTING, 0, &booleanResult) == FcResultMatch && !booleanResult)
        cairo_font_options_set_hint_style(options, CAIRO_HINT_STYLE_NONE);

    // FreeType's metric hinting uses rounding
    cairo_font_options_set_hint_metrics(options, CAIRO_HINT_METRICS_OFF);
}

static cairo_font_options_t* getDefaultCairoFontOptions()
{
#if 0 && PLATFORM(GTK)
    if (GdkScreen* screen = gdk_screen_get_default()) {
        const cairo_font_options_t* screenOptions = gdk_screen_get_font_options(screen);
        if (screenOptions)
            return cairo_font_options_copy(screenOptions);
    }
#endif
    return cairo_font_options_create();
}

static FcPattern* getDefaultFontconfigOptions()
{
    // Get some generic default settings from fontconfig for web fonts. Strategy
    // from Behdad Esfahbod in https://code.google.com/p/chromium/issues/detail?id=173207#c35
    // For web fonts, the hint style is overridden in FontCustomPlatformData::FontCustomPlatformData
    // so Fontconfig will not affect the hint style, but it may disable hinting completely.
    static FcPattern* pattern = 0;
    if(!pattern) {
        pattern = FcPatternCreate();
        FcConfigSubstitute(0, pattern, FcMatchPattern);
        FcDefaultSubstitute(pattern);
        FcPatternDel(pattern, FC_FAMILY);
        FcConfigSubstitute(0, pattern, FcMatchFont);
    }
    return pattern;
}

DLL_PUBLIC FontPlatformData::FontPlatformData(const FontDescription& fontDescription, FcPattern* fcPattern, const String& familyName)
: m_pattern(fcPattern)
, m_fontDescription(fontDescription)
, m_familyName(familyName)
, m_syntheticBold(false)
, m_syntheticOblique(false)
, m_fixedWidth(false)
, m_spaceWidth(0)
, m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
{
    ASSERT(m_pattern);
	if(!m_pattern)
		return;

    // Disable bitmap strikes in all local fonts if outlines are available.
    // Bitmap strikes are 1-bit black-and-white images, so properly hinted
    // anti-aliased glyphs are usually better even at very small sizes.
    FcPatternDel(m_pattern.get(), FC_EMBEDDED_BITMAP);
    FcPatternAddBool(m_pattern.get(), FC_EMBEDDED_BITMAP, false);

    m_fontFace = adoptRef(cairo_ft_font_face_create_for_pattern(m_pattern.get()));

    bool descriptionAllowsSyntheticBold = fontDescription.fontSynthesis() & FontSynthesisWeight;
    if (descriptionAllowsSyntheticBold && fontDescription.weight() >= FontWeightBold) {
        // The FC_EMBOLDEN property instructs us to fake the boldness of the font.
        FcBool fontConfigEmbolden;
        if (FcPatternGetBool(m_pattern.get(), FC_EMBOLDEN, 0, &fontConfigEmbolden) == FcResultMatch)
            m_syntheticBold = fontConfigEmbolden;
    }

    int spacing;
    if (FcPatternGetInteger(m_pattern.get(), FC_SPACING, 0, &spacing) == FcResultMatch && spacing == FC_MONO) {
        m_fixedWidth = true;
    }

    // We requested an italic font, but Fontconfig gave us one that was neither oblique nor italic.
    int actualFontSlant;
    bool descriptionAllowsSyntheticOblique = fontDescription.fontSynthesis() & FontSynthesisStyle;
    if (descriptionAllowsSyntheticOblique && fontDescription.italic()
        && FcPatternGetInteger(m_pattern.get(), FC_SLANT, 0, &actualFontSlant) == FcResultMatch) {
        m_syntheticOblique = actualFontSlant == FC_SLANT_ROMAN;
    }

    buildScaledFont(m_fontFace.get());
}


FontPlatformData::FontPlatformData(const FontDescription& fontDescription, bool bold, bool italic)
: m_fontDescription(fontDescription)
, m_syntheticBold(bold)
, m_syntheticOblique(italic)
, m_fixedWidth(false)
, m_spaceWidth(0)
, m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
{
}

FontPlatformData::FontPlatformData(const FontDescription& fontDescription, cairo_font_face_t* fontFace, bool bold, bool italic)
: m_fontFace(fontFace)
, m_fontDescription(fontDescription)
, m_syntheticBold(bold)
, m_syntheticOblique(italic)
, m_fixedWidth(false)
, m_spaceWidth(0)
, m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
{
    buildScaledFont(fontFace);

	if(!m_scaledFont)
		return;

	CairoFtFaceLocker cairoFtFaceLocker(m_scaledFont.get());
	FT_Face freeTypeFace = cairoFtFaceLocker.ftFace();
    if(!freeTypeFace)
    	return;

	m_fixedWidth = freeTypeFace->face_flags & FT_FACE_FLAG_FIXED_WIDTH;
}

FontPlatformData::FontPlatformData(const FontPlatformData& other)
: m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
{
    *this = other;
}

FontPlatformData& FontPlatformData::operator=(const FontPlatformData& other)
{
    // Check for self-assignment.
    if (this == &other)
        return *this;

    clearCachedScaledFonts();

    if(m_cachePruneTimer.isActive())
    	m_cachePruneTimer.stop();

    m_fontDescription = other.m_fontDescription;
    m_syntheticBold = other.m_syntheticBold;
    m_syntheticOblique = other.m_syntheticOblique;
    m_fixedWidth = other.m_fixedWidth;
    m_spaceWidth = other.m_spaceWidth;
    m_charReplaceMap = other.m_charReplaceMap;
    m_familyName = other.m_familyName;
    m_scaledFont = other.m_scaledFont;
    m_pattern = other.m_pattern;
    m_fontFace = other.m_fontFace;

#if USE(HARFBUZZ_NG)
    m_harfbuzzFace = other.m_harfbuzzFace;
#endif

	// This will be re-created on demand.
	m_fallbacks.clear();

    return *this;
}

DLL_PUBLIC FontPlatformData::~FontPlatformData()
{
	fontPlatformDataCache().remove(this);

    clearCachedScaledFonts();
}

FontPlatformData FontPlatformData::smallCapsFontData(const FontDescription& fontDescription) const
{
	return FontPlatformData(fontDescription, m_pattern.get());
}

#if USE(HARFBUZZ_NG)
HarfBuzzNGFace* FontPlatformData::harfbuzzFace() const
{
	if(!m_scaledFont)
		return 0;

    if (!m_harfbuzzFace) {
    	uint64_t uniqueID = reinterpret_cast<uint64_t>(m_scaledFont.get());
        m_harfbuzzFace = HarfBuzzNGFace::create(const_cast<FontPlatformData*>(this), uniqueID);
    }

    return m_harfbuzzFace.get();
}
#endif

DLL_PUBLIC int FontPlatformData::totalCachedScaledFonts()
{
	int total = 0;
	HashSet<FontPlatformData*>& cache = fontPlatformDataCache();

	for(HashSet<FontPlatformData*>::iterator it = cache.begin(); it != cache.end(); ++it) {
		FontPlatformData* data = *it;
		total += data->m_cachedScaledFonts.size();
	}

	return total;
}

void FontPlatformData::cacheScaledFont(cairo_scaled_font_t* scaledFont)
{
    if(!m_cachePruneTimer.isActive())
    	m_cachePruneTimer.startOneShot(0.5);

	for(unsigned int i = 0; i < m_cachedScaledFonts.size(); ++i) {
		if(m_cachedScaledFonts[i].first == scaledFont) {
			m_cachedScaledFonts[i].second = WTF::monotonicallyIncreasingTime();
			return;
		}
	}

	HashSet<FontPlatformData*>& cache = fontPlatformDataCache();

	FontPlatformData* candidateData = 0;
	double candidateTime = 0;
	int candidateIndex = 0;
	int totalCachedScaledFonts = 0;

	for(HashSet<FontPlatformData*>::iterator it = cache.begin(); it != cache.end(); ++it) {
		FontPlatformData* data = *it;
		totalCachedScaledFonts += data->m_cachedScaledFonts.size();

		for(int j = 0; j < data->m_cachedScaledFonts.size(); ++j) {
			if(!candidateData || (candidateTime > data->m_cachedScaledFonts[j].second)) {
				candidateData = data;
				candidateTime = data->m_cachedScaledFonts[j].second;
				candidateIndex = j;
			}
		}
	}

	if(candidateData && totalCachedScaledFonts >= 128) {
		cairo_scaled_font_destroy(candidateData->m_cachedScaledFonts[candidateIndex].first);
		candidateData->m_cachedScaledFonts.remove(candidateIndex);
	}

	m_cachedScaledFonts.append(std::pair<cairo_scaled_font_t*, double>(cairo_scaled_font_reference(scaledFont), WTF::monotonicallyIncreasingTime()));

	if(!cache.contains(this))
		cache.add(this);
}

void FontPlatformData::cachePruneTimerCallback(Timer<FontPlatformData>*)
{
	double now = WTF::monotonicallyIncreasingTime();

	for(int i = m_cachedScaledFonts.size() - 1; i >= 0; --i) {
		if((now - m_cachedScaledFonts[i].second) > 30) { // not used in 30 seconds
			cairo_scaled_font_destroy(m_cachedScaledFonts[i].first);
			m_cachedScaledFonts.remove(i);
		}
	}

	if(m_cachedScaledFonts.size() > 0)
		m_cachePruneTimer.startOneShot(0.5);
	else
		fontPlatformDataCache().remove(this);
}

void FontPlatformData::clearCachedScaledFonts()
{
    if(m_cachedScaledFonts.size() == 0)
    	return;

	for(unsigned int i = 0; i < m_cachedScaledFonts.size(); ++i)
		cairo_scaled_font_destroy(m_cachedScaledFonts[i].first);

	m_cachedScaledFonts.clear();

	fontPlatformDataCache().remove(this);
}

void FontPlatformData::buildScaledFont(cairo_font_face_t* fontFace)
{
	if(size() <= 0)
		return;

	cairo_font_options_t* options = getDefaultCairoFontOptions();
    FcPattern* optionsPattern = m_pattern ? m_pattern.get() : getDefaultFontconfigOptions();
    setCairoFontOptionsFromFontConfigPattern(options, optionsPattern);

	cairo_matrix_t ctm;
	cairo_matrix_init_identity(&ctm);

    // FontConfig may return a list of transformation matrices with the pattern, for instance,
    // for fonts that are oblique. We use that to initialize the cairo font matrix.
    cairo_matrix_t fontMatrix;
    FcMatrix fontConfigMatrix, *tempFontConfigMatrix;
    FcMatrixInit(&fontConfigMatrix);

    // These matrices may be stacked in the pattern, so it's our job to get them all and multiply them.
    for (int i = 0; FcPatternGetMatrix(optionsPattern, FC_MATRIX, i, &tempFontConfigMatrix) == FcResultMatch; i++)
        FcMatrixMultiply(&fontConfigMatrix, &fontConfigMatrix, tempFontConfigMatrix);

    cairo_matrix_init(&fontMatrix, fontConfigMatrix.xx, -fontConfigMatrix.yx,
        -fontConfigMatrix.xy, fontConfigMatrix.yy, 0, 0);

    // The matrix from FontConfig does not include the scale. Scaling a font with width zero size leads
    // to a failed cairo_scaled_font_t instantiations. Instead we scale we scale the font to a very tiny
    // size and just abort rendering later on.
    float realSize = size() ? size() : 1;
    cairo_matrix_scale(&fontMatrix, realSize, realSize);

    if(m_fontDescription.computedHorizontalSize() > 0) {
    	if(m_fontDescription.computedHorizontalSize() != size()) {
    		if(orientation() == Vertical)
        		cairo_matrix_scale(&fontMatrix, 1.0, m_fontDescription.computedHorizontalSize() / size());
    		else
    			cairo_matrix_scale(&fontMatrix, m_fontDescription.computedHorizontalSize() / size(), 1.0);
    	}
    }

    if (m_syntheticOblique) {
        static const float syntheticObliqueSkew = -tanf(14 * acosf(0) / 90);
        static const cairo_matrix_t skew = {1, 0, syntheticObliqueSkew, 1, 0, 0};
		static const cairo_matrix_t verticalSkew = {1, -syntheticObliqueSkew, 0, 1, 0, syntheticObliqueSkew};
		cairo_matrix_multiply(&fontMatrix, orientation() == Vertical ? &verticalSkew : &skew, &fontMatrix);
	}

    if (m_fontDescription.orientation() == Vertical) {
        // The resulting transformation matrix for vertical glyphs (V) is a
        // combination of rotation (R) and translation (T) applied on the
        // horizontal matrix (H). V = H . R . T, where R rotates by -90 degrees
        // and T translates by font size towards y axis.
        cairo_matrix_rotate(&fontMatrix, -piOverTwoDouble);
        cairo_matrix_translate(&fontMatrix, 0.0, 1.0);
    }

	// for BML
	//cairo_font_options_set_hint_style(options, CAIRO_HINT_STYLE_NONE);
	//cairo_font_options_set_hint_metrics(options, CAIRO_HINT_METRICS_OFF);

#if (CAIRO_HAS_GLESV2_SURFACE || CAIRO_HAS_GL_SURFACE)
	if(sFontAntialiased)
		cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_GOOD);
#endif

    m_scaledFont = adoptRef(cairo_scaled_font_create(fontFace, &fontMatrix, &ctm, options));
	cairo_font_options_destroy(options);
}

bool FontPlatformData::init()
{
    static bool initialized = false;
    if (initialized)
        return true;

    initialized = true;

    if (!FcInit())
        return false;

    return true;
}

static void addCharactersToFontConfigPattern(const UChar* characters, int length, FcPattern* pattern)
{
	FcCharSet* fontConfigCharSet = FcCharSetCreate();
	for (int i = 0; i < length; ++i) {
		if (U16_IS_SURROGATE(characters[i]) && U16_IS_SURROGATE_LEAD(characters[i])
				&& i != length - 1 && U16_IS_TRAIL(characters[i + 1])) {
			FcCharSetAddChar(fontConfigCharSet, U16_GET_SUPPLEMENTARY(characters[i], characters[i+1]));
			i++;
		} else
			FcCharSetAddChar(fontConfigCharSet, characters[i]);
	}
	FcPatternAddCharSet(pattern, FC_CHARSET, fontConfigCharSet);
	FcCharSetDestroy(fontConfigCharSet);
}

static RefPtr<FcPattern> findBestFontGivenFallbacks(const FontPlatformData& fontData, FcPattern* pattern, FcResult& fontConfigResult)
{
    if (!fontData.fcPattern())
        return nullptr;

    if (!fontData.m_fallbacks) {
        FcResult fontConfigResult;
        fontData.m_fallbacks = adoptPtr(FcFontSort(0, fontData.fcPattern(), FcTrue, 0, &fontConfigResult));
    }

    if (!fontData.m_fallbacks)
        return nullptr;

    FcFontSet* sets[] = { fontData.m_fallbacks.get() };
    return adoptRef(FcFontSetMatch(0, sets, 1, pattern, &fontConfigResult));
}

PassRefPtr<SimpleFontData> FontCache::systemFallbackForCharacters(const FontDescription& description, const SimpleFontData* originalFontData, bool, const UChar* characters, int length)
{
	if(m_perCharactersPattern) {
		addCharactersToFontConfigPattern(characters, length, m_perCharactersPattern.get());
	}
	else {
		m_perCharactersPattern = adoptRef(FcPatternCreate());

		addCharactersToFontConfigPattern(characters, length, m_perCharactersPattern.get());
	    FcPatternAddBool(m_perCharactersPattern.get(), FC_SCALABLE, FcTrue);

	    // the following two functions are very slow
	    FcConfigSubstitute(0, m_perCharactersPattern.get(), FcMatchPattern);
	    FcDefaultSubstitute(m_perCharactersPattern.get());
	}

	FcResult fontConfigResult;
    if(RefPtr<FcPattern> fallbackPattern = findBestFontGivenFallbacks(originalFontData->platformData(), m_perCharactersPattern.get(), fontConfigResult)) {
    	FcChar8* fontConfigFamilyNameAfterMatching;
		FcPatternGetString(fallbackPattern.get(), FC_FAMILY, 0, &fontConfigFamilyNameAfterMatching);
		String familyNameAfterMatching = String::fromUTF8(reinterpret_cast<char*>(fontConfigFamilyNameAfterMatching));
		if(RefPtr<SimpleFontData> simpleFontData = fontForFamily(description, familyNameAfterMatching)) {
			FcPatternDel(m_perCharactersPattern.get(), FC_CHARSET);
			return simpleFontData.release();
		}
    }

	if(RefPtr<FcPattern> resultPattern = adoptRef(FcFontMatch(0, m_perCharactersPattern.get(), &fontConfigResult))) {
		FcChar8* fontConfigFamilyNameAfterMatching;
		FcPatternGetString(resultPattern.get(), FC_FAMILY, 0, &fontConfigFamilyNameAfterMatching);
		String familyNameAfterMatching = String::fromUTF8(reinterpret_cast<char*>(fontConfigFamilyNameAfterMatching));
		if(RefPtr<SimpleFontData> simpleFontData = fontForFamily(description, familyNameAfterMatching)) {
			FcPatternDel(m_perCharactersPattern.get(), FC_CHARSET);
			return simpleFontData.release();
		}
	}

    FcPatternDel(m_perCharactersPattern.get(), FC_CHARSET);
    return nullptr;
}

DLL_PUBLIC String getFamilyNameStringFromFontDescriptionAndFamily(const FontDescription& fontDescription, const AtomicString& family)
{
    // If we're creating a fallback font (e.g. "-webkit-monospace"), convert the name into
    // the fallback name (like "monospace") that fontconfig understands.
    if (family.length() && !family.startsWith("-webkit-"))
        return family.string();

    if (family == standardFamily || family == serifFamily)
        return "serif";
    if (family == sansSerifFamily)
        return "sans-serif";
    if (family == monospaceFamily)
        return "monospace";
    if (family == cursiveFamily)
        return "cursive";
    if (family == fantasyFamily)
        return "fantasy";
    return "";
}

#if ENABLE(OPENTYPE_VERTICAL)
namespace {
typedef HashMap<FT_Face, RefPtr<OpenTypeVerticalData>> FontVerticalDataCache;

static FontVerticalDataCache& fontVerticalDataCacheInstance()
{
    DEFINE_STATIC_LOCAL(FontVerticalDataCache, fontVerticalDataCache, ());
    return fontVerticalDataCache;
}

static void releaseVerticalData(void* data)
{
    FontVerticalDataCache& fontVerticalDataCache = fontVerticalDataCacheInstance();
    FontVerticalDataCache::iterator iter = fontVerticalDataCache.find(reinterpret_cast<FT_Face>(data));
    if (iter != fontVerticalDataCache.end())
    	fontVerticalDataCache.remove(iter);
}

static inline PassRefPtr<OpenTypeVerticalData> getVerticalData(cairo_font_face_t* fontFace, const FontPlatformData& platformData)
{
	cairo_scaled_font_t* scaledFont = platformData.scaledFont();
	if(!scaledFont || !fontFace)
		return nullptr;

	CairoFtFaceLocker cairoFtFaceLocker(scaledFont);
	FT_Face freeTypeFace = cairoFtFaceLocker.ftFace();
	if(!freeTypeFace)
		return nullptr;

	FontVerticalDataCache& fontVerticalDataCache = fontVerticalDataCacheInstance();
	FontVerticalDataCache::iterator result = fontVerticalDataCache.find(freeTypeFace);
	if (result != fontVerticalDataCache.end())
		return result.get()->second;

	RefPtr<OpenTypeVerticalData> verticalData = OpenTypeVerticalData::create(platformData);
	if (!verticalData->isOpenType())
		verticalData.clear();
	fontVerticalDataCache.set(freeTypeFace, verticalData);

	static cairo_user_data_key_t verticalDataKey;
	cairo_font_face_set_user_data(fontFace, &verticalDataKey, freeTypeFace, releaseVerticalData);

	return verticalData.release();
}
}
#endif

DLL_PUBLIC unsigned verticalDataCount()
{
	return fontVerticalDataCacheInstance().size();
}

PassRefPtr<OpenTypeVerticalData> FontPlatformData::verticalData() const
{
	return getVerticalData(m_fontFace.get(), *this);
}

PassRefPtr<SharedBuffer> FontPlatformData::openTypeTable(uint32_t table) const
{
	if(!m_scaledFont)
		return nullptr;

	CairoFtFaceLocker cairoFtFaceLocker(m_scaledFont.get());
	FT_Face freeTypeFace = cairoFtFaceLocker.ftFace();
    if(!freeTypeFace)
    	return nullptr;

	FT_ULong tableSize = 0;
	// Tag bytes need to be reversed because OT_MAKE_TAG uses big-endian order.
	uint32_t tag = FT_MAKE_TAG((table & 0xff), (table & 0xff00) >> 8, (table & 0xff0000) >> 16, table >> 24);
	if (FT_Load_Sfnt_Table(freeTypeFace, tag, 0, 0, &tableSize))
		return nullptr;

	RefPtr<SharedBuffer> buffer = SharedBuffer::create(tableSize);
	FT_ULong expectedTableSize = tableSize;
	if (buffer->size() != tableSize)
		return nullptr;

	FT_Error error = FT_Load_Sfnt_Table(freeTypeFace, tag, 0, reinterpret_cast<FT_Byte*>(const_cast<char*>(buffer->data())), &tableSize);
	if (error || tableSize != expectedTableSize)
		return nullptr;

    return buffer.release();
}

void FontPlatformData::setOrientation(FontOrientation orientation)
{
    if (!m_scaledFont || (m_fontDescription.orientation() == orientation))
        return;

    ASSERT(m_scaledFont);
    m_fontDescription.setOrientation(orientation);
    buildScaledFont(cairo_scaled_font_get_font_face(m_scaledFont.get()));
}

void FontPlatformData::setSyntheticOblique(bool newSyntheticObliqueValue)
{
    if (!m_scaledFont || newSyntheticObliqueValue == syntheticOblique())
        return;

    ASSERT(m_scaledFont);
    m_syntheticOblique = newSyntheticObliqueValue;
    buildScaledFont(cairo_scaled_font_get_font_face(m_scaledFont.get()));
}

}
