/*
 * Copyright (c) 2012 Google Inc. All rights reserved.
 * Copyright (c) 2012 Intel Corporation
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "HarfBuzzNGFace.h"

#include "CairoUtilities.h"
#include "FontCache.h"
#include "FontPlatformData.h"
#include "GlyphBuffer.h"
#include "HarfBuzzShaper.h"
#include "SimpleFontData.h"
#include "TextEncoding.h"
#include <cairo-ft.h>
#include <cairo.h>
#include <freetype/freetype.h>
#include <freetype/tttables.h>
#include <hb.h>
#include <wtf/text/CString.h>
#include <wtf/text/StringView.h>

namespace WebCore {

struct HarfBuzzFontData {
    HarfBuzzFontData(WTF::HashMap<uint32_t, uint32_t>* glyphCacheForFaceCacheEntry, cairo_scaled_font_t* cairoScaledFont, FontCharReplaceMap& charReplaceMap)
        : m_glyphCacheForFaceCacheEntry(glyphCacheForFaceCacheEntry)
        , m_cairoScaledFont(cairoScaledFont)
    	, m_charReplaceMap(charReplaceMap)
    { }

    WTF::HashMap<uint32_t, uint32_t>* m_glyphCacheForFaceCacheEntry;
    cairo_scaled_font_t* m_cairoScaledFont;
    FontCharReplaceMap m_charReplaceMap;
};

static hb_position_t floatToHarfBuzzPosition(float value)
{
    return static_cast<hb_position_t>(value * (1 << 16));
}

static hb_position_t doubleToHarfBuzzPosition(double value)
{
    return static_cast<hb_position_t>(value * (1 << 16));
}

static void CairoGetGlyphWidthAndExtents(cairo_scaled_font_t* scaledFont, hb_codepoint_t codepoint, hb_position_t* advance, hb_glyph_extents_t* extents)
{
    cairo_text_extents_t glyphExtents;
    cairo_glyph_t glyph;
    glyph.index = codepoint;
    glyph.x = 0;
    glyph.y = 0;
    cairo_scaled_font_glyph_extents(scaledFont, &glyph, 1, &glyphExtents);

    bool hasVerticalGlyphs = glyphExtents.y_advance;
    if (advance)
        *advance = doubleToHarfBuzzPosition(hasVerticalGlyphs ? -glyphExtents.y_advance : glyphExtents.x_advance);

    if (extents) {
        extents->x_bearing = doubleToHarfBuzzPosition(glyphExtents.x_bearing);
        extents->y_bearing = doubleToHarfBuzzPosition(hasVerticalGlyphs ? -glyphExtents.y_bearing : glyphExtents.y_bearing);
        extents->width = doubleToHarfBuzzPosition(hasVerticalGlyphs ? -glyphExtents.height : glyphExtents.width);
        extents->height = doubleToHarfBuzzPosition(hasVerticalGlyphs ? glyphExtents.width : glyphExtents.height);
    }
}

static hb_bool_t harfbuzzGetNominalGlyph(hb_font_t* hbFont, void* fontData, hb_codepoint_t unicode, hb_codepoint_t* glyph, void* userData)
{
    HarfBuzzFontData* hbFontData = reinterpret_cast<HarfBuzzFontData*>(fontData);
    cairo_scaled_font_t* scaledFont = hbFontData->m_cairoScaledFont;
    ASSERT(scaledFont);

    auto result = hbFontData->m_glyphCacheForFaceCacheEntry->add(unicode, 0);
    if (result.isNewEntry) {
        CairoFtFaceLocker cairoFtFaceLocker(scaledFont);
        FT_Face ftFont = cairoFtFaceLocker.ftFace();
        if (UNLIKELY(!ftFont))
            return false;

    	// look for per-font replacement character
    	FontCharReplaceMap::const_iterator it = hbFontData->m_charReplaceMap.find(unicode);
    	if(UNLIKELY(it != hbFontData->m_charReplaceMap.end()))
    		unicode = it->second;

		FT_UInt glyphindex = FT_Get_Char_Index(ftFont, unicode);

		if (UNLIKELY(!glyphindex)) {
			// try similar character
	        FontCharReplaceMap::const_iterator it = fontCache().similarCharacters().find(unicode);
			if(it != fontCache().similarCharacters().end()) {
	    		unicode = it->second;
				glyphindex = FT_Get_Char_Index(ftFont, unicode);
				if(!glyphindex)
					return false;
			}
			else
				return false;
		}

		result.iterator->second = glyphindex;
    }
    *glyph = result.iterator->second;
    return !!*glyph;
}

static hb_bool_t harfbuzzGetVariationGlyph(hb_font_t *hbFont, void *fontData, hb_codepoint_t unicode, hb_codepoint_t variation_selector, hb_codepoint_t *glyph, void *userData)
{
    HarfBuzzFontData* hbFontData = reinterpret_cast<HarfBuzzFontData*>(fontData);
    cairo_scaled_font_t* scaledFont = hbFontData->m_cairoScaledFont;
    ASSERT(scaledFont);

    CairoFtFaceLocker cairoFtFaceLocker(scaledFont);
    FT_Face ftFont = cairoFtFaceLocker.ftFace();
    if (UNLIKELY(!ftFont))
        return false;

	// look for per-font replacement character
	FontCharReplaceMap::const_iterator it = hbFontData->m_charReplaceMap.find(unicode);
	if(UNLIKELY(it != hbFontData->m_charReplaceMap.end()))
		unicode = it->second;

	unsigned int g = FT_Face_GetCharVariantIndex(ftFont, unicode, variation_selector);

	if (UNLIKELY(!g)) {
		// try similar character
        FontCharReplaceMap::const_iterator it = fontCache().similarCharacters().find(unicode);
		if(it != fontCache().similarCharacters().end()) {
    		unicode = it->second;
    		g = FT_Face_GetCharVariantIndex(ftFont, unicode, variation_selector);
    		if (!g)
    			return false;
		}
		else
			return false;
	}

	*glyph = g;
	return true;
}

static hb_position_t harfbuzzGetGlyphHorizontalAdvance(hb_font_t* hbFont, void* fontData, hb_codepoint_t glyph, void* userData)
{
    HarfBuzzFontData* hbFontData = reinterpret_cast<HarfBuzzFontData*>(fontData);
    cairo_scaled_font_t* scaledFont = hbFontData->m_cairoScaledFont;
    ASSERT(scaledFont);

    hb_position_t advance = 0;
    CairoGetGlyphWidthAndExtents(scaledFont, glyph, &advance, 0);
    return advance;
}

static hb_bool_t harfbuzzGetGlyphHorizontalOrigin(hb_font_t* hbFont, void* fontData, hb_codepoint_t glyph, hb_position_t* x, hb_position_t* y, void* userData)
{
    // Just return true, following the way that Harfbuzz-FreeType
    // implementation does.
    return true;
}

static hb_bool_t harfbuzzGetGlyphExtents(hb_font_t* hbFont, void* fontData, hb_codepoint_t glyph, hb_glyph_extents_t* extents, void* userData)
{
    HarfBuzzFontData* hbFontData = reinterpret_cast<HarfBuzzFontData*>(fontData);
    cairo_scaled_font_t* scaledFont = hbFontData->m_cairoScaledFont;
    ASSERT(scaledFont);

    CairoGetGlyphWidthAndExtents(scaledFont, glyph, 0, extents);
    return true;
}

static hb_font_funcs_t* harfbuzzCairoTextGetFontFuncs()
{
    static hb_font_funcs_t* harfbuzzCairoFontFuncs = 0;

    // We don't set callback functions which we can't support.
    // Harfbuzz will use the fallback implementation if they aren't set.
    if (!harfbuzzCairoFontFuncs) {
        harfbuzzCairoFontFuncs = hb_font_funcs_create();
        hb_font_funcs_set_nominal_glyph_func(harfbuzzCairoFontFuncs, harfbuzzGetNominalGlyph, 0, 0);
        hb_font_funcs_set_variation_glyph_func(harfbuzzCairoFontFuncs, harfbuzzGetVariationGlyph, 0, 0);
        hb_font_funcs_set_glyph_h_advance_func(harfbuzzCairoFontFuncs, harfbuzzGetGlyphHorizontalAdvance, 0, 0);
        hb_font_funcs_set_glyph_h_origin_func(harfbuzzCairoFontFuncs, harfbuzzGetGlyphHorizontalOrigin, 0, 0);
        hb_font_funcs_set_glyph_extents_func(harfbuzzCairoFontFuncs, harfbuzzGetGlyphExtents, 0, 0);
        hb_font_funcs_make_immutable(harfbuzzCairoFontFuncs);
    }
    return harfbuzzCairoFontFuncs;
}

static const FT_Encoding fcFontEncodings[] = {
    FT_ENCODING_UNICODE,
    FT_ENCODING_MS_SYMBOL
};

#define NUM_DECODE  (int) (sizeof (fcFontEncodings) / sizeof (fcFontEncodings[0]))

static hb_blob_t* harfbuzzCairoGetTable(hb_face_t* face, hb_tag_t tag, void* userData)
{
    cairo_scaled_font_t* scaledFont = reinterpret_cast<cairo_scaled_font_t*>(userData);
    if (!scaledFont)
        return 0;

    CairoFtFaceLocker cairoFtFaceLocker(scaledFont);
    FT_Face ftFont = cairoFtFaceLocker.ftFace();
    if (!ftFont)
        return 0;

    FT_ULong tableSize = 0;
    FT_Error error = FT_Load_Sfnt_Table(ftFont, tag, 0, 0, &tableSize);
    if (error)
        return 0;

    FT_Byte* buffer = reinterpret_cast<FT_Byte*>(fastMalloc(tableSize));
    if (!buffer)
        return 0;
    FT_ULong expectedTableSize = tableSize;
    error = FT_Load_Sfnt_Table(ftFont, tag, 0, buffer, &tableSize);
    if (error || tableSize != expectedTableSize) {
        fastFree(buffer);
        return 0;
    }

    return hb_blob_create(reinterpret_cast<const char*>(buffer), tableSize, HB_MEMORY_MODE_WRITABLE, buffer, fastFree);
}

static void destroyHarfBuzzFontData(void* userData)
{
    HarfBuzzFontData* hbFontData = reinterpret_cast<HarfBuzzFontData*>(userData);
    delete hbFontData;
}

hb_face_t* HarfBuzzNGFace::createFace()
{
    hb_face_t* face = hb_face_create_for_tables(harfbuzzCairoGetTable, m_scaledFont.get(), 0);
    ASSERT(face);
    return face;
}

hb_font_t* HarfBuzzNGFace::createFont()
{
    hb_font_t* font = hb_font_create(m_face);
    HarfBuzzFontData* hbFontData = new HarfBuzzFontData(m_glyphCacheForFaceCacheEntry, m_scaledFont.get(), m_charReplaceMap);
	hb_font_set_funcs(font, harfbuzzCairoTextGetFontFuncs(), hbFontData, destroyHarfBuzzFontData);
	const float size = m_size;
    if (floorf(size) == size)
        hb_font_set_ppem(font, size, size);
    int scale = floatToHarfBuzzPosition(size);
    hb_font_set_scale(font, scale, scale);
    hb_font_make_immutable(font);

	CairoFtFaceLocker cairoFtFaceLocker(m_scaledFont.get());
	if(FT_Face ftFont = cairoFtFaceLocker.ftFace()) {
		for(int i = 0; i < NUM_DECODE; ++i) {
			if(ftFont->charmap && ftFont->charmap->encoding == fcFontEncodings[i])
				break;

			if (FT_Select_Charmap(ftFont, fcFontEncodings[i]))
				break;
		}
	}

    return font;
}

GlyphBufferAdvance HarfBuzzShaper::createGlyphBufferAdvance(float width, float height)
{
    return width;
}

} // namespace WebCore
