/*
 * Copyright (C) 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Google Inc. All rights reserved.
 * Copyright (C) 2011 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"

#if ENABLE(WEBGL)

#include "GraphicsContext.h"
#include "GraphicsContext3D.h"
#include "CanvasRenderingContext.h"
#include "Image.h"
#include <wtf/PassOwnPtr.h>
#include <stdio.h>
#include "BitmapImage.h"
#include "ImageDecoder.h"
#include "TiledImageOpenVG.h"
#include "VGUtils.h"

namespace WebCore {

bool GraphicsContext3D::getImageData(Image* image, unsigned int format, unsigned int type, bool premultiplyAlpha, bool ignoreGammaAndColorProfile, Vector<uint8_t>& outputVector)
{
    if (!image)
        return false;

    // if there is original (compressed) image data, it's way faster to decode it again
    // than to fetch the decoded data from video memory (since this stalls the pipeline.)
    if (image->data()) {

        OwnPtr<ImageDecoder> decoder = adoptPtr(ImageDecoder::create(*image->data(), "", NULL,
                    premultiplyAlpha ? ImageSource::AlphaPremultiplied : ImageSource::AlphaNotPremultiplied,
                    ignoreGammaAndColorProfile ? ImageSource::GammaAndColorProfileIgnored : ImageSource::GammaAndColorProfileApplied));
        if (!decoder)
            return false;
        decoder->setData(image->data(), true);

        ImageFrame* frame = decoder->frameBufferAtIndex(0);
        if (!frame || frame->status() == ImageFrame::FrameEmpty)
            return 0;
        IntSize size = decoder->size();
        int width = size.width();
        int height = size.height();

        int bpp = 4;
        if (!outputVector.tryReserveCapacity(width * height * bpp))
            return false;
        outputVector.resize(width * height * bpp);
        bool ret = packPixels(reinterpret_cast<uint8_t*>(frame->getAddr(0,0)), SourceFormatBGRA8,
                        width, height, 1, format, type, AlphaDoNothing, outputVector.data());
        return ret;

    }

    // if we have no data, then we have to vgGetImageSubData() from the image
    // tiles in video memory.
    
    SurfaceOpenVG::makeCompatibleCurrent();
    RefPtr<TiledImageOpenVG> tiledImage = image->nativeImageForCurrentFrame();
    if (!tiledImage)
        return false;

    AlphaOp alphaOp = AlphaDoNothing;
    if (!premultiplyAlpha)
        alphaOp = AlphaDoUnmultiply;

    int bpp = 4;
    int width = tiledImage->size().width();
    int height = tiledImage->size().height();
    const int numColumns = tiledImage->numColumns();
    const int numRows = tiledImage->numRows();

    Vector<uint8_t> sourceBytes;
    if (!sourceBytes.tryReserveCapacity(width * height * bpp))
        return false;
    sourceBytes.resize(width * height * bpp);
    if (!outputVector.tryReserveCapacity(width * height * bpp))
        return false;
    outputVector.resize(width * height * bpp);

    for (int yIndex = 0; yIndex < numRows; ++yIndex) {
        for (int xIndex = 0; xIndex < numColumns; ++xIndex) {
            IntRect tileRect = tiledImage->tileRect(xIndex, yIndex);
            VGImage tile = tiledImage->tile(xIndex, yIndex);

            const int bpp = 4;
            vgGetImageSubData(tile,
                sourceBytes.data() + ((tileRect.y() * width) + tileRect.x()) * bpp,
                width * bpp, VG_sARGB_8888,
                0, 0, tileRect.width(), tileRect.height());
            ASSERT_VG_NO_ERROR();
        }
    }

    return packPixels(sourceBytes.data(), SourceFormatBGRA8,
                      width, height, 1, format, type, alphaOp, outputVector.data());
}

bool GraphicsContext3D::getImageData(ImageBuffer* imageBuffer, unsigned int format, unsigned int type, bool premultiplyAlpha, bool ignoreGammaAndColorProfile, Vector<uint8_t>& outputVector)
{
	return false;
}

void GraphicsContext3D::paintToCanvas(const unsigned char* imagePixels, int imageWidth, int imageHeight, int canvasWidth, int canvasHeight, GraphicsContext* context)
{
    // TODO: this is currently broken. we might not need it anyway
    // since canvas is composited.
#if 0
    if (!imagePixels || imageWidth <= 0 || imageHeight <= 0 || canvasWidth <= 0 || canvasHeight <= 0 || !context)
        return;
    
    SurfaceOpenVG::makeCompatibleCurrent();

    // some duplication here with RGBA32Buffer::asNewNativeImage()
    
    const IntSize vgMaxImageSize(vgGeti(VG_MAX_IMAGE_WIDTH), vgGeti(VG_MAX_IMAGE_HEIGHT));
    ASSERT_VG_NO_ERROR();
    RefPtr<TiledImageOpenVG> tiledImage = adoptRef(new TiledImageOpenVG(
        IntSize(imageWidth, imageHeight), vgMaxImageSize));

    const int numColumns = tiledImage->numColumns();
    const int numRows = tiledImage->numRows();

    for (int yIndex = 0; yIndex < numRows; ++yIndex) {
        for (int xIndex = 0; xIndex < numColumns; ++xIndex) {
            IntRect tileRect = tiledImage->tileRect(xIndex, yIndex);
            VGImage tile = vgCreateImage(VG_sARGB_8888_PRE,
                tileRect.width(), tileRect.height(), VG_IMAGE_QUALITY_FASTER);
            ASSERT_VG_NO_ERROR();

            const int bpp = 4;
            vgImageSubData(tile,
                const_cast<unsigned char*>(imagePixels + ((tileRect.y() * imageWidth) + tileRect.x()) * bpp),
                imageWidth * bpp, VG_sARGB_8888_PRE,
                0, 0, tileRect.width(), tileRect.height());
            ASSERT_VG_NO_ERROR();

            tiledImage->setTile(xIndex, yIndex, tile);
        }
    }

    RefPtr<Image> image = BitmapImage::create(tiledImage);
    context->save();

    // OpenGL keeps the pixels stored bottom up, so we need to flip the image here.
    context->translate(0, imageHeight);
    context->scale(FloatSize(1.0f, -1.0f));

    context->clearRect(FloatRect(0, 0, canvasWidth, canvasHeight));
    context->drawImage(image.get(), DeviceColorSpace, IntRect(0, 0, canvasWidth, canvasHeight));
    context->restore();
#endif
}

} // namespace WebCore

#endif // ENABLE(WEBGL)

