/*
 * Copyright (C) Research In Motion Limited 2009-2010. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "GraphicsContext.h"

#include "AffineTransform.h"
#include "Gradient.h"
#include "KURL.h"
#include "NotImplemented.h"
#include "PaintOpenVG.h"
#include "PainterOpenVG.h"
#include "PlatformPathOpenVG.h"
#include "PatternOpenVG.h"
#include "SurfaceOpenVG.h"
#include "TransformationMatrix.h"

#include <wtf/Assertions.h>
#include <wtf/MathExtras.h>
#include <wtf/UnusedParam.h>
#include <wtf/Vector.h>


namespace WebCore {

class GraphicsContextPlatformPrivate : public PainterOpenVG {
public:
    GraphicsContextPlatformPrivate(SurfaceOpenVG* surface)
        : PainterOpenVG(surface)
    {
    }
};

void GraphicsContext::platformInit(SurfaceOpenVG* surface) {
    m_data = surface ? new GraphicsContextPlatformPrivate(surface) : 0;

    setPaintingDisabled(!surface);

    if (!m_data)
        return;

    // Align the initial settings in platform data
    m_data->setTextDrawingMode(m_state.textDrawingMode);
    m_data->setStrokeStyle(m_state.strokeStyle);
    m_data->setStrokeThickness(m_state.strokeThickness);
    m_data->setStrokeColor(m_state.strokeColor);
    m_data->setFillColor(m_state.fillColor);
}

void GraphicsContext::platformDestroy()
{
    delete m_data;
}

DLL_PUBLIC PlatformGraphicsContext* GraphicsContext::platformContext() const
{
    if (UNLIKELY(paintingDisabled()))
        return 0;

    return m_data->baseSurface();
}

DLL_PUBLIC AffineTransform GraphicsContext::getCTM(IncludeDeviceScale) const
{
    if (UNLIKELY(paintingDisabled()))
        return AffineTransform();

    return m_data->transformation();
}

DLL_PUBLIC void GraphicsContext::setCTM(const AffineTransform& transform)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setTransformation(transform);
}

void GraphicsContext::savePlatformState()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->save();
}

void GraphicsContext::restorePlatformState()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->restore();
}

void GraphicsContext::drawRect(const IntRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawRect(rect);
}

DLL_PUBLIC void GraphicsContext::drawLine(const IntPoint& from, const IntPoint& to)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawLine(from, to);
}

/**
 * Draw the largest ellipse that fits into the given rectangle.
 */
void GraphicsContext::drawEllipse(const IntRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawEllipse(rect);
}

void GraphicsContext::strokeArc(const IntRect& rect, int startAngle, int angleSpan)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawArc(rect, startAngle, angleSpan, VG_STROKE_PATH);
}

void GraphicsContext::drawConvexPolygon(size_t numPoints, const FloatPoint* points, bool shouldAntialias)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawPolygon(numPoints, points);

    UNUSED_PARAM(shouldAntialias); // FIXME
}

void GraphicsContext::fillPath(const Path& path)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawPath(path, VG_FILL_PATH, m_state.fillRule);
    m_data->beginPath();
}

#if 0
void GraphicsContext::fillPath()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawPath(*m_data->currentPath(), VG_FILL_PATH, m_state.fillRule);
    m_data->beginPath();
}
#endif

void GraphicsContext::strokePath(const Path& path)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawPath(path, VG_STROKE_PATH, m_state.fillRule);
    m_data->beginPath();
}

#if 0
void GraphicsContext::strokePath()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawPath(*m_data->currentPath(), VG_STROKE_PATH, m_state.fillRule);
    m_data->beginPath();
}
#endif

#if 0
void GraphicsContext::drawPath()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawPath(*m_data->currentPath(), VG_FILL_PATH | VG_STROKE_PATH, m_state.fillRule);
    m_data->beginPath();
}
#endif

void GraphicsContext::fillRect(const FloatRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawRect(rect, VG_FILL_PATH);
}

DLL_PUBLIC void GraphicsContext::fillRect(const FloatRect& rect, const Color& color, ColorSpace colorSpace)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    PaintOpenVG currentPaint = m_data->fillPaint();
    m_data->setFillColor(color);
    m_data->drawRect(rect, VG_FILL_PATH);
    m_data->setFillPaint(currentPaint);

    UNUSED_PARAM(colorSpace); // FIXME
}

void GraphicsContext::fillRoundedRect(const IntRect& rect, const IntSize& topLeft, const IntSize& topRight, const IntSize& bottomLeft, const IntSize& bottomRight, const Color& color, ColorSpace colorSpace)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    PaintOpenVG currentPaint = m_data->fillPaint();
    m_data->setFillColor(color);
    m_data->drawRoundedRect(rect, topLeft, topRight, bottomLeft, bottomRight, VG_FILL_PATH);
    m_data->setFillPaint(currentPaint);

    UNUSED_PARAM(colorSpace); // FIXME
}

#if 0
void GraphicsContext::beginPath()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->beginPath();
}
#endif

#if 0
void GraphicsContext::addPath(const Path& path)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->addPath(path);
}
#endif

DLL_PUBLIC void GraphicsContext::clip(const FloatRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->clipRect(rect, PainterOpenVG::IntersectClip);
}

void GraphicsContext::clipPath(const Path& path, WindRule clipRule)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->clipPath(path, PainterOpenVG::IntersectClip, clipRule);
}

void GraphicsContext::clipConvexPolygon(size_t numPoints, const FloatPoint* points, bool antialiased)
{
    notImplemented();
}

void GraphicsContext::drawFocusRing(const Vector<IntRect>& rects, int width, int offset, const Color& color)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    if (rects.isEmpty())
        return;

    // FIXME: We should outline the edge of the full region.
    PaintOpenVG currentFillPaint = m_data->fillPaint();
    CompositeOperator currentOp = m_data->compositeOperation();
    m_data->setFillColor(0xffa3c8fe);
    m_data->setCompositeOperation(CompositePlusDarker);

    for (unsigned i = 0; i < rects.size(); i++) {
        IntRect focusRect = rects[i];
        focusRect.inflate(2);
        m_data->drawRect(focusRect, VG_FILL_PATH);
    }
    m_data->setFillPaint(currentFillPaint);
    m_data->setCompositeOperation(currentOp);
}

void GraphicsContext::drawFocusRing(const Path& path, int width, int offset, const Color& color)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    UNUSED_PARAM(width);
    UNUSED_PARAM(offset);

    StrokeStyle currentStyle = m_data->strokeStyle();
    float currentThickness = m_data->strokeThickness();
    PaintOpenVG currentPaint = m_data->strokePaint();

    m_data->setStrokeStyle(SolidStroke);
    m_data->setStrokeThickness(1);
    m_data->setStrokeColor(color);

    m_data->drawPath(path, VG_STROKE_PATH, RULE_NONZERO);

    m_data->setStrokeStyle(currentStyle);
    m_data->setStrokeThickness(currentThickness);
    m_data->setStrokePaint(currentPaint);
}

void GraphicsContext::drawLineForText(const FloatPoint& origin, float width, bool printing)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    if (width <= 0)
        return;

    StrokeStyle oldStyle = m_data->strokeStyle();
    m_data->setStrokeStyle(SolidStroke);
    drawLine(roundedIntPoint(origin), roundedIntPoint(origin) + IntSize(width, 0));
    m_data->setStrokeStyle(oldStyle);

    UNUSED_PARAM(printing);
}

// FIXME:  Style should probably be pulled from the render theme or the paint pushed there.
static const RGBA32 spellingUnderline = 0xFFFF0000;
static const RGBA32 grammarUnderline = 0xFF008000;

void GraphicsContext::drawLineForDocumentMarker(const FloatPoint& origin, float width, DocumentMarkerLineStyle style)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    if (width <= 0)
        return;

    StrokeStyle currentStyle = m_data->strokeStyle();
    PaintOpenVG currentPaint = m_data->strokePaint();
    m_data->setStrokeStyle(DottedStroke);

    switch (style) {
        case DocumentMarkerGrammarLineStyle:
            m_data->setStrokeColor(grammarUnderline);
            break;
        case DocumentMarkerSpellingLineStyle:
        default:
            m_data->setStrokeColor(spellingUnderline);
            break;
    }

    drawLine(roundedIntPoint(origin), roundedIntPoint(origin) + IntSize(width, 0));

    m_data->setStrokeStyle(currentStyle);
    m_data->setStrokePaint(currentPaint);
}

FloatRect GraphicsContext::roundToDevicePixels(const FloatRect& rect, RoundingMode)
{
    if (UNLIKELY(paintingDisabled()))
        return FloatRect();

    return FloatRect(enclosingIntRect(m_data->transformation().mapRect(rect)));
}

void GraphicsContext::setPlatformShadow(const FloatSize& size, float blur, const Color& color, ColorSpace colorSpace)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    // FIXME: This still does not address the issue that shadows
    // within canvas elements should ignore transforms.
    if (m_state.shadowsIgnoreTransforms)  {
        // Currently only the GraphicsContext associated with the
        // CanvasRenderingContext for HTMLCanvasElement have shadows ignore
        // Transforms. So with this flag set, we know this state is associated
        // with a CanvasRenderingContext. CG uses natural orientation for
        // Y axis, but the HTML5 canvas spec does not. So we now flip the
        // height since it was flipped in CanvasRenderingContext in order to
        // work with CG.
        m_data->setShadow(IntSize(size.width(), -size.height()), blur, color);
    } else
        m_data->setShadow(IntSize(size.width(), size.height()), blur, color);

    UNUSED_PARAM(colorSpace); // FIXME
}

void GraphicsContext::clearPlatformShadow()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->clearShadow();
}

bool GraphicsContext::supportsTransparencyLayers()
{
    return true;
}

void GraphicsContext::beginPlatformTransparencyLayer(float opacity)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->beginTransparencyLayer(opacity);
}

void GraphicsContext::endPlatformTransparencyLayer()
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->endTransparencyLayer();
}

DLL_PUBLIC void GraphicsContext::clearRect(const FloatRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    CompositeOperator op = m_data->compositeOperation();
    float opacity = m_data->opacity();
    m_data->setCompositeOperation(CompositeClear);
    m_data->setOpacity(1.0);
    m_data->drawRect(rect, VG_FILL_PATH);
    m_data->setOpacity(opacity);
    m_data->setCompositeOperation(op);
}

#if 0
void GraphicsContext::strokeRect(const FloatRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->drawRect(rect, VG_STROKE_PATH);
}
#endif

DLL_PUBLIC void GraphicsContext::strokeRect(const FloatRect& rect, float lineWidth)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    float oldThickness = m_data->strokeThickness();
    m_data->setStrokeThickness(lineWidth);
    m_data->drawRect(rect, VG_STROKE_PATH);
    m_data->setStrokeThickness(oldThickness);
}

void GraphicsContext::setLineCap(LineCap lc)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setLineCap(lc);
}

void GraphicsContext::setLineDash(const DashArray& dashes, float dashOffset)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    // SVG code sets an empty line dash for solid strokes, don't rely on
    // painter internals to get that one right.
    if (dashes.isEmpty()) {
        m_data->setStrokeStyle(SolidStroke);
        return;
    }
    m_data->setLineDash(dashes, dashOffset);
    m_data->setStrokeStyle(DashedStroke);
}

void GraphicsContext::setLineJoin(LineJoin lj)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setLineJoin(lj);
}

void GraphicsContext::setMiterLimit(float limit)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setMiterLimit(limit);
}

void GraphicsContext::setAlpha(float opacity)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setOpacity(opacity);
}

void GraphicsContext::setPlatformCompositeOperation(CompositeOperator op, BlendMode)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setCompositeOperation(op);
}

void GraphicsContext::clip(const Path& path)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->clipPath(path, PainterOpenVG::IntersectClip, m_state.fillRule);
}

void GraphicsContext::canvasClip(const Path& path)
{
    clip(path);
}

void GraphicsContext::clipOut(const Path& path)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->clipPath(path, PainterOpenVG::SubtractClip, m_state.fillRule);
}

DLL_PUBLIC void GraphicsContext::scale(const FloatSize& scaleFactors)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->scale(scaleFactors);
}

void GraphicsContext::rotate(float radians)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->rotate(radians);
}

DLL_PUBLIC void GraphicsContext::translate(float dx, float dy)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->translate(dx, dy);
}

#if 0
IntPoint GraphicsContext::origin()
{
    if (UNLIKELY(paintingDisabled()))
        return IntPoint();

    AffineTransform transformation = m_data->transformation();
    return IntPoint(roundf(transformation.e()), roundf(transformation.f()));
}
#endif

void GraphicsContext::clipOut(const IntRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->clipRect(rect, PainterOpenVG::SubtractClip);
}

#if 0
void GraphicsContext::clipOutEllipseInRect(const IntRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    Path path;
    path.addEllipse(rect);
    m_data->clipPath(path, PainterOpenVG::SubtractClip, m_state.fillRule);
}
#endif

#if 0
DLL_PUBLIC void GraphicsContext::clipToImageBuffer(const FloatRect& rect, const ImageBuffer* imageBuffer)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    notImplemented();
    UNUSED_PARAM(rect);
    UNUSED_PARAM(imageBuffer);
}

DLL_PUBLIC void GraphicsContext::clipToImageBuffer(ImageBuffer* imageBuffer, const FloatRect& rect)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    notImplemented();
    UNUSED_PARAM(rect);
    UNUSED_PARAM(imageBuffer);
}
#endif

void GraphicsContext::addInnerRoundedRectClip(const IntRect& rect, int thickness)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    Path path;
    path.addEllipse(rect);
    path.addEllipse(FloatRect(rect.x() + thickness, rect.y() + thickness,
        rect.width() - (thickness * 2), rect.height() - (thickness * 2)));

    m_data->clipPath(path, PainterOpenVG::IntersectClip, m_state.fillRule);
}

void GraphicsContext::concatCTM(const AffineTransform& transformation)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->concatTransformation(transformation);
}

void GraphicsContext::setURLForRect(const KURL& link, const IntRect& destRect)
{
    notImplemented();
    UNUSED_PARAM(link);
    UNUSED_PARAM(destRect);
}

void GraphicsContext::setPlatformStrokeColor(const Color& color, ColorSpace colorSpace)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setStrokeColor(color);

    UNUSED_PARAM(colorSpace); // FIXME
}

void GraphicsContext::setPlatformStrokeGradient(Gradient* gradient)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setStrokeGradient(*(gradient->platformGradient()));
}

void GraphicsContext::setPlatformStrokePattern(Pattern* p)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    PatternOpenVG* platformPattern = p->createPlatformPattern(AffineTransform());
    m_data->setStrokePattern(*platformPattern);
    delete platformPattern;
}

void GraphicsContext::setPlatformStrokeStyle(StrokeStyle strokeStyle)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setStrokeStyle(strokeStyle);
}

void GraphicsContext::setPlatformStrokeThickness(float thickness)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setStrokeThickness(thickness);
}

void GraphicsContext::setPlatformFillColor(const Color& color, ColorSpace colorSpace)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setFillColor(color);

    UNUSED_PARAM(colorSpace); // FIXME
}

void GraphicsContext::setPlatformFillGradient(Gradient* gradient)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setFillGradient(*(gradient->platformGradient()));
}

void GraphicsContext::setPlatformFillPattern(Pattern* p)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    PatternOpenVG* platformPattern = p->createPlatformPattern(AffineTransform());
    m_data->setFillPattern(*platformPattern);
    delete platformPattern;
}

void GraphicsContext::setPlatformShouldAntialias(bool enable)
{
    if (UNLIKELY(paintingDisabled()))
        return;

    m_data->setAntialiasingEnabled(enable);
}

DLL_PUBLIC void GraphicsContext::setImageInterpolationQuality(InterpolationQuality)
{
    notImplemented();
}

InterpolationQuality GraphicsContext::imageInterpolationQuality() const
{
    notImplemented();
    return InterpolationDefault;
}

bool GraphicsContext::isAcceleratedContext() const
{
#if USE(ACCELERATED_COMPOSITING)
    return platformContext()->isAccelerated();
#else
    return false;
#endif
}

}
