/*
 * Copyright (C) Research In Motion Limited 2009-2010. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "PainterOpenVG.h"

#include "AffineTransform.h"
#include "Color.h"
#include "DashArray.h"
#include "FloatPoint.h"
#include "FloatQuad.h"
#include "FloatRect.h"
#include "IntRect.h"
#include "IntSize.h"
#include "NotImplemented.h"
#include "PaintOpenVG.h"
#include "PlatformPathOpenVG.h"
#include "SurfaceOpenVG.h"
#include "TiledImageOpenVG.h"
#include "VGUtils.h"

#include <vgu.h>

#include <wtf/Assertions.h>
#include <wtf/MathExtras.h>
#include <wtf/RefPtr.h>

// These bits can be ORed together for a total of 8 possible text drawing modes.
static const int cTextInvisible = 0;
static const int cTextFill = 1;
static const int cTextStroke = 2;
static const int cTextClip = 4;

namespace WebCore {

#if 0
static bool isNonRotatedAffineTransformation(const AffineTransform& t)
{
    return fabsf(t.b()) <= FLT_EPSILON && fabsf(t.c()) <= FLT_EPSILON;
}
#endif

static VGCapStyle toVGCapStyle(LineCap lineCap)
{
    switch (lineCap) {
    case RoundCap:
        return VG_CAP_ROUND;
    case SquareCap:
        return VG_CAP_SQUARE;
    case ButtCap:
    default:
        return VG_CAP_BUTT;
    }
}

static VGJoinStyle toVGJoinStyle(LineJoin lineJoin)
{
    switch (lineJoin) {
    case RoundJoin:
        return VG_JOIN_ROUND;
    case BevelJoin:
        return VG_JOIN_BEVEL;
    case MiterJoin:
    default:
        return VG_JOIN_MITER;
    }
}

static VGFillRule toVGFillRule(WindRule fillRule)
{
    return fillRule == RULE_EVENODD ? VG_EVEN_ODD : VG_NON_ZERO;
}

struct TransparencyLayer {
	WTF_MAKE_FAST_ALLOCATED;
public:
    TransparencyLayer(SurfaceOpenVG* surf, const IntRect& absRect, float alpha)
        : surface(surf)
        , rect(absRect)
        , opacity(alpha)
    {
    }

    ~TransparencyLayer()
    {
        delete surface;
    }

    SurfaceOpenVG* surface;
    IntRect rect;
    float opacity;
};

struct CachedClippingPath {
    WTF_MAKE_FAST_ALLOCATED;
public:
    CachedClippingPath(const Path& path, PainterOpenVG::ClipOperation clipOperation, const AffineTransform& surfaceTransformation);
    Path path;
    PainterOpenVG::ClipOperation clipOperation;
    AffineTransform surfaceTransformation;
};

CachedClippingPath::CachedClippingPath(const Path& path, PainterOpenVG::ClipOperation clipOperation, const AffineTransform& surfaceTransformation)
    : path(path)
    , clipOperation(clipOperation)
    , surfaceTransformation(surfaceTransformation)
{}

struct PlatformPainterState {
    WTF_MAKE_FAST_ALLOCATED;
public:
    PainterOpenVG::PainterStateCategoryFlags modifiedStateCategories;

    AffineTransform surfaceTransformation;
    CompositeOperator compositeOperation;
    float opacity;

    bool scissoringEnabled;
    FloatRect scissorRect;

    bool maskingEnabled;
    #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
    VGMaskLayer mask;
    IntRect maskBounds;
    Vector<CachedClippingPath*, 4> maskPaths;
    #endif

    PaintOpenVG fillPaint;
    StrokeStyle strokeStyle;
    PaintOpenVG strokePaint;
    float strokeThickness;
    LineCap strokeLineCap;
    LineJoin strokeLineJoin;
    float strokeMiterLimit;
    DashArray strokeDashArray;
    float strokeDashOffset;

    IntSize shadowOffset;
    float shadowBlur;
    Color shadowColor;

    int textDrawingMode;
    bool antialiasingEnabled;

    PlatformPainterState()
        : modifiedStateCategories(PainterOpenVG::NoStateCategories)
        , compositeOperation(CompositeSourceOver)
        , opacity(1.0)
        , scissoringEnabled(false)
        , maskingEnabled(false)
        #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
        , mask(VG_INVALID_HANDLE)
        #endif
        , fillPaint(Color::black)
        , strokeStyle(NoStroke)
        , strokeThickness(0.0)
        , strokeLineCap(ButtCap)
        , strokeLineJoin(MiterJoin)
        , strokeMiterLimit(4.0)
        , strokeDashOffset(0.0)
        , shadowBlur(0.0)
        , textDrawingMode(cTextFill)
        , antialiasingEnabled(true)
    {
    }

    ~PlatformPainterState()
    {
        #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
        if (mask != VG_INVALID_HANDLE) {
            vgDestroyMaskLayer(mask);
            ASSERT_VG_NO_ERROR();
        }

        for (size_t i = 0; i < maskPaths.size(); ++i)
            delete maskPaths[i];
        #endif
    }

    PlatformPainterState(const PlatformPainterState& state)
    {
        modifiedStateCategories = PainterOpenVG::NoStateCategories;
        surfaceTransformation = state.surfaceTransformation;

        scissoringEnabled = state.scissoringEnabled;
        scissorRect = state.scissorRect;

        maskingEnabled = state.maskingEnabled;
        maskBounds = state.maskBounds;
        #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
        mask = VG_INVALID_HANDLE;
        for (size_t i = 0; i < state.maskPaths.size(); ++i)
            maskPaths.append(new CachedClippingPath(*state.maskPaths[i]));
        #endif

        copyPaintState(&state);
    }

    void copyPaintState(const PlatformPainterState* other)
    {
        compositeOperation = other->compositeOperation;
        opacity = other->opacity;

        fillPaint = other->fillPaint;
        strokeStyle = other->strokeStyle;
        strokePaint = other->strokePaint;
        strokeThickness = other->strokeThickness;
        strokeLineCap = other->strokeLineCap;
        strokeLineJoin = other->strokeLineJoin;
        strokeMiterLimit = other->strokeMiterLimit;
        strokeDashArray = other->strokeDashArray;
        strokeDashOffset = other->strokeDashOffset;

        shadowOffset = other->shadowOffset;
        shadowBlur = other->shadowBlur;
        shadowColor = other->shadowColor;

        textDrawingMode = other->textDrawingMode;
        antialiasingEnabled = other->antialiasingEnabled;
    }

    void applyState(PainterOpenVG* painter, PainterOpenVG::PainterStateCategoryFlags modifiedStateCategories)
    {
        ASSERT(painter);

        if (modifiedStateCategories & PainterOpenVG::StrokeState) {
            strokePaint.apply(VG_STROKE_PATH, painter->m_vgStrokePaint);
            ASSERT_VG_NO_ERROR();
            vgSetf(VG_STROKE_LINE_WIDTH, strokeThickness);
            vgSeti(VG_STROKE_CAP_STYLE, toVGCapStyle(strokeLineCap));
            vgSeti(VG_STROKE_JOIN_STYLE, toVGJoinStyle(strokeLineJoin));
            vgSetf(VG_STROKE_MITER_LIMIT, strokeMiterLimit);
            ASSERT_VG_NO_ERROR();
            applyStrokeStyle();
        }

        if (modifiedStateCategories & PainterOpenVG::FillState) {
            fillPaint.apply(VG_FILL_PATH, painter->m_vgFillPaint);
        }

        if (modifiedStateCategories & PainterOpenVG::QualitySettingState) {
            if (antialiasingEnabled)
                vgSeti(VG_RENDERING_QUALITY, WEBKIT_VG_RENDERING_QUALITY);
            else
                vgSeti(VG_RENDERING_QUALITY, VG_RENDERING_QUALITY_NONANTIALIASED);
        }
        ASSERT_VG_NO_ERROR();

        if (modifiedStateCategories & PainterOpenVG::BlendingState)
            applyBlending(painter);

        if (modifiedStateCategories & PainterOpenVG::ScissorState)
            applyScissorRect(painter);

        #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
        if (modifiedStateCategories & PainterOpenVG::MaskState) {
            if (maskingEnabled) {
                vgSeti(VG_MASKING, VG_TRUE);
                vgMask(VG_INVALID_HANDLE, VG_CLEAR_MASK, 0, 0, painter->currentSurface()->width(), painter->currentSurface()->height());
                
                // if our mask is cached, restore it
                if (mask != VG_INVALID_HANDLE) {
                    vgMask(mask, VG_SET_MASK, maskBounds.x(), maskBounds.y(), maskBounds.width(), maskBounds.height());

                // otherwise rebuild it from the path list
                } else if (!maskBounds.isEmpty()) {
                    vgMask(mask, VG_FILL_MASK, maskBounds.x(), maskBounds.y(), maskBounds.width(), maskBounds.height());
                    for (size_t i = 0; i < maskPaths.size(); ++i) {
                        const CachedClippingPath* cached = maskPaths[i];
                        applyTransformation(painter, cached->surfaceTransformation);
                        vgRenderToMask(cached->path.platformPath()->vgPath(), VG_FILL_PATH, (VGMaskOperation)cached->clipOperation);
                    }
                    modifiedStateCategories |= PainterOpenVG::TransformationState;

                } else {
                    ASSERT(0 && "cannot restore clip!");
                }

            } else {
                vgSeti(VG_MASKING, VG_FALSE);
            }
        }
        #endif
        ASSERT_VG_NO_ERROR();

        if (modifiedStateCategories & PainterOpenVG::TransformationState)
            applyTransformation(painter);

    }

    void applyBlending(PainterOpenVG* painter)
    {
        VGBlendMode blendMode = VG_BLEND_SRC_OVER;

        switch (compositeOperation) {
        case CompositeClear: {
            // Clear means "set to fully transparent regardless of SRC".
            // We implement that by multiplying DST with white color
            // (= no changes) and an alpha of 1.0 - opacity, so the destination
            // pixels will be fully transparent when opacity == 1.0 and
            // unchanged when opacity == 0.0.
            blendMode = VG_BLEND_DST_IN;
            const VGfloat values[] = { 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0 - opacity };
            vgSetfv(VG_COLOR_TRANSFORM_VALUES, 8, values);
            vgSeti(VG_COLOR_TRANSFORM, VG_TRUE);
            ASSERT_VG_NO_ERROR();
            break;
        }
        case CompositeCopy:
            blendMode = VG_BLEND_SRC;
            break;
        case CompositeSourceOver:
            blendMode = VG_BLEND_SRC_OVER;
            break;
        case CompositeSourceIn:
            blendMode = VG_BLEND_SRC_IN;
            break;
        case CompositeSourceOut:
            notImplemented();
            break;
        case CompositeSourceAtop:
            notImplemented();
            break;
        case CompositeDestinationOver:
            blendMode = VG_BLEND_DST_OVER;
            break;
        case CompositeDestinationIn:
            blendMode = VG_BLEND_DST_IN;
            break;
        case CompositeDestinationOut:
            notImplemented();
            break;
        case CompositeDestinationAtop:
            notImplemented();
            break;
        case CompositeXOR:
            notImplemented();
            break;
        case CompositePlusDarker:
            blendMode = VG_BLEND_DARKEN;
            break;
        case CompositePlusLighter:
            blendMode = VG_BLEND_LIGHTEN;
            break;
        default:
            notImplemented();
            break;
        }

        if (compositeOperation != CompositeClear) {
            if (opacity >= (1.0 - FLT_EPSILON) && !painter->m_drawingShadow)
                vgSeti(VG_COLOR_TRANSFORM, VG_FALSE);
            else if (blendMode == VG_BLEND_SRC) {
                blendMode = VG_BLEND_SRC_OVER;
                VGfloat values[] = { 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, opacity };
                if (painter->m_drawingShadow) {
                    values[0] = 0.0;
                    values[1] = 0.0;
                    values[2] = 0.0;
                    values[4] = shadowColor.red() / 255.0;
                    values[5] = shadowColor.green() / 255.0;
                    values[6] = shadowColor.blue() / 255.0;
                    values[7] *= shadowColor.alpha() / 255.0;
                }
                vgSetfv(VG_COLOR_TRANSFORM_VALUES, 8, values);
                vgSeti(VG_COLOR_TRANSFORM, VG_TRUE);
            } else {
                VGfloat values[] = { 1.0, 1.0, 1.0, opacity, 0.0, 0.0, 0.0, 0.0 };
                if (painter->m_drawingShadow) {
                    values[0] = 0.0;
                    values[1] = 0.0;
                    values[2] = 0.0;
                    values[3] *= shadowColor.alpha() / 255.0;
                    values[4] = shadowColor.red() / 255.0;
                    values[5] = shadowColor.green() / 255.0;
                    values[6] = shadowColor.blue() / 255.0;
                }
                vgSetfv(VG_COLOR_TRANSFORM_VALUES, 8, values);
                vgSeti(VG_COLOR_TRANSFORM, VG_TRUE);
            }
            ASSERT_VG_NO_ERROR();
        }

        vgSeti(VG_BLEND_MODE, blendMode);
        ASSERT_VG_NO_ERROR();
    }

    void applyTransformation(PainterOpenVG* painter, const AffineTransform& transform)
    {
        // There are *five* separate transforms that can be applied to OpenVG
        // as of 1.1. We use the same "world transformation" for paths, images
        // and glyphs, and only fill/stroke transformations (which are applied
        // before the others take effect) are separate ones.
        VGMatrix vgMatrix(transform);
        const VGfloat* vgFloatArray = vgMatrix.toVGfloat();

        vgSeti(VG_MATRIX_MODE, VG_MATRIX_PATH_USER_TO_SURFACE);
        VGUtils::loadMatrix(vgFloatArray);
        ASSERT_VG_NO_ERROR();

        vgSeti(VG_MATRIX_MODE, VG_MATRIX_IMAGE_USER_TO_SURFACE);
        VGUtils::loadMatrix(vgFloatArray);
        ASSERT_VG_NO_ERROR();

#ifdef OPENVG_VERSION_1_1
        vgSeti(VG_MATRIX_MODE, VG_MATRIX_GLYPH_USER_TO_SURFACE);
        VGUtils::loadMatrix(vgFloatArray);
        ASSERT_VG_NO_ERROR();
#endif
    }

    void applyTransformation(PainterOpenVG* painter)
    {
        applyTransformation(painter, surfaceTransformation);
    }

    void applyScissorRect(PainterOpenVG* painter)
    {
        if (scissoringEnabled) {
            vgSeti(VG_SCISSORING, VG_TRUE);

            // scissor regions are inverted on some platforms
            // TODO: is this inverted only on the primary surface or on all of them?
            if (VGUtils::bugInvertedScissor) {
                FloatRect rect(scissorRect);
                rect.setY(painter->currentSurface()->height() - rect.y() - rect.height());
                vgSetfv(VG_SCISSOR_RECTS, 4, VGRect(rect).toVGfloat());
            } else {
                vgSetfv(VG_SCISSOR_RECTS, 4, VGRect(scissorRect).toVGfloat());
            }

        } else
            vgSeti(VG_SCISSORING, VG_FALSE);

        ASSERT_VG_NO_ERROR();
    }

    void applyStrokeStyle()
    {
        if (strokeStyle == DottedStroke) {
            VGfloat vgFloatArray[2] = { 1.0, 1.0 };
            vgSetfv(VG_STROKE_DASH_PATTERN, 2, vgFloatArray);
            vgSetf(VG_STROKE_DASH_PHASE, 0.0);
        } else if (strokeStyle == DashedStroke) {
            if (!strokeDashArray.size()) {
                VGfloat vgFloatArray[2] = { 4.0, 3.0 };
                vgSetfv(VG_STROKE_DASH_PATTERN, 2, vgFloatArray);
            } else {
                Vector<VGfloat, 8> vgFloatArray(strokeDashArray.size());
                for (size_t i = 0; i < strokeDashArray.size(); ++i)
                    vgFloatArray[i] = strokeDashArray[i];

                vgSetfv(VG_STROKE_DASH_PATTERN, vgFloatArray.size(), vgFloatArray.data());
            }
            vgSetf(VG_STROKE_DASH_PHASE, strokeDashOffset);
        } else {
            vgSetfv(VG_STROKE_DASH_PATTERN, 0, 0);
            vgSetf(VG_STROKE_DASH_PHASE, 0.0);
        }

        ASSERT_VG_NO_ERROR();
    }

    inline bool strokeDisabled() const
    {
        return strokeStyle == NoStroke || strokeThickness < FLT_EPSILON
            || (compositeOperation == CompositeSourceOver && strokePaint.isEmpty());
    }

    inline bool fillDisabled() const
    {
        return (compositeOperation == CompositeSourceOver && fillPaint.isEmpty());
    }

    void saveMask()
    {
        #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
        // if we've already cached the mask, do nothing; it will be destroyed
        // if it changes.
        if (mask != VG_INVALID_HANDLE)
            return;

        // we can't save the mask if vgCopyMask() is bugged; we'll just
        // rebuild it from the path cache instead.
        if (maskingEnabled && !maskBounds.isEmpty() && !VGUtils::bugDisableCopyMask) {
            mask = vgCreateMaskLayer(maskBounds.width(), maskBounds.height());
            ASSERT_VG_NO_ERROR();
            vgCopyMask(mask, 0, 0, maskBounds.x(), maskBounds.y(), maskBounds.width(), maskBounds.height());
            ASSERT_VG_NO_ERROR();
        }
        #endif
    }
};

inline bool PainterOpenVG::debugDisabled() {
    return false
        #if PAINTER_DISABLE_TRANSPARENCY_LAYERS
        || debugTransparencyLevel > 0
        #endif
        #if PAINTER_DISABLE_MASKING
        || (m_state && m_state->maskingEnabled)
        #endif
        ;
}

PainterOpenVG::PainterOpenVG()
    : m_state(0)
    , m_baseSurface(0)
    , m_currentSurface(0)
    , m_currentPath(0)
    , m_drawingShadow(false)
    , m_didPaint(false)
    #if PAINTER_DISABLE_TRANSPARENCY_LAYERS
    , debugTransparencyLevel(0)
    #endif
{
    platformInit(NULL);

    m_vgStrokePaint = vgCreatePaint();
    m_vgFillPaint = vgCreatePaint();
    ASSERT_VG_NO_ERROR();
}

PainterOpenVG::PainterOpenVG(SurfaceOpenVG* surface)
    : m_state(0)
    , m_baseSurface(0)
    , m_currentSurface(0)
    , m_currentPath(0)
    , m_drawingShadow(false)
    #if PAINTER_DISABLE_TRANSPARENCY_LAYERS
    , debugTransparencyLevel(0)
    #endif
{
    ASSERT(surface);
    platformInit(surface);

    m_vgStrokePaint = vgCreatePaint();
    m_vgFillPaint = vgCreatePaint();
    ASSERT_VG_NO_ERROR();

    begin(surface);
}

PainterOpenVG::~PainterOpenVG()
{
    end();
    delete m_currentPath;

    vgDestroyPaint(m_vgStrokePaint);
    vgDestroyPaint(m_vgFillPaint);
    ASSERT_VG_NO_ERROR();

    platformDestroy();
}

void PainterOpenVG::begin(SurfaceOpenVG* surface)
{
    if (surface == m_baseSurface)
        return;

    ASSERT(surface);
    ASSERT(!m_state);

    m_baseSurface = surface;
    m_currentSurface = m_baseSurface;
    m_drawingShadow = false;

    m_stateStack.append(new PlatformPainterState());
    m_state = m_stateStack.last();

    m_currentSurface->setActivePainter(this);
    makeCurrent();
}

DLL_PUBLIC void PainterOpenVG::end()
{
    if (!m_baseSurface)
        return;

    destroyTransparencyLayers();

    m_baseSurface->setActivePainter(0);
    m_baseSurface = 0;
    m_currentSurface = 0;

    destroyPainterStates();
}

DLL_PUBLIC void PainterOpenVG::setStateModified(PainterStateCategoryFlags modifiedStateCategories)
{
    m_state->modifiedStateCategories |= modifiedStateCategories;
}

void PainterOpenVG::destroyTransparencyLayers()
{
    TransparencyLayer* layer = 0;
    while (!m_layers.isEmpty()) {
        layer = m_layers.last();
        m_layers.removeLast();
        delete layer;
    }
    m_currentSurface = m_baseSurface;
}

void PainterOpenVG::destroyPainterStates()
{
    PlatformPainterState* state = 0;
    while (!m_stateStack.isEmpty()) {
        state = m_stateStack.last();
        m_stateStack.removeLast();
        delete state;
    }
    m_state = 0;
}

/**
 * Copy the current back buffer image onto the surface.
 *
 * Call this method when all painting operations have been completed,
 * otherwise the surface won't visibly change.
 */
void PainterOpenVG::blitToSurface()
{
    ASSERT(m_state); // implies m_baseSurface
    m_baseSurface->flush();
}

void PainterOpenVG::beginTransparencyLayer(float opacity)
{
    #if PAINTER_DISABLE_TRANSPARENCY_LAYERS
    save();
    ++debugTransparencyLevel;
    return;
    #endif

    ASSERT(m_state);

    // Retrieve scissor rect and transformation matrix using the offset of the
    // current layer, and apply the new offset later.
    AffineTransform surfaceIndependentTransformation = transformation();
    IntRect layerRect(0, 0, m_currentSurface->width(), m_currentSurface->height());

    if (m_state->scissoringEnabled) {
        layerRect.intersect(enclosingIntRect(m_state->scissorRect));
        if (layerRect.isEmpty())
            layerRect.setSize(IntSize(1, 1));
    }
    if (!m_layers.isEmpty()) {
        const IntRect& currentLayerRect = m_layers.last()->rect;
        layerRect.move(currentLayerRect.x(), currentLayerRect.y());
    }
    ASSERT(!layerRect.isEmpty());

    m_state->saveMask();

    save();

    // clear mask and scissor from transparency layer
    m_state->modifiedStateCategories |= ScissorState;
    m_state->scissoringEnabled = false;
    m_state->scissorRect = IntRect();
    m_state->maskingEnabled = false;
    #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
    m_state->modifiedStateCategories |= MaskState;
    m_state->mask = VG_INVALID_HANDLE;
    m_state->maskBounds = IntRect();
    for (size_t i = 0; i < m_state->maskPaths.size(); ++i)
        delete m_state->maskPaths[i];
    m_state->maskPaths.clear();
    #endif
    m_state->applyState(this, m_state->modifiedStateCategories);

    SurfaceOpenVG* surface = new SurfaceOpenVG(layerRect.size());

    surface->setActivePainter(this);
    TransparencyLayer* layer = new TransparencyLayer(surface, layerRect, opacity);
    m_layers.append(layer);

    m_currentSurface = surface;
    makeCurrent();

    // Clear the surface from whatever color was set before to transparent.
    // No need to turn off scissoring, it's already disabled in our new state.
    const VGfloat transparent[4] = { 0.0, 0.0, 0.0, 0.0 };
    vgSetfv(VG_CLEAR_COLOR, 4, transparent);
    vgClear(0, 0, layer->rect.width(), layer->rect.height());

    // Apply offset to the transformation.
    setTransformation(surfaceIndependentTransformation);
}

void PainterOpenVG::endTransparencyLayer()
{
    endTransparencyLayer(0);
}

void PainterOpenVG::endTransparencyLayer(int blurRadius)
{
    #if PAINTER_DISABLE_TRANSPARENCY_LAYERS
    restore();
    --debugTransparencyLevel;
    return;
    #endif

    ASSERT(m_state);
    ASSERT(!m_layers.isEmpty());

    TransparencyLayer* layer = m_layers.last();
    IntRect layerRectAtZero(IntPoint(0, 0), layer->rect.size());
    m_layers.removeLast();
    layer->surface->setActivePainter(0);

    // Make the underlying surface current so that we can access the
    // surface image of our transparency layer on top.
    m_currentSurface = m_layers.isEmpty() ? m_baseSurface : m_layers.last()->surface;
    restore();
    save();

    // Get the image to paint
    VGImage image = layer->surface->image();
    VGImage blurredImage = VG_INVALID_HANDLE;

    // Blur it if necessary
    if (blurRadius && !VGUtils::bugDisableBlur) {
        VGImageFormat format = (VGImageFormat)vgGetParameteri(image, VG_IMAGE_FORMAT);
        blurredImage = vgCreateImage(format, layer->rect.width(), layer->rect.height(), VG_IMAGE_QUALITY_NONANTIALIASED);
        ASSERT_VG_NO_ERROR();

        setFillColor(Color::transparent);
        vgGaussianBlur(blurredImage, image, blurRadius, blurRadius, VG_TILE_FILL);
        ASSERT_VG_NO_ERROR();

        image = blurredImage;
    }

    AffineTransform originalTransformation = m_state->surfaceTransformation;
    CompositeOperator op = m_state->compositeOperation;
    float opacity = m_state->opacity;
    Color shadowColor = m_state->shadowColor;

    m_state->compositeOperation = CompositeSourceOver;
    m_state->opacity = layer->opacity;
    m_state->shadowColor = Color(); // make sure not to try to draw any shadows
    m_state->applyBlending(this);
    setSurfaceTransformation(AffineTransform());

    if (m_currentSurface != m_baseSurface) {
        const IntRect& lowerRect = m_layers.last()->rect;
        layer->rect.move(-lowerRect.x(), -lowerRect.y());
    }
    translate(layer->rect.x(), layer->rect.y());

    // some platforms have inverted pbuffer surfaces
    if (layer->surface->tiledImage()->inverted()) {
        translate(0.0f, layer->rect.height());
        scale(FloatSize(1.0f, -1.0f));
    }

    if (!debugDisabled()) { 
        if (m_layers.size() > 0 && !m_state->maskingEnabled) {                
            vgSeti(VG_MASKING, VG_TRUE);
            VGImage fimage = m_currentSurface->image();    
            vgClear(0, 0, m_currentSurface->width(), m_currentSurface->height());
            vgMask(image, VG_SET_MASK, 0, 0, layer->surface->width(), layer->surface->height());
            vgDrawImage(fimage);
            ASSERT_VG_NO_ERROR();            
            vgSeti(VG_MASKING, VG_FALSE);
        } else {
            vgDrawImage(image);
        }
    }
    painted();
    ASSERT_VG_NO_ERROR();

    setSurfaceTransformation(originalTransformation);
    m_state->compositeOperation = op;
    m_state->opacity = opacity;
    m_state->shadowColor = shadowColor;
    m_state->applyBlending(this);

    if (blurredImage)
        vgDestroyImage(blurredImage);
    delete layer;
    restore();
}

AffineTransform PainterOpenVG::transformation() const
{
    ASSERT(m_state);

    if (m_currentSurface == m_baseSurface) // a.k.a. m_layers.isEmpty()
        return m_state->surfaceTransformation;

    AffineTransform transformationWithoutOffset;
    const IntRect& layerRect = m_layers.last()->rect;
    transformationWithoutOffset.translate(layerRect.x(), layerRect.y());
    transformationWithoutOffset.multiply(m_state->surfaceTransformation);
    return transformationWithoutOffset;
}

void PainterOpenVG::setTransformation(const AffineTransform& transformation)
{
    if (m_currentSurface == m_baseSurface) // a.k.a. m_layers.isEmpty()
        setSurfaceTransformation(transformation);
    else {
        // If a transparency layer surface is put on top of the base surface,
        // take the layer offset into account so the transformation results in
        // correct rendering for the layer surface.
        AffineTransform transformationWithOffset;
        const IntRect& layerRect = m_layers.last()->rect;
        transformationWithOffset.translate(-layerRect.x(), -layerRect.y());
        transformationWithOffset.multiply(transformation);
        setSurfaceTransformation(transformationWithOffset);
    }
}

void PainterOpenVG::concatTransformation(const AffineTransform& transformation)
{
    ASSERT(m_state);
    m_state->modifiedStateCategories |= TransformationState;
    makeCurrent();

    // We do the multiplication ourself using WebCore's AffineTransform rather
    // than offloading this to VG via vgMultMatrix() to keep things simple and
    // so we can maintain state ourselves.
    m_state->surfaceTransformation.multiply(transformation);
    m_state->applyTransformation(this);
}

void PainterOpenVG::setSurfaceTransformation(const AffineTransform& transformation)
{
    ASSERT(m_state);
    m_state->modifiedStateCategories |= TransformationState;
    makeCurrent();

    m_state->surfaceTransformation = transformation;
    m_state->applyTransformation(this);
}

void PainterOpenVG::transformPath(VGPath dst, VGPath src, const AffineTransform& transformation)
{
    #if PAINTER_DISABLE_PATHS 
    return;
    #endif
    vgSeti(VG_MATRIX_MODE, VG_MATRIX_PATH_USER_TO_SURFACE);

    // Save the transform state
    VGfloat currentMatrix[9];
    vgGetMatrix(currentMatrix);
    ASSERT_VG_NO_ERROR();

    // Load the new transform
    VGUtils::loadMatrix(VGMatrix(transformation).toVGfloat());
    ASSERT_VG_NO_ERROR();

    // Apply the new transform
    vgTransformPath(dst, src);
    ASSERT_VG_NO_ERROR();

    // Restore the transform state
    VGUtils::loadMatrix(currentMatrix);
    ASSERT_VG_NO_ERROR();
}

CompositeOperator PainterOpenVG::compositeOperation() const
{
    ASSERT(m_state);
    return m_state->compositeOperation;
}

void PainterOpenVG::setCompositeOperation(CompositeOperator op)
{
    ASSERT(m_state);

    if (op == m_state->compositeOperation)
        return;

    m_state->modifiedStateCategories |= BlendingState;
    makeCurrent();

    m_state->compositeOperation = op;
    m_state->applyBlending(this);
}

float PainterOpenVG::opacity() const
{
    ASSERT(m_state);
    return m_state->opacity;
}

void PainterOpenVG::setOpacity(float opacity)
{
    ASSERT(m_state);

    if (opacity == m_state->opacity)
        return;

    m_state->modifiedStateCategories |= BlendingState;
    makeCurrent();

    m_state->opacity = opacity;
    m_state->applyBlending(this);
}

float PainterOpenVG::strokeThickness() const
{
    ASSERT(m_state);
    return m_state->strokeThickness;
}

void PainterOpenVG::setStrokeThickness(float thickness)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    ASSERT(m_state);
    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokeThickness = thickness;
    vgSetf(VG_STROKE_LINE_WIDTH, thickness);
    ASSERT_VG_NO_ERROR();
}

StrokeStyle PainterOpenVG::strokeStyle() const
{
    ASSERT(m_state);
    return m_state->strokeStyle;
}

void PainterOpenVG::setStrokeStyle(const StrokeStyle& style)
{
    ASSERT(m_state);
    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokeStyle = style;
    m_state->applyStrokeStyle();
}

void PainterOpenVG::setLineDash(const DashArray& dashArray, float dashOffset)
{
    ASSERT(m_state);

    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokeDashArray = dashArray;
    m_state->strokeDashOffset = dashOffset;
    m_state->applyStrokeStyle();
}

void PainterOpenVG::setLineCap(LineCap lineCap)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    ASSERT(m_state);
    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokeLineCap = lineCap;
    vgSeti(VG_STROKE_CAP_STYLE, toVGCapStyle(lineCap));
    ASSERT_VG_NO_ERROR();
}

void PainterOpenVG::setLineJoin(LineJoin lineJoin)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    ASSERT(m_state);
    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokeLineJoin = lineJoin;
    vgSeti(VG_STROKE_JOIN_STYLE, toVGJoinStyle(lineJoin));
    ASSERT_VG_NO_ERROR();
}

void PainterOpenVG::setMiterLimit(float miterLimit)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    ASSERT(m_state);
    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokeMiterLimit = miterLimit;
    vgSetf(VG_STROKE_MITER_LIMIT, miterLimit);
    ASSERT_VG_NO_ERROR();
}

const PaintOpenVG& PainterOpenVG::strokePaint() const
{
    ASSERT(m_state);
    return m_state->strokePaint;
}

void PainterOpenVG::setStrokePaint(const PaintOpenVG& paint)
{
    ASSERT(m_state);

    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokePaint = paint;
    m_state->strokePaint.apply(VG_STROKE_PATH, m_vgStrokePaint);
}

void PainterOpenVG::setStrokeColor(const Color& color)
{
    if (!color.isValid()) {
        setStrokeColor(Color::transparent);
        return;
    }

    ASSERT(m_state);

    if (m_state->strokePaint.type() == PaintOpenVG::ColorPaint && color == m_state->strokePaint.color())
        return;

    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokePaint.setColor(color);
    m_state->strokePaint.apply(VG_STROKE_PATH, m_vgStrokePaint);
}

void PainterOpenVG::setStrokeGradient(const GradientOpenVG& gradient)
{
    if (gradient.isEmpty()) {
        // See http://www.w3.org/TR/2dcontext/#colors-and-styles:
        // "When there are no stops, the gradient is transparent black."
        setStrokeColor(Color::transparent);
        return;
    }

    ASSERT(m_state);

    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokePaint.setGradient(gradient);
    m_state->strokePaint.apply(VG_STROKE_PATH, m_vgStrokePaint);
}

void PainterOpenVG::setStrokePattern(const PatternOpenVG& pattern)
{
    if (!pattern.isValid()) {
        setStrokeColor(Color::transparent);
        return;
    }

    ASSERT(m_state);

    m_state->modifiedStateCategories |= StrokeState;
    makeCurrent();

    m_state->strokePaint.setPattern(pattern);
    m_state->strokePaint.apply(VG_STROKE_PATH, m_vgStrokePaint);
}

const PaintOpenVG& PainterOpenVG::fillPaint() const
{
    ASSERT(m_state);
    return m_state->fillPaint;
}

void PainterOpenVG::setFillPaint(const PaintOpenVG& paint)
{
    ASSERT(m_state);

    m_state->modifiedStateCategories |= FillState;
    makeCurrent();

    m_state->fillPaint = paint;
    m_state->fillPaint.apply(VG_FILL_PATH, m_vgFillPaint);
}

void PainterOpenVG::setFillColor(const Color& color)
{
    if (!color.isValid()) {
        setFillColor(Color::transparent);
        return;
    }

    ASSERT(m_state);

    if (m_state->fillPaint.type() == PaintOpenVG::ColorPaint && color == m_state->fillPaint.color())
        return;

    m_state->modifiedStateCategories |= FillState;
    makeCurrent();

    m_state->fillPaint.setColor(color);
    m_state->fillPaint.apply(VG_FILL_PATH, m_vgFillPaint);
}

void PainterOpenVG::setFillGradient(const GradientOpenVG& gradient)
{
    if (gradient.isEmpty()) {
        // See http://www.w3.org/TR/2dcontext/#colors-and-styles:
        // "When there are no stops, the gradient is transparent black."
        setFillColor(Color::transparent);
        return;
    }

    ASSERT(m_state);

    m_state->modifiedStateCategories |= FillState;
    makeCurrent();

    m_state->fillPaint.setGradient(gradient);
    m_state->fillPaint.apply(VG_FILL_PATH, m_vgFillPaint);
}

void PainterOpenVG::setFillPattern(const PatternOpenVG& pattern)
{
    if (!pattern.isValid()) {
        setFillColor(Color::transparent);
        return;
    }

    ASSERT(m_state);

    m_state->modifiedStateCategories |= FillState;
    makeCurrent();

    m_state->fillPaint.setPattern(pattern);
    m_state->fillPaint.apply(VG_FILL_PATH, m_vgFillPaint);
}

bool PainterOpenVG::shadowEnabled() const
{
    #if PAINTER_DISABLE_SHADOWS
    return false;
    #endif
    ASSERT(m_state);
    return m_state->shadowColor.isValid();
}

void PainterOpenVG::setShadow(const IntSize& offset, float blur, const Color& color)
{
    ASSERT(m_state);

    if (color.isValid() && color.alpha() && (blur > FLT_EPSILON || offset.width() || offset.height())) {
        m_state->shadowOffset = offset;
        m_state->shadowBlur = blur;
        m_state->shadowColor = color;
    } else
        clearShadow();
}

void PainterOpenVG::clearShadow()
{
    m_state->shadowOffset = IntSize();
    m_state->shadowBlur = 0.0;
    m_state->shadowColor = Color();
}

void PainterOpenVG::beginDrawShadow()
{
    ASSERT(shadowEnabled());

    if (m_state->shadowBlur > FLT_EPSILON && !VGUtils::bugDisableBlur)
        beginTransparencyLayer(1.0);
    else
        save();

    m_drawingShadow = true;
    m_state->modifiedStateCategories |= BlendingState;
    m_state->applyBlending(this);
    translate(m_state->shadowOffset.width(), m_state->shadowOffset.height());
}

void PainterOpenVG::endDrawShadow()
{
    ASSERT(shadowEnabled());
    m_drawingShadow = false;

    if (m_state->shadowBlur > FLT_EPSILON && !VGUtils::bugDisableBlur)
        endTransparencyLayer(static_cast<int>((m_state->shadowBlur + 1) / 2));
    else
        restore();
}

int PainterOpenVG::textDrawingMode() const
{
    ASSERT(m_state);
    return m_state->textDrawingMode;
}

void PainterOpenVG::setTextDrawingMode(int mode)
{
    ASSERT(m_state);
    m_state->textDrawingMode = mode;
}

bool PainterOpenVG::antialiasingEnabled() const
{
    ASSERT(m_state);
    return m_state->antialiasingEnabled;
}

void PainterOpenVG::setAntialiasingEnabled(bool enabled)
{
    // TODO: ?
    return;

    ASSERT(m_state);
    m_state->modifiedStateCategories |= QualitySettingState;
    makeCurrent();

    m_state->antialiasingEnabled = enabled;

    if (enabled)
        vgSeti(VG_RENDERING_QUALITY, WEBKIT_VG_RENDERING_QUALITY);
    else
        vgSeti(VG_RENDERING_QUALITY, VG_RENDERING_QUALITY_NONANTIALIASED);
    vgSeti(VG_PIXEL_LAYOUT, VG_PIXEL_LAYOUT_UNKNOWN);
}

void PainterOpenVG::scale(const FloatSize& scaleFactors)
{
    ASSERT(m_state);
    makeCurrent();

    AffineTransform transformation = m_state->surfaceTransformation;
    transformation.scaleNonUniform(scaleFactors.width(), scaleFactors.height());
    setSurfaceTransformation(transformation);
}

void PainterOpenVG::rotate(float radians)
{
    ASSERT(m_state);
    makeCurrent();

    AffineTransform transformation = m_state->surfaceTransformation;
    transformation.rotate(rad2deg(radians));
    setSurfaceTransformation(transformation);
}

void PainterOpenVG::shear(float sx, float sy) {
	ASSERT(m_state);
	makeCurrent();

	AffineTransform transformation = m_state->surfaceTransformation;
    transformation.shear(sx, sy);
    setSurfaceTransformation(transformation);
}
void PainterOpenVG::translate(float dx, float dy)
{
    ASSERT(m_state);
    makeCurrent();

    AffineTransform transformation = m_state->surfaceTransformation;
    transformation.translate(dx, dy);
    setSurfaceTransformation(transformation);
}

void PainterOpenVG::beginPath()
{
    if (!m_currentPath) {
        m_currentPath = new Path();
        return;
    }
    m_currentPath->clear();
}

void PainterOpenVG::addPath(const Path& path)
{
    addPath(path.platformPath()->vgPath());
}

void PainterOpenVG::addPath(VGPath path)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    m_currentPath->platformPath()->makeCompatibleContextCurrent();

    vgAppendPath(m_currentPath->platformPath()->vgPath(), path);
    ASSERT_VG_NO_ERROR();
}

Path* PainterOpenVG::currentPath() const
{
    return m_currentPath;
}

void PainterOpenVG::drawPath(const Path& path, VGbitfield specifiedPaintModes, WindRule fillRule)
{
    if (path.isNull())
        return;
    if (debugDisabled())
        return;
    drawPath(path.platformPath()->vgPath(), specifiedPaintModes, fillRule);
}

void PainterOpenVG::drawPath(VGPath path, VGbitfield specifiedPaintModes, WindRule fillRule)
{
    if (debugDisabled())
        return;
    #if PAINTER_DISABLE_PATHS 
    return;
    #endif

    ASSERT(m_state);

    VGbitfield paintModes = 0;
    if (!m_state->strokeDisabled())
        paintModes |= VG_STROKE_PATH;
    if (!m_state->fillDisabled())
        paintModes |= VG_FILL_PATH;

    paintModes &= specifiedPaintModes;

    if (!paintModes)
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();
        drawPath(path, paintModes, fillRule);
        endDrawShadow();
    } else
        makeCurrent();

    vgSeti(VG_FILL_RULE, toVGFillRule(fillRule));
    vgDrawPath(path, paintModes);
    painted();
    ASSERT_VG_NO_ERROR();
}

void PainterOpenVG::intersectScissorRect(const FloatRect& rect)
{
    // Scissor rectangles are defined by float values, but e.g. painting
    // something red to a float-clipped rectangle and then painting something
    // white to the same rectangle will leave some red remnants as it is
    // rendered to full pixels in between. Also, some OpenVG implementations
    // are likely to clip to integer coordinates anyways because of the above
    // effect. So considering the above (and confirming through tests) the
    // visual result is better if we clip to the enclosing integer rectangle
    // rather than the exact float rectangle for scissoring.
    FloatRect intersectingRect = enclosingIntRect(rect);

    if (m_state->scissoringEnabled) {
        if (intersectingRect.contains(m_state->scissorRect))
            return;

        m_state->scissorRect.intersect(intersectingRect);
    } else {
        m_state->scissoringEnabled = true;
        m_state->scissorRect = intersectingRect;
    }

    m_state->modifiedStateCategories |= ScissorState;
    m_state->applyScissorRect(this);
}

void PainterOpenVG::clipRect(const FloatRect& rect, PainterOpenVG::ClipOperation maskOp)
{
    ASSERT(m_state);
    makeCurrent();

    if (maskOp != PainterOpenVG::IntersectClip) {

        // just clip out the rect as any other path
        Path rectPath;
        rectPath.addRect(rect);
        clipPath(rectPath, maskOp);

        return;
    }

    if (m_state->surfaceTransformation.isIdentity()) {
        // No transformation required, skip all the complex stuff.
        intersectScissorRect(rect);
        return;
    }

    // Check if the actual destination rectangle is still rectilinear (can be
    // represented as FloatRect) so we could apply scissoring instead of
    // (potentially more expensive) path clipping. Note that scissoring is not
    // subject to transformations, so we need to do the transformation to
    // surface coordinates by ourselves.
    FloatQuad effectiveScissorQuad = m_state->surfaceTransformation.mapQuad(FloatQuad(rect));

    if (effectiveScissorQuad.isRectilinear())
        intersectScissorRect(effectiveScissorQuad.boundingBox());
    else {
        // The transformed scissorRect cannot be represented as FloatRect
        // anymore, so we need to perform masking instead.
        Path scissorRectPath;
        scissorRectPath.addRect(rect);
        clipPath(scissorRectPath, PainterOpenVG::IntersectClip);
    }
}

void PainterOpenVG::clipPath(const Path& path, PainterOpenVG::ClipOperation maskOp, WindRule clipRule)
{
    #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
    ASSERT(m_state);
    VGPath vgPath = path.platformPath()->vgPath();

    m_state->modifiedStateCategories |= MaskState;
    makeCurrent();

    // if we've cached the current mask, destroy it
    if (m_state->mask != VG_INVALID_HANDLE) {
        vgDestroyMaskLayer(m_state->mask);
        m_state->mask = VG_INVALID_HANDLE;
        ASSERT_VG_NO_ERROR();
    }

    // if we have a parent state, tell it to save its mask because it will need to restore it
    if (m_stateStack.size() >= 2)
        m_stateStack[m_stateStack.size() - 2]->saveMask();

    // if we're first enabling masking, clear the mask
    if (!m_state->maskingEnabled) {
        m_state->maskingEnabled = true;
        m_state->maskBounds = IntRect(0, 0, m_currentSurface->width(), m_currentSurface->height());
        vgSeti(VG_MASKING, VG_TRUE);
        vgMask(VG_INVALID_HANDLE, VG_FILL_MASK, 0, 0, m_currentSurface->width(), m_currentSurface->height());
    }

    // if we can't use vgCopyMask(), store the path instead to rebuild the mask. we
    // limit the total paths stored to 8; if there are more, we evict the oldest under
    // the assumption that the newer paths are smaller sub-regions intersecting it.
    if (VGUtils::bugDisableCopyMask) {
        if (m_state->maskPaths.size() >= 8) {
            delete m_state->maskPaths[0];
            m_state->maskPaths.remove(0);
        }
        m_state->maskPaths.append(new CachedClippingPath(path, maskOp, m_state->surfaceTransformation));
    }

    // Intersect the path from the mask, or subtract it from there.
    // (In either case we always decrease the visible area, never increase it,
    // which means masking never has to modify scissor rectangles.)
    vgSeti(VG_FILL_RULE, toVGFillRule(clipRule));
    ASSERT_VG_NO_ERROR();
    vgRenderToMask(vgPath, VG_FILL_PATH, (VGMaskOperation) maskOp);
    ASSERT_VG_NO_ERROR();

    // If the path is intersected, we can also intersect the bounds to
    // decrease the portion of the mask that needs to be saved.
    if (maskOp == IntersectClip) {

        // get path bounds
        VGfloat minX;
        VGfloat minY;
        VGfloat width;
        VGfloat height;
        vgPathBounds(vgPath, &minX, &minY, &width, &height);
        FloatRect bounds(minX, minY, width, height);

        // transform rect
        VGfloat data[9];
        vgSeti(VG_MATRIX_MODE, VG_MATRIX_PATH_USER_TO_SURFACE);
        vgGetMatrix(data);
        bounds = AffineTransform(VGMatrix(data)).mapRect(bounds);

        m_state->maskBounds.intersect(enclosingIntRect(bounds));
    }
    #else
    m_state->maskingEnabled = true;
    notImplemented();
    #endif
}

DLL_PUBLIC void PainterOpenVG::setCustomMaskApplied(const IntRect& bounds, ClipOperation maskOp)
{
    m_state->maskingEnabled = true;

    #if defined(OPENVG_VERSION_1_1) && !PAINTER_DISABLE_MASKING
    // note that custom masks can't be rebuilt if copy mask is disabled.

    if (!m_state->maskingEnabled)
        m_state->maskBounds = IntRect(0, 0, m_currentSurface->width(), m_currentSurface->height());
    if (maskOp == IntersectClip)
        m_state->maskBounds.intersect(bounds);

    m_state->modifiedStateCategories |= MaskState;

    // destroy any cached mask
    if (m_state->mask != VG_INVALID_HANDLE) {
        vgDestroyMaskLayer(m_state->mask);
        m_state->mask = VG_INVALID_HANDLE;
        ASSERT_VG_NO_ERROR();
    }
    #endif
}

void PainterOpenVG::drawRect(const FloatRect& rect, VGbitfield specifiedPaintModes)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    if (debugDisabled())
        return;
    ASSERT(m_state);

    if (rect.isEmpty())
        return;

    VGbitfield paintModes = 0;
    if (!m_state->strokeDisabled())
        paintModes |= VG_STROKE_PATH;
    if (!m_state->fillDisabled())
        paintModes |= VG_FILL_PATH;

    paintModes &= specifiedPaintModes;

    if (!paintModes)
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();

        // Improve draw performance for shadows of patterns and gradients.
        if (m_state->compositeOperation == CompositeCopy
            && m_state->fillPaint.type() != PaintOpenVG::ColorPaint) {
            // Fill with any opaque color, the color transformation does the rest.
            setFillColor(Color::black);
        }

        drawRect(rect, specifiedPaintModes);
        endDrawShadow();
    } else
        makeCurrent();

    VGPath path = vgCreatePath(
        VG_PATH_FORMAT_STANDARD, VG_PATH_DATATYPE_F,
        1.0 /* scale */, 0.0 /* bias */,
        5 /* expected number of segments */,
        5 /* expected number of total coordinates */,
        VG_PATH_CAPABILITY_APPEND_TO);
    ASSERT_VG_NO_ERROR();

    if (vguRect(path, rect.x(), rect.y(), rect.width(), rect.height()) == VGU_NO_ERROR) {
        vgDrawPath(path, paintModes);
        painted();
        ASSERT_VG_NO_ERROR();
    }

    vgDestroyPath(path);
    ASSERT_VG_NO_ERROR();

}

void PainterOpenVG::drawRoundedRect(const FloatRect& rect, const IntSize& topLeft, const IntSize& topRight, const IntSize& bottomLeft, const IntSize& bottomRight, VGbitfield specifiedPaintModes)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    if (debugDisabled())
        return;
    ASSERT(m_state);

    VGbitfield paintModes = 0;
    if (!m_state->strokeDisabled())
        paintModes |= VG_STROKE_PATH;
    if (!m_state->fillDisabled())
        paintModes |= VG_FILL_PATH;

    paintModes &= specifiedPaintModes;

    if (!paintModes)
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();
        drawRoundedRect(rect, topLeft, topRight, bottomLeft, bottomRight, specifiedPaintModes);
        endDrawShadow();
    } else
        makeCurrent();

    VGPath path = vgCreatePath(
        VG_PATH_FORMAT_STANDARD, VG_PATH_DATATYPE_F,
        1.0 /* scale */, 0.0 /* bias */,
        10 /* expected number of segments */,
        25 /* expected number of total coordinates */,
        VG_PATH_CAPABILITY_APPEND_TO);
    ASSERT_VG_NO_ERROR();

    // clamp corner arc sizes
    FloatSize clampedTopLeft = FloatSize(topLeft).shrunkTo(rect.size()).expandedTo(FloatSize());
    FloatSize clampedTopRight = FloatSize(topRight).shrunkTo(rect.size()).expandedTo(FloatSize());
    FloatSize clampedBottomLeft = FloatSize(bottomLeft).shrunkTo(rect.size()).expandedTo(FloatSize());
    FloatSize clampedBottomRight = FloatSize(bottomRight).shrunkTo(rect.size()).expandedTo(FloatSize());

    // As OpenVG's coordinate system is flipped in comparison to WebKit's,
    // we have to specify the opposite value for the "clockwise" value.
    static const VGubyte pathSegments[] = {
        VG_MOVE_TO_ABS,
        VG_HLINE_TO_REL,
        VG_SCCWARC_TO_REL,
        VG_VLINE_TO_REL,
        VG_SCCWARC_TO_REL,
        VG_HLINE_TO_REL,
        VG_SCCWARC_TO_REL,
        VG_VLINE_TO_REL,
        VG_SCCWARC_TO_REL,
        VG_CLOSE_PATH
    };
    // Also, the rounded rectangle path proceeds from the top to the bottom,
    // requiring height distances and clamped radius sizes to be flipped.
    const VGfloat pathData[] = {
        rect.x() + clampedTopLeft.width(), rect.y(),
        rect.width() - clampedTopLeft.width() - clampedTopRight.width(),
        clampedTopRight.width(), clampedTopRight.height(), 0, clampedTopRight.width(), clampedTopRight.height(),
        rect.height() - clampedTopRight.height() - clampedBottomRight.height(),
        clampedBottomRight.width(), clampedBottomRight.height(), 0, -clampedBottomRight.width(), clampedBottomRight.height(),
        -(rect.width() - clampedBottomLeft.width() - clampedBottomRight.width()),
        clampedBottomLeft.width(), clampedBottomLeft.height(), 0, -clampedBottomLeft.width(), -clampedBottomLeft.height(),
        -(rect.height() - clampedTopLeft.height() - clampedBottomLeft.height()),
        clampedTopLeft.width(), clampedTopLeft.height(), 0, clampedTopLeft.width(), -clampedTopLeft.height(),
    };

    vgAppendPathData(path, 10, pathSegments, pathData);
    vgDrawPath(path, paintModes);
    painted();
    vgDestroyPath(path);
    ASSERT_VG_NO_ERROR();
}

void PainterOpenVG::drawLine(const IntPoint& from, const IntPoint& to)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    if (debugDisabled())
        return;
    ASSERT(m_state);

    if (m_state->strokeDisabled())
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();
        drawLine(from, to);
        endDrawShadow();
    } else
        makeCurrent();

    VGPath path = vgCreatePath(
        VG_PATH_FORMAT_STANDARD, VG_PATH_DATATYPE_F,
        1.0 /* scale */, 0.0 /* bias */,
        2 /* expected number of segments */,
        4 /* expected number of total coordinates */,
        VG_PATH_CAPABILITY_APPEND_TO);
    ASSERT_VG_NO_ERROR();

    if (vguLine(path, from.x(), from.y(), to.x(), to.y()) == VGU_NO_ERROR) {
        vgDrawPath(path, VG_STROKE_PATH);
        painted();
        ASSERT_VG_NO_ERROR();
    }

    vgDestroyPath(path);
    ASSERT_VG_NO_ERROR();
}

void PainterOpenVG::drawArc(const IntRect& rect, int startAngle, int angleSpan, VGbitfield specifiedPaintModes)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    if (debugDisabled())
        return;
    ASSERT(m_state);

    VGbitfield paintModes = 0;
    if (!m_state->strokeDisabled())
        paintModes |= VG_STROKE_PATH;
    if (!m_state->fillDisabled())
        paintModes |= VG_FILL_PATH;

    paintModes &= specifiedPaintModes;

    if (!paintModes)
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();
        drawArc(rect, startAngle, angleSpan, paintModes);
        endDrawShadow();
    } else
        makeCurrent();

    VGPath path = vgCreatePath(
        VG_PATH_FORMAT_STANDARD, VG_PATH_DATATYPE_F,
        1.0 /* scale */, 0.0 /* bias */,
        2 /* expected number of segments */,
        4 /* expected number of total coordinates */,
        VG_PATH_CAPABILITY_APPEND_TO);
    ASSERT_VG_NO_ERROR();

    if (vguArc(path, rect.x() + rect.width() / 2.0, rect.y() + rect.height() / 2.0, rect.width(), rect.height(), -startAngle, -angleSpan, VGU_ARC_OPEN) == VGU_NO_ERROR) {
        vgDrawPath(path, VG_STROKE_PATH);
        painted();
        ASSERT_VG_NO_ERROR();
    }

    vgDestroyPath(path);
    ASSERT_VG_NO_ERROR();
}

void PainterOpenVG::drawEllipse(const IntRect& rect, VGbitfield specifiedPaintModes)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    if (debugDisabled())
        return;
    ASSERT(m_state);

    VGbitfield paintModes = 0;
    if (!m_state->strokeDisabled())
        paintModes |= VG_STROKE_PATH;
    if (!m_state->fillDisabled())
        paintModes |= VG_FILL_PATH;

    paintModes &= specifiedPaintModes;

    if (!paintModes)
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();
        drawEllipse(rect, paintModes);
        endDrawShadow();
    } else
        makeCurrent();

    VGPath path = vgCreatePath(
        VG_PATH_FORMAT_STANDARD, VG_PATH_DATATYPE_F,
        1.0 /* scale */, 0.0 /* bias */,
        4 /* expected number of segments */,
        12 /* expected number of total coordinates */,
        VG_PATH_CAPABILITY_APPEND_TO);
    ASSERT_VG_NO_ERROR();

    if (vguEllipse(path, rect.x() + rect.width() / 2.0, rect.y() + rect.height() / 2.0, rect.width(), rect.height()) == VGU_NO_ERROR) {
        vgDrawPath(path, paintModes);
        painted();
        ASSERT_VG_NO_ERROR();
    }

    vgDestroyPath(path);
    ASSERT_VG_NO_ERROR();
}

void PainterOpenVG::drawPolygon(size_t numPoints, const FloatPoint* points, VGbitfield specifiedPaintModes)
{
    #if PAINTER_DISABLE_PATHS
    return;
    #endif
    if (debugDisabled())
        return;
    ASSERT(m_state);

    VGbitfield paintModes = 0;
    if (!m_state->strokeDisabled())
        paintModes |= VG_STROKE_PATH;
    if (!m_state->fillDisabled())
        paintModes |= VG_FILL_PATH;

    paintModes &= specifiedPaintModes;

    if (!paintModes)
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();
        drawPolygon(numPoints, points, paintModes);
        endDrawShadow();
    } else
        makeCurrent();

    // Path segments: all points + "close path".
    const VGint numSegments = numPoints + 1;
    const VGint numCoordinates = numPoints * 2;

    VGPath path = vgCreatePath(
        VG_PATH_FORMAT_STANDARD, VG_PATH_DATATYPE_F,
        1.0 /* scale */, 0.0 /* bias */,
        numSegments /* expected number of segments */,
        numCoordinates /* expected number of total coordinates */,
        VG_PATH_CAPABILITY_APPEND_TO);
    ASSERT_VG_NO_ERROR();

    Vector<VGfloat, 32> vgPoints(numCoordinates);
    for (size_t i = 0; i < numPoints; ++i) {
        vgPoints[i*2]     = points[i].x();
        vgPoints[i*2 + 1] = points[i].y();
    }

    if (vguPolygon(path, vgPoints.data(), numPoints, VG_TRUE /* closed */) == VGU_NO_ERROR) {
        vgDrawPath(path, paintModes);
        painted();
        ASSERT_VG_NO_ERROR();
    }

    vgDestroyPath(path);
    ASSERT_VG_NO_ERROR();
}

DLL_PUBLIC void PainterOpenVG::drawImage(TiledImageOpenVG* tiledImage, const FloatRect& dst, const FloatRect& src)
{
    #if PAINTER_DISABLE_IMAGES
    return;
    #endif
    if (debugDisabled())
        return;
    if (debugDisabled()) return;
    ASSERT(m_state);

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();

        if (m_state->compositeOperation == CompositeCopy) {
            // Fill with any opaque color, the color transformation does the rest.
            setFillColor(Color::black);
            drawRect(dst, VG_FILL_PATH);
        } else
            drawImage(tiledImage, dst, src);

        endDrawShadow();
    } else
        makeCurrent();

    // If buffers can be larger than the maximum OpenVG image sizes,
    // we split them into tiles.
    IntRect drawnTiles = tiledImage->tilesInRect(src);
    AffineTransform srcToDstTransformation = makeMapBetweenRects(
        FloatRect(FloatPoint(0.0, 0.0), src.size()), dst);
    srcToDstTransformation.translate(-src.x(), -src.y());

    // some platforms have inverted pbuffer surfaces
    if (tiledImage->inverted()) {
        srcToDstTransformation.translate(0.0f, dst.height());
        srcToDstTransformation.scale(1.0f, -1.0f);
    }

    for (int yIndex = drawnTiles.y(); yIndex < drawnTiles.maxY(); ++yIndex) {
        for (int xIndex = drawnTiles.x(); xIndex < drawnTiles.maxX(); ++xIndex) {
            // The srcTile rectangle is an aligned tile cropped by the src rectangle.
            FloatRect tile(tiledImage->tileRect(xIndex, yIndex));
            FloatRect srcTile = intersection(src, tile);

            save();

            // If the image is drawn in full, all we need is the proper transformation
            // in order to get it drawn at the right spot on the surface.
            concatTransformation(AffineTransform(srcToDstTransformation).translate(tile.x(), tile.y()));

            // If only a part of the tile is drawn, we also need to clip the surface.
            if (srcTile != tile) {
                // Put boundaries relative to tile origin, as we already
                // translated to (x, y) with the transformation matrix.
                srcTile.move(-tile.x(), -tile.y());
                clipRect(srcTile, PainterOpenVG::IntersectClip);
            }

            VGImage image = tiledImage->tile(xIndex, yIndex);
            if (image != VG_INVALID_HANDLE) {
                vgDrawImage(image);
                painted();
                ASSERT_VG_NO_ERROR();
            }

            restore();
        }
    }
}

#if 0
void vgDrawGlyphs(VGFont font, VGint glyphCount, const VGuint *glyphIndices,
        const VGfloat *adjustments_x, const VGfloat *adjustments_y,
        VGbitfield paintModes, VGboolean allowAutoHinting)
{
    ESPIAL_PROFILE();
    ::vgDrawGlyphs(font, glyphCount, glyphIndices, adjustments_x, adjustments_y, paintModes, allowAutoHinting);
}
#endif

#ifdef OPENVG_VERSION_1_1
void PainterOpenVG::drawText(VGFont vgFont, size_t count, VGuint* characters, VGfloat* adjustmentsX, VGfloat* adjustmentsY, const FloatPoint& point)
{
    #if PAINTER_DISABLE_TEXT
    return;
    #endif
    if (debugDisabled())
        return;
    ASSERT(m_state);

    VGbitfield paintModes = 0;

    if (m_state->textDrawingMode & cTextClip)
        return; // unsupported for every port except CG at the time of writing
    if (m_state->textDrawingMode & cTextFill && !m_state->fillDisabled())
        paintModes |= VG_FILL_PATH;
    if (m_state->textDrawingMode & cTextStroke && !m_state->strokeDisabled())
        paintModes |= VG_STROKE_PATH;

    if (!paintModes)
        return;

    if (shadowEnabled() && !m_drawingShadow) {
        beginDrawShadow();
        drawText(vgFont, count, characters, adjustmentsX, adjustmentsY, point);
        endDrawShadow();
    } else
        makeCurrent();

    FloatPoint p = point;

    // disabled for the same reason as autohinting
#if 0
    FloatPoint effectivePoint = m_state->surfaceTransformation.mapPoint(point);
    AffineTransform* originalTransformation = 0;

    // In case the font isn't drawn at a pixel-exact baseline and we can easily
    // fix that (which is the case for non-rotated affine transforms), let's
    // align the starting point to the pixel boundary in order to prevent
    // font rendering issues such as glyphs that appear off by a pixel.
    // This causes us to have inconsistent spacing between baselines in a
    // larger paragraph, but that seems to be the least of all evils.
    if ((fmod(effectivePoint.x() + 0.01, 1.0) > 0.02 || fmod(effectivePoint.y() + 0.01, 1.0) > 0.02)
        && isNonRotatedAffineTransformation(m_state->surfaceTransformation))
    {
        originalTransformation = new AffineTransform(m_state->surfaceTransformation);
        setSurfaceTransformation(AffineTransform(
            m_state->surfaceTransformation.a(), 0,
            0, m_state->surfaceTransformation.d(),
            roundf(effectivePoint.x()), roundf(effectivePoint.y())));
        p = FloatPoint();
    }
#endif

    const VGfloat vgPoint[2] = { p.x(), p.y() };
    vgSetfv(VG_GLYPH_ORIGIN, 2, vgPoint);

    if(VGUtils::enableFontAntialiased)
    	vgSetf (VG_RENDERING_QUALITY, VG_RENDERING_QUALITY_BETTER);
    else
    	vgSetf (VG_RENDERING_QUALITY, VG_RENDERING_QUALITY_NONANTIALIASED);

    ASSERT_VG_NO_ERROR();

    // I've disabled autohinting for now. Hinting will make things look worse when
    // the browser surface isn't aligned to the pixel grid, and it will throw off
    // our text size metrics because we don't use OpenVG to measure string sizes
    // (and OpenVG hinting is allowed to modify escapement values as it measures
    // and draws hinted glyphs.)

    // Some platforms (e.g. Broadcom) will freeze if too many glyphs are drawn in
    // one call to vgDrawGlyphs, and some glyphs will just not be drawn if they
    // aren't passed one at a time.
    if (VGUtils::bugIndividualGlyphs) {
        for (size_t i = 0; i < count; ++i) {
            vgDrawGlyphs(vgFont,
                         1,
                         characters + i,
                         adjustmentsX ? adjustmentsX + i : NULL,
                         adjustmentsY ? adjustmentsY + i : NULL,
                         paintModes,
                         VG_FALSE /* disable autohinting */);
        }
    } else {
        vgDrawGlyphs(vgFont, count, characters,
            adjustmentsX, adjustmentsY, paintModes, VG_FALSE /* disable autohinting */);
    }
    painted();

    // Drawing glyphs currently fails on some fonts for unknown reasons.
    // For now we will ignore glyph drawing failures.
    VGErrorCode error = vgGetError();
    if (error != VG_NO_ERROR) {
        // ignore
    }

#if 0
    if (originalTransformation) {
        setSurfaceTransformation(*originalTransformation);
        delete originalTransformation;
    }
#endif

	vgSetf(VG_RENDERING_QUALITY, VG_RENDERING_QUALITY_BETTER);

    ASSERT_VG_NO_ERROR();
}
#endif

NativeImagePtr PainterOpenVG::asNewNativeImage(const IntRect& src, VGImageFormat format)
{
    return asNewNativeImage(m_baseSurface, src, format);
}

NativeImagePtr PainterOpenVG::asNewNativeImage(SurfaceOpenVG* surface, const IntRect& src, VGImageFormat format)
{
    ASSERT(m_state);

    // We want our own surface/context current to create VGImages and read pixels
    makeCurrent(surface);

    const IntSize vgMaxImageSize(vgGeti(VG_MAX_IMAGE_WIDTH), vgGeti(VG_MAX_IMAGE_HEIGHT));
    ASSERT_VG_NO_ERROR();

    const IntRect rect = intersection(src, IntRect(0, 0, surface->width(), surface->height()));
    TiledImageOpenVG* tiledImage = new TiledImageOpenVG(rect.size(), vgMaxImageSize);

    const int numColumns = tiledImage->numColumns();
    const int numRows = tiledImage->numRows();

    // Create the images as resources of the shared surface/context.
    for (int yIndex = 0; yIndex < numRows; ++yIndex) {
        for (int xIndex = 0; xIndex < numColumns; ++xIndex) {
            IntRect tileRect = tiledImage->tileRect(xIndex, yIndex);
            VGImage image = vgCreateImage(format, tileRect.width(), tileRect.height(), WEBKIT_VG_IMAGE_QUALITY);
            ASSERT_VG_NO_ERROR();

            tiledImage->setTile(xIndex, yIndex, image);
        }
    }

    // Fill the image contents with our own surface/context being current.
    for (int yIndex = 0; yIndex < numRows; ++yIndex) {
        for (int xIndex = 0; xIndex < numColumns; ++xIndex) {
            IntRect tileRect = tiledImage->tileRect(xIndex, yIndex);
            #if PAINTER_DISABLE_READBACK
            continue;
            #endif
            vgGetPixels(tiledImage->tile(xIndex, yIndex), 0, 0,
                rect.x() + tileRect.x(), rect.y() + tileRect.y(),
                tileRect.width(), tileRect.height());
            ASSERT_VG_NO_ERROR();
        }
    }

    return tiledImage;
}

DLL_PUBLIC void PainterOpenVG::save()
{
    ASSERT(m_state);
    makeCurrent();

    PlatformPainterState* state = new PlatformPainterState(*m_state);
    m_stateStack.append(state);
    m_state = m_stateStack.last();
}

void PainterOpenVG::restore()
{
    ASSERT(m_stateStack.size() >= 2);
    makeCurrent();

    PlatformPainterState* state = m_stateStack.last();
    PainterStateCategoryFlags modifiedStateCategories = state->modifiedStateCategories;
    m_stateStack.removeLast();
    delete state;

    m_state = m_stateStack.last();
    m_state->applyState(this, modifiedStateCategories);
}

void PainterOpenVG::painted()
{
    // we do a quick check to make sure the scissor rect is not empty so we
    // don't trigger painting for stuff that is clipped to off-screen
    if (!m_state->scissoringEnabled || !m_state->scissorRect.isEmpty())
        m_didPaint = true;
}

}
