/*
 * Copyright (C) 2009, 2011, 2012 Google Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "SocketStreamError.h"
#include "SocketStreamHandle.h"
#include "SocketStreamHandleClient.h"
#include <wtf/CurrentTime.h>

namespace WebCore {

const static double sPollInterval = 0.025;
const static unsigned int MAX_BUF_SIZE = 32 * 1024 * 1024; // 32M

DLL_PUBLIC bool SocketStreamHandleBase::send(const char* data, int length)
{
    if (m_state == Closed || m_state == Closing)
        return false;

    if (m_buffer.isEmpty()) {
        int bytesWritten = 0;

        if (m_state == Open){
            bytesWritten = platformSend(data, length);

            if (bytesWritten < 0)
            	return false;
        }

        if (m_buffer.size() + length - bytesWritten > MAX_BUF_SIZE) {
			m_client->didFailSocketStream(static_cast<SocketStreamHandle*>(this), SocketStreamError(0, String(), String("send error : exceeding maximum pending buffer size ") + String::number(MAX_BUF_SIZE)));
            return false;
        }

        if (bytesWritten < length) {
            m_buffer.append(data + bytesWritten, length - bytesWritten);

            if (m_client)
                m_client->didUpdateBufferedAmount(static_cast<SocketStreamHandle*>(this), bufferedAmount());
        }
    }
    else{
        if (m_buffer.size() + length > MAX_BUF_SIZE){
			m_client->didFailSocketStream(static_cast<SocketStreamHandle*>(this), SocketStreamError(0, String(), String("send error : exceeding maximum pending buffer size ") + String::number(MAX_BUF_SIZE)));
            return false;
        }

        m_buffer.append(data, length);

        if (m_client)
            m_client->didUpdateBufferedAmount(static_cast<SocketStreamHandle*>(this), bufferedAmount());
    }

    if (!m_buffer.isEmpty() && !m_sendtimer.isActive())
		m_sendtimer.startOneShot(sPollInterval);

    return true;
}

DLL_PUBLIC void SocketStreamHandleBase::close()
{
    if (m_state == Closed)
        return;
    m_state = Closing;
    if (!m_buffer.isEmpty())
        return;
    disconnect();
}

void SocketStreamHandleBase::disconnect()
{
    RefPtr<SocketStreamHandle> protect(static_cast<SocketStreamHandle*>(this));

    platformClose();
    m_state = Closed;
}

void SocketStreamHandleBase::setClient(SocketStreamHandleClient* client)
{
    ASSERT(!client || (!m_client && m_state == Connecting));
    m_client = client;
}

bool SocketStreamHandleBase::sendPendingDataOnce()
{
    if (m_state != Open && m_state != Closing)
        return false;

    if (m_buffer.isEmpty())
    	return false;

	int bytesWritten = platformSend(m_buffer.firstBlockData(), m_buffer.firstBlockSize());

	if (bytesWritten <= 0)
		return false;

	m_buffer.consume(bytesWritten);

    if (m_client)
        m_client->didUpdateBufferedAmount(static_cast<SocketStreamHandle*>(this), bufferedAmount());

    return true;
}

void SocketStreamHandleBase::sendTimerFired(Timer<SocketStreamHandleBase>*)
{
	RefPtr<SocketStreamHandle> protector(static_cast<SocketStreamHandle*>(this));

	double endTime = monotonicallyIncreasingTime() + sPollInterval;

    while(!m_buffer.isEmpty()){
    	if(!sendPendingDataOnce())
    		break;

    	if(monotonicallyIncreasingTime() >= endTime)
    		break;
    }

    if (!m_buffer.isEmpty() && !m_sendtimer.isActive())
		m_sendtimer.startOneShot(sPollInterval);
}

}  // namespace WebCore
