/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2012 Espial Group Inc. All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ResourceHandleManager_h
#define ResourceHandleManager_h

#include "PlatformString.h"
#include "Timer.h"
#include "ResourceHandleClient.h"

#if PLATFORM(WIN)
#include <winsock2.h>
#include <windows.h>
#endif

#if HAVE(OPENSSL)
#include <openssl/x509.h>
#endif
extern "C" {
typedef void CURL;
typedef void CURLM;
typedef void CURLSH;
struct curl_slist;
}
#include <wtf/DoublyLinkedList.h>
#include <wtf/Vector.h>
#include <wtf/text/CString.h>
#include <ResourceHandle.h>
#include <wtf/HashMap.h>

namespace WebCore {

class Document;
class KURL;
class Frame;
class ResourceHandle;
class ResourceRequest;

class ResourceHandleManager {
	WTF_MAKE_FAST_ALLOCATED;
	WTF_MAKE_NONCOPYABLE(ResourceHandleManager);

public:
    enum ProxyType {
        HTTP,
        Socks4,
        Socks4A,
        Socks5,
        Socks5Hostname
    };
    static ResourceHandleManager& sharedInstance();
    void add(ResourceHandle*);
    void cancel(ResourceHandle*);
    void setCookieJarFileName(const char* cookieJarFileName);
    void setCookieJarNetTVFileName(const char* cookieJarFileName);
    
    void defersLoading(ResourceHandle* job, bool defers);

    void init();
    void shutdown();

#if HAVE(OPENSSL)
    void pauseForSSLVerify();
    void resumeFromSSLVerify();
#endif
    
    void dispatchSynchronousJob(ResourceHandle*);
    
    void rootCertificatesChanged();

    void setCookie(const char* domain, const char* path, const char* line, bool fixed);
    void deleteCookie(const char* domain, const char* path, const char* name);

    void setProxyInfo(const String& httpHost = "",
                      unsigned long httpPort = 0,
                      const String& httpsHost = "",
                      unsigned long httpsPort = 0,
                      ProxyType type = HTTP);
    
    void setExcludeProxyList(const String& excludeProxyList);

    void setCookies(const KURL&, const String&);
    String cookies(const KURL&, bool httpOnly);
    String cookies(const char* host, const char* path, bool secure, bool httpOnly);
    
    void downloadTimerCallback();

    bool doSelect();

    void setCookiesEnabled(bool enable);
    void setThirdPartyCookiesEnabled(bool enable);
    void clearCookies();
    bool cookieEnabled() const { return m_cookiesEnabled; }

    void setExtendedHttpHeaders(const Vector<String>& headers, const String& domain);
    void addExtendedHttpHeaders(const Vector<String>& headers, const String& domain);

    void addExtendedHttpHeader(const String& header, const String& domain);

    void initializeHttpAuthentication(ResourceHandle*, const ProtectionSpace*);
    void initializeProxy(ResourceHandle*) const;
    void initializeExtendedHeaders(ResourceHandle*);

    //! Returns true if any downloads are in progress or scheduled to start;
    //! false if there are no resources to download or if all downloads are
    //! paused for SSL verification or HTTP authentication.
    bool downloadsRunning();

    void saveCookiesTimerCallback(Timer<ResourceHandleManager>*);
    
    void connectionInactiveTimerCallback(Timer<ResourceHandleManager>*);

private:
    ResourceHandleManager();
    ~ResourceHandleManager();
    bool removeFromCurl(ResourceHandle*);
    bool removeScheduledJob(ResourceHandle*);
    bool startJob(ResourceHandle*, long useFreshConnect);
    bool startScheduledJobs(bool async);

    bool initializeHandle(ResourceHandle*, const ProtectionSpace*, long useFreshConnect);
    
#if HAVE(OPENSSL)
    static int ssl_verify_callback(int preverify_ok, X509_STORE_CTX *ctx);
#endif
    
    static ResourceHandle* findHandleByCURL(CURL* curl /* easy handle */);

    static ResourceHandle* findHandleBySSL(void *ssl /* actually an OpenSSL SSL */);
    
#if HAVE(OPENSSL)
    static int ssl_ctx_callback(	CURL *curl,    /* easy handle */
    									void *ssl_ctx, /* actually an OpenSSL SSL_CTX */
								        void *ssl,     /* actually an OpenSSL SSL */
								        void *userptr);
#endif

	enum {
		TransferType_Header = 0x1,
		TransferType_Data   = 0x2,
		TransferType_Msg    = 0x4
	};

	struct TransferBuffer : public DoublyLinkedListNode<TransferBuffer>
	{
		friend class DoublyLinkedListNode<TransferBuffer>;
		WTF_MAKE_NONCOPYABLE(TransferBuffer);
	public:
		TransferBuffer(RefPtr<ResourceHandle> job, int type);
		TransferBuffer(char* ptr, size_t totalSize, const String& effectiveURL, RefPtr<ResourceHandle> job, int type);
		TransferBuffer(int result, unsigned long errdetail, const String& effectiveURL, RefPtr<ResourceHandle> job, int type);
		~TransferBuffer();

	private:
		TransferBuffer* m_prev;
		TransferBuffer* m_next;

	public:
		RefPtr<ResourceHandle> m_job;
		int m_transferGeneration;
		int m_type;
    	int m_result;
    	unsigned long m_errdetail;
		String m_effectiveURL;
		Vector<char> m_ptr;
	};

	void appendTransferBuffer(TransferBuffer* buffer);
	void appendTransferBufferHeader(RefPtr<ResourceHandle> job);
	void appendTransferBufferData(char* ptr, size_t totalSize, const String& effectiveURL, RefPtr<ResourceHandle> job);
	void appendTransferBufferMsg(int result, unsigned long errdetail, const String& effectiveURL, RefPtr<ResourceHandle> job);

    static size_t curlWriteCallback(void* ptr, size_t size, size_t nmemb, void* data, int flags);
    static size_t curlHeaderCallback(char* ptr, size_t size, size_t nmemb, void* data, int flags);
    
    void headerCallback(RefPtr<ResourceHandle> job);
    void writeCallback(char* ptr, size_t totalSize, const String& effectiveURL, RefPtr<ResourceHandle> job);

    void handleCURLDone(int result, unsigned long errdetail, const String& effectiveURL, RefPtr<ResourceHandle> job);

    void readCURLMsgs();

    static void curl_ready_http_callback(CURL *curl, /* easy handler */
    										void* SSL /*if secure*/);
    
    static int curl_send_cookie_function(CURL *curl,       /* easy handler */
								  		const char* domain,
								        const char* name,
								        int cookie_type);
    static int curl_cookie_type_callback(CURL *curl, const char* domain, const char* name, int replace_old);

    void callCURLMultiPerform();

    void checkStallTimeout();

    void ensureDummyCurlhandle();

    CURLM* m_curlMultiHandle;
    CURLSH* m_curlShareHandle;
    CURL* m_dummyCurlhandle;
    bool m_rootCertificatesChanged;
    CString m_cookieJarFileName;
    CString m_cookieJarNetTVFileName;
    Vector<RefPtr<ResourceHandle> > m_scheduledResourceHandles;
    Vector<RefPtr<ResourceHandle> > m_runningResourceHandles;
    const CString m_certificatePath;

    bool m_cookiesEnabled;
    CString m_httpProxy;
    CString m_httpsProxy;
    ProxyType m_proxyType;
    
    CString m_preferredSSLCiphers;
    CString m_ciphers;

    bool m_pausedForSSLVerify;

    HashMap<String, Vector<String> > m_extenedHttpHeaders;
    Vector<String> m_allDomainExtendedHttpHeaders;

    HashMap<String, String> m_replaceURLsMap;

    int m_changedRegularCookies;
    int m_changedSpecialCookies;
    Timer<ResourceHandleManager> m_cookiesSaveTimer;
    double m_lastCookieSaveTime;

    Timer<ResourceHandleManager> m_connectionInactiveTimer;

    void handleTransferBuffer(TransferBuffer* cur, bool& shouldScheduleImmediately);
    void handleAllTransferBuffers(bool& shouldScheduleImmediately);

    DoublyLinkedList<TransferBuffer> m_transferBuffers;

    struct HttpAuth
    {
    	HttpAuth(RefPtr<ResourceHandle> job, long httpCode)
    	: m_job(job)
    	, m_httpCode(httpCode)
    	, m_done(false)
    	, m_result(0)
    	, m_errdetail(0)
    	{
    	}

		RefPtr<ResourceHandle> m_job;
    	long m_httpCode;
    	bool m_done;
    	int m_result;
    	unsigned long m_errdetail;
		String m_effectiveURL;
    	Vector<char> m_receivedPtr;
    };

    Vector<HttpAuth> m_pendingHttpAuths;
    Vector<HttpAuth*> m_processingHttpAuths;

    void addHttpAuth(RefPtr<ResourceHandle> job, long httpCode);
    void updateHttpAuth(RefPtr<ResourceHandle> job, char* receivedPtr, size_t receivedSize);
    void updateHttpAuth(RefPtr<ResourceHandle> job, bool done, int result, unsigned long errdetail, const String& effectiveURL);

    void doHttpAuth(const HttpAuth&);
    void httpAuthTimerCallback(Timer<ResourceHandleManager>*);
    Timer<ResourceHandleManager> m_httpAuthTimer;
};

}

#endif
