/*
 * Copyright (C) 2011 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef RenderGrid_h
#define RenderGrid_h

#include "Grid.h"
#include "GridTrackSizingAlgorithm.h"
#include "RenderBlock.h"

namespace WebCore {

class GridArea;
class GridSpan;

struct ContentAlignmentData;

enum GridAxisPosition {GridAxisStart, GridAxisEnd, GridAxisCenter};

class RenderGrid : public RenderBlock {
public:
    explicit RenderGrid(Element&, PassRefPtr<RenderStyle> style);
    virtual ~RenderGrid();

    Element& existingElement() const { return *RenderBlock::element(); }

    virtual void styleDidChange(StyleDifference, const RenderStyle* oldStyle) OVERRIDE;
    virtual void layoutBlock(bool relayoutChildren, LayoutUnit pageLogicalHeight = 0) OVERRIDE;

    virtual bool avoidsFloats() const OVERRIDE { return true; }
    virtual bool canDropAnonymousBlockChild() const OVERRIDE { return false; }

    void dirtyGrid();
    Vector<LayoutUnit> trackSizesForComputedStyle(GridTrackSizingDirection) const;

    const Vector<LayoutUnit>& columnPositions() const { return m_columnPositions; }
    const Vector<LayoutUnit>& rowPositions() const { return m_rowPositions; }

    unsigned autoRepeatCountForDirection(GridTrackSizingDirection direction) const { return m_grid.autoRepeatTracks(direction); }

    // Required by GridTrackSizingAlgorithm. Keep them under control.
    bool isOrthogonalChild(const RenderBox&) const;
    LayoutUnit guttersSize(const Grid&, GridTrackSizingDirection, unsigned startLine, unsigned span) const;
private:
    void element() const WTF_DELETED_FUNCTION;

    virtual const char* renderName() const OVERRIDE;
    virtual bool isRenderGrid() const OVERRIDE { return true; }
    virtual void computeIntrinsicLogicalWidths(LayoutUnit& minLogicalWidth, LayoutUnit& maxLogicalWidth) const OVERRIDE;
	virtual void computePreferredLogicalWidths() OVERRIDE;

	virtual void addChild(RenderObject* newChild, RenderObject* beforeChild) FINAL;
	virtual void removeChild(RenderObject*) FINAL;

    bool explicitGridDidResize(const RenderStyle&) const;
    bool namedGridLinesDefinitionDidChange(const RenderStyle&) const;

	virtual Optional<LayoutUnit> computeIntrinsicLogicalContentHeightUsing(Length logicalHeightLength, Optional<LayoutUnit> intrinsicContentHeight, LayoutUnit borderAndPadding) const OVERRIDE;

    unsigned computeAutoRepeatTracksCount(GridTrackSizingDirection, SizingOperation) const;

    PassOwnPtr<OrderedTrackIndexSet> computeEmptyTracksForAutoRepeat(Grid&, GridTrackSizingDirection) const;

    void placeItemsOnGrid(Grid&, SizingOperation) const;
    void populateExplicitGridAndOrderIterator(Grid&) const;
    PassOwnPtr<GridArea> createEmptyGridAreaAtSpecifiedPositionsOutsideGrid(Grid&, const RenderBox*, GridTrackSizingDirection, const GridSpan&) const;
    void placeSpecifiedMajorAxisItemsOnGrid(Grid&, Vector<RenderBox*>) const;
    void placeAutoMajorAxisItemsOnGrid(Grid&, Vector<RenderBox*>) const;
    typedef std::pair<size_t, size_t> AutoPlacementCursor;
    void placeAutoMajorAxisItemOnGrid(Grid&, RenderBox*, AutoPlacementCursor&) const;
    GridTrackSizingDirection autoPlacementMajorAxisDirection() const;
    GridTrackSizingDirection autoPlacementMinorAxisDirection() const;

    virtual bool canPerformSimplifiedLayout() const FINAL;
	void prepareChildForPositionedLayout(RenderBox&);
	virtual void layoutPositionedObject(RenderBox*, bool relayoutChildren, bool fixedPositionObjectsOnly) OVERRIDE;
	void offsetAndBreadthForPositionedChild(const RenderBox&, GridTrackSizingDirection, LayoutUnit& offset, LayoutUnit& breadth);

    void computeTrackSizesForDefiniteSize(GridTrackSizingDirection, LayoutUnit availableSpace);
    void computeTrackSizesForIndefiniteSize(GridTrackSizingAlgorithm&, GridTrackSizingDirection, Grid&, LayoutUnit& minIntrinsicSize, LayoutUnit& maxIntrinsicSize) const;
    LayoutUnit computeTrackBasedLogicalHeight() const;

    void repeatTracksSizingIfNeeded(LayoutUnit availableSpaceForColumns, LayoutUnit availableSpaceForRows);

    void layoutGridItems();
    void populateGridPositionsForDirection(GridTrackSizingDirection);

    GridAxisPosition columnAxisPositionForChild(const RenderBox&) const;
	GridAxisPosition rowAxisPositionForChild(const RenderBox&) const;
    LayoutUnit columnAxisOffsetForChild(const RenderBox&) const;
    LayoutUnit rowAxisOffsetForChild(const RenderBox&) const;
    ContentAlignmentData computeContentPositionAndDistributionOffset(GridTrackSizingDirection, const LayoutUnit& availableFreeSpace, unsigned numberOfGridTracks) const;
	LayoutPoint findChildLogicalPosition(const RenderBox&) const;
	GridSpan cachedGridSpan(const RenderBox&, GridTrackSizingDirection) const;

    LayoutUnit gridAreaBreadthForChildIncludingAlignmentOffsets(const RenderBox&, GridTrackSizingDirection) const;

    void applyStretchAlignmentToTracksIfNeeded(GridTrackSizingDirection);

    virtual void paintChildren(PaintInfo& forSelf, const LayoutPoint& paintOffset, PaintInfo& forChild, bool usePrintRect) OVERRIDE FINAL;
	bool needToStretchChildLogicalHeight(const RenderBox&) const;
	LayoutUnit marginLogicalHeightForChild(const RenderBox&) const;
	LayoutUnit computeMarginLogicalSizeForChild(GridTrackSizingDirection, const RenderBox&) const;
	LayoutUnit availableAlignmentSpaceForChildBeforeStretching(LayoutUnit gridAreaBreadthForChild, const RenderBox&) const;
	StyleSelfAlignmentData justifySelfForChild(const RenderBox&) const;
	StyleSelfAlignmentData alignSelfForChild(const RenderBox&) const;
	void applyStretchAlignmentToChildIfNeeded(RenderBox&);
	bool hasAutoSizeInColumnAxis(const RenderBox& child) const { return isHorizontalWritingMode() ? child.style()->height().isAuto() : child.style()->width().isAuto(); }
	bool hasAutoSizeInRowAxis(const RenderBox& child) const { return isHorizontalWritingMode() ? child.style()->width().isAuto() : child.style()->height().isAuto(); }
	bool allowedToStretchChildAlongColumnAxis(const RenderBox& child) const { return alignSelfForChild(child).position() == ItemPositionStretch && hasAutoSizeInColumnAxis(child) && !hasAutoMarginsInColumnAxis(child); }
	bool allowedToStretchChildAlongRowAxis(const RenderBox& child) const { return justifySelfForChild(child).position() == ItemPositionStretch && hasAutoSizeInRowAxis(child) && !hasAutoMarginsInRowAxis(child); }
	bool hasAutoMarginsInColumnAxis(const RenderBox&) const;
	bool hasAutoMarginsInRowAxis(const RenderBox&) const;
    void resetAutoMarginsAndLogicalTopInColumnAxis(RenderBox& child);
    void updateAutoMarginsInColumnAxisIfNeeded(RenderBox&);
    void updateAutoMarginsInRowAxisIfNeeded(RenderBox&);

    virtual int baselinePosition(FontBaseline, bool firstLine, LineDirectionMode, LinePositionMode = PositionOnContainingLine) const FINAL;
    virtual Optional<int> firstLineBaseline() const FINAL;
    virtual Optional<int> inlineBlockBaseline(LineDirectionMode) const FINAL;
    bool isInlineBaselineAlignedChild(const RenderBox&) const;

    LayoutUnit gridGapForDirection(GridTrackSizingDirection) const;

    unsigned numTracks(GridTrackSizingDirection, const Grid&) const;

    LayoutUnit translateRTLCoordinate(LayoutUnit) const;

    GridTrackSizingDirection flowAwareDirectionForChild(const RenderBox&, GridTrackSizingDirection) const;

    Grid m_grid;

    GridTrackSizingAlgorithm m_trackSizingAlgorithm;

    Vector<LayoutUnit> m_columnPositions;
    Vector<LayoutUnit> m_rowPositions;
    LayoutUnit m_offsetBetweenColumns;
    LayoutUnit m_offsetBetweenRows;

    Optional<LayoutUnit> m_minContentHeight;
    Optional<LayoutUnit> m_maxContentHeight;
};

inline RenderGrid* toRenderGrid(RenderObject* object)
{
    ASSERT(!object || object->isRenderGrid());
    return static_cast<RenderGrid*>(object);
}

inline const RenderGrid* toRenderGrid(const RenderObject* object)
{
    ASSERT(!object || object->isRenderGrid());
    return static_cast<const RenderGrid*>(object);
}

inline RenderGrid& toRenderGrid(RenderObject& object)
{
    ASSERT(object.isRenderGrid());
    return static_cast<RenderGrid&>(object);
}

inline const RenderGrid& toRenderGrid(const RenderObject& object)
{
    ASSERT(object.isRenderGrid());
    return static_cast<const RenderGrid&>(object);
}

// This will catch anyone doing an unnecessary cast.
void toRenderGrid(const RenderGrid*);
void toRenderGrid(const RenderGrid&);

template<> inline bool isRendererOfType<const RenderGrid>(const RenderObject& renderer) { return renderer.isRenderGrid(); }

} // namespace WebCore

#endif // RenderGrid_h
