/*
 * Copyright (C) 2007 Apple Inc.
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2008 Collabora Ltd.
 * Copyright (C) 2009 Kenneth Rohde Christiansen
 * Copyright (C) 2009-2012, Espial Group Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"

#include "RenderBoxModelObject.h"
#include "Path.h"
#include "Color.h"
#include "FloatPoint.h"
#include "IntRect.h"
#include "TransformationMatrix.h"
#include "GraphicsContext.h"
#include "NotImplemented.h"
#include "RenderObject.h"
#include "RenderView.h"
#include "PaintInfo.h"
#include "Image.h"
#include "HTMLNames.h"
#include "RenderThemeEspial.h"
#include <wtf/text/StringBuilder.h>

#if ENABLE(VIDEO)
#include "HTMLMediaElement.h"
#endif

namespace WebCore {

#if ENABLE(VIDEO)
static HTMLMediaElement* getMediaElementFromRenderObject(RenderObject* o)
{
    Node* node = o->node();
    Node* mediaNode = node ? node->shadowAncestorNode() : 0;
    if (!mediaNode || (!mediaNode->hasTagName(HTMLNames::videoTag) && !mediaNode->hasTagName(HTMLNames::audioTag)))
        return 0;

    return static_cast<HTMLMediaElement*>(mediaNode);
}
#endif

DLL_PUBLIC RenderTheme* theRenderTheme()
{
	static RenderThemeEspial theTheme;	
	return &theTheme;
}

RenderThemeEspial::~RenderThemeEspial()
{	
}

RenderThemeEspial::RenderThemeEspial()
{
}

PassRefPtr<RenderTheme> RenderTheme::themeForPage(Page* page)
{
    return theRenderTheme();
}

static void drawImage(GraphicsContext* ctx, Image* image, const IntRect& rect)
{	
	ctx->drawImage(image, rect);
}

static float min(float x, float y){
	return (x<y ? x : y);
}

enum {
    topPadding,
    rightPadding,
    bottomPadding,
    leftPadding
};

enum NSControlSize{
	NSRegularControlSize = 0,
	NSSmallControlSize = 1,    
    NSMiniControlSize = 2
};


static NSControlSize controlSizeForFont(RenderStyle* style)
{
    unsigned fontSize = style->computedFontPixelSize();
    
    if (fontSize >= 16){
        return NSRegularControlSize;
    }
    else if (fontSize >= 11){
        return NSSmallControlSize;
    }
    else{
    	return NSMiniControlSize;
    }
}

static void setSizeFromFont(RenderStyle* style, const IntSize* sizes)
{
    // FIXME: Check is flawed, since it doesn't take min-width/max-width into account.
    IntSize size = sizes[controlSizeForFont(style)];
    
    if (style->width().isIntrinsicOrAuto() && size.width() > 0)
        style->setWidth(Length((int)(size.width() * style->effectiveZoom()), Fixed));
    
    if (style->height().isAuto() && size.height() > 0)
        style->setHeight(Length((int)(size.height() * style->effectiveZoom()), Fixed));
}


bool RenderThemeEspial::supportsFocusRing(const RenderStyle* style) const
{
    return false;
}

static void setButtonPaddingFromControlSize(RenderStyle* style, NSControlSize size)
{
    style->setPaddingTop(Length(0, Fixed));
    style->setPaddingBottom(Length(0, Fixed));
    
	// If the width and height are both specified, then we have nothing to do.
    if (!style->width().isIntrinsicOrAuto() && !style->height().isAuto())
        return;
    
    // Just use 8px.  AppKit wants to use 11px for mini buttons, but that padding is just too large
    // for real-world Web sites (creating a huge necessary minimum width for buttons whose space is
    // by definition constrained, since we select mini only for small cramped environments.
    // This also guarantees the HTML4 <button> will match our rendering by default, since we're using a consistent
    // padding.
    const int padding = 8;
    style->setPaddingLeft(Length(padding, Fixed));
    style->setPaddingRight(Length(padding, Fixed));
}


void RenderThemeEspial::setCheckboxSize(RenderStyle* style) const
{
    static const IntSize sizes[3] = { IntSize(14, 14), IntSize(12, 12), IntSize(10, 10) };
        
    // If the width and height are both specified, then we have nothing to do.
    if (!style->width().isIntrinsicOrAuto() && !style->height().isAuto())
        return;

    // Use the font size to determine the intrinsic width of the control.
    setSizeFromFont(style, sizes);
}

bool RenderThemeEspial::paintCheckbox(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* const ctx = i.context;

	ctx->save();	
	ctx->clip(rect);
	
	if(RenderTheme::isChecked(o)){
		if(m_checkboxYesImage){
			drawImage(ctx, m_checkboxYesImage.get(), rect);
			goto done;
		}
	}
	else{
		if(m_checkboxNoImage){
			drawImage(ctx, m_checkboxNoImage.get(), rect);
			goto done;
		}
	}
	
	// fill white
	ctx->fillRect(rect, Color(255, 255, 255));

	//draw border
	ctx->setStrokeStyle(SolidStroke);
	ctx->setStrokeColor(defaultBorderColor());
	ctx->strokeRect(rect, 1);
		
	if(RenderTheme::isChecked(o)){
		ctx->drawLine(IntPoint(rect.x(), rect.y()), IntPoint(rect.x()+rect.width(), rect.y()+rect.height()));
		ctx->drawLine(IntPoint(rect.x()+rect.width(), rect.y()), IntPoint(rect.x(), rect.y()+rect.height()));
	}

	//std::cout << "RenderTheme::paintCheckbox at(" << rect.x() << "," << rect.y() << ") size(" << rect.width() << "," << rect.height() << ")\n";

done:
	ctx->restore();
	
	return false;
}

void RenderThemeEspial::setRadioSize(RenderStyle* style) const
{
    static const IntSize sizes[3] = { IntSize(14, 15), IntSize(12, 13), IntSize(10, 10) };
    
    // If the width and height are both specified, then we have nothing to do.
    if (!style->width().isIntrinsicOrAuto() && !style->height().isAuto())
        return;

    // Use the font size to determine the intrinsic width of the control.
    setSizeFromFont(style, sizes);
}

bool RenderThemeEspial::paintRadio(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	//std::cout << "RenderTheme::paintRadio at(" << rect.x() << "," << rect.y() << ") size(" << rect.width() << "," << rect.height() << ")\n";
	
	GraphicsContext* const ctx = i.context;

	ctx->save();		
	ctx->clip(rect);

	if(RenderTheme::isChecked(o)){
		if(m_radiobuttonYesImage){
			drawImage(ctx, m_radiobuttonYesImage.get(), rect);
			goto done;
		}
	}
	else{
		if(m_radiobuttonNoImage){
			drawImage(ctx, m_radiobuttonNoImage.get(), rect);
			goto done;
		}
	}
	
	{
		const float halfWidth = rect.width()/2.0;
		const float halfHeight = rect.height()/2.0;
	
		const FloatPoint center(rect.x() + rect.width() - halfWidth, rect.y() + rect.height() - halfHeight);
	
	    ctx->setStrokeStyle(SolidStroke);
	    ctx->setStrokeColor(defaultBorderColor());
	    ctx->setStrokeThickness(1);
	    
	    ctx->strokeArc(IntRect(rect.x()+1, rect.y()+1, rect.width()-2, rect.height()-2), 0, 360);
	    
		if(RenderTheme::isChecked(o)){
			ctx->setFillColor(Color(0, 100, 50));
			
			const int radius = 3;
			
			const IntRect circleRect(center.x() - radius, center.y() - radius,  radius * 2, radius * 2);
		    
		    ctx->drawEllipse(circleRect);
		}
	}
	
done:
	ctx->restore();	
	return false;
}

void RenderThemeEspial::setButtonSize(RenderStyle* style) const
{
    static const IntSize sizes[3] = { IntSize(0, 21), IntSize(0, 18), IntSize(0, 15) };
    
    // If the width and height are both specified, then we have nothing to do.
    if (!style->width().isIntrinsicOrAuto() && !style->height().isAuto())
        return;

    // Use the font size to determine the intrinsic width of the control.
    setSizeFromFont(style, sizes);
}

void RenderThemeEspial::adjustButtonStyle(StyleResolver* selector, RenderStyle* style, WebCore::Element* e) const
{
    // There are three appearance constants for buttons.
    // (1) Push-button is the constant for the default Aqua system button.  Push buttons will not scale vertically and will not allow
    // custom fonts or colors.  <input>s use this constant.  This button will allow custom colors and font weights/variants but won't
    // scale vertically.
    // (2) square-button is the constant for the square button.  This button will allow custom fonts and colors and will scale vertically.
    // (3) Button is the constant that means "pick the best button as appropriate."  <button>s use this constant.  This button will
    // also scale vertically and allow custom fonts and colors.  It will attempt to use Aqua if possible and will make this determination
    // solely on the rectangle of the control.

    // Determine our control size based off our font.
    NSControlSize controlSize = controlSizeForFont(style);

    if(style->appearance() == PushButtonPart) {
        // Ditch the border.
        style->resetBorder();

        style->setBorderRadius(IntSize(buttonCornerRadius() * style->effectiveZoom(), buttonCornerRadius() * style->effectiveZoom()));
        
        style->setBorderLeftStyle(OUTSET);
        style->setBorderRightStyle(OUTSET);
        style->setBorderBottomStyle(OUTSET);
        style->setBorderTopStyle(OUTSET);
        
        style->setBorderLeftColor(buttonBorderColor());
        style->setBorderRightColor(buttonBorderColor());
        style->setBorderBottomColor(buttonBorderColor());
        style->setBorderTopColor(buttonBorderColor());
        
        style->setBorderLeftWidth(buttonBorderWidth() * style->effectiveZoom());
        style->setBorderRightWidth(buttonBorderWidth() * style->effectiveZoom());
        style->setBorderBottomWidth(buttonBorderWidth() * style->effectiveZoom());
        style->setBorderTopWidth(buttonBorderWidth() * style->effectiveZoom());
        
        style->setBackgroundColor(buttonBackgroundColor());
        
        // Height is locked to auto.
        style->setHeight(Length(Auto));

        // White-space is locked to pre
        style->setWhiteSpace(PRE);

        // Set the button's vertical size.
        setButtonSize(style);

        // Add in the padding that we'd like to use.
        setButtonPaddingFromControlSize(style, controlSize);

        style->setMinHeight(Length(style->computedFontSize() + 8, Fixed));
    } else {
        // Set a min-height so that we can't get smaller than the mini button.
        style->setMinHeight(Length(15, Fixed));

        // Reset the top and bottom borders.
        style->resetBorderTop();
        style->resetBorderBottom();
    }
}

bool RenderThemeEspial::paintButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* ctx = i.context;
	RenderStyle* style = o->style();
	
	ctx->save();	
	
	if(style && style->hasBorderRadius()){
		Path path;
		
		path.addRoundedRect(style->getRoundedBorderFor(rect));
		
        ctx->clip(path);
        
	    //fill background
	    ctx->fillRect(rect, buttonBackgroundColor());
	    
	    if(o->isBoxModelObject()){
	    	static_cast<RenderBoxModelObject*>(o)->paintBorder(i, rect, style);
	    }
	    else{
		    //draw border
		    ctx->setStrokeColor(buttonBorderColor());
		    ctx->strokePath(path);
	    }
	}
	else{
		ctx->clip(rect);
		
	    //fill background
	    ctx->fillRect(rect, buttonBackgroundColor());
	    
	    //draw border
	    ctx->setStrokeColor(buttonBorderColor());
	    ctx->strokeRect(rect, buttonBorderWidth());
	        
		//std::cout << "RenderTheme::paintButton at(" << rect.x() << "," << rect.y() << ") size(" << rect.width() << "," << rect.height() << ")\n";
	}
	
	ctx->restore();
	
	return false;
}

const int sliderThumbWidth = 16;
const int sliderThumbHeight = 16;

void RenderThemeEspial::adjustSliderThumbStyle(StyleResolver*, RenderStyle* style, Element* element) const
{
    adjustSliderThumbSize(style, element);
}

void RenderThemeEspial::adjustSliderThumbSize(RenderStyle* style, Element*) const
{
    if (style->appearance() == SliderThumbHorizontalPart || style->appearance() == SliderThumbVerticalPart) {
        style->setWidth(Length(sliderThumbWidth, Fixed));
        style->setHeight(Length(sliderThumbHeight, Fixed));
    }
#if ENABLE(VIDEO)
    else if (style->appearance() == MediaSliderThumbPart) {
        style->setWidth(Length(sliderThumbWidth, Fixed));
        style->setHeight(Length(sliderThumbHeight, Fixed));
    }
#endif
}

static const float baseFontSize = 11.0f;
static const float baseArrowHeight = 5.0f;
static const float baseArrowWidth = 7.0f;
static const int arrowPaddingLeft = 5;
static const int arrowPaddingRight = 5;
static const int paddingBeforeSeparator = 4;
static const int baseBorderRadius = 5;
static const int styledPopupPaddingLeft = 8;
static const int styledPopupPaddingTop = 1;
static const int styledPopupPaddingBottom = 2;

static const int* popupButtonPadding(NSControlSize size)
{
    static const int padding[3][4] =
    {
        { 2, 26, 3, 8 },
        { 2, 23, 3, 8 },
        { 2, 22, 3, 10 }
    };
    return padding[size];
}

int RenderThemeEspial::popupInternalPaddingLeft(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[leftPadding];
    
    if (style->appearance() == MenulistButtonPart)
        return styledPopupPaddingLeft;
    return 0;
}

int RenderThemeEspial::popupInternalPaddingRight(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[rightPadding];
    
    if (style->appearance() == MenulistButtonPart) {
        float fontScale = style->computedFontSize() / baseFontSize;
        float arrowWidth = baseArrowWidth * fontScale;
        return static_cast<int>(ceilf(arrowWidth + arrowPaddingLeft + arrowPaddingRight + paddingBeforeSeparator));
    }
    return 0;
}

int RenderThemeEspial::popupInternalPaddingTop(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[topPadding];
    if (style->appearance() == MenulistButtonPart)
        return styledPopupPaddingTop;
    return 0;
}

int RenderThemeEspial::popupInternalPaddingBottom(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[bottomPadding];
    if (style->appearance() == MenulistButtonPart)
        return styledPopupPaddingBottom;
    return 0;
}


void RenderThemeEspial::adjustMenuListStyle(StyleResolver* selector, RenderStyle* style, WebCore::Element* e) const
{
    style->resetBorder();
    style->resetPadding();
    style->setHeight(Length(Auto));
    style->setWhiteSpace(PRE);
}

bool RenderThemeEspial::paintMenuListButton(RenderObject* o, const PaintInfo& paintInfo, const IntRect& r)
{
	//std::cout << "RenderTheme::paintMenuListButton at(" << r.x() << "," << r.y() << ") size(" << r.width() << "," << r.height() << ")\n";

	GraphicsContext* ctx = paintInfo.context;
	
    IntRect bounds = IntRect(r.x() + o->style()->borderLeftWidth(),
                             r.y() + o->style()->borderTopWidth(),
                             r.width() - o->style()->borderLeftWidth() - o->style()->borderRightWidth(),
                             r.height() - o->style()->borderTopWidth() - o->style()->borderBottomWidth());
        
    // Since we actually know the size of the control here, we restrict the font scale to make sure the arrow will fit vertically in the bounds
    float fontScale = min(o->style()->computedFontSize() / baseFontSize, bounds.height() / baseArrowHeight);
    float centerY = bounds.y() + bounds.height() / 2.0f;
    float arrowHeight = baseArrowHeight * fontScale;
    float arrowWidth = baseArrowWidth * fontScale;
    float leftEdge = bounds.x() + bounds.width() - arrowPaddingRight - arrowWidth;

    if (bounds.width() < arrowWidth + arrowPaddingLeft)
        return false;

    ctx->setFillColor(o->style()->visitedDependentColor(CSSPropertyColor));
    ctx->setStrokeColor(NoStroke);

    FloatPoint arrow[3];
    arrow[0] = FloatPoint(leftEdge, centerY - arrowHeight / 2.0f);
    arrow[1] = FloatPoint(leftEdge + arrowWidth, centerY - arrowHeight / 2.0f);
    arrow[2] = FloatPoint(leftEdge + arrowWidth / 2.0f, centerY + arrowHeight / 2.0f);

    // Draw the arrow
    ctx->drawConvexPolygon(3, arrow, true);

    Color leftSeparatorColor(0, 0, 0, 40);
    Color rightSeparatorColor(255, 255, 255, 40);
    
    // FIXME: Should the separator thickness and space be scaled up by fontScale?
    int separatorSpace = 2;
    int leftEdgeOfSeparator = static_cast<int>(leftEdge - arrowPaddingLeft); // FIXME: Round?

    // Draw the separator to the left of the arrows
    ctx->setStrokeThickness(1.0f);
    ctx->setStrokeStyle(SolidStroke);
    ctx->setStrokeColor(leftSeparatorColor);
    ctx->drawLine(IntPoint(leftEdgeOfSeparator, bounds.y()), IntPoint(leftEdgeOfSeparator, bounds.y() + bounds.height()));

    ctx->setStrokeColor(rightSeparatorColor);
    ctx->drawLine(IntPoint(leftEdgeOfSeparator + separatorSpace, bounds.y()), IntPoint(leftEdgeOfSeparator + separatorSpace, bounds.y() + bounds.height()));

    return false;
}

bool RenderThemeEspial::paintMenuList(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{	
	//std::cout << "RenderTheme::paintMenuList at(" << rect.x() << "," << rect.y() << ") size(" << rect.width() << "," << rect.height() << ")\n";

	GraphicsContext* ctx = i.context;
	
	ctx->save();	
	ctx->clip(rect);
			
	ctx->fillRect(rect, o->style()->visitedDependentColor(CSSPropertyBackgroundColor));
	
	//draw border
    ctx->setStrokeStyle(SolidStroke);
	ctx->setStrokeColor(defaultBorderColor());
	ctx->strokeRect(rect, 1);

	if(o && o->isMenuList()){
		paintMenuListButton(o, i, rect);
	}
	
	ctx->restore();
	
	return false;
}

bool RenderThemeEspial::paintSliderTrack(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

    ctx->fillRect(rect, m_sliderTrackBackgroundColor);

	ctx->restore();

	return false;
}

bool RenderThemeEspial::paintSliderThumb(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	if(m_mediaSliderThumbImage)
		drawImage(ctx, m_mediaSliderThumbImage.get(), rect);

	ctx->restore();

	return false;
}

void RenderThemeEspial::adjustTextFieldStyle(StyleResolver* selector, RenderStyle* style, Element* e) const
{
#if ENABLE(BML)
	if(e && e->document()->isBMLDocument())
		return;
#endif
            
    style->setWhiteSpace(PRE);
}

bool RenderThemeEspial::paintTextField(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* const ctx = i.context;
	
	ctx->save();
	
	ctx->clip(rect);
	
	ctx->setStrokeStyle(SolidStroke);
    ctx->setStrokeColor(defaultBorderColor());
    
    ctx->strokeRect(rect, 1);
    
	ctx->restore();
	
	return false;
}

void RenderThemeEspial::adjustTextAreaStyle(StyleResolver* selector, RenderStyle* style, Element* e) const
{
    adjustTextFieldStyle(selector, style, e);
}

bool RenderThemeEspial::paintTextArea(RenderObject* o, const PaintInfo& i, const IntRect& r)
{
    return paintTextField(o, i, r);
}

void RenderThemeEspial::adjustSearchFieldResultsButtonStyle(StyleResolver* selector, RenderStyle* style, Element* e) const
{
    adjustSearchFieldCancelButtonStyle(selector, style, e);
}

bool RenderThemeEspial::paintSearchFieldResultsButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{

	notImplemented();
	return false;
}

void RenderThemeEspial::adjustSearchFieldResultsDecorationStyle(StyleResolver* selector, RenderStyle* style, Element* e) const
{
    style->resetBorder();
    style->resetPadding();

    // FIXME: This should not be hard-coded.
    IntSize size = IntSize(14, 14);
    style->setWidth(Length(size.width(), Fixed));
    style->setHeight(Length(size.height(), Fixed));
}

bool RenderThemeEspial::paintSearchFieldResultsDecoration(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	notImplemented();
	return false;
}

void RenderThemeEspial::adjustSearchFieldCancelButtonStyle(StyleResolver* selector, RenderStyle* style, Element* e) const
{
    style->resetBorder();
    style->resetPadding();

    // FIXME: This should not be hard-coded.
    IntSize size = IntSize(14, 14);
    style->setWidth(Length(size.width(), Fixed));
    style->setHeight(Length(size.height(), Fixed));
}

bool RenderThemeEspial::paintSearchFieldCancelButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	notImplemented();
	return false;
}

void RenderThemeEspial::adjustSearchFieldStyle(StyleResolver* selector, RenderStyle* style, Element* e) const
{
    adjustTextFieldStyle(selector, style, e);
}

bool RenderThemeEspial::paintSearchField(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
    return paintTextField(o, i, rect);
}


void RenderThemeEspial::systemFont(CSSValueID, FontDescription&) const
{
    // If you remove this notImplemented(), replace it with an comment that explains why.
    notImplemented();
}

bool RenderThemeEspial::paintMediaFullscreenButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	ctx->fillRect(rect, m_mediaControlsBackgroundColor);

	HTMLMediaElement* mediaElement = getMediaElementFromRenderObject(o);

	if(mediaElement){
		if(mediaElement->isFullscreen()) {
			if(m_mediaFullScreenOffImage)
				drawImage(ctx, m_mediaFullScreenOffImage.get(), rect);
		}
		else {
			if(m_mediaFullScreenOnImage)
				drawImage(ctx, m_mediaFullScreenOnImage.get(), rect);
		}
	}

	ctx->restore();
#endif
	
	return false;
}

bool RenderThemeEspial::paintMediaPlayButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	ctx->fillRect(rect, m_mediaControlsBackgroundColor);

	HTMLMediaElement* mediaElement = getMediaElementFromRenderObject(o);

	if(mediaElement){
		if(mediaElement->canPlay()){
			if(m_mediaPlayImage)
				drawImage(ctx, m_mediaPlayImage.get(), rect);
		}
		else{
			if(m_mediaPauseImage)
				drawImage(ctx, m_mediaPauseImage.get(), rect);
		}
	}

	ctx->restore();
#endif
	
	return false;
}

bool RenderThemeEspial::paintMediaToggleClosedCaptionsButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	ctx->fillRect(rect, m_mediaControlsBackgroundColor);

	HTMLMediaElement* mediaElement = getMediaElementFromRenderObject(o);

	if(mediaElement){
		if(mediaElement->closedCaptionsVisible()){
			if(m_mediaCCOffImage)
				drawImage(ctx, m_mediaCCOffImage.get(), rect);
		}
		else{
			if(m_mediaCCOnImage)
				drawImage(ctx, m_mediaCCOnImage.get(), rect);
		}
	}

	ctx->restore();
#endif

	return false;
}

bool RenderThemeEspial::paintMediaMuteButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	ctx->fillRect(rect, m_mediaControlsBackgroundColor);

	HTMLMediaElement* mediaElement = getMediaElementFromRenderObject(o);

	if(mediaElement){
		if(mediaElement->muted()){
			if(m_mediaVolumeOffImage)
				drawImage(ctx, m_mediaVolumeOffImage.get(), rect);
		}
		else{
			if(m_mediaVolumeOnImage)
				drawImage(ctx, m_mediaVolumeOnImage.get(), rect);
		}
	}

	ctx->restore();
#endif
	
	return false;
}

bool RenderThemeEspial::paintMediaSeekBackButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* ctx = i.context;

	ctx->save();		
	ctx->clip(rect);

	ctx->fillRect(rect, m_mediaControlsBackgroundColor);

	if(m_mediaSeekBackwardImage)
		drawImage(ctx, m_mediaSeekBackwardImage.get(), rect);

	ctx->restore();
	
	return false;
}

bool RenderThemeEspial::paintMediaSeekForwardButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* ctx = i.context;

	ctx->save();		
	ctx->clip(rect);

	ctx->fillRect(rect, m_mediaControlsBackgroundColor);

	if(m_mediaSeekForwardImage)
		drawImage(ctx, m_mediaSeekForwardImage.get(), rect);

	ctx->restore();
	
	return false;
}

bool RenderThemeEspial::paintMediaSliderTrack(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	ctx->fillRect(rect, m_sliderTrackBackgroundColor);

	ctx->restore();
#endif
	
	return false;
}

bool RenderThemeEspial::paintMediaSliderThumb(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	if(m_mediaSliderThumbImage)
		drawImage(ctx, m_mediaSliderThumbImage.get(), rect);

	ctx->restore();
#endif
	
	return false;
}

bool RenderThemeEspial::paintMediaVolumeSliderContainer(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	return false;
}

bool RenderThemeEspial::paintMediaVolumeSliderTrack(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	ctx->fillRect(rect, m_sliderTrackBackgroundColor);

	ctx->restore();
#endif
	
	return false;
}

bool RenderThemeEspial::paintMediaVolumeSliderThumb(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
#if ENABLE(VIDEO)
	GraphicsContext* ctx = i.context;

	ctx->save();
	ctx->clip(rect);

	if(m_mediaSliderThumbImage)
		drawImage(ctx, m_mediaSliderThumbImage.get(), rect);

	ctx->restore();
#endif
	
	return false;
}

bool RenderThemeEspial::paintMediaRewindButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	return false;
}

bool RenderThemeEspial::paintMediaReturnToRealtimeButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	return false;
}

bool RenderThemeEspial::paintMediaControlsBackground(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* ctx = i.context;

	ctx->save();		
	ctx->clip(rect);
	
    ctx->fillRect(rect, m_mediaControlsBackgroundColor);

	ctx->restore();
	
	return false;
}

bool RenderThemeEspial::paintMediaCurrentTime(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	return false;
}

bool RenderThemeEspial::paintMediaTimeRemaining(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	return false;
}

#if ENABLE(VIDEO)
bool RenderThemeEspial::supportsClosedCaptioning() const
{
	return true;
}

bool RenderThemeEspial::hasOwnDisabledStateHandlingFor(ControlPart part) const
{
	switch(part){
	case MediaMuteButtonPart:
		return true;
	default:
		return false;
	}
}
#endif

#if ENABLE(DATALIST_ELEMENT)
IntSize RenderThemeEspial::sliderTickSize() const
{
    // FIXME: We need to set this to the size of one tick mark.
    return IntSize(0, 0);
}

int RenderThemeEspial::sliderTickOffsetFromTrackCenter() const
{
    // FIXME: We need to set this to the position of the tick marks.
    return 0;
}
#endif

}
