/*
 * Copyright (C) 2000 Lars Knoll (knoll@kde.org)
 *           (C) 2000 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2003, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2006 Graham Dennis (graham.dennis@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef FillLayer_h
#define FillLayer_h

#include "GraphicsTypes.h"
#include "LengthSize.h"
#include "RenderStyleConstants.h"
#include "StyleImage.h"
#include <wtf/PassOwnPtr.h>
#include <wtf/RefPtr.h>

namespace WebCore {

class RenderElement;

struct FillSize {
    FillSize()
        : type(SizeLength)
    {
    }

    FillSize(EFillSizeType type, const LengthSize& size)
        : type(type)
        , size(size)
    {
    }

    EFillSizeType type;
    LengthSize size;
};

inline bool operator==(const FillSize& a, const FillSize& b)
{
    return a.type == b.type && a.size == b.size;
}

inline bool operator!=(const FillSize& a, const FillSize& b)
{
    return !(a == b);
}

class FillLayer {
    WTF_MAKE_FAST_ALLOCATED;
public:
    explicit FillLayer(EFillLayerType);
    ~FillLayer();

    ALWAYS_INLINE StyleImage* image() const { return m_image.get(); }
    const Length& xPosition() const { return m_xPosition; }
    const Length& yPosition() const { return m_yPosition; }
    ALWAYS_INLINE Edge backgroundXOrigin() const { return static_cast<Edge>(m_backgroundXOrigin); }
    ALWAYS_INLINE Edge backgroundYOrigin() const { return static_cast<Edge>(m_backgroundYOrigin); }
    ALWAYS_INLINE EFillAttachment attachment() const { return static_cast<EFillAttachment>(m_attachment); }
    ALWAYS_INLINE EFillBox clip() const { return static_cast<EFillBox>(m_clip); }
    ALWAYS_INLINE EFillBox origin() const { return static_cast<EFillBox>(m_origin); }
    ALWAYS_INLINE EFillRepeat repeatX() const { return static_cast<EFillRepeat>(m_repeatX); }
    ALWAYS_INLINE EFillRepeat repeatY() const { return static_cast<EFillRepeat>(m_repeatY); }
    ALWAYS_INLINE CompositeOperator composite() const { return static_cast<CompositeOperator>(m_composite); }
    ALWAYS_INLINE BlendMode blendMode() const { return static_cast<BlendMode>(m_blendMode); }
    ALWAYS_INLINE const LengthSize& sizeLength() const { return m_sizeLength; }
    ALWAYS_INLINE EFillSizeType sizeType() const { return static_cast<EFillSizeType>(m_sizeType); }
    ALWAYS_INLINE FillSize size() const { return FillSize(static_cast<EFillSizeType>(m_sizeType), m_sizeLength); }
    ALWAYS_INLINE EMaskSourceType maskSourceType() const { return static_cast<EMaskSourceType>(m_maskSourceType); }

    ALWAYS_INLINE const FillLayer* next() const { return m_next.get(); }
    ALWAYS_INLINE FillLayer* next() { return m_next.get(); }

    ALWAYS_INLINE bool isImageSet() const { return m_imageSet; }
    ALWAYS_INLINE bool isXPositionSet() const { return m_xPosSet; }
    ALWAYS_INLINE bool isYPositionSet() const { return m_yPosSet; }
    ALWAYS_INLINE bool isBackgroundOriginSet() const { return m_backgroundOriginSet; }
    ALWAYS_INLINE bool isAttachmentSet() const { return m_attachmentSet; }
    ALWAYS_INLINE bool isClipSet() const { return m_clipSet; }
    ALWAYS_INLINE bool isOriginSet() const { return m_originSet; }
    ALWAYS_INLINE bool isRepeatXSet() const { return m_repeatXSet; }
    ALWAYS_INLINE bool isRepeatYSet() const { return m_repeatYSet; }
    ALWAYS_INLINE bool isCompositeSet() const { return m_compositeSet; }
    ALWAYS_INLINE bool isBlendModeSet() const { return m_blendModeSet; }
    ALWAYS_INLINE bool isSizeSet() const { return m_sizeType != SizeNone; }
    ALWAYS_INLINE bool isMaskSourceTypeSet() const { return m_maskSourceTypeSet; }
    
    ALWAYS_INLINE void setImage(PassRefPtr<StyleImage> image) { m_image = image; m_imageSet = true; }
    ALWAYS_INLINE void setXPosition(const Length& l) { m_xPosition = l; m_xPosSet = true; }
    ALWAYS_INLINE void setYPosition(const Length& l) { m_yPosition = l; m_yPosSet = true; }
    ALWAYS_INLINE void setBackgroundXOrigin(Edge o) { m_backgroundXOrigin = static_cast<unsigned>(o); m_backgroundOriginSet = true; }
    ALWAYS_INLINE void setBackgroundYOrigin(Edge o) { m_backgroundYOrigin = static_cast<unsigned>(o); m_backgroundOriginSet = true; }
    ALWAYS_INLINE void setAttachment(EFillAttachment attachment) { m_attachment = attachment; m_attachmentSet = true; }
	ALWAYS_INLINE void setClip(EFillBox b) { m_clip = b; m_clipSet = true; }
    ALWAYS_INLINE void setOrigin(EFillBox b) { m_origin = b; m_originSet = true; }
    ALWAYS_INLINE void setRepeatX(EFillRepeat r) { m_repeatX = r; m_repeatXSet = true; }
    ALWAYS_INLINE void setRepeatY(EFillRepeat r) { m_repeatY = r; m_repeatYSet = true; }
    ALWAYS_INLINE void setComposite(CompositeOperator c) { m_composite = c; m_compositeSet = true; }
    ALWAYS_INLINE void setBlendMode(BlendMode b) { m_blendMode = b; m_blendModeSet = true; }
	ALWAYS_INLINE void setSizeType(EFillSizeType b) { m_sizeType = b; }
    ALWAYS_INLINE void setSizeLength(const LengthSize& l) { m_sizeLength = l; }
    ALWAYS_INLINE void setSize(const FillSize& f) { m_sizeType = f.type; m_sizeLength = f.size; }
    ALWAYS_INLINE void setMaskSourceType(EMaskSourceType m) { m_maskSourceType = m; m_maskSourceTypeSet = true; }
    
    ALWAYS_INLINE void clearImage() { m_image.clear(); m_imageSet = false; }
    ALWAYS_INLINE void clearXPosition() { m_xPosSet = false; m_backgroundOriginSet = false; }
    ALWAYS_INLINE void clearYPosition() { m_yPosSet = false; m_backgroundOriginSet = false; }

    ALWAYS_INLINE void clearAttachment() { m_attachmentSet = false; }
    ALWAYS_INLINE void clearClip() { m_clipSet = false; }
    ALWAYS_INLINE void clearOrigin() { m_originSet = false; }
    ALWAYS_INLINE void clearRepeatX() { m_repeatXSet = false; }
    ALWAYS_INLINE void clearRepeatY() { m_repeatYSet = false; }
    ALWAYS_INLINE void clearComposite() { m_compositeSet = false; }
    ALWAYS_INLINE void clearBlendMode() { m_blendModeSet = false; }
    ALWAYS_INLINE void clearSize() { m_sizeType = SizeNone; }
    ALWAYS_INLINE void clearMaskSourceType() { m_maskSourceTypeSet = false; }

    void setNext(PassOwnPtr<FillLayer> n) { m_next = n; }

    FillLayer& operator=(const FillLayer&);
    FillLayer(const FillLayer&);

    bool operator==(const FillLayer&) const;
    ALWAYS_INLINE bool operator!=(const FillLayer& other) const { return !(*this == other); }

    bool containsImage(StyleImage&) const;
    bool imagesAreLoaded() const;
    bool hasImage() const;
    bool hasMaskImage() const { return false; }
    bool hasFixedImage() const;
    bool hasOpaqueImage(const RenderElement&) const;
    bool hasRepeatXY() const;
    bool clipOccludesNextLayers(bool firstLayer) const;

    ALWAYS_INLINE EFillLayerType type() const { return static_cast<EFillLayerType>(m_type); }

    void fillUnsetProperties();
    void cullEmptyLayers();

    static bool imagesIdentical(const FillLayer*, const FillLayer*);

    static ALWAYS_INLINE EFillAttachment initialFillAttachment(EFillLayerType) { return ScrollBackgroundAttachment; }
    static ALWAYS_INLINE EFillBox initialFillClip(EFillLayerType) { return BorderFillBox; }
    static ALWAYS_INLINE EFillBox initialFillOrigin(EFillLayerType type) { return type == BackgroundFillLayer ? PaddingFillBox : BorderFillBox; }
    static ALWAYS_INLINE EFillRepeat initialFillRepeatX(EFillLayerType) { return RepeatFill; }
    static ALWAYS_INLINE EFillRepeat initialFillRepeatY(EFillLayerType) { return RepeatFill; }
    static ALWAYS_INLINE CompositeOperator initialFillComposite(EFillLayerType) { return CompositeSourceOver; }
    static ALWAYS_INLINE BlendMode initialFillBlendMode(EFillLayerType) { return BlendModeNormal; }
    static ALWAYS_INLINE EFillSizeType initialFillSizeType(EFillLayerType) { return SizeNone; }
    static ALWAYS_INLINE LengthSize initialFillSizeLength(EFillLayerType) { return LengthSize(); }
    static ALWAYS_INLINE FillSize initialFillSize(EFillLayerType type) { return FillSize(initialFillSizeType(type), initialFillSizeLength(type)); }
    static ALWAYS_INLINE Length initialFillXPosition(EFillLayerType) { return Length(0.0f, Percent); }
    static ALWAYS_INLINE Length initialFillYPosition(EFillLayerType) { return Length(0.0f, Percent); }
    static ALWAYS_INLINE StyleImage* initialFillImage(EFillLayerType) { return 0; }
    static EMaskSourceType initialMaskSourceType(EFillLayerType) { return MaskAlpha; }

private:
    friend class RenderStyle;

    void computeClipMax() const;

    FillLayer() { }

    OwnPtr<FillLayer> m_next;

    RefPtr<StyleImage> m_image;

    Length m_xPosition;
    Length m_yPosition;

    LengthSize m_sizeLength;

    unsigned m_attachment : 2; // EFillAttachment
    unsigned m_clip : 2; // EFillBox
    unsigned m_origin : 2; // EFillBox
    unsigned m_repeatX : 3; // EFillRepeat
    unsigned m_repeatY : 3; // EFillRepeat
    unsigned m_composite : 4; // CompositeOperator
    unsigned m_sizeType : 2; // EFillSizeType
    unsigned m_blendMode : 5; // BlendMode
    unsigned m_maskSourceType : 1; // EMaskSourceType
    
    unsigned m_imageSet : 1;
    unsigned m_attachmentSet : 1;
    unsigned m_clipSet : 1;
    unsigned m_originSet : 1;
    unsigned m_repeatXSet : 1;
    unsigned m_repeatYSet : 1;
    unsigned m_xPosSet : 1;
    unsigned m_yPosSet : 1;
    unsigned m_backgroundOriginSet : 1;
    unsigned m_backgroundXOrigin : 2; // Edge
    unsigned m_backgroundYOrigin : 2; // Edge
    unsigned m_compositeSet : 1;
    unsigned m_blendModeSet : 1;
    unsigned m_maskSourceTypeSet : 1;
    
    unsigned m_type : 1; // EFillLayerType

    mutable unsigned m_clipMax : 2; // EFillBox, maximum m_clip value from this to bottom layer
};

} // namespace WebCore

#endif // FillLayer_h
