/** @file GalioConfiguration.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "GalioConfiguration.h"
#include "GalioConfigurationItem.h"
#include "GalioConfigurationComponent.h"

#include "UtilsMOM.h"

extern "C" {
#include "layers.h"
#include "layers_string_common.h"
#include "confman.h"
#include "cm_internal.h"
}

// in Galio this is a configuration item, in case it's needed in
// future keep the functionality but controlled by a static
static const bool config_dom_config_readonly = false;

namespace WebCore {

PassRefPtr<GalioConfiguration> GalioConfiguration::create()
{
    return adoptRef(new GalioConfiguration());
}

GalioConfiguration::GalioConfiguration()
{
}

void GalioConfiguration::persist( int delayInMilliSecs, ExceptionCode& ec )
{
    if (config_dom_config_readonly)
    {
        ec = NO_MODIFICATION_ALLOWED_ERR;
        return;
    }

    confman_persist_delayed(delayInMilliSecs/10);
}

void GalioConfiguration::persist(ExceptionCode& ec )
{
	persist(0, ec);
}

PassRefPtr<GalioConfigurationItem> GalioConfiguration::get( const String & name )
{
    return GalioConfigurationItem::create(name);
}

void GalioConfiguration::set( const String & name, const String & value, ExceptionCode& ec )
{
    if (config_dom_config_readonly)
    {
        ec = NO_MODIFICATION_ALLOWED_ERR;
        return;
    }

    const char           *prefix = NULL;
    const char           *tag = NULL;
    confman_dataelement  **base = NULL;
    confman_datatype_tag type;
    int                  index;
    confman_dataelement  *config;
    
    layers_string lname = UtilsMOM::LayersStringFromString(name);
    BOOL found = confman_lookup_item(layers_string_lock(lname),
                                     &prefix, &tag, &base, &type, &index, &config );
    layers_string_unlock(lname);
    layers_string_dispose(lname);

    if (!found)
    {
        ec = NOT_FOUND_ERR;
        return;
    }

    layers_string lvalue = UtilsMOM::LayersStringFromString(value);
    const char *strvalue = layers_string_lock(lvalue);

    // note this code doesn't check for syntax errors, unlike the
    // original Galio code.
    switch (type)
    {
    case confman_datatype_BOOLEAN:
        boolean_text_restore(strvalue, config);
        break;
    case confman_datatype_SIGNED_DECIMAL:
        signed_decimal_text_restore(strvalue, config);
        break;
    case confman_datatype_UNSIGNED_DECIMAL:
        unsigned_decimal_text_restore(strvalue, config);
        break;
    case confman_datatype_HEX:
        hex_text_restore(strvalue, config);
        break;
    case confman_datatype_CHARACTER:
        character_text_restore(strvalue, config);
        break;
    case confman_datatype_COLOUR:
        colour_text_restore(strvalue, config);
        break;
    case confman_datatype_STRING:
        mm_freeconst(CONFMAN_DATA(config, string));
        CONFMAN_DATA(config, string) = mm_strdup(strvalue);
        break;
    case confman_datatype_LAYERS_STRING:
        layers_string_dispose(CONFMAN_DATA(config, lstring));
        CONFMAN_DATA(config, lstring) = lvalue ? layers_string_copy(lvalue) : layers_string_common_empty;
        break;
    case confman_datatype_STRING_LIST:
        confman_string_list_text_restore(strvalue, &CONFMAN_DATA(config, string_list));
        break;
    case confman_datatype_SIGNED_DECIMAL_LIST:
        confman_signed_list_text_restore(strvalue, &CONFMAN_DATA(config, signed_decimal_list), 10);
        break;
    case confman_datatype_UNSIGNED_DECIMAL_LIST:
    case confman_datatype_HEX_LIST:
        confman_unsigned_list_text_restore(strvalue, &CONFMAN_DATA(config, unsigned_decimal_list), 10);
        break;
    case confman_datatype_COLOUR_LIST:
        confman_colour_list_text_restore(strvalue, &CONFMAN_DATA(config, colour_list));
        confman_colour_list_fixup(CONFMAN_DATA(config, colour_list));
        break;

    case confman_datatype_END:
    case confman_datatype_PREFIX:
    case confman_datatype_COMMENT:
    case confman_datatype_COUNT:
        break;
    }
    
    layers_string_unlock(lvalue);
    layers_string_dispose(lvalue);

#if CONFMAN_USE_ACCESS_FUNCTIONS
    layers_configuration_item_set(prefix, tag, base, type, index );
#endif
}

PassRefPtr<GalioConfigurationComponent> GalioConfiguration::createComponent( ScriptExecutionContext* context, const String & name, ExceptionCode& ec )
{
    if (config_dom_config_readonly)
    {
        ec = NO_MODIFICATION_ALLOWED_ERR;
        return nullptr;
    }

    layers_string lname = UtilsMOM::LayersStringFromString(name);
    confman_dynamic_component comp = confman_dynamic_component_create(lname);
    layers_string_dispose(lname);

    return comp ? GalioConfigurationComponent::create(context, comp) : nullptr;
}

PassRefPtr<GalioConfigurationComponent> GalioConfiguration::getComponent( ScriptExecutionContext* context, const String & name, ExceptionCode& ec )
{
    if (config_dom_config_readonly)
    {
        ec = NO_MODIFICATION_ALLOWED_ERR;
        return nullptr;
    }

    layers_string lname = UtilsMOM::LayersStringFromString(name);
    confman_dynamic_component comp = confman_dynamic_component_get(lname);
    layers_string_dispose(lname);

    return comp ? GalioConfigurationComponent::create(context, comp) : nullptr;
}

} // namespace WebCore

#endif // ENABLE(RMAI)
