/** @file GalioWafer.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "GalioWafer.h"
#include "GalioWaferCollection.h"
#include "EventTarget.h"
#include "DOMWindow.h"
#include "Event.h"
#include "UIEvent.h"
#include "HTMLNames.h"
#include "Frame.h"
#include "HTMLIFrameElement.h"
#include "HTMLBodyElement.h"

#include "RenderObject.h"
#include "RenderStyle.h"

namespace WebCore {

PassRefPtr<GalioWafer> GalioWafer::create(Document* document)
{
    return adoptRef(new GalioWafer(document));
}

PassRefPtr<GalioWafer> GalioWafer::create(GalioWaferCollection* waferCollection, const String & url)
{
    RefPtr<GalioWafer> wafer = GalioWafer::create(waferCollection->document());

    wafer->m_url = url;
    wafer->m_waferCollection = waferCollection;

    return wafer.release();
}

GalioWafer::GalioWafer(Document* document)
    : ContainerNode(document, CreateContainer),
      m_destroyed(0),
      m_deleting(0),
      m_visible(0),
      m_active(0),
      m_waferCollection(NULL),
      m_url()
{
}

GalioWafer::~GalioWafer()
{
    destroyWafer();
}

// Node methods
String GalioWafer::nodeName() const
{
    return String("#wafer");
}

Node::NodeType GalioWafer::nodeType() const
{
    return WAFER_NODE;
}

RefPtr<Node> GalioWafer::cloneNodeInternal(Document&, CloningOperation)
{
   // TODO if needed
    return nullptr;
}

DOMWindow* GalioWafer::window() const
{
    return m_destroyed ? NULL : m_iframe->contentWindow();
}

Document* GalioWafer::contentDocument() const
{
    return m_destroyed ? NULL : m_iframe->contentDocument();
}

bool GalioWafer::visible() const
{
    return !m_destroyed && m_visible;
}

bool GalioWafer::active() const
{
    return !m_destroyed && m_active;
}

bool GalioWafer::isPrimaryReceiver() const
{
    return m_waferCollection->activeWafer() == this;
}

bool GalioWafer::isVisibleTopmost() const
{
    return m_waferCollection->topmostVisible() == this;
}

/*
bool GalioWafer::hasFocus() const
{
    return 0;
}
*/

GalioWaferCollection* GalioWafer::apis() const
{
    return m_waferCollection;
}

/*
bool GalioWafer::scrollable() const
{
    return 0;
}

void GalioWafer::setScrollable(bool scrollable)
{
}
*/

/*
Script* GalioWafer::onLowMemory() const
{
    return NULL;
}
void GalioWafer::setOnLowMemory(Script* onLowMemory)
{
}
*/

PassRefPtr<GalioWafer> GalioWafer::createWafer( const String & url )
{
    return GalioWafer::create(m_waferCollection, url);
}

void GalioWafer::makeVisible(bool visible)
{
    m_visible = visible;

    m_iframe->setInlineStyleProperty(CSSPropertyDisplay, visible ? String("block") : String("none"), true);

    int zIndex = visible ? m_waferCollection->nextZIndex() : 0;
    m_iframe->setInlineStyleProperty(CSSPropertyZIndex, String::number(zIndex), true);
}

void GalioWafer::show()
{
    if (!m_destroyed && !m_deleting && !m_visible)
        m_waferCollection->show(this);
}

void GalioWafer::hide()
{
    if (!m_destroyed && m_visible)
        m_waferCollection->hide(this);
}

void GalioWafer::deactivate()
{
    if (!m_destroyed)
        m_waferCollection->makeInactive(this);
}

void GalioWafer::activate()
{
    if (!m_destroyed)
        m_waferCollection->makeActive(this);
}

void GalioWafer::destroyWafer()
{
    if (!m_destroyed)
    {
        m_deleting = true;

        m_waferCollection->makeInactive(this);
        m_waferCollection->hide(this);

        ExceptionCode ec;
        m_waferCollection->frame()->document()->body()->removeChild(m_iframe, ec);
        m_iframe = NULL;

        m_destroyed = true;
        m_deleting = false;
    }
}

bool GalioWafer::dispatchSystemEvent(PassRefPtr<Event> event, PassRefPtr<Node> target)
{
    return m_waferCollection->dispatchSystemEvent(event, target);
}

void GalioWafer::blur()
{
    deactivate();
}

void GalioWafer::focus()
{
    activate();
}

void GalioWafer::restartApplicationTree()
{
    // TODO
}

PassRefPtr<Event> GalioWafer::createUIEvent(const AtomicString& eventName)
{
    ASSERT(m_iframe);
    return UIEvent::create(eventName, false, false, m_iframe->contentDocument()->defaultView(), 0);
}

void GalioWafer::dispatchWaferEvent(const AtomicString& eventName )
{
    if ( !m_deleting && !m_destroyed )
    {
        RefPtr<Event> event;
        if (eventName == eventNames().loadEvent || eventName == eventNames().unloadEvent ||
            eventName == eventNames().abortEvent || eventName == eventNames().errorEvent)
        {
            event = Event::create(eventName, false, false);
        }
        else
        {
            event = createUIEvent(eventName);
        }

        dispatchEvent(event.release());
    }
}

bool GalioWafer::isRootWafer()
{
    Node* parent = parentNode();
    return parent && parent->nodeType() == DOCUMENT_NODE;
}

void GalioWafer::start()
{
    ExceptionCode ec;

    // wafer node has been inserted into tree, create new iframe and
    // load url into it, making iframe invisible and inactive. this
    // could be the rootwafer or a child wafer, it doesn't make a
    // difference
    Document* containingDocument = m_waferCollection->frame()->document();
    if (!containingDocument->body())
        containingDocument->setBody(HTMLBodyElement::create(HTMLNames::bodyTag, containingDocument), ec);
    ASSERT(ec == 0);
    RefPtr<HTMLIFrameElement> iframe = HTMLIFrameElement::create(HTMLNames::iframeTag, containingDocument);

    iframe->setLocation(m_url);
    iframe->setIdAttribute(m_url);
    m_iframe = iframe.get();

    containingDocument->body()->appendChild(iframe.release(), ec);
    ASSERT(ec == 0);
}

bool GalioWafer::childTypeAllowed(NodeType nodeType) const
{
    return nodeType == WAFER_NODE;
}

PassRefPtr<Node> GalioWafer::appendChild(PassRefPtr<Node> prpNewChild, ExceptionCode& ec)
{
    RefPtr<Node> newChild(prpNewChild);
    if (!Node::appendChild(newChild, ec))
    {
        return nullptr;
    }

    RefPtr<GalioWafer> wafer = static_pointer_cast<GalioWafer>(newChild);
    wafer->start();
    return wafer.release();
}

void GalioWafer::documentLoaded(Document* doc)
{
    dispatchSystemEvent(createUIEvent(eventNames().ApplicationLoadedEvent), doc);
}

void GalioWafer::documentUnloaded(Document* doc)
{
    dispatchSystemEvent(createUIEvent(eventNames().ApplicationUnloadedEvent), doc);
}

static void setInlineStylePropertyPx(HTMLIFrameElement* iframe, CSSPropertyID id, float value)
{
	iframe->setInlineStyleProperty(id, String::number(value), true);
}

void GalioWafer::moveBy(float x, float y)
{
    // TODO: if necessary
}

void GalioWafer::moveTo(float x, float y)
{
    setInlineStylePropertyPx(m_iframe, CSSPropertyLeft, x);
    setInlineStylePropertyPx(m_iframe, CSSPropertyTop, y);
}

void GalioWafer::resizeBy(float x, float y)
{
    setInlineStylePropertyPx(m_iframe, CSSPropertyWidth, m_iframe->width() + x);
    setInlineStylePropertyPx(m_iframe, CSSPropertyHeight, m_iframe->height() + y);
}

void GalioWafer::resizeTo(float x, float y)
{
    setInlineStylePropertyPx(m_iframe, CSSPropertyWidth, x);
    setInlineStylePropertyPx(m_iframe, CSSPropertyHeight, y);
}

} // namespace WebCore

#endif // ENABLE(RMAI)
