/** @file GalioWaferCollection.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "GalioWaferCollection.h"
#include "GalioWafer.h"

#include "SchemeRegistry.h"
#include "SecurityOrigin.h"
#include "DOMWindow.h"
#include "Element.h"
#include "ElementTraversal.h"
#include "HTMLNames.h"
#include "Frame.h"
#include "HTMLIFrameElement.h"
#include "KeyboardEvent.h"
#include "PlatformKeyboardEvent.h"

extern "C" {
#include "layers.h"
#include "layers_flex.h"
#include "useropts.h"
#include "confman.h"
extern void layers_log_printf( const char *fmt, ... );
}

namespace WebCore {

static GalioWaferCollection* gWaferCollection;
    
GalioWaferCollection* GalioWaferCollection::instance()
{
    if (!gWaferCollection)
        gWaferCollection = new GalioWaferCollection();
    return gWaferCollection;
}

GalioWaferCollection::GalioWaferCollection()
    : m_activeList(Vector<GalioWafer*>()),
      m_showList(Vector<GalioWafer*>()),
      m_document(0),
      m_frame(0),
      m_zIndex(0)
{
}

GalioBrowser* GalioWaferCollection::browser() const
{
    if (!m_browser)
        m_browser = GalioBrowser::create(const_cast<GalioWaferCollection*>(this));
    return m_browser.get();
}

GalioConfiguration* GalioWaferCollection::config() const
{
    if (!m_config)
        m_config = GalioConfiguration::create();
    return m_config.get();
}
/*
MOM* GalioWaferCollection::mom() const
{
    if (!m_mom)
        m_mom = MOM::create(m_context);
    return m_mom.get();
}
*/
GalioUtility* GalioWaferCollection::utility() const
{
    if (!m_utility)
        m_utility = GalioUtility::create();
    return m_utility.get();
}

/* add wafer to head of list and return the previous head */
GalioWafer* GalioWaferCollection::link(Vector<GalioWafer*>& list, GalioWafer* wafer)
{
    GalioWafer* previousHead = list.isEmpty() ? NULL : list[0];
    list.insert(0, wafer);
    return previousHead;
}
    
/* remove wafer from list and return true if it was at the head of the list */
bool GalioWaferCollection::delink(Vector<GalioWafer*>& list, GalioWafer* wafer)
{
    size_t location = list.find(wafer);
    if (location != notFound)
        list.remove(location);
    return location == 0;
}
    
GalioWafer* GalioWaferCollection::activeWafer() const
{
    return m_activeList.isEmpty() ? NULL : m_activeList[0];
}

GalioWafer* GalioWaferCollection::topmostVisible() const
{
    return m_showList.isEmpty() ? NULL : m_showList[0];
}
   
void GalioWaferCollection::show(GalioWafer* wafer)
{
    GalioWafer* previousHead = link(m_showList, wafer);
    wafer->makeVisible(true);
    wafer->dispatchWaferEvent(eventNames().ApplicationShownEvent);
    if (previousHead && topmostVisible() != previousHead)
    {
        previousHead->dispatchWaferEvent(eventNames().ApplicationNotTopmostEvent);
    }
    if (topmostVisible() == wafer)
    {
        wafer->dispatchWaferEvent(eventNames().ApplicationTopmostEvent);
    }   
}

void GalioWaferCollection::hide(GalioWafer* wafer)
{
    bool changed_head = delink(m_showList, wafer);
    wafer->makeVisible(false);
    wafer->dispatchWaferEvent(eventNames().ApplicationHiddenEvent);
    if (changed_head)
    {
        wafer->dispatchWaferEvent(eventNames().ApplicationNotTopmostEvent);
        if (GalioWafer* top = topmostVisible())
            top->dispatchWaferEvent(eventNames().ApplicationTopmostEvent);
    }
}

void GalioWaferCollection::makeActive(GalioWafer* wafer)
{
    //layers_log_printf("-- makeActive: %s\n", wafer->url().latin1().data());
    /* Nothing to do - this application is already frontmost active */
    if (activeWafer() == wafer) return;

    bool was_already_active = wafer->active();
    if ( was_already_active )
    {
        /* Already active!  Unlink from the application chain */
        delink(m_activeList, wafer);
    }
    else
    {
        wafer->setActive(true);
    }

    GalioWafer* previousHead = link(m_activeList, wafer);
    if ( !was_already_active )
    {
        wafer->dispatchWaferEvent(eventNames().ApplicationActivatedEvent);
    }
    if ( previousHead && m_activeList[0] != previousHead )
    {
        previousHead->dispatchWaferEvent(eventNames().ApplicationNotPrimaryReceiverEvent);
    }
    if ( m_activeList[0] == wafer )
    {
        wafer->dispatchWaferEvent(eventNames().ApplicationPrimaryReceiverEvent);
    }
}
    
void GalioWaferCollection::makeInactive(GalioWafer* wafer)
{
    //layers_log_printf("-- makeInactive: %s\n", wafer->url().latin1().data());
    if ( !wafer->active() ) return;

    bool changed_head = delink(m_activeList, wafer);
    wafer->setActive(false);
    wafer->dispatchWaferEvent(eventNames().ApplicationDeactivatedEvent);
    if ( changed_head )
    {
        wafer->dispatchWaferEvent(eventNames().ApplicationNotPrimaryReceiverEvent );
        if (GalioWafer* active = activeWafer())
            active->dispatchWaferEvent(eventNames().ApplicationPrimaryReceiverEvent );
    }
}

void GalioWaferCollection::createApplicationTree(Frame* frame)
{
    if (m_document)
        return;

    String rootUrl(useropts_default_app());
    if (rootUrl.isEmpty())
        return;

    // remove security blocks between antclient and package, and file
    SchemeRegistry::registerURLSchemeAsLocal(String("antclient"));
    SchemeRegistry::registerURLSchemeAsLocal(String("package"));

    HTMLIFrameElement* wafersFrameEl = static_cast<HTMLIFrameElement*>(frame->document()->getElementById(String("wafers")));
    if (!wafersFrameEl)
        return;

    m_document = wafersFrameEl->contentDocument();
    m_frame = frame;

    m_document->removeChildren();

    RefPtr<GalioWafer> rootWafer = GalioWafer::create(this, rootUrl);
    GalioWafer* rootWaferPtr = rootWafer.get();
    ExceptionCode ec;
    if (m_document->appendChild(rootWafer.release(), ec))
    {
        gWaferCollection->link(m_activeList, rootWaferPtr);
        rootWaferPtr->setActive(true);

        rootWaferPtr->start();
    }
}

GalioWafer* GalioWaferCollection::findOwnerWafer(const Document* appDoc) const
{
    if (!m_document)
        return NULL;

    Node* baseNode = m_document->firstChild();
    for (Node* node = baseNode; node; node = NodeTraversal::next(node, baseNode))
    {
        GalioWafer* wafer = static_cast<GalioWafer*>(node);
        Document* waferDoc = wafer->contentDocument();
        if (waferDoc == appDoc)
            return wafer;
    }
    return NULL;
}

bool GalioWaferCollection::dispatchSystemEventToActiveList(PassRefPtr<Event> prpEvent, Document* exceptDoc, GalioWafer *startingFromWafer)
{
    RefPtr<Event> event(prpEvent);
    Vector<GalioWafer*> listCopy = gWaferCollection->m_activeList;

    /* iterate over list copy so that newly active apps aren't included */
    Vector<GalioWafer*>::iterator end = listCopy.end();
    bool seenStartWafer = false;
    for (Vector<GalioWafer*>::iterator it = listCopy.begin(); it != end; ++it)
    {
        GalioWafer* wafer = *it;

        /* iterate from the old activeWafer, ignore anything newly active */
        if (!seenStartWafer && wafer == startingFromWafer) {
            seenStartWafer = true;
        }
        if (!seenStartWafer) {
            continue;
        }
   
        DOMWindow* window = wafer->window();
        /* check wafer is still active, don't call it if it isn't */
        if (window && wafer->active())
        {
            Document* doc = window->document();
            if (doc && doc != exceptDoc)
            {
                Node* node = doc->focusedElement();
                if (!node)
                    node = doc;
                //layers_log_printf("-- sending system event to wafer %s doc %p nodeIsDoc %d\n", (*it)->url().latin1().data(), doc, node == doc);
                event->clearState();
                if (!node->dispatchEvent(event))
                    return false;
            }
        }
    }

    return true;
}

bool GalioWaferCollection::dispatchSystemEvent(PassRefPtr<Event> prpEvent, PassRefPtr<Node> target)
{
    RefPtr<Event> event(prpEvent);

    /* send to the active wafer first */
    GalioWafer *currentActiveWafer = activeWafer();
    if (currentActiveWafer)
    {
        event->clearState();
        if (!currentActiveWafer->dispatchEvent(event))
            return false;
    }

    /* then to the target if any */
    Document* target_doc = NULL;
    if (target)
    {
        target_doc = target->document();
        event->clearState();
        if (!target->dispatchEvent(event))
            return false;
    }

    return dispatchSystemEventToActiveList(event, target_doc, currentActiveWafer);
}

bool GalioWaferCollection::handleKeyEvent(const PlatformKeyboardEvent& kevt, Frame* frame)
{
    layers_log_printf("\n-- sending key %d --\n", kevt.windowsVirtualKeyCode());

    RefPtr<KeyboardEvent> event = KeyboardEvent::create(kevt, frame->document()->defaultView());

    /* protect the wafer during the call. Can be 0 if frame is the
     * mainFrame. */
    RefPtr<GalioWafer> frameWafer(frame->document()->ownerWafer());

    /* send to the active wafer first */
    GalioWafer *currentActiveWafer = activeWafer();
    if (currentActiveWafer)
    {
        //layers_log_printf("-- sending key to active wafer %s\n", wafer->url().latin1().data());
        event->clearState();
        if (!currentActiveWafer->dispatchEvent(event))
            return true;
    }

    /* check that original frame has not been deleted */
    Document *frameDoc = frameWafer ? frameWafer->contentDocument() : NULL;
    if (frameDoc)
    {
        /* then do the standard key handling */
        //layers_log_printf("-- sending key to target frame %p in %s\n", frame, wafer?wafer->url().latin1().data():"");
        bool defaultPrevented = false;
        frame->eventHandler()->keyEvent(kevt, &defaultPrevented);
        if (defaultPrevented)
            return defaultPrevented;
    }
    else
    {
        //layers_log_printf("-- not calling key handler, frame deleted %s\n", frameWafer?frameWafer->url().latin1().data():"");
    }

    /* then to the documents on the active list */
    //layers_log_printf("-- sending key to active list\n");
    dispatchSystemEventToActiveList(event, frameDoc, currentActiveWafer);

    return false;
}

} // namespace WebCore

#endif // ENABLE(RMAI)
