/** @file MOM.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOM.h"
#include "MOMDVR.h"
#include "MOMConfiguration.h"
#include "MOMPlayer.h"
#include "MOMRenderer.h"
#include "MOMChannelManager.h"
#include "MOMVODManager.h"
#include "MOMContentItem.h"
#include "MOMContentList.h"
#include "MOMPackageManager.h"
#include "MOMDeviceCollection.h"
#include "MOMLocalSystem.h"
#include "MOMDRMAgent.h"
#include "MOMParentalRatingSchemeCollection.h"
#include "MOMNetworkInterface.h"

#include "wtf/text/CString.h"
#include "UtilsMOM.h"
#include "MOMException.h"
#include "EventNames.h"

extern "C" {
#include "layers.h"
#include "layers_string.h"
#include "layers_util.h"
#include "cdf/mom_core.h"
#include "cdf/mom_localsystem.h"
#include "cdf/mom_resource_conflict_list.h"
#include "cdf/mom_content_catalogue_collection.h"
}

namespace WebCore {

static MOM* gMOM;
    
PassRefPtr<MOM> MOM::create(ScriptExecutionContext* context)
{
    if (!gMOM)
        gMOM = new MOM(context);
    return gMOM;
}

MOM::MOM(ScriptExecutionContext* context)
    : ActiveDOMObject(context),
      m_asyncEventQueue(GenericEventQueue::create(this))
{
}

MOM::~MOM()
{
}

// event handling

PassRefPtr<EventTarget> MOM::findTarget(const mom_event_params* params, MOMEvent::TargetType type)
{
    switch(type)
    {
    case MOMEvent::TARGET_MOM:
        return this;

    case MOMEvent::TARGET_PACKAGE_MANAGER:
        return packageManager();

    case MOMEvent::TARGET_NETWORK_INTERFACE:
        return localSystem()->networkInterfaces()->item(params->args[mom_event_NETWORKTEST_INDEX]);

    case MOMEvent::TARGET_DVR:
        return dvr();

    case MOMEvent::TARGET_DRM_AGENT:
        return drmAgent();
    }
    return nullptr;
}

void MOM::enqueueEventOnAll(MOMEvent::TargetType type, mom_event_t which, mom_event_params* params)
{
    if (!gMOM) return;
    RefPtr<EventTarget> target = gMOM->findTarget(params, type);
    RefPtr<MOMEvent> event = MOMEvent::create(which, params, target->scriptExecutionContext());
    event->setTarget(target);
    gMOM->enqueueEvent(event);
}
    
void MOM::dispatchEventOnAll(MOMEvent::TargetType type, mom_event_t which, mom_event_params* params)
{
    if (!gMOM) return;
    RefPtr<EventTarget> target = gMOM->findTarget(params, type);
    RefPtr<MOMEvent> event = MOMEvent::create(which, params, target->scriptExecutionContext());
    target->dispatchEvent(event);
}

void MOM::enqueueEvent(PassRefPtr<MOMEvent> event)
{
    m_asyncEventQueue->enqueueEvent(event);
}

void MOM::enqueueEventOnMatch(PassRefPtr<MOMEvent> prpEvent)
{
    if (!gMOM) return;
    gMOM->enqueueEvent(prpEvent);
}
    
void MOM::stop()
{
}

// EventTarget
ScriptExecutionContext* MOM::scriptExecutionContext() const
{
    return ActiveDOMObject::scriptExecutionContext();
}

EventTargetData* MOM::eventTargetData()
{
    return &m_eventTargetData;
}

EventTargetData* MOM::ensureEventTargetData()
{
    return &m_eventTargetData;
}

EventTargetInterface MOM::eventTargetInterface() const
{
    return MOMEventTargetInterfaceType;
}

// MOM    
MOMDeviceCollection* MOM::devices() const
{
    if (!m_devices)
        m_devices = MOMDeviceCollection::create(scriptExecutionContext(), MOMDeviceCollection::STRATEGY_ALL, FALSE);
    return m_devices.get();
}

MOMDVR* MOM::dvr() const
{
    if (!m_dvr)
        m_dvr = MOMDVR::create(scriptExecutionContext());
    return m_dvr.get();
}

MOMDRMAgent* MOM::drmAgent() const
{
    if (!m_drmAgent)
        m_drmAgent = MOMDRMAgent::create(scriptExecutionContext());
    return m_drmAgent.get();
}

MOMConfiguration* MOM::configuration() const
{
    if (!m_configuration)
        m_configuration = MOMConfiguration::create();
    return m_configuration.get();
}

MOMLocalSystem* MOM::localSystem() const
{
    if (!m_localSystem)
        m_localSystem = MOMLocalSystem::create(scriptExecutionContext());
    return m_localSystem.get();
}

MOMParentalRatingSchemeCollection* MOM::parentalRatingSchemes() const
{
    if (!m_parentalRatingsSchemes)
        m_parentalRatingsSchemes = MOMParentalRatingSchemeCollection::create();
    return m_parentalRatingsSchemes.get();
}

MOMPackageManager* MOM::packageManager() const
{
    if (!m_packageManager)
        m_packageManager = MOMPackageManager::create(scriptExecutionContext());
    return m_packageManager.get();
}

MOMChannelManager* MOM::channels() const
{
    return MOM::channelManager();
}

MOMChannelManager* MOM::channelManager() const
{
    if (!m_channelManager)
        m_channelManager = MOMChannelManager::create(scriptExecutionContext());
    return m_channelManager.get();
}

MOMContentItem* MOM::catalogues() const
{
    if (!m_catalogues)
        m_catalogues = MOMContentList::create(scriptExecutionContext(), mom_content_catalogue_collection_lookup());
    return m_catalogues.get();
}

MOMVODManager* MOM::vodManager() const
{
    if (!m_vodManager)
        m_vodManager = MOMVODManager::create(scriptExecutionContext());
    return m_vodManager.get();
}


bool MOM::hasFeature( const String & feature, const String & )
{
    if (feature == "guide" || feature == "core" || feature == "pvr" || feature == "vod")
        return true;
    return false;
}

/* map from the id value passed to getDevicesByType to the strategy */
static struct
{
    int id;                     /* values in the XML interface definition */
    MOMDeviceCollection::CollectionStrategy strategy;
} s_devicecollection_strategies[] =
{
    { MOM::SOURCE,   MOMDeviceCollection::STRATEGY_SOURCE },
    { MOM::RECORDER, MOMDeviceCollection::STRATEGY_RECORDER },
    { MOM::RF,       MOMDeviceCollection::STRATEGY_RF },
    { MOM::IPTV,     MOMDeviceCollection::STRATEGY_IPTV },
    { MOM::UPNP,     MOMDeviceCollection::STRATEGY_UPNP },
    { MOM::VOD,      MOMDeviceCollection::STRATEGY_VOD },
    /*{ MOM::REMOVABLE, MOMDeviceCollection::STRATEGY_REMOVABLE },*/
};

static MOMDeviceCollection::CollectionStrategy toStrategy( int id )
{
    unsigned int i;
    for ( i = 0; i < sizeof(s_devicecollection_strategies)/sizeof(s_devicecollection_strategies[0]); i++ )
    {
        if ( s_devicecollection_strategies[i].id == id )
            return s_devicecollection_strategies[i].strategy;
    }

    return MOMDeviceCollection::STRATEGY_UNKNOWN;
}

    // TODO: should raise exception if type is out of range
PassRefPtr<MOMDeviceCollection> MOM::getDevicesByType( int type, bool localOnly )
{
    return MOMDeviceCollection::create(scriptExecutionContext(), toStrategy(type), localOnly);
}

PassRefPtr<MOMDeviceCollection> MOM::getDevicesByURI( const String & uri, bool localOnly )
{
    layers_string luri = UtilsMOM::LayersStringFromString(uri);
    RefPtr<MOMDeviceCollection> collection =  MOMDeviceCollection::create(scriptExecutionContext(), luri, localOnly);
    layers_string_dispose(luri);
    return collection;
}

PassRefPtr<MOMRenderer> MOM::getRenderer(ExceptionCode& ec)
{
    mom_renderer mrend;
    mom_error err = mom_renderer_create(&mrend);
    if (err != ERR_NONE)
    {
        MOMException::setFromMOMError(err, ec);
        return nullptr;
    }
    RefPtr<MOMRenderer> renderer = MOMRenderer::create(scriptExecutionContext(), mrend, false);
    mom_renderer_dispose(mrend);
    return renderer.release();
}

PassRefPtr<MOMPlayer> MOM::createPlayer( const String & source, bool trickplay, ExceptionCode& ec )
{
    mom_player mplay;
    mom_error err = mom_player_create(trickplay, &mplay);
    if (err != ERR_NONE)
    {
        MOMException::setFromMOMError(err, ec);
        return nullptr;
    }
    RefPtr<MOMPlayer> player = MOMPlayer::create(scriptExecutionContext(), mplay);
    mom_player_dispose(mplay);

    player.get()->connect(source, 0, 0, ec);

    return player.release();
}

void MOM::setText( const String & name, const String & value )
{
    layers_string lname = UtilsMOM::LayersStringFromString(name);
    layers_string lvalue = UtilsMOM::LayersStringFromString(value);
    mom_set_text(lname, lvalue);
    layers_string_dispose(lname);
    layers_string_dispose(lvalue);
}

String MOM::getText( const String & name )
{
    layers_string lname = UtilsMOM::LayersStringFromString(name);
    layers_string ltitle = mom_get_text(lname);
    layers_string_dispose(lname);
    return UtilsMOM::StringFromLayersString(ltitle);
}

String MOM::formatTime( const String & format, int time, bool formatAsUTC )
{
    layers_string lformat = UtilsMOM::LayersStringFromString(format);
    char buffer[256];
    layers_strftime( buffer, sizeof(buffer), layers_string_lock(lformat), time, formatAsUTC );
    layers_string_unlock(lformat);
    layers_string_dispose(lformat);
    return String::fromUTF8(buffer);
}

String MOM::formatTime( const String & format, int time)
{
	return formatTime(format, time, false);
}

bool MOM::downloadImage()
{
    return mom_resource_conflict_list_perform_download() == ERR_NONE;
}

bool MOM::initialised()
{
    // if this isn't set then it's a fair guess that MOM hasn't
    // been initialised.
    return mom_localsystem_software_version() != NULL;
}

void MOM::init()
{
    MOMEvent::init();
    MOMPackageManager::init();
}

} // namespace WebCore

#endif // ENABLE(RMAI)
