/** @file MOM.h
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#ifndef MOM_h
#define MOM_h

#if ENABLE(RMAI)

#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/text/AtomicString.h>

#include "ActiveDOMObject.h"
#include "ExceptionCode.h"
#include "EventTarget.h"
#include "GenericEventQueue.h"
#include "MOMEvent.h"

namespace WebCore {

class MOMDVR;
class MOMConfiguration;
class MOMPlayer;
class MOMRenderer;
class MOMChannelManager;
class MOMVODManager;
class MOMContentItem;
class MOMContentList;
class MOMPackageManager;
class MOMDeviceCollection;
class MOMLocalSystem;
class MOMDRMAgent;
class MOMParentalRatingSchemeCollection;
class Event;
class ScriptExecutionContext;

class MOM : public RefCounted<MOM>, public ActiveDOMObject, public EventTarget {
public:
    enum {
        NONE = 0,
        SOURCE = 1,
        RECORDER = 2,
        RF = 10,
        IPTV = 11,
        UPNP = 12,
        PORTABLE = 13,
        VOD = 14,
        IDLE = 0,
        CONNECTING = 101,
        CONNECTED = 102,
        PLAYING = 103,
        STOPPED = 104,
        FAILED = 105,
        PROBLEM = 106,
        RECORDING = 107,
        SCHEDULED = 108,
        DELETED = 109,
        DUE_SOON = 110,
        UPDATED = 111,
        PARTIAL = 112,
        REMINDER = 113,
        REMINDER_DUE = 114,
        ADDED = 115,
        CANCELLED = 116,
        PLAYING_INTERSTITIAL = 117,
        PLAYING_WITHOUT_PRESENTATION = 118,
        BUFFERING = 119,
        OK = 0,
        INVALID = 1,
        BUSY = 2,
        NOTPRESENT = 3,
        DISCFULL = 4,
        FAILURE = 5,
        NOTAUTHORISED = 6,
        RATING = 7,
        DIALOG = 8,
        START_OF_STREAM = 9,
        END_OF_STREAM = 10,
        TSB_WRAP = 11,
        PIN_WRONG = 12,
        PIN_LOCKED = 13,
        EXPIRED = 14,
        SERVER_ERROR = 15,
        CLIENT_ERROR = 16,
        PROGRESSED = 17,
        VIEW = 0,
        PREVIEW = 1,
        PURCHASE = 2,
        SORT_UID = 0,
        SORT_START_TIME = 1,
        SORT_TITLE = 2,
        SORT_DVR_AND_TITLE = 3,
        SORT_TITLE_AND_DESC_DATE = 4,
        SORT_CHANNEL = 5,
        SORT_CHANNEL_AND_DESC_DATE = 6,
    };

    static PassRefPtr<MOM> create(ScriptExecutionContext*);
    virtual ~MOM();

    MOMDeviceCollection* devices() const;
    MOMDVR* dvr() const;
    MOMDRMAgent* drmAgent() const;
    MOMConfiguration* configuration() const;
    MOMLocalSystem* localSystem() const;
    MOMParentalRatingSchemeCollection* parentalRatingSchemes() const;
    MOMPackageManager* packageManager() const;
    MOMChannelManager* channels() const;
    MOMChannelManager* channelManager() const;
    MOMContentItem* catalogues() const;
    MOMVODManager* vodManager() const;

    bool hasFeature( const String & feature, const String & version );
    PassRefPtr<MOMDeviceCollection> getDevicesByType( int type, bool localOnly );
    PassRefPtr<MOMDeviceCollection> getDevicesByURI( const String & uri, bool localOnly );
    PassRefPtr<MOMRenderer> getRenderer(ExceptionCode& ec);
    PassRefPtr<MOMPlayer> createPlayer( const String & source, bool trickplay, ExceptionCode& ec );
    void setText( const String & name, const String & value );
    String getText( const String & name );
    String formatTime( const String & format, int time );
    String formatTime( const String & format, int time, bool formatAsUTC );
    bool downloadImage();
    static bool initialised();

    // ActiveDOMObject interface
    bool hasPendingActivity() const /*OVERRIDE*/ { return const_cast<MOM*>(this)->hasEventListeners(); }
    void stop() /*OVERRIDE*/;
    virtual const char* activeDOMObjectName() const OVERRIDE { return "MOM"; }

    // EventTarget interface
    MOM* toMOM() { return this; }
    ScriptExecutionContext* scriptExecutionContext() const /*OVERRIDE*/;
    EventTargetData* eventTargetData() /*OVERRIDE*/;
    EventTargetData* ensureEventTargetData() /*OVERRIDE*/;
    void refEventTarget() /*OVERRIDE*/ { ref(); }
    void derefEventTarget() /*OVERRIDE*/ { deref(); }

    virtual EventTargetInterface eventTargetInterface() const;

    using RefCounted<MOM>::ref;
    using RefCounted<MOM>::deref;

    void enqueueEvent(PassRefPtr<MOMEvent> event);
    static void enqueueEventOnAll(MOMEvent::TargetType type, mom_event_t which, mom_event_params* params);
    static void dispatchEventOnAll(MOMEvent::TargetType type, mom_event_t which, mom_event_params* params);

    static void enqueueEventOnMatch(PassRefPtr<MOMEvent> event);

    static void init();

protected:
    MOM(ScriptExecutionContext*);

private:
    PassRefPtr<EventTarget> findTarget(const mom_event_params* event, MOMEvent::TargetType type);

    mutable RefPtr<MOMDeviceCollection> m_devices;
    mutable RefPtr<MOMDVR> m_dvr;
    mutable RefPtr<MOMDRMAgent> m_drmAgent;
    mutable RefPtr<MOMConfiguration> m_configuration;
    mutable RefPtr<MOMLocalSystem> m_localSystem;
    mutable RefPtr<MOMParentalRatingSchemeCollection> m_parentalRatingsSchemes;
    mutable RefPtr<MOMPackageManager> m_packageManager;
    mutable RefPtr<MOMChannelManager> m_channelManager;
    mutable RefPtr<MOMContentList> m_catalogues;
    mutable RefPtr<MOMVODManager> m_vodManager;

    EventTargetData m_eventTargetData;
    OwnPtr<GenericEventQueue> m_asyncEventQueue;
};

} // namespace WebCore

#endif // ENABLE(RMAI)

#endif // MOM_h
