/** @file MOMChannel.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMChannel.h"
#include "MOMPlayer.h"
#include "MOMStringCollection.h"

#include "MOMException.h"
#include "UtilsMOM.h"

extern "C" {
#include "layers.h"
#include "cdf/mom_channel_extension.h"
#include "cdf/mom_channel_list_local.h"
}

namespace WebCore {

PassRefPtr<MOMChannel> MOMChannel::createAdopt(ScriptExecutionContext *context, mom_channel chan)
{
    return static_pointer_cast<MOMChannel>(MOMContentItem::createAdopt(context, (mom_content_item)chan));
}
    
PassRefPtr<MOMChannel> MOMChannel::create(ScriptExecutionContext *context, mom_channel chan)
{
    return static_pointer_cast<MOMChannel>(MOMContentItem::create(context, (mom_content_item)chan));
}
    
PassRefPtr<MOMChannel> MOMChannel::create(ScriptExecutionContext *context, const String& chan, const String& name, ExceptionCode& ec)
{
    mom_channel channel;

    layers_string lchan = UtilsMOM::LayersStringFromString(chan);
    mom_error err = mom_channel_create_dummy(lchan, 0, MDA_CHANNEL_TYPE_URI, &channel);
    layers_string_dispose(lchan);

    if (err != ERR_NONE)
    {
        MOMException::setFromMOMError(err, ec);
        return nullptr;
    }

    layers_string lname = UtilsMOM::LayersStringFromString(name);
    if (lname)
    {
        mom_content_item_set_name((mom_content_item)channel, lname);
        layers_string_dispose(lname);
    }

    mom_content_list_add_item((mom_content_list)mom_channel_list_local(), (mom_content_item)channel, NULL);
    mom_channel_set_local(channel, TRUE);

    return MOMChannel::createAdopt(context, channel);
}

MOMChannel::MOMChannel(ScriptExecutionContext *context, mom_content_item item)
    : MOMContentItem(context, item)
{
}

int MOMChannel::channelNumber() const
{
    return mom_channel_channel_number(itemAsChannel());
}

String MOMChannel::source() const
{
    return UtilsMOM::StringFromLayersString(mom_content_item_get_uri(m_item));
}

int MOMChannel::channelType() const
{
    return mom_channel_get_channel_type(itemAsChannel());
}

String MOMChannel::serviceId() const
{
    return UtilsMOM::StringFromLayersString(mom_channel_get_service_id(itemAsChannel()));
}

bool MOMChannel::authorised() const
{
    return mom_channel_is_authorised(itemAsChannel());
}

MOMStringCollection* MOMChannel::genre() const
{
    if (!m_genre)
        m_genre = MOMStringCollection::create(mom_channel_genres(itemAsChannel()));
    return m_genre.get();
}

bool MOMChannel::manualBlock() const
{
    return mom_content_item_get_manual_block(m_item);
}

bool MOMChannel::favourite() const
{
    return mom_content_item_get_favourite(m_item);
}

bool MOMChannel::hidden() const
{
    return mom_content_item_get_hidden(m_item);
}

bool MOMChannel::canRecord() const
{
    return (mom_channel_get_recordable(itemAsChannel()) & MDA_CHANNEL_CAN_RECORD) != 0;
}

bool MOMChannel::canTrickPlay() const
{
    return (mom_channel_get_recordable(itemAsChannel()) & MDA_CHANNEL_CAN_USE_TSB) != 0;
}

bool MOMChannel::displayInGuide() const
{
    return mom_channel_get_display_in_guide(itemAsChannel());
}

bool MOMChannel::isVirtual() const
{
    return mom_channel_is_virtual(itemAsChannel());
}

MOMChannel* MOMChannel::equivalentHDChannel() const
{
    if (!m_equivalentHDChannel)
        m_equivalentHDChannel = MOMChannel::create(scriptExecutionContext(), mom_channel_get_hd_equivalent(itemAsChannel()));
    return m_equivalentHDChannel.get();
}

String MOMChannel::packageID() const
{
    return UtilsMOM::StringFromLayersString(mom_channel_get_package_id(itemAsChannel()));
}

bool MOMChannel::subscribed() const
{
    return mom_channel_extension_get_subscribed(itemAsChannel());
}

void MOMChannel::connect( MOMPlayer* player, int mode, ExceptionCode& ec )
{
    if ( mode >= mom_stream_connect_mode__MAX )
    {
        MOMException::setFromMOMError(mom_err_INVALID, ec);
        return;
    }

    mom_error err = mom_player_connect_content_item(player->player(), itemAsItem(), static_cast<mom_stream_connect_mode>(mode), 0);
    MOMException::setFromMOMError(err, ec);
}

void MOMChannel::connect( MOMPlayer* player, ExceptionCode& ec )
{
	connect(player, 0, ec);
}

String MOMChannel::getLogo( int width, int height )
{
    return UtilsMOM::StringFromLayersString(mom_channel_get_logo(itemAsChannel(), width, height));
}

} // namespace WebCore

#endif // ENABLE(RMAI)
