/** @file MOMChannelList.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMChannelList.h"
#include "MOMChannel.h"
#include "MOMContentItem.h"
#include "MOMException.h"
#include "UtilsMOM.h"

extern "C" {
#include "layers.h"
#include "cdf/mom_channel_list.h"
#include "cdf/mom_channel_list_all.h"
#include "cdf/mom_content_item.h"
#include "cdf/mom_content_list.h"
}

namespace WebCore {

PassRefPtr<MOMChannelList> MOMChannelList::createAdopt(ScriptExecutionContext *context, mom_channel_list list)
{
    return static_pointer_cast<MOMChannelList>(MOMContentItem::createAdopt(context, (mom_content_item)list));
}
    
PassRefPtr<MOMChannelList> MOMChannelList::create(ScriptExecutionContext *context, mom_channel_list list)
{
    return static_pointer_cast<MOMChannelList>(MOMContentItem::create(context, (mom_content_item)list));
}
    
// constructor exposed to JavaScript
PassRefPtr<MOMChannelList> MOMChannelList::create(ScriptExecutionContext *context, ExceptionCode& ec)
{
    mom_channel_list list;
    mom_error err = mom_channel_list_create(&list);
    if (err != ERR_NONE)
    {
        MOMException::setFromMOMError(err, ec);
        return nullptr;
    }
    return MOMChannelList::createAdopt(context, list);
}
    
MOMChannelList::MOMChannelList(ScriptExecutionContext *context, mom_content_item item)
    : MOMContentList(context, item)
{
}

PassRefPtr<MOMChannel> MOMChannelList::getChannel( int channelNumber, ExceptionCode& ec)
{
    mom_channel chan = mom_channel_list_get_channel(itemAsChannelList(), channelNumber);
    if (chan == NULL)
    {
        ec = MOMException::NOTPRESENT;
        return PassRefPtr<MOMChannel>();
    }
    return MOMChannel::create(scriptExecutionContext(), chan);
}

PassRefPtr<MOMChannel> MOMChannelList::getChannelByName( const String & name )
{
    layers_string lname = UtilsMOM::LayersStringFromString(name);
    // ignores current channel list, as original Galio code
    RefPtr<MOMChannel> channel = MOMChannel::create(scriptExecutionContext(), mom_channel_list_all_get_channel_by_name(lname));
    layers_string_dispose(lname);
    return channel.release();
}

void MOMChannelList::setLength(int length, ExceptionCode& ec)
{
    if (mom_content_item_is_readonly(m_item)) {
        ec = MOMException::INVALID;
        return;
    }
    mom_content_list_truncate(itemAsContentList(), length);
}

void MOMChannelList::append( MOMChannel* newChan, ExceptionCode& ec )
{
    if (mom_content_item_is_readonly(m_item)) {
        ec = MOMException::INVALID;
        return;
    }
    mom_content_list_add_item(itemAsContentList(), newChan->itemAsItem(), NULL);
}

void MOMChannelList::remove( MOMChannel* chan, ExceptionCode& ec )
{
    if (mom_content_item_is_readonly(m_item)) {
        ec = MOMException::INVALID;
        return;
    }
    mom_content_list_remove_item(itemAsContentList(), chan->itemAsItem());
}

void MOMChannelList::insertAt( int index, MOMChannel* newChan, ExceptionCode& ec )
{
    if (mom_content_item_is_readonly(m_item)) {
        ec = MOMException::INVALID;
        return;
    }

    int length = mom_content_list_length( itemAsContentList() );
    /* off the end of the list so extend list by 1 */
    if( index > length )
        index = length;
    mom_content_list_add_item_at(itemAsContentList(), newChan->itemAsItem(), index);
}

void MOMChannelList::replace( MOMChannelList* newList, ExceptionCode& ec )
{
    if (mom_content_item_is_readonly(m_item)) {
        ec = MOMException::INVALID;
        return;
    }

    /* clear 'list' and re-populate from newList */
    mom_content_list_clear( itemAsContentList() ); /* clear the list */

    if ( newList )
    {
        /* re-populate the list */
        mom_error err = mom_content_list_merge( itemAsContentList(), newList->itemAsContentList() );
        if (err != ERR_NONE)
            ec = err;
    }
}

} // namespace WebCore

#endif // ENABLE(RMAI)
