/** @file MOMContentItem.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMCable.h"
#include "MOMChannel.h"
#include "MOMChannelList.h"
#include "MOMChannelScanOptions.h"
#include "MOMContentCatalogue.h"
#include "MOMContentItem.h"
#include "MOMContentList.h"
#include "MOMContentPagedList.h"
#include "MOMContentSearch.h"
#include "MOMEPGProgramme.h"
#include "MOMParentalRating.h"
#include "MOMPackageDescriptor.h"
#include "MOMPackageList.h"
#include "MOMProgramme.h"
#include "MOMProgrammeList.h"
#include "MOMPurchasableAsset.h"
#include "MOMPurchasableContentCatalogue.h"
#include "MOMReminder.h"
#include "MOMResourceConflictItem.h"
#include "MOMResourceConflictList.h"
#include "MOMSatellite.h"
#include "MOMScheduledRecording.h"
#include "MOMTransportStream.h"
#include "MOMVODAsset.h"
#include "MOMVODCatalogue.h"
#include "MOMVODService.h"

#include "UtilsMOM.h"
#include "MOMException.h"
#include "EventNames.h"

extern "C" {
#include "layers.h"
#include "layers_class.h"
#include "cdf/mom_content_item_defs.h"
#include "cdf/mom_content_item.h"
#include "cdf/mom_content_catalogue.h"
#include "cdf/mom_content_paged_list.h"
#include "cdf/mom_core.h"
#include "cdf/mom_content_search.h"
#include "cdf/mom_epgprogramme.h"
#include "cdf/mom_resource_conflict_list.h"
}

namespace WebCore {

PassRefPtr<MOMContentItem> MOMContentItem::createAdopt(ScriptExecutionContext* context, mom_content_item item)
{
    RefPtr<MOMContentItem> witem = MOMContentItem::create(context, item);
    mom_content_item_dispose(item);
    return witem.release();
}

PassRefPtr<MOMContentItem> MOMContentItem::create(ScriptExecutionContext* context, mom_content_item item)
{
    if (item == NULL) return nullptr;

    // if we've already create a MOMContentItem for this
    // mom_content_item then return that instead of creating a new
    // one. Saves memory and existing JS expects object equality
    RefPtr<MOMContentItem> mci = static_cast<MOMContentItem*>(mom_content_item_dom_obj(item));
    if (mci)
        return mci.release();
    
    // keep this list in sync with the one in JSMOMContentItemCustom.cpp
    switch (mom_content_item_get_item_type(item))
    {
    case mom_content_item_type_CONTENT_LIST:
        mci = adoptRef(new MOMContentList(context, item));
        break;
    case mom_content_item_type_VOD_ASSET:
        mci = adoptRef(new MOMVODAsset(context, item));
        break;
    case mom_content_item_type_VOD_SERVICE:
        mci = adoptRef(new MOMVODService(context, item));
        break;
    case mom_content_item_type_CHANNEL:
        mci = adoptRef(new MOMChannel(context, item));
        break;
    case mom_content_item_type_PROGRAMME:
        if (layers_class_is_instance( FLEX_USE(item), mom_epgprogramme )) {
            mci = adoptRef(new MOMEPGProgramme(context, item));
        }
        else {
            mci = adoptRef(new MOMProgramme(context, item));
        }
        break;
    case mom_content_item_type_CHANNEL_LIST:
        mci = adoptRef(new MOMChannelList(context, item));
        break;
    case mom_content_item_type_PROGRAMME_LIST:
        mci = adoptRef(new MOMProgrammeList(context, item));
        break;
    case mom_content_item_type_RECORDING:
        mci = adoptRef(new MOMScheduledRecording(context, item));
        break;
    case mom_content_item_type_REMINDER:
        mci = adoptRef(new MOMReminder(context, item));
        break;
    case mom_content_item_type_PACKAGE:
        mci = adoptRef(new MOMPackageDescriptor(context, item));
        break;
    case mom_content_item_type_PACKAGE_LIST:
        mci = adoptRef(new MOMPackageList(context, item));
        break;
    case mom_content_item_type_RESOURCE_CONFLICT_ITEM:
        mci = adoptRef(new MOMResourceConflictItem(context, item));
        break;
    case mom_content_item_type_DATA_DOWNLOAD:
        mci = adoptRef(new MOMContentItem(context, item));
        //mci = adoptRef(new MOMDataDownload(context, item));
        // TODO: this needs checking
        break;
    case mom_content_item_type_VOD_CATALOGUE:
        mci = adoptRef(new MOMVODCatalogue(context, item));
        break;
    case mom_content_item_type_PURCHASABLE_ASSET:
        mci = adoptRef(new MOMPurchasableAsset(context, item));
        break;
    case mom_content_item_type_PURCHASABLE_CONTENT_CATALOGUE:
        mci = adoptRef(new MOMPurchasableContentCatalogue(context, item));
        break;
    case mom_content_item_type_SATELLITE:
        mci = adoptRef(new MOMSatellite(context, item));
        break;
    case mom_content_item_type_TRANSPORTSTREAM:
        mci = adoptRef(new MOMTransportStream(context, item));
        break;
    case mom_content_item_type_CABLE:
        //TODO: this shouldn't get here with any luck as
        //mom_content_item is used internally but not exposed
        //mci = adoptRef(new MOMCable(context, item));
        break;
    case mom_content_item_type_CHANNELSCANOPTIONS:
        mci = adoptRef(new MOMChannelScanOptions(context, item));
        break;
    default:
        /* mom_content_item subclasses that don't have a specific
         * content type enumeration */
        if (layers_class_is_instance(FLEX_USE(item), mom_content_catalogue)) {
            mci = adoptRef(new MOMContentCatalogue(context, item));
        }
        else if (layers_class_is_instance(FLEX_USE(item), mom_content_paged_list)) {
            mci = adoptRef(new MOMContentPagedList(context, item));
        }
        else if (layers_class_is_instance(FLEX_USE(item), mom_content_search)) {
            mci = adoptRef(new MOMContentSearch(context, item));
        }
        else if (layers_class_is_instance(FLEX_USE(item), mom_resource_conflict_list)) {
            mci = adoptRef(new MOMResourceConflictList(context, item));
        }
        else {
            mci = adoptRef(new MOMContentItem(context, item));
        }
        break;
    }

    return mci.release();
}

MOMContentItem::MOMContentItem(ScriptExecutionContext* context, mom_content_item item)
    : ActiveDOMObject(context),
      m_item(mom_content_item_keep(item))
{
    mom_content_item_set_dom_obj(item, this);
}

MOMContentItem::~MOMContentItem()
{
    mom_content_item_set_dom_obj(m_item, NULL);
    mom_content_item_dispose(m_item);
}

EventTargetInterface MOMContentItem::eventTargetInterface() const
{
    return MOMContentItemEventTargetInterfaceType;
}

String MOMContentItem::name() const
{
    return UtilsMOM::StringFromLayersString(mom_content_item_get_name(m_item));
}

void MOMContentItem::setName(const String & name)
{
    layers_string lname = UtilsMOM::LayersStringFromString(name);
    mom_content_item_set_name(m_item, lname);
    layers_string_dispose(lname);
}

String MOMContentItem::longName() const
{
    return UtilsMOM::StringFromLayersString(mom_content_item_get_long_name(m_item));
}

void MOMContentItem::setLongName(const String & longName)
{
    layers_string lname = UtilsMOM::LayersStringFromString(longName);
    mom_content_item_set_long_name(m_item, lname);
    layers_string_dispose(lname);
}

String MOMContentItem::description() const
{
    return UtilsMOM::StringFromLayersString(mom_content_item_get_description(m_item));
}

void MOMContentItem::setDescription(const String & description)
{
    layers_string ldescr = UtilsMOM::LayersStringFromString(description);
    mom_content_item_set_description(m_item, ldescr);
    layers_string_dispose(ldescr);
}

String MOMContentItem::uri() const
{
    return UtilsMOM::StringFromLayersString(mom_content_item_get_uri(m_item));
}

String MOMContentItem::uid() const
{
    return UtilsMOM::StringFromLayersString(mom_content_item_get_uid(m_item));
}

String MOMContentItem::thumbnailUri() const
{
    return UtilsMOM::StringFromLayersString(mom_content_item_get_thumbnail_uri(m_item));
}

int MOMContentItem::type() const
{
    return mom_content_item_get_item_type(m_item);
}

MOMParentalRating* MOMContentItem::parentalRating() const
{
    if (!m_parentalRating)
    {
        if ( mom_content_item_get_parental_rating_scheme(m_item) == NULL )
            return NULL;
        m_parentalRating = MOMParentalRating::create(m_item);
    }
    return m_parentalRating.get();
}

bool MOMContentItem::locked() const
{
    return mom_content_item_is_locked(m_item);
}

bool MOMContentItem::blocked() const
{
    return mom_content_item_is_blocked(m_item);
}

int MOMContentItem::requiredBandwidth() const
{
    return mom_content_item_get_required_bandwidth(m_item);
}


String MOMContentItem::getProperty( const String & property )
{
    layers_string lprop = UtilsMOM::LayersStringFromString(property);
    layers_string lvalue = mom_content_item_get_property(m_item, lprop);
    String value = UtilsMOM::StringFromLayersString(lvalue);
    layers_string_dispose(lprop);
    layers_string_dispose(lvalue);
    return value;
}

int MOMContentItem::setProperty( const String & property, const String & value, ExceptionCode& ec )
{
    layers_string lprop = UtilsMOM::LayersStringFromString(property);
    layers_string lvalue = UtilsMOM::LayersStringFromString(value);
    mom_error err = mom_content_item_set_property(m_item, lprop, lvalue);
    layers_string_dispose(lprop);
    layers_string_dispose(lvalue);
    MOMException::setFromMOMError(err, ec);
    return err;
}

int MOMContentItem::setProperty( const String & property, ExceptionCode& ec )
{
	return setProperty(property, String(), ec);
}

} // namespace WebCore

#endif // ENABLE(RMAI)
