/** @file MOMContentSearch.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMContentSearch.h"
#include "MOMContentItem.h"
#include "MOMContentList.h"
#include "MOMStringCollection.h"

#include "UtilsMOM.h"

extern "C" {
#include "layers.h"
#include "cdf/mom_content_search.h"
}

namespace WebCore {

MOMContentSearch::MOMContentSearch(ScriptExecutionContext *context, mom_content_item item)
    : MOMContentItem(context, item)
{
}

MOMContentItem* MOMContentSearch::catalogue() const
{
    if (!m_catalogue)
        m_catalogue = MOMContentItem::create(scriptExecutionContext(), (mom_content_item)mom_content_search_get_catalogue(itemAsContentSearch()));
    return m_catalogue.get();
}

MOMContentItem* MOMContentSearch::results() const
{
    if (!m_results)
        m_results = MOMContentItem::create(scriptExecutionContext(), (mom_content_item)mom_content_search_get_results(itemAsContentSearch()));
    return m_results.get();
}

String MOMContentSearch::sort() const
{
    layers_string sort = mom_content_search_get_sort(itemAsContentSearch());
    return sort ? UtilsMOM::StringFromLayersString(sort) : String();
}

void MOMContentSearch::setSort(const String & sort)
{
    layers_string lsort = UtilsMOM::LayersStringFromString(sort);
    mom_content_search_set_sort(itemAsContentSearch(), lsort );
}

int MOMContentSearch::searchType() const
{
    return mom_content_search_get_search_type(itemAsContentSearch());
}

void MOMContentSearch::find( const String & searchKey, const String & searchValue )
{
    layers_string lkey = UtilsMOM::LayersStringFromString(searchKey);
    layers_string lvalue = UtilsMOM::LayersStringFromString(searchValue);
    mom_content_search_find(itemAsContentSearch(), lkey, lvalue);
    layers_string_dispose(lkey);
    layers_string_dispose(lvalue);
}

void MOMContentSearch::abort()
{
    mom_content_search_abort(itemAsContentSearch());
}

void MOMContentSearch::reset()
{
    mom_content_search_reset(itemAsContentSearch());
}

void MOMContentSearch::addSearchTerm( const String & key, const String & value )
{
    layers_string lkey = UtilsMOM::LayersStringFromString(key);
    layers_string lvalue = UtilsMOM::LayersStringFromString(value);
    mom_content_search_add_search_term(itemAsContentSearch(), lkey, lvalue);
    layers_string_dispose(lkey);
    layers_string_dispose(lvalue);
}

void MOMContentSearch::clearSearchTerm( const String & key, const String & value )
{
    layers_string lkey = UtilsMOM::LayersStringFromString(key);
    layers_string lvalue = UtilsMOM::LayersStringFromString(value);
    if (lkey == NULL)
        mom_content_search_clear_all_search_terms(itemAsContentSearch());
    else
        mom_content_search_clear_search_term(itemAsContentSearch(), lkey, lvalue);
    layers_string_dispose(lkey);
    layers_string_dispose(lvalue);
}

void MOMContentSearch::clearSearchTerm()
{
	clearSearchTerm(String(), String());
}

void MOMContentSearch::clearSearchTerm( const String & key)
{
	clearSearchTerm(key, String());
}

PassRefPtr<MOMStringCollection> MOMContentSearch::getSearchTerm( const String & key )
{
    layers_string lkey = UtilsMOM::LayersStringFromString(key);
    mom_urilist values = mom_content_search_get_search_terms(itemAsContentSearch(), lkey);
    layers_string_dispose(lkey);
    return values ? MOMStringCollection::createAdopt(values) : nullptr;
}

int MOMContentSearch::setParentalControlEnable( const String & PIN, bool enable )
{
    layers_string lpin = UtilsMOM::LayersStringFromString(PIN);
    int result = mom_content_search_set_parental_control_enable(itemAsContentSearch(), lpin, enable);
    layers_string_dispose(lpin);
    return result;
}

bool MOMContentSearch::getParentalControlEnable()
{
    return mom_content_search_get_parental_control_enable(itemAsContentSearch());
}

void MOMContentSearch::addGenreConstraint( const String & genre )
{
    layers_string lgenre = UtilsMOM::LayersStringFromString(genre);
    mom_content_search_add_genre_constraint(itemAsContentSearch(), lgenre);
    layers_string_dispose(lgenre);
}

void MOMContentSearch::clearGenreConstraint( const String & genre )
{
    layers_string lgenre = UtilsMOM::LayersStringFromString(genre);
    mom_content_search_clear_genre_constraint(itemAsContentSearch(), lgenre);
    layers_string_dispose(lgenre);
}

void MOMContentSearch::clearGenreConstraint()
{
	return clearGenreConstraint(String());
}

PassRefPtr<MOMStringCollection> MOMContentSearch::getGenreConstraint()
{
    mom_urilist values = mom_content_search_get_genre_constraint(itemAsContentSearch());
    return values ? MOMStringCollection::createAdopt(values) : nullptr;
}

void MOMContentSearch::addChannelConstraint( MOMContentItem* channels )
{
    mom_content_search_add_channel_constraint(itemAsContentSearch(), channels->itemAsItem());
}

void MOMContentSearch::clearChannelConstraint( MOMContentItem* channels )
{
    mom_content_search_clear_channel_constraint(itemAsContentSearch(), channels ? channels->itemAsItem() : NULL);
}

PassRefPtr<MOMContentItem> MOMContentSearch::getChannelConstraint()
{
    mom_content_list channels = mom_content_search_get_channel_constraint(itemAsContentSearch());
    return channels ? MOMContentItem::create(scriptExecutionContext(), (mom_content_item)channels) : nullptr;
}

void MOMContentSearch::setStartTimeConstraint( int startTime )
{
    mom_content_search_set_start_time_constraint(itemAsContentSearch(), startTime);
}

void MOMContentSearch::setEndTimeConstraint( int endTime )
{
    mom_content_search_set_end_time_constraint(itemAsContentSearch(), endTime);
}

int MOMContentSearch::getStartTimeConstraint()
{
    return mom_content_search_get_start_time_constraint(itemAsContentSearch());
}

int MOMContentSearch::getEndTimeConstraint()
{
    return mom_content_search_get_end_time_constraint(itemAsContentSearch());
}

void MOMContentSearch::addConstraint( const String & key, const String & value )
{
    layers_string lkey = UtilsMOM::LayersStringFromString(key);
    layers_string lvalue = UtilsMOM::LayersStringFromString(value);
    mom_content_search_add_constraint(itemAsContentSearch(), lkey, lvalue);
    layers_string_dispose(lkey);
    layers_string_dispose(lvalue);
}

void MOMContentSearch::clearConstraint( const String & key, const String & value )
{
    layers_string lkey = UtilsMOM::LayersStringFromString(key);
    layers_string lvalue = UtilsMOM::LayersStringFromString(value);
    if (lkey == NULL)
        mom_content_search_clear_all_constraints(itemAsContentSearch());
    else
        mom_content_search_clear_constraint(itemAsContentSearch(), lkey, lvalue);
    layers_string_dispose(lkey);
    layers_string_dispose(lvalue);
}

void MOMContentSearch::clearConstraint( const String & key)
{
	clearConstraint(key, String());
}

void MOMContentSearch::clearConstraint()
{
	clearConstraint(String(), String());
}

PassRefPtr<MOMStringCollection> MOMContentSearch::getConstraint( const String & key )
{
    layers_string lkey = UtilsMOM::LayersStringFromString(key);
    mom_urilist values = mom_content_search_get_constraints(itemAsContentSearch(), lkey);
    layers_string_dispose(lkey);
    return values ? MOMStringCollection::createAdopt(values) : nullptr;
}

} // namespace WebCore

#endif // ENABLE(RMAI)
