/** @file MOMDevice.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMDevice.h"
#include "MOMStream.h"

#include "MOMException.h"
#include "MOMSource.h"
#include "MOMRecorder.h"
#include "UtilsMOM.h"
#include "EventNames.h"

extern "C" {
#include "layers.h"
#include "cdf/mom_player.h"
}

namespace WebCore {

PassRefPtr<MOMDevice> MOMDevice::create(ScriptExecutionContext* context, mom_driver driver)
{
    RefPtr<MOMDevice> dev;

    switch ( mom_driver_type( driver ) )
    {
    case MDA_SOURCE:
        dev = MOMSource::create(context, driver);
        break;
    case MDA_RECORDER:
        dev = MOMRecorder::create(context, driver);
        break;
    default:
        break;
    }
    return dev.release();
}

MOMDevice::MOMDevice(ScriptExecutionContext* context, mom_driver driver)
    : ActiveDOMObject(context),
      m_driver(mom_driver_keep(driver))
{
}

MOMDevice::~MOMDevice()
{
    mom_driver_dispose(m_driver);
}

EventTargetInterface MOMDevice::eventTargetInterface() const
{
    return MOMDeviceEventTargetInterfaceType;
}

String MOMDevice::stringAttr(MDAAttr attr) const
{
    MDADevice dev;
    layers_string lstr = NULL;
    if ( mom_mda_driver_device_open( m_driver, &dev, 0, NULL ) == MDA_OK )
    {
        lstr = mom_mda_driver_device_get_string( m_driver, dev, attr );
        mom_mda_driver_device_close( m_driver, dev );
    }

    String str = UtilsMOM::StringFromLayersString(lstr);
    layers_string_dispose(lstr);
    return str;
}

int MOMDevice::intAttr(MDAAttr attr) const
{
    MDADevice dev;
    int val = 0;
    if ( mom_mda_driver_device_open( m_driver, &dev, 0, NULL ) == MDA_OK )
    {
        val = mom_mda_driver_device_get( m_driver, dev, attr );
        mom_mda_driver_device_close( m_driver, dev );
    }
    return val;
}
    
String MOMDevice::uri() const
{
    return stringAttr(MDA_ATTR_SRC);
}

String MOMDevice::id() const
{
    return stringAttr(MDA_ATTR_BOX_ID);
}

String MOMDevice::name() const
{
    return stringAttr(MDA_ATTR_NAME);
}

bool MOMDevice::enabled() const
{
    return intAttr(MDA_ATTR_DEVICE_ENABLE) == 1;
}

void MOMDevice::setEnabled(bool enabled)
{
    MDADevice dev;
    if ( mom_mda_driver_device_open( m_driver, &dev, 0, NULL ) == MDA_OK )
    {
        mom_mda_driver_device_set( m_driver, dev, MDA_ATTR_DEVICE_ENABLE, (enabled ? 1 : 0) );
        mom_mda_driver_device_close( m_driver, dev );
    }
}

bool MOMDevice::isLocal()
{
    return mom_driver_is_local(m_driver);
}

PassRefPtr<MOMStream> MOMDevice::open(const String & uri, ExceptionCode& ec)
{
    layers_string luri = UtilsMOM::LayersStringFromString(uri);

    mom_stream mstrh;
    mom_error err = mom_player_create_stream( m_driver, luri, &mstrh );
    layers_string_dispose(luri);

    if (err != ERR_NONE)
    {
        MOMException::setFromMOMError(err, ec);
        return nullptr;
    }

    RefPtr<MOMStream> stream = MOMStream::create(scriptExecutionContext(), mstrh);
    mom_stream_dispose( mstrh );
    return stream.release();
}

void MOMDevice::close( MOMStream* stream )
{
    mom_stream_dispose(stream->stream());
}

} // namespace WebCore

#endif // ENABLE(RMAI)
