/** @file MOMEvent.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMEvent.h"
#include "MOMPlayEvent.h"
#include "MOMRecordEvent.h"
#include "MOMGuideSearchEvent.h"
#include "MOMUpdateEvent.h"
#include "MOMAccessEvent.h"
#include "MOMContentEvent.h"
#include "MOMContentCatalogueEvent.h"
#include "MOMDVREvent.h"
#include "MOMChannelScanEvent.h"
#include "MOMEASEvent.h"
#include "MOMResourceConflictEvent.h"
#include "MOMMessageEvent.h"
#include "MOMRequestEvent.h"
#include "MOMDeviceEvent.h"
#include "MOMPackageEvent.h"
#include "MOMPurchasableEvent.h"
#include "MOMPurchaseEvent.h"
#include "MOMDiscEvent.h"
#include "MOMNetworkTestEvent.h"
#include "MOMComponentEvent.h"
#include "MOMDRMMessageResultEvent.h"

#include "MOM.h"
#include "MOMGuideSearch.h"
#include "MOMRecorder.h"
#include "MOMRenderer.h"

#include "UtilsMOM.h"

extern "C" {
#include "cdf/mom_parentalcontrols.h"
#include "cdf/mom_guide_parental_controls.h"
#include "cdf/mom_mda_channel_list.h"
#include "cdf/mom_channel_filter.h"
}

namespace WebCore {

PassRefPtr<MOMEvent> MOMEvent::create(mom_event_t which, const mom_event_params* params, ScriptExecutionContext* context)
{
    RefPtr<MOMEvent> event;
    switch (which)
    {
    case mom_event_PLAY:
        event = adoptRef(new MOMPlayEvent(params, context));
        break;
    case mom_event_RECORD:
        event = adoptRef(new MOMRecordEvent(params, context));
        break;
    case mom_event_GUIDESEARCH:
        event = adoptRef(new MOMGuideSearchEvent(params, context));
        break;
    case mom_event_UPDATE:
        event = adoptRef(new MOMUpdateEvent(params, context));
        break;
    case mom_event_ACCESS:
        event = adoptRef(new MOMAccessEvent(params, context));
        break;
    case mom_event_CONTENTACTION:
        event = adoptRef(new MOMContentEvent(params, context));
        break;
    case mom_event_CONTENTCATALOGUE:
        event = adoptRef(new MOMContentCatalogueEvent(params, context));
        break;
    case mom_event_DVR:
        event = adoptRef(new MOMDVREvent(params, context));
        break;
    case mom_event_CHANNELSCAN:
        event = adoptRef(new MOMChannelScanEvent(params, context));
        break;
    case mom_event_EAS:
        event = adoptRef(new MOMEASEvent(params, context));
        break;
    case mom_event_RESOURCECONFLICT:
        event = adoptRef(new MOMResourceConflictEvent(params, context));
        break;
    case mom_event_MESSAGE:
        event = adoptRef(new MOMMessageEvent(params, context));
        break;
    case mom_event_REQUEST:
        event = adoptRef(new MOMRequestEvent(params, context));
        break;
    case mom_event_DEVICE:
        event = adoptRef(new MOMDeviceEvent(params, context));
        break;
    case mom_event_PACKAGE:
        event = adoptRef(new MOMPackageEvent(params, context));
        break;
    case mom_event_PURCHASABLE:
        event = adoptRef(new MOMPurchasableEvent(params, context));
        break;
    case mom_event_PURCHASE:
        event = adoptRef(new MOMPurchaseEvent(params, context));
        break;
    case mom_event_DISC:
        event = adoptRef(new MOMDiscEvent(params, context));
        break;
    case mom_event_NETWORKTEST_RESULT:
        event = adoptRef(new MOMNetworkTestEvent(params, context));
        break;
    case mom_event_COMPONENT:
        event = adoptRef(new MOMComponentEvent(params, context));
        break;
    case mom_event_DRMMESSAGERESULT:
        event = adoptRef(new MOMDRMMessageResultEvent(params, context));
        break;

    // OIPF only events
    case mom_event_STREAM:
    case mom_event_CHANNELCHANGESUCCEEDED:
    case mom_event_CHANNELCHANGEERROR:
    case mom_event_FULLSCREENCHANGE:
    case mom_event_PLAYSTATECHANGED:
    case mom_event_PROGRAMMESCHANGED:
    case mom_event_PARENTALRATINGCHANGE:
    case mom_event_PARENTALRATINGERROR:
    case mom_event_OIPFPLAYSPEEDCHANGED:
    case mom_event_OIPFPLAYPOSITIONCHANGED:
    case mom_event_OIPFRECORDINGEVENT:
    case mom_event_MAX:
        break;
    }

    if (event) event->m_momEventType = which;

    return event.release();
}

MOMEvent::MOMEvent(const AtomicString& type, const mom_event_params* params)
    : Event(type, false, false)
{
    if (params)
    {
        for (int i = 0; i < params->arg_count; i++)
            setIntParam(i, params->args[i]);
    }
}

MOMEvent::~MOMEvent()
{
}

MOMEvent::TargetType MOMEvent::getTargetTypeFromMOMEventT(mom_event_t which)
{
    switch (which)
    {
    case mom_event_UPDATE:
    case mom_event_ACCESS:
    case mom_event_EAS:
    case mom_event_RESOURCECONFLICT:
    case mom_event_CHANNELSCAN:
    case mom_event_MESSAGE:
    case mom_event_REQUEST:
    case mom_event_DEVICE:
        return TARGET_MOM;

    case mom_event_DVR:
        return TARGET_DVR;

    case mom_event_PACKAGE:
        return TARGET_PACKAGE_MANAGER;

    case mom_event_NETWORKTEST_RESULT:
        return TARGET_NETWORK_INTERFACE;

    case mom_event_DRMMESSAGERESULT:
        return TARGET_DRM_AGENT;

        // these items will always have a target passed and so should
        // never get here
    case mom_event_PLAY:
    case mom_event_RECORD:
    case mom_event_GUIDESEARCH:
    case mom_event_CONTENTACTION:
    case mom_event_CONTENTCATALOGUE:
    case mom_event_PURCHASABLE:
    case mom_event_PURCHASE:
    case mom_event_COMPONENT:
    case mom_event_DISC:
        break;

        // OIPF events - unused
    case mom_event_CHANNELCHANGESUCCEEDED:
    case mom_event_CHANNELCHANGEERROR:
    case mom_event_STREAM:
    case mom_event_FULLSCREENCHANGE:
    case mom_event_PLAYSTATECHANGED:
    case mom_event_PROGRAMMESCHANGED:
    case mom_event_PARENTALRATINGCHANGE:
    case mom_event_PARENTALRATINGERROR:
    case mom_event_OIPFPLAYSPEEDCHANGED:
    case mom_event_OIPFPLAYPOSITIONCHANGED:
    case mom_event_OIPFRECORDINGEVENT:
    case mom_event_MAX:
        break;
    }

    return TARGET_MOM;
}
    
/* This function is registered with mom_core_register_dom_dispatcher().
 * It is called when events are raised with mom_event_dispatch() or
 * mom_event_dispatch_queue().
 * We have to deduce the actual EventTarget object based on the event
 * type.
 * The target passed may be NULL, or maybe should be ignored anyway,
 * in which case the event will be dispatched to all MOM instances.
 */
mom_error MOMEvent::dispatchMOMEventToUntypedTarget(mom_event_t which, void* target, mom_event_params* params, unsigned int queue)
{
    switch (which)
    {
        // For events raised on mom_content_item subclasses we are
        // passed a MOMContentItem* wrapper (stored in the
        // mom_content_item.dom_obj property)
        // Note in some cases (DVR) target could be NULL, if it is then
        // broadcast event
    case mom_event_CONTENTACTION:
    case mom_event_CONTENTCATALOGUE:
    case mom_event_PURCHASABLE:
    case mom_event_PURCHASE:
    case mom_event_DVR:
        dispatchMOMEvent(which, static_cast<MOMContentItem*>(target), params, queue);
        break;

        // For component events we are passed the renderer on which to
        // raise the event.
    case mom_event_COMPONENT:
        dispatchMOMEvent(which, static_cast<MOMRenderer*>(target), params, queue);
        break;

        // these events are broadcast, so the target should be NULL
        // and is ignored.
    case mom_event_UPDATE:
    case mom_event_ACCESS:
    case mom_event_EAS:
    case mom_event_RESOURCECONFLICT:
    case mom_event_CHANNELSCAN:
    case mom_event_MESSAGE:
    case mom_event_REQUEST:
    case mom_event_DEVICE:
    case mom_event_PACKAGE:
    case mom_event_NETWORKTEST_RESULT:
    case mom_event_DRMMESSAGERESULT:
        dispatchMOMEvent(which, NULL, params, queue);
        break;

    default:
        // shouldn't get here
        break;
    }

    return static_cast<mom_error>(ERR_NONE);
}

mom_error MOMEvent::dispatchMOMEvent(mom_event_t which, EventTarget* target, mom_event_params* params, unsigned int queue)
{
    if (target)
    {
        RefPtr<MOMEvent> event = MOMEvent::create(which, params, target->scriptExecutionContext());
        if (queue)
        {
            event->setTarget(target);
            MOM::enqueueEventOnMatch(event);
        }
        else
        {
            target->dispatchEvent(event);
        }
    }
    else
    {
        MOMEvent::TargetType type = getTargetTypeFromMOMEventT(which);
        if (queue)
        {
            MOM::enqueueEventOnAll(type, which, params);
        }
        else
        {
            MOM::dispatchEventOnAll(type, which, params);
        }
    }

    return static_cast<mom_error>(ERR_NONE);
}

static void dispatch_update_event(int arg1, int arg2, mom_content_item item)
{
    mom_event_params event_params;
    int args[mom_event_UPDATE_COUNT];

    args[0] = arg1;
    args[1] = arg2;

    event_params.args = args;
    event_params.arg_count = mom_event_UPDATE_COUNT;
    event_params.object = item;

    MOMEvent::dispatchMOMEvent(mom_event_UPDATE, NULL, &event_params, FALSE);
}

/* handle an update event for a channel list from MDA */
static void mom_guide_handle_update(mom_internal_event_t event, void*)
{
    switch ( event )
    {
    case mom_event_CHANNEL_LIST_INVALIDATED:
        dispatch_update_event(MOMUpdateEvent::CHANNEL_LIST_INVALIDATED, 0, NULL);
        break;

    case mom_event_PROGRAMMES_INVALIDATED:
        dispatch_update_event(MOMUpdateEvent::PROGRAMMES_INVALIDATED, 0, NULL);
        break;

    default:
        break;
    }
}

/** This callback is made when changes to the parental control mean
 * that the visible status of some or all objects may have changed.
 */
static void mom_guide_pc_update(uint32_t changes, mom_content_item object, void*, void*)
{
    uint32_t info = 0;

    ASSERT( changes );
    if ( object )
        ASSERT_FLEX_CLASS( mom_content_item, object );

    if ( changes & mom_parentalcontrols_update_LOCK )
        info |= MOMUpdateEvent::LOCK;
    if ( changes & mom_parentalcontrols_update_BLOCK )
        info |= MOMUpdateEvent::BLOCK;
    if ( changes & mom_parentalcontrols_update_TITLES )
        info |= MOMUpdateEvent::TITLES;

    dispatch_update_event( MOMUpdateEvent::ACCESS_CHANGE, info, object );
}

static void channel_filter_callback( uint32_t changes, void*, void* )
{
    uint32_t info = 0;

    if ( changes & mom_channel_filter_changed_BLOCKED )
        info |= MOMUpdateEvent::BLOCKED_CHANNELS;

    if ( changes & mom_channel_filter_changed_FAVOURITE )
        info |= MOMUpdateEvent::FAVOURITE_CHANNELS;

    if ( changes & mom_channel_filter_changed_HIDDEN )
        info |= MOMUpdateEvent::HIDDEN_CHANNELS;

    if ( changes & mom_channel_filter_changed_UNAUTHORISED )
        info |= MOMUpdateEvent::UNAUTHORISED_CHANNELS;

    dispatch_update_event(MOMUpdateEvent::CHANNEL_FILTER_CHANGE, info, NULL);
}

void MOMEvent::init()
{
    mom_core_register_dom_dispatcher( dispatchMOMEventToUntypedTarget );
    mom_guide_parental_controls_set_callback( dispatchMOMEventToUntypedTarget );

    mom_mda_channel_list_set_update_handler( mom_guide_handle_update, NULL );
    mom_parentalcontrols_register_callback( mom_guide_pc_update, NULL, NULL );
    mom_channel_filter_register_callback( channel_filter_callback, NULL, NULL );
}

} // namespace WebCore

#endif // ENABLE(RMAI)
