/** @file MOMProgramme.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMProgramme.h"
#include "MOMPurchasableAsset.h"
#include "MOMContentItem.h"
#include "MOMReminder.h"
#include "MOMStringCollection.h"
#include "MOMChannel.h"

#include "MOMException.h"
#include "UtilsMOM.h"

extern "C" {
#include "layers.h"
#include "cdf/mom_programme.h"
#include "cdf/mom_programme_extension.h"
}

namespace WebCore {

RefPtr<MOMProgramme> MOMProgramme::create(ScriptExecutionContext *context, MOMChannel* chan, int startTime, int duration, ExceptionCode& ec)
{
    mom_programme prog = mom_programme_create_full(NULL, NULL, NULL, NULL, startTime, duration);
    if (prog == NULL)
    {
        MOMException::setFromMOMError((mom_error)ERR_NO_MEMORY, ec);
        return nullptr;
    }
    mom_programme_set_channel(prog, chan->itemAsItem());
    mom_programme_set_source_id(prog, mom_content_item_get_uri(chan->itemAsItem()));
    return MOMProgramme::createAdopt(context, prog);
}

RefPtr<MOMProgramme> MOMProgramme::createAdopt(ScriptExecutionContext *context, mom_programme prog)
{
    return static_pointer_cast<MOMProgramme>(MOMContentItem::createAdopt(context, (mom_content_item)prog));
}

MOMProgramme::MOMProgramme(ScriptExecutionContext *context, mom_content_item item)
    : MOMContentItem(context, item)
{
}

int MOMProgramme::startTime() const
{
    return mom_programme_get_start_time(itemAsProgramme());
}

int MOMProgramme::duration() const
{
    return mom_programme_get_duration(itemAsProgramme());
}

int MOMProgramme::filler() const
{
    return mom_programme_get_filler(itemAsProgramme());
}

int MOMProgramme::releaseYear() const
{
    return mom_programme_get_release_year(itemAsProgramme());
}

bool MOMProgramme::hasRecording() const
{
    return mom_programme_has_recording(itemAsProgramme());
}

MOMContentItem* MOMProgramme::channel() const
{
    if (!m_channel)
        m_channel = MOMContentItem::create(scriptExecutionContext(), mom_programme_get_channel(itemAsProgramme()));
    return m_channel.get();
}

PassRefPtr<MOMContentItem> MOMProgramme::scheduledRecording() const
{
    mom_content_item rec = mom_programme_get_recording(itemAsProgramme(), -1);
    return rec ? MOMContentItem::create(scriptExecutionContext(), rec) : nullptr;
}

MOMContentItem* MOMProgramme::recordings() const
{
    return const_cast<MOMProgramme*>(this);
}

int MOMProgramme::length() const
{
    return mom_programme_get_recording_count(itemAsProgramme());
}

PassRefPtr<MOMReminder> MOMProgramme::reminder() const
{
    mom_reminder rem = mom_reminder_get(itemAsProgramme());
    return rem ? MOMReminder::create(scriptExecutionContext(), rem) : nullptr;
}

int MOMProgramme::starRating() const
{
    return mom_programme_star_rating(itemAsProgramme());
}

bool MOMProgramme::subtitles() const
{
    return (mom_programme_flags(itemAsProgramme()) & mom_programme_SUBTITLES) != 0;
}

bool MOMProgramme::sapAvailable() const
{
    return (mom_programme_flags(itemAsProgramme()) & mom_programme_SAP) != 0;
}

bool MOMProgramme::isMultilingual() const
{
    return (mom_programme_flags(itemAsProgramme()) & mom_programme_MULTILINGUAL) != 0;
}

String MOMProgramme::episodeTitle() const
{
    return UtilsMOM::StringFromLayersString(mom_programme_episode_title(itemAsProgramme()));
}

MOMStringCollection* MOMProgramme::contentAdvisory() const
{
    if (!m_contentAdvisory)
        m_contentAdvisory = MOMStringCollection::create(mom_programme_content_advisory(itemAsProgramme()));
    return m_contentAdvisory.get();
}

int MOMProgramme::episode() const
{
    return mom_programme_episode(itemAsProgramme());
}

int MOMProgramme::totalEpisodes() const
{
    return mom_programme_total_episodes(itemAsProgramme());
}

int MOMProgramme::showType() const
{
    return mom_programme_show_type(itemAsProgramme());
}

bool MOMProgramme::isHD() const
{
    return (mom_programme_video_type(itemAsProgramme()) & MDA_VIDEO_TYPE_HD) != 0;
}

bool MOMProgramme::isWidescreen() const
{
    return (mom_programme_video_type(itemAsProgramme()) & (MDA_VIDEO_TYPE_HD|MDA_VIDEO_TYPE_SD_WIDESCREEN)) != 0;
}

int MOMProgramme::audioType() const
{
    return mom_programme_audio_type(itemAsProgramme());
}

MOMStringCollection* MOMProgramme::genre() const
{
    if (!m_genre)
        m_genre = MOMStringCollection::create(mom_programme_get_genres(itemAsProgramme()));
    return m_genre.get();
}

int MOMProgramme::purchased() const
{
    return mom_programme_extension_get_purchased(itemAsProgramme());
}

PassRefPtr<MOMContentItem> MOMProgramme::item( int index )
{
    mom_content_item rec = mom_programme_get_recording(itemAsProgramme(), index);
    return rec ? MOMContentItem::create(scriptExecutionContext(), rec) : nullptr;
}

void MOMProgramme::setBookmark( int bookmark )
{
    mom_programme_set_bookmark(itemAsProgramme(), bookmark);
}

} // namespace WebCore

#endif // ENABLE(RMAI)
