/** @file MOMRecorder.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "MOMRecorder.h"
#include "MOMDiscInfo.h"
#include "MOMStream.h"
#include "MOMStringCollection.h"

#include "MOMException.h"
#include "MOMEvent.h"
#include "UtilsMOM.h"

extern "C" {
#include "layers.h"
}

namespace WebCore {

PassRefPtr<MOMRecorder> MOMRecorder::create(ScriptExecutionContext* context, mom_driver driver)
{
    return adoptRef(new MOMRecorder(context, driver));
}

mom_error MOMRecorder::eventDispatcher(mom_event_t which, void* target, mom_event_params* params, unsigned int queue)
{
    return MOMEvent::dispatchMOMEvent(which, static_cast<MOMRecorder*>(target), params, queue);
}

MOMRecorder::MOMRecorder(ScriptExecutionContext* context, mom_driver driver)
    : MOMDevice(context, driver)
{
    if (mom_recorder_create( driver, &m_recorder ) == ERR_NONE)
    {
        mom_recorder_set_events(m_recorder, this, MOMRecorder::eventDispatcher);
    }
}

MOMRecorder::~MOMRecorder()
{
    mom_recorder_set_events(m_recorder, NULL, 0);
    mom_recorder_dispose(m_recorder);
}
    
PassRefPtr<MOMStringCollection> MOMRecorder::recordings() const
{
    return MOMStringCollection::create(mom_driver_urilist(driver()));
}

bool MOMRecorder::isRemovable() const
{
    return mom_recorder_is_removable(m_recorder);
}

MOMDiscInfo* MOMRecorder::discInfo() const
{
    if (!m_discInfo)
        m_discInfo = MOMDiscInfo::create(m_recorder);
    return m_discInfo.get();
}

int MOMRecorder::tsbMaxLength() const
{
    return mom_recorder_get_tsb_maxlength(m_recorder);
}

void MOMRecorder::setTsbMaxLength(int tsbMaxLength, ExceptionCode& ec)
{
    if (tsbMaxLength < 0)
    {
        MOMException::setFromMOMError(mom_err_INVALID, ec);
        return;
    }
    mom_recorder_set_tsb_maxlength(m_recorder, tsbMaxLength);
}

String MOMRecorder::boxName() const
{
    return UtilsMOM::StringFromLayersString(mom_recorder_get_box_name(m_recorder));
}

String MOMRecorder::boxId() const
{
    return UtilsMOM::StringFromLayersString(mom_recorder_get_box_id(m_recorder));
}

String MOMRecorder::record( MOMStream* source, ExceptionCode& ec )
{
    layers_string pvr_uri;
    mom_error err = mom_recorder_record(m_recorder, source->stream(), TRUE, &pvr_uri );
    if (err != ERR_NONE)
    {
        MOMException::setFromMOMError(err, ec);
        return String();
    }
    String str = UtilsMOM::StringFromLayersString(pvr_uri);
    layers_string_dispose(pvr_uri);
    return str;
}

void MOMRecorder::close( MOMStream* stream )
{
    /* close the recorder if recording or the stream if not (ie we
     * sourced the stream) */
    if (stream->stream() == mom_recorder_get_stream(m_recorder))
        mom_recorder_disconnect(m_recorder);
    else
        mom_stream_dispose(stream->stream());
}

void MOMRecorder::erase( const String & uri, ExceptionCode& ec )
{
    layers_string luri = UtilsMOM::LayersStringFromString(uri);
    mom_error err = mom_recorder_erase(m_recorder, luri);
    layers_string_dispose(luri);
    MOMException::setFromMOMError(err, ec);
}

void MOMRecorder::beginDiscOperation( int operation, ExceptionCode& ec )
{
    mom_error err = mom_recorder_begin_disc_operation(m_recorder, operation);
    MOMException::setFromMOMError(err, ec);
}

void MOMRecorder::cancelDiscOperation(ExceptionCode& ec)
{
    mom_error err = mom_recorder_cancel_disc_operation(m_recorder);
    MOMException::setFromMOMError(err, ec);
}

} // namespace WebCore

#endif // ENABLE(RMAI)
