/** @file PluginViewMOM.cpp
 *
 * Copyright (C) Espial Limited 2014 - Company confidential - All rights reserved.
 *
 */

#include "config.h"

#if ENABLE(RMAI)

#include "GraphicsContext.h"
#include "HTMLPlugInElement.h"
#include "Frame.h"
#include "FrameView.h"
#include "RenderBlock.h"

#include "MOMPlayer.h"

#include "PluginViewMOM.h"

extern "C" {
#include "cdf/mom_localsystem.h"
}

namespace WebCore
{

static Vector<PluginViewMOM*> gPluginInstances;

static void plugin_player_events( mom_player player, mom_player_event_t event, void *handle )
{
    PluginViewMOM *pluginView = static_cast<PluginViewMOM*>(handle);

    switch ( event )
    {
    case mom_player_event_LOCKED:
    case mom_player_event_UNLOCKED:
    case mom_player_event_DISCONNECT:
    case mom_player_event_CONNECT:
        /* just cause redraw - the redraw code reads the lock
         * status */
        pluginView->invalidate();
        break;

    case mom_player_event_DESTROYED:
        /* shouldn't happen as plugin owns the player */
        break;

    default:
        break;
    }
}

PluginViewMOM::PluginViewMOM(Frame*, const IntSize&, Element* element, const KURL&, const Vector<String>& paramNames, const Vector<String>& paramValues, const String&, bool)
    : Widget(),
      m_element(element),
      m_name(),
      m_windowRect()
{
	for(unsigned int i = 0; i < paramNames.size(); ++i)
    {
		if(equalIgnoringCase(paramNames[i], "playerID"))
        {
			m_name = paramValues[i];
            break;
		}
	}

    /* see if there is already a master plugin for this name */
    PluginViewMOM* master = NULL;
    if (!m_name.isEmpty())
    {
        for (Vector<PluginViewMOM*>::iterator it = gPluginInstances.begin(); it != gPluginInstances.end(); ++it)
        {
            if (m_name == (*it)->name() && (*it)->master())
            {
                master = (*it);
                break;
            }
        }
    }

    /* either copy the MOMPlayer or create a new one */
    if (master)
    {
        m_player = master->player();
        m_master = false;
        m_active = false;
    }
    else
    {
        m_player = MOMPlayer::create(element->scriptExecutionContext());
        m_master = true;
        m_active = true;
    }

    mom_player_register_callback(m_player->player(), plugin_player_events, this);

    gPluginInstances.append(this);
}

PluginViewMOM::~PluginViewMOM()
{
    if (m_player)
        mom_player_unregister_callback(m_player->player(), plugin_player_events, this);

    gPluginInstances.remove(gPluginInstances.find(this));
}

static void pluginview_commit_changes(layers_monotime_t, void*)
{
    mom_mda_driver_commit(mom_localsystem_get_sys_driver(), NULL, NULL);
}

void PluginViewMOM::updateMOMPlayer(bool force)
{
    if (!parent())
        return;

    if (m_player && m_active)
    {
        ASSERT(parent()->isFrameView());
        FrameView* frameView = static_cast<FrameView*>(parent());

        // get our location on the screen relative to the main browser window
        // 0,0 top left, +ve downwards
        IntRect windowRect = IntRect(frameView->contentsToWindow(frameRect().location()), frameRect().size());

        if (force || windowRect != m_windowRect)
        {
            //printf("PluginViewMOM::updateMOMPlayer: %p force %d windowRect %d,%d %d,%d\n", this, force, windowRect.x(), windowRect.y(), windowRect.right(), windowRect.bottom());
            layers_box box;
            box.x0 = windowRect.x();
            box.y0 = windowRect.y();
            box.x1 = windowRect.x() + windowRect.width();
            box.y1 = windowRect.y() + windowRect.height();
            mom_renderer_set_position(mom_player_renderer(m_player->player()), &box);

            m_windowRect = windowRect;
            layers_alarm_reschedule_after( 0, pluginview_commit_changes, NULL );
        }
    }
}

void PluginViewMOM::setFrameRect(const IntRect& r)
{
    //printf("PluginViewMOM::setFrameRect %p\n", this);
    if (r != frameRect())
        Widget::setFrameRect(r);
    updateMOMPlayer(false);
}

// this is called when the page or container scrolls and so the
// absolute rect changes
void PluginViewMOM::frameRectsChanged()
{
    //printf("PluginViewMOM::frameRectsChanged %p\n", this);
    updateMOMPlayer(false);
}

void PluginViewMOM::invalidateRect(const IntRect& r)
{
    if (!isSelfVisible())
        return;

    //printf("PluginViewMOM::invalidateRect: %p r %d,%d %d,%d renderer %p\n", this, r.x(), r.y(), r.right(), r.bottom(), m_element->renderer());
    if (m_element->renderer()){
		toRenderBox(m_element->renderer())->repaintRectangle(r);
    }
}

void PluginViewMOM::paint(GraphicsContext* ctx, const IntRect& r)
{
    if (!isSelfVisible())
        return;

    if (auto* renderer = m_element->renderer()) {
        if (renderer->opacity() == 0.0)
            return;
    }

    const IntRect widgetRect(frameRect());

	if(r.intersects(widgetRect))
    {
        //printf("PluginViewMOM::paint: %p r %d,%d %d,%d\n", this, r.x(), r.y(), r.right(), r.bottom());

        layers_colour colour = mom_player_get_bg_colour(m_player->player());
        if (colour == layers_colour_TRANSPARENT)
        {
            ctx->clearRect(intersection(widgetRect, r));
        }
        else
        {
            Color c = Color(layers_colour_REDNESS(colour),
                            layers_colour_GREENNESS(colour),
                            layers_colour_BLUENESS(colour));
            ctx->setFillColor(c);
            ctx->fillRect(intersection(widgetRect, r));
        }
    }
}

void PluginViewMOM::setPlayer(PassRefPtr<MOMPlayer> player)
{
    //printf("PluginViewMOM::setPlayer: %p was %p now %p\n", this, m_player.get(), player.get());
    if (player != m_player)
    {
        if ( m_player )
            mom_player_unregister_callback( m_player->player(), plugin_player_events, this);

        if ( player )
            mom_player_register_callback( player->player(), plugin_player_events, this);
        
        m_player = player;
    }
}

void PluginViewMOM::setActive(bool active)
{
    //printf("PluginViewMOM::setActive: %p active %d\n", this, active);
    if ( active != m_active )
    {
        if ( active )
        {
            /* deactivate any other plugins that use the same player */
            for (Vector<PluginViewMOM*>::iterator it = gPluginInstances.begin(); it != gPluginInstances.end(); ++it)
            {
                if ((*it)->player() == player())
                {
                    (*it)->deactivate();
                }
            }

            /* activate us */
            m_active = true;
            updateMOMPlayer(true);
            invalidate();
        }
        else
        {
            /* reactivate the master plugin */
            for (Vector<PluginViewMOM*>::iterator it = gPluginInstances.begin(); it != gPluginInstances.end(); ++it)
            {
                if ((*it)->player() == player() && (*it)->master())
                {
                    (*it)->setActive(true);
                    break;
                }
            }
        }
    }
}

void PluginViewMOM::show()
{
    setSelfVisible(true);
    Widget::show();
}

void PluginViewMOM::hide()
{
    setSelfVisible(false);
    Widget::hide();
}

} // namespace WebEspial

#endif
