/*
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "Storage.h"

#include "Frame.h"
#include "Page.h"
#include "Settings.h"
#include "StorageMap.h"
#include "PlatformString.h"
#include <wtf/PassRefPtr.h>
#include <wtf/HashSet.h>
#include "StorageNamespaceImpl.h"
#include "FileSystem.h"
#include <wtf/NonCopyingSort.h>
#include <utility>

namespace WebCore {

DLL_PUBLIC Storage::LRUMap& Storage::totalLRUSet()
{
    DEFINE_STATIC_LOCAL(LRUMap, totalLRUSet, ());
    return totalLRUSet;    
}

// Remove item from total storage
void Storage::totalLRUSetDelete() const
{
    if (! trusted())
    {
        ListHashSet<StorageArea*>::iterator i = totalLRUSet().first.find(const_cast<StorageArea*>(m_storageArea.get()));

        if (i != totalLRUSet().first.end())
            totalLRUSet().first.remove(i);
    }
    else
    {
        ListHashSet<StorageArea*>::iterator i = totalLRUSet().second.find(const_cast<StorageArea*>(m_storageArea.get()));

        if (i != totalLRUSet().second.end())
            totalLRUSet().second.remove(i);
    }
}

// Put storage at the top of the list
void Storage::totalLRUSetTop() const
{
    totalLRUSetDelete();

    if (! trusted())
        totalLRUSet().first.insertBefore(totalLRUSet().first.end(), const_cast<StorageArea*>(m_storageArea.get()));
    else
        totalLRUSet().second.insertBefore(totalLRUSet().second.end(), const_cast<StorageArea*>(m_storageArea.get()));
}

// Trim total storage
bool Storage::totalLRUSetTrim() const
{
    if (! trusted())
    {
        for (ListHashSet<StorageArea*>::iterator i = totalLRUSet().first.begin(); i != totalLRUSet().first.end(); ++ i)
            if (*i && *i != m_storageArea.get() && (*i)->used())
			{
				ExceptionCode ec;
				(*i)->clear(ec, frame());
				return true;
			}
                    
        return false;
    }
    else
    {
        for (ListHashSet<StorageArea*>::iterator i = totalLRUSet().second.begin(); i != totalLRUSet().second.end(); ++ i)
            if (*i && *i != m_storageArea.get() && (*i)->used())
			{
				ExceptionCode ec;
				(*i)->clear(ec, frame());
				return true;
			}
                    
        return false;
    }
}

unsigned Storage::totalUsed() const
{
    if (! trusted())
    {
        unsigned n = 0;
        
        for (ListHashSet<StorageArea*>::iterator i = totalLRUSet().first.begin(); i != totalLRUSet().first.end(); ++ i)
            if (*i)
                n += (*i)->used();

        return n;
    }
    else
    {
        unsigned n = 0;
        
        for (ListHashSet<StorageArea*>::iterator i = totalLRUSet().second.begin(); i != totalLRUSet().second.end(); ++ i)
            if (*i)
                n += (*i)->used();

        return n;
    }
}

PassRefPtr<Storage> Storage::create(Frame* frame, PassRefPtr<StorageArea> storageArea)
{
    return adoptRef(new Storage(frame, storageArea));
}

Storage::Storage(Frame* frame, PassRefPtr<StorageArea> storageArea)
    : DOMWindowProperty(frame)
    , m_storageArea(storageArea)
{
    ASSERT(m_frame);
    ASSERT(m_storageArea);
    if (m_storageArea)
        m_storageArea->incrementAccessCount();

	if (type() == LocalStorage)
		totalLRUSetTop();
}

Storage::~Storage()
{
    if (m_storageArea)
        m_storageArea->decrementAccessCount();
}

unsigned Storage::used() const
{
    return m_storageArea->used();
}

unsigned Storage::total() const
{
    return m_storageArea->total();
}

bool Storage::trusted() const
{
    return m_storageArea->trusted();
}

StorageType Storage::type() const
{
    return m_storageArea->type();
}

String Storage::url() const
{
    if (!m_frame)
        return String();

    return m_storageArea->url();
}

String Storage::key(unsigned index, ExceptionCode& ec) const
{
	if (type() == LocalStorage)
		totalLRUSetTop();

	return m_storageArea->key(index, ec, m_frame);
}

String Storage::getItem(const String& key, ExceptionCode& ec) const
{
	if (type() == LocalStorage)
		totalLRUSetTop();

    return m_storageArea->getItem(key, ec, m_frame);
}

void Storage::setItem(const String& key, const String& value, ExceptionCode& ec)
{
    ec = 0;
    if (!m_frame)
        return;

	if (type() == LocalStorage) {
		totalLRUSetTop();
    
		if (total() == 0)
		{
			ec = QUOTA_EXCEEDED_ERR;
			return;
		}
	}

    m_storageArea->setItem(key, value, ec, m_frame);

    if (type() == LocalStorage){
    	if (! trusted())
    		clearIfNeeded(totalLRUSet().first, StorageNamespaceImpl::localStorageNamespaceMap().first, total());
    	else
    		clearIfNeeded(totalLRUSet().second, StorageNamespaceImpl::localStorageNamespaceMap().second, total());
    }
}

void Storage::removeItem(const String& key, ExceptionCode& ec)
{
    if (!m_frame)
        return;

	if (type() == LocalStorage)
		totalLRUSetTop();

    m_storageArea->removeItem(key, ec, m_frame);
}

// Clears everything from total storage
DLL_PUBLIC void Storage::clearAll()
{
	clearAll(totalLRUSet().first, StorageNamespaceImpl::localStorageNamespaceMap().first);
	clearAll(totalLRUSet().second, StorageNamespaceImpl::localStorageNamespaceMap().second);
}

void Storage::clearAll(const LRUList& itLRU, const StorageNamespaceImpl::LocalStorageNamespace& itNamespace)
{
	HashSet<String> used;

	for (LRUList::const_iterator i = itLRU.begin(); i != itLRU.end(); ++ i)
		if (*i)
		{
			String filename((*i)->fullDatabaseFilename());
			if(filename.length() > 0)
				used.add(filename);
			ExceptionCode ec;
			(*i)->clear(ec);
		}

	const String& path = itNamespace.first;

	Vector<String> files = listDirectory(path, "*.localstorage");
	for (Vector<String>::iterator file = files.begin(); file != files.end(); ++ file)
		if (used.find(*file) == used.end())
			deleteFile(*file);
}

struct Predicate
{
       bool operator () (const pair<time_t, String>& o1, const pair<time_t, String>& o2)
       {
               return o1.first < o2.first;
       }
};

// Clears 10% off the total storage
DLL_PUBLIC void Storage::clearIfNeeded(const LRUList& itLRU, const StorageNamespaceImpl::LocalStorageNamespace& itNamespace, unsigned totalQuota)
{
	if (totalQuota == StorageMap::noQuota)
		return;

	HashSet<String> used;
	for (LRUList::const_iterator i = itLRU.begin(); i != itLRU.end(); ++ i){
		if (*i){
			String filename((*i)->fullDatabaseFilename());
			if(filename.length() > 0)
				used.add(filename);
		}
	}

	const String& path = itNamespace.first;

	// Get the files regarding the associated section
	Vector<String> files = listDirectory(path, "*.localstorage");
	time_t time;
	Vector< pair<time_t, String> > unused;
	unused.reserveCapacity(files.size());

	for (int i = 0; i < files.size(); ++i){
		const String file = files[i];

		if(file.length() == 0)
			continue;

		if (used.find(file) == used.end())
			if (getFileModificationTime(file, time))
				unused.append(std::make_pair(time, file));
	}

	nonCopyingSort(unused.begin(), unused.end(), Predicate());

	// Get total usage
	long long size;
	long long totalUsed = 0;
	for (Vector< pair<time_t, String> >::iterator file = unused.begin(); file != unused.end(); ++ file)
		if (getFileSize(file->second, size))
			totalUsed += size;

	for (LRUList::const_iterator i = itLRU.begin(); i != itLRU.end(); ++ i)
		if (*i && (*i)->used())
			totalUsed += (*i)->used() * sizeof(UChar);

	// Dangling files are always the least recently used so delete them first
	for (Vector< pair<time_t, String> >::iterator file = unused.begin(); file != unused.end(); ++ file)
		if (!totalQuota || totalUsed > totalQuota)
		{
			if (getFileSize(file->second, size))
				totalUsed -= size;

			deleteFile(file->second);
		}

	for (LRUList::const_iterator it = itLRU.begin(); it != itLRU.end(); ++ it)
		if (*it && (*it)->used()){
			if (!totalQuota || totalUsed > totalQuota)
			{
				ExceptionCode ec;
				totalUsed -= (*it)->used() * sizeof(UChar);
				(*it)->clear(ec);
			}
		}
}

// Clears item from total storage
DLL_PUBLIC void Storage::clearDomain(const String& url)
{
	for (ListHashSet<StorageArea*>::iterator it = totalLRUSet().first.begin(); it != totalLRUSet().first.end(); ++ it)
		if (*it)
			if ((*it)->url() == url)
			{
				ExceptionCode ec;
				(*it)->clear(ec);
				return;
			}
				
	for (ListHashSet<StorageArea*>::iterator it = totalLRUSet().second.begin(); it != totalLRUSet().second.end(); ++ it)
		if (*it)
			if ((*it)->url() == url)
			{
				ExceptionCode ec;
				(*it)->clear(ec);
				return;
			}
}

bool Storage:: contains(const String& key, ExceptionCode& ec) const
{
	if (type() == LocalStorage)
		totalLRUSetTop();

	return m_storageArea->contains(key, ec, m_frame);
}

}
