/*
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "StorageSyncManager.h"

#include "EventNames.h"
#include "FileSystem.h"
#include "Frame.h"
#include "FrameTree.h"
#include "StorageTask.h"
#include "StorageThread.h"
#include "Page.h"
#include "PageGroup.h"
#include "StorageAreaSync.h"
#include <wtf/MainThread.h>
#include <wtf/StdLibExtras.h>
#include <wtf/text/CString.h>

namespace WebCore {

PassRefPtr<StorageSyncManager> StorageSyncManager::create(const String& path)
{
    return adoptRef(new StorageSyncManager(path));
}

StorageSyncManager::StorageSyncManager(const String& path)
: m_path(path.isolatedCopy())
, m_syncTimer(this, &StorageSyncManager::syncTimerFired)
, m_closed(false)
{
    ASSERT(isMainThread());
    ASSERT(!m_path.isEmpty());
}

StorageSyncManager::~StorageSyncManager()
{
    ASSERT(isMainThread());
}

// Called on a background thread.
String StorageSyncManager::fullDatabaseFilename(const String& databaseIdentifier)
{
    if (!makeAllDirectories(m_path)) {
        LOG_ERROR("Unabled to create LocalStorage database path %s", m_path.utf8().data());
        return String();
    }

    return pathByAppendingComponent(m_path, databaseIdentifier + ".localstorage");
}

void StorageSyncManager::close()
{
    ASSERT(isMainThread());

	while(m_queue.size()){
		runOneTask();
	}

    m_syncTimer.stop();
    m_closed = true;
}

bool StorageSyncManager::scheduleImport(PassRefPtr<StorageAreaSync> area)
{
	if(m_closed)
		return false;

    ASSERT(isMainThread());
    m_queue.append(StorageTask::createImport(area.get()));

    if(!m_syncTimer.isActive())
    	m_syncTimer.startOneShot(0);

    return true;
}

void StorageSyncManager::scheduleSync(PassRefPtr<StorageAreaSync> area)
{
	if(m_closed)
		return;

    ASSERT(isMainThread());
    m_queue.append(StorageTask::createSync(area.get()));

    if(!m_syncTimer.isActive())
    	m_syncTimer.startOneShot(0);
}

void StorageSyncManager::scheduleDeleteEmptyDatabase(PassRefPtr<StorageAreaSync> area)
{
	if(m_closed)
		return;

    ASSERT(isMainThread());
    m_queue.append(StorageTask::createDeleteEmptyDatabase(area.get()));

    if(!m_syncTimer.isActive())
    	m_syncTimer.startOneShot(0);
}

void StorageSyncManager::runOneTask()
{
	if(m_queue.size() == 0)
		return;

	OwnPtr<StorageTask> task = m_queue[0].release();

	m_queue.remove(0);

	task->performTask();
}

void StorageSyncManager::syncTimerFired(Timer<StorageSyncManager>*)
{
	if(m_closed)
		return;

	runOneTask();

    if(m_queue.size() == 0)
    	return;

    if(m_syncTimer.isActive())
    	return;

    m_syncTimer.startOneShot(0);
}

} // namespace WebCore
